/*
	File: AfhFMmake.c
	C source file for the AFH FilterMeister standalone filter "maker".

    Copyright 2000-2014 AFH Systems.

  FilterMeister is free software. You can redistribute and modify it
  under the terms of the MPL2 License.
  See: https://www.mozilla.org/en-US/MPL/2.0/
  Any changes to the FilterMeister files have to be published or sent
  to harry@thepluginsite.com. The filter plugin code itself does not
  need to be disclosed.

    16-Jan-2000  1. Extracted from AfhFMUI.c.
	07-Oct-2013	 1. Fix up for X64.
	30-Oct-2013  1. Fix bug in use of X64GEN macro.
	19-Dec-2013  1. Separate FMCODEs for x32 and x64.
	27-Dec-2013	 1. Use FMTEMPLATE 164 when generating
					X64 code.
	12-Jan-2014	 1. Add PIMaxSize32Property to PiPL to allow pixel
					dimensions > 30,000.
	28-Jan-2014	 1. Add the target platform to the completion msgBox.
	30-Jan-2014	 1. Save and restore first word of filterFilenameText,
					which gets clobbered by GetSaveFileName() when
					file name is too long.

*/

#ifndef STANDALONE

#include <windows.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <stddef.h>     /* offsetof */
#include <stdio.h>
#include <assert.h>

#include "AfhFM.h"
#include "AfhFMcontext.h"

#include "resource.h"

#ifdef SCRIPTABLE
    #include "AfhFMScripting.h"
#endif

#define FLAG(id) (1 << ((id) ^ 7))
#define BIT(id)  (1 << (id))

  // Some macros for switching between the NT UpdateResource API
  // and our own homebrew UpdateResource API for Win 95/98

#ifdef USE_NT_UPDATERESOURCE
#define UR_BeginUpdateResource BeginUpdateResource
#define UR_UpdateResource UpdateResource
#define UR_EndUpdateResource EndUpdateResource
#else
#define UR_BeginUpdateResource afhBeginUpdateResource
#define UR_UpdateResource afhUpdateResource
#define UR_EndUpdateResource afhEndUpdateResource
#endif

/*****************************************************************************/

#ifdef FULL

static void SysErrorMessage(LPCTSTR prefix) {
    LPVOID lpMsgBuf;
    DWORD msgid = GetLastError();

    FormatMessage( 
        FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
        NULL,	// pointer to  message source 
        msgid,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
        (LPTSTR) &lpMsgBuf,
        0,	// minimum size of message buffer 
        NULL  	// address of array of message inserts 
    );

    ErrorOk("%s\n%s", prefix, lpMsgBuf);

    if (LocalFree(lpMsgBuf) != NULL) {
        ErrorOk("LocalFree failed in SysErrorMessage.");
    }
} //SysErrorMessage

#endif //ifdef FULL
 
/*****************************************************************************/

#ifdef FULL

void *FmLoadFile(const char *szFileName, int *pcbSize)
{
    FILE    *file;
    int     iLength;
    const char  *pPath;
    void    *p;

    /* Search for file in PATH and FM_PATH environment variables: */
    pPath = FmSearchFile(szFileName);

    if (!(file = fopen (pPath, "rb"))) {
        msgBox(MB_ICONERROR, "FilterMeister I/O Error",
             "Could not open file\n%s", pPath);
        return NULL;
    }

    // Determine file length...
    fseek (file, 0, SEEK_END);
    iLength = ftell (file);
    rewind (file);

    if (!(p = malloc (iLength)))
    {
        fclose (file);
        msgBox(MB_ICONERROR, "FilterMeister System Error",
           "Could not allocate %d bytes of memory.", iLength);
        return NULL;
    }

    *pcbSize = (int)fread (p, 1, iLength, file);
    fclose (file);
    if (*pcbSize != iLength) {
        //incorrect file size???
        free(p);
        return NULL;
    }

    return p;
} /*FmLoadFile*/


static int FmCodeSize(const CB *cb)
{
	if (X64GEN) {
#ifdef _WIN64
		//use real CB64
		return (sizeof(CB_SECTION_HEADER) + sizeof(CB64) + cb->pc + 3)
			   & ~3; //round up to multiple of 4 bytes
#else //Win32
		//use dummy CB64
		return (sizeof(CB_SECTION_HEADER) + sizeof(CB64) + (cb->pc - cb->pcStart) + 3)
			   & ~3; //round up to multiple of 4 bytes
#endif
	}
	else { //X32GEN
#ifdef _WIN64
		//use dummy CB32
		return (sizeof(CB_SECTION_HEADER) + sizeof(CB32) + cb->pc + 3)
			   & ~3; //round up to multiple of 4 bytes
#else //Win32
		//use real CB32
		return (sizeof(CB_SECTION_HEADER) + sizeof(CB32) + (cb->pc - cb->pcStart) + 3)
			   & ~3; //round up to multiple of 4 bytes
#endif
	}
} /*FmCodeSize*/


static char *AddFmCodeBytes(char *lp, const void *pBytes, int nBytes)
{
    memcpy(lp, pBytes, nBytes);
    return lp + nBytes;
} /*AddFmCodeBytes*/


static char *AddFmCode(char *lp, const CB *cb, CBSECTIONKIND iKind)
{
    CB_SECTION_HEADER header;

    assert(cb->pcStart != 0);	//make sure buffer allocated
	if (X64GEN) {
#ifdef _WIN64
	    header.nBytes = sizeof(CB64) + cb->pc;
#else //Win32
	    header.nBytes = sizeof(CB64) + (cb->pc - cb->pcStart);
#endif
	}
	else { //X32GEN
#ifdef _WIN64
	    header.nBytes = sizeof(CB32) + cb->pc;
#else //Win32
	    header.nBytes = sizeof(CB32) + (cb->pc - cb->pcStart);
#endif
	}
    if (header.nBytes < sizeof(CB32) + 0) {   //+2 would fail if no user functions.	//CB32 is temp kludge!!!
        //Implausible or uninitialized...
        ErrorOk("Bad code section byte count in AddFmCode.");
    }
    header.iKind = iKind;
    header.iEncrypt = 0;
    header.iKey = 0;
    header.iChecksum = 0;
    lp = AddFmCodeBytes(lp, &header, sizeof(header));
	if (X64GEN) {
	    lp = AddFmCodeBytes(lp, cb, sizeof(CB64));	//cb should be cb64??
	}
	else { //X32GEN
	    lp = AddFmCodeBytes(lp, cb, sizeof(CB32));	//cb should be cb32??
	}
#ifdef _WIN64
    lp = AddFmCodeBytes(lp, cb->pcStart, cb->pc);
#else //Win32
    lp = AddFmCodeBytes(lp, cb->pcStart, cb->pc - cb->pcStart);
#endif
    lp = AddFmCodeBytes(lp, "PAD", -header.nBytes & 3); //0-3 pad bytes.
    return lp;
} /*AddFmCode*/

#endif //ifdef FULL

/*****************************************************************************/

#ifdef FULL

static BOOL EmbedOtherFile(HANDLE hExe, const char *szFilename, int iResType)
{
    void *pData;
    int cbFileSize;
    char *pszUpperCase;
    char *pszResType;

    pData = FmLoadFile(szFilename, &cbFileSize);
    if (!pData) {
        UR_EndUpdateResource(hExe, TRUE);    //discard updates
        ErrorOk("Can't load embed file %s",
                szFilename);
        return FALSE;
    }
    if (cbFileSize < 10) {
        UR_EndUpdateResource(hExe, TRUE);    //discard updates
        ErrorOk("Not a credible file size (%d) for embed file\n%s",
                cbFileSize,
                szFilename);
        free(pData);
        return FALSE;
    }
    // Resource names must be uppercase -- otherwise,
    // FindResource() won't find them... (Note that
    // strdup() allocates a new string with malloc().)
    pszUpperCase = _strdup(szFilename);
    if (!pszUpperCase) {
        UR_EndUpdateResource(hExe, TRUE);    //discard updates
        free(pData);
        ErrorOk("strdup failed while embedding resources");
        return FALSE;
    }
    pszUpperCase = _strupr(pszUpperCase);
    //Write resource data...
    if (iResType == 'W') {
        pszResType = "FMMETAFILE";
    }
    else if (iResType == 'A') {
        pszResType = "WAVE";
    }
    else {
        pszResType = "FMDATA"; //for now
    }
    if (!UR_UpdateResource(
        hExe,	                // update-file handle 
        pszResType,	        // predefined resource type
        pszUpperCase,         // resource name
        MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_US),	// language identifier of resource 
        pData,	            // address of resource data 
        cbFileSize            // length of resource data, in bytes 
       ))
    {
        SysErrorMessage("UpdateResource failed while embedding resources");
        UR_EndUpdateResource(hExe, TRUE);    //discard updates
        free(pszUpperCase);
        free(pData);
        return FALSE;
    }
    free(pszUpperCase);
    free(pData);

    return TRUE;    //success!

} //EmbedOtherFile


static BOOL EmbedCursorFile(HANDLE hExe, const char *szFilename)
{
    void *pData;
    int cbFileSize;
    int cursorNumber = 104; //Base Cursor Number

    pData = FmLoadFile(szFilename, &cbFileSize);
    if (!pData) {
        UR_EndUpdateResource(hExe, TRUE);    //discard updates
        ErrorOk("Can't load embed file %s",
                szFilename);
        return FALSE;
    }
    if (cbFileSize < 10) {
        UR_EndUpdateResource(hExe, TRUE);    //discard updates
        ErrorOk("Not a credible file size (%d) for embed file\n%s",
                cbFileSize,
                szFilename);
        free(pData);
        return FALSE;
    }

    //Find next free cursor number in resource
    while (FindResource(hDllInstance,MAKEINTRESOURCE(cursorNumber),RT_CURSOR) != NULL) cursorNumber++;

    //Info ("cursorNumber: %d",cursorNumber);

    //Write resource data...
    if (!UR_UpdateResource(
        hExe,	                // update-file handle 
        RT_CURSOR,	        // predefined resource type
        MAKEINTRESOURCE(cursorNumber),  // resource name
        MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_US),	// language identifier of resource 
        pData,	            // address of resource data 
        cbFileSize            // length of resource data, in bytes 
       ))
    {
        SysErrorMessage("UpdateResource failed while embedding resources");
        UR_EndUpdateResource(hExe, TRUE);    //discard updates
        free(pData);
        return FALSE;
    }
    free(pData);

    return TRUE;    //success!

} //EmbedCursorFile



static BOOL EmbedBitmapFile(HANDLE hExe, const char *szFilename)
{
    PBITMAPFILEHEADER pbmfh;
    PBITMAPCOREHEADER pbmch;
    PBITMAPINFOHEADER pbmih;
    PBITMAPV4HEADER   pbmv4h;
    int cbFileSize;
    char *pszUpperCase;

    pbmfh = FmLoadFile(szFilename, &cbFileSize);
    if (!pbmfh) {
        UR_EndUpdateResource(hExe, TRUE);    //discard updates
        ErrorOk("Can't load bitmap file %s", szFilename);
        return FALSE;
    }
    if (cbFileSize < sizeof(BITMAPFILEHEADER) ||
        pbmfh->bfType != 'MB') {
        UR_EndUpdateResource(hExe, TRUE);    //discard updates
        ErrorOk("Not a valid DIB bitmap file: %s", szFilename);
        free(pbmfh);
        return FALSE;
    }
    pbmch = (PBITMAPCOREHEADER)(pbmfh + 1); //scaled ptr addition!!
    pbmih = (PBITMAPINFOHEADER)pbmch;
    pbmv4h = (PBITMAPV4HEADER)pbmch;
    if (pbmch->bcSize == sizeof(BITMAPCOREHEADER)) {
        //It's a Win 2.x Version 2 (16-bit) DIB...
#ifdef VERBOSE
        Info("Embedding a Windows Version 2 DIB from %s", szFilename);
#endif
    }
    else if (pbmih->biSize == sizeof(BITMAPINFOHEADER)) {
        //It's a Win 3.x Version 3 (32-bit) DIB...
#ifdef VERBOSE
        Info("Embedding a Windows Version 3 DIB from %s", szFilename);
#endif
    }
    else if (pbmv4h->bV4Size == sizeof(BITMAPV4HEADER)) {
        //It's a Win 95 Version 4 (32-bit) DIB...
#ifdef VERBOSE
        Info("Embedding a Windows 95 Version 4 DIB from %s", szFilename);
#endif
    }
    else {
        //Unrecognized DIB version????
        if (Warn("Unrecognized DIB version in file %s\n"
                 "version size = %x",
                 szFilename,
                 pbmch->bcSize) == IDCANCEL)
        {
            UR_EndUpdateResource(hExe, TRUE);    //discard updates
            free(pbmfh);
            return FALSE;
        }
    }
    // Resource names must be uppercase -- otherwise,
    // FindResource() won't find them... (Note that
    // strdup() allocates a new string with malloc().)
    pszUpperCase = _strdup(szFilename);
    if (!pszUpperCase) {
        UR_EndUpdateResource(hExe, TRUE);    //discard updates
		free(pbmfh);
        ErrorOk("strdup failed while embedding resources");
        return FALSE;
    }
    pszUpperCase = _strupr(pszUpperCase);
    //Write BITMAP resource data, stripped of file header...
    if (!UR_UpdateResource(
        hExe,	                // update-file handle 
        RT_BITMAP,	        // predefined resource type
        pszUpperCase,         // resource name
        MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_US),	// language identifier of resource 
        pbmch,	            // address of resource data 
        cbFileSize - sizeof(BITMAPFILEHEADER) // length of resource data, in bytes 
       ))
    {
		SysErrorMessage("UpdateResource failed while embedding resources");
        UR_EndUpdateResource(hExe, TRUE);    //discard updates
        free(pszUpperCase);
        free(pbmfh);
        return FALSE;
    }
    free(pszUpperCase);
    free(pbmfh);
    return TRUE;    //success!
} //EmbedBitmapFile

#endif //ifdef FULL

/*****************************************************************************/

#ifdef FULL

/* Following typedefs are from IconPro.htm. */

/* There exists one ICONDIRENTRY for each icon image in the     */
/* file, providing details about its location in the file, size */
/* and color depth. The ICONDIRENTRY structure is defined as:   */

typedef struct
{
    BYTE        bWidth;          // Width, in pixels, of the image
    BYTE        bHeight;         // Height, in pixels, of the image
    BYTE        bColorCount;     // Number of colors in image (0 if >=8bpp)
    BYTE        bReserved;       // Reserved ( must be 0)
    WORD        wPlanes;         // Color Planes (0 for icons, XHotSpot for cursors)
    WORD        wBitCount;       // Bits per pixel (0 for icons, YHotSpot for cursors)
    DWORD       dwBytesInRes;    // How many bytes in this resource?
    DWORD       dwImageOffset;   // Where in the file is this image?
} ICONDIRENTRY, *LPICONDIRENTRY;

/* The file begins with an icon directory.  The idCount member  */
/* indicates how many images are present in the icon resource.  */
/* The size of the idEntries array is determined by idCount.    */

// #pragmas are used here to insure that the structure's
// packing in memory matches the packing of the ICO file.
#pragma pack( push )
#pragma pack( 2 )
typedef struct
{
    WORD           idReserved;   // Reserved (must be 0)
    WORD           idType;       // Resource Type (1 for icons, 2 for cursors)
    WORD           idCount;      // How many images?
    ICONDIRENTRY   idEntries[1]; // An entry for each image (idCount of 'em)
} ICONDIR, *LPICONDIR;
#pragma pack( pop )

/* For each ICONDIRENTRY, the file contains an icon image.       */
/* The dwBytesInRes member indicates the size of the image data. */
/* This image data can be found dwImageOffset bytes from the     */
/* beginning of the file, and is stored in the following format: */

typedef struct
{
   BITMAPINFOHEADER   icHeader;      // DIB header
   RGBQUAD         icColors[1];   // Color table
   BYTE            icXOR[1];      // DIB bits for XOR mask
   BYTE            icAND[1];      // DIB bits for AND mask
} ICONIMAGE, *LPICONIMAGE;

/* The icHeader member has the form of a DIB BITMAPINFOHEADER.  */
/* Only the following members are used: biSize, biWidth,        */
/* biHeight, biPlanes, biBitCount, biSizeImage. All other       */
/* members must be 0. The biHeight member specifies the         */
/* combined height of the XOR and AND masks. The members of     */
/* icHeader define the contents and sizes of the other elements */
/* of the ICONIMAGE structure in the same way that the          */
/* BITMAPINFOHEADER structure defines a CF_DIB format DIB.      */


/* Analogous to the ICONDIR data in the ICO file is the         */
/* RT_GROUP_ICON resource. In fact, one RT_GROUP_ICON resource  */
/* is created for each ICO file bound to the EXE or DLL.        */
/* The RT_GROUP_ICON resource is simply a GRPICONDIR structure: */

/* The idCount member indicates how many images are present in  */
/* the icon resource. The size of the idEntries array is        */
/* determined by idCount. There exists one GRPICONDIRENTRY for  */
/* each icon image in the resource, providing details about its */
/* size and color depth.                                        */
/* The GRPICONDIRENTRY structure is defined as:                 */

// #pragmas are used here to insure that the structure's
// packing in memory matches the packing of the EXE or DLL.
#pragma pack( push )
#pragma pack( 2 )
typedef struct
{
   BYTE   bWidth;               // Width, in pixels, of the image
   BYTE   bHeight;              // Height, in pixels, of the image
   BYTE   bColorCount;          // Number of colors in image (0 if >=8bpp)
   BYTE   bReserved;            // Reserved
   WORD   wPlanes;              // Color Planes (obtained from BITMAPINFOHEADER.biPlanes)
   WORD   wBitCount;            // Bits per pixel (obtained from BITMAPINFOHEADER.biBitCount)
   DWORD  dwBytesInRes;         // how many bytes in this resource?
   WORD   nID;                  // the ID
} GRPICONDIRENTRY, *LPGRPICONDIRENTRY;
#pragma pack( pop )

/* The dwBytesInRes member indicates the total size of the      */
/* RT_ICON resource referenced by the nID member. nID is the    */
/* RT_ICON identifier that can be passed to FindResource(),     */
/* LoadResource() and LockResource() to obtain a pointer to the */
/* ICONIMAGE structure (defined above) for this image.          */

#pragma pack( push )
#pragma pack( 2 )
typedef struct 
{
   WORD            idReserved;   // Reserved (must be 0)
   WORD            idType;       // Resource type (1 for icons)
   WORD            idCount;      // How many images?
   GRPICONDIRENTRY   idEntries[1]; // The entries for each image
} GRPICONDIR, *LPGRPICONDIR;
#pragma pack( pop )

/*****************************************************************************
 From WinRes.txt:

 There are only 3 differences between an ICON in an .ICO file and .EXE file.
  1] Offset to the DIB
	In an ICO file, the offset to the DIB is the last
	item in the ICONDIRENTRY structure - dwImageOffset.
	In an EXE file, the offset is found in the Offset for
	each individual ICON, the rc_desc.wOffset field.
  2] The two reserved WORD fields BEFORE the dwBytesInRes field in the ICO
	file are filled in.
	wPlanes   - filled from the BITMAPINFOHEADER.biPlanes
	wBitCount - filled from the BITMAPINFOHEADER.biBitCount
	This allows an application to determine which ICON is
	best suited for the display without having to read the DIB.
  3] The dwImageOffset is changed from a DWORD in the ICO file
	to a WORD (nID) in the EXE file.  The offset value in the ICO
	file is changed to represent an ordinal	number that
	starts with 1 and ends with n for the total number of
	ICONs that are in the EXE file.  Be aware that one ICON
	group (value RT_GROUP_ICON), one icon with two different
	resolutions, are counted as two separate ICONs (value RT_ICON).
*****************************************************************************/

static int nID;

static BOOL EmbedIconFile(HANDLE hExe, const char *szFilename)
{
    LPICONDIR pID;
    LPICONDIRENTRY pIDE;
    LPICONIMAGE pImage;
    LPGRPICONDIR pGID;
    int cbGID;
    unsigned int cbFileSize;
    char *pszUpperCase;
    int i;

    pID = FmLoadFile(szFilename, &cbFileSize);
    if (!pID) {
        UR_EndUpdateResource(hExe, TRUE);    //discard updates
        ErrorOk("Can't load ICON file %s", szFilename);
        return FALSE;
    }
    if (cbFileSize < sizeof(*pID) ||
        pID->idReserved != 0 ||
        pID->idType != 1 ||
        pID->idCount < 1 ||
        pID->idCount > 255) {
        UR_EndUpdateResource(hExe, TRUE);    //discard updates
        ErrorOk("Not a valid ICON file: %s", szFilename);
        free(pID);
        return FALSE;
    }
    // Allocate memory to construct the Group Icon Directory...
    cbGID = sizeof(GRPICONDIR) + sizeof(GRPICONDIRENTRY)*(pID->idCount - 1);
    cbGID = (cbGID + 3) & ~3;   //round up to dword boundary
    pGID = calloc(cbGID, 1);
    if (!pGID) {
        //calloc failed
        ErrorOk("calloc failed in EmbedIconFile");
        free(pID);
        return FALSE;
    }
    // Fill in Group Icon Directory header.
    pGID->idReserved = 0;
    pGID->idType = 1;       // 1=icon, 2=cursor
    pGID->idCount = pID->idCount;

    // Iterate through all icon directory entries...
    pIDE = pID->idEntries;  //pointer to first icon directory entry

    for (i = 0;  i < pID->idCount;  i++, pIDE++) {
        // Process next icon directory entry in file.
        // First some consistency checks to make sure we don't point
        // beyond the end of the file in memory...
        if ((char *)pIDE - (char *)pID + sizeof(ICONDIRENTRY) > cbFileSize) {
            //file too small!!
            ErrorOk("Not a valid ICON file: %s", szFilename);
            free(pGID);
            free(pID);
            return FALSE;
        }
        if (pIDE->dwImageOffset + pIDE->dwBytesInRes > cbFileSize) {
            //file too small!!
            ErrorOk("Not a valid ICON file: %s", szFilename);
            free(pGID);
            free(pID);
            return FALSE;
        }
        pImage = (LPICONIMAGE)((char *)pID + pIDE->dwImageOffset); //pointer to icon image
        // Fill in some info in the group icon directory entry
        // from the BITMAPINFOHEADER for this icon.
        pGID->idEntries[i].wPlanes = pImage->icHeader.biPlanes;
        pGID->idEntries[i].wBitCount = pImage->icHeader.biBitCount;
        // Fill in rest of Group Icon Directory entry from file Icon Directory entry.
        pGID->idEntries[i].bWidth = pIDE->bWidth;
        pGID->idEntries[i].bHeight = pIDE->bHeight;
        pGID->idEntries[i].bColorCount = pIDE->bColorCount;
        pGID->idEntries[i].bReserved = 0;
        pGID->idEntries[i].dwBytesInRes = pIDE->dwBytesInRes;
        pGID->idEntries[i].nID = nID;       // Assign a new resource ID
        ///// Write the icon data to an RT_ICON resource ////
        if (!UR_UpdateResource(
            hExe,	                                        // update-file handle 
            RT_ICON,	                                    // predefined resource type
            MAKEINTRESOURCE(nID),                           // resource ID
            MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_US),   // language identifier of resource 
            pImage,	                                        // address of resource data 
            pIDE->dwBytesInRes                              // length of resource data, in bytes 
           ))
        {
            SysErrorMessage("UpdateResource failed while embedding an RT_ICON resource");
            UR_EndUpdateResource(hExe, TRUE);    //discard updates
            free(pGID);
            free(pID);
            return FALSE;
        }
        nID++;      //increment RT_ICON resource ID
    } //for i

    // Now write the RT_GROUP_ICON resource...

    // Resource names must be uppercase -- otherwise,
    // FindResource() won't find them... (Note that
    // strdup() allocates a new string with malloc().)
    pszUpperCase = _strdup(szFilename);
    if (!pszUpperCase) {
        UR_EndUpdateResource(hExe, TRUE);    //discard updates
        free(pGID);
        free(pID);
        ErrorOk("strdup failed while embedding resources");
        return FALSE;
    }
    pszUpperCase = _strupr(pszUpperCase);
    // Write RT_GROUP_ICON resource directory...
    if (!UR_UpdateResource(
        hExe,	                                        // update-file handle 
        RT_GROUP_ICON,	                                // predefined resource type
        pszUpperCase,                                   // resource name
        MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_US),   // language identifier of resource 
        pGID,	                                        // address of resource data 
        cbGID                                           // length of resource data, in bytes 
       ))
    {
        SysErrorMessage("UpdateResource failed while embedding an RT_GROUP_ICON resource");
        UR_EndUpdateResource(hExe, TRUE);    //discard updates
        free(pszUpperCase);
        free(pGID);
        free(pID);
        return FALSE;
    }
    free(pszUpperCase);
    free(pGID);
    free(pID);
    return TRUE;    //success!
} //EmbedIconFile

#endif FULL

/*****************************************************************************/

void FmInitPiPLContents(void)
{
    // Set default supported image modes...
    gParams->SupportedModes = FLAG(plugInModeGrayScale)    |
                              FLAG(plugInModeRGBColor)     |
                              FLAG(plugInModeCMYKColor)    |
                              FLAG(plugInModeHSLColor)     |
                              FLAG(plugInModeHSBColor)     |
                              FLAG(plugInModeMultichannel) |
                              FLAG(plugInModeDuotone)      |
#if 1
                              //include ALL modes...
                              FLAG(plugInModeBitmap)       |
                              FLAG(plugInModeIndexedColor) |
                              FLAG(plugInModeGray16)       |
                              FLAG(plugInModeRGB48)        |
							  //Added by Harald Heim, December 6, 2002
							  FLAG(plugInModeLab48)        |
							  FLAG(plugInModeCMYK64)       |
							  FLAG(plugInModeDeepMultichannel)|
							  FLAG(plugInModeDuotone16)    |
#endif
                              FLAG(plugInModeLabColor);

    // No Enable Info by default...
    gParams->EI_specified = FALSE;
    gParams->EI_start_index = 0;
    gParams->EI_count = 0;

    // Set default 'fici' contents...
    gParams->fici[filterCaseFlatImageNoSelection-1].inputHandling = filterDataHandlingNone;
    gParams->fici[filterCaseFlatImageNoSelection-1].outputHandling = filterDataHandlingNone;
    gParams->fici[filterCaseFlatImageNoSelection-1].flags1 = BIT(PIFilterDontCopyToDestinationBit);
    gParams->fici[filterCaseFlatImageNoSelection-1].flags2 = 0;    //reserved

    gParams->fici[filterCaseFlatImageWithSelection-1].inputHandling = filterDataHandlingNone;
    gParams->fici[filterCaseFlatImageWithSelection-1].outputHandling = filterDataHandlingNone;
    gParams->fici[filterCaseFlatImageWithSelection-1].flags1 = BIT(PIFilterDontCopyToDestinationBit);
    gParams->fici[filterCaseFlatImageWithSelection-1].flags2 = 0;    //reserved

    gParams->fici[filterCaseFloatingSelection-1].inputHandling = filterDataHandlingBackgroundZap;
    gParams->fici[filterCaseFloatingSelection-1].outputHandling = filterDataHandlingNone;
    gParams->fici[filterCaseFloatingSelection-1].flags1 = BIT(PIFilterDontCopyToDestinationBit);
    gParams->fici[filterCaseFloatingSelection-1].flags2 = 0;    //reserved

    gParams->fici[filterCaseEditableTransparencyNoSelection-1].inputHandling = filterDataHandlingBackgroundZap;
    gParams->fici[filterCaseEditableTransparencyNoSelection-1].outputHandling = filterDataHandlingNone;
    gParams->fici[filterCaseEditableTransparencyNoSelection-1].flags1 = BIT(PIFilterDontCopyToDestinationBit) |
                                                                        BIT(PIFilterWorksWithBlankDataBit);
    gParams->fici[filterCaseEditableTransparencyNoSelection-1].flags2 = 0;    //reserved

    gParams->fici[filterCaseEditableTransparencyWithSelection-1].inputHandling = filterDataHandlingBackgroundZap;
    gParams->fici[filterCaseEditableTransparencyWithSelection-1].outputHandling = filterDataHandlingNone;
    gParams->fici[filterCaseEditableTransparencyWithSelection-1].flags1 = BIT(PIFilterDontCopyToDestinationBit) |
                                                                          BIT(PIFilterWorksWithBlankDataBit);
    gParams->fici[filterCaseEditableTransparencyWithSelection-1].flags2 = 0;    //reserved

    gParams->fici[filterCaseProtectedTransparencyNoSelection-1].inputHandling = filterDataHandlingBackgroundZap;
    gParams->fici[filterCaseProtectedTransparencyNoSelection-1].outputHandling = filterDataHandlingNone;
    gParams->fici[filterCaseProtectedTransparencyNoSelection-1].flags1 = BIT(PIFilterDontCopyToDestinationBit);
    gParams->fici[filterCaseProtectedTransparencyNoSelection-1].flags2 = 0;    //reserved

    gParams->fici[filterCaseProtectedTransparencyWithSelection-1].inputHandling = filterDataHandlingBackgroundZap;
    gParams->fici[filterCaseProtectedTransparencyWithSelection-1].outputHandling = filterDataHandlingNone;
    gParams->fici[filterCaseProtectedTransparencyWithSelection-1].flags1 = BIT(PIFilterDontCopyToDestinationBit);
    gParams->fici[filterCaseProtectedTransparencyWithSelection-1].flags2 = 0;    //reserved
} //FmInitPiPLContents

/*****************************************************************************/


BOOL FmMakeStandalone(HWND hDlg)
{
#ifdef FULL
    int cbFmCode;
    char *pFmCode;
    char *pFmText;
    OSVERSIONINFO osvi;
    char *pszPlatform;
#endif

#ifdef BETA
    {
    Info("This beta release cannot make standalone filters.\n\n"
        "Upgrade soon to FilterMeister Lite or Pro!");
    return FALSE;
    }
#else
    // Check authorization
    // (Force re-authorization if SHIFT key is depressed.)
    if (!DoAuthorize(GetKeyState(VK_SHIFT) < 0)) return FALSE;

    {//scope

        memset(&osvi, 0, sizeof(OSVERSIONINFO));
        osvi.dwOSVersionInfoSize = sizeof (OSVERSIONINFO);
        GetVersionEx (&osvi);

        if (osvi.dwPlatformId == VER_PLATFORM_WIN32s)
            pszPlatform = "Win32s";
        else if (osvi.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS)
            pszPlatform = "Windows 95/98";
        else if (osvi.dwPlatformId == VER_PLATFORM_WIN32_NT)
            pszPlatform = "Windows NT";
        else
            pszPlatform = "Unknown Platform";

#ifdef USE_NT_UPDATERESOURCE
        //Check that platform = NT for now (the updateResource API
        //is just a stub on Win95).
        if (osvi.dwPlatformId != VER_PLATFORM_WIN32_NT) {
            Info("Microsoft %s %d.%d (Build %d %s)\n\n"
                 "Sorry!\n\n"
                 "Creating a standalone filter requires Windows NT.",
                 pszPlatform,
                 osvi.dwMajorVersion,
                 osvi.dwMinorVersion,
                 osvi.dwBuildNumber & 0xFFFF,
                 osvi.szCSDVersion);
            return FALSE;
        }
#endif //USE_NT_UPDATERESOURCE
    }//scope

    // If source needs to be recompiled, ask user and recompile
    // (and re-init) if requested...
    // (Keep this message in sync with the same message in
    // DoFilterRect().
    if (!gParams->fValidCompile) {	//Does ValidCompile imply Inited??
        char szMessage[512];
        // Get the message string...
        // (e.g., "The compiled code is not up-to-date.\n\n"
        //        "Would you like to compile it first?"
        LoadString(hDllInstance, IDS_RECOMPILE, szMessage, sizeof szMessage);
        switch (YesNo(szMessage))
        {
        case IDYES:
        case IDOK:
            // (re-)codegen and init before making the standalone filter...
            SendMessage(MyHDlg, WM_COMMAND, IDC_INIT, 0);  //We need Init to set some fmc fields.
            // cancel if compilation fails...
            if (!gParams->fValidCompile) return FALSE;
            break;  //proceed after successful re-compile
        case IDCANCEL:
        case IDABORT:
        case IDNO:
            return FALSE;
            break;
        default:
            break;  //proceed without re-compiling
        }//switch
    }//if

    // Another quick authorization check...
    if (dwAuthorized2 != RegCodeFingerPrint[2]) return FALSE;

    // Warn if USE_MMX, USE_CMOV or USE_SSE is in effect...
    {//scope
        int iReply;
        if ((gParams->flags & AFH_USE_MMX) & !X64GEN) {
            iReply = Warn(
                      "A standalone filter made with \"Use MMX\"\n"
                      "in effect may not work properly on all systems.\n\n"
					  "To turn off \"Use MMX\", go to Advanced Mode (>>>)\n"
                      "and deselect the \"Use MMX\" check box."
                      );
            if (iReply != IDYES && iReply != IDOK) {
                return FALSE; //abandon ship
            }
        }
        if (gParams->flags & AFH_USE_CMOV) {	//DOES X64GEN IMPLY CMOV AVAILABLE???
            iReply = Warn(
                      "A standalone filter made with \"Use CMOV\"\n"
                      "in effect may not work properly on all systems.\n\n"
					  "To turn off \"Use CMOV\", go to Advanced Mode (>>>)\n"
                      "and deselect the \"Use CMOV\" check box."
                      );
            if (iReply != IDYES && iReply != IDOK) {
                return FALSE; //abandon ship
            }
        }
        if ((gParams->flags & AFH_USE_SSE) && !X64GEN) {
            iReply = Warn(
                      "A standalone filter made with \"Use SSE\"\n"
                      "in effect may not work properly on all systems.\n\n"
					  "To turn off \"Use SSE\", go to Advanced Mode (>>>)\n"
                      "and deselect the \"Use SSE\" check box."
                      );
            if (iReply != IDYES && iReply != IDOK) {
                return FALSE; //abandon ship
            }
        }
    }//scope

    // Warn if Dialog:image= was specified, but there are no
    // embedded images...
    {//scope
        int iReply;
        //Later, could actually search embeds[] for the image file name???
        if (gParams->gDialogImage_save && gParams->gImagesEmbedded == 0) {
            iReply = Warn(
                "A dialog background image was specified,\n"
                "but no images have been embedded with an\n"
                "Embed: specifier.\n\n"
                "Is that OK?"
                );
            if (iReply != IDYES && iReply != IDOK) {
                return FALSE; //abandon ship
            }
        }
    }//scope

    // Warn if any calls to setCtlImage were generated (except for
    // the built-in logo2.bmp image), but there are no
    // embedded images...
    {//scope
        int iReply;
        if (gParams->gCtlImagesSet && gParams->gImagesEmbedded == 0) {
            iReply = Warn(
                "One or more control images have been specified,\n"
                "but no images have been embedded with an\n"
                "Embed: specifier.\n\n"
                "Is that OK?"
                );
            if (iReply != IDYES && iReply != IDOK) {
                return FALSE; //abandon ship
            }
        }
    }//scope

    // Warn if any calls to playWave() were generated, but there are no
    // embedded waves...
    {//scope
        int iReply;
        if (gParams->gWavesSpecified && gParams->gWavesEmbedded == 0) {
            iReply = Warn(
                "One or more Wave files have been specified,\n"
                "but no waves have been embedded with an\n"
                "Embed:Wave= specifier.\n\n"
                "Is that OK?"
                );
            if (iReply != IDYES && iReply != IDOK) {
                return FALSE; //abandon ship
            }
        }
    }//scope

    // Another quick authorization check...
    if (dwAuthorized3 != RegCodeFingerPrint[3]) return FALSE;

    //Create the FMCODE resource in memory....
    {//scope
        #define DEA 2   //default encryption algorithm
        CB_SECTION_HEADER header;
        char *pHeader;
        char *lp;

		// Sanity checks...
		assert( sizeof(CB32) == 84 );
		assert( sizeof(CB64) == 96 );

        //Compute the total size of the FMCODE resource...
        cbFmCode = 0;
        cbFmCode += FmCodeSize(&cbCodeBuffer);    //Channel code (ForEveryPixel)
        cbFmCode += FmCodeSize(&cbInitCode);      //init code
        cbFmCode += FmCodeSize(&cbDivCheck);      //Divide Check code
        cbFmCode += FmCodeSize(&cbZeroDivide);    //OnZeroDivide code
        cbFmCode += FmCodeSize(&cbDivideOverflow);//OnDivideOverflow code
        cbFmCode += FmCodeSize(&cbOnCtl);         //OnCtl code
        cbFmCode += FmCodeSize(&cbOnFilterStart); //OnFilterStart code
        cbFmCode += FmCodeSize(&cbForEveryTile);  //ForEveryTile code
        cbFmCode += FmCodeSize(&cbOnFilterEnd);   //OnFilterEnd code
        cbFmCode += FmCodeSize(&cbUserFunctions); //UserFunctions code
        //size of literal pool section...
        cbFmCode += (sizeof (CB_SECTION_HEADER) + sizeof(gFmcp->ndlits) +
                     (sizeof(gFmcp->dlitPool[0]) * gFmcp->ndlits) + 3)
                    & ~3; //round up to multiple of 4 bytes
        //size of fmc initial values section...
        cbFmCode += (sizeof (CB_SECTION_HEADER)
                     + offsetof(FMcontext, use_CMOV)		//DANGER WILL ROBINSON!!
                     - offsetof(FMcontext, need_iuv)
                     + sizeof(gFmcp->use_CMOV)
                     + 3)
                    & ~3; //round up to multiple of 4 bytes

        //Allocate memory for the FMCODE resource.
        if (!(pFmCode = malloc (cbFmCode)))
        {
            msgBox(MB_ICONERROR, "FilterMeister System Error",
                   "Could not allocate %d bytes of memory.", cbFmCode);
            return FALSE;
        }

        lp = pFmCode; //traveling pointer

        //add channel (ForEveryPixel) code
        pHeader = lp;
        lp = AddFmCode(lp, &cbCodeBuffer, CBK_FOREVERYPIXEL);
        EncryptFmCode(pHeader, DEA, RegCodeFingerPrint[0]);

        //add init code
        pHeader = lp;
        lp = AddFmCode(lp, &cbInitCode, CBK_INITCODE);
        EncryptFmCode(pHeader, DEA, RegCodeFingerPrint[1]);

        //add div check code
        pHeader = lp;
        lp = AddFmCode(lp, &cbDivCheck, CBK_DIVCHECK);
        EncryptFmCode(pHeader, DEA, RegCodeFingerPrint[2]);

        //add zero divide code
        pHeader = lp;
        lp = AddFmCode(lp, &cbZeroDivide, CBK_ONZERODIVIDE);
        EncryptFmCode(pHeader, DEA, RegCodeFingerPrint[3]);

        //add divide overflow code
        pHeader = lp;
        lp = AddFmCode(lp, &cbDivideOverflow, CBK_ONDIVIDEOVERFLOW);
        EncryptFmCode(pHeader, DEA, RegCodeFingerPrint[2]);

        //add OnCtl code
        pHeader = lp;
        lp = AddFmCode(lp, &cbOnCtl, CBK_ONCTL);
        EncryptFmCode(pHeader, DEA, RegCodeFingerPrint[1]);

        //add OnFilterStart code
        pHeader = lp;
        lp = AddFmCode(lp, &cbOnFilterStart, CBK_ONFILTERSTART);
        EncryptFmCode(pHeader, DEA, RegCodeFingerPrint[0]);

        //add ForEveryTile code
        pHeader = lp;
        lp = AddFmCode(lp, &cbForEveryTile, CBK_FOREVERYTILE);
        EncryptFmCode(pHeader, DEA, RegCodeFingerPrint[1]);

        //add OnFilterEnd code
        pHeader = lp;
        lp = AddFmCode(lp, &cbOnFilterEnd, CBK_ONFILTEREND);
        EncryptFmCode(pHeader, DEA, RegCodeFingerPrint[2]);

        //add UserFunctions code
        pHeader = lp;
        lp = AddFmCode(lp, &cbUserFunctions, CBK_USERFUNCTIONS);
        EncryptFmCode(pHeader, DEA, RegCodeFingerPrint[3]);

        //add the literal pool
        pHeader = lp;
        header.nBytes = sizeof(gFmcp->ndlits) +
                        (sizeof(gFmcp->dlitPool[0]) * gFmcp->ndlits);
        header.iKind = CBK_LITPOOL;
        header.iEncrypt = 0;
        header.iKey = 0;
        header.iChecksum = 0;
        lp = AddFmCodeBytes(lp, &header, sizeof(header));
        lp = AddFmCodeBytes(lp, &gFmcp->ndlits, sizeof(gFmcp->ndlits));
        lp = AddFmCodeBytes(lp, gFmcp->dlitPool, sizeof(gFmcp->dlitPool[0]) * gFmcp->ndlits);
        lp = AddFmCodeBytes(lp, "PAD", -header.nBytes & 3); //0-3 pad bytes.
        EncryptFmCode(pHeader, DEA, RegCodeFingerPrint[3]);

        //add the initialized flags from the context record
        pHeader = lp;
        header.nBytes = offsetof(FMcontext, use_CMOV) -
                        offsetof(FMcontext, need_iuv) +
                        sizeof(gFmcp->use_CMOV);
        header.iKind = CBK_INITFMC;
        header.iEncrypt = 0;
        header.iKey = 0;
        header.iChecksum = 0;
        lp = AddFmCodeBytes(lp, &header, sizeof(header));
        lp = AddFmCodeBytes(lp, &gFmcp->need_iuv, header.nBytes);
        lp = AddFmCodeBytes(lp, "PAD", -header.nBytes & 3); //0-3 pad bytes.
        EncryptFmCode(pHeader, DEA, RegCodeFingerPrint[2]);

#if 1
        //Consistency check...
        if (lp - pFmCode != cbFmCode) {
            msgBox(MB_ICONERROR, "FilterMeister Error",
                "Bad FMCODE byte count:\n"
                "Should be %d\n"
                "Actually is %d", cbFmCode, lp - pFmCode);
            free(pFmCode);
            return FALSE;
        }
#endif
        #undef DEA
    }//scope
    {//scope

        // Now create the standalone template file and
        // update its resources.
        HANDLE hExe;
#ifdef USE_NT_UPDATERESOURCE
        FILE *file;
#endif
        int cbPipl;
        char *pPipl;
        DWORD dwProperty;
        int cbPropertyLength;
        BYTE pstr[260];   //bytecount-prefixed string (with room for ellipsis)
        HRSRC hRes;         // handle/ptr. to res. info.
        HRSRC hResLoad;     // handle to loaded resource  
        char *pFmTemplate;  // pointer to FMTEMPLATE resource data
		int TemplateSubID;	// Resource subtype for FMTEMPLATE and FMCODE
        int cb;             // Size of resource (bytes)
		WORD wFFTsave;		// Saves first word of filterFilenameText.
        char *szTemplateFile = gFmcp->filterFilenameText;
        static OPENFILENAME sfn;
        static char szFilter[] = 
            "FilterMeister Standalone Filter (*.8bf)\0*.8bf\0";  //1

        // Find the FMTEMPLATE resource and use it to create
        // the standalone template file.

        // Locate the FMTEMPLATE 132 or 164 resource for X32GEN or X64GEN, resp.
		TemplateSubID = X64GEN ? 164 : 132;
        hRes = FindResource(hDllInstance, MAKEINTRESOURCE(TemplateSubID), "FMTEMPLATE");

        if (hRes == NULL) { 
            free(pFmCode);
            ErrorOk("Could not find FMTEMPLATE %d resource.", TemplateSubID);
            return FALSE;
        } 

        // Check for reasonable size of resource.
        if ((cb=SizeofResource(hDllInstance, hRes)) < 1000)
        {
            free(pFmCode);
            ErrorOk("FMTEMPLATE %d resource is too short: %d bytes", TemplateSubID, cb);
            return FALSE;
        }

        // Load the FMTEMPLATE resource into global memory. 
        hResLoad = LoadResource(hDllInstance, hRes); 

        if (hResLoad == NULL) { 
            free(pFmCode);
            ErrorOk("Could not load FMTEMPLATE %d resource.", TemplateSubID);
            return FALSE;
        } 

        // Lock the FMTEMPLATE resource into global memory. 
        pFmTemplate = LockResource(hResLoad); 

        if (pFmTemplate == NULL) { 
            free(pFmCode);
            ErrorOk("Could not lock FMTEMPLATE %d resource.", TemplateSubID);
            return FALSE;
        }

        // Create the standalone template file...

#if 0 //This logic was moved to Mdc_evaluate_sdl...
        if (gFmcp->filterFilenameText[0] == '\0') {
            // Supply a default filter module name...
            strcpy(gFmcp->filterFilenameText, "MyFilter.8bf");
        }
#endif

		// Save the first word of filterFilenameText for later.
		wFFTsave = *(WORD *)gFmcp->filterFilenameText;

        sfn.lStructSize       = sizeof (OPENFILENAME) ;
        sfn.hwndOwner         = hDlg ;
        sfn.hInstance         = NULL ;
        sfn.lpstrFilter       = szFilter ;
        sfn.lpstrCustomFilter = NULL ;
        sfn.nMaxCustFilter    = 0 ;
        sfn.nFilterIndex      = 0 ;
        sfn.nMaxFile          = sizeof(gFmcp->filterFilenameText) ;
        sfn.nMaxFileTitle     = _MAX_FNAME + _MAX_EXT ;
        sfn.lpstrInitialDir   = NULL ;
        sfn.lpstrTitle        = "Make Standalone Filter" ;
        sfn.nFileOffset       = 0 ;
        sfn.nFileExtension    = 0 ;
        sfn.lpstrDefExt       = "8bf" ;
        sfn.lCustData         = 0L ;
        sfn.lpfnHook          = NULL ;
        sfn.lpTemplateName    = NULL ;
        sfn.lpstrFile         = gFmcp->filterFilenameText ;
        sfn.lpstrFileTitle    = szTitleName ;
        sfn.Flags             = OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT ;

        if (!GetSaveFileName(&sfn)) {
            free(pFmCode);
            switch (CommDlgExtendedError()) {
            case FNERR_BUFFERTOOSMALL:
				//First word of filterFilenameText contains the
				//actual char count.
                msgBox(MB_ICONERROR, "FilterMeister I/O Error",
                       "File name is too long (%d characters)",
                       *(WORD *)gFmcp->filterFilenameText);
                //Now restore first word of filterFilenameText,
                //which was clobbered with the actual char count!
				*(WORD *)gFmcp->filterFilenameText = wFFTsave;
                break;
            case 0:   //User canceled the dialog.
            default:
                //Just fail silently...
                break;
            } //switch
            return FALSE;
        }

        // Update dialog title (re-interpolate)...
        fm_setDialogText((INT_PTR) gParams->gDialogTitle);

#ifdef USE_NT_UPDATERESOURCE
        // Create the standalone template file...
        if (!(file = fopen (szTemplateFile, "wb"))) {
            free(pFmCode);
            msgBox(MB_ICONERROR, "FilterMeister I/O Error",
                   "Could not create file\n%s", szTemplateFile);
            return FALSE;
        }

        // Write the template to the file...
        fwrite(pFmTemplate, 1, cb, file);

        // Close the template file...
        if (fclose(file)) {
            free(pFmCode);
            msgBox(MB_ICONERROR, "FilterMeister I/O Error",
                   "Could not close file\n%s", szTemplateFile);
            return FALSE;
        }
        else {
#ifdef VERBOSE
            Info("Wrote template to file\n%s", szTemplateFile);
#endif
        }
#endif //USE_NT_UPDATERESOURCE

        // Now update the resources in the template file...
        hExe = UR_BeginUpdateResource(
                szTemplateFile,
#ifdef USE_NT_UPDATERESOURCE
                FALSE	// do not delete existing resources
#else
                pFmTemplate   //address of RAM file template
#endif
            );

        if (hExe == NULL) { 
            free(pFmCode);
#if 0
            // This puts out a bogus message on NT if
            // the template file is not found (says
            // successful completion, although it failed!).
            SysErrorMessage("BeginUpdateResource failed:");
#else
            // So we put out our own message instead, and
            // simply assume the cause of failure is the
            // missing template file.  (On Win95, the failure
            // is due to the unsupported API, but we catch
            // that problem earlier now...)
            msgBox(MB_ICONERROR, "FilterMeister I/O Error",
                   "Could not open template file\n%s", szTemplateFile);
#endif
            return FALSE;
        } 

        //////// Update the FMCODE resource... ////////
        if (!UR_UpdateResource(
            hExe,	                        // update-file handle 
            "FMCODE",	                    // resource type
            MAKEINTRESOURCE(TemplateSubID),	// resource name (132 or 164)
            MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_US),	// language identifier of resource 
            pFmCode,	                    // address of resource data 
            cbFmCode 	                    // length of resource data, in bytes 
           )) {

            free(pFmCode);
            SysErrorMessage("UpdateResource (FMCODE) failed:");
            UR_EndUpdateResource(hExe, TRUE);    //discard updates
            return FALSE;
        }

        free(pFmCode);     //Free the FMCODE resource data

        //////// Update the FMTEXT ABOUT resource... ////////
        if (gParams->filterAboutText[0] != '\0') {
            //Use About Text that was set programmatically.
            pFmText = gParams->filterAboutText;
        }
        else {
            //Use default About Text.
            pFmText = defaultAboutText;
        }
        //Must interpolate the About Text string now, since it
        //won't be interpolated by the standalone filter.
        pFmText = fm_formatString(pFmText);
        if (!UR_UpdateResource(
            hExe,	                        // update-file handle 
            "FMTEXT",	                    // resource type
            "ABOUT",	                    // resource name
            MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_US),	// language identifier of resource 
            pFmText,	                    // address of resource data 
            (int)strlen(pFmText)+1          // length of resource data, in bytes 
           )) {
            SysErrorMessage("UpdateResource (FMTEXT ABOUT) failed:");
            UR_EndUpdateResource(hExe, TRUE);    //discard updates
            return FALSE;
        }

        //////// Write any embedded resources ////////
        {//scope
            int i;

            // Initialize the starting resource ID for RT_ICON resources...
            nID = 30100;    //arbitrary starting point; MSVC linker starts with 30000.

            for (i = 0; gParams->embeds[i] != '\0'; i++) {
                int iResType = gParams->embeds[i];
                switch (iResType) {
                case 'B':
                    // BITMAP resource
					i++;
                    if (!EmbedBitmapFile(hExe, &gParams->embeds[i])) {
                        return FALSE;   //failed
                    }
                    i += (int)strlen(&gParams->embeds[i]);
                    break;
                case 'I': //icon
                    i++;
                    if (!EmbedIconFile(hExe, &gParams->embeds[i])) {
                        return FALSE;   //failed
                    }
                    i += (int)strlen(&gParams->embeds[i]);
                    break;
                case 'C': //cursor
                    i++;
                    //ErrorOk("Embedding a CURSOR file is not yet implemented!");
                    if (!EmbedCursorFile(hExe, &gParams->embeds[i])) {
                        return FALSE;   //failed
                    }
                    i += (int)strlen(&gParams->embeds[i]);
                    break;
                case 'W': //wmf file
                case 'A': //audio file
				case 'O': //other file
                    // METAFILE, AUDIO, or OTHER resource
                    i++;
                    if (!EmbedOtherFile(hExe, &gParams->embeds[i], iResType)) {
                        return FALSE;   //failed
                    }
                    i += (int)strlen(&gParams->embeds[i]);
                    break;
                default:
                    // Unrecognized resource type
                    UR_EndUpdateResource(hExe, TRUE);    //discard updates
                    ErrorOk("Unrecognized resource type in Embed key: %x",
                        gParams->embeds[i]);
					return FALSE;
                }//switch
            }//for
            assert(gParams->cbEmbeds == i + 1);
        }//scope
 
        //////// Update the PiPL resource... ////////
        if (!CreatePipl()) {
            UR_EndUpdateResource(hExe, TRUE);    //discard updates
            msgBox(MB_ICONERROR, "FilterMeister System Error",
                   "Could not create PIPL resource");
            return FALSE;
        }

        // Add 'kind' property...
        dwProperty = '8BFM';          // Adobe Photoshop Filter Module
        if (!AddPiplProp(
                PIKindProperty,       // 'kind'
                0,                    // propertyID
                4,                    // propertyLength
                &dwProperty           // propertyData
            )) {
            UR_EndUpdateResource(hExe, TRUE);    //discard updates
            msgBox(MB_ICONERROR, "FilterMeister System Error",
                   "Could not add PiPL 'kind' property");
            return FALSE;
        }

        // Add 'name' property...
        // Create a bytecount-prefixed string (with ellipsis added).
        // Should use a default if null????????
        strcpy(&pstr[1], fm_stripEllipsis(gFmcp->filterTitleText));
#if 1
        // Add an ellipsis if user controls exist,
        // else no ellipsis for the case in which no
        // dialog is popped.  NYI!!!
        strcpy(&pstr[1], fm_appendEllipsis(&pstr[1]));
#endif
        pstr[0] = min(255, (int)strlen(&pstr[1])); //max pstr is 255 chars
        cbPropertyLength = (pstr[0] & 0xff) + 1;
        if (!AddPiplProp(
                PINameProperty,       // 'name'
                0,                    // propertyID
                cbPropertyLength,     // propertyLength
                pstr                  // propertyData
            )) {
            UR_EndUpdateResource(hExe, TRUE);    //discard updates
            msgBox(MB_ICONERROR, "FilterMeister System Error",
                   "Could not add PiPL 'name' property");
            return FALSE;
        }

        // Add 'catg' property...
        // Create a bytecount-prefixed string.
        // Should use a default if null????????
        strcpy(&pstr[1], gFmcp->filterCategoryText);
        pstr[0] = (int)strlen(gFmcp->filterCategoryText);
        cbPropertyLength = (int)strlen(gFmcp->filterCategoryText) + 1;
        if (!AddPiplProp(
                PICategoryProperty,   // 'catg'
                0,                    // propertyID
                cbPropertyLength,     // propertyLength
                pstr                  // propertyData
            )) {
            UR_EndUpdateResource(hExe, TRUE);    //discard updates
            msgBox(MB_ICONERROR, "FilterMeister System Error",
                   "Could not add PiPL 'catg' property");
            return FALSE;
        }

        // Add 'vers' property...
        dwProperty = (latestFilterVersion << 16) | latestFilterSubVersion;
        if (!AddPiplProp(
                PIVersionProperty,    // 'vers'
                0,                    // propertyID
                4,                    // propertyLength
                &dwProperty           // propertyData
            )) {
            UR_EndUpdateResource(hExe, TRUE);    //discard updates
            msgBox(MB_ICONERROR, "FilterMeister System Error",
                   "Could not add PiPL 'vers' property");
            return FALSE;
        }

#ifndef PIMaxSize32Property
#define PIMaxSize32Property 'ms32'
#endif
		// Add 'ms32', property for pixel dimensions > 30,000
		{//scope
			//set max image height and width to 300,000
			//which is the max currently supported by CS.
			static const struct {
				int maxHeight;
				int maxWidth;
			} maxSize = {300000, 300000};
			if (!AddPiplProp(
					PIMaxSize32Property,	// 'ms32'
					0,						// propertyID
					8,						// propertyLength
					&maxSize				// propertyData
				)) {
				UR_EndUpdateResource(hExe, TRUE);    //discard updates
				msgBox(MB_ICONERROR, "FilterMeister System Error",
					   "Could not add PiPL 'ms32' property");
				return FALSE;
			}
		}//scope

#if 1 //Support Big Documents (pixel dimension > 30000)
		"MIB8",		// '8BIM'
		"23sm",		// 'ms32'	PIMaxSize32Property
		0L,			// Index
		8L,			// Length
		300000L,	// Maximum height
		300000L,	// Maximum width
#endif
        // Add 'wx86' or '8664' property...
        // Create a bytecount-prefixed string.
        cbPropertyLength = (int)strlen("ENTRYPOINT00") + 1;
        if (!AddPiplProp(
                X64GEN ?
					PIWin64X86CodeProperty    // '8664'
					:
                    PIWin32X86CodeProperty,   // 'wx86'
                0,                        // propertyID
                cbPropertyLength,         // propertyLength
                "ENTRYPOINT00"            // propertyData
            )) {
            UR_EndUpdateResource(hExe, TRUE);    //discard updates
            msgBox(MB_ICONERROR, "FilterMeister System Error",
				X64GEN ?
                   "Could not add PiPL '8664' property"
				   :
                   "Could not add PiPL 'wx86' property"
				);
            return FALSE;
        }

        // Add 'mode' property...
#if 0
        //debugging...
        Info("PiPL.SupportedModes = 0x%8.8x", gParams->SupportedModes);
#endif
        if (!AddPiplProp(
                PIImageModesProperty,               // 'mode'
                0,                                  // propertyID
                sizeof(gParams->SupportedModes),    // propertyLength
                &gParams->SupportedModes            // propertyData
            )) {
            UR_EndUpdateResource(hExe, TRUE);    //discard updates
            msgBox(MB_ICONERROR, "FilterMeister System Error",
                   "Could not add PiPL 'mode' property");
            return FALSE;
        }

        // Add 'enbl' property if specified...
        if (gParams->EI_specified) {
            // convert to a null-terminated string (CString)...
            int nChars = gParams->EI_count;
            // should check nChars for reasonableness???
            char *p = _alloca(nChars + 1);
            memcpy(p, &gParams->szSrcCode[gParams->EI_start_index], nChars);
            p[nChars] = 0;  //null-terminate the string
#if 0
            //debugging...
            Info("PiPL.EnableInfo = \"%s\"", p);
#endif
            // should convert non-printing chars to ' ' (e.g., CR, LF, TAB, etc.)???
            // should strip leading/trailing white space?
            if (!AddPiplProp(
                    PIEnableInfoProperty,       // 'enbl'
                    0,                          // propertyID
                    nChars + 1,                 // propertyLength
                    p                           // propertyData
                )) {
                UR_EndUpdateResource(hExe, TRUE);    //discard updates
                msgBox(MB_ICONERROR, "FilterMeister System Error",
                       "Could not add PiPL 'enbl' property");
                return FALSE;
            }
        }


#ifdef SCRIPTABLE

        // Add 'hstm' property...
        {
            int i;
            for (i = 0; i < 56; i++) pstr[i] = 0;
        }
        strcpy(&pstr[4], "tliFtliF"); // Class ID + Event ID
        pstr[12] = 0x80; //16000 -> AETE Ressource number
        pstr[13] = 0x3E; //16000 -> AETE Ressource number
        strcpy(&pstr[14], gFmcp->filterUniqueID);
        
        if (!AddPiplProp(
                PIHasTerminologyProperty, //'hstm'
                0,                        // propertyID
                56,  //0x38       // propertyLength
                pstr                      // propertyData
            )) {
            UR_EndUpdateResource(hExe, TRUE);    //discard updates
            msgBox(MB_ICONERROR, "FilterMeister System Error",
                   "Could not add PiPL 'hstm' property");
            return FALSE;
        }
        
        
        // Add 'flly' property
        dwProperty = 0x80;
        if (!AddPiplProp(
                PIFilterLayerSupport, //0x666c6c79L, // 'flly'
                0,                    // propertyID
                4,                    // propertyLength
                &dwProperty           // propertyData
            )) {
            UR_EndUpdateResource(hExe, TRUE);    //discard updates
            msgBox(MB_ICONERROR, "FilterMeister System Error",
                   "Could not add PiPL 'flly' property");
            return FALSE;
        }

#endif //SCRIPTABLE


        // Add 'fici' property...
#if 0
        //debugging...
        {   int i;
            for (i = 0; i < 7; i++) {
                Info("fici[%d].inputHanlding = %d\n"
                     "fici[%d].outputHandling = %d\n"
                     "fici[%d].flags1 = 0x%.2x\n"
                     "fici[%d].flags2 = 0x%.2x",
                     i, gParams->fici[i].inputHandling,
                     i, gParams->fici[i].outputHandling,
                     i, gParams->fici[i].flags1,
                     i, gParams->fici[i].flags2);
            }
        }
#endif
        if (!AddPiplProp(
                PIFilterCaseInfoProperty, // 'fici'
                0,                        // propertyID
                sizeof(gParams->fici),    // propertyLength
                &gParams->fici            // propertyData
            )) {
            UR_EndUpdateResource(hExe, TRUE);    //discard updates
            msgBox(MB_ICONERROR, "FilterMeister System Error",
                   "Could not add PiPL 'fici' property");
            return FALSE;
        }

        pPipl = FinishPipl(&cbPipl);

        if (!UR_UpdateResource(
            hExe,	                        // update-file handle 
            "PIPL",	                    // resource type
            MAKEINTRESOURCE(16000),       // resource name
            MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_US), // language identifier of resource 
            pPipl,	                    // address of resource data 
            cbPipl 	                    // length of resource data, in bytes 
           )) {

            free(pPipl);
            SysErrorMessage("UpdateResource (PIPL) failed:");
            UR_EndUpdateResource(hExe, TRUE);    //discard updates
            return FALSE;
        }

        free(pPipl);     //Free the PiPL resource data


#ifdef SCRIPTABLE

        //////// Update the AETE resource... ////////
        {

            char *pAete = NULL;
            int cbAete = 0;
            int pos, length, i;
            char sNumber[10], sString[20];

            //Calc resource length
            cbAete = 10 + (int)strlen(gFmcp->filterCategoryText)+1 + 
						  (int)strlen(gFmcp->filterDescriptionText)+1 + 10 +
                          (int)strlen(gFmcp->filterTitleText)+1 +
                          (int)strlen(gFmcp->filterDescriptionText)+1 + 24 +
                          256*27 + 8;

            //Allocate resource
            pAete = calloc(cbAete, 1);
            if (pAete == NULL) {
                SysErrorMessage("UpdateResource (AETE) failed:");
                UR_EndUpdateResource(hExe, TRUE);    //discard updates
                free(pAete);
                return FALSE;
            }

            pAete[0] = 1; /* First byte must always be 1 */
            pAete[1] = 0;
            pAete[2] = 1; /* majorVersion|minorVersion */
            pAete[3] = 0;
            pAete[4] = 0; /* language */ 
            pAete[5] = 0;
            pAete[6] = 0; /* script */
            pAete[7] = 0;
            pAete[8] = 1; /* Suite Count */
            pAete[9] = 0;
            pos = 10;

            length = (int)strlen(gFmcp->filterCategoryText);
            pAete[pos] = length;
            strcpy(&pAete[pos+1],gFmcp->filterCategoryText);
            pos += length+1;

            length = (int)strlen(gFmcp->filterDescriptionText);
            pAete[pos] = length;
            strcpy(&pAete[pos+1],gFmcp->filterDescriptionText);
            pos += length+1;

            /* Suite ID */
            strcpy(&pAete[pos],"tliF");
            pos += 4;

            pAete[pos] = 1;   /* Suite level */
            pAete[pos+1] = 0;
            pAete[pos+2] = 1; /* Suite version */
            pAete[pos+3] = 0;
            pAete[pos+4] = 1; /* Event Count */
            pAete[pos+5] = 0;
            pos += 6;

            length = (int)strlen(gFmcp->filterTitleText);
            pAete[pos] = length;
            strcpy(&pAete[pos+1],gFmcp->filterTitleText);
            pos += length+1;
            
            length = (int)strlen(gFmcp->filterDescriptionText);
            pAete[pos] = length;
            strcpy(&pAete[pos+1],gFmcp->filterDescriptionText);
            pos += length+1;

            /* Event class */
            strcpy(&pAete[pos],"tliF");
            pos += 4;

            /* Descriptor type */
            strcpy(&pAete[pos],"tliF");
            pos += 4;

            /* Reply type */
            strcpy(&pAete[pos],"llun");
            pos += 4;
	        
            pAete[pos] = 0;   
            pAete[pos+1] = 0; 
            pAete[pos+2] = 0; /* Flags */
            pos += 3;
	    
		    /* Direct param type */
            strcpy(&pAete[pos],"RmI#");
            pos += 4;

            pAete[pos] = 0;  
            pAete[pos+1] = 0; /* Direct params */
            pAete[pos+2] = (char)0xB0; 
            pAete[pos+3] = 0; /* Parameter Count */   //0x0001 = 256 parameters in this case 
            pAete[pos+4] = 1; 
            pos += 5; 


            //Start with Params
            for (i=0;i<256;i++)
            { 
                pAete[pos] = 6;

                _itoa( i, sNumber, 10 );
                //Add zeros in front of number
                if (i<10){
                    sNumber[3] = 0;
                    sNumber[2] = sNumber[0];
                    sNumber[1] = 0x30;
                    sNumber[0] = 0x30;
                } else if (i<100) {
                    sNumber[3] = 0;
                    sNumber[2] = sNumber[1];
                    sNumber[1] = sNumber[0];
                    sNumber[0] = 0x30;
                }

                strcpy (sString,"ctl");
                strcat (sString,sNumber);
                strcpy(&pAete[pos+1],sString); //"ctl000" //Displayed Name

                pAete[pos+7] = sNumber[2]; //"0000" //Internal Name
                pAete[pos+8] = sNumber[1];
                pAete[pos+9] = sNumber[0];
                pAete[pos+10] = 0x30;

                strcpy(&pAete[pos+11],"gnol"); //Param Type

                pAete[pos+15] = 9;
                strcpy (sString,"contro");
                strcat (sString,sNumber);
                strcpy(&pAete[pos+16],sString); //"contro000" //Description

                pAete[pos+25] = 0;
                pAete[pos+26] = 0;
                pos += 27;
            }



            //End
            pAete[pos] = 0;  
            pAete[pos+1] = 0;
            pAete[pos+2] = 0;
            pAete[pos+3] = 0;
            pAete[pos+4] = 0;
            pAete[pos+5] = 0;
            pAete[pos+6] = 0;
            pAete[pos+7] = 0;
            pos += 8;

            //cbAete = pos; //Size of the Resource (already computed above!)



        
            if (!UR_UpdateResource(
                hExe,	                        // update-file handle 
                "AETE",	                    // resource type
                MAKEINTRESOURCE(16000),       // resource name
                MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_US), // language identifier of resource 
                pAete,	                    // address of resource data 
                cbAete 	                    // length of resource data, in bytes 
               )) {

                free(pAete);
                SysErrorMessage("UpdateResource (AETE) failed:");
                UR_EndUpdateResource(hExe, TRUE);    //discard updates
                return FALSE;
            }

            free(pAete);     //Free the PiPL resource data

        }
#endif


        //////// Update the VS_VERSION_INFO resource... ////////
        {//scope
            VS_BLOCK_HEADER *pVS;
            static VS_FIXEDFILEINFO fixedFileInfo = {
                VS_FFI_SIGNATURE,           /* dwSignature; e.g. 0xfeef04bd */
                VS_FFI_STRUCVERSION,        /* dwStrucVersion; e.g. 0x00000042 = "0.42", 0x00010000 = "1.00" */
                // The following 2 fields should really be parsed from the FD-specified
                // text in the filterVersionText (and productVersionText) strings ????
				// Try reading the filterVersionText string with the following sscanf formats:
				//		"%*[^0-9] %u.%u.%u.%u" (and count valid fields)
				//		"%*[^0-9] %u.%u.%u" (for fewer fields is field count not implemented)
				//		"%*[^0-9] %u.%u"
				// until a match is found
				// Alternately, try the following instead of &*[^0-9] :
				//		%[^0-9] and eat the string
				//		%*[^0123456789] if '-' is not accepted (non-ANSI)
				//		%*[a-zA-Z] if '^' is not accepted
				//		%*[VERSIONversion] to be pickier (Accepts "V1.0","Version 1.2.3.4", "NoiseR 3.2", etc.
#if 0			// 0.4.21.0
                0x00000004, 0x00150000,     // File version MS/LS   -- should agree with text
                0x00000004, 0x00150000,     // Product version MS/LS-- versions below!!
#elif 0			// 1.0.0.0
                0x00010000, 0x00000000,     // File version MS/LS   -- should agree with text
                0x00010000, 0x00000000,     // Product version MS/LS-- versions below!!
#else			// parse from filterVersionText, default is 0.0.0.0
                0x00000000, 0x00000000,     // File version MS/LS   -- should agree with text
                0x00000000, 0x00000000,     // Product version MS/LS-- versions below!!
#endif
                VS_FFI_FILEFLAGSMASK,       /* dwFileFlagsMask = 0x3F for version "0.42" (and 1.00 also?) */
                0,                          /* dwFileFlags; e.g. VFF_DEBUG | VFF_PRERELEASE */ //FD could specify?
                VOS_NT_WINDOWS32,           /* dwFileOS; e.g. VOS_DOS_WINDOWS16 */
                VFT_DLL,                    /* dwFileType; e.g. VFT_DRIVER */
                VFT2_UNKNOWN,               /* dwFileSubtype; e.g. VFT2_DRV_KEYBOARD */
                0,                          /* dwFileDateMS; e.g. 0 */
                0                           /* dwFileDateLS; e.g. 0 */
            };
            static struct {
                WORD langID;
                WORD charsetID;
            } translation = {0x409, 1200};  // = 0x409, 0x4b0 -- English(United States)
            char szFmNameAndVersion[100];
            char szBuffer[128];
			DWORD fver[4];
			int nFields;

			// Parse File Version from filterVersionText.
			// Assumes missing fields are not altered by sscanf;
			// is that correct??
			fver[0] = fver[1] = fver[2] = fver[3] = 0; //default is 0.0.0.0
			nFields = sscanf(gFmcp->filterVersionText, "%*[^0-9]%u.%u.%u.%u",
							 &fver[0], &fver[1], &fver[2], &fver[3]);
			if (nFields <= 0) {
				//maybe filterVersionText starts with a digit...
				nFields = sscanf(gFmcp->filterVersionText, "%u.%u.%u.%u",
					             &fver[0], &fver[1], &fver[2], &fver[3]);
			}
			//pack into big-endian dwords...
			fixedFileInfo.dwFileVersionMS = (fver[0] << 16) | (WORD)fver[1];
			fixedFileInfo.dwFileVersionLS = (fver[2] << 16) | (WORD)fver[3];
			//for now, product version is same as file version
			fixedFileInfo.dwProductVersionMS = fixedFileInfo.dwFileVersionMS;
			fixedFileInfo.dwProductVersionLS = fixedFileInfo.dwFileVersionLS;

            if (!VS_init()) {
                UR_EndUpdateResource(hExe, TRUE);    //discard updates
                msgBox(MB_ICONERROR, "FilterMeister System Error",
                       "Could not create VERSIONINFO resource");
                return FALSE;
            }

            if (!VS_begin_binary_block("VS_VERSION_INFO",
                                       &fixedFileInfo,
                                       sizeof(fixedFileInfo))) {
                UR_EndUpdateResource(hExe, TRUE);    //discard updates
                msgBox(MB_ICONERROR, "FilterMeister System Error",
                       "Could not add VS_VERSION_INFO");
                return FALSE;
            }

                if (!VS_begin_string_block("StringFileInfo", NULL)) {
                    UR_EndUpdateResource(hExe, TRUE);    //discard updates
                    msgBox(MB_ICONERROR, "FilterMeister System Error",
                           "Could not add StringFileInfo");
                    return FALSE;
                }

                    if (!VS_begin_string_block("040904b0", NULL)) {
                        UR_EndUpdateResource(hExe, TRUE);    //discard updates
                        msgBox(MB_ICONERROR, "FilterMeister System Error",
                               "Could not add lang-charset info");
                        return FALSE;
                    }
#if 0
                    //////// Fixed data for regression testing ////////
                        if (!VS_begin_string_block("Comments",
                            "Beta version -- do not redistribute!")) {
                            UR_EndUpdateResource(hExe, TRUE);    //discard updates
                            msgBox(MB_ICONERROR, "FilterMeister System Error",
                                   "Could not add Comments info");
                            return FALSE;
                        }
                        VS_end_block();

                        if (!VS_begin_string_block("CompanyName",
                            "AFH Systems, Mountain View, CA 94040, http://www.afh.com")) {
                            UR_EndUpdateResource(hExe, TRUE);    //discard updates
                            msgBox(MB_ICONERROR, "FilterMeister System Error",
                                   "Could not add CompanyName info");
                            return FALSE;
                        }
                        VS_end_block();

                        if (!VS_begin_string_block("FileDescription",
                            "FilterMeister Plug-in Filter")) {
                            UR_EndUpdateResource(hExe, TRUE);    //discard updates
                            msgBox(MB_ICONERROR, "FilterMeister System Error",
                                   "Could not add FileDescription info");
                            return FALSE;
                        }
                        VS_end_block();

                        if (!VS_begin_string_block("FileVersion",
                            "0, 4, 13, 0")) {
                            UR_EndUpdateResource(hExe, TRUE);    //discard updates
                            msgBox(MB_ICONERROR, "FilterMeister System Error",
                                   "Could not add FileVersion info");
                            return FALSE;
                        }
                        VS_end_block();

                        if (!VS_begin_string_block("InternalName",
                            "AfhFM04")) {
                            UR_EndUpdateResource(hExe, TRUE);    //discard updates
                            msgBox(MB_ICONERROR, "FilterMeister System Error",
                                   "Could not add InternalName info");
                            return FALSE;
                        }
                        VS_end_block();

                        if (!VS_begin_string_block("LegalCopyright",
                            "Copyright  1999-2001 AFH Systems")) {
                            UR_EndUpdateResource(hExe, TRUE);    //discard updates
                            msgBox(MB_ICONERROR, "FilterMeister System Error",
                                   "Could not add LegalCopyright info");
                            return FALSE;
                        }
                        VS_end_block();

                        if (!VS_begin_string_block("OriginalFilename",
                            "AfhFm04.8bf")) {
                            UR_EndUpdateResource(hExe, TRUE);    //discard updates
                            msgBox(MB_ICONERROR, "FilterMeister System Error",
                                   "Could not add OriginalFilename info");
                            return FALSE;
                        }
                        VS_end_block();

                        if (!VS_begin_string_block("ProductName",
                            "FilterMeister Plug-in Filter")) {
                            UR_EndUpdateResource(hExe, TRUE);    //discard updates
                            msgBox(MB_ICONERROR, "FilterMeister System Error",
                                   "Could not add ProductName info");
                            return FALSE;
                        }
                        VS_end_block();

                        if (!VS_begin_string_block("ProductVersion",
                            "0, 4, 13, 0")) {
                            UR_EndUpdateResource(hExe, TRUE);    //discard updates
                            msgBox(MB_ICONERROR, "FilterMeister System Error",
                                   "Could not add ProductVersion info");
                            return FALSE;
                        }
                        VS_end_block();
#else
                        //////// Here's the real version info... ////////

                        if (!VS_begin_string_block("Author",
                            gFmcp->filterAuthorText)) {
                            UR_EndUpdateResource(hExe, TRUE);    //discard updates
                            msgBox(MB_ICONERROR, "FilterMeister System Error",
                                   "Could not add Author info");
                            return FALSE;
                        }
                        VS_end_block();
#if 1
                        // Append FM version # (IDS_FMVERSION) to Comments string...
                        //
                        // Bare version number string is resource # IDS_FMVERSION.
                        // (Or should we use string resource # 2, which includes both
                        // the FM product name and version number?)						//TODO--done
                        //
                        if (LoadString(hDllInstance, 2, szFmNameAndVersion, sizeof szFmNameAndVersion) <= 0) {
                            // Version number not found, or null.
#ifdef VERBOSE
                            ErrorOk("FM version number resource string not found or corrupt.");
#endif
                            strncpy(szFmNameAndVersion, "FM 1.0", sizeof(szFmNameAndVersion));  //default version
                        }
                        szFmNameAndVersion[sizeof(szFmNameAndVersion)-1] = 0; //ensure null-terminated
                        _snprintf(szBuffer, sizeof(szBuffer), "Generated by %s", szFmNameAndVersion);
                        szBuffer[sizeof(szBuffer)-1] = 0; //ensure null-terminated
                        if (!VS_begin_string_block("Comments", szBuffer)) {
                            UR_EndUpdateResource(hExe, TRUE);    //discard updates
                            msgBox(MB_ICONERROR, "FilterMeister System Error",
                                   "Could not add Comments info");
                            return FALSE;
                        }
                        VS_end_block();
#endif
                        if (!VS_begin_string_block("CompanyName",
                            gFmcp->filterOrganizationText)) {
                            UR_EndUpdateResource(hExe, TRUE);    //discard updates
                            msgBox(MB_ICONERROR, "FilterMeister System Error",
                                   "Could not add CompanyName info");
                            return FALSE;
                        }
                        VS_end_block();

                        if (!VS_begin_string_block("FileDescription",
                            gFmcp->filterDescriptionText)) {
                            UR_EndUpdateResource(hExe, TRUE);    //discard updates
                            msgBox(MB_ICONERROR, "FilterMeister System Error",
                                   "Could not add FileDescription info");
                            return FALSE;
                        }
                        VS_end_block();

                        if (!VS_begin_string_block("FileVersion",
                            gFmcp->filterVersionText)) {
                            UR_EndUpdateResource(hExe, TRUE);    //discard updates
                            msgBox(MB_ICONERROR, "FilterMeister System Error",
                                   "Could not add FileVersion info");
                            return FALSE;
                        }
                        VS_end_block();

                        if (!VS_begin_string_block("InternalName",
                            "AfhFM04s")) {
                            UR_EndUpdateResource(hExe, TRUE);    //discard updates
                            msgBox(MB_ICONERROR, "FilterMeister System Error",
                                   "Could not add InternalName info");
                            return FALSE;
                        }
                        VS_end_block();

                        if (!VS_begin_string_block("LegalCopyright",
                            gFmcp->filterCopyrightText)) {
                            UR_EndUpdateResource(hExe, TRUE);    //discard updates
                            msgBox(MB_ICONERROR, "FilterMeister System Error",
                                   "Could not add LegalCopyright info");
                            return FALSE;
                        }
                        VS_end_block();
#if 1
                        if (!VS_begin_string_block("LegalTrademarks",
                            "FilterMeister is a trademark of AFH Systems")) {
                            UR_EndUpdateResource(hExe, TRUE);    //discard updates
                            msgBox(MB_ICONERROR, "FilterMeister System Error",
                                   "Could not add LegalTrademarks info");
                            return FALSE;
                        }
                        VS_end_block();
#endif
                        if (!VS_begin_string_block("OriginalFilename",
                            szTitleName)) {
                            UR_EndUpdateResource(hExe, TRUE);    //discard updates
                            msgBox(MB_ICONERROR, "FilterMeister System Error",
                                   "Could not add OriginalFilename info");
                            return FALSE;
                        }
                        VS_end_block();

                        if (!VS_begin_string_block("ProductName",
                            gFmcp->filterTitleText)) {
                            UR_EndUpdateResource(hExe, TRUE);    //discard updates
                            msgBox(MB_ICONERROR, "FilterMeister System Error",
                                   "Could not add ProductName info");
                            return FALSE;
                        }
                        VS_end_block();

                        if (!VS_begin_string_block("ProductVersion",
                            gFmcp->filterVersionText)) {
                            UR_EndUpdateResource(hExe, TRUE);    //discard updates
                            msgBox(MB_ICONERROR, "FilterMeister System Error",
                                   "Could not add ProductVersion info");
                            return FALSE;
                        }
                        VS_end_block();

                        if (gFmcp->filterURLText[0] != 0) {
                            if (!VS_begin_string_block("URL",
                                gFmcp->filterURLText)) {
                                UR_EndUpdateResource(hExe, TRUE);    //discard updates
                                msgBox(MB_ICONERROR, "FilterMeister System Error",
                                       "Could not add URL info");
                                return FALSE;
                            }
                            VS_end_block();
                        }
#endif
                    VS_end_block();   //040904b0

                VS_end_block();   //StringFileInfo

                if (!VS_begin_string_block("VarFileInfo", NULL)) {
                    UR_EndUpdateResource(hExe, TRUE);    //discard updates
                    msgBox(MB_ICONERROR, "FilterMeister System Error",
                           "Could not add VarFileInfo");
                    return FALSE;
                }

                    if (!VS_begin_binary_block("Translation",
                                               &translation,
                                               sizeof(translation))) {
                        UR_EndUpdateResource(hExe, TRUE);    //discard updates
                        msgBox(MB_ICONERROR, "FilterMeister System Error",
                               "Could not add Translation info");
                        return FALSE;
                    }
                    VS_end_block();

                VS_end_block();   //VarFileInfo

            VS_end_block();   //VS_VERSION_INFO

            pVS = VS_finish();

            if (!UR_UpdateResource(
                hExe,	                        // update-file handle 
                RT_VERSION,	                // resource type
                MAKEINTRESOURCE(1),           // resource name
                MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_US), // language identifier of resource 
                pVS,	                        // address of resource data 
                pVS->wLength                  // length of resource data, in bytes 
               )) {

                free(pVS);
                SysErrorMessage("UpdateResource (VERSIONINFO) failed:");
                UR_EndUpdateResource(hExe, TRUE);    //discard updates
                return FALSE;
            }

            free(pVS);     //Free the VERSIONINFO resource data

        }//scope (VS_VERSION_INFO)

        //////// Finished updating resources! ////////

        if (!UR_EndUpdateResource(
            hExe,	// update-file handle  
            FALSE 	// do not discard changes 
           )) {

            SysErrorMessage("EndUpdateResource failed:");
            return FALSE;
        }

#if 1
		{//scope
			//Set the target name (safe to use X64gen, or some other flag?
			//X64GEN macro may no longer be correct???
			char *pszTarget = X64GEN ? "X64" : "X32";

			Info("Microsoft %s %d.%d (Build %d %s)\n\n"
				 "The %s standalone filter was successfully created in file:\n"
				 "\"%s\"\n\n"
#if 1
				 "Copy this file to the host application's plug-in folder.\n\n"
#endif
				 "Exit and (re)start the %s host application (e.g., !H) to use "
				 "the new filter, which should appear as:\n\n"
				 "\t\"!t...\"\n\n"
				 "under the filter menu category:\n\n"
				 "\t\"!C\"",
				 pszPlatform,
				 osvi.dwMajorVersion,
				 osvi.dwMinorVersion,
				 osvi.dwBuildNumber & 0xFFFF,
				 osvi.szCSDVersion,
				 pszTarget,
				 szTemplateFile,
				 pszTarget);
		}//scope
#endif
    }//scope
#endif //!BETA
    return TRUE;        //success!
} //FmMakeStandalone

#undef UR_BeginUpdateResource
#undef UR_UpdateResource
#undef UR_EndUpdateResource

#endif //ifndef STANDALONE
