/*************************************************************************/
/*                                                                       */
/*                        AfhFMsymtab.c                                  */
/*                                                                       */
/*  This module contains symbol table routines for the FilterMeister.    */
/*                                                                       */
/*  Copyright 1999-2014 AFH Systems                                      */
/*                                                                       */
/*  FilterMeister is free software. You can redistribute and modify it   */
/*  under the terms of the MPL2 License.                                 */
/*  See: https://www.mozilla.org/en-US/MPL/2.0/                          */
/*  Any changes to the FilterMeister files have to be published or sent  */
/*  to harry@thepluginsite.com. The filter plugin code itself does not   */
/*  need to be disclosed.                                                */
/*                                                                       */
/*************************************************************************/
/*                                                                       */
/*              R E V I S I O N   H I S T O R Y                          */
/*                                                                       */
/*  02Jan99  Alex Hunter  1. Created.                                    */
/*  01Sep12  Alex Hunter  1. Add minimal X64 support.					 */
/*  16Jul13  Alex Hunter  1. Modify for WIN64.							 */
/*	17Sep13  Alex Hunter  1. Make enhancements to display_symbol_scope.	 */
/*	03Oct13	 Alex Hunter  1. Add HALF 16-bit floating-point type.		 */
/*						  2. Add INT_PTR and UINT_PTR types.			 */
/*	04Oct13	 Alex Hunter  1. Test allocate actual size of variable.		 */
/*						  2. Allocate vars with "native" alignment.		 */
/*	21Oct13	 Alex Hunter  1. Change 'treal' to 'tempreal'.				 */
/*	24May14	 Alex Hunter  1. Avoid redefinition of PRINTF* macros.		 */
/*  06Jun14  Alex Hunter  1. Change DCL_TREAL to DCL_TEMPREAL.			 */
/*	10Jun14	 Alex Hunter  1. Add fmfNONNEG for unsigned types.			 */
/*							 N.B. fmfSIGNED and fmfUNSIGNED now reflect  */
/*							 only the actual presence of the respective  */
/*							 keywords. fmfNONNEG specifies the actual    */
/*							 signed or unsigned state.					 */
/*						  2. Remove fmfSIGNED from fundamental types.	 */
/*						  3. Add 'const' versions of fundamental types.	 */
/*																		 */
/*					T O D O   L I S T									 */
/*                                                                       */
/*	- Drop the 'size' param to create_scalar_type(); let this function	 */
/*	  determine the size from the 'subtype' field.						 */
/*	- Maybe ditto for the 'flags' param to create_scalar_type?			 */
/*	- Implement 'saturated' types.										 */
/*	- If DEBUG then keep track of all memory allocations, report leaks.	 */
/*                                                                       */
/*************************************************************************/

////// Do uses of pList overlap??????

#include "plmcvax.h"
#include "AfhFM.h"            /* msgBox.                                 */
#include "AfhFMcontext.h"     /* gFmcp, etc.                             */
#include "AfhFMsymtab.h"      /* Symbol table definitions.               */
#include "md386def.hc"        /* MD386__ message codes.                  */
#include "md386lit.h"         /* MD386 common definitions.               */
#include "mdg.h"              /* TASK_REGS.                              */
#include "mdd.h"
#include "mdy.h"              /* QUALIFIERSTRUCTURE.                     */
#include "md386proto.h"       /* MD386 intermodule function prototypes.  */

#include <stdio.h>            /* sprintf.                                */
#include <ctype.h>            /* toupper.                                */
#include <stddef.h>           /* offsetof.                               */
#include <string.h>           /* memcpy, memmove, memset, strcat,        */
                              /*  strcpy, strlen, strncat.               */
#include "mdp.h"              /* Semantic descriptors.                   */
#include <assert.h>


/******************* Static data and variables ************************/



    /* Formatted print routines for FM */

#ifndef BETA
#ifndef PRINTF

void FMprint(const char *buff);

#define PRINTF_BUFF_SIZE 256

#ifdef NDEBUG
#define CHECK_BUFF(exp) (exp)
#else
#define CHECK_BUFF(exp) assert((exp) < PRINTF_BUFF_SIZE)
#endif

#define PRINTF(f) \
    do {\
        FMprint(f);\
    } while(0)

#define PRINTF2(f,a) \
    do {\
        char buff[PRINTF_BUFF_SIZE];\
        CHECK_BUFF( sprintf(buff, f, a) );\
        FMprint(buff);\
    } while(0)

#define PRINTF3(f,a,b) \
    do {\
        char buff[PRINTF_BUFF_SIZE];\
        CHECK_BUFF( sprintf(buff, f, a, b) );\
        FMprint(buff);\
    } while(0)

#endif //ifndef PRINTF
#endif //ifndef BETA

/****** Error Reporting ******/

#ifdef _DEBUG
#define BUG(n) Mde_abort_command(MD386__BUG, (n), 11, (UINT4) "AfhFMsymtab")
#else
#define BUG(n) (void)
#endif


/****** Hash coding stuff ******/

#define NHASH 64

#if (NHASH & -NHASH)==NHASH
#define NHASH_IS_POWER_OF_2 1
#else
#define NHASH_IS_POWER_OF_2 0
#endif

static int fmHash(const char *name)
{
    int h = 0;
    while (*name) {
        h += *name++;	//awful hash???
    }
#if NHASH_IS_POWER_OF_2
    return h & (NHASH - 1);
#else
    return (h & 0x7fffffff) % NHASH;
#endif
} /*fmHash*/

/****** Memory Allocation Routines ******/

static void *fmNew(int nBytes)
{
    //if debug, keep track of amount allocated/deallocated to check
    //for memory leaks!!!!
	//Allocates only 3 different sized objects, so could use a simple
	//array allocator a la md386, to make wholesale deletion more
	//convenient and efficient...
    void *p = calloc(1, nBytes);
    if (!p) Mde_fatal(MD386__EMEM, 0, 0, 0);
    return p;
} /*fmNew*/

static void fmDelete(void *p)
{
    //if debug, keep track of amount allocated/deallocated to check
    //for memory leaks!!!!
	//Also should check errno...
    free(p);
} /*fmDelete*/

pfm_OBJ newOBJ(fm_WHAT what)
{
    pfm_OBJ p = fmNew(sizeof(*p));
    p->what = what;
    return p;
} /*newOBJ*/

/****** Object constructors ******/

pfm_OBJ create_scalar_type(fm_BASETYPE basetype, fm_SUBTYPE subtype,
                           int size, const char *name, int flags)
{
    pfm_OBJ p = newOBJ(fmSCALAR);
    p->basetype = basetype;
    p->subtype = subtype;
    p->am = amUNDEF;
    p->name = name;
    p->pType = NULL;
    p->pList = NULL;
    p->flags = flags;
    p->nCount = 0;
    p->offset = 0;
    p->size = size;
    return p;
} /*create_scalar_type*/

/****** Block-scoped symbol table ******/

typedef pfm_OBJ HASH_BUCKET[NHASH];

typedef struct fm_SCOPE {
    struct fm_SCOPE *pOuterScope;
    HASH_BUCKET *hb;
    int cbStackFrameOffset;
	int nBlock;			//block # at next level
    //char *nextString;	//string store marker
} fm_SCOPE, *pfm_SCOPE;

static pfm_SCOPE top_scope = NULL;
static int cbStackFrameOffset;
static int cbMaxStackDepth;
static int cbMaxGlobalStack;
static int cbStaticFrameOffset;

#ifndef BETA

static void print_block_no(pfm_SCOPE scope)
{
    assert(scope != NULL);
	if (scope->pOuterScope) {
		print_block_no(scope->pOuterScope);
		PRINTF(".");
	}
	PRINTF2("%d", scope->nBlock);
} /*print_block_no*/

static void display_symbol_scope(pfm_SCOPE scope)
{
    int i;
    pfm_OBJ p;
    assert(scope != NULL);
    if (scope->pOuterScope) {
        PRINTF("\r\n=== begin block scope ");
		print_block_no(scope->pOuterScope);
		PRINTF(" ===\r\n");
    } else {
        PRINTF("\r\n=== begin GLOBAL scope ===\r\n");
    }
    PRINTF2("beginning stack frame offset = %d\r\n", scope->cbStackFrameOffset);
    if (scope->hb) {
        for (i = 0; i < NHASH; i++) {
            for (p = (*scope->hb)[i]; p != NULL; p = p->pList) {
                PRINTF("   ");
                display_OBJ(p);
                PRINTF("\r\n");
            }//for p
        }//for i
    }
    else {
        PRINTF(" *** no declarations ***\r\n");
    }//if
    PRINTF2("stack depth of this block = %d\r\n", cbStackFrameOffset);
    if (scope->pOuterScope) {
        PRINTF2("max stack depth = %d\r\n", cbMaxStackDepth);
    } else {
        PRINTF2("max global stack = %d\r\n", cbMaxGlobalStack);
		PRINTF2("total global variable size allocated = %d bytes\r\n",
			cbStaticFrameOffset - offsetof(FMcontext, globvars.user_vars));
		PRINTF2("total global variable size allowed = %d bytes\r\n",
			sizeof(gFmcp->globvars.user_vars));
    }
    PRINTF("==== end block scope ====\r\n");
} /*display_symbol_scope*/

#endif //ifndef BETA

void enter_scope(void)
{
    pfm_SCOPE new_scope = fmNew(sizeof(*new_scope));
    new_scope->pOuterScope = top_scope;
	if (top_scope)
		top_scope->nBlock++;	//bump # children of outer block
	new_scope->nBlock = 0;		//no nested blocks so far
    new_scope->cbStackFrameOffset = cbStackFrameOffset;
    // reset max stack depth at start of each level-2 block...
    if (top_scope && !top_scope->pOuterScope) cbMaxStackDepth = 0;
    top_scope = new_scope;
} /*enter_scope*/

int leave_scope(void)
{
    pfm_SCOPE this_scope = top_scope;
    assert(top_scope != NULL);
	if (cbStackFrameOffset < cbMaxStackDepth) {
		//stack frame grows downward..
        cbMaxStackDepth = cbStackFrameOffset;
	}
#ifndef BETA
#if 1
     if ((gParams->flags & (AFH_EDITMODE|AFH_ADVANCED|AFH_DUMP_SDL)) ==
                           (AFH_EDITMODE|AFH_ADVANCED|AFH_DUMP_SDL)) {
        display_symbol_scope(top_scope);
     }
#endif
#endif //ifndef BETA
    cbStackFrameOffset = top_scope->cbStackFrameOffset;
    top_scope = top_scope->pOuterScope;
    if (this_scope->hb) fmDelete(this_scope->hb);
    //should also delete all objects from top scope,
    //pop string store, types, etc!!!!             ///// HUGE MEM LEAK AFTER EACH COMPILATION???
    fmDelete(this_scope);
    if (top_scope && !top_scope->pOuterScope) {
        //update global stack depth at end of each level-2 block...
		if (cbMaxStackDepth < cbMaxGlobalStack) {
			//stack frame grows downward..
			//AN OVERESTIMATE!!! (INCLUDES function/handler stacks)
			//EH???  Is cbMaxGlobalStack supposed to be the max stack
			//used by auto vars in global blocks??
			//or are nested global vars allocated in the FMC???
            cbMaxGlobalStack = cbMaxStackDepth;
		}
    }
#if 1	//either of these should give the same results???
	return -cbMaxStackDepth; //max stack depth for this block (including parents)
#else
	return -cbStackFrameOffset; //max stack depth for this block
#endif
} /*leave_scope*/

pfm_OBJ enter_symbol(const char *name, fm_WHAT what, pfm_OBJ pType)
{
    pfm_OBJ p;
    int h = fmHash(name);
    assert(top_scope != NULL);
    if (top_scope->hb == NULL) {
        //allocate a hash bucket for this scope
        top_scope->hb = fmNew(sizeof(HASH_BUCKET));
    }
    else {
        //Check for duplicate symbol definition
        for (p = (*top_scope->hb)[h]; p != NULL; p = p->pList) {
            if (strcmp(name, p->name) == 0) {
                Mde_warn(MD386__DUPSYM, (UINT_PTR)name, 0, 0);
                break;
            }
        } //for
    } //if
    //Enter the symbol regardless...
    p = newOBJ(what);
    p->pList = (*top_scope->hb)[h];
    (*top_scope->hb)[h] = p;
    p->name = name;
    p->pType = pType;
    p->basetype = btUNDEF;
    p->subtype = stUNDEF;
    p->am = amUNDEF;
    if (what == fmVAR) {
        int size;
		assert(pType != NULL);	//vars must have a type
        p->flags |= pType->flags;   //actually only need storage class and type qualifier flags???
        // Vars at outermost level (level 0) are 'GLOBAL'...
        if (top_scope->pOuterScope == NULL) p->flags |= fmfGLOBAL;
        if (pType->what == fmSCALAR && pType->basetype == btREAL) {
            //for now, all reals are allocated 8 bytes					//EH??
            size = 8;
        }
#ifdef X64
		else if (pType->what == fmSCALAR && pType->basetype == btINTEGRAL &&
				 (pType->subtype == stINT64 || pType->subtype == stQWORD)) {
			size = 8;		//64-bit integer
		}
		else if (pType->what == fmSCALAR && pType->basetype == btINTEGRAL &&
				 (pType->subtype == stINT_PTR || pType->subtype == stUINT_PTR)) {
			if (X64GEN) {
				size = 8;	//X64: INT_PTR types are 64-bits
			}
			else {
				size = 4;	//X86: INT_PTR types are 32-bits
			}
		}
		else if (pType->what == fmPOINTER) {
			if (X64GEN) {
				size = 8;	//X64 pointers are 64-bits
			}
			else {
				size = 4;	//X86 pointers are 32-bits
			}
		}
#endif //X64
		else if (pType->what == fmARRAY) {
			ErrorOk("Arrays are not yet implemented.");
			size = 4;
		}
 		else if (pType->what == fmSTRUCT) {
			ErrorOk("Structures are not yet implemented.");
			size = 4;
		}
 		else if (pType->what == fmUNION) {
			ErrorOk("Unions are not yet implemented.");
			size = 4;
		}
 		else if (pType->what == fmENUM) {
			size = 4;
		}
        else {
            //all others are allocated 4 bytes for now...
            size = 4;
        }

		// Determine actual size
#if 1
		// test code: override allocation size with "theoretical" size...
		size = pType->size;
#endif
#if 1
		p->size = pType->size;	//theoretically correct size
#else
		p->size = size;	//actually allocated size (for now)
#endif

        // Determine access method:
        if (p->flags & fmfSTATIC) {
            p->am = amSTATIC;
        }
        else if (p->flags & fmfAUTO) {
            if (p->flags & fmfGLOBAL) {
                // Symbol is global (outermost scope).
                ErrorOk("Global symbol '%s' must be static, not 'auto'.", name);
				p->flags &= ~fmfAUTO;
				p->am = amSTATIC;
            }
            else p->am = amSTACK;   //Auto vars are allocated on stack
        }
        else if (p->flags & fmfREGISTER) {
            if (p->flags & fmfGLOBAL) {
                // Symbol is global (outermost scope).
                ErrorOk("Global symbol '%s' must be static, not 'register'.", name);
				p->flags &= ~fmfREGISTER;
				p->am = amSTATIC;
            }
            else p->am = amSTACK;   //Register vars not yet implemented, so use auto
        }

        if (p->am == amUNDEF) {
            //Use default access method...
            if (p->flags & fmfGLOBAL) {
                //Default for global (outermost scope) vars is static
                p->am = amSTATIC;
            }
            else {
                //Default for all other vars is auto
                p->am = amSTACK;
            }
        }

        // Now allocate storage for the variable...
        if (p->am == amSTATIC) {
            //allocate static variable
			int nAlign = min(size, 16);
			if (nAlign == 10) nAlign = 2;	//align tempreal's mod 2
			assert(IS_POWER_OF_TWO(nAlign));
			cbStaticFrameOffset = (cbStaticFrameOffset + nAlign - 1) & -nAlign; //align mod nAlign
            p->offset = cbStaticFrameOffset;
            cbStaticFrameOffset += size;
            if (cbStaticFrameOffset - offsetof(FMcontext, globvars.user_vars) > sizeof(gFmcp->globvars.user_vars)) {
                // too many global/static vars...
                ErrorOk("Too many global or static variables: %s", name);
                //Mde_fatal(MD386__DICTOVFL, 0, 0, 0);
            }
        }
        else {
            //allocate variable on stack
			int nAlign = min(size, 16);
			if (nAlign == 10) nAlign = 2;	//align tempreal's mod 2
			assert(IS_POWER_OF_TWO(nAlign));
            assert(p->am = amSTACK);    //register vars not yet implemented
            cbStackFrameOffset -= size;
			cbStackFrameOffset &= -nAlign;	//align to "native" size
            p->offset = cbStackFrameOffset;
            if (cbStackFrameOffset < -(MAX_LOCALS*4)) {		//is this still necessary???
                // too many locals...
                ErrorOk("Too many local (auto) variables: %s", name);
                //Mde_fatal(MD386__DICTOVFL, 0, 0, 0);
            }
        }
    }//fmVAR
	else {
		// fmFUNCTION, fmLABEL, etc.
		ErrorOk("Bug: Symbol type not yet handled in AfhFMsymtab.");
	}
    return p;
} /*enter_symbol*/


pfm_OBJ lookup_symbol(const char *name)
{
    int h = fmHash(name);
    pfm_SCOPE scope = top_scope;
    while (scope) {
        if (scope->hb) {
            pfm_OBJ p;
            for (p = (*scope->hb)[h]; p != NULL; p = p->pList) {
                if (strcmp(name, p->name) == 0)
                    return p;   //found it
            }
        } //if
        scope = scope->pOuterScope;
    } //while
    return NULL;    //not found
} /*lookup_symbol*/


pfm_OBJ lookup_member(const char *name, pfm_OBJ pSym)
{
    pfm_OBJ pList;
    for (pList = pSym->pList; pList != NULL; pList = pList->pList) {	//isn't pList the hash chain???
        if (strcmp(name, pList->name) == 0) break;
    } //while
    return pList;
} /*lookup_member*/


pfm_OBJ enter_member(const char *name, pfm_OBJ pSym)
{
    pfm_OBJ p, pTrail;
    //check for duplicate member name
    pTrail = pSym;
    for (p = pSym->pList; p != NULL; p = p->pList) {			//isn't pList the hash chain???		
        if (name != NULL && strcmp(name, p->name) == 0) {
            //Mde_warn(MD386_DUPMEMB, (UINT4)name, 0, 0);
            Mde_warn(MD386__DUPSYM, (UINT_PTR)name, 0, 0); //for now
        } //if
        pTrail = p;
    } //for
    p = newOBJ(fmMEMBER);
    p->name = name; //for now
    p->basetype = btUNDEF;
    p->subtype = stUNDEF;
    p->am = amUNDEF;
#if 1
    //Add new member to tail of list, regardless
    p->pList = NULL;
    pTrail->pList = p;
#else
    //Add new member to head of list, regardless
    p->pList = pSym->pList;
    pSym->pList = p;
#endif
    return p;
} /*enter_member*/


void reset_symbol_table(void)
{
    while (top_scope) {
        leave_scope();
    } //while
} /*reset_symbol_table*/


void init_symbol_table(void)
{
    reset_symbol_table();
    assert(top_scope == NULL);
    cbStackFrameOffset = 0;
    cbMaxStackDepth = 0;
    cbMaxGlobalStack = 0;
    cbStaticFrameOffset = offsetof(FMcontext, globvars.user_vars);
    enter_scope();  //create global (file-level) scope

    //create the fundamental types
    pfmVOID   = create_scalar_type(btVOID,     stVOID,   0, "void",   0);
    pfmBOOL   = create_scalar_type(btINTEGRAL, stBOOL,   1, "bool",   fmfNONNEG);
    pfmCHAR   = create_scalar_type(btINTEGRAL, stCHAR,   1, "char",   0);
    pfmBYTE   = create_scalar_type(btINTEGRAL, stBYTE,   1, "byte",   fmfNONNEG);
    pfmSHORT  = create_scalar_type(btINTEGRAL, stSHORT,  2, "short",  0);
    pfmWORD   = create_scalar_type(btINTEGRAL, stWORD,   2, "word",   fmfNONNEG);
    pfmINT    = create_scalar_type(btINTEGRAL, stINT,    4, "int",    0);
    pfmDWORD  = create_scalar_type(btINTEGRAL, stDWORD,  4, "dword",  fmfNONNEG);
    pfmLONG   = create_scalar_type(btINTEGRAL, stLONG,   4, "long",   0);
#ifdef X64
    pfmINT64  = create_scalar_type(btINTEGRAL, stINT64,  8, "int64",  0);
    pfmQWORD  = create_scalar_type(btINTEGRAL, stQWORD,  8, "qword",  fmfNONNEG);
	if (X64GEN) {	//INT_PTR types are 8 bytes
		pfmINT_PTR  = create_scalar_type(btINTEGRAL, stINT_PTR,  8, "INT_PTR",  0);
		pfmUINT_PTR = create_scalar_type(btINTEGRAL, stUINT_PTR, 8, "UINT_PTR", fmfNONNEG);
	} else
	{// X32GEN		INT_PTR types are 4 bytes
		pfmINT_PTR  = create_scalar_type(btINTEGRAL, stINT_PTR,  4, "INT_PTR",  0);
		pfmUINT_PTR = create_scalar_type(btINTEGRAL, stUINT_PTR, 4, "UINT_PTR", fmfNONNEG);
	}
#endif //X64
    pfmHALF   = create_scalar_type(btREAL,     stHALF,   2, "half",		0);
    pfmFLOAT  = create_scalar_type(btREAL,     stFLOAT,  4, "float",	0);
    pfmDOUBLE = create_scalar_type(btREAL,     stDOUBLE, 8, "double",   0);
    pfmTREAL  = create_scalar_type(btREAL,     stTREAL, 10, "tempreal", 0);

#ifndef DONT_USE_PREDEFINED_CONST_TYPES
	//We also need some predefined 'const' types.
    pfmCONST_VOID   = create_scalar_type(btVOID,     stVOID,   0, "void",   fmfCONST);
    pfmCONST_BOOL   = create_scalar_type(btINTEGRAL, stBOOL,   1, "bool",   fmfCONST|fmfNONNEG);
    pfmCONST_CHAR   = create_scalar_type(btINTEGRAL, stCHAR,   1, "char",   fmfCONST);
    pfmCONST_BYTE   = create_scalar_type(btINTEGRAL, stBYTE,   1, "byte",   fmfCONST|fmfNONNEG);
    pfmCONST_SHORT  = create_scalar_type(btINTEGRAL, stSHORT,  2, "short",  fmfCONST);
    pfmCONST_WORD   = create_scalar_type(btINTEGRAL, stWORD,   2, "word",   fmfCONST|fmfNONNEG);
	pfmCONST_INT	= create_scalar_type(btINTEGRAL, stINT,    4, "int",    fmfCONST);
    pfmCONST_DWORD  = create_scalar_type(btINTEGRAL, stDWORD,  4, "dword",  fmfCONST|fmfNONNEG);
    pfmCONST_LONG   = create_scalar_type(btINTEGRAL, stLONG,   4, "long",   fmfCONST);
#ifdef X64
    pfmCONST_INT64  = create_scalar_type(btINTEGRAL, stINT64,  8, "int64",  fmfCONST);
    pfmCONST_QWORD  = create_scalar_type(btINTEGRAL, stQWORD,  8, "qword",  fmfCONST|fmfNONNEG);
	if (X64GEN) {	//INT_PTR types are 8 bytes
		pfmCONST_INT_PTR  = create_scalar_type(btINTEGRAL, stINT_PTR,  8, "INT_PTR",  fmfCONST);
		pfmCONST_UINT_PTR = create_scalar_type(btINTEGRAL, stUINT_PTR, 8, "UINT_PTR", fmfCONST|fmfNONNEG);
	} else
	{// X32GEN		INT_PTR types are 4 bytes
		pfmCONST_INT_PTR  = create_scalar_type(btINTEGRAL, stINT_PTR,  4, "INT_PTR",  fmfCONST);
		pfmCONST_UINT_PTR = create_scalar_type(btINTEGRAL, stUINT_PTR, 4, "UINT_PTR", fmfCONST|fmfNONNEG);
	}
#endif //X64
    pfmCONST_HALF   = create_scalar_type(btREAL,     stHALF,   2, "half",	  fmfCONST);
    pfmCONST_FLOAT  = create_scalar_type(btREAL,     stFLOAT,  4, "float",	  fmfCONST);
    pfmCONST_DOUBLE = create_scalar_type(btREAL,     stDOUBLE, 8, "double",   fmfCONST);
    pfmCONST_TREAL  = create_scalar_type(btREAL,     stTREAL, 10, "tempreal", fmfCONST);
#endif //CONST_TYPES

} /*init_symbol_table*/


pfm_OBJ newUNDEFtype(void)
{
    return create_scalar_type(btUNDEF, stUNDEF,  0, "undefined",   0);
} /*newUNDEFtype*/


pfm_OBJ compose_type(pfm_OBJ typeObj, int type, int dim)
{
    pfm_OBJ t = typeObj;
    switch (type) {
    case SDTYPE(DCL_TYPEDEF):
        t->flags |= fmfTYPEDEF;
        break;
    case SDTYPE(DCL_EXTERN):
        t->flags |= fmfEXTERN;
        break;
    case SDTYPE(DCL_STATIC):
        if (t->flags & fmfAUTO) {
            ErrorOk("Storage class 'static' conflicts with storage class 'auto'.");
        }
        else if (t->flags & fmfREGISTER) {
            ErrorOk("Storage class 'static' conflicts with storage class 'register'.");
        }
        else {
            t->flags |= fmfSTATIC;
        }
        break;
    case SDTYPE(DCL_AUTO):
        if (t->flags & fmfSTATIC) {
            ErrorOk("Storage class 'auto' conflicts with storage class 'static'.");
        }
        else if (t->flags & fmfREGISTER) {
            ErrorOk("Storage class 'auto' conflicts with storage class 'register'.");
        }
        else {
            t->flags |= fmfAUTO;
        }
        break;
    case SDTYPE(DCL_REGISTER):
        if (t->flags & fmfSTATIC) {
            ErrorOk("Storage class 'register' conflicts with storage class 'static'.");
        }
        else if (t->flags & fmfAUTO) {
            ErrorOk("Storage class 'register' conflicts with storage class 'auto'.");
        }
        else {
            t->flags |= fmfREGISTER;
        }
        break;
    case SDTYPE(DCL_SIGNED):
        t->flags |= fmfSIGNED;
        break;
    case SDTYPE(DCL_UNSIGNED):
        t->flags |= fmfUNSIGNED;
		t->flags |= fmfNONNEG;
        break;
    case SDTYPE(DCL_SATURATED):
        t->flags |= fmfSATURATED;
        break;
    case SDTYPE(DCL_CONST):
        t->flags |= fmfCONST;
        break;
    case SDTYPE(DCL_VOLATILE):
        t->flags |= fmfVOLATILE;
        break;
    case SDTYPE(DCL_VOID):
        if (t->subtype == stUNDEF) {
            t->subtype = stVOID;
        }
        else if (t->subtype != stVOID) {
            ErrorOk("Type specifier 'void' conflicts with other type specifiers.");
        }
        break;
    case SDTYPE(DCL_CHAR):
        if (t->subtype == stUNDEF) {
            t->subtype = stCHAR;
        }
        else if (t->subtype != stCHAR) {
            ErrorOk("Type specifier 'char' conflicts with other type specifiers.");
        }
        break;
    case SDTYPE(DCL_SHORT):
        t->flags |= fmfSHORT;
        break;
    case SDTYPE(DCL_INT):
        if (t->subtype == stUNDEF) {
            t->subtype = stINT;
        }
        else if (t->subtype != stINT) {
            ErrorOk("Type specifier 'int' conflicts with other type specifiers.");
        }
        break;
    case SDTYPE(DCL_LONG):
#ifdef X64
		if (t->subtype == stUNDEF &&
			(t->flags & fmfLONG)) {
			// 2 occurrences of 'long' ==> int64.
			t->subtype = stINT64;
			t->flags &= ~fmfLONG;
		}
		else
#endif //X64
		{
			t->flags |= fmfLONG;
		}
        break;
    case SDTYPE(DCL_BOOL):
        if (t->subtype == stUNDEF) {
            t->subtype = stBOOL;
			t->flags |= fmfNONNEG;	//redundant to finalize_type??
        }
        else if (t->subtype != stBOOL) {
            ErrorOk("Type specifier 'bool' conflicts with other type specifiers.");
        }
        break;
    case SDTYPE(DCL_BYTE):
        if (t->subtype == stUNDEF) {
            t->subtype = stBYTE;
 			t->flags |= fmfNONNEG;
        }
        else if (t->subtype != stBYTE) {
            ErrorOk("Type specifier 'byte' conflicts with other type specifiers.");
        }
        break;
    case SDTYPE(DCL_WORD):
        if (t->subtype == stUNDEF) {
            t->subtype = stWORD;
			t->flags |= fmfNONNEG;
        }
        else if (t->subtype != stWORD) {
            ErrorOk("Type specifier 'word' conflicts with other type specifiers.");
        }
        break;
    case SDTYPE(DCL_DWORD):
        if (t->subtype == stUNDEF) {
            t->subtype = stDWORD;
			t->flags |= fmfNONNEG;
        }
        else if (t->subtype != stDWORD) {
            ErrorOk("Type specifier 'dword' conflicts with other type specifiers.");
        }
        break;
#ifdef X64
	case SDTYPE(DCL_INT64):
		if (t->subtype == stUNDEF) {
			t->subtype = stINT64;
		}
		else if (t->subtype != stINT64) {
			ErrorOk("Type specifier 'int64' conflicts with other type specifiers.");
		}
		break;
    case SDTYPE(DCL_QWORD):
        if (t->subtype == stUNDEF) {
            t->subtype = stQWORD;
			t->flags |= fmfNONNEG;
        }
        else if (t->subtype != stQWORD) {
            ErrorOk("Type specifier 'qword' conflicts with other type specifiers.");
        }
        break;
    case SDTYPE(DCL_INT_PTR):
        if (t->subtype == stUNDEF) {
            t->subtype = stINT_PTR;
        }
        else if (t->subtype != stINT_PTR) {
            ErrorOk("Type specifier 'INT_PTR' conflicts with other type specifiers.");
        }
        break;
    case SDTYPE(DCL_UINT_PTR):
        if (t->subtype == stUNDEF) {
            t->subtype = stUINT_PTR;
			t->flags |= fmfNONNEG;
        }
        else if (t->subtype != stUINT_PTR) {
            ErrorOk("Type specifier 'UINT_PTR' conflicts with other type specifiers.");
        }
        break;
#endif //X64
    case SDTYPE(DCL_HALF):
        if (t->subtype == stUNDEF) {
            t->subtype = stHALF;
        }
        else if (t->subtype != stHALF) {
            ErrorOk("Type specifier 'half' conflicts with other type specifiers.");
        }
        break;
    case SDTYPE(DCL_FLOAT):
        if (t->subtype == stUNDEF) {
            t->subtype = stFLOAT;
        }
        else if (t->subtype != stFLOAT) {
            ErrorOk("Type specifier 'float' conflicts with other type specifiers.");
        }
        break;
    case SDTYPE(DCL_DOUBLE):
        if (t->subtype == stUNDEF) {
            t->subtype = stDOUBLE;
        }
        else if (t->subtype != stDOUBLE) {
            ErrorOk("Type specifier 'double' conflicts with other type specifiers.");
        }
        break;
    case SDTYPE(DCL_TEMPREAL):
        if (t->subtype == stUNDEF) {
            t->subtype = stTREAL;
        }
        else if (t->subtype != stTREAL) {
            ErrorOk("Type specifier 'tempreal' conflicts with other type specifiers.");
        }
        break;
    case SDTYPE(DCL_STRUCT):
        ErrorOk("'struct' is not yet implemented");
        break;
    case SDTYPE(DCL_UNION):
        ErrorOk("'union' is not yet implemented");
        break;
    case SDTYPE(DCL_ENUM):
        ErrorOk("'enum' is not yet implemented");
        break;
    case SDTYPE(DCL_POINTER):
        finalize_type(typeObj);
        t = newOBJ(fmPOINTER);
        t->basetype = btUNDEF;
        t->subtype = stUNDEF;
        t->am = amUNDEF;
        t->name = "pointer";
        t->pType = typeObj;
        t->pList = NULL;
        t->flags = typeObj->flags & fmfSTORAGE_CLASS_FLAGS; //propagate storage class flags
#if 0
        //would need to duplicate the type object itself when duplicating an SDL entry
        //containing a pointer to a type object (OP_DUPL) if this is to work...
        //for now, it's probably harmless to have leftover storage class
        //flags in type objects ... just ignore them!!
        typeObj->flags &= ~fmfSTORAGE_CLASS_FLAGS;
#endif
        t->nCount = 0;
        t->offset = 0;
		t->flags |= fmfNONNEG;
#ifdef X64
		if (X64GEN) {
			//pointer is 64-bits for X64 -- is this the right place???
			t->size = 8;
		}
		else
#endif //X64
		{
	        t->size = 4;
		}
        break;
    case SDTYPE(DCL_ARRAYDIM):
        finalize_type(typeObj);
        t = newOBJ(fmARRAY);
        t->basetype = btUNDEF;
        t->subtype = stUNDEF;
        t->am = amUNDEF;
        t->name = "array";
        t->pType = typeObj;
        t->pList = NULL;
        t->flags = typeObj->flags & fmfSTORAGE_CLASS_FLAGS; //propagate storage class flags
#if 0
        //would need to duplicate the type object itself when duplicating an SDL entry
        //containing a pointer to a type object (OP_DUPL) if this is to work...
        //for now, it's probably harmless to have leftover storage class
        //flags in type objects ... just ignore them!!
        typeObj->flags &= ~fmfSTORAGE_CLASS_FLAGS;
#endif
        t->nCount = dim;
        t->offset = 0;
        t->size = dim * typeObj->size;	//X64 -- need 64-bit math!!!
        break;
    case SDTYPE(DCL_ARRAYDIMNUL):
        finalize_type(typeObj);
        t = newOBJ(fmARRAY);
        t->basetype = btUNDEF;
        t->subtype = stUNDEF;
        t->am = amUNDEF;
        t->name = "open array";
        t->pType = typeObj;
        t->pList = NULL;
        t->flags = typeObj->flags & fmfSTORAGE_CLASS_FLAGS; //propagate storage class flags
#if 0
        //would need to duplicate the type object itself when duplicating an SDL entry
        //containing a pointer to a type object (OP_DUPL) if this is to work...
        //for now, it's probably harmless to have leftover storage class
        //flags in type objects ... just ignore them!!
        typeObj->flags &= ~fmfSTORAGE_CLASS_FLAGS;
#endif
        t->flags |= fmfOPENDIM;
        t->nCount = 0;
        t->offset = 0;
        t->size = 0;
        break;
    case SDTYPE(DCL_PARAM):
        ErrorOk("Function parameters are not yet implemented");
        break;
    case SDTYPE(DCL_MEMBER):
        ErrorOk("List members are not yet implemented");
        break;
    case SDTYPE(DCL_FUNCTION):
        ErrorOk("Function declarations are not yet implemented");
        break;
    default:
        ErrorOk("Bug: Unknown declaration specifier in compose_type.");
        break;
    } //switch
    return t;
} /*compose_type*/


pfm_OBJ finalize_type(pfm_OBJ type)
{
    // Appply final defaults and consistency checks.

    switch (type->what) {
    case fmSCALAR:
        // Check for consistent SCALAR attributes...
        if ((type->flags & fmfSIGNED) && (type->flags & fmfUNSIGNED)) {
            Warn("Type specifier 'signed' conflicts with 'unsigned'.");
            type->flags &= ~fmfSIGNED;
        }
        if ((type->flags & fmfSIGNED) && (type->flags & fmfNONNEG)) {
            Warn("Type specifier 'signed' cannot be applied to an unsigned type.");
            type->flags &= ~fmfSIGNED;
        }
        if ((type->flags & fmfSHORT) && (type->flags & fmfLONG)) {
            Warn("Type specifier 'short' conflicts with 'long'.");
            type->flags &= ~fmfSHORT;
        }
        if ((type->flags & fmfCONST) && (type->flags & fmfVOLATILE)) {
            Warn("Type specifier 'const' conflicts with 'volatile'.");
            type->flags &= ~fmfCONST;
        }
		//// should be other conficts: short/int64, byte/int,...  !!!

        // Compute final scalar subtype...
        if (type->subtype == stUNDEF) {
            if (type->flags & fmfUNSIGNED) {
                if (type->flags & fmfSHORT) {
					// unsigned short = word
                    type->subtype = stWORD;
                }
				else if (type->flags & fmfLONG) {
					// unsigned long = dword
					type->subtype = stDWORD;
				}
#if 0
				/// ditto.... ???? these aren't defined!!
				else if (type->flags & fmfINT) {
					// unsigned int = dword
					type->subtype = stDWORD;
				}
				else if (type->flags & fmfINT64) {
					// unsigned int64 = qword
					type->subtype = stQWORD;
				}
				else if (type->flags & fmfCHAR) {
					// unsigned char = byte
					type->subtype = stBYTE;
				}
				else if (type->flags & fmfINT_PTR) {
					// unsigned INT_PTR = UINT_PTR
					type->subtype = stUINT_PTR;
				}
				///  ditto....??? and more below???
#endif
                else {
					// unsigned = dword
                    type->subtype = stDWORD;
                }
            }
            else {
                ///type->flags |= fmfSIGNED;
                if (type->flags & fmfSHORT) {
					// [signed] short = short
                    type->subtype = stSHORT;
                }
                else if (type->flags & fmfLONG) {
					// [signed] long = long
                    type->subtype = stLONG;
                }
                else {
					// [signed] = int
                    type->subtype = stINT;
                }
            }
        }
        else {			//following are okay???
            if (type->flags & fmfLONG) {
                if (type->subtype == stINT) {
					// long int = long
                    type->subtype = stLONG;
                }
                else if (type->subtype == stDOUBLE) {
					// long double = tempreal
                    type->subtype = stTREAL;
                }
#if 0
                else {
#else
                else if (type->subtype != stLONG &&
                         type->subtype != stTREAL) {
#endif
                    Warn("Type specifier 'long' conflicts with other type specifiers.");
                    type->flags &= ~fmfLONG;
                }
            }
            if (type->flags & fmfSHORT) {
                if (type->subtype == stINT) {
					// short int = short
                    type->subtype = stSHORT;
                }
                else if (type->subtype == stDWORD) {
					// short dword = word ??
                    type->subtype = stWORD;
                }
#if 0
                else {
#else
                else if (type->subtype != stSHORT &&
                         type->subtype != stWORD) {
#endif
                    Warn("Type specifier 'short' conflicts with other type specifiers.");
                    type->flags &= ~fmfSHORT;
                }
            }
        }

        //Compute base type, size, name...
        switch (type->subtype) {
        case stVOID:
            type->basetype = btVOID;
            type->size = 0;
            type->name = "void";
            break;
        case stBOOL:
            type->basetype = btINTEGRAL;
			type->flags |= fmfNONNEG;
            type->size = 1;					//or 4 for MS compat???
            type->name = "bool";
            break;
        case stCHAR:
            type->basetype = btINTEGRAL;
            type->size = 1;
            type->name = "char";
            break;
        case stBYTE:
            type->basetype = btINTEGRAL;
			type->flags |= fmfNONNEG;
            type->size = 1;
            type->name = "byte";
            break;
        case stSHORT:
            type->basetype = btINTEGRAL;
            type->size = 2;
            type->name = "short";
            break;
        case stWORD:
            type->basetype = btINTEGRAL;
			type->flags |= fmfNONNEG;
            type->size = 2;
            type->name = "word";
            break;
        case stINT:
            type->basetype = btINTEGRAL;
            type->size = 4;
            type->name = "int";
            break;
        case stDWORD:
            type->basetype = btINTEGRAL;
			type->flags |= fmfNONNEG;
            type->size = 4;
            type->name = "dword";
            break;
        case stLONG:
            type->basetype = btINTEGRAL;
            type->size = 4;
            type->name = "long";
			break;
#ifdef X64
		case stINT64:
			type->basetype = btINTEGRAL;
			type->size = 8;
			type->name = "int64";
			break;
		case stQWORD:
			type->basetype = btINTEGRAL;
			type->flags |= fmfNONNEG;
			type->size = 8;
			type->name = "qword";
			break;
		case stINT_PTR:
			type->basetype = btINTEGRAL;
			type->size = X64GEN ? 8 : 4;
			type->name = "INT_PTR";
			break;
		case stUINT_PTR:
			type->basetype = btINTEGRAL;
			type->flags |= fmfNONNEG;
			type->size = X64GEN ? 8 : 4;
			type->name = "UINT_PTR";
			break;
#endif //X64
        case stHALF:
            type->basetype = btREAL;
            type->size = 2;
            type->name = "half";
            break;
        case stFLOAT:
            type->basetype = btREAL;
            type->size = 4;
            type->name = "float";
            break;
        case stDOUBLE:
            type->basetype = btREAL;
            type->size = 8;
            type->name = "double";
            break;
        case stTREAL:
            type->basetype = btREAL;
            type->size = 10;
            type->name = "tempreal";
            break;
        default:
            ErrorOk("Bug: Unknown subtype in finalize_type.");
            break;
        } //switch subtype
        break; //end case fmSCALAR

    case fmENUM:
    case fmFUNCTION:
        if (type->flags & fmfCONST) {
            Warn("Type specifier 'const' is not valid in this declaration.");
            type->flags &= ~fmfCONST;
        }
        if (type->flags & fmfVOLATILE) {
            Warn("Type specifier 'volatile' is not valid in this declaration.");
            type->flags &= ~fmfVOLATILE;
        }
        if (type->flags & fmfAUTO) {
            Warn("Type specifier 'auto' is not valid in this declaration.");
            type->flags &= ~fmfAUTO;
        }
        //fall through...
    case fmARRAY:
    case fmSTRUCT:
    case fmUNION:
        if (type->flags & fmfREGISTER) {
            Warn("Type specifier 'register' is not valid in this declaration.");
            type->flags &= ~fmfREGISTER;
        }
        //fall through...
    case fmPOINTER:
        if (type->flags & fmfSIGNED) {
            Warn("Type specifier 'signed' is not valid in this declaration.");
            type->flags &= ~fmfSIGNED;
        }
        if (type->flags & fmfUNSIGNED) {
            Warn("Type specifier 'unsigned' is not valid in this declaration.");
            type->flags &= ~fmfUNSIGNED;
        }
        if (type->flags & fmfSHORT) {
            Warn("Type specifier 'short' is not valid in this declaration.");
            type->flags &= ~fmfSHORT;
        }
        if (type->flags & fmfLONG) {
            Warn("Type specifier 'long' is not valid in this declaration.");
            type->flags &= ~fmfLONG;
        }
        if (type->flags & fmfSATURATED) {
            Warn("Type specifier 'saturated' is not valid in this declaration.");
            type->flags &= ~fmfSATURATED;
        }
        break;

    default:
        ErrorOk("Bug: Unhandled object kind in finalize_type.");
        break;
    } //switch kind

    return type;
} /*finalize_type*/


#ifndef BETA

void display_flags(int flags)
{
    if (flags & fmfNONNEG) PRINTF("nonnegative ");	///later display as 'unsigned "??
    if (flags & fmfSIGNED) PRINTF("signed ");
    if (flags & fmfUNSIGNED) PRINTF("unsigned ");
    if (flags & fmfSHORT) PRINTF("short ");
    if (flags & fmfLONG) PRINTF("long ");
    if (flags & fmfSATURATED) PRINTF("saturated ");
    if (flags & fmfCONST) PRINTF("const ");
    if (flags & fmfVOLATILE) PRINTF("volatile ");
    if (flags & fmfTYPEDEF) PRINTF("typedef ");
    if (flags & fmfEXTERN) PRINTF("extern ");
    if (flags & fmfSTATIC) PRINTF("static ");
    if (flags & fmfAUTO) PRINTF("auto ");
    if (flags & fmfREGISTER) PRINTF("register ");
    if (flags & fmfSTACK) PRINTF("stack ");
    if (flags & fmfHEAP) PRINTF("heap ");
    //if (flags & fmfOPENDIM) PRINTF("opendim ");
    if (flags & fmfGLOBAL) PRINTF("global ");
} /*display_flags*/


void display_name(const char *name)
{
    if (name && *name) PRINTF2("\'%s\' ", name);
} /*display_name*/

void display_size(int size)
{
#ifndef DONT_DISPLAY_SIZE
	PRINTF2("size(%d) ", size);
#endif
} /*display_size*/


void display_list(pfm_OBJ list)
{
    PRINTF("{");	//or PRINTF("("); ???
    if (list) {
        assert(list->what == fmMEMBER);	//could be a function param???
        display_OBJ(list);
		list = list->pList;
        while (list) {
            PRINTF(", ");	//or PRINTF("; "); ???
            assert(list->what == fmMEMBER);
            display_OBJ(list);
			list = list->pList;
        }//while
    }//if
    PRINTF("} ");	//or PRINTF(" )"); ???
} /*display_list*/


void display_OBJ(pfm_OBJ t)
{
    switch (t->what) {
    case fmUNDEF:
        display_flags(t->flags);
        PRINTF("undefined ");
        display_name(t->name);
        break;
    case fmSCALAR:
        display_flags(t->flags);
        PRINTF("scalar ");
        switch (t->basetype) {
        case btUNDEF:
            PRINTF("undefined ");
            break;
        case btVOID:
            PRINTF("void ");
            break;
        case btINTEGRAL:
            PRINTF("integral ");
            break;
        case btREAL:
            PRINTF("real ");
            break;
        default:
            PRINTF("unrecognized ");
            break;
        }//switch basetype
        switch (t->subtype) {
        case stUNDEF:
            PRINTF("undefined ");
            break;
        case stVOID:
            PRINTF("void ");
            break;
        case stBOOL:
            PRINTF("bool ");
            break;
        case stCHAR:
            PRINTF("char ");
            break;
        case stBYTE:
            PRINTF("byte ");
            break;
        case stSHORT:
            PRINTF("short ");
            break;
        case stWORD:
            PRINTF("word ");
            break;
        case stINT:
            PRINTF("int ");
            break;
        case stDWORD:
            PRINTF("dword ");
            break;
        case stLONG:
            PRINTF("long ");
            break;
#ifdef X64
		case stINT64:
			PRINTF("int64 ");
			break;
		case stQWORD:
			PRINTF("qword ");
			break;
		case stINT_PTR:
			PRINTF("INT_PTR ");
			break;
		case stUINT_PTR:
			PRINTF("UINT_PTR ");
			break;
#endif //X64
        case stHALF:
            PRINTF("half ");
            break;
        case stFLOAT:
            PRINTF("float ");
            break;
        case stDOUBLE:
            PRINTF("double ");
            break;
        case stTREAL:
            PRINTF("tempreal ");
            break;
        default:
            PRINTF("unrecognized ");
            break;
        }//switch subtype
        display_name(t->name);
		display_size(t->size);
        break;
    case fmPOINTER:
        display_flags(t->flags);
        PRINTF("pointer ");
		display_size(t->size);
        display_name(t->name);
        PRINTF("to ");
        display_OBJ(t->pType);
        break;
    case fmARRAY:
        display_flags(t->flags);
        PRINTF("array ");
        display_name(t->name);
        PRINTF("[");
        if ((t->flags & fmfOPENDIM) == 0) {
            PRINTF2("%d", t->nCount);
        }
        PRINTF("] "); 
		display_size(t->size);
		PRINTF("of ");
        display_OBJ(t->pType);
        break;
    case fmSTRUCT:
        display_flags(t->flags);
        PRINTF("struct ");
        display_name(t->name);
		display_size(t->size);
        display_list(t->pList);
        break;
    case fmUNION:
        display_flags(t->flags);
        PRINTF("union ");
        display_name(t->name);
		display_size(t->size);
        display_list(t->pList);
        break;
    case fmENUM:
        display_flags(t->flags);
        PRINTF("enum ");
        display_name(t->name);
		display_size(t->size);
        display_list(t->pList);
        break;
    case fmFUNCTION:
        display_flags(t->flags);
        display_OBJ(t->pType);
        PRINTF("function ");
        display_name(t->name);
        display_list(t->pList);
        break;
    case fmTYPEDEF:
        display_flags(t->flags);
        PRINTF("typedef ");
        display_name(t->name);
		display_size(t->size);
        PRINTF("is ");
        display_OBJ(t->pType);
        break;
    case fmVAR:
        display_flags(t->flags);
        PRINTF("variable ");
        display_name(t->name);
		display_size(t->size);
        PRINTF("at ");
        switch (t->am) {
        case amUNDEF:
            PRINTF2("(*undefined*) offset %+d ", t->offset);
            break;
        case amSTATIC:
#ifdef X64
			if (X64GEN) {
				PRINTF2("[rbx%+d] ", t->offset);
			}
			else
#endif //X64
			{
				PRINTF2("[ebx%+d] ", t->offset);
			}
            break;
        case amSTACK:
#ifdef X64
			if (X64GEN) {
				PRINTF2("[rbp%+d] ", t->offset);
			}
			else
#endif //X64
			{
				PRINTF2("[ebp%+d] ", t->offset);
			}
            break;
        case amHEAP:
            PRINTF2("[%x] (heap) ", t->offset);
            break;
        case amBASED:
            PRINTF2("[%x] (based) ", t->offset);
            break;
        case amREG:
            PRINTF2("register %d ", t->offset);
            break;
        default:
            PRINTF2("BUG am=%d ", t->am);
            break;
        }//switch am
        PRINTF("is ");
        display_OBJ(t->pType);
        break;
    case fmMEMBER:
        display_flags(t->flags);
        display_OBJ(t->pType);
        display_name(t->name);
        break;
    case fmLABEL:
        display_flags(t->flags);
        PRINTF("label ");
        display_name(t->name);
        break;
    default:
        display_flags(t->flags);
        PRINTF2("unrecognized obj (%d) ", t->what);
        display_name(t->name);
        break;
    } //switch what
} /*display_OBJ*/

#endif //ifndef BETA