/*
 *   EmitX86.h
 *
 *   Copyright 1997-2014 AFH Systems
 *
 *   FilterMeister is free software. You can redistribute and modify it
 *   under the terms of the MPL2 License.
 *   See: https://www.mozilla.org/en-US/MPL/2.0/
 *   Any changes to the FilterMeister files have to be published or sent
 *   to harry@thepluginsite.com. The filter plugin code itself does not
 *   need to be disclosed.
 *
 *	05JUL2013	A. Hunter	1. For X64, define rip = eip = ip = 16, for use with
 *							   rip-relative addressing.
 *  14JUL2013	A. Hunter	1. Changes for WIN64.
 *	25JUL2013	A. Hunter	1. Add OC1M, OC2M, and OC2 macros.
 *	01AUG2013	A. Hunter	1. Add FCOMI* and FUCOMI* opcodes (P6+ only).
 *	18AUG2013	A. Hunter	1. Add SSE3 opcodes for FISTTP (truncate to int and store/pop).
 *	20AUG2013	A. Hunter	1. Sort out various encodings of MOVD/MOVQ.
 *	04SEP2013	A. Hunter	1. Add MOVSS/MOVSSR, MOVSDX/MOVSDXR, CVT*, CVTT*
 *  05SEP2013	A. Hunter	1. Add FCMOVcc opcodes (P6+ only).
 *  10SEP2013	A. Hunter	1. Add MOVAPS and MOVAPD opcodes (SSE2).
 *  13SEP2013	A. Hunter	1. Add GPR_* macros for gpr display names.
 *							2. Add MOV*X* alternate mnemonics.
 *							3. Treat FWAIT (9Bh) as an opcode prefix, so other prefixes
 *							   such as REX will follow, not precede, the FWAIT.
 *	16SEP2013	A. Hunter	1. Add ASIZE(op) macro.
 *							2. Use more #ifndef X64ONLY sections.
 *	17SEP2013	A. Hunter	1. Add several more MOVZ* (zero-extend) and MOVS* (sign-
 *							   extend) opcodes.
 *							2. Add an nLocals parameter to the X86_emit_*_epilog routines.
 *	30Sep2013	A. Hunter	1. Make func struct local to CB.
 *	26Jan2014	A. Hunter	1. Make TMP macro work correctly for X64.
 *	23Feb2014	A. Hunter	1. Add prototype for X86_abort2.
 *  24Feb2014	A. Hunter	1. Add prototype for X86_error2.
 *	23Mar2014   A. Hunter	1. Add (undocumented) FFREEP instruction.
 *	28May2014	A. Hunter	1. Add X86_emit_op_TiR.
 *  13Jul2014	A. Hunter	1. Add MOVI_Q and MOVI_Q_REG.
 *
 *  To do:
 *      - add op_W variants
 *      - add corresponding X86_emit_*_simm16 (and _imm16?) routines
 *      - implement CMOVcc_W vs. CMOVcc_D
 *      - synthesize CMOVcc_B ?
 *      - add SSE2 opcodes and XMM registers (also update the disassembler!!!)
 *      - allow (limited) chaining of disp8 fixups (should be within reach afterall??)
 *      - special (load-time or run-time) fixups to replace, e.g., CMOVcc with synthesized
 *          sequence when cpu doesn't support CMOV (BSWAP, etc.)
 *      - synthesize BSWAP_REG on 386 (any other cpu dependencies?)
 *      - add branch hint prefixes (HNT)
 *      - distinguish MOV ES, AX vs. MOV ES, EAX (latter has no operand override, so faster)
 *		- add undocumented(?) SALC opcode??
 *	X64 to do:
 *		- encode RIP-relative addressing
 *		- encode reference to new byte regs (SPL etc.)
 *		- automatic ndp stack tracking???
 */

#ifndef EMITX86_H
#define EMITX86_H 1

#ifdef X64
#define PREDEFSYMOFF(sym) (X64gen?offsetof(FMcontext64, sym):offsetof(FMcontext32, sym))
#else
#define PREDEFSYMOFF(sym) offsetof(FMcontext, sym)
#endif

#define FMC(x) ebx, no_reg, 0, PREDEFSYMOFF(x)
#define FMCX(v, xreg, scale) ebx, xreg, scale, PREDEFSYMOFF(v)
#define TMPSIZE (X64gen?8:4)		//sizeof(temp)=8 if X64, else 4
#define TMP(n) ebp, no_reg, 0, TMPSIZE*(n)
#define MEM(n) int base_reg##n, int index_reg##n, int scaleFactor##n, int disp##n
#define MEM1 MEM(1)
#define MEM2 MEM(2)
#define mem(n) base_reg##n, index_reg##n, scaleFactor##n, disp##n
#define mem1 mem(1)
#define mem2 mem(2)

// Encoding opcode mnemonics for EmitX86/X64:
// An opcode mnemonic has a 32-bit value.
// The high order byte (byte 3) contains option bits for generating opcode prefixes:
//
// x86/x64 prefix byte definitions.
#define REX		0x40	// the bare REX prefix byte (for SPL etc.)
#define REX_W	0x08	// REX.W flag => operand size = 64 bits
#define REX_R	0x04	// REX.R flag => use extended GPR or XMM in reg field
#define REX_X	0x02	// REX.X flag => use extended GPR or XMM in index reg field
#define REX_B	0x01	// REX.B flag => use extended GPR or XMM in base reg field
#define REX_D	0x10	// (16-bit) data size override (0x66)
#define REX_A	0x20	// (32-bit) address size override (0x67)	-- NEEDED???
#define REP_N	0x01	// REPNE/REPNZ (0xF2) prefix, also used by SSE+ ops		*** or use these 2 bits
#define REP_E	0x02	// REPE/REPZ/REP (0xF3) prefix, also used by SSE+ ops	*** for ndp stack tracking??
#define LOCK_PB	0x04	// LOCK (0xF0) prefix			-- NEEDED???
#define WAIT_PB	0x80	// FWAIT(0x9B) prefix (actually the WAIT op, but treated like a pefix)
#if 0 //NYI
#define GS_PB	0x??	// GS:(0x65) prefix -- NYI
#define FS_PB	0x??	// FS:(0x64) prefix -- NYI
#endif //NYI
// Note 0x80 is now used by WAIT, and is NO LONGER available for reuse for other legacy prefixes,
// e.g. FS (0x64), GS (0x65),...
// Also, do we actually need REX_A or LOCK_PB??? REP_N or REP_E???

// Byte 0 contains the opcode value for a 1-byte opcode.
// A 2-byte opcode value is in byte 1 and byte 0, with the
// bytes reversed from the actual value in memory, so byte 1
// is emitted *before* byte 0.
// In general, a 3-byte opcode (e.g. some SSE2 ops) is in bytes 2-0,
// and is emitted with byte 2 first.
// It is assumed that the first opcode byte to be emitted is nonzero
// (unless it is the only byte).  For example:
//		0x00002A emits 0x2A
//		0x00DD3B emits 0xDD 0x3B
//		0x660F6E emits 0x66 0x0F 0x6E
// etc.
//
// Sometimes an opcode with two variants is encoded with the main
// 1-byte variant in byte 0, and the alternate 1- or 2-byte variant
// in bytes 2-1.

// Some handy macros for parsing/defining 4-byte opcode values:

#define PF(op)	 ((op) & 0xFF000000)	//Prefix part of opcode
#define NP(op)	 ((op) & 0x00FFFFFF)	//Non-Prefix part of opcode
#define OC1M(op) ((op) & 0x000000FF)	//1st opcode mask
#define OC2M(op) ((op) & 0x00FFFF00)	//2nd opcode mask
#define OC1(op)	 ((op) & 0xFF0000FF)	//1st opcode with prefixes preserved
#define OC2(op)  (PF(op)|(NP(op) >> 8))	//2nd opcode with prefixes preserved

// In general, QWORD opcodes can be formed by setting the REX.W bit in the
// high-order byte of the corresponding DWORD opcode encoding.
#define Q(op)	((op)|(REX_W << 24))

// In general, WORD opcodes can be formed by setting the D16 bit in the
// high-order byte of the corresponding DWORD opcode encoding.
// Also used to modify some media opcodes (e.g. XMOVD).
#define W(op)	((op)|(REX_D << 24))

// The default address size in X64 mode is 64 bits. This can be overridden
// to 32 bits by setting the REX_A bit in the opcode definition, which
// causes an Address-Size prefix byte (0x67) to be emitted.  This can be 
// used to force opcodes such as JrCXZ and LOOPcc to test the CX register
// or ECX register instead of the default register (ECX in 32-bit mode,
// RCX in X64 mode).
#define ASIZE(op) ((op)|(REX_A << 4))

// A macro to automatically select the 32-bit or 64-bit encoding of an op,
// depending on the setting of the X64gen flag.
#define V(op)	(X64gen?Q(op):(op))

// A macro to force a bare REX prefix if needed,
// for extended low byte registers (SPL etc.)
#define XL(op)	((op)|(REX << 24))

// A macro to emit a 0xF2 prefix for certain sse+ opcodes
#define F2(op) ((op)|(REP_N << 24))
// A macro to emit a 0xF3 prefix for certain sse+ opcodes
#define F3(op) ((op)|(REP_E << 24))
// A macro to emit a REPNE/REPNZ prefix for a string instruction
#define REPN_OP(op) ((op)|(REP_N << 24))
// A macro to emit a REPE/REPZ/REP prefix for a string instruction
#define REPE_OP(op) ((op)|(REP_E << 24))
// A macro to emit a REP prefix for a string instruction
#define REP_OP(op) ((op)|(REP_E << 24))
// A macro to emit a LOCK prefix for a lockable instruction
#define LOCKED(op) ((op)|(LOCK_PB << 24))
// A macro to emit an FWAIT prefix for a floating point instruction
#define WAIT_OP(op) ((op)|(WAIT_PB << 24))
#if 0	//NYI
// A macro to emit a GS: segment register override prefix
#define GS(op) ((op)|(GS_PB << 24))
// A macro to emit an FS: segment register override prefix
#define FS(op) ((op)|(FS_PB << 24))
#endif //NYI

// Since the EmitX86/X64 assembly code syntax does not specify the type
// of an instruction's operands, we add the following suffixes to the
// EmitX84/X64 opcode mnemonics, in addition to the usual Intel suffixes
// (such as condition code names, 'ZX' for zero-extend, 'SX' for sign-
// extend, etc.) to specify the explicit operand type, et al.
//
// By convention the EmitX86/X64 mnemonic suffixes indicate:
//	I		Immediate operand (or Indirect for calls/jumps)
//	R		Reversed operands (e.g. 'ADDR reg,mem' to add reg to mem)
// (N.B. Some redundant "reversed" aliases may be Invalid for X64???)
//	X		(as prefix) XMM operand1; (as suffix) MMX or XMM operand2
//	0		Implicit 0 operand (e.g. RET0_N)
//	1		1-bit shift
//	_B		Byte (8-bit) op
//	_W		Word (16-bit) op
//	_D		Dword (32-bit) op
//	_Q		Qword (64-bit) op
//	_O		Oword (128-bit) op
//	_V		A chameleon op, which automatically changes operand size
//			from _D to _Q (and maybe _W) depending on the current code
//			generation mode (X86gen, X64gen, or maybe X16gen, resp.)
//	_REG	REGister-only op
//	_SHORT	8-bit jump displacement
//	_N		Near (offset-only) op
//	_F		Far (offset+selector) op
//	_CS, _DS, _ES, _FS, _GS, _SS	Implicit selector name in op
//
// For X87 mnemonics, the operand types are indicated by the
// following suffixes:
//  _F		Float (32-bit real)
//  _D      Double (64-bit real)
//  _T      Tbyte or Tempreal (80-bit real)
//  _S      Short (16-bit integer)
//  _I      Integer (32-bit integer)
//  _L      Long (64-bit integer)
//  _B      BCD (80-bit packed decimal)
//  _STN    operands are ST,ST(n)
//  _ST0    operands are ST(n),ST
//	_AX		Implicit register AX operand

// General Purpose Register display names.
#define GPR_Q(reg) X86_reg64_name(reg)		//64-bit (QWORD) register name
#define GPR_D(reg) X86_reg32_name(reg)		//32-bit (DWORD) register name
#define GPR_W(reg) X86_reg16_name(reg)		//16-bit (WORD) register name
#define GPR_BH(reg) X86_reg8H_name(reg)		//8-bit (BYTE) High/Low register name (x86)
#define GPR_BL(reg) X86_reg8L_name(reg)		//8-bit (BYTE) orthogonal register name (x64)
#define GPR_V(reg) X86_regv_name(reg)		//GPR_Q if X64gen, GPR_D if X32gen, else GPR_W
#define MMX(reg) X86_mmxreg_name(reg)		//64-bit MMX register name
#define XMM(reg) X86_xmmreg_name(reg)		//128-bit XMM register name
#define YMM(reg) X86_ymmreg_name(reg)		//256-bit YMM register name
#define STN(reg) X86_stnreg_name(reg)		//80-bit ndp stack register name


//Opcodes for op_M_imm8, op_T_imm8, op_R_imm8
#define ADDI_B      0x8000
#define ORI_B       0x8001
#define ADCI_B      0x8002
#define SBBI_B      0x8003
#define ANDI_B      0x8004
#define SUBI_B      0x8005
#define XORI_B      0x8006
#define CMPI_B      0x8007

#define ROLI_B      0xC000
#define RORI_B      0xC001
#define RCLI_B      0xC002
#define RCRI_B      0xC003
#define SHLI_B      0xC004
#define SHRI_B      0xC005
#define SALI_B      SHLI_B
#define SARI_B      0xC007

#define ROLI_D      0xC100
#define RORI_D      0xC101
#define RCLI_D      0xC102
#define RCRI_D      0xC103
#define SHLI_D      0xC104
#define SHRI_D      0xC105
#define SALI_D      SHLI_D
#define SARI_D      0xC107

#define MOVI_B      0xC600
#define TESTI_B     0xF600

#define BTI         0x0FBA04
#define BTSI        0x0FBA05
#define BTRI        0x0FBA06
#define BTCI        0x0FBA07

//Opcodes for op_imm8
#define MOVI_B_REG  0xB0
#define INTI_B      0xCD    //('INT' conflicts with windef.h)

//Opcodes for op_imm16
#define RET_N       0xC2
#ifndef X64ONLY
#define RET_F       0xCA
#endif //!X64ONLY

//Opcodes for op_imm32
#define MOVI_D_REG  0xB8			//zero-extends imm32 to reg64
#define MOVI_Q_REG	Q(MOVI_D_REG)	//moves imm64 to reg64

//Opcodes for op_simm32
#define PUSHI_D     0x68

//Opcodes for op_M_imm32, op_T_imm32, op_R_imm32
#define MOVI_D      0xC700		//zero-extends imm32 to reg64; moves imm32 to mem32
#define MOVI_Q      Q(MOVI_D)	//sign-extends imm32 to reg/mem64
#define TESTI_D     0xF700

//Opcodes for op_MR_imm8, op_TR_imm8, op_RR_imm8
#define SHLDI_D     0x0FA4
#define SHRDI_D     0x0FAC

//Opcodes for op_RM_simm32, op_RT_simm32, op_RR_simm32,
//            op_MM_simm32, op_TT_simm32
#define IMULI_D     0x69        //can also use with OP_M_simm32, etc.

//Opcodes for op_M_simm32, op_T_simm32, op_R_simm32
#define ADDI_D      0x8100
#define ORI_D       0x8101
#define ADCI_D      0x8102
#define SBBI_D      0x8103
#define ANDI_D      0x8104
#define SUBI_D      0x8105
#define XORI_D      0x8106
#define CMPI_D      0x8107

//Opcodes for op_M, op_T, op_R
#define POP_D       0x8F00		//X64 automatically promotes to POP_Q

#define ROL1_B      0xD000
#define ROR1_B      0xD001
#define RCL1_B      0xD002
#define RCR1_B      0xD003
#define SHL1_B      0xD004
#define SHR1_B      0xD005
#define SAL1_B      SHL1_B
#define SAR1_B      0xD007

#define ROL1_D      0xD100
#define ROR1_D      0xD101
#define RCL1_D      0xD102
#define RCR1_D      0xD103
#define SHL1_D      0xD104
#define SHR1_D      0xD105
#define SAL1_D      SHL1_D
#define SAR1_D      0xD107

#define ROL_B       0xD200
#define ROR_B       0xD201
#define RCL_B       0xD202
#define RCR_B       0xD203
#define SHL_B       0xD204
#define SHR_B       0xD205
#define SAL_B       SHL_B
#define SAR_B       0xD207

#define ROL_D       0xD300
#define ROR_D       0xD301
#define RCL_D       0xD302
#define RCR_D       0xD303
#define SHL_D       0xD304
#define SHR_D       0xD305
#define SAL_D       SHL_D
#define SAR_D       0xD307

#define NOT_B       0xF602
#define NEG_B       0xF603
#define MUL_B       0xF604
#define IMUL_B      0xF605	//misnamed -- should be IMUL_AL???
#define DIV_B       0xF606
#define IDIV_B      0xF607

#define NOT_D       0xF702
#define NEG_D       0xF703
#define MUL_D       0xF704
#define IMUL_D      0xF705	//misnamed -- should be IMUL_EAX???
#define DIV_D       0xF706
#define IDIV_D      0xF707

#define INC_B       0xFE00
#define DEC_B       0xFE01
#define INC_D       0xFF00
#define DEC_D       0xFF01
#define CALLI_N     0xFF02      //call indirect near
#ifndef X64ONLY
#define CALLI_F     0xFF03      //call indirect far
#define JMPI_F      0xFF05      //jump indirect far
#endif //!X64ONLY
#define JMPI_N      0xFF04      //jump indirect near
#define PUSH_D      0xFF06		//X64 automatically promotes to PUSH_Q

#define SLDT        0x0F0000
#define STR         0x0F0001
#define LLDT        0x0F0002
#define LTR         0x0F0003
#define VERR        0x0F0004
#define VERW        0x0F0005
#define SGDT        0x0F0100
#define SIDT        0x0F0101
#define LGDT        0x0F0102
#define LIDT        0x0F0103
#define SMSW        0x0F0104
#define LMSW        0x0F0106

#define SETO_B      0x0F9000
#define SETNO_B     0x0F9100
#define SETB_B      0x0F9200
#define SETNB_B     0x0F9300
#define SETZ_B      0x0F9400
#define SETNZ_B     0x0F9500
#define SETBE_B     0x0F9600
#define SETNBE_B    0x0F9700
#define SETS_B      0x0F9800
#define SETNS_B     0x0F9900
#define SETP_B      0x0F9A00
#define SETNP_B     0x0F9B00
#define SETL_B      0x0F9C00
#define SETNL_B     0x0F9D00
#define SETLE_B     0x0F9E00
#define SETNLE_B    0x0F9F00

#ifndef NOALTMNEMS
#define SETA_B      SETNBE_B
#define SETAE_B     SETNB_B
#define SETC_B      SETB_B
#define SETE_B      SETZ_B
#define SETG_B      SETNLE_B
#define SETGE_B     SETNL_B
#define SETNA_B     SETBE_B
#define SETNAE_B    SETB_B
#define SETNC_B     SETNB_B
#define SETNE_B     SETNZ_B
#define SETNG_B     SETLE_B
#define SETNGE_B    SETL_B
#define SETPE_B     SETP_B
#define SETPO_B     SETNP_B
#endif

// (following are synthetic...)
#define SETO_D      0x0F9001
#define SETNO_D     0x0F9101
#define SETB_D      0x0F9201
#define SETNB_D     0x0F9301
#define SETZ_D      0x0F9401
#define SETNZ_D     0x0F9501
#define SETBE_D     0x0F9601
#define SETNBE_D    0x0F9701
#define SETS_D      0x0F9801
#define SETNS_D     0x0F9901
#define SETP_D      0x0F9A01
#define SETNP_D     0x0F9B01
#define SETL_D      0x0F9C01
#define SETNL_D     0x0F9D01
#define SETLE_D     0x0F9E01
#define SETNLE_D    0x0F9F01

#ifndef NOALTMNEMS
#define SETA_D      SETNBE_D
#define SETAE_D     SETNB_D
#define SETC_D      SETB_D
#define SETE_D      SETZ_D
#define SETG_D      SETNLE_D
#define SETGE_D     SETNL_D
#define SETNA_D     SETBE_D
#define SETNAE_D    SETB_D
#define SETNC_D     SETNB_D
#define SETNE_D     SETNZ_D
#define SETNG_D     SETLE_D
#define SETNGE_D    SETL_D
#define SETPE_D     SETP_D
#define SETPO_D     SETNP_D
#endif

//Opcodes for op_RM, OP_MR, OP_RR, OP_RT, OP_TR, etc.
#define ADDR_B      0x00		// =ADD_B - 0x02
#define ADDR_D      0x01		// =ADD_D - 0x02
#define ADD_B       0x02
#define ADD_D       0x03

#define ORR_B       0x08		// =OR_B - 0x02
#define ORR_D       0x09		// =OR_D - 0x02
#define OR_B        0x0A
#define OR_D        0x0B

#define ADCR_B      0x10		// =ADC_B - 0x02
#define ADCR_D      0x11		// =ADC_D - 0x02
#define ADC_B       0x12
#define ADC_D       0x13

#define SBBR_B      0x18		// =SBB_B - 0x02
#define SBBR_D      0x19		// =SBB_D - 0x02
#define SBB_B       0x1A
#define SBB_D       0x1B

#define ANDR_B      0x20		// =AND_B - 0x02
#define ANDR_D      0x21		// =AND_D - 0x02
#define AND_B       0x22
#define AND_D       0x23

#define SUBR_B      0x28		// =SUB_B - 0x02
#define SUBR_D      0x29		// =SUB_D - 0x02
#define SUB_B       0x2A
#define SUB_D       0x2B

#define XORR_B      0x30		// =XOR_B - 0x02
#define XORR_D      0x31		// =XOR_D - 0x02
#define XOR_B       0x32
#define XOR_D       0x33

#define CMPR_B      0x38		// =CMP_B - 0x02
#define CMPR_D      0x39		// =CMP_D - 0x02
#define CMP_B       0x3A
#define CMP_D       0x3B

#ifndef X64ONLY
#define BOUND       0x62
#define ARPL        0x63
#endif //X64ONLY

#define TEST_B      0x84
#define TEST_D      0x85

#define XCHG_B      0x86
#define XCHG_D      0x87

#define MOVR_B      0x88		// =MOV_B - 0x02
#define MOVR_D      0x89		// =MOV_D - 0x02
#define MOV_B       0x8A
#define MOV_D       0x8B

// 64-bit (QWORD) opcodes
#ifdef X64
#define LEA_Q		Q(LEA)
#define ADDR_Q		Q(ADDR_D)
#define ADD_Q		Q(ADD_D)
#define ORR_Q		Q(ORR_D)
#define OR_Q		Q(OR_D)
#define ADCR_Q		Q(ADCR_D)
#define ADC_Q		Q(ADC_D)
#define SBBR_Q		Q(SBBR_D)
#define SBB_Q		Q(SBB_D)
#define ANDR_Q		Q(ANDR_D)
#define AND_Q		Q(AND_D)
#define SUBR_Q		Q(SUBR_D)
#define SUB_Q		Q(SUB_D)
#define XORR_Q		Q(XORR_D)
#define XOR_Q		Q(XOR_D)
#define CMPR_Q		Q(CMPR_D)
#define CMP_Q		Q(CMP_D)
#define MOVR_Q		Q(MOVR_D)
#define MOV_Q		Q(MOV_D)
#define XCHG_Q		Q(XCHG_D)
#define TEST_Q		Q(TEST_D)
#define INC_Q		Q(INC_D)
#define DEC_Q		Q(DEC_D)
#define IMUL32_Q	Q(IMUL32)	//misnamed -- should be IMUL_Q and IMUL_D,resp.???
///...etc...
#endif //X64

// 16-bit (WORD) opcodes
#ifndef NO_W_OPS
#define ADDR_W		W(ADDR_D)
#define ADD_W		W(ADD_D)
#define ORR_W		W(ORR_D)
#define OR_W		W(OR_D)
#define ADCR_W		W(ADCR_D)
#define ADC_W		W(ADC_D)
#define SBBR_W		W(SBBR_D)
#define SBB_W		W(SBB_D)
#define ANDR_W		W(ANDR_D)
#define AND_W		W(AND_D)
#define SUBR_W		W(SUBR_D)
#define SUB_W		W(SUB_D)
#define XORR_W		W(XORR_D)
#define XOR_W		W(XOR_D)
#define CMPR_W		W(CMPR_D)
#define CMP_W		W(CMP_D)
#define MOVR_W		W(MOVR_D)
#define MOV_W		W(MOV_D)
#define XCHG_W		W(XCHG_D)
#define TEST_W		W(TEST_D)
#define INC_W		W(INC_D)
#define DEC_W		W(DEC_D)
#define IMUL32_W	W(IMUL32)	//misnamed -- should be IMUL_W and IMUL_D,resp.???
///...etc...
#endif //NO_W_OPS

#define LEA         0x8D

#ifndef X64ONLY
#define LES         0xC4
#define LDS         0xC5
#endif //!X64ONLY

#define LAR         0x0F02
#define LSL         0x0F03

#define SHLD        0x0FA5
#define SHRD        0x0FAD

#define IMUL32      0x0FAF	//misnamed -- should be IMUL_D???

#define LSS         0x0FB2
#define LFS         0x0FB4
#define LGS         0x0FB5

#define MOVZB_D     0x0FB6	//Intel mnemonic = MOVZX (see MOVZXB_D alternate below)
#define MOVZW_D     0X0FB7	//Intel mnemonic = MOVZX (see MOVZXW_D alternate below)
#define MOVZB_W		W(MOVZB_D)	//D16 override
#ifdef X64
#define MOVZB_Q		MOVZB_D	//REX.W not needed since MOVZB_D zero-extends result to 64 bits
#define MOVZW_Q		MOVZW_D	//REX.W not needed since MOVZW_D zero-extends result to 64 bits
#define	MOVZD_Q		MOV_D	//Synthetic version (MOV_D zero-extends to 64-bits) (see MOVZXD_Q alternate below)
#endif //X64

#define BSF_D       0x0FBC
#define BSR_D       0x0FBD

#define MOVSB_D     0x0FBE	//Intel mnemonic = MOVSX (see MOVSXB_D alternate below)
#define MOVSB_W		W(MOVSB_D)	//D16 override; sign-extends int8 to int16
#define MOVSW_D     0x0FBF	//Intel mnemonic = MOVSX (see MOVSXW_D alternate below)
#ifdef X64
#define MOVSB_Q		Q(MOVSB_D)	//REX.W override; sign-extends int8 to int64
#define MOVSW_Q		Q(MOVSW_D)	//REX.W override; sign-extends int16 to int64
//AMD mnemonic = MOVSXD (usurps ARPL in 64-bit mode)
#define MOVSD_D		0x63		//MOVSD_D actually zero-extends dword to qword!!!(but MOV_D is just as good) (use MOVSXD_D as alternate??)
#define MOVSD_Q		Q(MOVSD_D)	//MOVSD_Q sign-extends dword to qword -- useful!! (see MOVSXD_Q alternate below)
#endif //X64

#ifndef NOALTMNEMS
#define	MOVZXB_D	MOVZB_D     //AMD/Intel alternate mnemonic
#define	MOVZXW_D	MOVZW_D     //AMD/Intel alternate mnemonic
#define MOVSXB_W	MOVSB_W     //AMD/Intel alternate mnemonic
#define	MOVSXB_D	MOVSB_D     //AMD/Intel alternate mnemonic
#define	MOVSXW_D	MOVSW_D     //AMD/Intel alternate mnemonic
///#define MOVSXD_D	MOVSD_D		//AMD/Intel alternate mnemonic //MOVSD_D has same effect as
#define MOVSXD_D	MOV_D		//AMD/Intel alternate mnemonic //MOV_D, but MOV_D may be faster
#ifdef X64
#define MOVZXB_Q	MOVZB_Q		//FM alternate mnemonic
#define MOVZXW_Q	MOVZW_Q		//FM alternate mnemonic
#define	MOVZXD_Q	MOVZD_Q		//FM alternate mnemonic
#define	MOVSXB_Q	MOVSB_Q     //AMD/Intel alternate mnemonic
#define	MOVSXW_Q	MOVSW_Q     //AMD/Intel alternate mnemonic
#define MOVSXD_Q	MOVSD_Q		//AMD/Intel alternate mnemonic
#endif //X64
#endif //NOALTMNEMS

#ifndef NOMMX
//MMX opcodes for use with op_RM, OP_MR, OP_RR, etc.

//where 'r' in 'r/m' is a GPR...
#define MOVD        0x0F6E		//MMX	movd mmx,r32/m32
#define MOVDR       0x0F7E		//MMX	movd r32/m32,mmx	=MOVD | 0x10
#define MOVQ		Q(MOVD)		//MMX	movq mmx,r64/m64
#define MOVQR		Q(MOVDR)	//MMX	movq r64/m64,mmx	=MOVQ | 0x10
//where 'r' in 'r/m' is an MMX register...
#define MOVQX		0x0F6F		//MMX	movq mmx,mmx/m64
#define MOVQXR		0x0F7F		//MMX	movq mmx/m64,mmx	=MOVQX | 0x10

#define PACKSSWB    0x0F63  //MMX
#define PACKSSDW    0x0F6B  //MMX
#define PACKUSWB    0x0F67  //MMX

#define PADDB       0x0FFC  //MMX
#define PADDW       0x0FFD  //MMX
#define PADDD       0x0FFE  //MMX

#define PSUBB       0x0FF8  //MMX
#define PSUBW       0x0FF9  //MMX
#define PSUBD       0x0FFA  //MMX

#define PADDSB      0x0FEC  //MMX
#define PADDSW      0x0FED  //MMX

#define PSUBSB      0x0FE8  //MMX
#define PSUBSW      0x0FE9  //MMX

#define PADDUSB     0x0FDC  //MMX
#define PADDUSW     0x0FDD  //MMX

#define PSUBUSB     0x0FD8  //MMX
#define PSUBUSW     0x0FD9  //MMX

#define PAND        0x0FDB  //MMX
#define PANDN       0x0FDF  //MMX
#define POR         0x0FEB  //MMX
#define PXOR        0x0FEF  //MMX

#define PCMPGTB     0x0F64  //MMX
#define PCMPGTW     0x0F65  //MMX
#define PCMPGTD     0x0F66  //MMX

#define PCMPEQB     0x0F74  //MMX
#define PCMPEQW     0x0F75  //MMX
#define PCMPEQD     0x0F76  //MMX

#define PMADDWD     0x0FF5  //MMX
#define PMULHW      0x0FE5  //MMX
#define PMULLW      0x0FD5  //MMX

#define PSLLW       0x0FF1  //MMX
#define PSLLD       0x0FF2  //MMX
#define PSLLQ       0x0FF3  //MMX

#define PSRAW       0x0FE1  //MMX
#define PSRAD       0x0FE2  //MMX

#define PSRLW       0x0FD1  //MMX
#define PSRLD       0x0FD2  //MMX
#define PSRLQ       0x0FD3  //MMX

#define PUNPCKHBW   0x0F68  //MMX
#define PUNPCKHWD   0x0F69  //MMX
#define PUNPCKHDQ   0x0F6A  //MMX

#define PUNPCKLBW   0x0F60  //MMX
#define PUNPCKLWD   0x0F61  //MMX
#define PUNPCKLDQ   0x0F62  //MMX

//MMX opcodes for use with op_R_imm8
#define PSLLWI      0x0F7106  //MMX
#define PSLLDI      0x0F7206  //MMX
#define PSLLQI      0x0F7306  //MMX

#define PSRAWI      0x0F7104  //MMX
#define PSRADI      0x0F7204  //MMX

#define PSRLWI      0x0F7102  //MMX
#define PSRLDI      0x0F7202  //MMX
#define PSRLQI      0x0F7302  //MMX

#endif //!NOMMX

#ifndef NOSSE2
#ifndef NOXMM
    /**************************************+++++********/
    /* Opcodes for SSE1/2 move/conversion instructions */
    /***************************************************/
//XMM (SSE1/2) opcodes for use with op_RM, OP_MR, OP_RR, etc.

//where 'r' in 'r/m' is a GPR...
#define XMOVD		W(MOVD)		//SSE2	movd xmm,r32/m32
#define XMOVDR      W(MOVDR)	//SSE2	movd r32/m32,xmm	=XMOVD | 0x10
#define XMOVQ		W(MOVQ)		//SSE2	movq xmm,r64/m64
#define XMOVQR		W(MOVQR)	//SSE2	movq r64/m64,xmm	=XMOVQ | 0x10
//where 'r' in 'r/m' is an XMM register...
#define XMOVQX		F3(MOVDR)	//SSE2	movq xmm,xmm/m64	[sic!]
#define XMOVQXR		W(0x0FD6)	//SSE2	movq xmm/m64,xmm	[sic!]
#define MOVSDX		F2(0x0F10)	//SSE2	movsdx xmm,xmm/m64	MSVC uses this, AMD uses MOVSD
#define MOVSDXR		F2(0x0F11)	//SSE2	movsdx xmm/m64,xmm	=MOVSDX | 0x01
#define MOVSS		F3(0x0F10)	//SSE1	movss xmm,xmm/m32
#define MOVSSR		F3(0x0F11)	//SSE1	movss xmm/m32,xmm	=MOVSS | 0x01
#define MOVUPD		W(0x0F10)	//SSE2	movupd xmm,xmm/m128
#define MOVUPDR		W(0x0F11)	//SSE2	movupd xmm/m128,xmm	=MOVUPD | 0x01
#define MOVUPS		0x0F10		//SSE1	movups xmm,xmm/m128
#define MOVUPSR		0x0F11		//SSE1	movups xmm/m128,xmm	=MOVUPS | 0x01
#define MOVAPD		W(0x0F28)	//SSE2	movapd xmm,xmm/m128
#define MOVAPDR		W(0x0F29)	//SSE2	movapd xmm/m128,xmm	=MOVAPD | 0x01
#define MOVAPS		0x0F28		//SSE1	movaps xmm,xmm/m128
#define MOVAPSR		0x0F29		//SSE1	movaps xmm/m128,xmm	=MOVAPS | 0x01
//conversions...
#define CVTSD2SI	F2(0X0F2D)	//SSE2	double to int32		Q(CVTSD2SI) for int64
#define CVTSD2SS	F2(0x0F5A)	//SSE2	double to float
#define CVTSI2SD	F2(0x0F2A)	//SSE2	int32 to double		Q(CVTSI2SD) for int64
#define CVTSI2SS	F3(0x0F2A)	//SSE1	int32 to float		Q(CVTSI2SS) for int64
#define CVTSS2SD	F3(0x0F5A)	//SSE2	float to double
#define CVTSS2SI	F3(0x0F2D)	//SSE1	float to int32		Q(CVTSS2SI) for int64
//truncate to integer
#define	CVTTSD2SI	F2(0x0F2C)	//SSE2	double to int32 chopped	Q(CVTTSD2SI) for int64
#define CVTTSS2SI	F3(0x0F2C)	//SSE1	float to int32 chopped	Q(CVTTSS2SI) for int64

#endif //!NOXMM
#endif //!NOSSE2

#ifndef NOCMOV
//CMOVcc implemented in Pentium Pro and up...
//Opcodes for op_RM, OP_MR, OP_RR, OP_RT, OP_TR, etc.
#define CMOVO       0x0F40
#define CMOVNO      0x0F41
#define CMOVB       0x0F42
#define CMOVNB      0x0F43
#define CMOVZ       0x0F44
#define CMOVNZ      0x0F45
#define CMOVBE      0x0F46
#define CMOVNBE     0x0F47
#define CMOVS       0x0F48
#define CMOVNS      0x0F49
#define CMOVP       0x0F4A
#define CMOVNP      0x0F4B
#define CMOVL       0x0F4C
#define CMOVNL      0x0F4D
#define CMOVLE      0x0F4E
#define CMOVNLE     0x0F4F

#ifndef NOALTMNEMS
#define CMOVA       CMOVNBE
#define CMOVAE      CMOVNB
#define CMOVC       CMOVB
#define CMOVE       CMOVZ
#define CMOVG       CMOVNLE
#define CMOVGE      CMOVNL
#define CMOVNA      CMOVBE
#define CMOVNAE     CMOVB
#define CMOVNC      CMOVNB
#define CMOVNE      CMOVNZ
#define CMOVNG      CMOVLE
#define CMOVNGE     CMOVL
#define CMOVPE      CMOVP
#define CMOVPO      CMOVNP
#endif //!NOALTMNEMS
#endif //!NOCMOV

//1-byte opcodes
#define PUSH_ES     0x06
#define POP_ES      0x07
#define PUSH_CS     0x0E
#define PUSH_SS     0x16
#define POP_SS      0x17
#define PUSH_DS     0x1E
#define POP_DS      0x1F
#ifndef X64ONLY
#define ES_         0x26    //segment override prefix
#define CS_         0x2E    //segment override prefix
#define SS_         0x36    //segment override prefix
#define DS_         0x3E    //segment override prefix
#define DAA         0x27
#define DAS         0x2F
#define AAA         0x37
#define AAS         0x3F
#define INC_REG     0x40	// X64 usurps INC_REG and DEC_REG for REX codes.
#define DEC_REG     0x48
#define PUSHA       0x60
#define POPA        0x61
#endif //!X64ONLY
#define PUSH_REG    0x50
#define POP_REG     0x58
#define FS_         0x64    //segment override prefix
#define GS_         0x65    //segment override prefix
#define D16         0x66    //operand size override prefix
#define A32         0x67    //address size override prefix
#define INS_B       0x6C
#define INS_D       0x6D
#define OUTS_B      0x6E
#define OUTS_D      0x6F
#define NOP         0x90    //=XCHG EAX,EAX
#define XCHG_REG    0x90
#define CBW         W(0x98)	//ax=sext(al)
#define CWDE        0x98	//eax=sext(ax)
#define	CDQE		Q(0x98)	//rax=sext(eax)
#define CWD         W(0x99) //dx:ax=sext(ax)
#define CDQ         0x99	//edx:eax=sext(eax)
#define	CQO			Q(0x99)	//rdx:rax=sext(rax)
#define WAIT        0x9B
#define PUSHF       0x9C
#define POPF        0x9D
#define SAHF        0x9E
#define LAHF        0x9F
#define MOVS_B      0xA4
#define MOVS_D      0xA5
#define CMPS_B      0xA6
#define CMPS_D      0xA7
#define STOS_B      0xAA
#define STOS_D      0xAB
#define LODS_B      0xAC
#define LODS_D      0xAD
#define SCAS_B      0xAE
#define SCAS_D      0xAF
#define RET0_N      0xC3
#define LEAVE       0xC9
#ifndef X64ONLY
#define RET0_F      0xCB
#define INTO        0xCE
#endif
#define INT3        0xCC
#define IRET        0xCF
#define IRETD       0xCF
#ifdef X64
#define IRETQ       0xCF
#endif //X64
#define XLAT        0xD7
#define ESC         0xD8
#define LOCK        0xF0
#define REPNE       0xF2
#define REPE        0xF3
#define REP         0xF3
#define HLT         0xF4
#define CMC         0xF5
#define CLC         0xF8
#define STC         0xF9
#define CLI         0xFA
#define STI         0xFB
#define CLD         0xFC
#define STD         0xFD

//2-byte opcodes
#define CLTS        0x0F06
#define INVD        0x0F08  //486+
#define WBINVD      0x0F09  //486+
#define UD2         0x0F0B  //all??
#define WRMSR       0x0F30  //p5+
#define RDTSC       0x0F31  //p5+
#define RDMSR       0x0F32  //p5+
#define RDPMC       0x0F33  //p5mmx/p6+
#define SYSENTER    0x0F34  //p6.3.3sep+
#define SYSEXIT     0x0F35  //p6.3.3sep+
#define PUSH_FS     0x0FA0
#define POP_FS      0x0FA1
#define CPUID       0x0FA2  //later model 486+ only!!!
#define PUSH_GS     0x0FA8
#define POP_GS      0x0FA9
#define RSM         0x0FAA  //386SL/486SL+
#define BSWAP_REG   0x0FC8  //486+ only!!!
#ifndef X64ONLY
#define AAM         0xD40A
#define AAD         0xD50A
#endif //!X64ONLY

#ifndef NOMMX
#define EMMS        0x0F77  //Pentium MMX and later...
#endif //!NOMMX

//Opcodes for op_lab
//high-order bytes are for disp32 (0 if N/A),
//low-order byte is for disp8 (0 if N/A).
//For LOOPcc and JrCXZ, the register tested
//(CX, ECX, or RCX) depends on the address size
//(16, 32, or 64 bits, respectively).
#define LOOPNZ      0xE0
#define LOOPZ       0xE1
#define LOOP        0xE2
#ifndef X64ONLY
#define JCXZ        ASIZE(0xE3)
#endif //!X64ONLY
#ifdef X64
#define JECXZ       ASIZE(0xE3)
#define JRCXZ       0xE3
#else //X86
#define JECXZ       0xE3
#endif //!X64
#define CALL        0xE800      //no disp8 form
#define JMP         0xE9EB
#define JMP_SHORT   0xEB        //force disp8 form

// No short-form-only for the following?
// How about the following macro to force the short form.
// Then JMP_SHORT above is SHORT(JMP), and short-form only
// of JO is SHORT(JO) for example.
// Or just use X86_emit_op_lab_disp8() to force use of short form.

#define SHORT(jmp) ((jmp) & 0xff)

#define JO          0x0F8070
#define JNO         0x0F8171
#define JB          0x0F8272
#define JNB         0x0F8373
#define JZ          0x0F8474
#define JNZ         0x0F8575
#define JBE         0x0F8676
#define JNBE        0x0F8777
#define JS          0x0F8878
#define JNS         0x0F8979
#define JP          0x0F8A7A
#define JNP         0x0F8B7B
#define JL          0x0F8C7C
#define JNL         0x0F8D7D
#define JLE         0x0F8E7E
#define JNLE        0x0F8F7F

#ifndef NOALTMNEMS
#define JA       JNBE
#define JAE      JNB
#define JC       JB
#define JE       JZ
#define JG       JNLE
#define JGE      JNL
#define JNA      JBE
#define JNAE     JB
#define JNC      JNB
#define JNE      JNZ
#define JNG      JLE
#define JNGE     JL
#define JPE      JP
#define JPO      JNP
#endif //!NOALTMNEMS

#ifndef NO_FP
    /*******************************************/
    /* Opcodes for Floating-Point instructions */
    /*******************************************/

//operand types:
//  _F          Float (32-bit real)
//  _D          Double (64-bit real)
//  _T          Tempreal (80-bit real)
//  _S          Short (16-bit integer)
//  _I          Integer (32-bit integer)
//  _L          Long (64-bit integer)
//  _B          BCD (80-bit packed decimal)
//  _STN        ST,ST(n)
//  _ST0        ST(n),ST
//	_AX			Implicit register AX

//For use with OP_M, OP_Tm
#define FLD_F       0xD900
#define FLD_I       0xDB00
#define FLD_D       0xDD00
#define FLD_S       0xDF00
#define FLD_L       0xDF05
#define FLD_T       0xDB05
#define FLD_B       0xDF04	/* FBLD bcd */

#define FST_F       0xD902
#define FST_I       0xDB02
#define FST_D       0xDD02	//=0xDDD0+i with opreg(FST_STN)??
#define FST_S       0xDF02

#define FSTP_F      0xD903
#define FSTP_I      0xDB03
#define FSTP_D      0xDD03 //=0xDDD8+i with opreg(FSTP_STN)??
#define FSTP_S      0xDF03
#define FSTP_L      0xDF07
#define FSTP_T      0xDB07
#define FSTP_B      0xDF06	/* FBSTP bcd */

#ifndef NOSSE3
#define FSTTP_S		0xDF01	/* SSE3 Truncate to int16 and store/pop */
#define FSTTP_I		0xDB01	/* SSE3 Truncate to int32 and store/pop */
#define FSTTP_L		0xDD01	/* SSE3 Truncate to int64 and store/pop */
#endif //!NOSSE3

#define FCOM_F      0xD802	/* FCOM float */
#define FCOM_I      0xDA02	/* FICOM int32 */
#define FCOM_D      0xDC02	/* FCOM double */
#define FCOM_S      0xDE02	/* FICOM int16 */

#define FCOMP_F     0xD803	/* FCOMP float */
#define FCOMP_I     0xDA03	/* FICOMP int32 */
#define FCOMP_D     0xDC03	/* FCOMP double */
#define FCOMP_S     0xDE03	/* FICOMP int16 */

#define FADD_F      0xD800
#define FADD_I      0xDA00
#define FADD_D      0xDC00
#define FADD_S      0xDE00

#define FSUB_F      0xD804
#define FSUB_I      0xDA04
#define FSUB_D      0xDC04
#define FSUB_S      0xDE04

#define FSUBR_F     0xD805
#define FSUBR_I     0xDA05
#define FSUBR_D     0xDC05
#define FSUBR_S     0xDE05

#define FMUL_F      0xD801
#define FMUL_I      0xDA01
#define FMUL_D      0xDC01
#define FMUL_S      0xDE01

#define FDIV_F      0xD806
#define FDIV_I      0xDA06
#define FDIV_D      0xDC06
#define FDIV_S      0xDE06

#define FDIVR_F     0xD807
#define FDIVR_I     0xDA07
#define FDIVR_D     0xDC07
#define FDIVR_S     0xDE07

#define FLDENV      0xD904
#define FLDCW       0xD905
#define FNSTENV     0xD906
#define FSTENV      WAIT_OP(FNSTENV)
#define FNSTCW      0xD907
#define FSTCW       WAIT_OP(FNSTCW)
#define FRSTOR      0xDD04
#define FNSAVE      0xDD06
#define FSAVE       WAIT_OP(FNSAVE)
#define FNSTSW      0xDD07
#define FSTSW       WAIT_OP(FNSTSW)

//For use with OP_R (op2, mod_reg_rm(3, op1, n)) = (op2, 3 << 6 | op1 >> 3 | n)
#define FLD_STN     0xD900
#define FXCH_STN    0xD901
#define FST_STN     0xDD02	/* or 0xDDD0+n for use with op_reg */
#define FSTP_STN    0xDD03	/* or 0xDDD8+n for use with op_reg */
#define FUCOM_STN   0xDD04	/* FUCOM ST(0),ST(n) */
#define FUCOMP_STN  0xDD05	/* FUCOMP ST(0),ST(n) */
#define FUCOMI_STN	0xDB05	/* FUCOMI ST(0),ST(n) -- P6+ only */
#define FUCOMIP_STN	0xDF05	/* FUCOMIP ST(0),ST(n) -- P6+ only */
#define FCOM_STN    0xD802	/* FCOM ST(0),ST(n) */
#define FCOMP_STN   0xD803	/* FCOMP ST(0),ST(n) */
#define FCOMI_STN	0xDB06	/* FCOMI ST(0),ST(n) -- P6+ only */
#define FCOMIP_STN	0xDF06	/* FCOMIP ST(0),ST(n) -- P6+ only */
#define FADD_STN    0xD800
#define FADD_ST0    0xDC00
#define FADDP_ST0   0xDE00
#define FSUB_STN    0xD804
#define FSUB_ST0    0xDC05
#define FSUBP_ST0   0xDE05
#define FSUBR_STN   0xD805
#define FSUBR_ST0   0xDC04
#define FSUBRP_ST0  0xDE04
#define FMUL_STN    0xD801
#define FMUL_ST0    0xDC01
#define FMULP_ST0   0xDE01
#define FDIV_STN    0xD806
#define FDIV_ST0    0xDC07
#define FDIVP_ST0   0xDE07
#define FDIVR_STN   0xD807
#define FDIVR_ST0   0xDC06
#define FDIVRP_ST0  0xDE06
#define FFREE       0xDD00
#ifdef USE_FFREEP
#define FFREEP		0xDFC0	//js FFREEP ST(0) a faster FPOP on some architectures??
#endif

#ifndef NOCMOV
//For use with opreg (op + n)
#define FCMOVB_STN	0xDAC0	/* FCMOVB ST(0),ST(n) -- P6+ only */
#define FCMOVE_STN	0xDAC8	/* FCMOVE ST(0),ST(n) -- P6+ only */
#define FCMOVBE_STN	0xDAD0	/* FCMOVBE ST(0),ST(n) -- P6+ only */
#define FCMOVU_STN	0xDAD8	/* FCMOVU ST(0),ST(n) -- P6+ only */
#define FCMOVNB_STN	0xDBC0	/* FCMOVNB ST(0),ST(n) -- P6+ only */
#define FCMOVNE_STN	0xDBC8	/* FCMOVNE ST(0),ST(n) -- P6+ only */
#define FCMOVNBE_STN 0xDBD0	/* FCMOVNBE ST(0),ST(n) -- P6+ only */
#define FCMOVNU_STN	0xDBD8	/* FCMOVNU ST(0),ST(n) -- P6+ only */

#ifndef NOALTMNEMS
#define FCMOVA_STN		FCMOVNBE_STN
#define FCMOVAE_STN		FCMOVNB_STN
#define FCMOVZ_STN		FCMOVE_STN
#define FCMOVNA_STN		FCMOVBE_STN
#define FCMOVNAE_STN	FCMOVB_STN
#define FCMOVNZ_STN		FCMOVNE_STN
#endif //!NOALTMNEMS
#endif //!NOCMOV

//For use with OP
#define FADD        0xDEC1  /* FADDP ST(1),ST */
#define FSUB        0xDEE9  /* FSUBP ST(1),ST */
#define FSUBR       0xDEE1  /* FSUBRP ST(1),ST */
#define FMUL        0xDEC9  /* FMULP ST(1),ST */
#define FDIV        0xDEF9  /* FDIVP ST(1),ST */
#define FDIVR       0xDEF1  /* FDIVRP ST(1),ST */
#define FXCH        0xD9C9  /* FXCH ST,ST(1) */
#define FCOMI		0xDBF1	/* FCOMI ST,ST(1) -- P6+ only */
#define	FCOMIP		0xDFF1	/* FCOMIP ST,ST(1) -- P6+ only */
#define FCOM        0xD8D1  /* FCOM ST,ST(1) */
#define FCOMP       0xD8D9  /* FCOMP ST,ST(1) */
#define FCOMPP      0xDED9	/* FCOMPP */
#define	FUCOMI		0xDBE9	/* FUCOMI ST,ST(1) -- P6+ only */
#define FUCOMIP		0xDFE9	/* FUCOMIP ST(0),ST(1) -- P6+ only */
#define FUCOM       0xDDE1  /* FUCOM ST,ST(1) */
#define FUCOMP      0xDDE9  /* FUCOMP ST,ST(1) */
#define FUCOMPP     0xDAE9	/* FUCOMPP */
#define FNOP        0xD9D0
#define FCHS        0xD9E0
#define FABS        0xD9E1
#define FTST        0xD9E4
#define FXAM        0xD9E5
#define FLD1        0xD9E8
#define FLDL2T      0xD9E9
#define FLDL2E      0xD9EA
#define FLDPI       0xD9EB
#define FLDLG2      0xD9EC
#define FLDLN2      0xD9ED
#define FLDZ        0xD9EE
#define F2XM1       0xD9F0
#define FYL2X       0xD9F1
#define FPTAN       0xD9F2
#define FPATAN      0xD9F3
#define FXTRACT     0xD9F4
#define FPREM1      0xD9F5
#define FDECSTP     0xD9F6
#define FINCSTP     0xD9F7
#define FPREM       0xD9F8
#define FYL2XP1     0xD9F9
#define FSQRT       0xD9FA
#define FSINCOS     0xD9FB
#define FRNDINT     0xD9FC
#define FSCALE      0xD9FD
#define FSIN        0xD9FE
#define FCOS        0xD9FF
#define FENI        0xDBE0
#define FDISI       0xDBE1
#define FNCLEX      0xDBE2
#define FCLEX       WAIT_OP(FNCLEX)
#define FNINIT      0xDBE3
#define FINIT       WAIT_OP(FNINIT)
#define FNSTSW_AX   0xDFE0
#define FSTSW_AX    WAIT_OP(FNSTSW_AX)
#define FWAIT       WAIT
/* A few handy abbreviations */
#define FPOP        0xDDD8    /* FSTP ST(0) -- is this optimal??? -- it's what VC++ 5.0 uses */
#define FDUP        0xD9C0    /* FLD ST(0) -- duplicate top of stack */
#define FSQR        0xD8C8    /* FMUL ST, ST(0) -- square top of stack */

//alternate (Intel-based) mnemonics
#if 1
#define FBLD        FLD_B
#define FILD_S      FLD_S
#define FILD_I      FLD_I
#define FILD_L      FLD_L
#define FIST_S      FST_S
#define FIST_I      FST_I
#define FISTP_S     FSTP_S
#define FISTP_I     FSTP_I
#define FISTP_L		FSTP_L

#ifndef NOSSE3
#define FISTTP_S    FSTTP_S		/* SSE3 Truncate to int16 and store/pop */
#define FISTTP_I    FSTTP_I		/* SSE3 Truncate to int32 and store/pop */
#define FISTTP_L	FSTTP_L		/* SSE3 Truncate to int64 and store/pop */
#endif //!NOSSE3
#endif //alternates

//illegals -- used for other ops
//#define FADDP_STN   0xDA00
//#define FSUBP_STN   0xDA04
//#define FSUBRP_STN  0xDA05
//#define FMULP_STN   0xDA01
//#define FDIVP_STN   0xDA06
//#define FDIVRP_STN  0xDA07

#endif //!NO_FP

//Miscellaneous defines and typedefs
#ifndef TRUE
typedef int BOOL;
#define TRUE 1
#define FALSE 0
#endif

//enums for X86 and X64 registers
#ifdef X64
typedef enum e_reg64  {            rax=0, rcx,  rdx,   rbx,  rsp,  rbp,  rsi,  rdi
					             , r8,    r9,   r10,   r11,  r12,  r13,  r14,  r15
								 , rip
								 } reg64;
#endif
					//note that (no_reg & 8) must = 0 for x64
typedef enum e_reg32  {no_reg=-16,  eax=0,  ecx,  edx,  ebx,  esp,  ebp,  esi,  edi
#ifdef X64
                                 , r8d,    r9d, r10d, r11d, r12d, r13d, r14d, r15d
								 , eip
#endif
									} reg32;
typedef enum e_reg16  {             ax=0,   cx,   dx,   bx,   sp,   bp,   si,   di
                                 ,  r8w,   r9w, r10w, r11w, r12w, r13w, r14w, r15w
								 ,  ip
									} reg16;
typedef enum e_reg8   {             al=0,   cl,   dl,   bl,   ah,   ch,   dh,   bh
									} reg8;
#ifdef X64
typedef enum e_xreg8  {				                        spl=4, bpl,  sil,  dil
								 ,  r8b,   r9b, r10b, r11b, r12b, r13b, r14b, r15b
									} xreg8;
#endif
typedef enum e_mmxreg {           mmx0=0, mmx1, mmx2, mmx3, mmx4, mmx5, mmx6, mmx7} mmxreg;
typedef enum e_xmmreg {           xmm0=0, xmm1, xmm2, xmm3, xmm4, xmm5, xmm6, xmm7
#ifdef X64
								, xmm8, xmm9, xmm10, xmm11, xmm12, xmm13, xmm14, xmm15
#endif
									} xmmreg;

/* Temp mapping stuff */
#ifdef X64
static reg64 tmpRegID[] = {rdi, rsi, r12, r13, r14, r15}; //"static" is quick kludge for now
//At start of a compilation, nTmps is defined as:
//int nTmps = X64gen?(sizeof(tmpRegID)/sizeof(tmpRegID[0])):2
//We then define isReg as:
#define isReg(tmp) ((tmp) < nTmps)
#define tmpReg(tmp) tmpRegID[tmp]
#else
#define isReg(tmp) ((tmp) < 2)
#define tmpReg(tmp) ((tmp)?esi:edi)			//copacetic with tmpRegID above
#endif

/* Typedefs */
typedef unsigned char BYTE;
typedef signed char SBYTE;
typedef unsigned short WORD;
typedef signed short SWORD;
//typedef unsigned int DWORD;
typedef signed int SDWORD;

#ifdef _WIN64
typedef DWORD ADDR;
#else
typedef BYTE *ADDR;
#endif

typedef int CGL;

// A structure to hold function-related info, local to the current CB.
// This data must be remembered between a function prolog and epilog,
// across the function body generation.
// This should be part of the CB structure, since function
// generation may be interrupted by code generation in a different CB;
// e.g., by init code generation.
typedef struct {
	int nTemps;					//number of temps used by this function
	int cbTemps;				//number of bytes to allocate for temps
	int nTempRegsUsed;			//number of temp registers used by this function (or a mask??)
	int nArgs;					//number of formal parameters (negated if VPL)	NOT USED???
	int cbStaticArglistSize;	//number of bytes to allocate (-1 means none; minimum is 32 for X64)
	int nargsNeeded;			//true iff "nargs" is referenced by function
	int needFrameReg;			//true iff a stack frame register is needed
	int homeArgsMask;			//mask of args which must be homed (first 4 args only)
	int argTypeMask;			//mask of arg types; 0=integral, 1=floating point
	int fpuUsed;				//true only if 1 or more fpu intructions were emitted
	int loadFMCP;				//true iff the FM Context pointer needs to be loaded into rBX
	int cbLocals;				//number of bytes to allocate on stack for local vars
	CGL ref_cbLocals;			//fwd reference label to fixup cbLocals when known
} FUNC;

typedef struct			//Shouldn't we use this for Win32 also???
{
    //CGL lab;
    ADDR fixup_list;    //index to (last) fixup target; could alias to addr?
    ADDR addr;			//addr or abs value once lab is resolved (i.e. defined is TRUE)???
    BOOL defined;
    BOOL isDisp8Fixup;
    BOOL isAbsVal;
} CGL_INFO64;

typedef struct		//DEPRECATE THIS???
{
    //CGL lab;
    ADDR *fixup_list;    //could alias to addr?
    BYTE *addr;			 //addr once lab is resolved (i.e. defined is true)???
    BOOL defined;
    BOOL isDisp8Fixup;
    BOOL isAbsVal;
} CGL_INFO32;

#ifdef _WIN64
typedef struct			//This is the real CB64 typedef
{
    const char *displayName;//display name of this CB, for error messages
    BYTE *pcStart;			//ptr to start of the code buffer
    ADDR pc;				//index of next byte to be emitted in buffer
    ADDR pc_max;			//index of end of code buffer + 1 (= sizeof(cb))
    CGL next_cgl;			//next CGL to be used for this CB
    CGL max_cgl;			//# of CGLs that will fit in current CGL_INFO array
    CGL_INFO64 *cgls;		//ptr to array of CGLs for this Code Buffer
	int tosmax;				//max # of temps used
	FUNC func;				//function-related info
} CB64;

typedef struct			//This is a dummy version of CB32 with same size as true CB32
{
    INT32 displayName;		//display name of this CB, for error messages
    INT32 pc;				//ptr to next byte to be emitted in buffer
    INT32 pcStart;			//ptr to start of the code buffer
    INT32 pc_max;			//ptr to end of code buffer + 1
    CGL next_cgl;			//next CGL to be used for this CB
    CGL max_cgl;			//# of CGLs that will fit in current CGL_INFO array
    INT32 cgls;				//ptr to array of CGLs for this Code Buffer
	int tosmax;				//max # of temps used
	FUNC func;				//function-related info
} CB32;
#else //Win32
typedef struct			//This is a dummy version of CB64 with same size as true CB64
{
    INT64 displayName;		//display name of this CB, for error messages
    INT64 pcStart;			//ptr to start of the code buffer
    ADDR pc;				//index of next byte to be emitted in buffer
    ADDR pc_max;			//index of end of code buffer + 1 (= sizeof(cb))
    CGL next_cgl;			//next CGL to be used for this CB
    CGL max_cgl;			//# of CGLs that will fit in current CGL_INFO array
    INT64 cgls;				//ptr to array of CGLs for this Code Buffer
	int tosmax;				//max # of temps used
	FUNC func;				//function-related info
} CB64;

typedef struct			//This is the real CB32 typedef
{
    const char *displayName;//display name of this CB, for error messages
    BYTE *pc;				//ptr to next byte to be emitted in buffer
    BYTE *pcStart;			//ptr to start of the code buffer
    BYTE *pc_max;			//ptr to end of code buffer + 1
    CGL next_cgl;			//next CGL to be used for this CB
    CGL max_cgl;			//# of CGLs that will fit in current CGL_INFO array
    CGL_INFO32 *cgls;		//ptr to array of CGLs for this Code Buffer
	int tosmax;				//max # of temps used
	FUNC func;				//function-related info
} CB32;
#endif

// For now, we use different CGL_INFO and CB structures
// for Win64 vs Win32.  Someday we should switch to using
// the same structs in both environments, and eliminate
// this kludge!!
#ifdef _WIN64
typedef CGL_INFO64 CGL_INFO;
typedef CB64 CB;
#else	//Win32
typedef CGL_INFO32 CGL_INFO;
typedef CB32 CB;
#endif //Win32

// Prototypes
void X86_print(const char *msg);
void X86_abort(const char *msg);
void X86_abort2(const char *fmt, int iarg1);
void X86_error(const char *msg);
void X86_error2(const char *fmt, int iarg1);
void X86_warn(const char *msg);
BOOL X86_MMX_available(void);
BOOL X86_CMOV_available(void);
int X86_cpuid_version(void);
int X86_cpuid_features(void);
char *X86_cpuid_vendor_id(void);
const char *X86_reg64_name(int reg);
const char *X86_reg32_name(int reg);
const char *X86_reg16_name(int reg);
const char *X86_reg8H_name(int reg);
const char *X86_reg8L_name(int reg);
const char *X86_regv_name(int reg);
const char *X86_mmxreg_name(int reg);
const char *X86_xmmreg_name(int reg);
const char *X86_ymmreg_name(int reg);
const char *X86_stnreg_name(int reg);
CB* X86_emit_init(CB *this_cbp, BYTE cb[], int cb_size, const char *displayName);
void X86_record_tos(int tos);
void X86_init_prolog(CB *this_cbp, int max_prolog_size);
CB* X86_switch_to(CB *this_cbp);
int X86_emit_finish(void);
void X86_emit_free_resources(CB *this_cbp);
void X86_disassemble(CB *this_cbp);
void X86_emit_overflow(void);
void X86_emit_byte(BYTE b);
void X86_emit_word(WORD w);
void X86_emit_dword(DWORD dw);
void X86_emit_qword(QWORD qw);
void X86_emit_bytes(BYTE *cb, int nBytes);
void X86_append_cb(CB *that_cbp);
void X86_emit_op(int op);
void X86_emit_opreg(int op, int reg);
void X86_emit_mod_reg_rm(int mod, int reg, int rm);
void X86_emit_sib(int ss, int index, int base);
void X86_emit_RR(int reg1, int reg2);
void X86_emit_RM(int reg, int base_reg, int index_reg, int scaleFactor, int disp);
void X86_emit_op_RR(int op, int reg1, int reg2);
void X86_emit_op_M(int op,
                   int base_reg, int index_reg, int scaleFactor, int disp);
void X86_emit_op_T(int op, int tmp);
void X86_emit_op_Tm(int op, int tmp);
void X86_emit_op_R(int op, int reg);
void X86_emit_op_RM(int op,
                    int reg,
                    int base_reg, int index_reg, int scaleFactor, int disp);
void X86_emit_op_MR(int op,
                    int base_reg, int index_reg, int scaleFactor, int disp,
                    int reg);
void X86_emit_op_RT(int op, int reg, int tmp);
void X86_emit_op_TR(int op, int tmp, int reg);
void X86_emit_op_TT(int op, int tmp1, int tmp2);
void X86_emit_op_MM(int op,
                    int base_reg1, int index_reg1, int scaleFactor1, int disp1,
                    int base_reg2, int index_reg2, int scaleFactor2, int disp2);
void X86_emit_op_TM(int op, int tmp,
                    int base_reg, int index_reg, int scaleFactor, int disp);
void X86_emit_op_MT(int op,
                    int base_reg, int index_reg, int scaleFactor, int disp,
                    int tmp);

void X86_emit_op_RMi(int op, int reg, MEM1);
void X86_emit_op_Mi(int op, MEM1);
void X86_emit_op_MMi(int op, MEM1, MEM2);
void X86_emit_op_TMi(int op, int tmp, MEM1);
void X86_emit_op_TTi(int op, int tmp1, int tmp2);
void X86_emit_op_Ti(int op, int tmp);

void X86_emit_op_MiR(int op, MEM1, int reg);
void X86_emit_op_MiM(int op, MEM1, MEM2);
void X86_emit_op_MiT(int op, MEM1, int tmp);
void X86_emit_op_TiT(int op, int tmp1, int tmp2);
void X86_emit_op_TiR(int op, int tmp, int reg);

void X86_emit_op_imm8(int op, BYTE immData);
void X86_emit_op_imm16(int op, WORD immData);
void X86_emit_op_imm32(int op, DWORD immData);
void X86_emit_op_simm32(int op, SDWORD immData);
void X86_emit_op_M_imm8(int op,
                        int base_reg, int index_reg, int scaleFactor, int disp,
                        BYTE immData);
void X86_emit_op_T_imm8(int op, int tmp, BYTE immData);
void X86_emit_op_R_imm8(int op, int reg, BYTE immData);
void X86_emit_op_MR_imm8(int op,
                         int base_reg, int index_reg, int scaleFactor, int disp,
                         int reg, BYTE immData);
void X86_emit_op_TR_imm8(int op, int tmp, int reg, BYTE immData);
void X86_emit_op_RR_imm8(int op, int reg1, int reg2, BYTE immData);
void X86_emit_op_RM_simm32(int op, int reg,
                           int base_reg, int index_reg, int scaleFactor, int disp,
                           SDWORD immData);
void X86_emit_op_RT_simm32(int op, int reg, int tmp, SDWORD immData);
void X86_emit_op_MR_simm32(int op,
                           int base_reg, int index_reg, int scaleFactor, int disp,
                           int reg,
                           SDWORD immData);
void X86_emit_op_TR_simm32(int op, int tmp, int reg, SDWORD immData);
void X86_emit_op_RR_simm32(int op, int reg1, int reg2, SDWORD immData);
void X86_emit_op_MM_simm32(int op,
                           int base_reg1, int index_reg1, int scaleFactor1, int disp1,
                           int base_reg2, int index_reg2, int scaleFactor2, int disp2,
                           SDWORD immData);
void X86_emit_op_TT_simm32(int op, int tmp1, int tmp2, SDWORD immData);
void X86_emit_op_M_imm32(int op,
                         int base_reg, int index_reg, int scaleFactor, int disp,
                         DWORD immData);
void X86_emit_op_T_imm32(int op, int tmp, DWORD immData);
void X86_emit_op_R_imm32(int op, int reg, DWORD immData);
void X86_emit_op_M_simm32(int op,
                          int base_reg, int index_reg, int scaleFactor, int disp,
                          SDWORD immData);
void X86_emit_op_T_simm32(int op, int tmp, SDWORD immData);
void X86_emit_op_Ti_simm32(int op, int tmp, SDWORD immData);
void X86_emit_op_R_simm32(int op, int reg, SDWORD immData);
CGL X86_genlab(void);
void X86_deflab(CGL lab);
void X86_def_abs32(CGL lab, int ival32);
ADDR X86_lab_offset(CGL lab);
int X86_lab_disp_size(CGL lab, int nOpcodeBytes);
void X86_emit_lab_disp32(CGL lab);
void X86_emit_lab_disp8(CGL lab);
void X86_emit_op_lab_disp32(int op, CGL lab);
void X86_emit_op_lab_disp8(int op, CGL lab);
void X86_emit_op_lab(int op, CGL lab);
void X86_emit_op_R_ref_simm32(int op, int reg, CGL lab);
void X86_emit_op_R_ref_imm32(int op, int reg, CGL lab);
void X86_emit_op_R_ref_imm8(int op, int reg, CGL lab);
void X86_align(int align);
INT_PTR X86_call_fmf1(BYTE *code, INT_PTR p1);
void X86_emit_prolog(int nTemps, int cbLocals);
void X86_emit_epilog(int nTemps, int cbLocals, int nArgs);
void X86_emit_handler_prolog(int nTemps, int cbLocals);
void X86_emit_handler_epilog(int nTemps, int cbLocals, int nArgs);
void X86_emit_DIVREM_TT(int op, int tmp1, int tmp2, BOOL check);
void X86_emit_divCheck(void);
void X86_emit_defaultHandler(void);
//void X86_emit_defaultOnZeroDivide(void);
//void X86_emit_onDivideOverflow(void);
//void X86_emit_defaultOnDivideOverflow(void);
void X86_emit_compute_iuv(void);
void X86_emit_compute_d(void);
void X86_emit_compute_m(void);
void X86_emit_doOneRow(int nTemps, int cbLocals);
void X86_emit_doOneRow_part1(int nTemps, int cbLocals);
void X86_emit_doOneRow_part2(int nTemps, int cbLocals);

/* Global variables */

CB *cbp;	//pointer to current code buffer


#endif //!EMITX86_H
