//////////////////////// BEGIN COLOR CONVERSIONS //////////////////////////

//
//	BOOL set_bitdepth_mode(int mode)
//
//  Sets current working bits per channel (bpc),
//	where mode = 8 or 16.
//	Returns true if successful, or
//	false if mode is not 8 or 16.
//
//Added by Harald Heim
//Modified by Alex Hunter on May 6, 2014
//
// Values which depend on the current working bit-depth mode:
//
// WARNING! THESE VALUES ARE NOT THREAD-SAFE!!
// Should they be? I.e., should each thread set its own working
// bit-depth mode? Does that even make sense?
//
// To make them thread safe, either:
// (1) add them to the FMcontext record (if they should be exposed
//     to the FD), or
// (2) use TLS by adding the THREAD_LOCAL attribute.
//
// Should these values be initialized to the proper default values
// each time the filter is invoked (by calling fm_set_bitdepth_mode)?
//
static int bitDepthMode = 8;	//add to fmc and expose to FD???
static double bitMultiply255 = 255.0;	//assume 8bpc initially
static double bitDivide255 = 1.0 / 255.0;

static INT_PTR fm_set_bitdepth_mode(INT_PTR mode)
{ 
	if (mode == 8) {
		//8 bits per channel
		bitDepthMode = (int)mode;
		bitMultiply255 = 1.0*255.0;
		bitDivide255 = 1.0 / bitMultiply255;
	}
	else if (mode == 16) {
		//16 bits per channel
		bitDepthMode = (int)mode;
		bitMultiply255 = 128.0*256.0;
		bitDivide255 = 1.0 / bitMultiply255;
	}
	else return false;	//invalid bpc

    return true;	//success
}/*fm_set_bitdepth_mode*/

//
//	int rgb2hsl(int r, int g, int b, int z)
//
//	Converts an (r,g,b) triple to HSL, and
//	returns one requested component (h, s, or l)
//	per call, where (z mod 4) specifies which 
//	value will be returned: 'h' for z=0 (mod 4),
//	's' for z=1 (mod 4), or 'l' for z=2 (mod 4).
//	Returns 0 if z=3 (mod 4).
//
//	Calculated (h,s,l) components are cached
//	across each call to rgb2hsl(), to speed 
//	things up by avoiding recalculating all
//	components if they have already been
//	calculated for the same (r,g,b) color
//	by the previous call.
//
//	The (r,g,b) and (h,s,l) components will be	
//	treated as 8 bpc or 16 bpc, as determined by
//	the last call to the set_bitdepth_mode() routine.
//
//function added by Harald Heim on Dec 23, 2002
//function as modified by Alex Hunter on Apr 24, 2014
static INT_PTR fm_rgb2hsl(INT_PTR r, INT_PTR g, INT_PTR b, INT_PTR z)
{
    double rf,gf,bf,hf,sf,lf;
	double cmax,cmin,diff;
    
	//Note that TLS access is relatively expensive,
	//so we try to minimize the number of accesses,
	//such as by combining (r,g,b) into a single
	//key value.
	//Question: Is it really worth the extra TLS overhead
	//to cache thread-local values, or is it just as fast
	//to always calculate h/s/l for each call??
	typedef union {
		__int64 key;
		WORD wrd[4];
	} key_t;
	key_t rgb, hsl;
	THREAD_LOCAL static key_t cached_rgb = {-1};
	THREAD_LOCAL static key_t cached_hsl;

	//create a unique 64-bit key from r, g, b
	rgb.wrd[0] = (WORD)r;
	rgb.wrd[1] = (WORD)g;
	rgb.wrd[2] = (WORD)b;
	rgb.wrd[3] = 0;	//last word must be determinate!

	if (rgb.key == cached_rgb.key) {
		//return the previously cached result
		return cached_hsl.wrd[z & 3];
	}
	else {
		//Not in the cache -- must calculate

        //Scale rgb to float
#if 0
		//would multiplication by inverse be faster?? (see Agner Fog)
        rf=(double)r / bitMultiply255;
        gf=(double)g / bitMultiply255;
        bf=(double)b / bitMultiply255;
#else
		//Should be faster. But what about roundoff error??
        rf=(double)r * bitDivide255;
        gf=(double)g * bitDivide255;
        bf=(double)b * bitDivide255;
#endif

        //Convert RGB to HSL

		//cmax = max(bf,rf,gf);
		//cmin = min(bf,rf,gf);
        cmax = bf;
        cmin = bf;
        if      (rf > cmax) cmax = rf;
        else if (rf < cmin) cmin = rf;
        if      (gf > cmax) cmax = gf;
        else if (gf < cmin) cmin = gf;

        lf = (cmax + cmin)/2.0;

        if (cmax == cmin){
            sf = 0.0;
            hf = 0.0;
        } else {
			
			diff = cmax - cmin;
			if (lf < 0.5) {
				sf = diff/(cmax + cmin); 
			} else {
				sf = diff/(2.0 - (cmax + cmin));
			}

			if (rf == cmax) {
				hf = (gf - bf)/diff;
			} else if (gf == cmax) {
				hf = 2.0 + (bf - rf)/diff;
			} else {
				hf = 4.0 + (rf - gf)/diff;
			}
			hf = hf/6.0;
			if (hf < 0.0) hf = hf + 1.0;
        }   
        
        //Scale hsl to int (with rounding!)
        hsl.wrd[0] = (int)(hf*bitMultiply255 + 0.5);
        hsl.wrd[1] = (int)(sf*bitMultiply255 + 0.5);
        hsl.wrd[2] = (int)(lf*bitMultiply255 + 0.5);
		hsl.wrd[3] = 0; //unused

		//and cache the results
		cached_rgb = rgb;	//lookup key
		cached_hsl = hsl;	//result key

    } //end of calculation

	//Return the desired result
	return hsl.wrd[z & 3];
}//fm_rgb2hsl
