/*
    File: AfhFM.c
    C source file for the AFH FilterMeister Photoshop Plug-in Filter.

    Copyright 1997-2014 AFH Systems
	Copyright 1997-2017 Alex Hunter, Harald Heim & Martijn Van der Lee

  Derived from Photoshop SDK source files which are:

    Copyright 1990 by Thomas Knoll.
    Copyright 1991-96 by Adobe Systems, Inc.

  FilterMeister is free software. You can redistribute and modify it
  under the terms of the MPL2 License.
  See: https://www.mozilla.org/en-US/MPL/2.0/
  Any changes to the FilterMeister files have to be published or sent
  to harry@thepluginsite.com. The filter plugin code itself does not
  need to be disclosed.

*/
//
//	02AUG2013  Alex Hunter  1. Changes for WIN64.
//	16AUG2013  Alex Hunter  1. Add _ftol2, _ftol2_sse.
//	21AUG2013  Alex Hunter  1. Fix bug in fm_set_psetp_mode (erroneous test of mode).
//	23AUG2013  Alex Hunter  1. Fix missing break in case 3 of fm_quickFill().
//							2. Add fm_quickFillArray(). (possible pluggable)
//	25AUG2013  Alex Hunter  1. Fix error accessing ExitCode in fm_isThreadActive().
//	04NOV2013  Alex Hunter  1. Fix typo in comments for fm_formatString.
//	26DEC2013  Alex Hunter	1. In fm_formatString, add !x format code which
//							   expands to the target architecture ("32" or "64").
//							2. Add !X format code which expands to the host app
//							   architecture ("32" for Win32 or "64" for Win64).
//	30DEC2013  Alex Hunter	1. Change MAX_SCALE_FACTOR to 30000 for now.
//	08JAN2014  Alex Hunter	1. Add a modifed version of FilterBigDocument.cpp
//							   from the googlecode WinGP project.
//	11JAN2014  Alex Hunter  1. Changes for 32-bit image coordinates.
//							2. Add a modified version of the routines from
//							   FilterBigDocument.cpp.
//	30JAN2014  Alex Hunter  1. Don't clobber filterFilenameText in resetAllCtls().
//	19FEB2014  Alex Hunter  1. Set gParams->fValidCompile = TRUE when LoadFmCode succeeds.
//	05MAR2014  Alex Hunter	1. Guard all calls to X86_call_fmf1 by a call to setjmp,
//							   while X64 exception handling is not yet working.
//	07MAR2014  Alex Hunter  1. Fix X64 bug which causes transparent control backgrounds
//							   to turn black upon reinvocation.
//	10MAR2014  Alex Hunter  1. Use fmc.abort_mode to select behavior of fm_abort().
//	31MAR2014  Alex Hunter  1. Add trampoline to fix bug in X64 CRT fmod routine.
//	12APR2014  Alex Hunter	1. Fix bug in getCtlRange(). Sense of IS_INUSE_CTL test was
//							   reversed.
//	14APR2014  Alex Hunter	1. In waitForThread, test only low 32 bits of 'ms' against
//							   INFINITE (otherwise it's not recognized as such).
//	23APR2014  Alex Hunter	1. Correct some errors and optimize color conversion routines
//							   such as rgb2hsl, etc.
//	07APR2014  Alex Hunter	1. Add more comments to fm_set_bitdepth_mode() and fm_rgb2hsl().
//	08APR2014  Alex Hunter	1. Set some parameters back to 'int' instead of 'INT_PTR' in
//							   waitForThread and waitForSync to fix errors when testing
//							   the timeout parameter against INFINITE.
//	24MAY2014  Alex Hunter	1. Add casts to some library routines in FMC to suppress warnings
//							   for now. Fix properly later!!
//							2. REMEMBER to fix later by adding thunks if necessary wherever
//							   you find a "need thunk" comment!!
//	04JUL2014  Alex Hunter  1. Change __int3264 to INT_PTR (or UINT_PTR) everywhere for
//							   consistency.
//							2. Many changes to ipow() to fix bad errno reporting, GP for
//							   integer divide by 0, etc.
//
//				TODOO
//
//	Allow negative coordinates for multiple screens!!
//
///#define TRIG_ON_FLY 1	//while debugging X64!!

#ifdef _DEBUG
#define VERBOSE 1
#endif

#define REG_DATA_LIMIT 2048	//maximum registry data entry we allow by user

#include <stdlib.h> // abs, rand
#include <malloc.h> // _expand, _msize
#include <math.h>
#include <string.h>
#include <time.h>   // clock, time.
#include <limits.h> // INT_MAX

#include <stdio.h>  // vsprintf, _vsnprintf
#include <stdarg.h> // va_list, etc.
#include <assert.h>
#include <errno.h>
#include <direct.h> //directory stuff
#include <shlobj.h> //For SHGetSpecialFolderLocation, must come before AfhFM.h

#include "AfhFM.h"
#include "AfhFMcontext.h"
#include "EmitX86.h"    // X86_MMX_available...
#include "plmcvax.h"

#include "PIUtilities.h"
#include <commctrl.h>   //PBM_SETPOS

#if defined(FM_LIFTCOV)
#include "liftcov/resliftcov.h"
#elif defined(FM_RADSINAL)
#include "RadSinal/resRadSinal.h"
#elif defined(FM_MONOMET)
#include "monomet/resmonomet.h"
#else
#include "resource.h"
#endif



#if 0
//Added by Harald Heim, jan 14, 2002 for getWindowsVersion
#include <windows.h>
#endif


#include "AfhFMScripting.h"

	/* macros */

#ifdef _WIN64 
#define INTV_MAX _I64_MAX
#define INTV_MIN _I64_MIN
#define IABS _abs64			//CRT 64-bit integer abs
#else
#define INTV_MAX INT_MAX
#define INTV_MIN INT_MIN
#define IABS abs			//CRT 32-bit integer abs
#endif

    /* externals */

///void *Mdc_evaluate_sdl(void *sdlp, int compile_flag);
///int Mdc_get_estack_val(int index);

    /* Module-local global constants. */

static const double twopi = 6.28318530717958647693;
static const double twobypi = 0.63661977236758134308;

    /* Module-local global variables. */

static int gNoMask;
static int gBevWidth;						//thread-safe???
static int minRow, maxRow, minCol, maxCol;
static int X0, Y0;

// Values which depend on the current working bit-depth mode.
//
// WARNING! THESE VALUES ARE NOT THREAD-SAFE!!
// Should they be? I.e., should each thread set its own working
// bit-depth mode? Does that even make sense?
//
// To make them thread safe, either:
// (1) add them to the FMcontext record (if they should be exposed
//     to the FD), or
// (2) use TLS by adding the THREAD_LOCAL attribute.
//
// Should these values be initialized to the proper default values
// each time the filter is invoked (by calling fm_set_bitdepth_mode)?
//
static int bitDepthMode = 8;	//add to fmc and expose to FD???
static double bitMultiply255 = 255.0;	//assume 8bpc initially
static double bitDivide255 = 1.0 / 255.0;

//Added by Harald Heim		// ARE THESE THREAD-SAFE?? -- add to FMcontext?? Init before compile???
static int edgeMode = 0;
static int psetpMode = 0;

static int Dont_Initialize_Cells = 0;	//THREAD-SAFE??
int MouseOverWhenInvisible = 1;
int EditBoxUpdate = 0;

// IS FOLLOWING THREAD-SAFE?? Are Arrays global or thread-local???
void *Array[N_ARRAYS]; //Initialized in DoStart
BufferID ArrayID[N_ARRAYS]; //for BufferProc //Initialized in DoStart
static INT_PTR ArraySize[N_ARRAYS];
static int ArrayBytes[N_ARRAYS];
static INT_PTR ArrayX[N_ARRAYS], ArrayY[N_ARRAYS], ArrayZ[N_ARRAYS]; //Initialized in DoStart
//Added by Harald Heim, Oct 12, 2003
static INT_PTR ArrayPadding[N_ARRAYS]; //Initialized in DoStart
int sizeGrip=0;
static int arrayMode = 0; //Is set to allocation by the host in DoStart
BOOL scaleLock=false;		//MUST be initialized dynamically before user Init Code??


static INT_PTR fm_set_array_mode(INT_PTR mode);

static INT_PTR *ThreadHandle = NULL;
static INT_PTR *ThreadMem = NULL;
static int ThreadCount = 0;


#define N_FONTS 32
HFONT Font[N_FONTS];	//THREAD-SAFE???



//Declarations for fm_pset/fm_pset16
static INT_PTR fm_msk(INT_PTR x, INT_PTR y);
static INT_PTR fm_src(const INT_PTR x0, const INT_PTR y0, const INT_PTR z0);
static INT_PTR fm_src16(const INT_PTR x0, const INT_PTR y0, const INT_PTR z0);

static INT_PTR fm_freeArray (INT_PTR nr);

static INT_PTR fm_getArrayAddress(INT_PTR nr);


INT_PTR fm_getAppTheme (void);
int appTheme = -1;	//THREAD-SAFE???


//Distance Array
static BufferID bufferID = 0;	//needs to be re-inited?  Thread-safe???
static struct Vect_s
{
    short x, y;
};

static struct Vect_s *dist;		//Thread-safe???
int ClickDrag=0;				//Should be re-inited across compiles!!!




    /* Globals */

GPtr globals;

// FilterMeister HOST-specific context record.	//X64???
static FMcontext fmc;
FMcontext * const gFmcp = &fmc;

/*****************************************************************************/
/*  Following is a modifed version of FilterBigDocument.cpp
/*  from the googlecode WinGP project at:
/*	http://wingp.googlecode.com/svn/photogp/common/sources/
/*****************************************************************************/
//maybe these should always update the fmc fields???
// E.g., GetFilterRect() and SetFilterRect() would copy also to fmc.filterRect32???
VPoint GetImageSize(void)
{
	VPoint returnPoint;

	gFilterRecord = gStuff;
	gBigDocumentData = gStuff->bigDocumentData;

	if (gBigDocumentData != NULL)
	{
		returnPoint = gBigDocumentData->imageSize32;
	}
	else
	{
		returnPoint.h = gFilterRecord->imageSize.h;
		returnPoint.v = gFilterRecord->imageSize.v;
	}

	return returnPoint;
}

VRect GetFilterRect(void)
{
	VRect returnRect;

	gFilterRecord = gStuff;
	gBigDocumentData = gStuff->bigDocumentData;

	if (gBigDocumentData != NULL)
	{
		returnRect = gBigDocumentData->filterRect32;
	}
	else
	{
		returnRect.right = gFilterRecord->filterRect.right;
		returnRect.top = gFilterRecord->filterRect.top;
		returnRect.left = gFilterRecord->filterRect.left;
		returnRect.bottom = gFilterRecord->filterRect.bottom;
	}
	
	return returnRect;
}

VRect GetInRect(void)
{
	VRect returnRect;

	gFilterRecord = gStuff;
	gBigDocumentData = gStuff->bigDocumentData;

	if (gBigDocumentData != NULL)
	{
		returnRect = gBigDocumentData->inRect32;
	}
	else
	{
		returnRect.right = gFilterRecord->inRect.right;
		returnRect.top = gFilterRecord->inRect.top;
		returnRect.left = gFilterRecord->inRect.left;
		returnRect.bottom = gFilterRecord->inRect.bottom;
	}
	
	return returnRect;
}

VRect GetOutRect(void)
{
	VRect returnRect;

	gFilterRecord = gStuff;
	gBigDocumentData = gStuff->bigDocumentData;

	if (gBigDocumentData != NULL)
	{
		returnRect = gBigDocumentData->outRect32;
	}
	else
	{
		returnRect.right = gFilterRecord->outRect.right;
		returnRect.top = gFilterRecord->outRect.top;
		returnRect.left = gFilterRecord->outRect.left;
		returnRect.bottom = gFilterRecord->outRect.bottom;
	}
	
	return returnRect;
}

VRect GetMaskRect(void)
{
	VRect returnRect;

	gFilterRecord = gStuff;
	gBigDocumentData = gStuff->bigDocumentData;

	if (gBigDocumentData != NULL)
	{
		returnRect = gBigDocumentData->maskRect32;
	}
	else
	{
		returnRect.right = gFilterRecord->maskRect.right;
		returnRect.top = gFilterRecord->maskRect.top;
		returnRect.left = gFilterRecord->maskRect.left;
		returnRect.bottom = gFilterRecord->maskRect.bottom;
	}
	
	return returnRect;
}

VPoint GetFloatCoord(void)
{
	VPoint returnPoint;

	gFilterRecord = gStuff;
	gBigDocumentData = gStuff->bigDocumentData;

	if (gBigDocumentData != NULL)
	{
		returnPoint = gBigDocumentData->floatCoord32;
	}
	else
	{
		returnPoint.h = gFilterRecord->floatCoord.h;
		returnPoint.v = gFilterRecord->floatCoord.v;
	}

	return returnPoint;
}

VPoint GetWholeSize(void)
{
	VPoint returnPoint;

	gFilterRecord = gStuff;
	gBigDocumentData = gStuff->bigDocumentData;

	if (gBigDocumentData != NULL)
	{
		returnPoint = gBigDocumentData->wholeSize32;
	}
	else
	{
		returnPoint.h = gFilterRecord->wholeSize.h;
		returnPoint.v = gFilterRecord->wholeSize.v;
	}

	return returnPoint;
}

void SetInRect(VRect inRect)
{
	gFilterRecord = gStuff;
	gBigDocumentData = gStuff->bigDocumentData;

	if (gBigDocumentData != NULL)
	{
		gBigDocumentData->PluginUsing32BitCoordinates = 1;	//temporary kludge!!!
		gBigDocumentData->inRect32 = inRect;
	}
	else
	{
		gFilterRecord->inRect.right = (int16)inRect.right;
		gFilterRecord->inRect.top = (int16)inRect.top;
		gFilterRecord->inRect.left = (int16)inRect.left;
		gFilterRecord->inRect.bottom = (int16)inRect.bottom;
	}
}

void SetOutRect(VRect inRect)
{
	gFilterRecord = gStuff;
	gBigDocumentData = gStuff->bigDocumentData;

	if (gBigDocumentData != NULL)
	{
		gBigDocumentData->PluginUsing32BitCoordinates = 1;	//temporary kludge!!!
		gBigDocumentData->outRect32 = inRect;
	}
	else
	{
		gFilterRecord->outRect.right = (int16)inRect.right;
		gFilterRecord->outRect.top = (int16)inRect.top;
		gFilterRecord->outRect.left = (int16)inRect.left;
		gFilterRecord->outRect.bottom = (int16)inRect.bottom;
	}
}

void SetMaskRect(VRect inRect)
{
	gFilterRecord = gStuff;
	gBigDocumentData = gStuff->bigDocumentData;

	if (gBigDocumentData != NULL)
	{
		gBigDocumentData->PluginUsing32BitCoordinates = 1;	//temporary kludge!!!
		gBigDocumentData->maskRect32 = inRect;
	}
	else
	{
		gFilterRecord->maskRect.right = (int16)inRect.right;
		gFilterRecord->maskRect.top = (int16)inRect.top;
		gFilterRecord->maskRect.left = (int16)inRect.left;
		gFilterRecord->maskRect.bottom = (int16)inRect.bottom;
	}
}
/*****************************************************************************/
// end of modified FilterBigDocument.cpp
/*****************************************************************************/

    /* Forward prototypes */

void InitGlobals (GPtr globals);
void ValidateParameters (GPtr globals);

void DoParameters (GPtr globals);
void DoPrepare (GPtr globals);
void DoStart (GPtr globals);
void DoContinue (GPtr globals);
void DoFinish (GPtr globals);

static INT_PTR fm_doForEveryRow(void);
void InitContextRecord(void);
static void computeBuddy2Pos(INT_PTR n, INT_PTR x, INT_PTR y, INT_PTR w, INT_PTR h, RECT *pr);

/*****************************************************************************/

/* All calls to the plug-in module come through this routine. It must be
   placed first in the resource. To achieve this, most development systems
   require that this be the first routine in the source. */

#define DllExport   __declspec( dllexport )

#ifndef _INTPTR_T_DEFINED
#ifdef _WIN64
   typedef __int64 intptr_t;
#else 
   typedef long intptr_t;
#endif
#define _INTPTR_T_DEFINED
#endif

DllExport void ENTRYPOINT (short selector,
                        FilterRecord *filterParamBlock,
                        intptr_t *data,
                        short *result)              
    {
    
    Globals globalValues;
    GPtr globals = &globalValues;
    
    if (!*data)
        {
        
        InitGlobals (globals);

        *data = (INT_PTR) NewHandle (sizeof (Globals));
        
        if (!*data)
            {
            *result = memFullErr;
            return;
            }
            
        ** (GHdl) *data = globalValues;
        
        }
        
    globalValues = ** (GHdl) *data;
    //Should be the only ptr to globals that we need!!!
    *(GPtr *)&MyGlobals = globals;    //cast away const on MyGlobals this one time!
        
    gStuff = filterParamBlock;
    gResult = noErr;
        
    switch (selector)
        {
        
        case filterSelectorAbout:
            DoAbout (globals, selector);
            break;
            
        case filterSelectorParameters:
            DoParameters (globals);
            break;
            
        case filterSelectorPrepare:
            DoPrepare (globals);
            break;
        
        case filterSelectorStart:
            DoStart (globals);
            break;

        case filterSelectorContinue:
#ifndef DEMO 
            DoContinue (globals);
#else
    gResult = 1;
#endif
            break;
        
        case filterSelectorFinish:
#ifndef DEMO       
            DoFinish (globals);
#else
    gResult = 1;
#endif
            break;

        default:
            gResult = filterBadParameters;  // NFG for PSP???
            break;
        
        }

    if (gResult == filterBadParameters)
    {   // PSP and others don't recognize the filterBadParameters error,
        // so put out our own message box...
        ShowAlert(16502);
        gResult = 1;    // indicates error message already displayed.
    }


    *result = gResult;
    
    ** (GHdl) *data = globalValues;
        
}


DllExport void ENTRYPOINT00 (short selector, FilterRecord *filterParamBlock, intptr_t *data, short *result)
{
	fmc.entryPoint = 0;
	ENTRYPOINT (selector, filterParamBlock, data, result);
}
DllExport void ENTRYPOINT01 (short selector, FilterRecord *filterParamBlock, intptr_t *data, short *result)
{
	fmc.entryPoint = 1;
	ENTRYPOINT (selector, filterParamBlock, data, result);
}
DllExport void ENTRYPOINT02 (short selector, FilterRecord *filterParamBlock, intptr_t *data, short *result)
{
	fmc.entryPoint = 2;
	ENTRYPOINT (selector, filterParamBlock, data, result);
}
DllExport void ENTRYPOINT03 (short selector, FilterRecord *filterParamBlock, intptr_t *data, short *result)
{
	fmc.entryPoint = 3;
	ENTRYPOINT (selector, filterParamBlock, data, result);
}
DllExport void ENTRYPOINT04 (short selector, FilterRecord *filterParamBlock, intptr_t *data, short *result)
{
	fmc.entryPoint = 4;
	ENTRYPOINT (selector, filterParamBlock, data, result);
}
DllExport void ENTRYPOINT05 (short selector, FilterRecord *filterParamBlock, intptr_t *data, short *result)
{
	fmc.entryPoint = 5;
	ENTRYPOINT (selector, filterParamBlock, data, result);
}
DllExport void ENTRYPOINT06 (short selector, FilterRecord *filterParamBlock, intptr_t *data, short *result)
{
	fmc.entryPoint = 6;
	ENTRYPOINT (selector, filterParamBlock, data, result);
}
DllExport void ENTRYPOINT07 (short selector, FilterRecord *filterParamBlock, intptr_t *data, short *result)
{
	fmc.entryPoint = 7;
	ENTRYPOINT (selector, filterParamBlock, data, result);
}
DllExport void ENTRYPOINT08 (short selector, FilterRecord *filterParamBlock, intptr_t *data, short *result)
{
	fmc.entryPoint = 8;
	ENTRYPOINT (selector, filterParamBlock, data, result);
}
DllExport void ENTRYPOINT09 (short selector, FilterRecord *filterParamBlock, intptr_t *data, short *result)
{
	fmc.entryPoint = 9;
	ENTRYPOINT (selector, filterParamBlock, data, result);
}
DllExport void ENTRYPOINT10 (short selector, FilterRecord *filterParamBlock, intptr_t *data, short *result)
{
	fmc.entryPoint = 10;
	ENTRYPOINT (selector, filterParamBlock, data, result);
}
DllExport void ENTRYPOINT11 (short selector, FilterRecord *filterParamBlock, intptr_t *data, short *result)
{
	fmc.entryPoint = 11;
	ENTRYPOINT (selector, filterParamBlock, data, result);
}
DllExport void ENTRYPOINT12 (short selector, FilterRecord *filterParamBlock, intptr_t *data, short *result)
{
	fmc.entryPoint = 12;
	ENTRYPOINT (selector, filterParamBlock, data, result);
}
DllExport void ENTRYPOINT13 (short selector, FilterRecord *filterParamBlock, intptr_t *data, short *result)
{
	fmc.entryPoint = 13;
	ENTRYPOINT (selector, filterParamBlock, data, result);
}
DllExport void ENTRYPOINT14 (short selector, FilterRecord *filterParamBlock, intptr_t *data, short *result)
{
	fmc.entryPoint = 14;
	ENTRYPOINT (selector, filterParamBlock, data, result);
}
DllExport void ENTRYPOINT15 (short selector, FilterRecord *filterParamBlock, intptr_t *data, short *result)
{
	fmc.entryPoint = 15;
	ENTRYPOINT (selector, filterParamBlock, data, result);
}
DllExport void ENTRYPOINT16 (short selector, FilterRecord *filterParamBlock, intptr_t *data, short *result)
{
	fmc.entryPoint = 16;
	ENTRYPOINT (selector, filterParamBlock, data, result);
}
DllExport void ENTRYPOINT17 (short selector, FilterRecord *filterParamBlock, intptr_t *data, short *result)
{
	fmc.entryPoint = 17;
	ENTRYPOINT (selector, filterParamBlock, data, result);
}
DllExport void ENTRYPOINT18 (short selector, FilterRecord *filterParamBlock, intptr_t *data, short *result)
{
	fmc.entryPoint = 18;
	ENTRYPOINT (selector, filterParamBlock, data, result);
}
DllExport void ENTRYPOINT19 (short selector, FilterRecord *filterParamBlock, intptr_t *data, short *result)
{
	fmc.entryPoint = 19;
	ENTRYPOINT (selector, filterParamBlock, data, result);
}
DllExport void ENTRYPOINT20 (short selector, FilterRecord *filterParamBlock, intptr_t *data, short *result)
{
	fmc.entryPoint = 20;
	ENTRYPOINT (selector, filterParamBlock, data, result);
}
DllExport void ENTRYPOINT21 (short selector, FilterRecord *filterParamBlock, intptr_t *data, short *result)
{
	fmc.entryPoint = 21;
	ENTRYPOINT (selector, filterParamBlock, data, result);
}
DllExport void ENTRYPOINT22 (short selector, FilterRecord *filterParamBlock, intptr_t *data, short *result)
{
	fmc.entryPoint = 22;
	ENTRYPOINT (selector, filterParamBlock, data, result);
}
DllExport void ENTRYPOINT23 (short selector, FilterRecord *filterParamBlock, intptr_t *data, short *result)
{
	fmc.entryPoint = 23;
	ENTRYPOINT (selector, filterParamBlock, data, result);
}
DllExport void ENTRYPOINT24 (short selector, FilterRecord *filterParamBlock, intptr_t *data, short *result)
{
	fmc.entryPoint = 24;
	ENTRYPOINT (selector, filterParamBlock, data, result);
}
DllExport void ENTRYPOINT25 (short selector, FilterRecord *filterParamBlock, intptr_t *data, short *result)
{
	fmc.entryPoint = 25;
	ENTRYPOINT (selector, filterParamBlock, data, result);
}
DllExport void ENTRYPOINT26 (short selector, FilterRecord *filterParamBlock, intptr_t *data, short *result)
{
	fmc.entryPoint = 26;
	ENTRYPOINT (selector, filterParamBlock, data, result);
}
DllExport void ENTRYPOINT27 (short selector, FilterRecord *filterParamBlock, intptr_t *data, short *result)
{
	fmc.entryPoint = 27;
	ENTRYPOINT (selector, filterParamBlock, data, result);
}
DllExport void ENTRYPOINT28 (short selector, FilterRecord *filterParamBlock, intptr_t *data, short *result)
{
	fmc.entryPoint = 28;
	ENTRYPOINT (selector, filterParamBlock, data, result);
}
DllExport void ENTRYPOINT29 (short selector, FilterRecord *filterParamBlock, intptr_t *data, short *result)
{
	fmc.entryPoint = 29;
	ENTRYPOINT (selector, filterParamBlock, data, result);
}
DllExport void ENTRYPOINT30 (short selector, FilterRecord *filterParamBlock, intptr_t *data, short *result)
{
	fmc.entryPoint = 30;
	ENTRYPOINT (selector, filterParamBlock, data, result);
}
DllExport void ENTRYPOINT31 (short selector, FilterRecord *filterParamBlock, intptr_t *data, short *result)
{
	fmc.entryPoint = 31;
	ENTRYPOINT (selector, filterParamBlock, data, result);
}



/*****************************************************************************/

void InitGlobals (GPtr globals)
    {
    
    /* None of the globals requires initialization. */
    
}

//-------------------------------------------------------------------------------
//
//  ValidateParameters
//
//  Initialize parameters to default values.
//
//  Inputs:
//      GPtr globals        Pointer to global structure.
//
//  Outputs:
//
//-------------------------------------------------------------------------------

void ValidateParameters (GPtr globals)
{
    int i;

    // Should add a version # to gParams, and ignore old params if version #
    // mismatch!!!

    if (!gStuff->parameters)
    {

        gStuff->parameters = NewHandle ((long) sizeof (TParameters));

        if (!gStuff->parameters)
        {
            gResult = memFullErr;
            return;
        }


// DEP Fix Start ---------------------------------------
#if 1
        {//scope
	        // "TParameters" contains the code buffers, which must be executable...
            DWORD oldProtect;
            BOOL iss;
            iss = VirtualProtect(gParams, sizeof(*gParams), PAGE_EXECUTE_READWRITE, &oldProtect);
#if 1
            if (!iss) {
                ErrorOk("VirtualProtect failed in ValidateParameters.\n"
                        "Last error code is %x", GetLastError());
            }
#if 0
            else {  //extra information while debugging this fix...
                MEMORY_BASIC_INFORMATION mbi;
                size_t cb;
                HGLOBAL hMem = GlobalHandle(gParams);

                Info("VirtualProtect succeeded in ValidateParameters.\n"
                     "Old protection was %x", oldProtect);

                Info("gParams = %p\nHandle = %x\nRequested size = %x (%d.)\nActual size = %x",
                     gParams, hMem, sizeof(*gParams), sizeof(*gParams), GlobalSize(hMem));

                memset(&mbi, 0, sizeof(mbi));
                cb = VirtualQuery(gParams, &mbi, sizeof(mbi));
                Info("Memory info for address %p (returned %d bytes of info)\n"
                     "Base Address: %p\n"
                     "Allocation Base: %p\n"
                     "Allocation Protect: %x\n"
                     "Region Size: %x (%d.)\n"
                     "State: %x\n"
                     "Protect: %x\n"
                     "Type: %x\n",
                     gParams, cb,
                     mbi.BaseAddress,
                     mbi.AllocationBase,
                     mbi.AllocationProtect,
                     mbi.RegionSize, mbi.RegionSize, //hex, dec
                     mbi.State,
                     mbi.Protect,
                     mbi.Type);
            }//else if iss
#endif
#endif
        }//scope
#endif
// DEP Fix End ---------------------------------------


#if 0
        // Note that NewHandle zeroes the allocated memory.  Otherwise,         /// DOES IT ZERO????? yes, sets GMEM_ZEROINIT
        // we should zero out the gParams here to make sure the ptrs in
        // the CB code buffer structures are set to NULL, to avoid
        // problems when calling X86_emit_free_resources() on an
        // otherwise uninitialized CB.
        memset(gParams, 0, sizeof(*gParams));
#endif

        // Better query for parameters, since they didn't exist!!
        gParams->queryForParameters = TRUE;

        gParams->version_major = 1;
        gParams->version_minor = 0;

        gParams->flags     = AFH_USE_ADVANCESTATE|
							 AFH_FLUSH_ADVANCESTATE|
                             AFH_INIT_OPTIMAL_ZOOM|
#ifdef _WIN64
							 //Default on Win64 is X64GEN
							 AFH_X64GEN|
#endif
                             AFH_ZERO_OUTPUT_DRAG| 
                             AFH_USE_INLINES|
                             AFH_SUPPORT_RGB; //|AFH_QUICK_TOGGLE ; /*|AFH_SUPPORT_GRAY| AFH_PROXY_BIG_GULP; */ 

        if (X86_MMX_available()) {
            //for now turn off MMX by default since it seems to crash PSP
            // -- possibly because of the fullword store at the end of
            //    the image???
#if 0
            gParams->flags |= AFH_USE_MMX;
#endif
        }
        fmc.use_MMX = (gParams->flags & AFH_USE_MMX) != 0;

        if (X86_CMOV_available()) {
            //turn off CMOV by default, so we don't accidentally create
            //standalone filters with CMOV instructions
#if 0
            gParams->flags |= AFH_USE_CMOV;
#endif
        }
        fmc.use_CMOV = (gParams->flags & AFH_USE_CMOV) != 0;

#if 0 //NYI
        if (X64GEN || X86_SSE_available()) {
            //turn off SSE by default, so we don't accidentally create
            //standalone filters with SSE instructions
            gParams->flags |= AFH_USE_SSE;
        }
#endif
        fmc.use_SSE = (gParams->flags & AFH_USE_SSE) != 0;

        gParams->bandWidth = 100;    // Slow FM down for now...

        for (i = 0;  i < N_CTLS;  i++)
        {   //default control settings
            //gParams->ctl[i].id = 0;
            gParams->ctl[i].initVal = 0;
            gParams->ctl[i].val = 0;
            gParams->ctl[i].defval = 0xffffffff;	//or -1 for X64???
            gParams->ctl[i].minval = 0;
            gParams->ctl[i].maxval = 255;
            gParams->ctl[i].lineSize = 1;
            gParams->ctl[i].pageSize = 10;
			gParams->ctl[i].thumbSize = 0;
            gParams->ctl[i].properties = 0;
            gParams->ctl[i].action = CA_NONE;
            gParams->ctl[i].label[0] = '\0';
            gParams->ctl[i].enabled = 3; //enabled and visible
            gParams->ctl[i].inuse = FALSE;
            gParams->ctl[i].tooltip[0] = '\0';
            gParams->ctl[i].image[0] = '\0';
            gParams->ctl[i].imageType = 0;
            
            gParams->ctl[i].editwidth = 24;
            gParams->ctl[i].editheight = 10;

			gParams->ctl[i].tabControl = -1;
            gParams->ctl[i].tabSheet = -1;
			gParams->ctl[i].anchor = 0;
			gParams->ctl[i].distWidth = 0;
			gParams->ctl[i].distHeight = 0;
			
            gParams->ctl[i].divisor = 1;
            gParams->ctl[i].gamma = 100;
			gParams->ctl[i].theme = -1;

            gParams->ctl[i].scripted = -1;
        }

        //Preview is always used
        //gParams->ctl[CTL_PREVIEW].inuse = TRUE;

        //Activate the preview events by default for now
        //until the problem with Photoshop is solved
        gParams->ctl[CTL_PREVIEW].properties = CTP_MOUSEMOVE | CTP_MOUSEOVER | CTP_PREVIEWDRAG;
		gParams->ctl[CTL_PREVIEW].bkColor = -1; //No preview color
		gParams->ctl[CTL_PREVIEW].xPos = -1;
		gParams->ctl[CTL_PREVIEW].yPos = -1;
		gParams->ctl[CTL_PREVIEW].width = -1;
		gParams->ctl[CTL_PREVIEW].height = -1;
		gParams->ctl[CTL_PREVIEW].enabled = 3;

		gParams->ctl[CTL_FRAME].xPos = -1;
		gParams->ctl[CTL_FRAME].yPos = -1;
		gParams->ctl[CTL_FRAME].width = -1;
		gParams->ctl[CTL_FRAME].height = -1;
		gParams->ctl[CTL_FRAME].enabled = 3;

		gParams->ctl[CTL_PROGRESS].xPos = -1;
		gParams->ctl[CTL_PROGRESS].yPos = -1;
		gParams->ctl[CTL_PROGRESS].width = -1;
		gParams->ctl[CTL_PROGRESS].height = -1;
		gParams->ctl[CTL_PROGRESS].enabled = 3;

		gParams->ctl[CTL_ZOOM].xPos = -1;
		gParams->ctl[CTL_ZOOM].yPos = -1;
		gParams->ctl[CTL_ZOOM].width = -1;
		gParams->ctl[CTL_ZOOM].height = -1;
		gParams->ctl[CTL_ZOOM].enabled = 3;



        gParams->fValidParse = FALSE;
        gParams->fValidCompile = FALSE;
#ifndef STANDALONE
        gParams->fGotSourceCode = FALSE;
        gParams->fNeedSourceSave = FALSE;
        gParams->szSrcCode[0] = '\0';  //Empty source code buffer
#endif //ifndef STANDALONE

        //Default dialog settings
        gParams->gInitialInvocation = TRUE;
        gParams->gDialogRegion = NULL;
#if 0
        gParams->gDialogGradientColor1 = RGB(0,0,0);        //black
        gParams->gDialogGradientColor2 = RGB(0,0,0);        //black
#else
        //Changed by Harald Heim, March 9, 2003
        gParams->gDialogGradientColor1 = GetSysColor(COLOR_BTNFACE);//RGB(0xD4,0xD0,0xC8);   //yellow grey
        gParams->gDialogGradientColor2 = GetSysColor(COLOR_BTNFACE);//RGB(0xD4,0xD0,0xC8);   //yellow grey
        //gParams->gDialogGradientColor1 = RGB(0x5F,0x9E,0xA0);   //X11.cadetblue
        //gParams->gDialogGradientColor2 = RGB(0x5F,0x9E,0xA0);   //X11.cadetblue
#endif
        gParams->gDialogGradientDirection = 0;              //vertical
        gParams->gDialogImage_save = gDialogImage = NULL;   //no background image
        gParams->gDialogImageFileName[0] = '\0';
        gParams->gDialogImageMode = DIM_TILED;              //TILED
        gParams->gDialogImageStretchMode = COLORONCOLOR;
        gParams->gDialogDragMode = 0;                       //drag by titlebar
#if 0
        gParams->gDialogTitle[0] = '\0';                    //default dialog title
#elif 1
        //default dialog title is "filter title (image mode, filter case)"
        strcpy(gParams->gDialogTitle, "!t (!M, !f)");
#else
        //default dialog title is "filter title (filter case) [image mode]"
        strcpy(gParams->gDialogTitle, "!t (!f) [!M]");
#endif

		gParams->gDialogTheme = 0; //Default Dialog Theme

		gParams->gDialogEvent = 0; //Default value: no event


        //initialize registry access...
        fm_setRegRoot(HKEY_CURRENT_USER);
        fm_setRegPath("Software\\!O\\!C\\!t");

        // Initialize built-in strings... (already zeroed anyway!)
        gParams->filterAuthorText[0] = 0;
        gParams->filterCategoryText[0] = 0;
        gParams->filterTitleText[0] = 0;
        gParams->filterFilenameText[0] = 0;
        gParams->filterCopyrightText[0] = 0;
        gParams->filterDescriptionText[0] = 0;
        gParams->filterOrganizationText[0] = 0;
        gParams->filterURLText[0] = 0;
        gParams->filterVersionText[0] = 0;
        gParams->filterAboutText[0] = 0;    //temporary
        gParams->filterUniqueID[0] = 0;

#ifndef STANDALONE
        // Initialize embedded resource array...
        gParams->embeds[0] = 0;
        gParams->cbEmbeds = 1;
#endif

    } //!gStuff->parameters
} // end ValidateParameters


/*****************************************************************************/

/* Asks the user for the plug-in filter module's parameters. Note that
   the image size information is not yet defined at this point. Also, do
   not assume that the calling program will call this routine every time the
   filter is run (it may save the data held by the parameters handle in
   a macro file). */

void DoParameters (GPtr globals)
{
    ValidateParameters (globals);

    if (gResult != noErr) return;

    gParams->queryForParameters = TRUE; //redundant!!!



} /*DoParameters*/

/*****************************************************************************/
static int topPad, bottomPad, leftPad, rightPad;	//THREAD-SAFE??
static VRect *mskRectPtr;

static void addPadding (GPtr globals)
{
#if 0   //OBSOLETE???
#ifdef STATIC_FILTER
    int padding = 20000;  // try to grab the entire rect for now
#else
    int padding = 103;  // upper bound for now... (100 pixel bevel + 3 pixel radius max blur)
#endif
#endif

    VRect scaledFilterRect;
    long  scaleFactor;
    int padding;

    if (fmc.isTileable)
        padding = fmc.needPadding;//0;        // for now, assume no padding needed
    else
        padding = 20000;    // try to grab the entire rect

	// Grab maskRect and inRect from host (necessary??)
	fmc.maskRect32 = GetMaskRect();
	fmc.inRect32 = GetInRect();

#ifndef DONT_USE_MASK
    if (fmc.maskRect32.right > fmc.maskRect32.left)
    {   // mask is nonempty
        mskRectPtr = &fmc.maskRect32;
    }
    else
#endif
    {   // using alpha (layer transparency mask) as mask
        mskRectPtr = &fmc.inRect32;
    }

    if (gParams->flags & AFH_USE_HOST_PADDING)
    {
        topPad = bottomPad = leftPad = rightPad = padding;
    }
    else
    {
		// Get filterRect from host and scale it down
        scaleFactor = gStuff->inputRate >> 16;
		scaledFilterRect = GetFilterRect();
        scaleRect32 (&scaledFilterRect, 1, scaleFactor);
		// Calculate padding at each edge
        topPad    = min(padding, mskRectPtr->top - scaledFilterRect.top);
        bottomPad = min(padding, scaledFilterRect.bottom - mskRectPtr->bottom);
        leftPad   = min(padding, mskRectPtr->left - scaledFilterRect.left);
        rightPad  = min(padding, scaledFilterRect.right - mskRectPtr->right);
    }
	// Add padding to either fmc.maskRect32 or fmc.inRect32
    mskRectPtr->top    -= topPad;
    mskRectPtr->bottom += bottomPad;
    mskRectPtr->left   -= leftPad;
    mskRectPtr->right  += rightPad;

#ifndef DONT_USE_MASK
    if (fmc.maskRect32.right > fmc.maskRect32.left)
    {   
		// Update maskRect for host
		SetMaskRect(fmc.maskRect32);
		// Mask was nonempty, so adjust input rectangle also
        fmc.inRect32.top    -= topPad;
        fmc.inRect32.bottom += bottomPad;
        fmc.inRect32.left   -= leftPad;
        fmc.inRect32.right  += rightPad;
    }
#endif
	// Update inRect for host
	SetInRect(fmc.inRect32);

} /*addPadding*/

static void subtractPadding (GPtr globals)
{
	// Remove padding from either fmc.maskRect32 or fmc.inRect32
    mskRectPtr->top    += topPad;
    mskRectPtr->bottom -= bottomPad;
    mskRectPtr->left   += leftPad;
    mskRectPtr->right  -= rightPad;

#ifndef DONT_USE_MASK
    if (fmc.maskRect32.right > fmc.maskRect32.left)
    {   
		// Update maskRect for host
		SetMaskRect(fmc.maskRect32);
		// Mask was nonempty, so adjust input rectangle also
        fmc.inRect32.top    += topPad;
        fmc.inRect32.bottom -= bottomPad;
        fmc.inRect32.left   += leftPad;
        fmc.inRect32.right  -= rightPad;
        // PSP 4.12 treats maskData as read-only, so we need to make
        // our own (adjusted) copy.
        MyAdjustedMaskData = (unsigned8 *) gStuff->maskData + (topPad*gStuff->maskRowBytes + leftPad);
                                                            // note: maskColumnBytes=1
    }
#endif
	// BUG FIX:
	// Some hosts (e.g., PSP X3) set inColumnBytes incorrectly,
	// so we set it correctly ourselves...
	if (gStuff->imageMode > 9) //16-bit image data
		gStuff->inColumnBytes = 2 * ((gStuff->inHiPlane - gStuff->inLoPlane) + 1);	
	else
		gStuff->inColumnBytes = (gStuff->inHiPlane - gStuff->inLoPlane) + 1;
	
	// PSP 4.12 treats inData as read-only, so we need to make
    // our own (adjusted) copy.
	MyAdjustedInData = (unsigned8 *)gStuff->inData + (topPad*gStuff->inRowBytes + leftPad*gStuff->inColumnBytes);
	// Update inRect for host
	SetInRect(fmc.inRect32);
} /*subtractPadding*/

/*****************************************************************************/
//
//  Stuff used to select "desirable" Proxy Preview scale factors...
//
/// 100%, 50%, 33.3%, 25%, 20%, 16.7%, 14.3%, 12.5%, 11.1%, 10%, 9.09%, 8.33%, (7.69%),
/// 7.14%, (6.67%), 6.25%, (5.88%), (5.55%), (5.26%), 5%, (4.76%),...,4%,...,3.33%,...,2.5%,
/// ...,2%,...,1.49%,...,1.33%,...,1%,...0.50%,...,0.33%,...,0.25%,...0.20%,...,0.17%,
/// ...,0.14%,...,0.12%,...,0.11%,...,0.10%,...0.05%,...0.03%,...0.02%,...,0.01%,...,
/// .005%, .002%, .001%, .0005%, .0002%, .0001%, .00005%, .00002%, .00001% <== brings 2G down to 200

#if 0
static int desirableScaleFactor[] = {
	1,		// 100%
	2,		// 50%
	3,		// 33.3%
	4,		// 25%
	5,		// 20%
	6,		// 16.7%
	7,		// 14.3%
	8,		// 12.5%
	9,		// 11.1%
	10,		// 10%
	11,		// 9.09%
	12,		// 8.33%
///	13,		//			7.69%	*skip*
	14,		// 7.14%
///	15,		//			6.67%	*skip*
	16,		// 6.25%
///	17,		//			5.88%	*skip*
///	18,		//			5.55%	*skip*
///	19,		//			5.26%	*skip*
	20,		// 5%
	25,		// 4%
	30,		// 3.33%
///	33,		//			3.03%	*skip*
	40,		// 2.5%
	50,		// 2%
	57,		// 1.75%
///	60,		//			1.67%	*skip*
	67,		// 1.49%
///	70,		//			1.43%	*skip*
	75,		// 1.33%
	80,		// 1.25%
///	90,		//			1.11%	*skip*
	91,		// 1.1%
	100,	// 1%
///	111,	//			0.901%	*skip*
	125,	// 0.8%
///	133,	//			0.752	*skip*
///	143,	//			0.699%	*skip*
///	167,	//			0.599	*skip*
	200,	// 0.5%
	250,	// 0.4%
///	300,	//			0.333%	*skip*
	333,	// 0.3%
	400,	// 0.25%
	500,	// 0.2%
///	600,	//			0.167%	*skip*
	625,	// 0.16%
	667,	// 0.15%
///	700,	//			0.143%	*skip*
	800,	// 0.125%
///	900,	//			0.111%	*skip*
	909,	// 0.11%
	1000,	// 0.1%
	1111,	// 0.09%
	1250,	// 0.08%
///	1333,	//			0.075	*skip*
	1429,	// 0.07%
	1667,	// 0.06%
	2000,	// 0.05%
	2500,	// 0.04%
///	3000,	//			0.0333%	*skip*
	3333,	// 0.03%
	5000,	// 0.02%
	10000,	// 0.01%
	11111,	// 0.009
	12500,	// 0.008%
///	13333,	//			0.0075%	*skip*
	14286,	// 0.007%
	16667,	// 0.006%
	20000,	// 0.005%
	25000,	// 0.004%
	33333,	// 0.003%
	50000,	// 0.002%
	100000,	// 0.001%
	111111,	// 0.0009%
	125000,	// 0.0008%
///	133333,	//			0.00075%	*skip*
	142857,	// 0.0007%
	166667, // 0.0006%
	200000,	// 0.0005%
	250000,	// 0.0004%
	333333,	// 0.0003%
	500000,	// 0.0002%
	1000000,// 0.0001%
	1111111,// 0.00009%
	1250000,// 0.00008%
///	1333333,//			0.000075%	*skip*
	1428571,// 0.00007%
	1666667,// 0.00006%
	2000000,// 0.00005%
	2500000,// 0.00004%
	3333333,// 0.00003%
	5000000,// 0.00002%
	10000000,// 0.00001%	//scales 2Gpx to 200px, so probably max reduction ever needed!
	INT_MAX	// ~infinity (stopper)
}; /**desirableScaleFactor*/
#else
//Using a roughly logarithmic series, with a step of 10**0.1 ~= 1.25 (1/0.8),
//which descends in one decade something like:
//100, 80, 64, 50, 40, 32, 25, 20, 16, 12.5, 10
static int desirableScaleFactor[] = {
	1,		// 100%
	//		// 80%		//no can do
	//		// 64%		//no can do
	2,		// 50%
	//		// 40%		//no can do
	3,		// 33.3%
	4,		// 25%
	5,		// 20%
	6,		// 16.7%
	8,		// 12.5%

	10,		// 10%
	12,		// 8.33%
	16,		// 6.25%
	20,		// 5%
	25,		// 4%
	30,		// 3.33%	//or 33 //3.03%??
	40,		// 2.5%
	50,		// 2%
	60,		// 1.67%
	80,		// 1.25%

	100,	// 1%
	125,	// 0.8%
	160,	// 0.625%	// or 156,	// 0.641%
	200,	// 0.5%
	250,	// 0.4%
	300,	// 0.333%
	400,	// 0.25%
	500,	// 0.2%
	625,	// 0.16%
	800,	// 0.125%

	1000,	// 0.1%
	1250,	// 0.08%
	1562,	// 0.064%
	2000,	// 0.05%
	2500,	// 0.04%
	3000,	// 0.0333%	//or 3125, // 0.032
	4000,	// 0.025%
	5000,	// 0.02%
	6250,	// 0.016%
	8000,	// 0.0125%

	10000,	// 0.01%
	12500,	// 0.008%
	15625,	// 0.0064%
	20000,	// 0.005%
	25000,	// 0.004%
	30000,	// 0.00333%
#if 0	///////////////// OOPS. For now scaleFactor can't exceed 30,000 ( or 32767).
		/////////////////       This limit is imposed because inputRate and MaskRate
		/////////////////		Are fixed point 16.16 numbers, and scaleFactor must
		/////////////////		fit in the high (signed) 16 bits!  This is a PS limitation,
		/////////////////		even in CS5.1 X64.  So if scaleFactor is <= 30000 and
		/////////////////		we want to scale an image to fit entirely within a 300 x 300 px
		/////////////////		frame, then no image dimension should exceeed 300px * 30000 =
		/////////////////		9,000,000 px, or approxmately 10Mpx.  Of course, this is well
		/////////////////		within the (published) CS5 limit of 300,000 x 300,000 px.
	40000,	// 0.0025%
	50000,	// 0.002%
	62500,	// 0.0016%
	80000,	// 0.00125%

	100000,	// 0.001%
	125000,	// 0.0008%
	156250,	// 0.00064%
	200000,	// 0.0005%
	250000,	// 0.0004%
	300000,	// 0.000333%
	400000,	// 0.00025%
	500000,	// 0.0002%
	625000,	// 0.00016%
	800000,	// 0.00125%

	1000000,// 0.0001%
	1250000,// 0.00008%
	1562500,// 0.000064%
	2000000,// 0.00005%
	2500000,// 0.00004%
	3000000,// 0.0000333%
	4000000,// 0.000025%
	5000000,// 0.00002%
	6250000,// 0.000016%
	8000000,// 0.000125%
	10000000,// 0.00001%	//scales 2Gpx to 200px, so probably max reduction ever needed!
#endif
	INT_MAX	// ~infinity (stopper)
}; /**desirableScaleFactor*/
#endif

/* Round up to a desirable scale factor */
int roundUpScaleFactor(int sf) {
	int i;
	for (i = 0; sf > desirableScaleFactor[i]; ++i) {
	}
	assert (sf <= desirableScaleFactor[i]);
	return desirableScaleFactor[i];
} /*roundUpScaleFactor*/

/* Round down to a desirable scale factor */
int roundDownScaleFactor(int sf) {
	int i;
	for (i = 0; sf > desirableScaleFactor[i]; ++i) {
	}
	assert (sf <= desirableScaleFactor[i]);
	if (sf < desirableScaleFactor[i])
		return desirableScaleFactor[i-1];
	return desirableScaleFactor[i];
} /*roundDownScaleFactor*/

#if 0
/* Round down to a desirable scale factor */
int roundDownScaleFactor2(int sf)
{
	int i = sizeof(desirableScaleFactor)/sizeof(desirableScaleFactor[0]) - 1;
	for ( ; i >= 0; --i) {
		if (sf >= desirableScaleFactor[i])
			return desirableScaleFactor[i];
	}
	return desirableScaleFactor[i];
} /*roundDownScaleFactor2*/
#endif

/*****************************************************************************/

void SetupFilterRecordForProxy (GPtr globals, long *scaleFactor, long prevScaleFactor,
                                POINT *offset)
{
    long    maxScaleFactor;
    int     proxW, proxH, inW, inH;
    VRect   scaledImage;
    VRect   savedImage, savedOut;
#ifndef DONT_USE_MASK
    VRect   savedMask;
#endif
    int     oops;
    static long oldScaleFactor = 9999;
    static BOOL scaleFactorChanged = FALSE;
    BOOL    proxyWasTrimmed;

    fmc.doingProxy = TRUE;

	fmc.filterRect32 = GetFilterRect();
    CalcProxyScaleFactor (&fmc.filterRect32, &gProxyRect, &maxScaleFactor);

    if (*scaleFactor == -888)
    {   // -888 means use calculated optimal scale factor...
        *scaleFactor = maxScaleFactor;
    }

#ifdef LIMIT_MAX_SCALEFACTOR
    // Optionally limit scalefactor to <= maxScaleFactor.
    if (*scaleFactor > maxScaleFactor) *scaleFactor = maxScaleFactor;
#endif

    // Make sure scale factor is in range [1,MAX_SCALE_FACTOR].
    if (*scaleFactor < 1) *scaleFactor = 1;
    if (*scaleFactor > MAX_SCALE_FACTOR) *scaleFactor = MAX_SCALE_FACTOR;
#if 0
    // Skip scalefactors of 13, 15, 17-19, 21-24, 26-39, 31-39, 41-49,
	//						51-59, 61-69, 71-79, 81-89, 91-99
	if (*scaleFactor < 13) {
		; //use 1 - 12 as is (no rounding up)
	} else if (*scaleFactor <= 14) {
		*scaleFactor = 14;	//round 13 or 14 up to 14
	} else if (*scaleFactor <= 16) {
		*scaleFactor = 16;	//round 15 or 16 up to 16
	} else if (*scaleFactor <= 20) {
		*scaleFactor = 20;	//round 17 thru 20 to 20, etc.
	} else if (*scaleFactor <= 25) {
		*scaleFactor = 25;
	} else if (*scaleFactor <= 30) {
		*scaleFactor = 30;
	} else if (*scaleFactor <= 40) {
		*scaleFactor = 40;
	} else if (*scaleFactor <= 50) {
		*scaleFactor = 50;
	} else if (*scaleFactor <= 60) {
		*scaleFactor = 60;
	} else if (*scaleFactor <= 70) {
		*scaleFactor = 70;
	} else if (*scaleFactor <= 80) {
		*scaleFactor = 80;
	} else if (*scaleFactor <= 90) {
		*scaleFactor = 90;
	} else if (*scaleFactor <= 100) {
		*scaleFactor = 100;
	}
#else
	*scaleFactor = roundUpScaleFactor(*scaleFactor);
#endif
    if (*scaleFactor != oldScaleFactor)
    {
        scaleFactorChanged = TRUE;
        oldScaleFactor = *scaleFactor;
        fmc.zoomFactor = *scaleFactor;
    }

#if 1
	fmc.inRect32 = fmc.filterRect32;
	scaleRect32(&fmc.inRect32, 1, *scaleFactor);
	scaledImage = fmc.inRect32;
#else
    copyRect (&gStuff->inRect, &gStuff->filterRect);
    scaleRect (&gStuff->inRect, 1, *scaleFactor);
    copyRect (&scaledImage, &gStuff->inRect);
#endif

    // Trim (scaled) inRect to fit within proxyRect if necessary...
    proxyWasTrimmed = FALSE;
    proxW = gProxyRect.right - gProxyRect.left;
    proxH = gProxyRect.bottom - gProxyRect.top;
    inW = fmc.inRect32.right - fmc.inRect32.left;
    inH = fmc.inRect32.bottom - fmc.inRect32.top;
    if (inW > proxW) {
        fmc.inRect32.left += (inW - proxW)/2;
        fmc.inRect32.right = fmc.inRect32.left + proxW;
        proxyWasTrimmed = TRUE;
    }
    if (inH > proxH) {
        fmc.inRect32.top += (inH - proxH)/2;
        fmc.inRect32.bottom = fmc.inRect32.top + proxH;
        proxyWasTrimmed = TRUE;
    }

    // Apply (scaled) offset...
#ifdef HIGHZOOM
    fmc.inRect32.left   += (offset->x / (*scaleFactor) / fmc.enlargeFactor);
    fmc.inRect32.right  += (offset->x / (*scaleFactor) / fmc.enlargeFactor);
    fmc.inRect32.top    += (offset->y / (*scaleFactor) / fmc.enlargeFactor);
    fmc.inRect32.bottom += (offset->y / (*scaleFactor) / fmc.enlargeFactor);
#else
	fmc.inRect32.left   += (offset->x / *scaleFactor);
    fmc.inRect32.right  += (offset->x / *scaleFactor);
    fmc.inRect32.top    += (offset->y / *scaleFactor);
    fmc.inRect32.bottom += (offset->y / *scaleFactor);
#endif

	//fm_setCtlTextv(0, "%d, %d - %d, %d", fmc.inRect32.left,fmc.inRect32.right,fmc.inRect32.top,fmc.inRect32.bottom ); 

    if (fmc.inRect32.left < scaledImage.left)
    {   // Oops.  Too far left...
        oops = scaledImage.left - fmc.inRect32.left;
        fmc.inRect32.left  += oops;
        fmc.inRect32.right += oops;
        offset->x += oops * (*scaleFactor);
    }
    else if (fmc.inRect32.right > scaledImage.right)
    {   // Oops.  Too far right...
        oops = fmc.inRect32.right - scaledImage.right;
        fmc.inRect32.left  -= oops;
        fmc.inRect32.right -= oops;
        offset->x -= oops * (*scaleFactor);
    }
    if (fmc.inRect32.top < scaledImage.top)
    {   // Oops.  Too far up...
        oops = scaledImage.top - fmc.inRect32.top;
        fmc.inRect32.top    += oops;
        fmc.inRect32.bottom += oops;
        offset->y += oops * (*scaleFactor);
    }
    else if (fmc.inRect32.bottom > scaledImage.bottom)
    {   // Oops.  Too far down...
        oops = fmc.inRect32.bottom - scaledImage.bottom;
        fmc.inRect32.top    -= oops;
        fmc.inRect32.bottom -= oops;
        offset->y -= oops * (*scaleFactor);
    }
	SetInRect(fmc.inRect32);

    if ((gParams->flags & (AFH_DRAGGING_PROXY|AFH_ZERO_OUTPUT_DRAG|AFH_DRAG_FILTERED_OUTPUT))
                       == (AFH_DRAGGING_PROXY|AFH_ZERO_OUTPUT_DRAG))
    {   // Don't need to set up output Rect.  Could save some time.
        PISetRect32(&fmc.outRect32, 0, 0, 0, 0);
    }
    else
    {
		fmc.outRect32 = fmc.inRect32;
    }
	SetOutRect(fmc.outRect32);

    gStuff->inLoPlane = gStuff->outLoPlane = 0;
    gStuff->inHiPlane = gStuff->outHiPlane = gStuff->planes - 1;

#ifndef DONT_USE_MASK
	if (gStuff->haveMask) {
		fmc.maskRect32 = fmc.inRect32;
	}
    else
#endif
	{
        PISetRect32 (&fmc.maskRect32, 0, 0, 0, 0);
	}
	SetMaskRect(fmc.maskRect32);

    gStuff->inputRate = gStuff->maskRate = long2fixed(*scaleFactor);
 
    if (gParams->flags & AFH_USE_HOST_PADDING)
    {
        gStuff->inputPadding = 0;   // 0=black.
        gStuff->outputPadding = 0;  // 0=black.
        gStuff->maskPadding = 0;    // 0=black (non-selected).
    }
    else
    {
        gStuff->inputPadding = plugInDoesNotWantPadding;
        gStuff->outputPadding = plugInDoesNotWantPadding;
        gStuff->maskPadding = plugInDoesNotWantPadding;
    }



    // If the Big Gulp option is set, try to speed up proxy
    // dragging by requesting the entire (scaled) image whenever
    // the scale factor first changes (and the scaled image
    // needs to be trimmed to fit in the proxy window).
    if ((gParams->flags & AFH_PROXY_BIG_GULP) && scaleFactorChanged && proxyWasTrimmed)
    {
        short tempResult;

        scaleFactorChanged = FALSE;
        // Save the current inRect, outRect, and (maybe) maskRect,
        // and set them to the full (scaled) image rectangle.
		savedImage = fmc.inRect32;
		fmc.inRect32 = scaledImage;
		SetInRect(fmc.inRect32);

		savedOut = fmc.outRect32;

        if ((gParams->flags & (AFH_DRAGGING_PROXY|AFH_ZERO_OUTPUT_DRAG|AFH_DRAG_FILTERED_OUTPUT))
                           == (AFH_DRAGGING_PROXY|AFH_ZERO_OUTPUT_DRAG))
        {   // Don't need to set up output Rect.  Could save some time.
            PISetRect32(&fmc.outRect32, 0, 0, 0, 0);
        }
        else
        {
			fmc.outRect32 = scaledImage;
        }
		SetOutRect(fmc.outRect32);

#ifndef DONT_USE_MASK
        if (gStuff->haveMask)
        {
			savedMask = fmc.maskRect32;
			fmc.maskRect32 = scaledImage;
			SetMaskRect(fmc.maskRect32);
        }
#endif

        addPadding(globals);

#ifndef NOTRACE
#if 0
        // Temporarily clear AFH_DRAGGING_PROXY so Trace can emit output.
		{//scope
		unsigned long savedFlags = gParams->flags;
        gParams->flags &= ~AFH_DRAGGING_PROXY;
#endif
        Trace (globals,
               "Calling AdvanceState for Big Gulp...\n"
               " inRect = %d, %d, %d, %d\n"
               " outRect = %d, %d, %d, %d\n"
               " maskRect = %d, %d, %d, %d",
               fmc.inRect32.left, fmc.inRect32.top,
               fmc.inRect32.right, fmc.inRect32.bottom,
               fmc.outRect32.left, fmc.outRect32.top,
               fmc.outRect32.right, fmc.outRect32.bottom,
               fmc.maskRect32.left, fmc.maskRect32.top,
               fmc.maskRect32.right, fmc.maskRect32.bottom);
#endif //NOTRACE

        // Call AdvanceState() to try to gulp in the entire (scaled) image.
        // The result is ignored, because we don't care if it fails (due to
        // insufficient memory).
        tempResult = AdvanceState();

#ifdef _DEBUG
        if (tempResult != noErr) {
            msgBox( MB_ICONHAND|MB_SYSTEMMODAL,
                    fmc.filterTitleText,//"FilterMeister",
                    "Big gulp failed.  AdvanceState returned %d",
                    tempResult
                    );
        }
#endif //_DEBUG
        //subtractPadding(globals);  // no need here

#ifndef NOTRACE
        Trace (globals, "AdvanceState returned %d\n"
                        " inData = %#x"
                        " outData = %#x"
                        " maskData = %#x",
                        gResult,
                        gStuff->inData,
                        gStuff->outData,
                        gStuff->maskData);
#if 0
        // Restore AFH_DRAGGING_PROXY flag.
        gParams->flags = savedFlags;
		}//end scope
#endif
#endif //NOTRACE

        // Restore the current values of inRect, outRect, and maskRect.
		fmc.inRect32 = savedImage;
		SetInRect(fmc.inRect32);
		fmc.outRect32 = savedOut;
		SetOutRect(fmc.outRect32);
#ifndef DONT_USE_MASK
        if (gStuff->haveMask)
        {
			fmc.maskRect32 = savedMask;
			SetMaskRect(fmc.maskRect32);
        }
#endif
    } /*if Big Gulp*/

    

    // Now call AdvanceState() for real, to get the actual proxy pixels.
    addPadding(globals);

#ifndef NOTRACE
    Trace (globals,
           "Calling AdvanceState from SetupFilterRecordForProxy...\n"
           " inRect = %d, %d, %d, %d\n"
           " outRect = %d, %d, %d, %d\n"
           " maskRect = %d, %d, %d, %d",
           fmc.inRect32.left, fmc.inRect32.top,
           fmc.inRect32.right, fmc.inRect32.bottom,
           fmc.outRect32.left, fmc.outRect32.top,
           fmc.outRect32.right, fmc.outRect32.bottom,
           fmc.maskRect32.left, fmc.maskRect32.top,
           fmc.maskRect32.right, fmc.maskRect32.bottom);
#endif

    gResult = AdvanceState();

    if (gResult != noErr) {
        // We use the special combo MB_ICONHAND|MB_SYSTEMMODAL
        // to ensure that the message box will be displayed
        // regardless of low memory...
        if (gResult == memFullErr) {
            msgBox( MB_ICONHAND|MB_SYSTEMMODAL,
                    fmc.filterTitleText,//"FilterMeister",
                    "Not enough memory to generate a preview\n"
                    "image at this zoom level."
                    );
        }
        else {
            msgBox( MB_ICONHAND|MB_SYSTEMMODAL,
                    fmc.filterTitleText,//"FilterMeister",
                    "Unknown error (%d) while trying to generate\n"
                    "a preview image.",
                    gResult
                    );
        }
    }
    
#ifndef NOTRACE
    Trace (globals, "AdvanceState returned %d\n"
                    " inData = %#x"
                    " outData = %#x"
                    " maskData = %#x",
                    gResult,
                    gStuff->inData,
                    gStuff->outData,
                    gStuff->maskData);
#endif
    subtractPadding(globals);

    // Need to set outRect correctly for ShowOutput();
	fmc.inRect32 = GetInRect();	//needed??
	fmc.outRect32 = fmc.inRect32;
	SetOutRect(fmc.outRect32);

    // Now gStuff->inData should have the (possibly padded) subsampled version of input data,
    // gStuff->maskData may have the (possibly padded) subsampled version of the mask data,
    // and gStuff->outData should be a buffer the size of input data that
    // could be used as a temporary buffer to do intermediate filtering
    // to display proxies

    // MyAdjustedInData and MyAdjustedMaskData will point to the start of the NON-padded
    // data.

} /*SetupFilterRecordForProxy*/

/*****************************************************************************/

/* Prepare to filter an image.  If the plug-in filter needs a large amount
   of buffer memory, this routine should set the bufferSpace field to the
   number of bytes required. */

void DoPrepare (GPtr globals)
{

	int maxSpace = gStuff->maxSpace;

    ValidateParameters (globals);

    if (gResult != noErr) return;

#ifndef NOTRACE
    Trace (globals, "DoPrepare:\n"
        " gStuff->bufferSpace = %d\n"
        " (before setting it to 0)",
        gStuff->bufferSpace);
#endif
    
    
    //Added by Harald Heim, February 2003
    //Suppress dialog when there is no filter parameter block
    //For Plugin Commander Pro
    if (gStuff == NULL)
        gParams->queryForParameters = FALSE; 


/*
#ifdef STANDALONE

	InitContextRecord();

	if (!fm_triggerEvent(0, FME_DOPREPARE, 0)){
		//Restore values in case false was returned;
		gStuff->maxSpace = maxSpace;	
		gStuff->bufferSpace = 0;
	}
#else */
	gStuff->bufferSpace = 0;
//#endif

    
    //Added by Harald Heim, Sep 26, 2003
    //Make Photoshop free up all the memory ??? -> Doesn't work
    //gStuff->maxSpace = 0;



    //Added by Harald Heim, Jul 6, 2002
    //if (gStuff->isFloating==FALSE && gStuff->haveMask==TRUE) 
    //      gStuff->autoMask = FALSE;

    
}

/*****************************************************************************/

//do initialization at start of each filter run...

static void InitBeforeFilter(void)
{
    int i;

    //initialize random number generator for the proxy image
    srand(1);

    // Set gamma = 1.0...
    fm_setGamma(1.0);

    //initialize anonymous cells
    if (Dont_Initialize_Cells == 0)
    {
        for (i = 0; i < N_CELLS; i++) {
            fmc.cell[i] = 0;
        }
    }

    // Cache current control values in fmc.pre_ctl[].
    for (i=0; i < N_CTLS; i++) {
        fmc.pre_ctl[i] = gParams->ctl[i].val;
    }

    //init fmc.scaleFactor, etc. ?????

	//Catch any longjmp's back...
    if (setjmp(Mde_error_exit_jb)) {
        //someone longjmp'ed back to us...
        return;   //abort
    }

    //Call the OnFilterStart handler...
    if (gParams->fValidCompile) {
        //for now, we ignore the return value
        INT_PTR res = X86_call_fmf1(gParams->CB_onFilterStart, (INT_PTR)&fmc);
    }

} /*InitBeforeFilter*/

/*****************************************************************************/

void DoStart (GPtr globals)
{

    //Added by Harald Heim
    int i;
	Boolean displayDialog = true;
	//OSErr err;


    ValidateParameters (globals);

    if (gResult != noErr) return;

#ifndef NOTRACE
    Trace (globals, "DoStart:\n"
        " gParams->queryForParameters = %#x",
        gParams->queryForParameters);
#endif
    
    // Initialize static parts of the FM context record.
    InitContextRecord();

    // Synchronize FM context flags to global (parameter) flags...
    fmc.use_MMX = (gParams->flags & AFH_USE_MMX) != 0;
    fmc.use_CMOV = (gParams->flags & AFH_USE_CMOV) != 0;
    fmc.use_SSE = (gParams->flags & AFH_USE_SSE) != 0;
    fmc.isTileable = (gParams->flags & AFH_IS_TILEABLE) != 0;

    // Copy info from host's filter record...
    
    fmc.hostSerialNumber = gStuff->serialNumber;
    fmc.displayDialog = gParams->queryForParameters;
    fmc.doingScripting = 0;
    fmc.planes = gStuff->planes;

	// Following fields are available in 32-bits if bigDocumentData is
	// non-NULL; else available only in 16 bits:
	fmc.bigDocument = gStuff->bigDocumentData != NULL;
#if 0	//old method
	if (fmc.bigDocument) {
		// Use the 32-bit host-suppplied fields.
		// Signal the host that we will be using 32-bit coordinates.
		gStuff->bigDocumentData->PluginUsing32BitCoordinates = 1;

		fmc.imageSize32		   = gStuff->bigDocumentData->imageSize32;
		fmc.filterRect32	   = gStuff->bigDocumentData->filterRect32;
		fmc.inRect32		   = gStuff->bigDocumentData->inRect32;
		fmc.outRect32		   = gStuff->bigDocumentData->outRect32;
		fmc.maskRect32		   = gStuff->bigDocumentData->maskRect32;
		fmc.floatCoord32	   = gStuff->bigDocumentData->floatCoord32;
		fmc.wholeSize32		   = gStuff->bigDocumentData->wholeSize32;
	}
	else {
		// Host app uses only 16-bit coordinates. Copy them to
		// 32-bit fields for our internal use.
		fmc.imageSize32.v = gStuff->imageSize.v;
		fmc.imageSize32.h = gStuff->imageSize.h;

		fmc.filterRect32.left   = gStuff->filterRect.left;
		fmc.filterRect32.top    = gStuff->filterRect.top;
		fmc.filterRect32.right  = gStuff->filterRect.right;
		fmc.filterRect32.bottom = gStuff->filterRect.bottom;

		fmc.inRect32.left   = gStuff->inRect.left;
		fmc.inRect32.top    = gStuff->inRect.top;
		fmc.inRect32.right  = gStuff->inRect.right;
		fmc.inRect32.bottom = gStuff->inRect.bottom;

		fmc.outRect32.left   = gStuff->outRect.left;
		fmc.outRect32.top    = gStuff->outRect.top;
		fmc.outRect32.right  = gStuff->outRect.right;
		fmc.outRect32.bottom = gStuff->outRect.bottom;

		fmc.maskRect32.left   = gStuff->maskRect.left;
		fmc.maskRect32.top    = gStuff->maskRect.top;
		fmc.maskRect32.right  = gStuff->maskRect.right;
		fmc.maskRect32.bottom = gStuff->maskRect.bottom;

		fmc.floatCoord32.v = gStuff->floatCoord.v;
		fmc.floatCoord32.h = gStuff->floatCoord.h;

		fmc.wholeSize32.v = gStuff->wholeSize.v;
		fmc.wholeSize32.h = gStuff->wholeSize.h;
	}
#else //Use the new Get*Rect functions
	// Following routines check at runtime whether to use 16- or 32-bit host fields...
	fmc.imageSize32	= GetImageSize();
	fmc.filterRect32 = GetFilterRect();
	fmc.inRect32 = GetInRect();
	fmc.outRect32 = GetOutRect();
	fmc.maskRect32 = GetMaskRect();
	fmc.floatCoord32 = GetFloatCoord();
	fmc.wholeSize32 = GetWholeSize();
#endif

	// Following two fields are only exposed in the FD's dictionary...
	fmc.filterRectWidth32   = fmc.filterRect32.right - fmc.filterRect32.left;
	fmc.filterRectHeight32  = fmc.filterRect32.bottom - fmc.filterRect32.top;

    
    if (gStuff->imageMode == plugInModeCMYKColor || gStuff->imageMode == plugInModeCMYK64)
        fmc.planesWithoutAlpha = 4;
    else if (gStuff->imageMode == plugInModeGrayScale || gStuff->imageMode == plugInModeGray16 
        || gStuff->imageMode == plugInModeDuotone || gStuff->imageMode == plugInModeDuotone16
        || gStuff->imageMode == plugInModeBitmap || gStuff->imageMode == plugInModeIndexedColor)
        fmc.planesWithoutAlpha = 1;
    else /*if (gStuff->imageMode ==  plugInModeMultichannel || gStuff->imageMode == plugInModeDeepMultichannel
        || gStuff->imageMode == plugInModeRGBColor || gStuff->imageMode == plugInModeLabColor
        || gStuff->imageMode == plugInModeHSLColor || gStuff->imageMode == plugInModeHSBColor
        || gStuff->imageMode == plugInModeRGB48 || gStuff->imageMode == plugInModeLab48)*/
        fmc.planesWithoutAlpha = 3;
    

    fmc.maxSpace = gStuff->maxSpace;
	fmc.bufferSpace = gStuff->bufferSpace;
    fmc.isFloating = gStuff->isFloating;
    fmc.haveMask = gStuff->haveMask;

    //Added by Harald Heim, Jul 5, 2002
    fmc.autoMask = gStuff->autoMask;
    
    fmc.backColor = (COLORREF)fm_RGBA(gStuff->backColor[0],
									  gStuff->backColor[1],
									  gStuff->backColor[2],
									  gStuff->backColor[3]);
    fmc.foreColor = (COLORREF)fm_RGBA(gStuff->foreColor[0],
									  gStuff->foreColor[1],
									  gStuff->foreColor[2],
									  gStuff->foreColor[3]);
    // Need to mask out garbage in unused bytes of back/foreColor...
    {   int mask = 0xffffffff;
        switch (gStuff->imageMode)
        {
        case plugInModeBitmap:
        case plugInModeGrayScale:
        case plugInModeIndexedColor:    //can this happen???
        case plugInModeGray16:          //correct???
            mask = 0x000000ff;
            break;

        case plugInModeDuotone:
        case plugInModeDuotone16:

            //mask = 0x0000ffff;          //is duotone always 2 channels???
            mask = 0x00ffffff;          //duotone can have RGB fore and background colors!
            break;

        case plugInModeRGBColor:
        case plugInModeLabColor:
        case plugInModeHSLColor:        //can this happen???
        case plugInModeHSBColor:        //can this happen???
        case plugInModeRGB48:           //correct???      
        case plugInModeLab48:   

            mask = 0x00ffffff;
            break;

        case plugInModeMultichannel:     
        case plugInModeDeepMultichannel:
            
            mask = 0x00ffffff;
            break;

        case plugInModeCMYKColor:
        case plugInModeCMYK64:

        default:
            mask = 0xffffffff;
            break;
        } /*switch*/
        fmc.backColor &= mask;
        fmc.foreColor &= mask;
    } //scope

    //scale 16-bit back/foreground values down to 8 bits...     //Is this correct ???
    fmc.bgColor = (COLORREF)fm_RGB(gStuff->background.red >> 8,
								  gStuff->background.green >> 8,
								  gStuff->background.blue >> 8);
    fmc.fgColor = (COLORREF)fm_RGB(gStuff->foreground.red >> 8,
								  gStuff->foreground.green >> 8,
								  gStuff->foreground.blue >> 8);
    fmc.hostSig = gStuff->hostSig;

    
    
    fmc.platformData = gStuff->platformData;	//X64
    fmc.imageHRes = (gStuff->imageHRes >> 16) + (gStuff->imageHRes & 0xffff) / 65536.0; //(int) gStuff->imageHRes; //>> 16
    fmc.imageVRes = (gStuff->imageVRes >> 16) + (gStuff->imageVRes & 0xffff) / 65536.0; //(int) gStuff->imageVRes; //>> 16


    fmc.imageMode = gStuff->imageMode;
    /*if (gStuff->filterCase == 0){
		if (gStuff->haveMask==0 && gStuff->isFloating==0) gStuff->filterCase = 1;
		if (gStuff->haveMask==1 && gStuff->isFloating==0) gStuff->filterCase = 2;
		if (gStuff->isFloating==1) gStuff->filterCase = 3;
	}*/
	fmc.filterCase = gStuff->filterCase;
	fmc.samplingSupport = gStuff->samplingSupport;
    fmc.zoomFactor = 0;     //not yet set
	fmc.enlargeFactor = 1;
	fmc.scrollFactor = 1;

    // Restore values in FM context record across invocations...
    fmc.need_iuv = gParams->need_iuv;
    fmc.need_d = gParams->need_d;
    fmc.need_m = gParams->need_m;
    fmc.need_tbuf = gParams->need_tbuf;
    fmc.need_t2buf = gParams->need_t2buf;
	fmc.need_t3buf = gParams->need_t3buf;
	fmc.need_t4buf = gParams->need_t4buf;
    fmc.needPremiere = gParams->needPremiere;
    fmc.needPadding = gParams->needPadding;
    fmc.bandWidth = gParams->bandWidth;
#ifndef DONT_SAVE_GLOBALS
    /* Restore the following predeclared global and user-declared global/static variables */
    fmc.globvars = gParams->globvars;
#endif //DONT_SAVE_GLOBALS
    memcpy(fmc.filterAuthorText, gParams->filterAuthorText, sizeof(fmc.filterAuthorText));
    memcpy(fmc.filterCategoryText, gParams->filterCategoryText, sizeof(fmc.filterCategoryText));
    memcpy(fmc.filterTitleText, gParams->filterTitleText, sizeof(fmc.filterTitleText));
    memcpy(fmc.filterFilenameText, gParams->filterFilenameText, sizeof(fmc.filterFilenameText));
    memcpy(fmc.filterCopyrightText, gParams->filterCopyrightText, sizeof(fmc.filterCopyrightText));
    memcpy(fmc.filterDescriptionText, gParams->filterDescriptionText, sizeof(fmc.filterDescriptionText));
    memcpy(fmc.filterOrganizationText, gParams->filterOrganizationText, sizeof(fmc.filterOrganizationText));
    memcpy(fmc.filterURLText, gParams->filterURLText, sizeof(fmc.filterURLText));
    memcpy(fmc.filterVersionText, gParams->filterVersionText, sizeof(fmc.filterVersionText));
    memcpy(fmc.filterUniqueID, gParams->filterUniqueID, sizeof(fmc.filterUniqueID));
    fmc.ndlits = gParams->ndlits;
    memcpy(fmc.dlitPool, gParams->dlitPool, sizeof(fmc.dlitPool));

    // Get filter installation directory...
    fm_getFilterInstallDir();

#ifndef STANDALONE
    // Restore CB's across invocations, for use by Make...
    // May need to un-relocate the code ptrs????
    cbCodeBuffer = gParams->cbCodeBuffer;
    cbInitCode = gParams->cbInitCode;
    cbDivCheck = gParams->cbDivCheck;
    cbZeroDivide = gParams->cbZeroDivide;
    cbDivideOverflow = gParams->cbDivideOverflow;
    cbOnCtl = gParams->cbOnCtl;
    cbOnFilterStart = gParams->cbOnFilterStart;
    cbForEveryTile = gParams->cbForEveryTile;
    cbOnFilterEnd = gParams->cbOnFilterEnd;
    cbUserFunctions = gParams->cbUserFunctions;
#endif

#ifdef STANDALONE
	fmc.DESIGNTIME=false;
#else
	fmc.DESIGNTIME=true;
#endif

	fmc.abort_mode = -1;	//use default for execution platform (also set at start of each compilation)

    // Check for valid mode...
#if 1
    ///// Temp kludge!!!!  Allow all modes for now...
    gResult = noErr;
#else
    gResult = filterBadMode;    // PSP 4.0 doesn't recognize this!!
#endif

    switch (gStuff->imageMode)
    {
    case plugInModeBitmap:
        strcpy(fmc.filterImageModeText, "Bitmap");
        if (gParams->flags & AFH_SUPPORT_BITMAP)
            gResult = noErr;
        break;

    case plugInModeGrayScale:
        strcpy(fmc.filterImageModeText, "Grayscale");
        if (gParams->flags & AFH_SUPPORT_GRAY)
            gResult = noErr;
        break;

    case plugInModeDuotone:
        strcpy(fmc.filterImageModeText, "Duotone");
        if (gParams->flags & AFH_SUPPORT_DUOTONE)
            gResult = noErr;
        break;

    case plugInModeIndexedColor:
        strcpy(fmc.filterImageModeText, "Indexed Color");
        if (gParams->flags & AFH_SUPPORT_INDEXED)
            gResult = noErr;
        break;

    case plugInModeRGBColor:
        strcpy(fmc.filterImageModeText, "RGB Color");
        if (gParams->flags & AFH_SUPPORT_RGB)
            gResult = noErr;
        break;

    case plugInModeCMYKColor:
        strcpy(fmc.filterImageModeText, "CMYK Color");
        if (gParams->flags & AFH_SUPPORT_CMYK)
            gResult = noErr;
        break;

    case plugInModeLabColor:
        strcpy(fmc.filterImageModeText, "Lab Color");
        if (gParams->flags & AFH_SUPPORT_LAB)
            gResult = noErr;
        break;

    case plugInModeMultichannel:
        strcpy(fmc.filterImageModeText, "Multichannel");
        if (gParams->flags & AFH_SUPPORT_MULTI)
            gResult = noErr;
        break;

    case plugInModeHSLColor:
        strcpy(fmc.filterImageModeText, "HSL Color");
        break;

    case plugInModeHSBColor:
        strcpy(fmc.filterImageModeText, "HSB Color");
        break;

    case plugInModeGray16:
        strcpy(fmc.filterImageModeText, "16-bit Grayscale");
        break;

    case plugInModeRGB48:
        strcpy(fmc.filterImageModeText, "48-bit RGB Color");
        break;

//Added by Harald Heim, December 6, 2002

    case plugInModeLab48:
        strcpy(fmc.filterImageModeText, "48-bit Lab Color");
        break;
    
    case plugInModeCMYK64:
        strcpy(fmc.filterImageModeText, "64-bit CMYK Color");
        break;

    case plugInModeDeepMultichannel:
        strcpy(fmc.filterImageModeText, "Deep Multichannel");
        break;

    case plugInModeDuotone16:
        strcpy(fmc.filterImageModeText, "16-bit Duotone");
        break;


    
    default:
        strcpy(fmc.filterImageModeText, "Unknown Image Mode");
        break;
    } /*switch*/


    if (gResult != noErr)
    {
        if (gResult == filterBadMode && gStuff->hostSig != '8BIM')
        {   // PSP and others don't recognize the filterBadMode error,
            // so put out our own message box...
            ShowAlert(16501);
            gResult = 1;    // indicates error message already displayed.
        }
        return;
    }

    
	//set default to allocation from the host
	fm_set_array_mode(1);

    //Initialize Arrays
    for (i=0; i<N_ARRAYS; i++){
                Array[i] = 0;
				ArrayID[i] = 0;
                ArrayPadding[i]=0;
				ArraySize[i]=0;
				ArrayBytes[i]=0;
				ArrayX[i]=0;
				ArrayY[i]=0; 
				ArrayZ[i]=0;
    }

    //Initialize Font Array
    for (i=0; i<N_FONTS; i++){
        Font[i] = 0;		
    }



	
	
#ifdef SCRIPTABLE

    //Initialize Script Array
    for (i=0; i<256; i++){
        ScriptParam[i] = 0;
        ScriptParamUsed[i] = false;
    }

    ReadScriptParameters(&displayDialog); //err = ReadScriptParameters(&displayDialog);
    fmc.displayDialog = displayDialog;
    fmc.doingScripting = ScriptParamRead;

#endif


	if (fmc.displayDialog)
    {
        //Display Dialog
        
        if (CheckProxiesSupported (gStuff->displayPixels) && WarnAdvanceStateAvailable())
            PromptUserForInput (globals);   // Show the UI
        else
            PromptUserForInputNoProxy (globals);


		//Unhook Key Capture
		if (KeyHook != NULL) UnhookWindowsHookEx(KeyHook);



#ifndef DEMO
        if (gResult == noErr){

            //user hit OK

            gParams->queryForParameters = FALSE;
			
			//Clean up threads
			fm_terminateThread(0);
    

        } else
#endif
        {

            //user hit Cancel, exit button or some other error

			if (gParams->gDialogEvent & 2) fm_triggerEvent(0, FME_CANCEL, 0);

			//Clean up threads
			fm_terminateThread(0);
			
            // Kill any async sound that might still be running...
            fm_playSoundWave(0);


            //Added by Harald Heim
            //Delete Array if available
            for (i=0; i<N_ARRAYS; i++) fm_freeArray(i);

            //Delete Fonts
            for (i=0; i<N_FONTS; i++){
                    if (Font[i] != 0) DeleteObject(Font[i]);
            }


            //Added by Harald Heim
            //free distance buffer if necessary
            if (bufferID != 0)
            {
                UnlockBuffer(bufferID);
                FreeBuffer(bufferID);
                bufferID=0;
            }


            #ifdef DEMO
                //Display when user presses OK
                if (gResult == noErr) 
                    Info ("This is a demo version which only lets you see\n"
                          "the final effect in the preview box.\n\n"
                          "To apply the effect to the image you need the full version.\n"
                          "You can order a full version of %s at\n\n"
                          "%s",
                          fmc.filterTitleText, fmc.filterURLText);

            #endif

            // Maybe should also copy the global vars to the param block here????

            return;

        }


    }
    else
    {
        // No Dialog will be displayed
        // don't need to ask for params
        // ...but may need to adjust passed-in params; e.g.,
        //    if the filterCase has changed...

    #ifdef STANDALONE
        if (!gParams->fValidCompile){
            if (LoadFmCode()) {
                gParams->fValidCompile = true;
            }
        }
        /*if (gParams->gInitialInvocation) {		//Who commented this out, and why??
            // Load code from FMCODE resource
			if (LoadFmCode()) {
                // Execute the initialization code...
                gParams->fValidCompile = TRUE;
                if (setjmp(Mde_error_exit_jb)) {
                    //Someone longjmp'ed back to us...
                    gParams->fValidCompile = FALSE;
                }
                else {
                    //execute init code
                    X86_call_fmf1(gParams->CB_initCode, (INT_PTR)gFmcp);
                }
            }
        }*/
    #endif


    #ifdef SCRIPTABLE
        //Write the script params to the controls
        if (ScriptParamRead){
            int n;
            for (n=0;n<256;n++) {
                if (ScriptParamUsed[n]) gFmcp->pre_ctl[n] = gParams->ctl[n].val = ScriptParam[n]; //fm_setCtlVal(n, ScriptParam[n]);
            }
        }
    #endif


    #ifdef FM_LIFTCOV
        // Force the "Transparent Background" checkbox to FALSE unless we
        // have an editable transparency layer (alpha channel).
        if (gStuff->filterCase != filterCaseEditableTransparencyNoSelection &&
            gStuff->filterCase != filterCaseEditableTransparencyWithSelection)
        {
            gParams->chk[0].val = FALSE;
        }
    #endif


    }


    //Finally apply filter to image...

#ifndef DEMO
    //set doingProxy FALSE again, since PromptUserForInput may have
    //set it TRUE...
    fmc.doingProxy = FALSE;

    //Perform pre-filter initializations...
    InitBeforeFilter();

    if (WarnAdvanceStateAvailable())
        StartWithAdvanceState(globals);
    else
        StartNoAdvanceState(globals);

	//Catch any longjmp's back...
    if (setjmp(Mde_error_exit_jb)) {
        //someone longjmp'ed back to us...
        return;   //abort -- or copy globals???
    }

    //Call the OnFilterEnd handler...
    if (gParams->fValidCompile) {
        //for now, we ignore the return value
        INT_PTR res = X86_call_fmf1(gParams->CB_onFilterEnd, (INT_PTR)&fmc);
    }
#endif

    // Maybe here is a good place to copy the global vars to the param block????
#ifndef DONT_SAVE_GLOBALS
    /* Save the following predeclared global and user-declared global/static variables */
    gParams->globvars = gFmcp->globvars;
#endif //DONT_SAVE_GLOBALS

} /*DoStart*/

/******************************************************************************/
/* Does all of the filtering with no advance state and proxies. */


#ifndef DEMO

void StartNoAdvanceState (GPtr globals)
{
    DoInitialRect(globals);

} /*StartNoAdvanceState*/
    
#endif


/*****************************************************************************/
/* Does all of the filtering. */

#ifndef DEMO

void StartWithAdvanceState (GPtr globals)
{
    int row;
    int total;
    int rowDelta = gParams->bandWidth;

    // Do the actual filtering operation on the original image
    fmc.filterRect32 = GetFilterRect();
    total = fmc.filterRect32.bottom - fmc.filterRect32.top;

    // If the algorithm isn't tileable, grab the entire image...
    if (!fmc.isTileable) {
        rowDelta = total;
    }

    gStuff->inLoPlane = gStuff->outLoPlane = 0;
    gStuff->inHiPlane = gStuff->outHiPlane = gStuff->planes - 1;
    
	// Set left and right edges of inRect and outRect to the full
	// width of the filterRect.
#if 0
    gStuff->inRect.left = gStuff->outRect.left = gStuff->filterRect.left;
    gStuff->inRect.right = gStuff->outRect.right = gStuff->filterRect.right;
#else
	fmc.inRect32 = fmc.outRect32 = fmc.filterRect32;
	SetInRect(fmc.inRect32);
	SetOutRect(fmc.outRect32);
#endif

    PISetRect32(&fmc.maskRect32, 0, 0, 0, 0);
	SetMaskRect(fmc.maskRect32);

    gStuff->inputRate = long2fixed (1L);
    gStuff->maskRate = long2fixed (1L);
    
    if (gParams->flags & AFH_USE_HOST_PADDING)
    {
        gStuff->inputPadding = 0;   // 0=black.
        gStuff->outputPadding = 0;  // 0=black.
        gStuff->maskPadding = 0;    // 0=black (non-selected).
    }
    else
    {
        gStuff->inputPadding = plugInDoesNotWantPadding;
        gStuff->outputPadding = plugInDoesNotWantPadding;
        gStuff->maskPadding = plugInDoesNotWantPadding;
    }
    
    for (row = fmc.filterRect32.top; row < fmc.filterRect32.bottom; row += rowDelta)
    {
        
        if (TestAbort ())
        {
            gResult = userCanceledErr;
            goto done;
        }
            
        fmc.inRect32.top = fmc.outRect32.top = row;
        if (rowDelta > fmc.filterRect32.bottom - row)
            rowDelta = fmc.filterRect32.bottom - row;
        fmc.inRect32.bottom = fmc.outRect32.bottom = row + rowDelta;
		SetInRect(fmc.inRect32);
		SetOutRect(fmc.outRect32);
        
#ifndef DONT_USE_MASK
		if (gStuff->haveMask) {
            fmc.maskRect32 = fmc.inRect32;
			SetMaskRect(fmc.maskRect32);
		}
#endif

        addPadding(globals);

#ifndef NOTRACE
        Trace (globals,
               "Calling AdvanceState from StartWithAdvanceState...\n"
               " inRect = %d, %d, %d, %d\n"
               " outRect = %d, %d, %d, %d\n"
               " maskRect = %d, %d, %d, %d",
               fmc.inRect32.left, fmc.inRect32.top,
               fmc.inRect32.right, fmc.inRect32.bottom,
               fmc.outRect32.left, fmc.outRect32.top,
               fmc.outRect32.right, fmc.outRect32.bottom,
               fmc.maskRect32.left, fmc.maskRect32.top,
               fmc.maskRect32.right, fmc.maskRect32.bottom);
#endif

        gResult = AdvanceState ();
        
#ifndef NOTRACE
        Trace (globals, "AdvanceState returned %d\n"
                        " inData = %#x"
                        " outData = %#x"
                        " maskData = %#x",
                        gResult,
                        gStuff->inData,
                        gStuff->outData,
                        gStuff->maskData);
#endif

        subtractPadding(globals);

        if (gResult != noErr)
        {
#if 0
            ShowAlert (16500); //debug PSP 4.12
#endif
            goto done;
        }
        
        DoFilterRect (globals);
        
        if (gResult != noErr)
        {
            goto done;
        }
        
    } //for
    
done:
    // Zero the request Rectangles to signal a stop
    PISetRect32(&fmc.inRect32, 0, 0, 0, 0);
    PISetRect32(&fmc.outRect32, 0, 0, 0, 0);
    PISetRect32(&fmc.maskRect32, 0, 0, 0, 0);
	SetInRect(fmc.inRect32);
	SetOutRect(fmc.outRect32);
	SetMaskRect(fmc.maskRect32);

    if ((gParams->flags & AFH_FLUSH_ADVANCESTATE) && (gResult == noErr))
    {
#ifndef NOTRACE
        Trace (globals,
               "Flushing AdvanceState...\n"
               " inRect = %d, %d, %d, %d\n"
               " outRect = %d, %d, %d, %d\n"
               " maskRect = %d, %d, %d, %d",
               fmc.inRect32.left, fmc.inRect32.top,
               fmc.inRect32.right, fmc.inRect32.bottom,
               fmc.outRect32.left, fmc.outRect32.top,
               fmc.outRect32.right, fmc.outRect32.bottom,
               fmc.maskRect32.left, fmc.maskRect32.top,
               fmc.maskRect32.right, fmc.maskRect32.bottom);
#endif

        // Call AdvanceState() one more time to flush out PSP 4.12...
        gResult = AdvanceState ();
    
#ifndef NOTRACE
        Trace (globals, "AdvanceState returned %d\n"
                        " inData = %#x"
                        " outData = %#x"
                        " maskData = %#x",
                        gResult,
                        gStuff->inData,
                        gStuff->outData,
                        gStuff->maskData);
#endif
    } //if
} /*StartWithAdvanceState*/

#endif //DEMO


//----------------------------------------------------------------
//
//  A few utility routines
//

/**********************************************************************/
/* Map4DBUsToPixels(&x1, &y1, &x2, &y2)
/*
/*  Maps 4 coordinates from Dialog Box Units to pixels.
/*  x1 and x2 are mapped as horizontal coordinates;
/*  y1 and y2 are mapped as vertical coordinates.
/*  (Horizontal and vertical mappings are generally identical,
/*  but not always!)
/**********************************************************************/
void Map4DBUsToPixels(int *x1, int *y1, int *x2, int *y2) {
    RECT r;
    r.left = *x1;
    r.top = *y1;
    r.right = *x2;
    r.bottom = *y2;
    MapDialogRect(fmc.hDlg, &r);
    *x1 = r.left;
    *y1 = r.top;
    *x2 = r.right;
    *y2 = r.bottom;
} /*Map4DBUsToPixels*/

/**********************************************************************/
/* Map2DBUsToPixels(&x1, &y1)
/*
/*  Same as Map4DBUsToPixels(), but maps only 2 coordinates instead of 4.
/*  x1 is mapped as a horizontal coordinate;
/*  y1 is mapped as a vertical coordinate.
/**********************************************************************/
void Map2DBUsToPixels(int *x1, int *y1) {
    RECT r;
    r.left = *x1;
    r.top = *y1;
    r.right = 0;
    r.bottom = 0;
    MapDialogRect(fmc.hDlg, &r);
    *x1 = r.left;
    *y1 = r.top;
} /*Map2DBUsToPixels*/

/**********************************************************************/
/* HDBUsToPixels(h)
/*
/*  Returns the number of pixels comprising h horizontal DBUs.
/*
/**********************************************************************/
INT_PTR fm_HDBUsToPixels(INT_PTR h) {
    RECT r;
    r.left = (int)h;
    r.top = 0;
    r.right = 0;
    r.bottom = 0;
    MapDialogRect(fmc.hDlg, &r);
    return r.left;
} /*fm_HDBUsToPixels*/

/**********************************************************************/
/* VDBUsToPixels(v)
/*
/*  Returns the number of pixels comprising v vertical DBUs.
/*
/**********************************************************************/
INT_PTR fm_VDBUsToPixels(INT_PTR v) {
    RECT r;
    r.left = 0;
    r.top = (int)v;
    r.right = 0;
    r.bottom = 0;
    MapDialogRect(fmc.hDlg, &r);
    return r.top;
} /*fm_VDBUsToPixels*/

/**********************************************************************/
/* setGamma(gamma)
/*
/*  Precomputes the gamma tables using the specified value of gamma.
/*  Returns true if successful; else returns false.
/*
/**********************************************************************/
INT_PTR fm_setGamma(double gamma) {
    double r = 1.0/gamma;
    int i;
    for (i = 0; i < 256; i++) {
        fmc.pre_gamma[i] = (int)(pow(i / 255.0, r) * 255.0 + 0.5);
    }
    return TRUE;
} /*fm_setGamma*/

/**********************************************************************/
/* gamma(i)
/*
/*  Returns the gamma-corrected value of i, where 0 <= i < 256.
/*  Result is undefined if i is out of range.
/*
/**********************************************************************/
INT_PTR fm_gamma(INT_PTR i) {
    return fmc.pre_gamma[i & 0xff];
} /*fm_gamma*/

//----------------------------------------------------------------
// Utility routines for simulating Filter Factory

static INT_PTR fm_pset(INT_PTR x, INT_PTR y, INT_PTR z, INT_PTR val)
{
    // Clamp value to [0,255].
#if 1
    if ((unsigned)val > 255) {
		//needs to be clamped
        val = (val <= 0) ? 0 : 255;
    }
#elif 0
	//possibly faster if compiler generates straightline code,
	//else maybe slower...
	val = (val < 0)?0:(val > 255)?255:val;
#elif 0
	//generates straightline code (no branches), so possibly
	//faster, but as yet untested!!
	#define BITSMINUS1 (sizeof(INT_PTR)*8-1)
	val = ((((255 - val) >> BITSMINUS1) & 0xff)|val) & ~(val >> BITSMINUS1);
	#undef BITSMINUS1
#elif 0
	//if target cpu supports the CMOVcc instruction, could probably
	//use inline or external assembler to generate faster code similar
	//to:
	 _asm {
		 mov eax, val
		 mov ecx, 0			;or==> xor ecx, ecx
		 cmp eax, ecx
		 cmovl eax, ecx
		 mov ecx, 255		;or==> not cl		
		 cmp eax, ecx
		 cmovg eax, ecx
		 mov val, eax
	 }
#endif
    if (x >= fmc.x_start && x < fmc.x_end &&
        y >= fmc.y_start && y < fmc.y_end &&
        z >= 0 && z < fmc.dstColDelta )
    {   //valid coordinate.

/*
#ifndef DONT_USE_MASK

        if (fmc.doingProxy && gStuff->haveMask){ //Simulate final result in preview
            
            INT_PTR MaskValue = fm_msk(x,y); //fm_src(x,y,-1);
            
            if (MaskValue>0){
                if (MaskValue<255){
                    val = (val*MaskValue + fm_src(x,y,z)*(255-MaskValue) )/255;
                }
                    
                ((unsigned8 *)gStuff->outData)
					[(x-fmc.x_start)*fmc.dstColDelta + (y-fmc.y_start)*fmc.dstRowDelta + z]
				= (unsigned8)val;
            } 

        } else { //regular pset

            ((unsigned8 *)gStuff->outData)
				[(x-fmc.x_start)*fmc.dstColDelta + (y-fmc.y_start)*fmc.dstRowDelta + z] 
			= (unsigned8)val;
        }

#else
*/
        ((unsigned8 *)gStuff->outData)
			[(x-fmc.x_start)*fmc.dstColDelta + (y-fmc.y_start)*fmc.dstRowDelta + z] 
		= (unsigned8)val;
        
//#endif


    }
    return val; //clamped value
} /*fm_pset*/

static INT_PTR fm_pset16(INT_PTR x, INT_PTR y, INT_PTR z, INT_PTR val)
{
    // 16-bit version.
    // Clamp value to [0,0x8000].
#if 0
    if (val > 0x8000) val = 0x8000;
    else if (val < 0) val = 0;
#elif 0 //using unsigned should be faster...
    val = ((unsigned)val <= 0x8000) ? val : (val <= 0) ? 0 : 0x8000;
#else //or is this faster??
    if ((unsigned)val > 0x8000) {
		//needs to be clamped
        val = (val <= 0) ? 0 : 0x8000;
    }
#endif
    if (x >= fmc.x_start && x < fmc.x_end &&
        y >= fmc.y_start && y < fmc.y_end &&
        z >= 0 && z < fmc.dstColDelta )
    {   //valid coordinates.
        
/*        
#ifndef DONT_USE_MASK

      if (fmc.doingProxy && gStuff->haveMask){ //Simulate final result in preview
            
            INT_PTR MaskValue = fm_msk(x,y); //fm_src(x,y,-1);
            
            if (MaskValue>0){
                if (MaskValue<65535){
                    val = (val*MaskValue/2 + fm_src16(x,y,z)*(32768-MaskValue/2) )/32768; //32768
                }
                    
                *(unsigned16 *)(((unsigned8 *)gStuff->outData) +
                        (((x-fmc.x_start)*fmc.dstColDelta + z) << 1) +
                        (y-fmc.y_start)*fmc.dstRowDelta)
                    = (unsigned16)val;

            } 

        } else { //regular pset

                *(unsigned16 *)(((unsigned8 *)gStuff->outData) +
                        (((x-fmc.x_start)*fmc.dstColDelta + z) << 1) +
                        (y-fmc.y_start)*fmc.dstRowDelta)
                = (unsigned16)val;
        }

#else
*/        
            *(unsigned16 *)(((unsigned8 *)gStuff->outData) +
                        (((x-fmc.x_start)*fmc.dstColDelta + z) << 1) +
                        (y-fmc.y_start)*fmc.dstRowDelta)
                = (unsigned16)val;
//#endif


    }
    return val; //clamped value
} /*fm_pset16*/

static INT_PTR fm_tset(INT_PTR x, INT_PTR y, INT_PTR z, INT_PTR val)
{
#if 1
    if (!fmc.tbuf) {
        // tbuf wasn't allocated????
        ErrorOk("No tile buffer!");
        return 0;
    }
#endif
    // Store channel value in (temporary) tile buffer.
    // Clamp value to [0,255].
#if 0
    if (val > 255) val = 255;
    else if (val < 0) val = 0;
#elif 0 //using unsigned should be faster...
    val = ((unsigned)val <= 255) ? val : (val <= 0) ? 0 : 255;
#else //or is this faster??
    if ((unsigned)val > 255) {
        val = (val <= 0) ? 0 : 255;
    }//needs to be clamped
#endif
#if 0
    //Min version (tile buffer is size of output tile)
    if (x >= fmc.x_start && x < fmc.x_end &&
        y >= fmc.y_start && y < fmc.y_end &&
        z >= 0 && z < fmc.Z )
    {   //valid coordinate.
        //(optimize this later, after it's been debugged!!!)
        fmc.tbuf[((y-fmc.y_start)*fmc.columns + (x-fmc.x_start))*fmc.Z + z] = (BYTE)val;
    }
#else
    //Max version (tile buffer is size of input tile)
    if (x >= 0 && x < fmc.X &&
        y >= 0 && y < fmc.Y &&
        z >= 0 && z < fmc.Z )
    {   //valid coordinate.
        //(optimize this later, after it's been debugged!!!)
        fmc.tbuf[(y*fmc.X + x)*fmc.Z + z] = (BYTE)val;
    }
#endif
    return val; //clamped value
} /*fm_tset*/

static INT_PTR fm_tset16(INT_PTR x, INT_PTR y, INT_PTR z, INT_PTR val)
{
#if 1
    if (!fmc.tbuf) {
        // tbuf wasn't allocated????
        ErrorOk("No tile buffer!");
        return 0;
    }
#endif
    // Store channel value in (temporary) tile buffer.
    // Clamp value to [0,255].
    if (val > 32768) val = 32768;
    else if (val < 0) val = 0;
#if 0
    //Min version (tile buffer is size of output tile)
    if (x >= fmc.x_start && x < fmc.x_end &&
        y >= fmc.y_start && y < fmc.y_end &&
        z >= 0 && z < fmc.Z )
    {   //valid coordinate.
        //(optimize this later, after it's been debugged!!!)
        fmc.tbuf[((y-fmc.y_start)*fmc.columns + (x-fmc.x_start))*fmc.Z + z] = (unsigned16)val;
    }
#else
    //Max version (tile buffer is size of input tile)
    if (x >= 0 && x < fmc.X &&
        y >= 0 && y < fmc.Y &&
        z >= 0 && z < fmc.Z )
    {   //valid coordinate.
        //(optimize this later, after it's been debugged!!!)
        *(unsigned16 *)(((unsigned8 *)fmc.tbuf) + 2*((y*fmc.X + x)*fmc.Z + z)) = (unsigned16)val;
    }
#endif
    return val; //clamped value
} /*fm_tset16*/

static INT_PTR fm_t2set(INT_PTR x, INT_PTR y, INT_PTR z, INT_PTR val)
{
#if 1
    if (!fmc.t2buf) {
        // t2buf wasn't allocated????
        ErrorOk("No tile buffer 2!");
        return 0;
    }
#endif
    // Store channel value in (temporary) tile buffer 2.
    // Clamp value to [0,255].
#if 0
    if (val > 255) val = 255;
    else if (val < 0) val = 0;
#elif 0 //using unsigned should be faster...
    val = ((unsigned)val <= 255) ? val : (val <= 0) ? 0 : 255;
#else //or is this faster??
    if ((unsigned)val > 255) {
        val = (val <= 0) ? 0 : 255;
    }//needs to be clamped
#endif
#if 0
    //Min version (tile buffer is size of output tile)
    if (x >= fmc.x_start && x < fmc.x_end &&
        y >= fmc.y_start && y < fmc.y_end &&
        z >= 0 && z < fmc.Z )
    {   //valid coordinate.
        //(optimize this later, after it's been debugged!!!)
        fmc.t2buf[((y-fmc.y_start)*fmc.columns + (x-fmc.x_start))*fmc.Z + z] = (BYTE)val;
    }
#else
    //Max version (tile buffer is size of input tile)
    if (x >= 0 && x < fmc.X &&
        y >= 0 && y < fmc.Y &&
        z >= 0 && z < fmc.Z )
    {   //valid coordinate.
        //(optimize this later, after it's been debugged!!!)
        fmc.t2buf[(y*fmc.X + x)*fmc.Z + z] = (BYTE)val;
    }
#endif
    return val; //clamped value
} /*fm_t2set*/

static INT_PTR fm_t2set16(INT_PTR x, INT_PTR y, INT_PTR z, INT_PTR val)
{
#if 1
    if (!fmc.t2buf) {
        // t2buf wasn't allocated????
        ErrorOk("No tile buffer 2!");
        return 0;
    }
#endif
    // Store channel value in (temporary) tile buffer 2.
    // Clamp value to [0,255].
    if (val > 32768) val = 32768;
    else if (val < 0) val = 0;
#if 0
    //Min version (tile buffer is size of output tile)
    if (x >= fmc.x_start && x < fmc.x_end &&
        y >= fmc.y_start && y < fmc.y_end &&
        z >= 0 && z < fmc.Z )
    {   //valid coordinate.
        //(optimize this later, after it's been debugged!!!)
        fmc.t2buf[((y-fmc.y_start)*fmc.columns + (x-fmc.x_start))*fmc.Z + z] = (unsigned16)val;
    }
#else
    //Max version (tile buffer is size of input tile)
    if (x >= 0 && x < fmc.X &&
        y >= 0 && y < fmc.Y &&
        z >= 0 && z < fmc.Z )
    {   //valid coordinate.
        //(optimize this later, after it's been debugged!!!)
        *(unsigned16 *)(((unsigned8 *)fmc.t2buf) + 2*((y*fmc.X + x)*fmc.Z + z)) = (unsigned16)val;
    }
#endif
    return val; //clamped value
} /*fm_t2set16*/

static INT_PTR fm_t3set(INT_PTR x, INT_PTR y, INT_PTR z, INT_PTR val)
{
#if 1
    if (!fmc.t3buf) {
        // t3buf wasn't allocated????
        ErrorOk("No tile buffer 3!");
        return 0;
    }
#endif
    // Store channel value in (temporary) tile buffer 2.
    // Clamp value to [0,255].
    if (val > 255) val = 255;
    else if (val < 0) val = 0;
#if 0
    //Min version (tile buffer is size of output tile)
    if (x >= fmc.x_start && x < fmc.x_end &&
        y >= fmc.y_start && y < fmc.y_end &&
        z >= 0 && z < fmc.Z )
    {   //valid coordinate.
        //(optimize this later, after it's been debugged!!!)
        fmc.t3buf[((y-fmc.y_start)*fmc.columns + (x-fmc.x_start))*fmc.Z + z] = (BYTE)val;
    }
#else
    //Max version (tile buffer is size of input tile)
    if (x >= 0 && x < fmc.X &&
        y >= 0 && y < fmc.Y &&
        z >= 0 && z < fmc.Z )
    {   //valid coordinate.
        //(optimize this later, after it's been debugged!!!)
        fmc.t3buf[(y*fmc.X + x)*fmc.Z + z] = (BYTE)val;
    }
#endif
    return val; //clamped value
} /*fm_t3set*/

static INT_PTR fm_t3set16(INT_PTR x, INT_PTR y, INT_PTR z, INT_PTR val)
{
#if 1
    if (!fmc.t3buf) {
        // t3buf wasn't allocated????
        ErrorOk("No tile buffer 3!");
        return 0;
    }
#endif
    // Store channel value in (temporary) tile buffer 2.
    // Clamp value to [0,255].
    if (val > 32768) val = 32768;
    else if (val < 0) val = 0;
#if 0
    //Min version (tile buffer is size of output tile)
    if (x >= fmc.x_start && x < fmc.x_end &&
        y >= fmc.y_start && y < fmc.y_end &&
        z >= 0 && z < fmc.Z )
    {   //valid coordinate.
        //(optimize this later, after it's been debugged!!!)
        fmc.t3buf[((y-fmc.y_start)*fmc.columns + (x-fmc.x_start))*fmc.Z + z] = val;
    }
#else
    //Max version (tile buffer is size of input tile)
    if (x >= 0 && x < fmc.X &&
        y >= 0 && y < fmc.Y &&
        z >= 0 && z < fmc.Z )
    {   //valid coordinate.
        //(optimize this later, after it's been debugged!!!)
        *(unsigned16 *)(((unsigned8 *)fmc.t3buf) + 2*((y*fmc.X + x)*fmc.Z + z)) = (unsigned16)val;
    }
#endif
    return val; //clamped value
} /*fm_t3set16*/

static INT_PTR fm_t4set(INT_PTR x, INT_PTR y, INT_PTR z, INT_PTR val)
{
#if 1
    if (!fmc.t4buf) {
        // t4buf wasn't allocated????
        ErrorOk("No tile buffer 4!");
        return 0;
    }
#endif
    // Store channel value in (temporary) tile buffer 2.
    // Clamp value to [0,255].
    if (val > 255) val = 255;
    else if (val < 0) val = 0;
#if 0
    //Min version (tile buffer is size of output tile)
    if (x >= fmc.x_start && x < fmc.x_end &&
        y >= fmc.y_start && y < fmc.y_end &&
        z >= 0 && z < fmc.Z )
    {   //valid coordinate.
        //(optimize this later, after it's been debugged!!!)
        fmc.t4buf[((y-fmc.y_start)*fmc.columns + (x-fmc.x_start))*fmc.Z + z] = val;
    }
#else
    //Max version (tile buffer is size of input tile)
    if (x >= 0 && x < fmc.X &&
        y >= 0 && y < fmc.Y &&
        z >= 0 && z < fmc.Z )
    {   //valid coordinate.
        //(optimize this later, after it's been debugged!!!)
        fmc.t4buf[(y*fmc.X + x)*fmc.Z + z] = (BYTE)val;
    }
#endif
    return val; //clamped value
} /*fm_t4set*/

static INT_PTR fm_t4set16(INT_PTR x, INT_PTR y, INT_PTR z, INT_PTR val)
{
#if 1
    if (!fmc.t4buf) {
        // t4buf wasn't allocated????
        ErrorOk("No tile buffer 4!");
        return 0;
    }
#endif
    // Store channel value in (temporary) tile buffer 2.
    // Clamp value to [0,255].
    if (val > 32768) val = 32768;
    else if (val < 0) val = 0;
#if 0
    //Min version (tile buffer is size of output tile)
    if (x >= fmc.x_start && x < fmc.x_end &&
        y >= fmc.y_start && y < fmc.y_end &&
        z >= 0 && z < fmc.Z )
    {   //valid coordinate.
        //(optimize this later, after it's been debugged!!!)
        fmc.t4buf[((y-fmc.y_start)*fmc.columns + (x-fmc.x_start))*fmc.Z + z] = val;
    }
#else
    //Max version (tile buffer is size of input tile)
    if (x >= 0 && x < fmc.X &&
        y >= 0 && y < fmc.Y &&
        z >= 0 && z < fmc.Z )
    {   //valid coordinate.
        //(optimize this later, after it's been debugged!!!)
        *(unsigned16 *)(((unsigned8 *)fmc.t4buf) + 2*((y*fmc.X + x)*fmc.Z + z)) = (unsigned16)val;
    }
#endif
    return val; //clamped value
} /*fm_t4set16*/

static INT_PTR fm_pget(const INT_PTR x0, const INT_PTR y0, const INT_PTR z0)
// Must not modify its input parameters due to cnvX/Y in-line optimizations!!!	//X64???
{
    register INT_PTR x = x0, y = y0, z = z0;
    if (z >= 0 && z < fmc.dstColDelta)
    {   // get an rgba pixel
        // pin the coordinates (i.e., replicate edges)...
        // Should replicate at IMAGE edge, not TILE edge???
        if (x < fmc.x_start) x = fmc.x_start;
        else if (x >= fmc.x_end) x = fmc.x_end - 1;
        if (y < fmc.y_start) y = fmc.y_start;
        else if (y >= fmc.y_end) y = fmc.y_end - 1;
        //(optimize this later, after it's been debugged!!!)
        return ((unsigned8 *)gStuff->outData)
            [(x-fmc.x_start)*fmc.dstColDelta + (y-fmc.y_start)*fmc.dstRowDelta + z];
    }
    else {
        // Bad channel number???
        return 0;
    }
} /*fm_pget*/


//Added by Harald Heim Feb 9, 2002
static INT_PTR fm_pget16(const INT_PTR x0, const INT_PTR y0, const INT_PTR z0)
// Must not modify its input parameters due to cnvX/Y in-line optimizations!!!	//X64???
{
    register INT_PTR x = x0, y = y0, z = z0;

    if (z >= 0 && z < fmc.dstColDelta)
    {   // get an rgba pixel
        // pin the coordinates (i.e., replicate edges)...
        // Should replicate at IMAGE edge, not TILE edge???
        if (x < fmc.x_start) x = fmc.x_start;
        else if (x >= fmc.x_end) x = fmc.x_end - 1;
        if (y < fmc.y_start) y = fmc.y_start;
        else if (y >= fmc.y_end) y = fmc.y_end - 1;
        //(optimize this later, after it's been debugged!!!)
        return *(unsigned16 *)  ( ((unsigned8 *)gStuff->outData) +
            (((x-fmc.x_start)*fmc.dstColDelta + z)<<1 ) + (y-fmc.y_start)*fmc.dstRowDelta  );
    }
    else {
        // Bad channel number???
        return 0;
    }
} /*fm_pget16*/


static INT_PTR fm_tget(const INT_PTR x0, const INT_PTR y0, const INT_PTR z0)
// Must not modify its input parameters due to cnvX/Y in-line optimizations!!!	//X64???
{
    register INT_PTR x = x0, y = y0, z = z0;
#if 1
    if (!fmc.tbuf) {
        // tbuf wasn't allocated????
        ErrorOk("No tile buffer!");
        return 0;
    }
#endif
    // Get channel value from (temporary) tile buffer.
    if (z >= 0 && z < fmc.Z)
    {   // get an rgba pixel
        // pin the coordinates (i.e., replicate edges)...
        // Should replicate at IMAGE edge, not TILE edge???
#if 0
        //Min version (tile buffer is size of output tile)
        if (x < fmc.x_start) x = fmc.x_start;
        else if (x >= fmc.x_end) x = fmc.x_end - 1;
        if (y < fmc.y_start) y = fmc.y_start;
        else if (y >= fmc.y_end) y = fmc.y_end - 1;
        //(optimize this later, after it's been debugged!!!)
        return fmc.tbuf[((y-fmc.y_start)*fmc.columns + (x-fmc.x_start))*fmc.Z + z];
#else
        //Max version (tile buffer is size of input tile)
        if (x < 0) x = 0;
        else if (x >= fmc.X) x = fmc.X - 1;
        if (y < 0) y = 0;
        else if (y >= fmc.Y) y = fmc.Y - 1;
        //(optimize this later, after it's been debugged!!!)
        return fmc.tbuf[(y*fmc.X + x)*fmc.Z + z];
#endif
    }
    else {
        // Bad channel number???
        return 0;
    }
} /*fm_tget*/


static INT_PTR fm_tget16(const INT_PTR x0, const INT_PTR y0, const INT_PTR z0)
// Must not modify its input parameters due to cnvX/Y in-line optimizations!!!	//X64???
{
    register INT_PTR x = x0, y = y0, z = z0;
#if 1
    if (!fmc.tbuf) {
        // tbuf wasn't allocated????
        ErrorOk("No tile buffer!");
        return 0;
    }
#endif
    // Get channel value from (temporary) tile buffer.
    if (z >= 0 && z < fmc.Z)
    {   // get an rgba pixel
        // pin the coordinates (i.e., replicate edges)...
        // Should replicate at IMAGE edge, not TILE edge???
#if 0
        //Min version (tile buffer is size of output tile)
        if (x < fmc.x_start) x = fmc.x_start;
        else if (x >= fmc.x_end) x = fmc.x_end - 1;
        if (y < fmc.y_start) y = fmc.y_start;
        else if (y >= fmc.y_end) y = fmc.y_end - 1;
        //(optimize this later, after it's been debugged!!!)
        return fmc.tbuf[((y-fmc.y_start)*fmc.columns + (x-fmc.x_start))*fmc.Z + z];
#else
        //Max version (tile buffer is size of input tile)
        if (x < 0) x = 0;
        else if (x >= fmc.X) x = fmc.X - 1;
        if (y < 0) y = 0;
        else if (y >= fmc.Y) y = fmc.Y - 1;
        //(optimize this later, after it's been debugged!!!)
        return *(unsigned16 *)(((unsigned8 *)fmc.tbuf) + 2*((y*fmc.X + x)*fmc.Z + z));

#endif
    }
    else {
        // Bad channel number???
        return 0;
    }
} /*fm_tget16*/


static INT_PTR fm_t2get(const INT_PTR x0, const INT_PTR y0, const INT_PTR z0)
// Must not modify its input parameters due to cnvX/Y in-line optimizations!!!	//X64???
{
    register INT_PTR x = x0, y = y0, z = z0;
#if 1
    if (!fmc.t2buf) {
        // t2buf wasn't allocated????
        ErrorOk("No tile buffer 2!");
        return 0;
    }
#endif
    // Get channel value from (temporary) tile buffer 2.
    if (z >= 0 && z < fmc.Z)
    {   // get an rgba pixel
        // pin the coordinates (i.e., replicate edges)...
        // Should replicate at IMAGE edge, not TILE edge???
#if 0
        //Min version (tile buffer is size of output tile)
        if (x < fmc.x_start) x = fmc.x_start;
        else if (x >= fmc.x_end) x = fmc.x_end - 1;
        if (y < fmc.y_start) y = fmc.y_start;
        else if (y >= fmc.y_end) y = fmc.y_end - 1;
        //(optimize this later, after it's been debugged!!!)
        return fmc.t2buf[((y-fmc.y_start)*fmc.columns + (x-fmc.x_start))*fmc.Z + z];
#else
        //Max version (tile buffer is size of input tile)
        if (x < 0) x = 0;
        else if (x >= fmc.X) x = fmc.X - 1;
        if (y < 0) y = 0;
        else if (y >= fmc.Y) y = fmc.Y - 1;
        //(optimize this later, after it's been debugged!!!)
        return fmc.t2buf[(y*fmc.X + x)*fmc.Z + z];
#endif
    }
    else {
        // Bad channel number???
        return 0;
    }
} /*fm_t2get*/

static INT_PTR fm_t2get16(const INT_PTR x0, const INT_PTR y0, const INT_PTR z0)
// Must not modify its input parameters due to cnvX/Y in-line optimizations!!!	//X64???
{
    register INT_PTR x = x0, y = y0, z = z0;
#if 1
    if (!fmc.t2buf) {
        // t2buf wasn't allocated????
        ErrorOk("No tile buffer 2!");
        return 0;
    }
#endif
    // Get channel value from (temporary) tile buffer 2.
    if (z >= 0 && z < fmc.Z)
    {   // get an rgba pixel
        // pin the coordinates (i.e., replicate edges)...
        // Should replicate at IMAGE edge, not TILE edge???
#if 0
        //Min version (tile buffer is size of output tile)
        if (x < fmc.x_start) x = fmc.x_start;
        else if (x >= fmc.x_end) x = fmc.x_end - 1;
        if (y < fmc.y_start) y = fmc.y_start;
        else if (y >= fmc.y_end) y = fmc.y_end - 1;
        //(optimize this later, after it's been debugged!!!)
        return fmc.t2buf[((y-fmc.y_start)*fmc.columns + (x-fmc.x_start))*fmc.Z + z];
#else
        //Max version (tile buffer is size of input tile)
        if (x < 0) x = 0;
        else if (x >= fmc.X) x = fmc.X - 1;
        if (y < 0) y = 0;
        else if (y >= fmc.Y) y = fmc.Y - 1;
        //(optimize this later, after it's been debugged!!!)
		return *(unsigned16 *)(((unsigned8 *)fmc.t2buf) + 2*((y*fmc.X + x)*fmc.Z + z));
#endif
    }
    else {
        // Bad channel number???
        return 0;
    }
} /*fm_t2get16*/

static INT_PTR fm_t3get(const INT_PTR x0, const INT_PTR y0, const INT_PTR z0)
// Must not modify its input parameters due to cnvX/Y in-line optimizations!!!	//X64???
{
    register INT_PTR x = x0, y = y0, z = z0;
#if 1
    if (!fmc.t3buf) {
        // t3buf wasn't allocated????
        ErrorOk("No tile buffer 3!");
        return 0;
    }
#endif
    // Get channel value from (temporary) tile buffer 2.
    if (z >= 0 && z < fmc.Z)
    {   // get an rgba pixel
        // pin the coordinates (i.e., replicate edges)...
        // Should replicate at IMAGE edge, not TILE edge???
#if 0
        //Min version (tile buffer is size of output tile)
        if (x < fmc.x_start) x = fmc.x_start;
        else if (x >= fmc.x_end) x = fmc.x_end - 1;
        if (y < fmc.y_start) y = fmc.y_start;
        else if (y >= fmc.y_end) y = fmc.y_end - 1;
        //(optimize this later, after it's been debugged!!!)
        return fmc.t3buf[((y-fmc.y_start)*fmc.columns + (x-fmc.x_start))*fmc.Z + z];
#else
        //Max version (tile buffer is size of input tile)
        if (x < 0) x = 0;
        else if (x >= fmc.X) x = fmc.X - 1;
        if (y < 0) y = 0;
        else if (y >= fmc.Y) y = fmc.Y - 1;
        //(optimize this later, after it's been debugged!!!)
        return fmc.t3buf[(y*fmc.X + x)*fmc.Z + z];
#endif
    }
    else {
        // Bad channel number???
        return 0;
    }
} /*fm_t3get*/

static INT_PTR fm_t3get16(const INT_PTR x0, const INT_PTR y0, const INT_PTR z0)
// Must not modify its input parameters due to cnvX/Y in-line optimizations!!!	//X64???
{
    register INT_PTR x = x0, y = y0, z = z0;
#if 1
    if (!fmc.t3buf) {
        // t3buf wasn't allocated????
        ErrorOk("No tile buffer 3!");
        return 0;
    }
#endif
    // Get channel value from (temporary) tile buffer 2.
    if (z >= 0 && z < fmc.Z)
    {   // get an rgba pixel
        // pin the coordinates (i.e., replicate edges)...
        // Should replicate at IMAGE edge, not TILE edge???
#if 0
        //Min version (tile buffer is size of output tile)
        if (x < fmc.x_start) x = fmc.x_start;
        else if (x >= fmc.x_end) x = fmc.x_end - 1;
        if (y < fmc.y_start) y = fmc.y_start;
        else if (y >= fmc.y_end) y = fmc.y_end - 1;
        //(optimize this later, after it's been debugged!!!)
        return fmc.t3buf[((y-fmc.y_start)*fmc.columns + (x-fmc.x_start))*fmc.Z + z];
#else
        //Max version (tile buffer is size of input tile)
        if (x < 0) x = 0;
        else if (x >= fmc.X) x = fmc.X - 1;
        if (y < 0) y = 0;
        else if (y >= fmc.Y) y = fmc.Y - 1;
        //(optimize this later, after it's been debugged!!!)
		return *(unsigned16 *)(((unsigned8 *)fmc.t3buf) + 2*((y*fmc.X + x)*fmc.Z + z));
#endif
    }
    else {
        // Bad channel number???
        return 0;
    }
} /*fm_t3get16*/

static INT_PTR fm_t4get(const INT_PTR x0, const INT_PTR y0, const INT_PTR z0)
// Must not modify its input parameters due to cnvX/Y in-line optimizations!!!	//X64???
{
    register INT_PTR x = x0, y = y0, z = z0;
#if 1
    if (!fmc.t4buf) {
        // t4buf wasn't allocated????
        ErrorOk("No tile buffer 4!");
        return 0;
    }
#endif
    // Get channel value from (temporary) tile buffer 2.
    if (z >= 0 && z < fmc.Z)
    {   // get an rgba pixel
        // pin the coordinates (i.e., replicate edges)...
        // Should replicate at IMAGE edge, not TILE edge???
#if 0
        //Min version (tile buffer is size of output tile)
        if (x < fmc.x_start) x = fmc.x_start;
        else if (x >= fmc.x_end) x = fmc.x_end - 1;
        if (y < fmc.y_start) y = fmc.y_start;
        else if (y >= fmc.y_end) y = fmc.y_end - 1;
        //(optimize this later, after it's been debugged!!!)
        return fmc.t4buf[((y-fmc.y_start)*fmc.columns + (x-fmc.x_start))*fmc.Z + z];
#else
        //Max version (tile buffer is size of input tile)
        if (x < 0) x = 0;
        else if (x >= fmc.X) x = fmc.X - 1;
        if (y < 0) y = 0;
        else if (y >= fmc.Y) y = fmc.Y - 1;
        //(optimize this later, after it's been debugged!!!)
        return fmc.t4buf[(y*fmc.X + x)*fmc.Z + z];
#endif
    }
    else {
        // Bad channel number???
        return 0;
    }
} /*fm_t4get*/

static INT_PTR fm_t4get16(const INT_PTR x0, const INT_PTR y0, const INT_PTR z0)
// Must not modify its input parameters due to cnvX/Y in-line optimizations!!!	//X64???
{
    register INT_PTR x = x0, y = y0, z = z0;
#if 1
    if (!fmc.t4buf) {
        // t4buf wasn't allocated????
        ErrorOk("No tile buffer 4!");
        return 0;
    }
#endif
    // Get channel value from (temporary) tile buffer 2.
    if (z >= 0 && z < fmc.Z)
    {   // get an rgba pixel
        // pin the coordinates (i.e., replicate edges)...
        // Should replicate at IMAGE edge, not TILE edge???
#if 0
        //Min version (tile buffer is size of output tile)
        if (x < fmc.x_start) x = fmc.x_start;
        else if (x >= fmc.x_end) x = fmc.x_end - 1;
        if (y < fmc.y_start) y = fmc.y_start;
        else if (y >= fmc.y_end) y = fmc.y_end - 1;
        //(optimize this later, after it's been debugged!!!)
        return fmc.t4buf[((y-fmc.y_start)*fmc.columns + (x-fmc.x_start))*fmc.Z + z];
#else
        //Max version (tile buffer is size of input tile)
        if (x < 0) x = 0;
        else if (x >= fmc.X) x = fmc.X - 1;
        if (y < 0) y = 0;
        else if (y >= fmc.Y) y = fmc.Y - 1;
        //(optimize this later, after it's been debugged!!!)
		return *(unsigned16 *)(((unsigned8 *)fmc.t4buf) + 2*((y*fmc.X + x)*fmc.Z + z));
#endif
    }
    else {
        // Bad channel number???
        return 0;
    }
} /*fm_t4get16*/


static INT_PTR fm_src(const INT_PTR x0, const INT_PTR y0, const INT_PTR z0)
// Must not modify its input parameters due to cnvX/Y in-line optimizations!!!	//X64???
{   ///// THIS VERSION ONLY WORKS WHEN FILTERING ENTIRE RECT AT ONCE!!!
    ///// OOPS- That's NFG for the proxy if it's clipped -- so in addPadding,
    /////     we set the pad amount astronomically high to always grab the
    /////     entire rect (assuming DONT_USE_MASK is set).
    register INT_PTR x = x0, y = y0, z = z0;
    if (z >= 0 && z < fmc.planesWithoutAlpha) //3 //gStuff->planes //gStuff->inHiPlane
    {   // get an rgb pixel

        // pin the coordinates (i.e., replicate edges)...
        //Added by Harald Heim, Feb 13, 02
        switch (edgeMode)
        {
            case 0:
                if (x < 0) x = 0;
                else if (x >= fmc.X) x = fmc.X - 1;
                if (y < 0) y = 0;
                else if (y >= fmc.Y) y = fmc.Y - 1;
                break;

            case 1:
                if (x < 0 || y < 0 || x >= fmc.X || y >= fmc.Y)
                    return 0;
				break;			//was missing!!
        
            case 2:
                if (x < 0) x = fmc.X + x;                   /// or use MOD ???
                else if (x >= fmc.X) x = x - fmc.X;
                if (y < 0) y = fmc.Y  + y;
                else if (y >= fmc.Y) y = y - fmc.Y;
                break;
            
            case 3: //two reflection modes, one of period 2N, other of period 2N-2 (as here)?
                if (x < 0) x = - x;
                else if (x >= fmc.X) x = (fmc.X - 1) - (x - (fmc.X - 1));  //optimize to 2*fmc.X - x - 2 ???
                if (y < 0) y = - y;
                else if (y >= fmc.Y) y = (fmc.Y - 1) - (y - (fmc.Y - 1));
                break;

			case 4: //the other reflection mode, or period 2N (not yet enabled)
				if (x < 0) x = -x - 1; //or x = ~x --; //still doesn't wrap below -X !!
				else if (x >= fmc.X) x = 2*fmc.X - x - 1; //still doesn't wrap above 2X!
				if (y < 0) y = -y - 1; //or y = ~y --; //still doesn't wrap below -Y !!
				else if (y >= fmc.Y) y = 2*fmc.Y - y - 1; //still doesn't wrap above 2Y!
				break;

			default: //no checking??? (here and elsewhere...)
				break;
        }//end switch

        return ((unsigned8 *)MyAdjustedInData)
            [(x-fmc.x_start)*fmc.srcColDelta + (y-fmc.y_start)*fmc.srcRowDelta + z];
        // note that srcPlaneDelta is always 1
    }

    else if (z == 3 || z < fmc.planes) //==3
    {   // get a mask pixel
#if 1
        // pin the coordinates (i.e., replicate edges)...
        //Added by Harald Heim, Jul 2, 02
        switch (edgeMode)
        {
            case 0:
                if (x < 0) x = 0;
                else if (x >= fmc.X) x = fmc.X - 1;
                if (y < 0) y = 0;
                else if (y >= fmc.Y) y = fmc.Y - 1;
                break;

            case 1:
                if (x < 0 || y < 0 || x >= fmc.X || y >= fmc.Y)
                    return 0;
				break;			//was missing!!
        
            case 2:
                if (x < 0) x = fmc.X + x;
                else if (x >= fmc.X) x = x - fmc.X;
                if (y < 0) y = fmc.Y + y;
                else if (y >= fmc.Y) y = y - fmc.Y;
                break;
            
            case 3:
                if (x < 0) x = - x;
                else if (x >= fmc.X) x = (fmc.X - 1) - (x - (fmc.X - 1));
                if (y < 0) y = - y;
                else if (y >= fmc.Y) y = (fmc.Y - 1) - (y - (fmc.Y - 1));
                break;
        }//end switch
#else
        // alpha outside image/selection is assumed 0...
        if (x < 0 || x >= fmc.X || y < 0 || y >= fmc.Y)
        {   // pin the coordinates (return 0 if outside, NOT replicate edges!!!)
            return 0;
        }
        else
#endif

        //Causes problems when reading the alpha of greyscale images, so z==3 was added
        if (gNoMask && z==3)
        {   // entire filter rect is presumed selected
            return 255;
        }

//Commented out by Harald Heim, Feb 09, 2002
/*
#ifndef DONT_USE_MASK
        else if (gStuff->haveMask)
        {   // get value from mask data
            return ((unsigned8 *)MyAdjustedMaskData)
                [(x-fmc.x_start)*fmc.mskColDelta + (y-fmc.y_start)*fmc.mskRowDelta];
        }
#endif
*/
        //else
        //{   // get value from alpha channel
            return ((unsigned8 *)MyAdjustedInData)
                [(x-fmc.x_start)*fmc.srcColDelta + (y-fmc.y_start)*fmc.srcRowDelta + z]; //3 //gStuff->planes
        //}
    }

    else
    {   // bad channel number??
        return 0;   // ???
    }
} /*fm_src*/

static INT_PTR fm_src16(const INT_PTR x0, const INT_PTR y0, const INT_PTR z0)
// Must not modify its input parameters due to cnvX/Y in-line optimizations!!!	//X664??
{   // 16-bit version.
    ///// THIS VERSION ONLY WORKS WHEN FILTERING ENTIRE RECT AT ONCE!!!
    ///// OOPS- That's NFG for the proxy if it's clipped -- so in addPadding,
    /////     we set the pad amount astronomically high to always grab the
    /////     entire rect (assuming DONT_USE_MASK is set).
    register INT_PTR x = x0, y = y0, z = z0;
    if (z >= 0 && z < fmc.planesWithoutAlpha) //3 //gStuff->planes
    {   // get an rgba... pixel
        // pin the coordinates (i.e., replicate edges)...
        
		switch (edgeMode)
        {
            case 0:
                if (x < 0) x = 0;
                else if (x >= fmc.X) x = fmc.X - 1;
                if (y < 0) y = 0;
                else if (y >= fmc.Y) y = fmc.Y - 1;
                break;

            case 1:
                if (x < 0 || y < 0 || x >= fmc.X || y >= fmc.Y)
                    return 0;
				break;			//was still missing!!
        
            case 2:
                if (x < 0) x = fmc.X + x;                   /// or use MOD ???
                else if (x >= fmc.X) x = x - fmc.X;
                if (y < 0) y = fmc.Y  + y;
                else if (y >= fmc.Y) y = y - fmc.Y;
                break;
            
            case 3: //two reflection modes, one of period 2N, other of period 2N-2 (as here)?
                if (x < 0) x = - x;
                else if (x >= fmc.X) x = (fmc.X - 1) - (x - (fmc.X - 1));       //optimize?
                if (y < 0) y = - y;
                else if (y >= fmc.Y) y = (fmc.Y - 1) - (y - (fmc.Y - 1));
                break;
        }//end switch

        return *(unsigned16 *)(((unsigned8 *)MyAdjustedInData) +
                               (((x-fmc.x_start)*fmc.srcColDelta + z) << 1) +
                               (y-fmc.y_start)*fmc.srcRowDelta);
    }


    // Must be upated for Photoshop CS where we have transparency in 16bit mode!!!

    else if (z == 3 || z < fmc.planes) //==3 //gStuff->planes
    {   // get a fictitious alpha channel pixel
        // alpha outside image/selection is assumed 0...
        //if (x < 0 || x >= fmc.X || y < 0 || y >= fmc.Y)
        //{   // pin the coordinates (return 0 if outside, NOT replicate edges!!!)
        //    return 0;
        //}
        //else {//still needs to be tested
            switch (edgeMode)
			{
				case 0:
					if (x < 0) x = 0;
					else if (x >= fmc.X) x = fmc.X - 1;
					if (y < 0) y = 0;
					else if (y >= fmc.Y) y = fmc.Y - 1;
					break;

				case 1:
					if (x < 0 || y < 0 || x >= fmc.X || y >= fmc.Y)
						return 0;
					break;			//was still missing!!
        
				case 2:
					if (x < 0) x = fmc.X + x;                   /// or use MOD ???
					else if (x >= fmc.X) x = x - fmc.X;
					if (y < 0) y = fmc.Y  + y;
					else if (y >= fmc.Y) y = y - fmc.Y;
					break;
            
				case 3: //two reflection modes, one of period 2N, other of period 2N-2 (as here)?
					if (x < 0) x = - x;
					else if (x >= fmc.X) x = (fmc.X - 1) - (x - (fmc.X - 1));       //optimize?
					if (y < 0) y = - y;
					else if (y >= fmc.Y) y = (fmc.Y - 1) - (y - (fmc.Y - 1));
					break;
			}//end switch


			//Causes problems when reading the alpha of greyscale images, so z==3 was added
			if (gNoMask && z==3)
			{   // entire filter rect is presumed selected
				return 32768;
			}
			
			return *(unsigned16 *)(((unsigned8 *)MyAdjustedInData) +
                               (((x-fmc.x_start)*fmc.srcColDelta + z) << 1) + //3 //+gStuff->planes
                               (y-fmc.y_start)*fmc.srcRowDelta);
            //return  65535; //0x8000;  // assume max opacity
        //}
    }
    else
    {   // invalid channel number
        return 0;
    }
} /*fm_src16*/


static INT_PTR fm_r2x(INT_PTR d, INT_PTR m)
{
    return (INT_PTR)((double)m * cos( (twopi/1024.0)*(double)d ) + 0.0); // trunc or round???
}

static INT_PTR fm_r2y(INT_PTR d, INT_PTR m)
{
    return (INT_PTR)((double)m * sin( (twopi/1024.0)*(double)d ) + 0.0); // trunc or round???
}

static INT_PTR fm_rad(INT_PTR d, INT_PTR m, INT_PTR z)
{
    return fmc.src(fm_r2x(d,m) + (fmc.X/2),
                   fm_r2y(d,m) + (fmc.Y/2),
                   z);
}

static INT_PTR fm_pgetr(INT_PTR d, INT_PTR m, INT_PTR z)
{
    return fm_pget(fm_r2x(d,m) + (fmc.X/2),
                   fm_r2y(d,m) + (fmc.Y/2),
                   z);
}

static INT_PTR fm_pgetr16(INT_PTR d, INT_PTR m, INT_PTR z)
{
    return fm_pget16(fm_r2x(d,m) + (fmc.X/2),
                   fm_r2y(d,m) + (fmc.Y/2),
                   z);
}

static INT_PTR fm_psetr(INT_PTR d, INT_PTR m, INT_PTR z, INT_PTR val)
{
    return fm_pset(fm_r2x(d,m) + (fmc.X/2),
                   fm_r2y(d,m) + (fmc.Y/2),
                   z, val);
}

static INT_PTR fm_tgetr(INT_PTR d, INT_PTR m, INT_PTR z)
{
    return fm_tget(fm_r2x(d,m) + (fmc.X/2),
                   fm_r2y(d,m) + (fmc.Y/2),
                   z);
}

static INT_PTR fm_tsetr(INT_PTR d, INT_PTR m, INT_PTR z, INT_PTR val)
{
    return fm_tset(fm_r2x(d,m) + (fmc.X/2),
                   fm_r2y(d,m) + (fmc.Y/2),
                   z, val);
}

static INT_PTR fm_t2getr(INT_PTR d, INT_PTR m, INT_PTR z)
{
    return fm_t2get(fm_r2x(d,m) + (fmc.X/2),
                    fm_r2y(d,m) + (fmc.Y/2),
                    z);
}

static INT_PTR fm_t2setr(INT_PTR d, INT_PTR m, INT_PTR z, INT_PTR val)
{
    return fm_t2set(fm_r2x(d,m) + (fmc.X/2),
                    fm_r2y(d,m) + (fmc.Y/2),
                    z, val);
}

static INT_PTR fm_t3getr(INT_PTR d, INT_PTR m, INT_PTR z)
{
    return fm_t3get(fm_r2x(d,m) + (fmc.X/2),
                    fm_r2y(d,m) + (fmc.Y/2),
                    z);
}

static INT_PTR fm_t3setr(INT_PTR d, INT_PTR m, INT_PTR z, INT_PTR val)
{
    return fm_t3set(fm_r2x(d,m) + (fmc.X/2),
                    fm_r2y(d,m) + (fmc.Y/2),
                    z, val);
}

static INT_PTR fm_t4getr(INT_PTR d, INT_PTR m, INT_PTR z)
{
    return fm_t4get(fm_r2x(d,m) + (fmc.X/2),
                    fm_r2y(d,m) + (fmc.Y/2),
                    z);
}

static INT_PTR fm_t4setr(INT_PTR d, INT_PTR m, INT_PTR z, INT_PTR val)
{
    return fm_t4set(fm_r2x(d,m) + (fmc.X/2),
                    fm_r2y(d,m) + (fmc.Y/2),
                    z, val);
}


static INT_PTR fm_ctl(INT_PTR i)
{
   /* Value of control i */ 
   if (IS_OK_CTL_NR(i))
      return fmc.pre_ctl[i];
   return 0;
}

static INT_PTR fm_scl(INT_PTR a, INT_PTR il, INT_PTR ih, INT_PTR ol, INT_PTR oh)
{
    //Does this work correctly when il > ih ????
#if 0
    //should really be this...
    if (ih - il + 1 != 0)
        return (a - il)*(oh - ol + 1)/(ih - il + 1) + ol;
    else
        return 0;   //or ol???
#else
    //but Filter Factory implements it this way...
    if (ih != il)
        return (a - il)*(oh - ol)/(ih - il) + ol;
    else
        return 0;   //per Jens and Filter Factory
#endif
}

static INT_PTR fm_val(INT_PTR i, INT_PTR a, INT_PTR b)
{
    //return fm_scl(fm_ctl(i), 0, 255, a, b);
#if 0
    //should really be this...
    return fm_ctl(i)*(b - a + 1)/256 + a;
#elif 0
    //or, better yet, this...
    return (fmc.ctl[i].val - fmc.ctl[i].lo) * (b - a + 1)/
           (fmc.ctl[i].hi - fmc.ctl[i].lo + 1)
           + a;
#elif 0
    //but Filter Factory implements it this way...
    return fm_ctl(i)*(b - a)/255 + a;
#else
    //final compromise:
    assert(IS_POWER_OF_TWO(N_CTLS));
    i &= N_CTLS-1;  //safe index
    if (gParams->ctl[i].maxval == gParams->ctl[i].minval)
        return a;   //why not?
    return (gParams->ctl[i].val - gParams->ctl[i].minval) * (b - a)/
           (gParams->ctl[i].maxval - gParams->ctl[i].minval)
           + a;
#endif
}

static INT_PTR fm_map(INT_PTR i, INT_PTR n)
{   
#if 0
    // Following is per Jens, but not what Adobe doc says!!!
    INT_PTR x, y;
    x = fm_ctl(i*2);
    y = fm_ctl(i*2 + 1);
    return n*(y-x)/255 + x;
#else
    // Okay, here's the way it really works (and what Adobe
    // says it should be!)
    INT_PTR H, L;
    H = fm_ctl(i*2);
    L = fm_ctl(i*2+1);
    //return (n < L) ? 0 : (n > H) ? 255 : (H == L) ? 255 : (n-L)*255/(H-L);
    //not correct when H < L
    return (n < L) ? 0 : (n > H) ? 255 : (H == L) ? 255 : (n-L)*255/(H-L);
    //following is sometimes off by 1 when ctl(0) < ctl(1); ie. rounding towards 0 vs -inf.
    //max(0,min(255,(ctl(1)==ctl(0)?((x-ctl(1))>=0?255:0):(x-ctl(1))*255/(ctl(0)-ctl(1)))))/2+64
#endif
}


static INT_PTR fm_min(INT_PTR a, INT_PTR b)
{
    return (a <= b) ? a : b;
}

static INT_PTR fm_max(INT_PTR a, INT_PTR b)
{
    return (a >= b) ? a : b;
}

/***********************************************/
/****** FM-implemented f.p. math routines ******/
/***********************************************/

/////////////////////////////////////////////////////////////////////
//
// fmax(x1,x2,...) : return maximum of x1, x2, ...
//
/////////////////////////////////////////////////////////////////////
#if 0   // 2 operands only
static double fm_fmax(double a, double b)
{
    return (a >= b) ? a : b;
}//fm_fmax
#else   // variable number of operands (2 or more)
static double fm_fmax(double a, double b, ...)
{
    double fmax;
    INT_PTR nargs;
#ifdef _WIN64
	nargs = X64_get_nargs();
#else
    _asm mov nargs, ecx;			//Win32
#endif
    fmax = (a >= b) ? a : b;
    if (nargs > 2) {
        va_list ap;
        va_start(ap, b);
        while (--nargs > 1) {
            double z = va_arg(ap, double);
            if (z > fmax) {
                fmax = z;
            }
        }
        va_end(ap);
    }
    return fmax;
}//fm_max
#endif

/////////////////////////////////////////////////////////////////////
//
// fmin(x1,x2,...) : return minimum of x1, x2, ...
//
/////////////////////////////////////////////////////////////////////
#if 0   // 2 operands only
static double fm_fmin(double a, double b)
{
    return (a <= b) ? a : b;
}//fm_min
#else   // variable number of operands (2 or more)
static double fm_fmin(double a, double b, ...)
{
    double fmin;
    INT_PTR nargs;
#ifdef _WIN64
	nargs = X64_get_nargs();
#else
    _asm mov nargs, ecx;			//Win32
#endif
    fmin = (a <= b) ? a : b;
    if (nargs > 2) {
        va_list ap;
        va_start(ap, b);
        while (--nargs > 1) {
            double z = va_arg(ap, double);
            if (z < fmin) {
                fmin = z;
            }
        }
        va_end(ap);
    }
    return fmin;
}//fm_min
#endif

/////////////////////////////////////////////////////////////////////
//
// round(x) : round x to nearest or even
//
/////////////////////////////////////////////////////////////////////
//
// UGH!  See the following reference:
//	Fast Rounding of Floating Point Numbers in C/C++ on Wintel Platform
//	Laurent de Soras
//	Updated on 2008.03.08
//		web: http://ldesoras.free.fr
//
#ifndef _WIN64
/************** FAIRLY FAST LIBRARY IMPLEMENTATION *****************/
// Disable "no return value" warnings
#pragma warning(disable:4035)
static double fm_round(double x)
{
    //code based on FM's in-line implementation...
    __asm {
        fld x                       ; load x
#if 0	//assumes user never alters the cw
        fnstcw fmc.current_cw       ; save current cw (probably redundant!)
        fldcw fmc.CW_ROUND          ; new cw with RC=round (00b)
        frndint                     ; round to nearest or even
        fldcw fmc.current_cw        ; restore current cw
#elif 1	//assumes user may alter cw
        fnstcw fmc.current_cw       ; save current cw
		mov eax, fmc.current_cw		; load cw
		and ax, 0xF3FF				; set RC=00b
		mov DWORD PTR fmc.temp, eax	; store new cw in temp
        fldcw fmc.temp				; load new cw with RC=round
        frndint                     ; round to nearest or even
        fldcw fmc.current_cw        ; restore current cw
#else	//minimal case
        frndint                     ; All we need if ROUND is always the default!!!
#endif
    }
}//fm_round
// restore "no return value" warnings
#pragma warning (default: 4035)
#elif 1
/************** FIRST REFERENCE IMPLEMENTATION, VERY SLOW **********/
static double fm_round(double x)
{
    // Kludgy but correct reference implementation,
    // with an early-out speedup:
    double round_near_or_up = floor(x + 0.5L);
    double round_near_or_down = ceil(x - 0.5L);
#if 1 //early out for speed?
    if (round_near_or_up == round_near_or_down) {
        return round_near_or_up;
    }
#endif
    if ( floor(round_near_or_up*0.5L)*2.0 == round_near_or_up ) {
        return round_near_or_up;    // round_near_or_up is even
    }
    else {
        assert ( floor(round_near_or_down*0.5L)*2.0 == round_near_or_down );
        return round_near_or_down;  // round_near_or_down is even
    }
}//fm_round
#else
/************** 2nd REFERENCE IMPLEMENTATION, EVEN SLOWER **********/
static double fm_round(double x)
{
    // Kludgy but correct reference implementation:
    double round_near_or_up = floor(x + 0.5L);
    if ( floor(round_near_or_up*0.5L)*2.0 == round_near_or_up )
        return round_near_or_up;    // round_near_or_up is even
    else
        return ceil(x - 0.5L);  // round_near_or_down must be even
}//fm_round
#endif

/////////////////////////////////////////////////////////////////////
//
// chop(x) : truncate x toward 0.0
//
////////////////////////////////////////////////////////////////////
#ifndef _WIN64
/************** FAIRLY FAST LIBRARY IMPLEMENTATION *****************/
// Disable "no return value" warnings
#pragma warning(disable:4035)
static double fm_chop(double x)
{
    //code based on FM's in-line implementation...
    __asm {
		fld x                       ; load x
#if 0	//assumes user never alters cw
        fnstcw fmc.current_cw       ; save current cw (probably redundant!)
        fldcw fmc.CW_CHOP           ; new cw with RC=chop (11b)
        frndint                     ; chop toward 0.0
        fldcw fmc.current_cw        ; restore current cw
#elif 1	//assumes user may alter cw
        fnstcw fmc.current_cw       ; save current cw
		mov eax, fmc.current_cw		; load cw
		or ax, 0x0C00				; set RC=11b
		mov DWORD PTR fmc.temp, eax	; store new cw in temp
        fldcw fmc.temp				; load new cw with RC=chop
        frndint                     ; chop toward 0.0
        fldcw fmc.current_cw        ; restore current cw
#endif
    }
}//fm_chop
// restore "no return value" warnings
#pragma warning (default: 4035)
#elif 1
/************** REFERENCE IMPLEMENTATION, FAIRLY SLOW **********/
static double fm_chop(double x)
{
    // Kludgy but correct reference implementation:
    return (x >= 0.0) ? floor(x) : -floor(-x);
}//fm_chop
#endif

/////////////////////////////////////////////////////////////////////
//
// iround(x) : round double x to nearest or even, return as 32-bit (x86)
//			   or 64-bit (x64) integer
//
////////////////////////////////////////////////////////////////////
#ifndef _WIN64
/************** FAIRLY FAST LIBRARY IMPLEMENTATION *****************/
// Disable "no return value" warnings
#pragma warning(disable:4035)
static INT_PTR fm_iround(double x)
{
    //code based on FM's in-line implementation...
    INT_PTR iRes;
    __asm {
        fld x                       ; load x
#if 0	//asumes user never alters cw
        fnstcw fmc.current_cw       ; save current cw (probably redundant!)
        fldcw fmc.CW_ROUND          ; new cw with RC=round (00b)
        fistp iRes                  ; convert to integer, rounding to nearest or even
        fldcw fmc.current_cw        ; restore current cw
#elif 1	//assumes user may alter cw
        fnstcw fmc.current_cw       ; save current cw
		mov eax, fmc.current_cw		; load cw
		and ax, 0xF3FF				; set RC=00b
		mov DWORD PTR fmc.temp, eax	; store new cw in temp
        fldcw fmc.temp				; load new cw with RC=round
        fistp iRes                  ; convert to integer, rounding to nearest or even
        fldcw fmc.current_cw        ; restore current cw
#else	//minimal case
        fistp iRes                  ; All we need if ROUND is always the default!!!
#endif
        mov eax, iRes               ; return int result in eax
    }
}//fm_iround
// restore "no return value" warnings
#pragma warning (default: 4035)
#elif 1
/************** REFERENCE IMPLEMENTATION, FAIRLY SLOW **********/
static INT_PTR fm_iround(double x)
{
    // Kludgy but correct reference implementation:
    return (INT_PTR)fm_round(x);
}//fm_iround
#endif

/////////////////////////////////////////////////////////////////////
//
// ichop(x) : truncate double x toward 0, return as 32-bit (x86) or
//			  64-bit (x64) integer
//
////////////////////////////////////////////////////////////////////
#ifndef _WIN64
/************** FAIRLY FAST LIBRARY IMPLEMENTATION *****************/
// Disable "no return value" warnings
#pragma warning(disable:4035)
static INT_PTR fm_ichop(double x)
{
    //code based on FM's in-line implementation...
    INT_PTR iRes;
    __asm {
        fld x                       ; load x
#if 0	//assumes user never alters cw
        fnstcw fmc.current_cw       ; save current cw (probably redundant!)
        fldcw fmc.CW_CHOP           ; new cw with RC=chop (11b)
        fistp iRes                  ; convert to integer, chopping toward 0
        fldcw fmc.current_cw        ; restore current cw
#elif 1	//assumes user may alter cw
        fnstcw fmc.current_cw       ; save current cw
		mov eax, fmc.current_cw		; load cw
		or ax, 0x0C00				; set RC=11b
		mov DWORD PTR fmc.temp, eax	; store new cw in temp
        fldcw fmc.temp				; load new cw with RC=chop
        fistp iRes                  ; convert to integer, chopping toward 0
        fldcw fmc.current_cw        ; restore current cw
#elif defined(USE_SSE3)
		fisttp iRes					; use new SSE3 integer truncate op
#endif
        mov eax, iRes               ; return INT_PTR result in eax
    }
}//fm_ichop
// restore "no return value" warnings
#pragma warning (default: 4035)
#elif 0
/************** REFERENCE IMPLEMENTATION, FAIRLY SLOW **********/
static INT_PTR fm_ichop(double x)
{
    // Kludgy but correct reference implementation:
    return (INT_PTR)fm_chop(x);
}//fm_ichop
#elif 1
/************** REFERENCE IMPLEMENTATION, FAST **********/
// In this version, we use the fact that C always truncates
// when converting float to int, so we let C do the dirty
// work here!
static INT_PTR fm_ichop(double x)
{
    return (INT_PTR)x;	//C compiler chops toward 0.
}//fm_ichop
#elif 0
/************** REFERENCE IMPLEMENTATION, FAIRLY FAST **********/
static INT_PTR fm_ichop(double x)
{
    // Similar to MSVCRT implementation of _ftol2:

	// Make a trial conversion to integer, regardless of rounding mode...
	INT_PTR ires = (INT_PTR)x;	//begs the question if C truncates (Win32 calls _ftol2)!!!

	//The trial result is either correct, or off by +/- 1...

	//if ires is 0 or integer QNaN, return it unchanged
	//(note that integer QNaN is 0x80000000[00000000])
	//(but what if ires overflows and ires-1 doesn't, or vice versa?)
	if ((ires << 1) != 0) {
		if (ires > 0) {
			//if result is strictly positive, trial result is either correct or off by +1
			if ((double)ires > x) return ires - 1;	//ires was too high
		}
		else {
			//if result is negative, trial result is either correct or off by -1
			if ((double)ires < x) return ires + 1;	//ires was too low
		}
	}
	return ires;	//ires is correct (or QNaN indicating overflow)
}//fm_ichop
#elif 1
/************** REFERENCE IMPLEMENTATION, EVEN SLIGHTLY FASTER **********/
static INT_PTR fm_ichop(double x)
{
    // Similar to MSVCRT implementation of _ftol2:

	// Make a trial conversion to integer, regardless of rounding mode...
	INT_PTR ires = (INT_PTR)x;	//begs the question if C truncates (Win32 calls _ftol2)!!!

	//The trial result is either correct, or off by +/- 1...

	//if ires is 0 or integer QNaN, return it unchanged
	//(note that integer QNaN is 0x80000000[00000000])
	//(but what if ires overflows and ires-1 doesn't, or vice versa?)

	if ((ires << 1) != 0 && (double)ires != x) {
		//ires is off by one
		//if result is negative, trial result is off by -1,
		//else trial result is off by +1
		return (ires < 0) ? (ires + 1) : (ires - 1);
	}
	return ires;	//ires is correct (or a QNaN indicating overflow)
}//fm_ichop
#endif

/////////////////////////////////////////////////////////////////////
//
// ifloor(x) : round double x down toward -Infinity,
//             return as a 32-bit (x86) or 64-bit (x64) integer.
//
////////////////////////////////////////////////////////////////////
#ifndef _WIN64
/************** FAIRLY FAST LIBRARY IMPLEMENTATION *****************/
// Disable "no return value" warnings
#pragma warning(disable:4035)
static INT_PTR fm_ifloor(double x)
{
    //code based on FM's in-line implementation...
    INT_PTR iRes;
    __asm {
        fld x                       ; load x
#if 0	///assumes user never alters cw
        fnstcw fmc.current_cw       ; save current cw (probably redundant!)
        fldcw fmc.CW_FLOOR          ; new cw with RC=floor (01b)
        fistp iRes                  ; convert to integer, rounding down toward -Infinity
        fldcw fmc.current_cw        ; restore current cw
#elif 1	//assumes user may alter cw
        fnstcw fmc.current_cw       ; save current cw
		mov eax, fmc.current_cw		; load cw
		and ax, 0xF3FF				; clear RC field
		or  ax, 0x0400				; set RC=01b
		mov DWORD PTR fmc.temp, eax	; store new cw in temp
        fldcw fmc.temp				; load new cw with RC=floor
        fistp iRes                  ; convert to integer, rounding down toward -Infinity
        fldcw fmc.current_cw        ; restore current cw
#endif
        mov eax, iRes               ; return INT_PTR result in eax
    }
}//fm_ifloor
// restore "no return value" warnings
#pragma warning (default: 4035)
#elif 1
/************** REFERENCE IMPLEMENTATION, FAIRLY SLOW **********/
static INT_PTR fm_ifloor(double x)
{
    // Kludgy but correct reference implementation:
    return (INT_PTR)floor(x);
}//fm_ifloor
#endif

/////////////////////////////////////////////////////////////////////
//
// iceil(x) : round double x up toward +Infinity, 
//            return as a 32-bit (x86) or 64-bit (x64) integer.
//
////////////////////////////////////////////////////////////////////
#ifndef _WIN64
/************** FAIRLY FAST LIBRARY IMPLEMENTATION *****************/
// Disable "no return value" warnings
#pragma warning(disable:4035)
static INT_PTR fm_iceil(double x)
{
    //code based on FM's in-line implementation...
    INT_PTR iRes;
    __asm {
        fld x                       ; load x
#if 0	//assumes user never alters cw
        fnstcw fmc.current_cw       ; save current cw (probably redundant!)
        fldcw fmc.CW_CEIL           ; new cw with RC=ceil (10b)
        fistp iRes                  ; convert to integer, rounding up toward +Infinity
        fldcw fmc.current_cw        ; restore current cw
#elif 1	//assumes user may alter cw
        fnstcw fmc.current_cw       ; save current cw
		mov eax, fmc.current_cw		; load cw
		and ax, 0xF3FF				; clear RC field
		or  ax, 0x0800				; set RC=10b
		mov DWORD PTR fmc.temp, eax	; store new cw in temp
        fldcw fmc.temp				; load new cw with RC=ceil
        fistp iRes                  ; convert to integer, rounding up toward +Infinity
        fldcw fmc.current_cw        ; restore current cw
#endif
        mov eax, iRes               ; return INT_PTR result in eax
    }
}//fm_iceil
// restore "no return value" warnings
#pragma warning (default: 4035)
#elif 1
/************** REFERENCE IMPLEMENTATION, FAIRLY SLOW **********/
static INT_PTR fm_iceil(double x)
{
    // Kludgy but correct reference implementation:
    return (INT_PTR)ceil(x);
}//fm_iceil
#endif

//Should put the following *pow* routines into the same module as OP_POW??!!!

//////////////////////////// ipow(i, n) ///////////////////////////////
//
// ipow(i, n) : calculate integer i to the power integer n, 
//              returns a 32-bit or 64-bit integer.
// we assume: ipow(i,0) = 1, even when i==0.
//
////////////////////////////////////////////////////////////////////
#if 0
/************** NAIVE INTEGER IMPLEMENTATION **********************/
static INT_PTR fm_ipow(INT_PTR i, INT_PTR n)
{
    if (n > 0) {
#if 0
        INT_PTR iRes = i;
        while (--n) 
        {// no overflow check
            iRes *= i;
        }
#elif 0
        INT_PTR iRes = i;
        while (--n) 
        {// overflow check
            iRes *= i;
            __asm jo oflo1;
        }
#elif 0
        INT_PTR iRes = i;
        char register oflo = 0;
        while (--n && !oflo) 
        {// overflow check
            iRes *= i;
            __asm  seto oflo;
        }
        if (oflo) goto oflo1;
#else
        INT_PTR iRes;
        __asm { //overflow check in asm
			 // register mapping:
			 // ecx = n, edx = i, eax = iRes
             mov     ecx, n		//ecx = n
	         mov	 edx, i		//edx = i
  		     dec	 ecx		//--n
  		     mov	 eax, edx	//eax = iRes = i
  		     je	     DONE		//if (n==0) goto DONE
LP01:							// do {
  	         imul	 eax, edx	//iRes *= i
             jo      oflo1		//if (overflow) goto oflo1
  		     dec	 ecx		//--n
  		     jne	 LP01		//} while(n!=0)
DONE:
             mov     iRes, eax	//iRes = eax
        }//asm
#endif
        return iRes;
    }
    else if (n == 0) {
        /* Assume ipow(0,0) = 1, not indefinite */
        if (i == 0) errno = EDOM; // but flag an EDOM error for ipow(0,0)
        return 1;
    }
    /* n < 0 */
    else if (i == 0) {
        /* Domain error: return approximation to +infinity */
        errno = EDOM;
        return INTV_MAX; //anything better??	//X64??
    }
    else if (i == 1) {
        return 1;
    }
    else if (i == -1) {
        /* return +1 if n is even, -1 if n is odd */
#if 0
        return (n & 1) ? -1 : 1;
#elif 1
        return 1 - ((n & 1) << 1);  // best code under Release
#elif 1
        return (-(n & 1) & -2) + 1;
#endif
    }
    else {
        return 0;
    }

oflo1:  /* Overflow => Range error: return approximation to +infinity */
        errno = ERANGE;
        return INTV_MAX; //anything better??	//X64??

}//fm_ipow
#elif 1
/************** LESS NAIVE INTEGER IMPLEMENTATION **********************/
// This one uses the binary exponentiation algorithm.	//COULD USE FASTCALL for x86??
static INT_PTR fm_ipow(INT_PTR i, INT_PTR n)
{
    INT_PTR iRes = 1;
	int isNeg = (i < 0) && (n & 1);	//sign of result

    /// But first, unroll the loop for the first 18 or so cases...

    if (n > 0) {  //how about --n, start with iRes = i (elim 1 mult, and also eliminate final i*=i if n>>1 is 0???
////#if 1 //for testing under _WIN32
#ifdef _WIN64	//for now, since no 64-bit inline asm... (could also use for 32-bit?)
#if 0	//do CERT-32 overflow checking
        do { // overflow checking -- check for overflow the hard way, for X64??
			// or just write the X64 version in ML64??
            if (n & 1) {
				if (iRes > INTV_MAX/i) goto oflo1;	//CERT-32 overflow test	//GP if i==0!!
                iRes *= i;  /* If n is odd, multiply in an i. */
            } 
			if (i > INTV_MAX/i) goto oflo2;			//CERT-32 overflow test
            i *= i;         /* Repeatedly square on each bit. */
            n >>= 1;        /* Iterate through all the bits. */ 
        } while (n);        /* If n == 0 then we are done. */ 
#elif 1	//Call an external X64 masm routine to do the hard lifting...
		extern INT64 X64_ipow_core(INT64 i, INT64 n);	//x64 asm helper code
		iRes = X64_ipow_core(i, n);
#endif //_Win64 cases
#elif 0
        do { // overflow check in asm
            if (n & 1) {
                iRes *= i;  /* If n is odd, multiply in an i. */
                __asm jo oflo1;
            } 
            i *= i;         /* Repeatedly square on each bit. */ 
            __asm jo oflo2; ///// oflo2: oflow iff n >> 1 != 0...
            n >>= 1;        /* Iterate through all the bits. */ 
        } while (n);        /* If n == 0 then we are done. */ 
#elif 1
        __asm { //overflow check in asm
			 // register mapping:
			 // ecx = n, edx = i, eax = iRes
             mov     ecx, n		//ecx = n
	         mov	 edx, i		//edx = i
			 mov	 eax, 1		//eax = iRes = 1
LP01:							//do {
			 test ecx, 1		//n & 1
			 jz		 LP02		//if (n & 1) {
			 imul	 eax, edx	//iRes *= i
			 jo		 oflo1		//if (overflow) goto oflo1
LP02:							//}
			 imul	 edx, edx	//i *= i
			 jo		 oflo3		//if (overflow) goto oflo3
			 shr	 ecx, 1		//n >>= 1
			 jnz	 LP01		//} while (n)
             mov     iRes, eax	//iRes = eax
        }//asm
#elif 1
        char register oflo = 0;
        char register oflo2 = 0;
        do { // overflow check
            if (n & 1) {
                iRes *= i;  /* If n is odd, multiply in an i. */
                __asm  seto oflo;
            } 
            i *= i;         /* Repeatedly square on each bit. */ 
            __asm  seto oflo2;
            n >>= 1;        /* Iterate through all the bits. */ 
        } while (n && !oflo && !oflo2);  /* If n == 0 then we are done. */ 
        if (oflo || n) goto oflo1;
#endif
        return iRes;

// Handle (possible) overflow here...
#ifndef _WIN64
oflo3:	/* store ecx back into n */
		_asm {mov n, ecx }
		/* now fall into oflo2 */
#endif
oflo2:  /* Overflow here only if n >> 1 != 0 */
        if ((n >> 1) == 0) return iRes;
        // fall through to overflow case...
oflo1:  /* Overflow => Range error: return approximation to +infinity */
        errno = ERANGE;
		//Still wrong: need to test *original* values of i and n!!
		if (isNeg) {
			/* i is negative and n is odd */
			return INTV_MIN;	//approximate -infinity
		}
		else {
	        return INTV_MAX;	//approximate +infinity
		}

    }//n > 0
    else if (n == 0) {
        /* Assume ipow(0,0) = 1, not indefinite */
        if (i == 0) errno = EDOM; // but flag an EDOM error for ipow(0,0)
        return 1;
    }
    /* n < 0 */
    else if (i == 0) {
        /* Domain error: return approximation to +infinity */
        errno = EDOM;
        return INTV_MAX; //anything better??
    }
    else if (i == 1) {
        return 1;
    }
    else if (i == -1) {
        /* return +1 if n is even, -1 if n is odd */
#if 1
		return isNeg ? -1 : 1;
#elif 1
        return (n & 1) ? -1 : 1;
#elif 1
        return 1 - ((n & 1) << 1);  // best code under Win32 Release
#elif 1
        return (-(n & 1) & -2) + 1;
#endif
    }
    else {
		/* i < -1 || i > 1 */
        return 0;
    }

}//fm_ipow
#elif 1
/************** FASTER INTEGER IMPLEMENTATION **********************/
#define T INT_PTR /* template type */
static T fm_ipow(T value, INT_PTR n)
{
   /* Explicitly compute the first few powers, this is never going to be as 
    * efficient as simply putting x*x (for example) in the code, but it is 
    * better than needing to go through initialization and loop constructs. 
    */ 
   T z = 1; 
   T y; 
    

   // In general, the binary-exponentiation algorithm for pow(x, n)
   // requires lg2(n) + onebits(n) - 1 multiply ops to compute, if n > 0,
   // where: lg2(n) is the largest integer i such that 2**i <= n, and
   //        onebits(n) is the number of 1-bits in the binary
   //        representation of n
   // e.g.:
   //                                         binary    add-chain   linear
   //           n  lg(2) + onebits(n) - 1  =  mulops      mulops    mulops
   //        ----- -----   ---------- ---     ------    ---------   ------
   //           1     0  +      1     - 1         0         0           0
   //           2     1  +      1     - 1         1         1           1
   //           3     1  +      2     - 1         2         2           2
   //           4     2  +      1     - 1         2         2           3
   //           5     2  +      2     - 1         3         3           4
   //           6     2  +      2     - 1         3         3           5
   //           7     2  +      3     - 1         4         4           6
   //           8     3  +      1     - 1         3         3           7
   //           9     3  +      2     - 1         4         4           8
   //          10     3  +      2     - 1         4         4           9
   //          11     3  +      3     - 1         5         5          10
   //          12     3  +      2     - 1         4         4          11
   //          13     3  +      3     - 1         5         5          12 
   //          14     3  +      3     - 1         5         5          13
   //          15     3  +      4     - 1         6         5          14
   //          16     4  +      1     - 1         4         4          15
   //          17     4  +      2     - 1         5         5          16
   //          18     4  +      2     - 1         5         5          17
   //
   // For comparison, we show the number of multiply ops required by an optimal
   // addition/multiplication chain and for linear calculation of the product
   // (mulops = n - 1) in the last two columns.  We see that the lowest n for
   // which the add-chain algorithm takes fewer mulops than binary exponentiation
   // is n = 15.
   // For n <= 18, we can always compute pow(x, n) with no more than 5 multiplies.
   //
   // Except for a few other special cases (such as n = power of 2), it's probably
   // not worth unrolling the loop for n > 18 in the long run...
   //
   switch( n ) 
   { 
     T register temp;
     case 0: 
       return z; 
     case 1:     // 0 mulops
       return value; 
     case 2:     // 1 mulop
       return value * value; 
     case 3:     // 2 mulops
       return value * value * value; 
     case 4:     // 2 mulops
       return temp = value * value, temp * temp;
     case 5:     // 3 mulops
       return temp = value * value, temp *= temp, temp * value;
     case 6:     // 3 mulops
       return temp = value * value * value, temp * temp;
     case 7:     // 4 mulops
       return temp = value * value * value, temp *= temp, temp * value;
     case 8:     // 3 mulops
       return temp = value * value, temp *= temp, temp * temp;
     case 9:     // 4 mulops
       return temp = value * value, temp *= temp, temp *= temp, temp * value;
     case 10:    // 4 mulops
       return temp = value * value, temp *= temp, temp *= value, temp * temp;
     case 11:    // 5 mulops
       return temp = value * value, temp *= temp, temp *= value, temp *= temp, temp * value;
     case 12:    // 4 mulops
       return temp = value * value * value, temp *= temp, temp * temp;
     case 13:    // 5 mulops
       return temp = value * value * value, temp *= temp, temp *= temp, temp * value;
     case 14:    // 5 mulops
       return temp = value * value * value, temp *= temp, temp *= value, temp * temp;
     case 15:    // 5 mulops (1st time add-chain beats binary exponentiation (6 mulops) per [Knuth]!!)
       return temp = value * value, temp *= temp, temp *= value, temp * temp * temp;
     case 16:    // 4 mulops
       return temp = value * value, temp *= temp, temp *= temp, temp * temp;
     case 17:    // 5 mulops
       return temp = value * value, temp *= temp, temp *= temp, temp *= temp, temp * value;
     case 18:    // 5 mulops
       return temp = value * value, temp *= temp, temp *= temp, temp *= value, temp * temp;
   } 
    
   if( n < 0 ) /* If n is negative then we need to compute the inverse */ 
   { 
     n = -n; 
     y = z / value; 
   } 
   else 
   { 
     y = value; 
   } 

   do 
   { 
     if( n & 0x1u ) /* 0x1u is the hexadecimal literal constant representing */ 
     {              /* an unsigned integer == 1. n & 0x1u means that the */ 
       z *= y;      /* smallest bit of n is zero, in this case the current */ 
     }              /* repeated square power of value (== y) is a component */ 
                    /* of the result. */ 
      
     y *= y;    /* Repeatedly square on each bit. */ 
     n >>= 1;   /* Iterate through all the bits. */ 
      
   }while( n ); /* If n == 0 then we are done. */ 
    
   return z; 
}//fm_ipow
#undef T
#elif 0
/************** FAIRLY FAST LIBRARY IMPLEMENTATION *****************/
// Disable "no return value" warnings
#pragma warning(disable:4035)
static INT_PTR fm_ipow(INT_PTR i, INT_PTR n)
{
#error NYI!!
}//fm_ipow
// restore "no return value" warnings
#pragma warning (default: 4035)
#elif 1
/************** REFERENCE IMPLEMENTATION, FAIRLY SLOW **********/
static INT_PTR fm_ipow(INT_PTR i, INT_PTR n)
{
	// Kludgy but (reasonably) correct reference implementation:
    return (INT_PTR)pow((double)i, (double)n);
}//fm_ipow
#endif


//////////////////////////// powi(x, n) //////////////////////////////////
//
// powi(x, n) : calculate double x to the power integer n, 
//              returns a double.
// we assume: powi(x,0) = 1.0, even when x==0.0.
//
////////////////////////////////////////////////////////////////////
#if 0	//why did I select this???
/************** NAIVE INTEGER IMPLEMENTATION **********************/
static double fm_powi(double x, INT_PTR n)
{
    if (n < 0) {
        if (x == 0.0) {
            /* Domain error: return +infinity */
            errno = EDOM;
            return HUGE_VAL;  ///////////// NO +INFINITY !!!!
        }
        else if (x == 1.0) { //speedy out?
            return 1.0;
        }
        else if (x == -1.0) { //speedy out?
            /* return +1 if n is even, -1 if n is odd */
            return (n & 1) ? -1.0 : 1.0;
        }
        /* compute powi(1/x, -n) */
        x = 1.0/x;
        n = -n;
    }
    assert(n >= 0);
    if (n > 0) {
        double res = x;
        while (--n > 0) res *= x;
        return res;
    }
    assert(n == 0);
    /* Assume powi(0.0,0) = 1.0, not indefinite */
    if (x == 0.0) errno = EDOM; // but flag an EDOM error for powi(0.0,0)
    return 1.0;

}//fm_powi
#elif 1	//should probably test and time this implementation, since it doesn't seem to have been in use!!
/************** FASTER INTEGER IMPLEMENTATION **********************/
#define T double /* template type */
static T fm_powi(T value, INT_PTR n)
{
   /* Explicitly compute the first few powers, this is never going to be as 
    * efficient as simply putting x*x (for example) in the code, but it is 
    * better than needing to go through initialization and loop constructs. 
    */ 
   T z = 1; 
   T y; 
    

   // In general, the binary-exponentiation algorithm for pow(x, n)
   // requires lg2(n) + onebits(n) - 1 multiply ops to compute, if n > 0,
   // where: lg2(n) is the largest integer i such that 2**i <= n, and
   //        onebits(n) is the number of 1-bits in the binary
   //        representation of n
   // e.g.:
   //                                         binary    add-chain   linear
   //           n  lg(2) + onebits(n) - 1  =  mulops      mulops    mulops
   //        ----- -----   ---------- ---     ------    ---------   ------
   //           1     0  +      1     - 1         0         0           0
   //           2     1  +      1     - 1         1         1           1
   //           3     1  +      2     - 1         2         2           2
   //           4     2  +      1     - 1         2         2           3
   //           5     2  +      2     - 1         3         3           4
   //           6     2  +      2     - 1         3         3           5
   //           7     2  +      3     - 1         4         4           6
   //           8     3  +      1     - 1         3         3           7
   //           9     3  +      2     - 1         4         4           8
   //          10     3  +      2     - 1         4         4           9
   //          11     3  +      3     - 1         5         5          10
   //          12     3  +      2     - 1         4         4          11
   //          13     3  +      3     - 1         5         5          12 
   //          14     3  +      3     - 1         5         5          13
   //          15     3  +      4     - 1         6         5          14
   //          16     4  +      1     - 1         4         6          15
   //          17     4  +      2     - 1         5         5          16
   //          18     4  +      2     - 1         5         5          17
   //
   // For comparison, we show the number of multiply ops required by an optimal
   // addition/multiplication chain and for linear calculation of the product
   // (mulops = n - 1) in the last two columns.  We see that the lowest n for
   // which the add-chain algorithm takes fewer mulops than binary exponentiation
   // is n = 15.
   // For n <= 18, we can always compute pow(x, n) with no more than 5 multiplies.
   //
   // Except for a few other special cases (such as n = power of 2), it's probably
   // not worth unrolling the loop for n > 18 in the long run...
   //
   switch( n ) 
   { 
     T register temp;
     case 0: 
       return z; 
     case 1:     // 0 mulops
       return value; 
     case 2:     // 1 mulop
       return value * value; 
     case 3:     // 2 mulops
       return value * value * value; 
     case 4:     // 2 mulops
       return temp = value * value, temp * temp;
     case 5:     // 3 mulops
       return temp = value * value, temp *= temp, temp * value;
     case 6:     // 3 mulops
       return temp = value * value * value, temp * temp;
     case 7:     // 4 mulops
       return temp = value * value * value, temp *= temp, temp * value;
     case 8:     // 3 mulops
       return temp = value * value, temp *= temp, temp * temp;
     case 9:     // 4 mulops
       return temp = value * value, temp *= temp, temp *= temp, temp * value;
     case 10:    // 4 mulops
       return temp = value * value, temp *= temp, temp *= value, temp * temp;
     case 11:    // 5 mulops
       return temp = value * value, temp *= temp, temp *= value, temp *= temp, temp * value;
     case 12:    // 4 mulops
       return temp = value * value * value, temp *= temp, temp * temp;
     case 13:    // 5 mulops
       return temp = value * value * value, temp *= temp, temp *= temp, temp * value;
     case 14:    // 5 mulops
       return temp = value * value * value, temp *= temp, temp *= value, temp * temp;
     case 15:    // 5 mulops (1st time add-chain beats binary exponentiation (6 mulops) per [Knuth]!!)
       return temp = value * value, temp *= temp, temp *= value, temp * temp * temp;
     case 16:    // 4 mulops
       return temp = value * value, temp *= temp, temp *= temp, temp * temp;
     case 17:    // 5 mulops
       return temp = value * value, temp *= temp, temp *= temp, temp *= temp, temp * value;
     case 18:    // 5 mulops
       return temp = value * value, temp *= temp, temp *= temp, temp *= value, temp * temp;
   } 
    
   if( n < 0 ) /* If n is negative then we need to compute the inverse */ 
   { 
     n = -n; 
     y = z / value; 
   } 
   else 
   { 
     y = value; 
   } 

   do 
   { 
     if( n & 0x1u ) /* 0x1u is the hexadecimal literal constant representing */ 
     {              /* an unsigned integer == 1. n & 0x1u means that the */ 
       z *= y;      /* smallest bit of n is one, in this case the current */ 
     }              /* repeated square power of value (== y) is a component */ 
                    /* of the result. */ 
      
     y *= y;    /* Repeatedly square on each bit. */ 
     n >>= 1;   /* Iterate through all the bits. */ 
      
   }while( n ); /* If n == 0 then we are done. */ 
    
   return z; 
}//fm_powi
#undef T
#elif 0
/************** FAIRLY FAST LIBRARY IMPLEMENTATION *****************/
// Disable "no return value" warnings
#pragma warning(disable:4035)
static double fm_powi(double x, INT_PTR n)
{
#error Not Yet Implemented
}//fm_powi
// restore "no return value" warnings
#pragma warning (default: 4035)
#elif 1
/************** REFERENCE IMPLEMENTATION, FAIRLY SLOW **********/
static double fm_powi(double x, INT_PTR n)
{
	// Kludgy but (reasonaby) correct reference implementation:
    return pow(x, (double)n);
}//fm_powi
#endif

/******************************************************/
/****** end of FM-implemented f.p. math routines ******/
/******************************************************/


static INT_PTR fm_abs(INT_PTR a)
{
    return (a >= 0) ? a : -a;
}

static INT_PTR fm_add(INT_PTR a, INT_PTR b, INT_PTR c)
{
    return (a+b < c) ? a+b : c;
}

static INT_PTR fm_sub(INT_PTR a, INT_PTR b, INT_PTR c)
{
    return (a-b > c) ? a-b : c; // or IABS(a-b) ???
}

static INT_PTR fm_dif(INT_PTR a, INT_PTR b)
{
    return (a-b >= 0) ? a-b : b-a;
}

static INT_PTR fm_rnd(INT_PTR a, INT_PTR b)
{
    // which is more uniformly distributed,
    // the high bits or the low bits of rand()?
#if 0
    return rand()*(b-a+1)/RAND_MAX + a;
#elif 1
    return a <= b ? rand() % (b-a+1) + a : rand() % (a-b+1) + b;
#else
    // Per Jens:  Wrong, returns [a,b) instead of [a,b].
    return a == b ? a : rand() % (b-a) + a;
#endif
}

static INT_PTR fm_rst(INT_PTR seed)	//X64??
{
    // (Re-)seed the pseudorandom number generator.
#if 1
    //for more FF-like behavior...
    seed += rand();
#endif
    srand((UINT32)seed);
    return seed;
} //fm_rnd

static INT_PTR fm_mix(INT_PTR a, INT_PTR b, INT_PTR n, INT_PTR d)
{
    // FF compatible version...
    return (d != 0) ? a*n/d + b*(d-n)/d : 0;
}

static INT_PTR fm_mix1(INT_PTR a, INT_PTR b, INT_PTR n, INT_PTR d)
{
#if 0
    return (d != 0) ? (a*n + b*(d-n))/d : 0;
#else
    return (d != 0) ? b - (b - a)*n/d : 0;
#endif
}

static INT_PTR fm_mix2(INT_PTR a, INT_PTR b, INT_PTR n, INT_PTR d)
{
    //return (d != 0) ? b - ((b - a)*n*2 + d)/(2*d) : 0;
    return (d == 0) ? 0 :
    (b - a)*n >= 0 ? b - ((b - a)*n*2 + d)/(2*d) :
                     b + ((a - b)*n*2 + d)/(2*d) ;
}

//FF benchmark for sqr = 8.0

static INT_PTR fm_sqr(INT_PTR x)
{
    //benchmark = 5.2
    //in-line fsqrt (OP_SQR) = 4.3
    return (x > 0) ? (INT_PTR)(sqrt((double)x) + 0.0) : 0;  // trunc or round???
}

#if 0
//from Michael:
INT_PTR fm_isqrt(unsigned INT_PTR x)
{
    //benchmark = 5.6
    register unsigned INT_PTR r, nr, m;

    r = 0;
#if 1
    m = 0x40000000;
#else
    m = 0x4000;
#endif
    do 
    {
        nr = r + m;
        if (nr <= x) 
        {
            x -= nr;
            r = nr + m;
        }
        r >>= 1;
        m >>= 2;
        } 
    while (m != 0);

    return (INT_PTR)r;
}

#else
/***********************************************************************/
/* from: http://www.research.microsoft.com/~hollasch/cgindex/math/introot.html
/* Integer Square Root
/* 
/* Ben Discoe (rodent@netcom.COM), comp.graphics, 6 Feb 92 
/*
/***********************************************************************/
UINT_PTR fm_isqrt(UINT_PTR v)	//X64 needs more work???
{
    //benchmark = 3.7 (3.4 when called from OP_SQR)
    register UINT_PTR t = 1L<<30, r = 0, s;

#   define STEP(k) s = t + r; r>>= 1; if (s <= v) { v -= s; r |= t;}

    STEP(15);  t >>= 2;
    STEP(14);  t >>= 2;
    STEP(13);  t >>= 2;
    STEP(12);  t >>= 2;
    STEP(11);  t >>= 2;
    STEP(10);  t >>= 2;
    STEP(9);   t >>= 2;
    STEP(8);   t >>= 2;
    STEP(7);   t >>= 2;
    STEP(6);   t >>= 2;
    STEP(5);   t >>= 2;
    STEP(4);   t >>= 2;
    STEP(3);   t >>= 2;
    STEP(2);   t >>= 2;
    STEP(1);   t >>= 2;
    STEP(0);

#   undef STEP

    return r;
} /*fm_isqrt*/
#endif

/******************** isqrt_fast: FASTCALL version *********************/
/* from: http://www.research.microsoft.com/~hollasch/cgindex/math/introot.html
/* Integer Square Root
/* 
/* Ben Discoe (rodent@netcom.COM), comp.graphics, 6 Feb 92 
/*
/***********************************************************************/
UINT_PTR __fastcall isqrt_fast(UINT_PTR v)  //fastcall passes 'v' in rCX
{											//X64 needs more work??
    //benchmark = 3.7 (3.4 when called from OP_SQR)
    register UINT_PTR t = 1L<<30, r = 0, s;
  
#   define STEP(k) s = t + r; r>>= 1; if (s <= v) { v -= s; r |= t;}

    STEP(15);  t >>= 2;
    STEP(14);  t >>= 2;
    STEP(13);  t >>= 2;
    STEP(12);  t >>= 2;
    STEP(11);  t >>= 2;
    STEP(10);  t >>= 2;
    STEP(9);   t >>= 2;
    STEP(8);   t >>= 2;
    STEP(7);   t >>= 2;
    STEP(6);   t >>= 2;
    STEP(5);   t >>= 2;
    STEP(4);   t >>= 2;
    STEP(3);   t >>= 2;
    STEP(2);   t >>= 2;
    STEP(1);   t >>= 2;
    STEP(0);

#   undef STEP

    return r;
} /*isqrt_fast*/


static INT_PTR fm_sin(INT_PTR x)
{
#ifdef TRIG_ON_FLY
    //compute sine on-the-fly...
    return (INT_PTR)(sin( (twopi/1024.0)*(double)x ) * 512.0 + 0.0);     // round?? can be neg!!
#else
    //use precomputed lookup table...
    return fmc.pre_sin[x & 1023];
#endif
}

static INT_PTR fm_cos(INT_PTR x)
{
#ifdef TRIG_ON_FLY
    //compute cosine on-the-fly...
    return (INT_PTR)(cos( (twopi/1024.0)*(double)x ) * 512.0 + 0.0);     // round?? can be neg!!
#else
    //use precomputed lookup table...
    return fmc.pre_sin[(x+256) & 1023];
#endif
}

static INT_PTR fm_tan(INT_PTR x)
{
#ifdef TRIG_ON_FLY
    //compute tangent on-the-fly...
    if ( (x & 511) == 256)
        return (INT_PTR)(tan( (twopi/1024.0)*(-255.5) ) * 1024.0 + 0.0);    // per Jens???
    else
        return (INT_PTR)(tan( (twopi/1024.0)*(double)x ) * 1024.0 + 0.0);   // round?? can be neg!!
#else
    //use precomputed lookup table...
    return fmc.pre_tan[x & 511];
#endif
}



static INT_PTR fm_c2d(INT_PTR x, INT_PTR y)
{
    //FF benchmark = 5.8
#if 0
    //benchmark = 7.1
    // per Jens (Gimp)...
    INT_PTR d;

    if (y != 0)
        d = (INT_PTR)(256.0 * twobypi * atan( (double)x/(double)IABS(y) ) + 0.0) - 256;   // round????
    else if (x < 0)
        d = 512;
    else
        d = 0;
    return (y <= 0) ? d : -d;   // Jens had y < 0
#else
    //following is more straightforward!!!
#if 0
	// VC doesn't fold this!!!	(try opt to fold float constant expressions???)
    return (INT_PTR)( (256.0 * twobypi) * atan2((double)y, (double)x) );
#else
    //benchmark = 6.1
    return (INT_PTR)( (256.0 * 0.63661977236758134308) * atan2((double)y, (double)x) );
#endif
#endif
}

static INT_PTR fm_c2m(INT_PTR x, INT_PTR y)
{
    //FF benchmark = 11.50 / 6.25 -- uses precomputed table for x/y!!!
#if 0
    //benchmark = 15.10
    //advantage: doesn't easily overflow -- IF convert x, y to double first!!!
    return (INT_PTR)( sqrt((double)(x*x + y*y)) );
#elif 1
    //benchmark = 12.62 / 8.47
    return fm_isqrt(x*x + y*y);
#elif 0
    //benchmark = 12.34 / 8.25
    register unsigned INT_PTR v = x*x + y*y;
    register unsigned INT_PTR t = 1L<<30, r = 0, s;

#   define STEP(k) s = t + r; r>>= 1; if (s <= v) { v -= s; r |= t;}

    STEP(15);  t >>= 2;
    STEP(14);  t >>= 2;
    STEP(13);  t >>= 2;
    STEP(12);  t >>= 2;
    STEP(11);  t >>= 2;
    STEP(10);  t >>= 2;
    STEP(9);   t >>= 2;
    STEP(8);   t >>= 2;
    STEP(7);   t >>= 2;
    STEP(6);   t >>= 2;
    STEP(5);   t >>= 2;
    STEP(4);   t >>= 2;
    STEP(3);   t >>= 2;
    STEP(2);   t >>= 2;
    STEP(1);   t >>= 2;
    STEP(0);

#   undef STEP

    return r;
#else
    //benchmark = 15.81 for 2 N-R's, 18.71 for 3 N-R's.
    INT_PTR d = IABS(x) + IABS(y);
    INT_PTR h;
    if (d == 0) return 0;
    h = x*x + y*y;      //easily overflows!
    d = (h/d + d) >> 1; //1st N-R iteration
    d = (h/d + d) >> 1; //2nd N-R iteration
    //d = (h/d + d) >> 1; //3rd N-R iteration
    return d;
#endif
}

static INT_PTR fm_get(INT_PTR i)
{
    return fmc.cell[i & (N_CELLS-1)];
}

static INT_PTR fm_put(INT_PTR v, INT_PTR i)
{
    return fmc.cell[i & (N_CELLS-1)] = (int)v;
}

static INT_PTR fm_cnv(INT_PTR m11, INT_PTR m12, INT_PTR m13,	//X64??
					  INT_PTR m21, INT_PTR m22, INT_PTR m23,
					  INT_PTR m31, INT_PTR m32, INT_PTR m33,
					  INT_PTR d)
{
    INT_PTR x = fmc.x;
    INT_PTR y = fmc.y;
    INT_PTR z = fmc.z;

    if (d == 0)
    {   // default divisor to sum of weights...
        d = m11+m12+m13+m21+m22+m23+m31+m32+m33;    // ???
        if (d == 0)
            return 255; // give up if divisor still 0
    }
    return (
        //could check for safe coordinates in toto and
        //then use unsafe src() to speed up...
        //or set up pointer to src(x-1,y-1,z), then
        //bump it through all 9 points without
        //recomputing index expression each time...
            m11*fmc.src(x-1, y-1, z) +
            m12*fmc.src(x  , y-1, z) +
            m13*fmc.src(x+1, y-1, z) +
            m21*fmc.src(x-1, y  , z) +
            m22*fmc.src(x  , y  , z) +
            m23*fmc.src(x+1, y  , z) +
            m31*fmc.src(x-1, y+1, z) +
            m32*fmc.src(x  , y+1, z) +
            m33*fmc.src(x+1, y+1, z)
           ) / d;
}

/********* FF+ built-in functions ***********/




//function added by Harald Heim on January 31, 2002
static UINT_PTR fm_pointer_to_buffer(INT_PTR a, INT_PTR x, INT_PTR y, INT_PTR z)
{
    if (a==0) //Input buffer
    {
        return (UINT_PTR)MyAdjustedInData +(x-fmc.x_start)*fmc.srcColDelta + (y-fmc.y_start)*fmc.srcRowDelta + z;
    }
    else if (a==1) //Temp1 buffer
    {
        return (UINT_PTR)fmc.tbuf + ((y-fmc.y_start)*fmc.columns + (x-fmc.x_start))*fmc.Z + z;
    }
    else if (a==2) //Temp2 buffer
    {
        return (UINT_PTR)fmc.t2buf + ((y-fmc.y_start)*fmc.columns + (x-fmc.x_start))*fmc.Z + z;
    }
    else if (a==3) //Output buffer
    {
        return (UINT_PTR)gStuff->outData + (x-fmc.x_start)*fmc.dstColDelta + (y-fmc.y_start)*fmc.dstRowDelta + z;
    }
    else if (a==4) //Temp3 buffer
    {
        return (UINT_PTR)fmc.t3buf + ((y-fmc.y_start)*fmc.columns + (x-fmc.x_start))*fmc.Z + z;
    }
    else if (a==5) //Temp4 buffer
    {
        return (UINT_PTR)fmc.t4buf + ((y-fmc.y_start)*fmc.columns + (x-fmc.x_start))*fmc.Z + z;
	}

	return false;

}


//function added by Harald Heim on Feb 5, 2002
static INT_PTR fm_phaseshift(INT_PTR a, INT_PTR b) 
    //a=Color Value, b=Shift Value from 0 to 512
{   
    if (b < 0) b=0;
    else if (b > 512) b=512;

    if (b==0)
        return a;
    else
        //return 255 - fm_scl( fm_sin( fm_scl(a,0,255,256-b*4,768+b*4) ) ,-512,511,0,255 );
        return 255 - ( fm_sin( fm_scl(a,0,255,256-b*4,768+b*4) ) +512 ) / 4;

}

//function added by Harald Heim on Feb 5, 2002
static INT_PTR fm_blend(INT_PTR a, INT_PTR b, INT_PTR z, INT_PTR m, INT_PTR r) 
    //a=Bottom Color Value, b=Top Color Value, m=Blend Mode, r = blending ratio from 0 (bottom) to 255 (top)
{   
    THREAD_LOCAL static INT_PTR noise, col;
    INT_PTR calc, scaled;
	INT_PTR maxVal,midVal,quarVal,bitMultiply;

	if (bitDepthMode==16){ 
		maxVal=32768;
		midVal=16384;
		quarVal=8192;
		bitMultiply=128;
	} else {
		maxVal=255;
		midVal=128;
		quarVal=64;
		bitMultiply=1;
	}


    if (a < 0) a=0;	//afh 01-may-2008
    else if (a > maxVal) a=maxVal;	//afh 01-may-2008
    if (b < 0) b=0;
    else if (b > maxVal) b=maxVal;
    if (r < 0) r=0;
    else if (r > maxVal) r=maxVal;

    if (r==255 && m<4)
        return b;
    else if (r==0)
        return a;
    else
    {
        
        switch (m)
        {
            case 0://Normal
                if (r==255)
					return b;
				else
					return ((255-r)*a + r*b)/255;
					//return fm_mix(a,b,255-r,255);

            case 1: //Dissolve
                if (z==0) noise=fm_rnd(0,255);
                return r < noise ? a : b;

            case 2: //Threshold
                if (z==0) col = a;
                return col < (255-r)*bitMultiply ? a : b;

            case 3: //Threshold 2
                if (z==0) col = b;
                return col < (255-r)*bitMultiply ? a : b;

            case 4: //Multiply
                calc = a*b/maxVal;
                break;

            case 5: //Screen
                calc = maxVal-((maxVal-a)*(maxVal-b))/maxVal;
                break;

            case 6: //Overlay
                calc = a<midVal ? (2*a*b)/maxVal : maxVal-(2*(maxVal-a)*(maxVal-b))/maxVal;
                break;

            case 7: //Soft Light
                //scaled = fm_scl(a,0,maxVal,64,192);
                scaled = a/2 + quarVal;
                calc =  a<midVal ? (2*b*scaled)/maxVal : maxVal-2*(maxVal-scaled)*(maxVal-b)/maxVal;
                break;
            
            case 8: //Hard Light
                calc = b<midVal ? (2*a*b)/maxVal : maxVal-(2*(maxVal-a)*(maxVal-b))/maxVal;
                break;

            case 9://Dodge
                //calc = maxVal*a/max(1,maxVal-b);
                calc = b >= maxVal? maxVal : min(a * maxVal / (maxVal - b), maxVal);
                //calc = (a * 256 / (256 - b)) - 1; //Faster Alternative
                break;

            case 10: //Burn
                //calc = maxVal-(maxVal*(maxVal-a)/(b+1));
                calc = b <= 0? 0 : max(maxVal - ((maxVal - a) * maxVal / b), 0);
                break;

            case 11://Darken
                calc = min(a,b);
                break;

            case 12: //Lighten
                calc = max(a,b);
                break;

            case 13://Exclusion
                calc = maxVal-(((maxVal-a)*(maxVal-b)/maxVal)+(a*b/maxVal));
                break;

            case 14://Difference
                calc = fm_dif(a,b);
                break;

            case 15://Negative Difference 1
                calc = maxVal-fm_dif(a,maxVal-b);
                break;

            case 16://Negative Difference 2
                calc = fm_dif(maxVal-a,b);
                break;

            case 17://Subtract
                calc = a-b;
                break;

            case 18://Add
                calc = a+b;
                break;

            case 19://Expose
                calc = a+a-b;
        }//end switch


        if (r==255)
			return calc;
		else
			return ((255-r)*a + r*calc)/255;
			//return fm_mix(a, calc, 255-r,255);

	}

}


//function added by Harald Heim on Feb 5, 2002
static INT_PTR fm_contrast(INT_PTR a, INT_PTR b) 
    //a=Color Value, b=Contrast Value from -127 to 128
{
    if (b < -127) b=-127;	//afh 01-may-2008
    else if (b > 128) b=128;	//afh 01-may-2008

    return fm_scl(a,b,257-b,0,255);
}

//function added by Harald Heim on Feb 5, 2002
static INT_PTR fm_posterize(INT_PTR a, INT_PTR b) 
    //a=Color Value, b=Posterize Value/Number of Colors from 0 to 255
{ 
    if (b < 0) b=0;
    else if (b > 255) b=255;

    //Bad formula
    /*
    if (b==0)
        return 0;
    if (b==255)
        return a;
    else
        return fm_scl( fm_scl(a,0,255,0,b) ,0,b,0,255 );
    */

    if (b==0)
        return a;
    else
        return a - (a % b);
}


//function added by Harald Heim on Feb 5, 2002
static INT_PTR fm_saturation(INT_PTR r, INT_PTR g, INT_PTR b, INT_PTR z, INT_PTR sat) 
    // r,g,b = color values, z = returned channel, sat = saturation value between -500 and 500
{
    INT_PTR gray,sat2,channel;

    //Which channel to return?
    if (z==0) channel=r;
    else if (z==1) channel=g;
    else if (z==2) channel=b;

    if (sat==0)
    {
        return channel;
    }
    else
    {
        //test limits
        if (sat < -500) b=-500;				//eh??? sat=-500???
        else if (sat > 500) b=500;			//eh??? sat=500???

        //Gray value
        //gray= (r+b+g)/3;
        gray =( max(max(r,g),b) + min(min(r,g),b) ) /2;

        if (sat==-500)
        {
            return gray;
        }
        else
        {           
            //Saturate or desaturate?
            if (sat < 0) sat2=(sat/5)+100;
            else sat2=sat+100;

            return ( (100-sat2)*gray + sat2*channel )/100;
        }

    }

}

//function added by Harald Heim on Feb 6, 2002
static INT_PTR fm_msk(INT_PTR x, INT_PTR y)
{
        if (x < 0 || x >= fmc.X || y < 0 || y >= fmc.Y)
        {   // pin the coordinates (return 0 if outside, NOT replicate edges!!!)
            return 0;
        }
        else if (gStuff->haveMask)
        { // get value from selection mask

            //if (gStuff->imageMode > 9
            //) {
            //    return *(unsigned16 *)(((unsigned8 *)MyAdjustedMaskData) +
            //                   (((x-fmc.x_start)*fmc.mskColDelta)) +  // <<1
            //                   (y-fmc.y_start)*fmc.mskRowDelta);    
            //} else {    
           
                return ((unsigned8 *)MyAdjustedMaskData)
                    [(x-fmc.x_start)*fmc.mskColDelta + (y-fmc.y_start)*fmc.mskRowDelta];    
            //}
        }
        /*else if (!gNoMask)
        {   // get value from alpha channel
            return ((unsigned8 *)MyAdjustedInData)
                [(x-fmc.x_start)*fmc.srcColDelta + (y-fmc.y_start)*fmc.srcRowDelta + 3];
        }*/
        else
        {
            //if (gStuff->imageMode > 9
            //  ) {
            //    return 65535; //32768;
            //}else {
                return 255;
            //}

        }
}


/**
//function added by Harald Heim on Feb 6, 2002
static INT_PTR fm_egw(INT_PTR b, INT_PTR t, INT_PTR v)
{
    INT_PTR swapvalue;

    if (t==0 && b==0)
        return 0;
    if (b==t)
        return b;

    //added by Harald Heim on Mar 8, 2002
    b=IABS(b);	//afh 01-may-2008
    t=IABS(t);	//afh 01-may-2008
    if (b>t)
    {
        swapvalue = t;
        t = b;
        b = swapvalue;
    }

    if (v > t)
        return b + (v-b) % (t-b) ;
    else if (v < b)
        return t - (b - v) % (t-b);
    else
        return v;

}/**/

//------
// routines
//------
// static INT_PTR fm_egw(INT_PTR edge_a, INT_PTR edge_b, INT_PTR value)
//		Arguments
//			edge_a
//				edge_a edge value
//			edge_b
//				edge_b edge value
//			value
//				value that will be edge wrapped.
//		Return
//			The edge-wrapped value will be returned.
//		Description
//			This function returns the value untouched if it lies between edge_a and edge_b.
//			If it lies outside edge_a and edge_b, the value will be edge wrapped to lie between
//			edge_a and edge_b as if the image is continuous.
//		Examples:
//			Input value of 21 will be returned as 11 if edge_a = 10 and edge_b = 20
//			Input value of 21 will be returned as 12 if edge_a = 10 and edge_b = 19
//			Input value of  8 will be returned as 18 if edge_a = 10 and edge_b = 20
//			Input value of  5 will be returned as 14 if edge_a = 10 and edge_b = 19.
//------
static INT_PTR fm_egw(INT_PTR edge_a, INT_PTR edge_b, INT_PTR value)
{
	INT_PTR swapvalue;

	//RC    if (edge_b == 0 && edge_a == 0)  return 0;   redundant with next statement??
	if (edge_a == edge_b) return edge_a;

	if (edge_a < 0) edge_a = -edge_a;
	if (edge_b < 0) edge_b = -edge_b;
	if (edge_a > edge_b)
	{
		swapvalue 	= edge_b;
		edge_b 		= edge_a;
		edge_a 		= swapvalue;
	}

	if (value > edge_b)
		return edge_a + (value - edge_a) % (edge_b - edge_a) ;
	else if (value < edge_a)
		return edge_b - (edge_a - value) % (edge_b - edge_a);
	else
		return value;
}


//------
// static INT_PTR fm_egm(INT_PTR edge_a, INT_PTR edge_b, INT_PTR value)
//
//		Arguments
//			edge_a
//				edge_a edge value
//			edge_b
//				edge_b edge value
//			value
//				value that will be edge mirrored.
//		Return
//			edge-mirrored value will be returned
//			-1 if after mirroring one time the computed value is outside the range.
//		Description
//			This function returns the value untouched if it lies between edge_a and edge_b.
//			If it lies outside edge_a and edge_b, the value will be edge mirrored to lie between
//			edge_a and edge_b as if the image is reflected back at each edge.
//		Examples:
//			Input value of 21 will be returned as 19 if edge_a = 10 and edge_b = 20
//			Input value of 21 will be returned as 18 if edge_a = 10 and edge_b = 19.
//			Input value of  8 will be returned as 12 if edge_a = 10 and edge_b = 20
//			Input value of  5 will be returned as 15 if edge_a = 10 and edge_b = 19.
//------
static INT_PTR fm_egm(INT_PTR edge_a, INT_PTR edge_b, INT_PTR value)
{
	INT_PTR swapvalue;

	if (edge_a == edge_b) return edge_a;

	if (edge_a < 0) edge_a = -edge_a;
	if (edge_b < 0) edge_b = -edge_b;
	if (edge_a > edge_b)
	{
		swapvalue 	= edge_b;
		edge_b 		= edge_a;
		edge_a 		= swapvalue;
	}

	if (value > edge_b)
	{
		value = edge_b * 2 - value;
		if(value < edge_a) value = -1;
	}
	else if (value < edge_a)
	{
		value = edge_a - value;
		if(value > edge_b) value = -1;
	}
	return value;
}



//function added by Harald Heim on Feb 6, 2002
static INT_PTR fm_gray(INT_PTR r, INT_PTR g, INT_PTR b, INT_PTR rweight, INT_PTR gweight, INT_PTR bweight)
{
    INT_PTR totalweight;
    totalweight = rweight + gweight + bweight;

    if (totalweight==0)
        return 0;
    if (r == g && g == b)
        return r;
    else
        return (rweight*r + gweight*g + bweight*b) / (totalweight);
}

//function added by Harald Heim on Feb 6, 2002
/*static INT_PTR fm_swap (INT_PTR & a, INT_PTR & b)
{
    INT_PTR c;

    c = a;
    a = b;
    b = c;

    return true;
}*/


//function added by Harald Heim on Feb 6, 2002
static INT_PTR fm_rgb2iuv(INT_PTR r, INT_PTR g, INT_PTR b, INT_PTR z)
{
    if (z == 0)
        return (76*r + 150*g + 29*b)/256;
    else if (z == 1)
        return (-19*r - 37*g + 56*b)/256;
    else if (z == 2)
        return (78*r - 65*g - 13*b)/256;
    else
        return 0;
}

//function added by Harald Heim on Feb 8, 2002
static INT_PTR fm_iuv2rgb(INT_PTR i, INT_PTR u, INT_PTR v, INT_PTR z)
{
    if (z == 0)
        return (39*i + 90*v)/39;
    else if (z == 1)
        return (39*i - 31*u - 46*v)/39;
    else if (z == 2)
        return (39*i + 158*u)/39;
    else
        return 0;
}

//function added by Harald Heim on Feb 8, 2002
static INT_PTR fm_rgb2cmyk(INT_PTR r, INT_PTR g, INT_PTR b, INT_PTR z)
{
    INT_PTR k;
    INT_PTR bitMultiply = (bitDepthMode == 16 ? 128 : 1);

    k=255*bitMultiply-max(r,max(g,b));

    if (z == 0)
        return 255*bitMultiply-r-k;
    else if (z == 1)
        return 255*bitMultiply-g-k;
    else if (z == 2)
        return 255*bitMultiply-b-k;
    else if (z == 3)
        return k;
    else
        return 0;
}

//function added by Harald Heim on Feb 8, 2002
static INT_PTR fm_cmyk2rgb(INT_PTR c, INT_PTR m, INT_PTR y, INT_PTR k, INT_PTR z)
{
    INT_PTR bitMultiply = (bitDepthMode == 16 ? 128 : 1);

    if (z == 0)
        return 255*bitMultiply-c-k;
    else if (z == 1)
        return 255*bitMultiply-m-k;
    else if (z == 2)
        return 255*bitMultiply-y-k;
    else
        return 0;
}

//function added by Harald Heim on Feb 6, 2002
static INT_PTR fm_solarize(INT_PTR a, INT_PTR s)
{
    if (s < 0) s=0;
    else if (s > 255) s=255;

    return IABS(s-a);
}


//function added by Harald Heim on Feb 7, 2002
static INT_PTR fm_tricos(INT_PTR x)
{ //triangle wave similiar to cos()
    x=IABS(x);	//afh 01-may-2008

    if (x%1024 < 512) 
        return 512 - ((x%512)*2);
    else 
        return ((x%512)*2) - 512;
}

//function added by Harald Heim on Feb 7, 2002
static INT_PTR fm_tri(INT_PTR x)
{
    return fm_tricos(x-256);
}

//function added by Harald Heim on Feb 9, 2002
static INT_PTR fm_sinbell(INT_PTR x)
{
    //Make sure x is between 0 and 1024 to make sure that the output is in the same range
    x = IABS(x%1025);

    return fm_sin(x-256)+512;
}


//function added by Harald Heim on Feb 7, 2002
static INT_PTR fm_grad2D(INT_PTR x, INT_PTR y, INT_PTR X, INT_PTR Y,
						 INT_PTR grad, INT_PTR dist, INT_PTR repeat)
// x / y = current x / y value
// X / Y = maximal value  of gradient field
//grad = gradient type
//dist = 0 for linear, 1 for sine distribution
//repeat = number of repetitions, 0 for no repetition
{
    INT_PTR calc, m, M, d;

    if (X==0 || Y==0)
        return 0;

    //Linear gradient formulas
    switch(grad)
    {
        case 0: //Horizontal
            calc = x*255/X;
            break;

        case 1: // Vertical
            calc = y*255/Y;
            break;

        case 2: //Diagonal
            calc = (x+y)*255/(X+Y); //or  calc = x*128/X + y*128/Y;
            break;

        case 3: //Diagonal 2
            calc = 128+(x*255/X - y*255/Y)/2;
            break;

        case 4: //Radial
            m = fm_c2m(x-X/2,y-Y/2);
            M = fm_c2m(X,Y)/2; 
            calc = m*255/M;
            break;

        case 5: //Ellipsoid
            //calc = fm_c2m(fm_scl(x,0,X,-1024,1024), fm_scl(y,0,Y,-1024,1024))/5;
            calc = fm_c2m(x*2048/X-1024, y*2048/Y-1024)/5;
            break;

        case 6: //Pyramid
            if (x<=X/2 && y<=Y/2)    calc= 2 * min(x*255/X,y*255/Y);
            else if (x>X/2 && y<Y/2) calc= 2 * min(255-x*255/X,y*255/Y);
            else if (x<X/2 && y>Y/2) calc= 2 * min(x*255/X,255-y*255/Y);
            else                     calc= 2 * min(255-x*255/X,255-y*255/Y);
            break;

        case 7: //Beam
            calc = fm_c2d(x,y);
            break;

        case 8: //Angular
            d = fm_c2d(x-X/2,y-Y/2);
            //calc = fm_scl(d,-512,512,255,0);
            calc = 255-(d+512)/4;
            break;

        case 9: //Star
            calc = 128-((128-x*255/X)*(128-y*255/Y)/128);   
            break;

        case 10: //Quarter Pyramid 1
            calc = min(x*255/X,y*255/Y);
            break;

        case 11: //Quarter Pyramid 2
            calc = 255-max(x*255/X,y*255/Y);
            break;

        case 12: //Quarter Radial
            M = fm_c2m(X,Y)/2; 
            calc = fm_c2m(x,y)*128/M;
            break;
            
    }//end switch


    //Sine distribution & repetition
    if (dist > 0)
    {
        if (repeat==0)
            repeat = 2;
        else
            repeat = repeat * 4;

        //calc = fm_scl( fm_sin(calc*repeat-256) ,-511,512,0,255);
        calc = ( fm_sin(calc*repeat-256) + 512) / 4  ;
    }

    //Linear repetitions
    else if  (dist ==0 && repeat>0)
    {
        repeat = repeat * 4;

        //calc = fm_scl( fm_tricos(calc*repeat+512),-511,512,0,255); 
        calc = ( fm_tricos(calc*repeat+512) + 512) / 4; 
    }   

    return calc;
}

//function added by Harald Heim on Feb 8, 2002
static INT_PTR fm_tone(INT_PTR a, INT_PTR h, INT_PTR m, INT_PTR d)
{
    if (h < -128) h=-128;
    else if (h > 128)  h=128;
    if (m < -128) m=-128;
    else if (m > 128)  m=128;
    if (d < -128) d=-128;
    else if (d > 128)  d=128;
    
    return a+(d*(256-IABS(a-43))/256)+(m*(256-2*IABS(a-128))/256)+(h*(256-IABS(a-213))/256);
}

//function added by Harald Heim on Feb 10, 2002
static INT_PTR fm_xyzcnv(INT_PTR x, INT_PTR y, INT_PTR z,
						 INT_PTR m11, INT_PTR m12, INT_PTR m13,
						 INT_PTR m21, INT_PTR m22, INT_PTR m23,
						 INT_PTR m31, INT_PTR m32, INT_PTR m33,
						 INT_PTR d)
{
    if (d == 0)
    {   // default divisor to sum of weights...
        d = m11+m12+m13+m21+m22+m23+m31+m32+m33;    // ???
        if (d == 0)
            return 255; // give up if divisor still 0
    }
    return (
        //could check for safe coordinates in toto and
        //then use unsafe src() to speed up...
        //or set up pointer to src(x-1,y-1,z), then
        //bump it through all 9 points without
        //recomputing index expression each time...
            m11*fmc.src(x-1, y-1, z) +
            m12*fmc.src(x  , y-1, z) +
            m13*fmc.src(x+1, y-1, z) +
            m21*fmc.src(x-1, y  , z) +
            m22*fmc.src(x  , y  , z) +
            m23*fmc.src(x+1, y  , z) +
            m31*fmc.src(x-1, y+1, z) +
            m32*fmc.src(x  , y+1, z) +
            m33*fmc.src(x+1, y+1, z)
           ) / d;
}

//function added by Harald Heim on Feb 10, 2002
static INT_PTR fm_cell_initialize(INT_PTR i)			//is sense reversed???
{
    if (i==0) 
        Dont_Initialize_Cells = 0; //Cells aren't initialized
    else
        Dont_Initialize_Cells = 1; //Cells are initialized
    
    return true;

}

/*
//function added by Harald Heim on Feb 10, 2002
static INT_PTR fm_tbuffer_initialize(INT_PTR i)
{
    if (i==0) 
        Dont_Initialize_Tbuffer = 0; //Tbuffer isn't initialized
    else
        Dont_Initialize_Tbuffer = 1; //Tbuffer is initialized
    
    return true;

}

//function added by Harald Heim on Feb 10, 2002
static INT_PTR fm_t2buffer_initialize(INT_PTR i)
{
    if (i==0) 
        Dont_Initialize_T2buffer = 0; //T2buffer isn't initialized
    else
        Dont_Initialize_T2buffer = 1; //T2buffer is initialized
    
    return true;

}
*/

//// Use an enum for edge mode, expose to FD???

//function added by Harald Heim on Feb 13, 2002
static INT_PTR fm_set_edge_mode(INT_PTR mode)
{ 
	// How about let mode=-1 mean no checking??
	// and mode==4 mean the "other" mirror?
//mode=0 for regular behaviour
//mode=1 for black border
//mode=2 for edge wrap
//mode=3 for edge mirror
    if (mode < 0 && mode > 4) return false;

    edgeMode = (int)mode;
    return true;
}


//
//	BOOL set_bitdepth_mode(int mode)
//
//  Sets current working bits per channel,
//	where mode = 8 or 16.
//	Returns true if successful, or
//	false if mode is not 8 or 16.
//
//Added by Harald Heim
//Modified by Alex Hunter on May 6, 2014
//
static INT_PTR fm_set_bitdepth_mode(INT_PTR mode)
{ 
	if (mode == 8) {
		//8 bits per channel
		bitDepthMode = (int)mode;
		bitMultiply255 = 1.0*255.0;
		bitDivide255 = 1.0 / bitMultiply255;
	}
	else if (mode == 16) {
		//16 bits per channel
		bitDepthMode = (int)mode;
		bitMultiply255 = 128.0*256.0;
		bitDivide255 = 1.0 / bitMultiply255;
	}
	else return false;	//invalid bpc

    return true;	//success
}/*fm_set_bitdepth_mode*/

/// Use pluggable variants of psetp() instead of testing a flag at run-time!!!

static INT_PTR fm_set_psetp_mode(INT_PTR mode)
{ 
//mode=0 for regular behaviour (not possible to set alpha channel to zero) for more performance and avoiding bugs
//mode=1 for always setting the alpha channel value

    if (mode < 0 || mode > 1) return false;

    psetpMode = (int)mode;
    return true;
}




//function added by Harald Heim on Feb 11, 2002
//
//	iget: return interpolated value at floating point coordinates (x, y)
//		  in channel z, from the specified 'buffer' using the
//		  specified interpolation 'mode', where:
//
//	buffer = 0 => source input buffer,
//		   = 1 => tbuf,
//		   = 2 => t2buf,
//		   = 3 => output buffer.
//		   = 4 => t3buf,
//		   = 5 => t4buf
//
//	mode   = 0 => nearest neighbor
//		   = 1 => bisquare
//	       = 2 => bicosine
//		   = 3 => bilinear
//		   = 4 => bicubic
//
// The current edgemode is used to determine wrapping, mirroring, etc.
//
static INT_PTR fm_iget(double x, double y, INT_PTR z, INT_PTR buffer, INT_PTR mode)
{ 
    INT_PTR x_int, y_int, x_int1, y_int1,v1,v2,v3,v4;
    double fracx, fracy;
    //Cached values
    THREAD_LOCAL static INT_PTR x_int_old, y_int_old, x_int1_old, y_int1_old,mode_old, edgeMode_old;
    THREAD_LOCAL static double fracx_old, fracy_old, x_old, y_old;


    //Info ("x: %f - y: %f",x,y);

    if (z >= 0 && z < fmc.Z)
    {
        
        if (mode==0) //Nearest Neighbor
        {
            
            x_int=(INT_PTR) x;
            y_int=(INT_PTR) y;

            switch(edgeMode)
            {
                case 0:
                    if (x_int < 0) x_int = 0;
                    else if (x_int >= fmc.X) x_int = fmc.X - 1;
                    if (y_int < 0) y_int = 0;
                    else if (y_int >= fmc.Y) y_int = fmc.Y - 1;
                    break;

                case 1:
                    if (x_int < 0 || y_int < 0 || x_int >= fmc.X || y_int >= fmc.Y)
                        return 0;
					break;				//was missing!!
                
                case 2:
                    if (x_int < 0) x_int = (fmc.X - 1) + x_int;		//No mod here & elsewhere?
                    else if (x_int >= fmc.X) x_int = (x_int - (fmc.X - 1)) % fmc.X;
                    if (y_int < 0) y_int = (fmc.Y - 1) + y_int;
                    else if (y_int >= fmc.Y) y_int = (y_int - (fmc.Y - 1)) % fmc.Y;
                    break;

                case 3:
                    if (x_int < 0) x_int = - x_int;
                    else if (x_int >= fmc.X) x_int = (fmc.X - 1) - (x_int - (fmc.X - 1));
                    if (y_int < 0) y_int = - y_int;
                    else if (y_int >= fmc.Y) y_int = (fmc.Y - 1) - (y_int - (fmc.Y - 1));
					break;	//good practice

				default:	//no special handling
					;	//afh 01-may-2008
            }//end switch 
        

            switch (buffer)
            {
				default:	//do something reasonable even if "buffer" isn't.
                case 0: //input buffer
                    if (gStuff->imageMode < 10)
                        return ((unsigned8 *)MyAdjustedInData)
                            [(x_int-fmc.x_start)*fmc.srcColDelta + (y_int-fmc.y_start)*fmc.srcRowDelta + z];
                    else
                        return *(unsigned16 *)(((unsigned8 *)MyAdjustedInData) +
                               (((x_int-fmc.x_start)*fmc.srcColDelta + z) << 1) +
                               (y_int-fmc.y_start)*fmc.srcRowDelta);
                case 1: //t buffer 1
                    if (gStuff->imageMode < 10)
						return fmc.tbuf[(y_int*fmc.X + x_int)*fmc.Z + z];
					else
						return ((unsigned16 *)(fmc.tbuf))[(y_int*fmc.X + x_int)*fmc.Z + z];
                case 2: //t buffer 2
                    if (gStuff->imageMode < 10)
						return fmc.t2buf[(y_int*fmc.X + x_int)*fmc.Z + z];
					else
						return ((unsigned16 *)(fmc.t2buf))[(y_int*fmc.X + x_int)*fmc.Z + z];
                case 3: //output buffer
                    if (gStuff->imageMode < 10)
                        return ((unsigned8 *)gStuff->outData)
                            [(x_int-fmc.x_start)*fmc.dstColDelta + (y_int-fmc.y_start)*fmc.dstRowDelta + z];
                    else
                        return *(unsigned16 *)  ( ((unsigned8 *)gStuff->outData) +
                            (((x_int-fmc.x_start)*fmc.dstColDelta + z)<<1 ) + (y_int-fmc.y_start)*fmc.dstRowDelta  );
				case 4: //t buffer 3
                    if (gStuff->imageMode < 10)
						return fmc.t3buf[(y_int*fmc.X + x_int)*fmc.Z + z];
					else
						return ((unsigned16 *)(fmc.t3buf))[(y_int*fmc.X + x_int)*fmc.Z + z];
				case 5: //t buffer 4
                    if (gStuff->imageMode < 10)
						return fmc.t4buf[(y_int*fmc.X + x_int)*fmc.Z + z];
					else
						return ((unsigned16 *)(fmc.t4buf))[(y_int*fmc.X + x_int)*fmc.Z + z];
            }


        }//mode 0

        else 
        {
            
            if (mode<4) 
            {	//mode 1,2,3 (or < 0 !)

                if (x==x_old && y==y_old && mode==mode_old && edgeMode==edgeMode_old)
                {
                    //Set cached values
                    fracx=fracx_old;
                    fracy=fracy_old;
                    x_int=x_int_old;
                    y_int=y_int_old; 
                    x_int1=x_int1_old;
                    y_int1=y_int1_old;
                }

                else //Calculate the values
                {

                    switch (edgeMode)
                    {
                        case 0:
                            if (x < 0) x = 0;
                            else if (x >= fmc.X) x = (double)(fmc.X - 1);
                            if (y < 0) y = 0;
                            else if (y >= fmc.Y) y = (double)(fmc.Y - 1);
                            break;

                        case 1:
                            if (x < 0 || y < 0 || x >= (fmc.X - 1)|| y >= (fmc.Y - 1))
                                return 0;
							break;				//was missing!!
                        
                        case 2:
                            if (x < 0) x = fmc.X + x;
                            else if (x >= fmc.X) x = fmod(x - fmc.X , fmc.X) ;
                            if (y < 0) y = fmc.Y + y;
                            else if (y >= fmc.Y) y = fmod(y - fmc.Y , fmc.Y);
                            break;

                        case 3:
                            if (x < 0) x = - x;
                            else if (x >= fmc.X) x = (fmc.X - 1) - (x - (fmc.X - 1));
                            if (y < 0) y = - y;
                            else if (y >= fmc.Y) y = (fmc.Y - 1) - (y - (fmc.Y - 1));
							break;

						default:	//no special handling
							;	//afh 01-may-2008
                    }//end switch


                    x_int=(INT_PTR) x;
                    y_int=(INT_PTR) y;

                    x_int1=x_int+1;
                    y_int1=y_int+1;
                    switch (edgeMode)
                    {
						default:
                        case 0: 
                            if (x_int1 >= fmc.X) x_int1 = (fmc.X - 1);
                            if (y_int1 >= fmc.Y) y_int1 = (fmc.Y - 1);
                            break;
                        case 1: //same as case 0??
                            if (x_int1 >= fmc.X) x_int1 = (fmc.X - 1);
                            if (y_int1 >= fmc.Y) y_int1 = (fmc.Y - 1);
                            break;
                        case 2:
                            if (x_int1 >= fmc.X) x_int1 = (x_int1  - fmc.X) % fmc.X;
                            if (y_int1 >= fmc.Y) y_int1 = (y_int1  - fmc.Y) % fmc.Y;
                            break;
                        case 3:
                            if (x_int1 >= fmc.X) x_int1 = (fmc.X - 1) - (x_int1 - (fmc.X - 1));
                            if (y_int1 >= fmc.Y) y_int1 = (fmc.Y - 1) - (y_int1 - (fmc.Y - 1));
							break;	//afh 01-may-2008
                    }//end switch
                    
                    //Info ("x_int: %d - y_ing: %d",x_int,y_int);

                    fracx=x-x_int;
                    fracy=y-y_int;

                    //Info ("fracx: %f - fracy: %f",fracx,fracy);
                    

                    if (mode==2) //Bicosine
                    {
                        if (fracx==fracy)
                        {
                            fracx= (1.0-cos(fracx*3.1415927))*0.5;
                            fracy=fracx;
                        }
                        else
                        {
                            fracx= (1.0-cos(fracx*3.1415927))*0.5;
                            fracy= (1.0-cos(fracy*3.1415927))*0.5;
                        }
                    }
                    else if (mode == 1) //Bisquare
                    {
                        fracx *= fracx;
                        fracy *= fracy;
                    }
                
                }//Calculate not cached

                
                switch (buffer)
                {
					default:
                    case 0: //input buffer
                        if (gStuff->imageMode < 10){
                            v1 = ((unsigned8 *)MyAdjustedInData)
                                [(x_int-fmc.x_start)*fmc.srcColDelta + (y_int-fmc.y_start)*fmc.srcRowDelta + z];
                            v2 = ((unsigned8 *)MyAdjustedInData)
                                [(x_int-fmc.x_start)*fmc.srcColDelta + (y_int1-fmc.y_start)*fmc.srcRowDelta + z];
                            v3 = ((unsigned8 *)MyAdjustedInData)
                                [(x_int1-fmc.x_start)*fmc.srcColDelta + (y_int-fmc.y_start)*fmc.srcRowDelta + z];
                            v4 = ((unsigned8 *)MyAdjustedInData)
                                [(x_int1-fmc.x_start)*fmc.srcColDelta + (y_int1-fmc.y_start)*fmc.srcRowDelta + z];
                        } else {
                            v1 = *(unsigned16 *)(((unsigned8 *)MyAdjustedInData) +
                               (((x_int-fmc.x_start)*fmc.srcColDelta + z) << 1) +
                               (y_int-fmc.y_start)*fmc.srcRowDelta);
                            v2 = *(unsigned16 *)(((unsigned8 *)MyAdjustedInData) +
                               (((x_int-fmc.x_start)*fmc.srcColDelta + z) << 1) +
                               (y_int1-fmc.y_start)*fmc.srcRowDelta);
                            v3 = *(unsigned16 *)(((unsigned8 *)MyAdjustedInData) +
                               (((x_int1-fmc.x_start)*fmc.srcColDelta + z) << 1) +
                               (y_int-fmc.y_start)*fmc.srcRowDelta);
                            v4 = *(unsigned16 *)(((unsigned8 *)MyAdjustedInData) +
                               (((x_int1-fmc.x_start)*fmc.srcColDelta + z) << 1) +
                               (y_int1-fmc.y_start)*fmc.srcRowDelta);
                        }
                        break;

                    case 1: //temp buffer 1
                        if (gStuff->imageMode < 10){
							v1 = fmc.tbuf[(y_int*fmc.X + x_int)*fmc.Z + z];
							v2 = fmc.tbuf[(y_int1*fmc.X + x_int)*fmc.Z + z];
							v3 = fmc.tbuf[(y_int*fmc.X + x_int1)*fmc.Z + z];
							v4 = fmc.tbuf[(y_int1*fmc.X + x_int1)*fmc.Z + z]; 
                        } else {
							v1 = ((unsigned16 *)(fmc.tbuf))[(y_int*fmc.X + x_int)*fmc.Z + z];
							v2 = ((unsigned16 *)(fmc.tbuf))[(y_int1*fmc.X + x_int)*fmc.Z + z];
							v3 = ((unsigned16 *)(fmc.tbuf))[(y_int*fmc.X + x_int1)*fmc.Z + z];
							v4 = ((unsigned16 *)(fmc.tbuf))[(y_int1*fmc.X + x_int1)*fmc.Z + z]; 
						}
                        break;

                    case 2: //temp buffer 2
                        if (gStuff->imageMode < 10){
							v1 = fmc.t2buf[(y_int*fmc.X + x_int)*fmc.Z + z];
							v2 = fmc.t2buf[(y_int1*fmc.X + x_int)*fmc.Z + z];
							v3 = fmc.t2buf[(y_int*fmc.X + x_int1)*fmc.Z + z];
							v4 = fmc.t2buf[(y_int1*fmc.X + x_int1)*fmc.Z + z]; 
                        } else {
							v1 = ((unsigned16 *)(fmc.t2buf))[(y_int*fmc.X + x_int)*fmc.Z + z];
							v2 = ((unsigned16 *)(fmc.t2buf))[(y_int1*fmc.X + x_int)*fmc.Z + z];
							v3 = ((unsigned16 *)(fmc.t2buf))[(y_int*fmc.X + x_int1)*fmc.Z + z];
							v4 = ((unsigned16 *)(fmc.t2buf))[(y_int1*fmc.X + x_int1)*fmc.Z + z]; 
						}
                        break;

                    case 3: //output buffer
                        if (gStuff->imageMode < 10){
                            v1 = ((unsigned8 *)gStuff->outData)
                                [(x_int-fmc.x_start)*fmc.dstColDelta + (y_int-fmc.y_start)*fmc.dstRowDelta + z];
                            v2 = ((unsigned8 *)gStuff->outData)
                                [(x_int-fmc.x_start)*fmc.dstColDelta + (y_int1-fmc.y_start)*fmc.dstRowDelta + z];
                            v3 = ((unsigned8 *)gStuff->outData)
                                [(x_int1-fmc.x_start)*fmc.dstColDelta + (y_int-fmc.y_start)*fmc.dstRowDelta + z];
                            v4 = ((unsigned8 *)gStuff->outData)
                                [(x_int1-fmc.x_start)*fmc.dstColDelta + (y_int1-fmc.y_start)*fmc.dstRowDelta + z];
                        } else {
                            v1 = *(unsigned16 *)  ( ((unsigned8 *)gStuff->outData) +
                                (((x_int-fmc.x_start)*fmc.dstColDelta + z)<<1 ) + (y_int-fmc.y_start)*fmc.dstRowDelta  );
                            v2 = *(unsigned16 *)  ( ((unsigned8 *)gStuff->outData) +
                                (((x_int-fmc.x_start)*fmc.dstColDelta + z)<<1 ) + (y_int1-fmc.y_start)*fmc.dstRowDelta  );
                            v3 = *(unsigned16 *)  ( ((unsigned8 *)gStuff->outData) +
                                (((x_int1-fmc.x_start)*fmc.dstColDelta + z)<<1 ) + (y_int-fmc.y_start)*fmc.dstRowDelta  );
                            v4 = *(unsigned16 *)  ( ((unsigned8 *)gStuff->outData) +
                                (((x_int1-fmc.x_start)*fmc.dstColDelta + z)<<1 ) + (y_int1-fmc.y_start)*fmc.dstRowDelta  );
                        }
						break;
					
					case 4: //temp buffer 3
                        if (gStuff->imageMode < 10){
							v1 = fmc.t3buf[(y_int*fmc.X + x_int)*fmc.Z + z];
							v2 = fmc.t3buf[(y_int1*fmc.X + x_int)*fmc.Z + z];
							v3 = fmc.t3buf[(y_int*fmc.X + x_int1)*fmc.Z + z];
							v4 = fmc.t3buf[(y_int1*fmc.X + x_int1)*fmc.Z + z]; 
                        } else {
							v1 = ((unsigned16 *)(fmc.t3buf))[(y_int*fmc.X + x_int)*fmc.Z + z];
							v2 = ((unsigned16 *)(fmc.t3buf))[(y_int1*fmc.X + x_int)*fmc.Z + z];
							v3 = ((unsigned16 *)(fmc.t3buf))[(y_int*fmc.X + x_int1)*fmc.Z + z];
							v4 = ((unsigned16 *)(fmc.t3buf))[(y_int1*fmc.X + x_int1)*fmc.Z + z]; 
						}
                        break;
					case 5: //temp buffer 4
                        if (gStuff->imageMode < 10){
							v1 = fmc.t4buf[(y_int*fmc.X + x_int)*fmc.Z + z];
							v2 = fmc.t4buf[(y_int1*fmc.X + x_int)*fmc.Z + z];
							v3 = fmc.t4buf[(y_int*fmc.X + x_int1)*fmc.Z + z];
							v4 = fmc.t4buf[(y_int1*fmc.X + x_int1)*fmc.Z + z]; 
                        } else {
							v1 = ((unsigned16 *)(fmc.t4buf))[(y_int*fmc.X + x_int)*fmc.Z + z];
							v2 = ((unsigned16 *)(fmc.t4buf))[(y_int1*fmc.X + x_int)*fmc.Z + z];
							v3 = ((unsigned16 *)(fmc.t4buf))[(y_int*fmc.X + x_int1)*fmc.Z + z];
							v4 = ((unsigned16 *)(fmc.t4buf))[(y_int1*fmc.X + x_int1)*fmc.Z + z]; 
						}
                        break;
                }//switch buffer
                


                //Set static variables for caching
                x_old = x;
                y_old = y;
                mode_old = mode;
                edgeMode_old=edgeMode;
                fracx_old=fracx;
                fracy_old=fracy;
                x_int_old=x_int;
                y_int_old=y_int; 
                x_int1_old=x_int1;
                y_int1_old=y_int1;


                return (INT_PTR) ((1.0-fracx)* ((1.0-fracy)*v1 + fracy*v2)  + fracx* ((1.0-fracy)*v3 + fracy*v4));
        
            }//mode<4

            else if (mode==4) //bicubic interpolation
            {

                //For Bicubic Interpolation 
               double nx,ny,ox3,oy3,nx3,ny3;
               double ix[4];//ix1, ix2, ix3, ix4;
               int j,k;
               INT_PTR i[4];//i1,i2,i3,i4;
               double x1,x2,x3,x4,y1,y2,y3,y4,y9;
            

                switch (edgeMode)
                {
                    case 0:
                        if (x < 0) x = 0;
                        else if (x >= fmc.X) x = (double)(fmc.X - 1);
                        if (y < 0) y = 0;
                        else if (y >= fmc.Y) y = (double)(fmc.Y - 1);
                        break;
                    case 1:
                        if (x < 0 || y < 0 || x >= (fmc.X - 1)|| y >= (fmc.Y - 1))
                            return 0;
						break;				//was missing!!
                    case 2:
                        if (x < 0) x = fmod(fmc.X + x , fmc.X);
                        else if (x >= fmc.X) x = fmod(x - fmc.X , fmc.X);
                        if (y < 0) y = fmod(fmc.Y + y , fmc.Y);
                        else if (y >= fmc.Y) y = fmod(y - fmc.Y , fmc.Y);
                        break;
                    case 3:
                        if (x < 0) x = - x;
                        else if (x >= fmc.X) x = (fmc.X - 1) - (x - (fmc.X - 1));
                        if (y < 0) y = - y;
                        else if (y >= fmc.Y) y = (fmc.Y - 1) - (y - (fmc.Y - 1));
						break;
					default:
						;	//afh 01-may-2008
                }//end switch


               y_int = (INT_PTR)y; //yy
               x_int = (INT_PTR)x; //xx

               fracx=x-x_int;
               fracy=y-y_int;

               nx = 1. - fracx;
               ny = 1. - fracy;
               ox3 = (fracx * fracx * fracx) - fracx;
               oy3 = (fracy * fracy * fracy) - fracy;
               nx3 = (nx * nx * nx) - nx;
               ny3 = (ny * ny * ny) - ny;

                
               for (j=-1;j<3;j++){

                   y_int1 = y_int + j;

                   switch (edgeMode)
                    {
                    case 0: 
                        if (y_int1 < 0) y_int1 = 0;
                        if (y_int1 >= fmc.Y) y_int1 = (fmc.Y - 1);
                        break;
                    case 1: 
                        if (y_int1 >= fmc.Y) y_int1 = (fmc.Y - 1);
                        break;
                    case 2:
                        if (y_int1 >= fmc.Y) y_int1 = (y_int1  - fmc.Y)% fmc.Y;
                        break;
                    case 3:
                        if (y_int1 >= fmc.Y) y_int1 = (fmc.Y - 1)- (y_int1 - (fmc.Y - 1));
						break;
					default:
						break;	//afh 01-may-2008
                    }//end switch

               
                   for (k=-1;k<3;k++) {
                        
                        x_int1 = x_int + k;

                        switch (edgeMode)
                        {
                            case 0: 
                                if (x_int1 < 0) x_int1 = 0;
                                if (x_int1 >= fmc.X) x_int1 = (fmc.X - 1);
                                break;
                            case 1: 
                                if (x_int1 >= fmc.X) x_int1 = (fmc.X - 1);
                                break;
                            case 2:
                                if (x_int1 >= fmc.X) x_int1 = (x_int1 - fmc.X) % fmc.X;
                                break;
                            case 3:
                                if (x_int1 >= fmc.X) x_int1 = (fmc.X - 1) - (x_int1 - (fmc.X - 1));
								break;
							default:
								break;	//afh 01-may-2008
                        }//end switch


                       switch (buffer){
							default:
                            case 0: //input buffer
                                if (gStuff->imageMode < 10)
                                    i[k+1] = ((unsigned8 *)MyAdjustedInData)
                                        [(x_int1 -fmc.x_start)*fmc.srcColDelta + (y_int1 -fmc.y_start)*fmc.srcRowDelta + z];
                                else
                                    i[k+1] = *(unsigned16 *)(((unsigned8 *)MyAdjustedInData) +
                                        (((x_int1-fmc.x_start)*fmc.srcColDelta + z) << 1) +
                                        (y_int1-fmc.y_start)*fmc.srcRowDelta);
                                break;
                            case 1: //temp buffer 1
                                if (gStuff->imageMode < 10)
									i[k+1] = fmc.tbuf[( y_int1 *fmc.X + x_int1 )*fmc.Z + z];
								else
									i[k+1] = ((unsigned16 *)(fmc.tbuf))[( y_int1 *fmc.X + x_int1 )*fmc.Z + z];
                                break;
                            case 2: //temp buffer 2
                                if (gStuff->imageMode < 10)
									i[k+1] = fmc.t2buf[( y_int1 *fmc.X + x_int1 )*fmc.Z + z];
								else
									i[k+1] = ((unsigned16 *)(fmc.t2buf))[( y_int1 *fmc.X + x_int1 )*fmc.Z + z];
                                break;
                            case 3: //output buffer
                                if (gStuff->imageMode < 10)
                                    i[k+1] = ((unsigned8 *)gStuff->outData)
                                        [(x_int1 -fmc.x_start)*fmc.dstColDelta + (y_int1 -fmc.y_start)*fmc.dstRowDelta + z];
                                else
                                    i[k+1] = *(unsigned16 *)  ( ((unsigned8 *)gStuff->outData) +
                                        (((x_int1-fmc.x_start)*fmc.dstColDelta + z)<<1 ) + (y_int1-fmc.y_start)*fmc.dstRowDelta  );
								break;
							case 4: //temp buffer 3
                                if (gStuff->imageMode < 10)
									i[k+1] = fmc.t3buf[( y_int1 *fmc.X + x_int1 )*fmc.Z + z];
								else
									i[k+1] = ((unsigned16 *)(fmc.t3buf))[( y_int1 *fmc.X + x_int1 )*fmc.Z + z];
                                break;
							case 5: //temp buffer 4
                                if (gStuff->imageMode < 10)
									i[k+1] = fmc.t4buf[( y_int1 *fmc.X + x_int1 )*fmc.Z + z];
								else
									i[k+1] = ((unsigned16 *)(fmc.t4buf))[( y_int1 *fmc.X + x_int1 )*fmc.Z + z];
                                break;
                        }//switch buffer

                    }//for k

                   
                   x1 = (double)((i[2] - i[1]) - (i[1] - i[0]));
                   x2 = (double)((i[3] - i[2]) - (i[2] - i[1]));
                   x3 = x1 - x2 / 4.;
                   x4 = x2 - x1 / 4.;
                   ix[j+1] = (fracx * i[2]) + (nx * i[1])
                    + (((x4 * ox3) + (x3 * nx3)) / 3.75);
               }//for j


               // Y
               y1 = (ix[2] - ix[1]) - (ix[1] - ix[0]);
               y2 = (ix[3] - ix[2]) - (ix[2] - ix[1]);
               y3 = y1 - y2 / 4.;
               y4 = y2 - y1 / 4.;
               y9 = (fracy * ix[2]) + (ny * ix[1])
                + (((y4 * oy3) + (y3 * ny3)) / 3.75);


                //Set static variables for caching
                //nx,ny,ox3,oy3,nx3,ny3

               return (INT_PTR)y9;

            }//end bicubic(mode 4)



        }// end interpolation methods


    }
    // bad channel number??
    return 0; 	//afh 01-may-2008
} /*fm_iget*/


//function added by Harald Heim on Feb 13, 2002
static INT_PTR fm_bCircle(INT_PTR x, INT_PTR y, INT_PTR centerx, INT_PTR centery, INT_PTR radius)
{
    return fm_c2m(x-centerx,y-centery) <= radius ? 1: 0;
    
}

//function added by Harald Heim on Feb 13, 2002
//static INT_PTR fm_bEllipse(INT_PTR x, INT_PTR y, INT_PTR centerx, INT_PTR centery, INT_PTR radiusx, INT_PTR radiusy)
//{
//  return fm_c2m(fm_scl(x,0,fmc.X,-radiusx*fmc.Y/fmc.X,radiusx*fmc.Y/fmc.X) -centerx , fm_scl(y,fmc.Y,0,-radiusy,radiusy) -centery ) >= fmc.Y/2 ? 1: 0;
//}



//function added by Harald Heim on Feb 13, 2002
static INT_PTR fm_bRect(INT_PTR x, INT_PTR y, INT_PTR centerx, INT_PTR centery, INT_PTR radiusx, INT_PTR radiusy)
{
    return  x >= centerx - radiusx  && x <= centerx + radiusx && y >= centery-radiusy && y <= centery+radiusy  ? 1: 0;
    
}

//function added by Harald Heim on Feb 13, 2002
static INT_PTR fm_bRect2(INT_PTR x, INT_PTR y, INT_PTR topx, INT_PTR topy, INT_PTR bottomx, INT_PTR bottomy)
{
    return  x >= topx && x <= bottomx && y >= topy && y <= bottomy  ? 1: 0;
    
}

//function added by Harald Heim on Feb 13, 2002
static INT_PTR fm_bTriangle(INT_PTR x, INT_PTR y, INT_PTR centerx, INT_PTR centery, INT_PTR radius)
{
    INT_PTR calc = y * fmc.X / fmc.Y / 2;
    INT_PTR topy = centery - radius;

    return  x - centerx + calc >= topy && x - centerx - calc <= -topy && y - centery < radius  ? 1 : 0;
    
}


//function added by Harald Heim on Apr 6, 2002
static INT_PTR fm_refreshWindow (void)
{
    ////return UpdateWindow(fmc.hDlg);
    //InvalidateRect(gParams->ctl[n].hCtl, NULL, TRUE );
    //return UpdateWindow(gParams->ctl[n].hCtl);

    //InvalidateRect(fmc.hDlg, NULL, TRUE );
    //return UpdateWindow(fmc.hDlg);

    return RedrawWindow(
            fmc.hDlg,   // handle of entire dialog window
            NULL,   // address of structure with update rectangle, NULL=>entire client area
            NULL,   // handle of update region
            RDW_ERASE|RDW_INVALIDATE|RDW_ERASENOW|RDW_UPDATENOW     // array of redraw flags
       );

}

//function added by Harald Heim on Apr 6, 2002
static INT_PTR fm_refreshRgn (INT_PTR Rgn)
{

    if (Rgn == 0) return 0;

    return RedrawWindow(
            fmc.hDlg,   // handle of entire dialog window
            NULL,   // address of structure with update rectangle, NULL=>entire client area
            (HRGN)Rgn,  // handle of update region
            RDW_ERASE|RDW_INVALIDATE|RDW_ERASENOW|RDW_UPDATENOW     // array of redraw flags
       );

}


static INT_PTR fm_setCtlOrder (INT_PTR n, INT_PTR order){

    HWND Control;
    
    if (!IS_OK_CTL_NR(n)) return false;
    
    //HWND_TOP        0
    //HWND_BOTTOM     1
    //HWND_TOPMOST    -1
    //HWND_NOTOPMOST  -2
    if (order < -2 || order > 1) return false;
    
    if (n == CTL_PREVIEW) Control = GetDlgItem(fmc.hDlg, 101);
    else if (n == CTL_FRAME) Control = GetDlgItem(fmc.hDlg, IDC_PROXY_GROUPBOX);
    else if (n == CTL_PROGRESS) Control = GetDlgItem(fmc.hDlg, IDC_PROGRESS1);
    else Control = gParams->ctl[n].hCtl;
    

    SetWindowPos(Control, (HWND)order, 0, 0, 0, 0, SWP_FRAMECHANGED|SWP_NOACTIVATE|SWP_NOSIZE|SWP_NOMOVE);
    
    InvalidateRect(gParams->ctl[n].hCtl, NULL, TRUE );
	UpdateWindow(gParams->ctl[n].hCtl);
    
    return true;
}


//function added by Harald Heim on Apr 6, 2002
static INT_PTR fm_refreshCtl (INT_PTR n)
{
    if (!IS_OK_CTL_NR(n)) return false;

	if (gParams->ctl[n].class == CC_TAB){
		int i;

		InvalidateRect(gParams->ctl[n].hCtl, NULL, TRUE );
		UpdateWindow(gParams->ctl[n].hCtl);

		for (i = 0; i < N_CTLS; ++i) {
			if (gParams->ctl[i].inuse && gParams->ctl[i].tabControl == n &&  gParams->ctl[i].tabSheet == gParams->ctl[n].val){
				InvalidateRect(gParams->ctl[i].hCtl, NULL, TRUE );
				UpdateWindow(gParams->ctl[i].hCtl);
				if (gParams->ctl[i].class == CC_STANDARD || gParams->ctl[i].class == CC_SLIDER){
					InvalidateRect(gParams->ctl[i].hBuddy1, NULL, TRUE );
					UpdateWindow(gParams->ctl[i].hBuddy1);
					SetWindowPos(gParams->ctl[i].hBuddy1, NULL, 0, 0, 0, 0,
                        SWP_FRAMECHANGED|SWP_NOZORDER|SWP_NOACTIVATE|SWP_NOSIZE|SWP_NOMOVE);
					
					InvalidateRect(gParams->ctl[i].hBuddy2, NULL, TRUE );
					UpdateWindow(gParams->ctl[i].hBuddy2);
				}
			}	
		}

	} else {
		////return UpdateWindow(fmc.hDlg);
        HWND Control;

        if (gParams->ctl[n].class == CC_STANDARD || gParams->ctl[n].class == CC_SLIDER){
				InvalidateRect(gParams->ctl[n].hBuddy1, NULL, TRUE );
				UpdateWindow(gParams->ctl[n].hBuddy1);
				SetWindowPos(gParams->ctl[n].hBuddy1, NULL, 0, 0, 0, 0,
                        SWP_FRAMECHANGED|SWP_NOZORDER|SWP_NOACTIVATE|SWP_NOSIZE|SWP_NOMOVE);

				InvalidateRect(gParams->ctl[n].hBuddy2, NULL, TRUE );
				UpdateWindow(gParams->ctl[n].hBuddy2);
		}
		
        if (n == CTL_PROGRESS) Control = GetDlgItem(fmc.hDlg, IDC_PROGRESS1);
        else Control = gParams->ctl[n].hCtl;

        InvalidateRect(Control, NULL, TRUE );
		return UpdateWindow(Control);
		
	}

	return true;

}

/*
//function added by Harald Heim on Apr 6, 2002
static int fm_lockCtl (int n, int c)
{

    if (c==1)
        return LockWindowUpdate(gParams->ctl[n].hCtl);
    else if (c==0)
        return LockWindowUpdate(NULL);

}
*/


//function added by Harald Heim on Apr 6, 2002
static INT_PTR fm_lockWindow (INT_PTR c)
{

    if (c==1)
        return LockWindowUpdate(fmc.hDlg);
    else if (c==0)
        return LockWindowUpdate(NULL);
	return -1;	//usage error -- afh 01-may-2008
}


//function added by Harald Heim on Apr 2, 2002
static INT_PTR fm_setZoom (INT_PTR n)
{
    if (!fmc.doingProxy) return 0;
    
    if (n < 1 && n !=- 888) n = 1;
    else if (n > MAX_SCALE_FACTOR) n = MAX_SCALE_FACTOR;

    //function added by Harald Heim on Jun 13, 2002
    fmc.scaleFactor = (int)n;
    
    return setZoom(fmc.hDlg, (int)n, 1);
}

static INT_PTR fm_setZoomEx (INT_PTR n, INT_PTR m)
{
#ifdef HIGHZOOM
    if (!fmc.doingProxy) return 0;
    
    if (n<1 && n!=-888) n=1;
    else if (n>16) n=16;

	if (n>1) m=1; //No enlargement for if < 100%

	if (m<1) m=1;
    else if (m>16) m=16;

    //function added by Harald Heim on Jun 13, 2002
    fmc.scaleFactor = (int)n;
	//fmc.enlargeFactor = (int)m; //is set in AfhFMUI
    
    return setZoom(fmc.hDlg, (int)n, (int)m);
#else

	return 0;

#endif
}



static INT_PTR fm_set_array_mode(INT_PTR mode)
{

	if (mode == 1){	
		if (gStuff->bufferProcs && 
			gStuff->bufferProcs->allocateProc &&
			gStuff->bufferProcs->lockProc &&
			gStuff->bufferProcs->unlockProc &&
			gStuff->bufferProcs->freeProc){
				arrayMode = 1;
				return true;
		} else {
			arrayMode = 0;
			return false;
		}
	} else {
		arrayMode = 0;
		return true;
	}
	
	return false;
}



static INT_PTR fm_allocHost (INT_PTR size)
{
    BufferID bufferid;

    if (AllocateBuffer((int32)size, &bufferid) == noErr){   
        return (INT_PTR)bufferid;	//X64??
    } else {
        return (INT_PTR)NULL;
    }
}

static INT_PTR fm_lockHost (INT_PTR bufferid)
{

    return (INT_PTR)LockBuffer((BufferID)bufferid, true);	//X64??
}


static INT_PTR fm_freeHost (INT_PTR bufferid)
{
    UnlockBuffer((BufferID)bufferid);	//X64??
	FreeBuffer((BufferID)bufferid);

    return true;
}



//function added by Harald Heim on Apr 9, 2002
static INT_PTR allocArray (INT_PTR nr, INT_PTR X, INT_PTR Y, INT_PTR Z,
						   INT_PTR bytes, INT_PTR padding)
{

    //BufferProcs* bufferProcs = gStuff->bufferProcs;
    INT_PTR size;

    if (!IS_OK_ARRAY_NR(nr)) return 0; //invalid Array index

    if (bytes==1 || bytes==2 || bytes==4 || bytes==8) ArrayBytes[nr]=(int)bytes;
    else return 0;

    if (X <= 0) return 0;
    ArrayX[nr]=X;
    if (Y <= 0) ArrayY[nr]=1; else ArrayY[nr]=Y;
    if (Z <= 0) ArrayZ[nr]=1; else ArrayZ[nr]=Z;

    ArrayX[nr] += 2*padding;
    ArrayY[nr] += 2*padding;    // ACK!!?? Could easily triple the size of a 1-dim array, or worse!!?? Harry:_ But who would use padding for a 1-dim array???
    ArrayPadding[nr] = padding;

    size = ArrayX[nr] * ArrayY[nr] * ArrayZ[nr] * ArrayBytes[nr];
    if (size == 0) return 0;

    //Avoid reallocation of the same size
    if (ArraySize[nr] == size && Array[nr]!=NULL) return 1;
	
	//Not necessary
	//Free array if it already exists and has a different size
	//if (ArraySize[nr] != size && Array[nr]!=NULL) fm_freeArray(nr);

    //Info ("%d", bufferProcs->spaceProc() );


	if (arrayMode == 1 && size <= INT_MAX){ //Host AllocateBuffer() can't handle size >= 2GB
										    //so revert to malloc/remalloc unless size < 2GB 
		//Free previous buffer
		if (ArrayID[nr]!=0) { 
			UnlockBuffer(ArrayID[nr]); //gStuff->bufferProcs->unlockProc(ArrayID[nr]);
			FreeBuffer(ArrayID[nr]); //gStuff->bufferProcs->freeProc(ArrayID[nr]);
		} else if (Array[nr]!=0){
			free (Array[nr]);
		}
		
		ArrayID[nr] = 0;
		Array[nr] = 0;
		
		if (AllocateBuffer((int32)size, &ArrayID[nr]) == noErr){ //if (gStuff->bufferProcs->allocateProc(size, &ArrayID[nr]) == noErr){
			Array[nr] = LockBuffer(ArrayID[nr], true);//Array[nr] = gStuff->bufferProcs->lockProc(ArrayID[nr], true);
		} else {
			Array[nr] = realloc (Array[nr], (size_t)size); //Fall back to Windows API
                                                   //(realloc could be malloc, since Array[nr]==0 here?-AFH)
            ArrayID[nr] = 0; //for safety?-afh
		}

	}else {
        //but what if ArrayID[nr] != 0 ?? Need to free it??-AFH	
		if (ArrayID[nr]!=0) { 
			UnlockBuffer(ArrayID[nr]); 
			FreeBuffer(ArrayID[nr]);
            ArrayID[nr] = 0;
		} 
		Array[nr] = realloc (Array[nr], (size_t)size);
	}
    

    if (Array[nr]==NULL)
    {
        ArrayID[nr]=0;
		ArraySize[nr]=0;
        ArrayBytes[nr]=0;
        ArrayX[nr]=0;
        ArrayY[nr]=0;
        ArrayZ[nr]=0;
        ArrayPadding[nr]=0;
        return 0;
    }
    else
    {   
        ArraySize[nr]=size;
        return 1;
    }

}


//function added by Harald Heim on Apr 9, 2002
static INT_PTR fm_allocArray (INT_PTR nr, INT_PTR X, INT_PTR Y, INT_PTR Z, INT_PTR bytes)
{
    
    return allocArray (nr, X, Y, Z, bytes, 0);
    
}

//function added by Harald Heim on Oct 12, 2003
static INT_PTR fm_allocArrayPad (INT_PTR nr, INT_PTR X, INT_PTR Y, INT_PTR Z, INT_PTR bytes, INT_PTR padding)
{

    return allocArray (nr, X, Y, Z, bytes, padding);

}

static INT_PTR fm_freeArray (INT_PTR nr)
{

    if (IS_ALLOCATED_ARRAY(nr))
    {

        if (ArrayID[nr]!=0){ //was allocated with Buffer Proc
			UnlockBuffer(ArrayID[nr]);//gStuff->bufferProcs->unlockProc(ArrayID[nr]);
			FreeBuffer(ArrayID[nr]);//gStuff->bufferProcs->freeProc(ArrayID[nr]); 
        } else {
			free (Array[nr]);
        }
        
        Array[nr] = NULL;
		ArrayID[nr]=0;
        ArraySize[nr]=0;
        ArrayBytes[nr]=0;
        ArrayX[nr]=0;
        ArrayY[nr]=0;
        ArrayZ[nr]=0;
        ArrayPadding[nr]=0;
		
        return 1;
    }

    return 0;

}




///////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2002, Industrial Light & Magic, a division of Lucas
// Digital Ltd. LLC
// 
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
// *       Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
// *       Redistributions in binary form must reproduce the above
// copyright notice, this list of conditions and the following disclaimer
// in the documentation and/or other materials provided with the
// distribution.
// *       Neither the name of Industrial Light & Magic nor the names of
// its contributors may be used to endorse or promote products derived
// from this software without specific prior written permission. 
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
///////////////////////////////////////////////////////////////////////////


__inline unsigned short floatToHalf (unsigned int i)
{
    // Our floating point number, f, is represented by the bit
    // pattern in integer i.  Disassemble that bit pattern into
    // the sign, s, the exponent, e, and the significand, m.
    // Shift s into the position where it will go in the
    // resulting half number.
    // Adjust e, accounting for the different exponent bias
    // of float and half (127 versus 15).

    register int s =  (i >> 16) & 0x00008000;
    register int e = ((i >> 23) & 0x000000ff) - (127 - 15);
    register int m =   i        & 0x007fffff;

    // Now reassemble s, e and m into a half:
    if (e <= 0){
		if (e < -10){
			// E is less than -10.  The absolute value of f is
			// less than HALF_MIN (f may be a small normalized
			// float, a denormalized float or a zero).
			//
			// We convert f to a half zero.

			return 0;
		}

		// E is between -10 and 0.  F is a normalized float,
		// whose magnitude is less than HALF_NRM_MIN.
		//
		// We convert f to a denormalized half. 

		m = (m | 0x00800000) >> (1 - e);

		// Round to nearest, round "0.5" up.
		//
		// Rounding may cause the significand to overflow and make
		// our number normalized.  Because of the way a half's bits
		// are laid out, we don't have to treat this case separately;
		// the code below will handle it correctly.

		if (m &  0x00001000)
			m += 0x00002000;

		// Assemble the half from s, e (zero) and m.

		return s | (m >> 13);

    } else if (e == 0xff - (127 - 15)) {
		if (m == 0){
			// F is an infinity; convert f to a half
			// infinity with the same sign as f.
			return s | 0x7c00;

		} else {
			// F is a NAN; we produce a half NAN that preserves
			// the sign bit and the 10 leftmost bits of the
			// significand of f, with one exception: If the 10
			// leftmost bits are all zero, the NAN would turn 
			// into an infinity, so we have to set at least one
			// bit in the significand.
			
			m >>= 13;
			return s | 0x7c00 | m | (m == 0);
		}

    } else {

		// E is greater than zero.  F is a normalized float.
		// We try to convert f to a normalized half.
		
		// Round to nearest, round "0.5" up
		
		if (m &  0x00001000) {

			m += 0x00002000;

			if (m & 0x00800000) {
				m =  0;		// overflow in significand,
				e += 1;		// adjust exponent
			}
		}

		// Handle exponent overflow

		if (e > 30) {
			// overflow ();	// Cause a hardware floating point overflow;
			return s | 0x7c00;	// if this returns, the half becomes an
		}   			// infinity with the same sign as f.

		// Assemble the half from s, e and m.
		return s | (e << 10) | (m >> 13);
    }
} 

__inline unsigned int halfToFloat (unsigned short y)
{

    int s = (y >> 15) & 0x00000001;
    int e = (y >> 10) & 0x0000001f;
    int m =  y        & 0x000003ff;

    if (e == 0) {
		if (m == 0) {
			// Plus or minus zero
			return s << 31;
		} else {
			// Denormalized number -- renormalize it
			while (!(m & 0x00000400)){
				m <<= 1;
				e -=  1;
			}
			e += 1;
			m &= ~0x00000400;
		}
    } else if (e == 31) {
		if (m == 0){
			// Positive or negative infinity
			return (s << 31) | 0x7f800000;
		} else {
			// Nan -- preserve sign and significand bits
			return (s << 31) | 0x7f800000 | (m << 13);
		}
    }

    // Normalized number
    e = e + (127 - 15);
    m = m << 13;

    // Assemble s, e and m.
    return (s << 31) | (e << 23) | m;
}




static INT_PTR fm_getArray (INT_PTR nr, INT_PTR x, INT_PTR y, INT_PTR z)
{

    INT_PTR pos;

    if (!IS_ALLOCATED_ARRAY(nr) || z<0 || z >= ArrayZ[nr] 
#ifndef _WIN64
		|| ArrayBytes[nr] == 8	//8-byte integer arrays are not implemented in Win32
#endif
		) return 0;

    // No check on x and y within bounds???? -- allows wrapping of bounds???
    // Harry: Isn't that done by the edgeMode stuff???
        
    switch (edgeMode)
        {
            case 0:
                // Replicate pixels at edge
                if (x < 0-ArrayPadding[nr]) x = 0- ArrayPadding[nr];
                else if (x >= ArrayX[nr]-ArrayPadding[nr]) x = ArrayX[nr] - 1-ArrayPadding[nr];
                if (y < 0-ArrayPadding[nr]) y = 0-ArrayPadding[nr];
                else if (y >= ArrayY[nr]-ArrayPadding[nr]) y = ArrayY[nr] - 1-ArrayPadding[nr];
                break;

            case 1:
                // Black (zero) padding
                if (x < 0-ArrayPadding[nr]|| y < 0-ArrayPadding[nr]|| x >= ArrayX[nr]-ArrayPadding[nr]|| y >= ArrayY[nr]-ArrayPadding[nr])
                    return 0;
				//break;			//needed, or is fall-thru intentional here??
        
            case 2:
                // Wrap around
                if (x < 0) x = ArrayX[nr]-ArrayPadding[nr] + x;   /// doesn't work for x < -2X ?? use MOD ???
                else if (x >= ArrayX[nr]-ArrayPadding[nr]) x = x - ArrayX[nr]-ArrayPadding[nr];
                if (y < 0) y = ArrayY[nr]-ArrayPadding[nr]  + y;
                else if (y >= ArrayY[nr]-ArrayPadding[nr]) y = y - ArrayY[nr]-ArrayPadding[nr];
                break;
            
            case 3: 
                // Two possible reflection modes, one of period 2N, other of period 2N-2 (as here)?
                if (x < 0) x = - x;
                else if (x >= ArrayX[nr]-ArrayPadding[nr]) x = (ArrayX[nr] - 1-ArrayPadding[nr]) - (x - (ArrayX[nr] - 1-ArrayPadding[nr]));       //optimize?
                if (y < 0) y = - y;
                else if (y >= ArrayY[nr]-ArrayPadding[nr]) y = (ArrayY[nr] - 1-ArrayPadding[nr]) - (y - (ArrayY[nr] - 1-ArrayPadding[nr]));
                break;
        }//end switch


    ///pos = y*ArrayX[nr]*ArrayZ[nr]*ArrayBytes[nr] + x*ArrayZ[nr]*ArrayBytes[nr] + z*ArrayBytes[nr];
    pos = (( (y+ArrayPadding[nr])*ArrayX[nr] + (x+ArrayPadding[nr]) )*ArrayZ[nr] + z)*ArrayBytes[nr];
    
    //if (pos <0 || pos >= ArraySize[nr]) return 0; // not necessary anymore because of edge mode **WRONG** unless mod used!!!


    // Following could be simplified if pos did not contain the ArrayBytes[nr] factor above!!
    if (ArrayBytes[nr] == 1)
        return ((unsigned8 *)Array[nr]) [pos];		//return zero-extended (UNSIGNED) BYTE

    else if (ArrayBytes[nr] == 2)
        //return ((unsigned16 *)Array[nr]) [pos];	//if pos doesn't include the ArrayBytes[nr] factor
        return  *(unsigned16 *)(((unsigned8 *)Array[nr]) + pos);	//return zero-extended unsigned16

    else if (ArrayBytes[nr] == 4)
        //return ((unsigned32 *)Array[nr]) [pos];	//if pos doesn't include the ArrayBytes[nr] factor
        return  *(unsigned32 *)(((unsigned8 *)Array[nr]) + pos);	//return zero-extended unsigned32

#ifdef _WIN64
    else if (ArrayBytes[nr] == 8)
        //return ((UINT64 *)Array[nr]) [pos];	//if pos doesn't include the ArrayBytes[nr] factor
        return  *(UINT64 *)(((unsigned8 *)Array[nr]) + pos);	//return unsigned int64
#endif

    else
        return 0; // or better yet, return -1 to indicate something really wrong???

}


//static int UNSAFE_get_pos[N_ARRAYS];

static INT_PTR fm_UNSAFE_getArray (INT_PTR nr, INT_PTR x, INT_PTR y, INT_PTR z)
{
    const INT_PTR pos = (( (y+ArrayPadding[nr])*ArrayX[nr] + (x+ArrayPadding[nr]) )*ArrayZ[nr] + z)*ArrayBytes[nr];
    //UNSAFE_get_pos[nr] = (( (y+ArrayPadding[nr])*ArrayX[nr] + (x+ArrayPadding[nr]) )*ArrayZ[nr] + z)*ArrayBytes[nr];

	switch(ArrayBytes[nr]) {
		case 1:	//return zero-extended (UNSIGNED) BYTE
	        return ((unsigned8 *)Array[nr]) [pos];
		case 2:	//return zero-extended unsigned16
	        return  *(unsigned16 *)(((unsigned8 *)Array[nr]) + pos);
		case 4:	//return zero-extended unsigned32
			return  *(unsigned32 *)(((unsigned8 *)Array[nr]) + pos);
#ifdef _WIN64
		case 8:	//return unsigned int64
			return  *(UINT64 *)(((unsigned8 *)Array[nr]) + pos);
#endif
	} //switch

	return 0;	//illegal # of bytes in ArrayBytes[nr]
}

/**
static INT_PTR fm_UNSAFE_getArrayNext8 (INT_PTR nr)
{  
#if 1	//should PRE-increment UNSAFE_get_pos[nr], not POST increment???
    return ((unsigned8 *)Array[nr]) [++UNSAFE_get_pos[nr]];
#else
    return ((unsigned8 *)Array[nr]) [UNSAFE_get_pos[nr]++];
#endif
}
static INT_PTR fm_UNSAFE_getArrayNext16 (INT_PTR nr)
{
    UNSAFE_get_pos[nr] += 2;
	//return zero-extended unsigned16
    return  *(unsigned16 *)(((unsigned8 *)Array[nr]) + UNSAFE_get_pos[nr]);
}
static INT_PTR fm_UNSAFE_getArrayNext32 (INT_PTR nr)
{  
    UNSAFE_get_pos[nr] += 4;
	//return zero-extended unsigned32
    return  *(unsigned32 *)(((unsigned8 *)Array[nr]) + UNSAFE_get_pos[nr]);
}
#ifdef _WIN64
static INT_PTR fm_UNSAFE_getArrayNext64 (INT_PTR nr)
{  
    UNSAFE_get_pos[nr] += 8;
	//return unsigned int64
    return  *(UINT64 *)(((unsigned8 *)Array[nr]) + UNSAFE_get_pos[nr]);
}
#endif //Win64
/**/


static double fm_fgetArray (INT_PTR nr, INT_PTR x, INT_PTR y, INT_PTR z)
{

    INT_PTR pos;

    if (!IS_ALLOCATED_ARRAY(nr) || z<0 || z >= ArrayZ[nr] || ArrayBytes[nr] < 2) return 0.0;

    // No edge effects as for fm_getArray above???


    pos = (( (y+ArrayPadding[nr])*ArrayX[nr] + (x+ArrayPadding[nr]) )*ArrayZ[nr] + z)*ArrayBytes[nr];
    if (pos <0 || pos >= ArraySize[nr]) return 0.0;

	if (ArrayBytes[nr] == 2){
		UIF fval;
		fval.i = halfToFloat( *(unsigned short *)(((unsigned8 *)Array[nr]) + pos) );
		return (double)fval.f;

    } else if (ArrayBytes[nr] == 4)
		return *(float *)(((unsigned8 *)Array[nr]) + pos);
	else if (ArrayBytes[nr] == 8)
		return *(double *)(((unsigned8 *)Array[nr]) + pos);
    else
        return 0.0; // or return some other value such as Indefinite Nan???
}


static double fm_UNSAFE_fgetArray (INT_PTR nr, INT_PTR x, INT_PTR y, INT_PTR z)
{
    const INT_PTR pos = (((y + ArrayPadding[nr]) * ArrayX[nr] + (x + ArrayPadding[nr])) * ArrayZ[nr] + z) * ArrayBytes[nr];

	switch(ArrayBytes[nr]) {
		case 4:
			return *(float *)(((unsigned8 *)Array[nr]) + pos);
		case 8:
			return *(double *)(((unsigned8 *)Array[nr]) + pos);
		case 2: {
			UIF fval;
			fval.i = halfToFloat( *(unsigned short *)(((unsigned8 *)Array[nr]) + pos) );
			return (double)fval.f;
		}
	}

    return 0.0;
}


//function added by Harald Heim on Apr 9, 2002
static INT_PTR fm_putArray (INT_PTR nr, INT_PTR x, INT_PTR y, INT_PTR z, INT_PTR val)
{

    INT_PTR pos;

    if (!IS_ALLOCATED_ARRAY(nr)
#ifndef _WIN64
		|| ArrayBytes[nr] == 8	//8-byte integer arrays are not implemented in Win32
#endif
		) return 0;
    //if ( x<0 || y<0 || z<0 || x >= ArrayX[nr] || y >= ArrayY[nr] || z >= ArrayZ[nr]  ) return 0;
    
    // clamp val to correct range...	//Do this within the if/else... below???
    if (ArrayBytes[nr]==1){
        if (val < 0) val = 0;
        else if (val > 255 ) val = 255;
    }
	else if (ArrayBytes[nr]==2){
        if (val < 0) val = 0;
        else if (val > 65535 ) val = 65535;
    }
#ifdef _WIN64
	else if (ArrayBytes[nr]==4){
        if (val < 0) val = 0;
        else if (val > UINT_MAX ) val = UINT_MAX;
    }
#endif

    //pos = y*ArrayX[nr]*ArrayZ[nr]*ArrayBytes[nr] + x*ArrayZ[nr]*ArrayBytes[nr] + z*ArrayBytes[nr];
    pos = (( (y+ArrayPadding[nr])*ArrayX[nr] + (x+ArrayPadding[nr]) )*ArrayZ[nr] + z)*ArrayBytes[nr];
    if (pos <0 || pos >= ArraySize[nr]) return 0;

    //no edge-wrap modes???


    if (ArrayBytes[nr] == 1)
        ((unsigned8 *)Array[nr]) [pos] = (unsigned8)val;

    else if (ArrayBytes[nr] == 2)
        //((unsigned16 *)) [pos] = val;
        *(unsigned16 *)(((unsigned8 *) Array[nr]) + pos) = (unsigned16)val;

    else if (ArrayBytes[nr] == 4)
        //((int *)Array[nr]) [pos] = val; 
        *(unsigned32 *)(((unsigned8 *) Array[nr]) + pos) = (unsigned32)val;
#ifdef _WIN64
    else if (ArrayBytes[nr] == 8)
        //((int *)Array[nr]) [pos] = val; 
        *(UINT64 *)(((unsigned8 *) Array[nr]) + pos) = val;
#endif
    else
        return 0; // return FALSE
    
    return 1;	//return TRUE

}

static INT_PTR fm_UNSAFE_putArray (INT_PTR nr, INT_PTR x, INT_PTR y, INT_PTR z, INT_PTR val)
{
    const INT_PTR pos = (( (y+ArrayPadding[nr])*ArrayX[nr] + (x+ArrayPadding[nr]) )*ArrayZ[nr] + z)*ArrayBytes[nr];

	switch(ArrayBytes[nr]) {
		case 1:
	        ((unsigned8 *)Array[nr]) [pos] = (unsigned8)val;
			return 1;
		case 2:
	        *(unsigned16 *)(((unsigned8 *) Array[nr]) + pos) = (unsigned16)val;
			return 1;
		case 4:
	        *(unsigned32 *)(((unsigned8 *) Array[nr]) + pos) = (unsigned32)val;
			return 1;
#ifdef _WIN64
		case 8:
			*(UINT64 *)(((unsigned8 *) Array[nr]) + pos) = val;
#endif
		default:
			return 0;
	}

	return 1;
}

static INT_PTR fm_fputArray (INT_PTR nr, INT_PTR x, INT_PTR y, INT_PTR z, double val)
{

    INT_PTR pos;

    if (!IS_ALLOCATED_ARRAY(nr) || ArrayBytes[nr] < 2) return 0;
    
    pos = (( (y+ArrayPadding[nr])*ArrayX[nr] + (x+ArrayPadding[nr]) )*ArrayZ[nr] + z)*ArrayBytes[nr];
    if (pos <0 || pos >= ArraySize[nr]) return 0;

	//if (val> 3.4E+38) val=3.4E+38; else if (val< 3.4E-38) val=3.4E-38;
    
	if (ArrayBytes[nr] == 2){
		UIF fval;
		fval.f = (float)val;
		*(unsigned short *)(((unsigned8 *) Array[nr]) + pos ) = floatToHalf( fval.i );
    } else if (ArrayBytes[nr] == 4)
		*(float *)(((unsigned8 *) Array[nr]) + pos ) = (float)val;
	else if (ArrayBytes[nr] == 8)
		*(double *)(((unsigned8 *) Array[nr]) + pos ) = val;
    else 
        return 0; //bad value for ArrayBytes[nr].
	
    return 1;
}

static INT_PTR fm_UNSAFE_fputArray (INT_PTR nr, INT_PTR x, INT_PTR y, INT_PTR z, double val)
{
    const INT_PTR pos = (( (y+ArrayPadding[nr])*ArrayX[nr] + (x+ArrayPadding[nr]) )*ArrayZ[nr] + z)*ArrayBytes[nr];
    
	switch (ArrayBytes[nr]) {
		case 4:
			*(float *)(((unsigned8 *) Array[nr]) + pos ) = (float)val;
			return 1;
		case 8:
			*(double *)(((unsigned8 *) Array[nr]) + pos ) = val;
			return 1;
		case 2: {
				UIF fval;
				fval.f = (float)val;
				*(unsigned short *)(((unsigned8 *) Array[nr]) + pos ) = floatToHalf( fval.i );
			}
			return 1;
		default:
			return 0;
	}
	
    return 1;
}



/**
__inline unsigned short float2le (double val)
{
  double v;
  int e;

  v = val;
  if (v < 1e-32) {
    return 0;
  }
  else {
    v = frexp(v,&e) * 256.0/v;
	return (unsigned char)(val * v)<<8 | (unsigned char)(e + 128);
  }
} 

__inline double le2float(unsigned short le)
{
  double f;
  unsigned char le1 = le >> 8;
  unsigned char le2 = le - (le1 << 8);	//afh--> le2 = le & 0xff; ???

  if (le1) {   //nonzero pixel
    f = ldexp(1.0,le2-(int)(128+8));
	return le1 * f;
  }
  else
    return 0;
} 


static INT_PTR fm_fputArrayLE (INT_PTR nr, INT_PTR x, INT_PTR y, INT_PTR z, double val)
{
    INT_PTR pos;

    if (!IS_ALLOCATED_ARRAY(nr) || ArrayBytes[nr] < 2) return 0;
    
    pos = (( (y+ArrayPadding[nr])*ArrayX[nr] + (x+ArrayPadding[nr]) )*ArrayZ[nr] + z)*ArrayBytes[nr];
    if (pos <0 || pos >= ArraySize[nr]) return 0;

	*(unsigned short *)(((unsigned8 *) Array[nr]) + pos ) = float2le (val);
	
    return 1;

}

static double fm_fgetArrayLE (INT_PTR nr, INT_PTR x, INT_PTR y, INT_PTR z)
{
    INT_PTR pos;

    if (!IS_ALLOCATED_ARRAY(nr) || z<0 || z >= ArrayZ[nr] || ArrayBytes[nr] < 2) return 0;

    pos = (( (y+ArrayPadding[nr])*ArrayX[nr] + (x+ArrayPadding[nr]) )*ArrayZ[nr] + z)*ArrayBytes[nr];
    if (pos <0 || pos >= ArraySize[nr]) return 0;

	return le2float ( *(unsigned short *)(((unsigned8 *)Array[nr]) + pos) );

}
/**/

// Usage of putArrayString(nr, i, str) and getArrayString(nr, i) appears
// to be:
//
// (1) Allocate an array with X-dim = #of strings to store, and Y-dim =
//     max length of a string (including the final NUL byte).  For example,
//     using Array #99 to store up to 50 strings of 255 chars each:
//
//		allocArray(99, 50, 256, 0, 1); /* 1 means the array stores bytes, */
//									   /* 256 is max string lenth including */
//									   /* the terminating NUL byte */
//
// (2) To put a string into Array 99 at location n ( 0 <= n < 50):
//
//		putArrayString (99, n, "Test string");
//
// (3) To retrieve a string from location n and copy it into a
//     global string (str0):
//
//		strcpy (str0, getArrayString(99, n)); /* str0 should be large enough to */
//											  /* hold a string of a least 255 chars */
//											  /* plus the final NUL byte */
//
// (4) To access individual characters of a string in an Array, use the y-coordinate
//	   to select which string to access, and use the x-coordinate to select the char
//     within the string (0-based indexing).  Using the above example,
//
//		getArray(99, 10, n, 0)
//
//	   will return the 11th char from the string "Test string"; i.e., 'g'. And the
//	   following will change the 4th character of the stored string to 'x'; i.e.,
//     the string stored at index n in Array 99 will be changed to "Text string".
//
//		putArray(99, 3, n, 0, 'x'); /* Changes string n to "Text string" */
//

static INT_PTR fm_putArrayString (INT_PTR nr, INT_PTR x, LPSTR string)
{

    INT_PTR pos;

    if (!IS_ALLOCATED_ARRAY(nr)) return false;
    
    pos = (x+ArrayPadding[nr])*ArrayY[nr];	//here roles of x and y are reversed from the convention elsewhere???
    if (pos <0 || pos + (INT_PTR)strlen(string) >= ArraySize[nr]) return false; //strlen+1 for NUL byte???-Alex //may overwrite array by 1 byte???

    strcpy ((char *)((unsigned8 *)Array[nr] + pos), string);

    return true;

}


static INT_PTR fm_getArrayString (INT_PTR nr, INT_PTR x)//, LPSTR string)
{

    INT_PTR pos;

    static const char dummystr[] = "Not Available"; //Return this string to avoid memory access violations

    if (!IS_ALLOCATED_ARRAY(nr)) return (INT_PTR)(unsigned8 *)dummystr;
    
    pos = (x+ArrayPadding[nr])*ArrayY[nr];
    if (pos <0 || pos >= ArraySize[nr]) return (INT_PTR)(unsigned8 *)dummystr;
		//should test that 1st NUL at pos or higher is within Array bounds???

    //strcpy (string, (char *)((unsigned8 *)Array[nr] + pos));

    return (INT_PTR)((unsigned8 *)Array[nr] + pos);//true;

}




//function added by Harald Heim on Jun 4, 2002
static unsigned8 gMskVal(int row, int rowOff, int col, int colOff)
{

    if ((gNoMask || !(gParams->flags & AFH_USE_HOST_PADDING)) &&
        (row+rowOff < minRow || row+rowOff > maxRow ||
         col+colOff < minCol || col+colOff > maxCol) )
        return 0;   // Simulate the padding...
    else if (gNoMask)
        return 255; // Entire filterRect is inside the mask.
    else
        return fmc.mskPtr[rowOff*fmc.mskRowDelta + colOff*fmc.mskColDelta];
} /*gMskVal*/



#define maxBlurRadius 3
#define mskVal(row,rowOff,col,colOff) \
    gMskVal(row,rowOff,col,colOff)


// "Distance to mask" array of vectors.
#define DIST(i,j) dist[(i)*(fmc.columns+2*maxBlurRadius) + (j)]



//function added by Harald Heim on Jun 4, 2002
static INT_PTR fm_calcSBD(INT_PTR bevWidth1)
{
    int i,j,i0,j0; 

    VRect scaledFilterRect;
    OSErr err = noErr;

    int32 distSize = (fmc.rows + 2*maxBlurRadius)*(fmc.columns + 2*maxBlurRadius)*sizeof(struct Vect_s);

    //Info("distSize: %d",distSize);

    //free buffer if necessary
    if (bufferID != 0)
    {
        UnlockBuffer(bufferID);
        FreeBuffer(bufferID);
        bufferID = 0;
    }

    // Allocate the dist array...
    err = AllocateBuffer(distSize, &bufferID);

    if (err != noErr) {
        // oops.  what now???
        gResult = err;
        return 0;
    }
    dist = (struct Vect_s *)LockBuffer(bufferID, FALSE);
    // Offset to account for non-0 lower bounds...
    dist += maxBlurRadius*(fmc.columns + 2*maxBlurRadius + 1);

#if 0
    copyRect (&scaledFilterRect, &gStuff->filterRect);
    scaleRect (&scaledFilterRect, 1, fmc.scaleFactor);
#else
	fmc.filterRect32 = GetFilterRect(); //necessary??
	scaledFilterRect = fmc.filterRect32;
    scaleRect32(&scaledFilterRect, 1, fmc.scaleFactor);
#endif

    // Compute min/max values for column and row indices that still lie
    // within the (scaled) filterRect (used to simulate padding of accesses
    // outside the bounds of filterRect).
	fmc.outRect32 = GetOutRect();	//needed???
    minCol = scaledFilterRect.left   - fmc.outRect32.left;
    maxCol = scaledFilterRect.right  - fmc.outRect32.left - 1;
    minRow = scaledFilterRect.top    - fmc.outRect32.top;
    maxRow = scaledFilterRect.bottom - fmc.outRect32.top - 1;

    { //scope
    //const int bevWidth = (fm_ctl(7)*2+fmc.scaleFactor)/(fmc.scaleFactor*2);  // round to nearest
	const int bevWidth = (int)bevWidth1;	//coerce to int if necessary
    const int bw = bevWidth;
    const int bwnz = bw ? bw : 1;  // Non-zero, so safe as a divisor...
    const int bw2 = bw*bw;
    const int bw2nz = bwnz*bwnz;   // Non-zero, so safe as a divisor...

    /* set local globals */
    gBevWidth = bevWidth;

        for (i = -maxBlurRadius; i < fmc.rows + maxBlurRadius; i++)
        {
            for (j = -maxBlurRadius; j < fmc.columns + maxBlurRadius; j++)
            {
                DIST(i,j).x = bw + 1;
                DIST(i,j).y = bw + 1;
            } // for j
        } //for i

        /*
        //Method1: TOO SLOW!!!
        // pass 1: find distance (vector) from nearest unmasked point...
        if (!(gParams->flags & AFH_OPTIMIZE_BEVEL))
        {   // method 1: O(n^4) brute force
            for (i=-maxBlurRadius; i < fmc.rows + maxBlurRadius; i++)
            {
                fmc.mskPtr =
                    fmc.mskData + (i * fmc.mskRowDelta) - (maxBlurRadius * fmc.mskColDelta);

                for (j=-maxBlurRadius; j < fmc.columns + maxBlurRadius; j++)
                {
                    if (mskVal(i,0,j,0))
                    {   // inside the mask
                        int d = DIST(i,j).x*DIST(i,j).x + DIST(i,j).y*DIST(i,j).y;

                        for (i0 = -bevWidth; i0 <= bevWidth; i0++)
                        {
                            for (j0 = -bevWidth; j0 <= bevWidth; j0++)
                            {
                                int a0 = mskVal(i,i0,j,j0);
                                int d0 = i0*i0 + j0*j0;
                                if (a0 == 0 && d0 <= bw2)
                                {   // found a point outside the mask
                                    if (d0 < d)
                                    {   // closer than previous...
                                        DIST(i,j).x = j0;
                                        DIST(i,j).y = i0;
                                        d = i0*i0 + j0*j0;
                                    }
                                }
                            } // for j0
                        } // for i0
                    } //if

                    fmc.mskPtr += fmc.mskColDelta;

                } // for j
            } // for i 
        } //bevel method 1
        */
        //else
        {


            // Method 2: O(n^2+) router simulation.
            // (j,i) = center of router bit.
            // For each (j,i) on the outside edge of the mask,
            // sweep out the path of the router blade, recording
            // any new minima to points hit inside the mask.
            // By optimizing the sweep, we can keep the whole
            // thing down to slightly worse than O(n^2).
            int pad = bevWidth + maxBlurRadius;

            for (i=-pad; i < fmc.rows + pad; i++)
            {
                fmc.mskPtr = fmc.mskData + (i * fmc.mskRowDelta) - (pad * fmc.mskColDelta);

                for (j=-pad; j < fmc.columns + pad; j++)
                {
                    if (mskVal(i,0,j,0) == 0)
                    {   // outside the mask...
                        // is it on the edge of the mask?
                        int i1 = -bevWidth;
                        int i2 =  bevWidth;
                        int i3 = -bevWidth;
                        int i4 =  bevWidth;
    #if 0
                        if (mskVal(i, 1, j, 0) ||
                            mskVal(i,-1, j, 0) ||
                            mskVal(i, 0, j, 1) ||
                          //mskVal(i, 1, j, 1) ||
                          //mskVal(i, 1, j,-1) ||
                          //mskVal(i,-1, j, 1) ||
                          //mskVal(i,-1, j,-1) ||
                            mskVal(i, 0, j,-1) )
    #else
                        if (i < -maxBlurRadius)
                            i1 = -i - maxBlurRadius;
                        else if (mskVal(i,-1,j,0))
                            i1 = -bevWidth;  // can tighten this...
                        else
                            i1 = 0;
                        if (i + i1 < -maxBlurRadius)    ////// redundant????????
                            i1 = -i - maxBlurRadius;

                        if (i > fmc.rows + maxBlurRadius - 1)
                            i2 = -i + fmc.rows + maxBlurRadius - 1;
                        else if (mskVal(i,1,j,0))
                            i2 = bevWidth;  // can tighten this...
                        else
                            i2 = 0;
                        if (i + i2 > fmc.rows + maxBlurRadius -1)
                            i2 = -i + fmc.rows + maxBlurRadius - 1;

                        if (j < -maxBlurRadius)
                            i3 = -j - maxBlurRadius;
                        else if (mskVal(i,0,j,-1))
                            i3 = -bevWidth;  // can tighten this...
                        else
                            i3 = 0;
                        if (j + i3 < -maxBlurRadius)
                            i3 = -j - maxBlurRadius;

                        if (j > fmc.columns + maxBlurRadius - 1)
                            i4 = -j + fmc.columns + maxBlurRadius - 1;
                        else if (mskVal(i,0,j,1))
                            i4 = bevWidth;  // can tighten this...
                        else
                            i4 = 0;
                        if (j + i4 > fmc.columns + maxBlurRadius -1)
                            i4 = -j + fmc.columns + maxBlurRadius - 1;

                        //i1 = mskVal(i,-1, j, 0) ? -bevWidth : 0;
                        //i2 = mskVal(i, 1, j, 0) ?  bevWidth : 0;
                        //i3 = mskVal(i, 0, j,-1) ? -bevWidth : 0;
                        //i4 = mskVal(i, 0, j, 1) ?  bevWidth : 0;
    #endif
                        {   // skip altogether if (i1 >= i2) && (i3 >= i4) ???
                            for (i0 = i1; i0 <= i2; i0++)
                            {
                                for (j0 = i3; j0 <= i4; j0++)
                                {
                                    int a0 = mskVal(i,i0,j,j0);
                                    int d0 = i0*i0 + j0*j0;
                                    if (a0 != 0 && d0 <= bw2)
                                    {   // found a point inside the mask
                                        int d = DIST(i+i0,j+j0).x*DIST(i+i0,j+j0).x + 
                                                DIST(i+i0,j+j0).y*DIST(i+i0,j+j0).y;
                                        if (d0 < d)
                                        {   // closer than previous...
                                            DIST(i+i0,j+j0).x = -j0;
                                            DIST(i+i0,j+j0).y = -i0;
                                        }
                                    }
                                } /* for j0 */
                            } /* for i0 */
                        } /*if on edge*/
                    } /*if*/

                    fmc.mskPtr += fmc.mskColDelta;

                } /* for j */
            } /* for i */
        } /*bevel method 2*/

    }//scope

    return 1;

}

static INT_PTR fm_freeSBD(void)
{
    // Release the dist array...
    UnlockBuffer(bufferID);
    FreeBuffer(bufferID);
    bufferID = 0;
    return 1;
}

static INT_PTR fm_getSBD(INT_PTR j0, INT_PTR i0)
{
	int j = (int)j0;
	int i = (int)i0;
    int valx,valy;

	if (j < fmc.x_start || j > fmc.x_end-1) return 0;
    if (i < fmc.y_start || i > fmc.y_end-1) return 0;
    i=i-fmc.y_start;
    j=j-fmc.x_start;
    if (i<0) i=0;
    else if (i> fmc.rows-1) i=fmc.rows-1;
    if (j<0) j=0;
    else if (j> fmc.columns-1) j=fmc.columns-1;
    valx=DIST(i,j).x;
    valy=DIST(i,j).y;

    return valx*valx + valy*valy;

}

static INT_PTR fm_getSBDangle(INT_PTR j0, INT_PTR i0, INT_PTR a10)
{
	int j = (int)j0;
	int i = (int)i0;
    int valx,valy;
    double phi = -((double) a10 / 100.0) * 6.28318530717958647693 / 360.0;

    if (j < fmc.x_start || j > fmc.x_end-1) return 0;
    if (i < fmc.y_start || i > fmc.y_end-1) return 0;
    i=i-fmc.y_start;
    j=j-fmc.x_start;
    if (i<0) i=0;
    else if (i> fmc.rows-1) i=fmc.rows-1;
    if (j<0) j=0;
    else if (j> fmc.columns-1) j=fmc.columns-1;
    valx=DIST(i,j).x;
    valy=DIST(i,j).y;

    //return (int)( (256.0 * 0.63661977236758134308) *  (atan2((double)valy, (double)valx) )   );
    return (INT_PTR)( 512.0 *  cos (atan2((double)valy, (double)valx) - phi)   );

}


static INT_PTR fm_getSBDX(INT_PTR j0, INT_PTR i0)
{
	int j = (int)j0;
	int i = (int)i0;

    if (j < fmc.x_start || j > fmc.x_end-1) return 0;
    if (i < fmc.y_start || i > fmc.y_end-1) return 0;
    i=i-fmc.y_start;
    j=j-fmc.x_start;
    if (i<0) i=0;
    else if (i> fmc.rows-1) i=fmc.rows-1;
    if (j<0) j=0;
    else if (j> fmc.columns-1) j=fmc.columns-1;
    return DIST(i,j).x;
}

static INT_PTR fm_getSBDY(INT_PTR j0, INT_PTR i0)
{
	int j = (int)j0;
	int i = (int)i0;

    if (j < fmc.x_start || j > fmc.x_end-1) return 0;
    if (i < fmc.y_start || i > fmc.y_end-1) return 0;
    i=i-fmc.y_start;
    j=j-fmc.x_start;
    if (i<0) i=0;
    else if (i> fmc.rows-1) i=fmc.rows-1;
    if (j<0) j=0;
    else if (j> fmc.columns-1) j=fmc.columns-1;
    return DIST(i,j).y;
}





static INT_PTR fm_getCtlPos (INT_PTR n, INT_PTR w){			//check valid n, elsewhere too!!!!!

    INT_PTR val=0;
    
    //RECT rcCtl;
    //GetWindowRect( gParams->ctl[n].hCtl, &rcCtl) 
    //ScreenToClient (fmc.hDlg, (LPPOINT)&rcCtl);
    //fm_PixelsToVDBUs 

    if (n == CTL_PREVIEW){

        RECT rcCtl;
    
        GetWindowRect( GetDlgItem(fmc.hDlg, 101), &rcCtl );
		ScreenToClient (fmc.hDlg, (LPPOINT)&rcCtl);
		ScreenToClient (fmc.hDlg, (LPPOINT)&(rcCtl.right));

		switch (w){
			case 0: val = fm_PixelsToHDBUs(rcCtl.left); break;
			case 1: val = fm_PixelsToVDBUs(rcCtl.top); break;
			case 2: val = fm_PixelsToHDBUs(rcCtl.right - rcCtl.left); break;
			case 3: val = fm_PixelsToVDBUs(rcCtl.bottom - rcCtl.top); break;
		}
	
	} else {

		switch (w){
			case 0: val = gParams->ctl[n].xPos; break;
			case 1: val = gParams->ctl[n].yPos; break;
			case 2: val = gParams->ctl[n].width; break;
			case 3: val = gParams->ctl[n].height; break;
		}
	}

    return val;
}


static INT_PTR fm_getCtlCoord (INT_PTR n, INT_PTR w)
{
    RECT rcCtl;
    POINT pt;

    if (n == CTL_PREVIEW){

        int proxW, inW, proxH, inH;
    
        if ( GetWindowRect( GetDlgItem(fmc.hDlg, 101), &rcCtl) )
        {   
            GetCursorPos(&pt);

            if (pt.x >= rcCtl.left && pt.x < rcCtl.right &&
                pt.y >= rcCtl.top  && pt.y < rcCtl.bottom) 
            {
                fmc.inRect32 = GetInRect();	//needed??
                if (w==0){ //X-Coordinate
                    proxW = gProxyRect.right - gProxyRect.left;
                    inW = fmc.inRect32.right - fmc.inRect32.left;
    
                    if (inW < proxW) //Proxy full used
                        return  pt.x - rcCtl.left - (proxW - inW)/2 - 1; 
                    else //Proxy partially used
                        return  fmc.x_start + pt.x - rcCtl.left - 1;

                } else { //Y-Coordinate
                    proxH = gProxyRect.bottom - gProxyRect.top;
                    inH = fmc.inRect32.bottom - fmc.inRect32.top;
                
                    if (inH < proxH)
                        return  pt.y - rcCtl.top - (proxH - inH)/2 - 1; 
                    else
                        return  fmc.y_start + pt.y - rcCtl.top - 1;
                }
                
            }
            else
            {
                return -1;
            }
        }
        return -1; //error from GetWindowRect afh 01-may2008
    
    } else { //Other Controls

        if ( GetWindowRect(gParams->ctl[n].hCtl, &rcCtl) ) //GetWindowRect( GetDlgItem(fmc.hDlg, 101), &rcCtl) )
        {   
            GetCursorPos(&pt);
            if (pt.x >= rcCtl.left && pt.x < rcCtl.right &&
                pt.y >= rcCtl.top  && pt.y < rcCtl.bottom) 
            {
                     if (w==0){ //X-Coordinate
                        return  pt.x - rcCtl.left;// - 1;

                     } else { 
                        return  pt.y - rcCtl.top;// -1;

                     }
            }
            else
            {
                return -1;
            }
            
        }
        return -1;	//error from GetWindowRect afh 01-may2008
    }
}



// function added by Harald Heim on Aug 9, 2002
static INT_PTR fm_getPreviewCoordX (void)
{
    
    //Redirect to new function
    return fm_getCtlCoord (CTL_PREVIEW, 0);
    
    /*RECT rcCtl;
    POINT pt;
    short   proxW, inW;


        if ( GetWindowRect( GetDlgItem(fmc.hDlg, 101), &rcCtl) )
        {   
            GetCursorPos(&pt);

            if (pt.x >= rcCtl.left && pt.x < rcCtl.right &&
                pt.y >= rcCtl.top  && pt.y < rcCtl.bottom) 
            {
                
                proxW = gProxyRect.right - gProxyRect.left;
                inW = gStuff->inRect.right - gStuff->inRect.left;
    
                if (inW < proxW) //Proxy full used
                    return  pt.x - rcCtl.left - (proxW - inW)/2 - 1; 
                else //Proxy partially used
                    return  fmc.x_start + pt.x - rcCtl.left - 1;
                
            }
            else
            {
                return -1;
            }
        }*/
    
}

// function added by Harald Heim on Aug 9, 2002
static INT_PTR fm_getPreviewCoordY (void)
{

    //Redirect to new function
    return fm_getCtlCoord (CTL_PREVIEW, 1);
    
    /*RECT rcCtl;
    POINT pt;
    short   proxH, inH;

        if ( GetWindowRect(GetDlgItem(fmc.hDlg, 101), &rcCtl) )
        {   
            GetCursorPos(&pt);

            if (pt.x >= rcCtl.left && pt.x < rcCtl.right &&
                pt.y >= rcCtl.top  && pt.y < rcCtl.bottom) 
            {
                
                proxH = gProxyRect.bottom - gProxyRect.top;
                inH = gStuff->inRect.bottom - gStuff->inRect.top;
                
                if (inH < proxH)
                    return  pt.y - rcCtl.top - (proxH - inH)/2 - 1; 
                else
                    return  fmc.y_start + pt.y - rcCtl.top - 1;

            }
            else
            {
                return -1;
            }
        }*/

}




// function added by Harald Heim on Aug 12, 2002
static INT_PTR fm_getAsyncKeyState (INT_PTR t)
{
        return GetAsyncKeyState ((int)t);
}


// function added by Harald Heim on Aug 12, 2002
static INT_PTR fm_getAsyncKeyStateF (INT_PTR t)
{
    if (GetForegroundWindow() == MyHDlg)
        return GetAsyncKeyState ((int)t);
    else
        return 0;

}


// function added by Harald Heim on Jun 6, 2002
static INT_PTR fm_mouseOverWhenInvisible (INT_PTR t)
{
    MouseOverWhenInvisible = (int)t;
    
    return 1;
}



static INT_PTR fm_getArrayDim (INT_PTR nr, INT_PTR dim)
{
	
	if (!IS_OK_ARRAY_NR(nr)) return 0;

	if (Array[nr]!=NULL){
		if (dim==0)
			return ArrayX[nr];
		else if (dim==1)
			return ArrayY[nr];
		else if (dim==2)
			return ArrayZ[nr];
#if 1   //extended return values...
        else if (dim==-1)
            return (INT_PTR)Array[nr];
        else if (dim==-2)
            return (INT_PTR)ArrayID[nr];
        else if (dim==-3)
            return ArraySize[nr];
        else if (dim==-4)
            return ArrayBytes[nr];
        else if (dim==-5)
            return ArrayPadding[nr];
#endif
	}
	
	return 0;

}

static INT_PTR fm_copyArray (INT_PTR src, INT_PTR dest)
{

    if (!IS_ALLOCATED_ARRAY(src)) return 0;	// afh 01-may2008
    if (ArraySize[src] == 0)	  return 0;
    if (!IS_OK_ARRAY_NR(dest))	  return 0;
    //Addded by Harald Heim, March 26, 2003
    if (src == dest) return 1;
    

	if (arrayMode == 1){
			
		//Free any previous buffer
		if (ArrayID[dest]!=0) { 
			gStuff->bufferProcs->unlockProc(ArrayID[dest]);
			gStuff->bufferProcs->freeProc(ArrayID[dest]); 
		} else if (Array[dest]!=0){
			free (Array[dest]);
		}
		
		ArrayID[dest] = 0;
		Array[dest] = 0;
		
		if (gStuff->bufferProcs->allocateProc((int32) ArraySize[src], &ArrayID[dest]) == noErr){
			Array[dest] = gStuff->bufferProcs->lockProc(ArrayID[dest], true);
		} else {
			Array[dest] = realloc (Array[dest], (size_t) ArraySize[src]); //Fall back to Windows API
			ArrayID[dest] = 0;	//for safety?-afh
		}

	}else {
        //but what if ArrayID[dest] != 0 ?? Need to free it??-AFH	
		if (ArrayID[dest]!=0) { 
			UnlockBuffer(ArrayID[dest]); 
			FreeBuffer(ArrayID[dest]);
            ArrayID[dest] = 0;
		} 
		Array[dest] = realloc (Array[dest], (size_t) ArraySize[src]);
	}

    

    if (Array[dest]==NULL)
    {
        ArrayID[dest]=0;
		ArraySize[dest]=0;
        ArrayX[dest] = 0;
        ArrayY[dest] = 0;
        ArrayZ[dest] = 0;
        ArrayPadding[dest]=0;
        return 0;
    }
    else
    {   
        memcpy (Array[dest], Array[src], ArraySize[src] );

        ArrayX[dest]		= ArrayX[src];
        ArrayY[dest]		= ArrayY[src];
        ArrayZ[dest]		= ArrayZ[src];
        ArrayBytes[dest]	= ArrayBytes[src];
        ArraySize[dest]		= ArraySize[src];
        ArrayPadding[dest]	= ArrayPadding[src];
        return 1;
    }
    return 0;

}


static INT_PTR fm_fillArray (INT_PTR nr, INT_PTR val)
{
    int i;

    if (!IS_ALLOCATED_ARRAY(nr) || ArraySize[nr] == 0) return 0;	// afh 01-may-2008

    if (ArrayBytes[nr]==1){
        
        if (val<0) val=0; 
        else if (val>255) val=255;

        memset( Array[nr], (int)val, ArraySize[nr] );

    } else if (ArrayBytes[nr]==2){

        if (val<0) val=0; 
        else if (val>65535) val=65535;

        //this could be optimized...
        for (i=0;i<ArraySize[nr];i+=2){
             *(unsigned16 *)(((unsigned8 *) Array[nr]) + i) = (unsigned16)val;
        }

    } else if (ArrayBytes[nr]==4){
        
        //this could be optimized...
        for (i=0;i<ArraySize[nr];i+=4){
            *(unsigned32 *)(((unsigned8 *) Array[nr]) + i ) = (unsigned32)val;
        }
    
	} else if (ArrayBytes[nr]==8){
#ifndef _WIN64
		//surely this never worked properly in Win32, but we'll
		//leave it this way now for backwards "incompatibility"
        double dval = (double)val;
        
        for (i=0;i<ArraySize[nr];i+=8){
            //could maybe do a 64-bit int copy??-afh
            *(double *)(((unsigned8 *) Array[nr]) + i ) = dval;
        }
#else
		//For Win64, we fill Array with an 8-byte integer value
		//(which we should really do also for Win32...).
		//To fill an Array with a double value, use ffillArray
		//instead!

        //this could be optimized...
        for (i=0;i<ArraySize[nr];i+=8){
            *(UINT64 *)(((unsigned8 *) Array[nr]) + i ) = (UINT64)val;
        }
#endif
	}

    return true;
}

//----------------------------------------------------------------
//  bool = ffillArray(nr, dval)
//
//      Fill an Array (nr) with a f.p. value dval.
//      dval will be converted to half, float, or double if
//      ArrayBytes[nr] is 2, 4, or 8, resp.
//
//  Returns false if failed (bad value for nr, Array not
//      allocated, ArrayBytes not 2, 4, or 8, etc.).
//  Returns true if successful.
//----------------------------------------------------------------
static INT_PTR fm_ffillArray (INT_PTR nr, double dval)
{
    int i;

    if (!IS_ALLOCATED_ARRAY(nr) || ArraySize[nr] == 0) return false;

    if (ArrayBytes[nr]==2){
        //fill with half (16-bit f.p.) value
        unsigned short hval;
		UIF fval;
		fval.f = (float)dval;
        hval = floatToHalf( fval.i );

        //this could be optimized...
        for (i=0;i<ArraySize[nr];i+=2){
             *(unsigned short *)(((unsigned8 *) Array[nr]) + i) = hval;
        }

    } else if (ArrayBytes[nr]==4){
        //fill with float (32-bit f.p.) value
        float fval = (float)dval;
        //this could be optimized...
        for (i=0;i<ArraySize[nr];i+=4){
            *(float *)(((unsigned8 *) Array[nr]) + i ) = fval;
        }
    
	} else if (ArrayBytes[nr]==8){
        //fill with double (64-bit f.p.) value
        for (i=0;i<ArraySize[nr];i+=8){
            //could maybe do a 64-bit int fill instead??
            *(double *)(((unsigned8 *) Array[nr]) + i ) = dval;
        }
    } else
        return false; //bad value for ArrayBytes[nr].

    return true;

}//fm_ffillArray




static INT_PTR fm_ctlEnabled(INT_PTR n)
{
//#if 0
    //// NO! Do not perpetuate this travesty!!!
    //// See note at fm_enableCtl(). AFH 20Aug2008
    int state= (int)gParams->ctl[n].enabled & 3;

    if (state==3) return -1;
    else return state;
//#else
    return gParams->ctl[n].enabled;
//#endif
} 

#define MAX_RECURSIVE_TAB_CONTROLS 16
/*	fm_ctlEnabledAs

	Returns the enabled state as the control is rendered on screen, not the internal state.
	Initially the "enabledAs" state depends on whether it has a Tab control assigned, but
	it could be easily changed to allow other parent controls or external criteria.
	
	Returns:
		0 (invisible and disabled)
		1 (visible but disabled)
		3 (visible and enabled)
*/ //MWVDL 2008-08-19
static INT_PTR ctlEnabledAs(INT_PTR n, INT_PTR depth)
{
	if(depth < 0) {
		ErrorOk("More than %d recursive tab controls", MAX_RECURSIVE_TAB_CONTROLS);
	}
	
    if ((!IS_INUSE_CTL(n)) && n != CTL_PROGRESS && n != CTL_FRAME && n != CTL_ZOOM) { //&& n != CTL_PREVIEW
        return 0;
    }

    {//scope
        INT_PTR level = gParams->ctl[n].enabled;				// by default, level is n.enabled
        const int t = gParams->ctl[n].tabControl;
        if (t >= 0) {											// is tabControl set
	        if (t < N_CTLS											// does tabControl exist? (in range)
		     && gParams->ctl[t].inuse								// is tabControl used?
	         && gParams->ctl[t].class == CC_TAB						// is tabControl a tab control?
	         && gParams->ctl[t].val == gParams->ctl[n].tabSheet) {	// is tabControl's sheet same as n's sheet?
		        level = min(level, ctlEnabledAs(t, --depth));		// lowest state of enabled, fm_ctlEnabledAs(tab)
	        } else {
		        level = 0;											// invisible
	        }
        }
	    return level;
    }//scope
}
INT_PTR fm_ctlEnabledAs(INT_PTR n)
{
	return ctlEnabledAs(n, MAX_RECURSIVE_TAB_CONTROLS);
}



//function added by Harald Heim on Dec 7, 2002
static INT_PTR fm_rgb2lab(INT_PTR r, INT_PTR g, INT_PTR b, INT_PTR z)
{

    double rVal, gVal, bVal;
    double xVal, yVal, zVal;
    double lVal, aVal;
    double fX, fY, fZ;

	// Cached values, should be per-thread.
    THREAD_LOCAL static int Cr=-1,Cg=-1,Cb=-1;	//force initial cache miss with -1's
    THREAD_LOCAL static int Cl,Ca,Cb2;


    if (r==Cr && g==Cg && b==Cb){ //Read from Cache

        if (z==0)
            return Cl;
        else if (z==1)
            return Ca;
        else
            return Cb2;   


    } else { //Calculate
        
        Cr=(int)r;
        Cg=(int)g;
        Cb=(int)b;

        rVal = (double) r;
        gVal = (double) g;
        bVal = (double) b;

#if 0	//calculate at run-time (unless compiler optimizes!)
        xVal = 0.412453 * rVal + 0.357580 * gVal + 0.180423 * bVal;
        xVal /= (255.0 * 0.950456);
        yVal = 0.212671 * rVal + 0.715160 * gVal + 0.072169 * bVal;
        yVal /=  255.0;
        zVal = 0.019334 * rVal + 0.119193 * gVal + 0.950227 * bVal;
        zVal /= (255.0 * 1.088754);

#else	//calculated in advance and double-checked against above;
		//should avoid any f.p. divides...
        //Provided by Bill MacBeth
        xVal = 0.00170178 *rVal + 0.00147537 * gVal + 0.000744423 * bVal;
        yVal = 0.000834004 * rVal + 0.00280455 * gVal + 0.000283016 * bVal;
        zVal = 0.0000696389 * rVal + 0.000429320 * gVal + 0.00342261 * bVal;
#endif

        if (yVal > 0.008856){
          fY = pow(yVal, 1.0 / 3.0);	//VC++ folds these and other constant divides at compile time...
          lVal = 116.0 * fY - 16.0;
        } else {
          fY = 7.787 * yVal + 16.0 / 116.0;
          lVal = 903.3 * yVal;
        }
        
        if (xVal > 0.008856)
            fX = pow(xVal, 1.0 / 3.0);
        else
            fX = 7.787 * xVal + 16.0 / 116.0;
                
        if (zVal > 0.008856)
            fZ = pow(zVal, 1.0 / 3.0);
        else
            fZ = 7.787 * zVal + 16.0 / 116.0;


        lVal = lVal * 2.56;
        aVal = 500.0 * (fX - fY)+128.0;
        bVal = 200.0 * (fY - fZ)+128.0;

        Cl=(int) lVal;
        Ca=(int) aVal;
        Cb2=(int) bVal;

        if (z==0)
            return Cl;
        else if (z==1)
            return Ca;
        else
            return Cb2;   
    }


}


//function added by Harald Heim on Dec 7, 2002
static INT_PTR fm_lab2rgb(INT_PTR l, INT_PTR a, INT_PTR b, INT_PTR z)
{
    double rVal, gVal, bVal;
    double xVal, yVal, zVal;
    double lVal, aVal;
    double fX, fY, fZ;

	// Cached values, should be per-thread.
    THREAD_LOCAL static int Cr,Cg,Cb;
    THREAD_LOCAL static int Cl=-999,Ca=-999,Cb2=-999; //force initial cache miss with impossible values


    if (l==Cl && a==Ca && b==Cb2){ //Read from Cache

        if (z==0) {         
            return Cr;
        } else if (z==1) {
            return Cg;
        } else {
            return Cb;
        }

    } else { //Calculate

        Cl=(int)l;
        Ca=(int)a;
        Cb2=(int)b;

        lVal = (double) l;
        aVal = (double) a;
        bVal = (double) b;


        lVal= lVal / 2.56;
        aVal = aVal - 128.0;
        bVal = bVal - 128.0;


        fY = pow((lVal + 16.0) / 116.0, 3.0);
        if (fY < 0.008856)
            fY = lVal / 903.3;
        yVal = fY;

        if (fY > 0.008856)
        fY = pow(fY, 1.0 / 3.0);
        else
        fY = 7.787 * fY + 16.0 / 116.0;

        fX = aVal / 500.0 + fY;
        if (fX > 0.206893)
            xVal = pow(fX, 3.0);
        else
            xVal = (fX - 16.0 / 116.0) / 7.787;

        fZ = fY - bVal /200.0;
        if (fZ > 0.206893)
            zVal = pow(fZ, 3.0);
        else
            zVal = (fZ - 16.0 / 116.0) / 7.787;

#if 1	//Will the VC++ compiler fold these constant operations at compile-time?
        xVal *= (0.950456 * 255.0);
        yVal *=             255.0;
        zVal *= (1.088754 * 255.0);
        rVal = xVal *  3.240479 - yVal * 1.537150 - zVal * 0.498535;
        gVal = xVal * -0.969256 + yVal * 1.875992 + zVal * 0.041556;
        bVal = xVal *  0.055648 - yVal * 0.204043 + zVal * 1.057311;
        
#else	//calculated in advance but only the first factor seems correct!!!
		//don't use for now, until this mismatch is sorted out!!!
		//(Are they based on different white points???)
        //Provided by Bill MacBeth
        rVal =   xVal * 7.853828E+02 - yVal * 6.342260E-03 - zVal * 2.056949E-03;
        gVal = - xVal * 2.471603E+02 + yVal * 7.356831E-03 + zVal * 1.629647E-04;
        bVal =   xVal * 1.544968E+01 - yVal * 7.349398E-04 + zVal * 3.808315E-03;
#endif

        Cr=(int) (rVal + 0.5);
        Cg=(int) (gVal + 0.5);
        Cb=(int) (bVal + 0.5);

        if (z==0) {         
            return Cr;
        } else if (z==1) {
            return Cg;
        } else {
            return Cb;
        }

    }


}

//function added by Harald Heim on Dec 13, 2002
static INT_PTR fm_setClickDrag(INT_PTR b)
{
    //b=0 for left click drag
    //b=1 for right click drag
    //b=2 for no click drag at all
    ClickDrag=(int)b;
    return 1;
}


/*
VOID CALLBACK MyTimerProc( 
    HWND hwnd,        // handle to window for timer messages 
    UINT message,     // WM_TIMER message 
    UINT idTimer,     // timer identifier 
    DWORD dwTime)     // current system time 
{ 

        fmc.n = idTimer-3;
        fmc.e = FME_TIMER;
        fmc.previous = 0;//previous;
        fmc.ctlMsg = WM_TIMER;//event;//WM_SIZE;
        fmc.ctlCode = 0;
        fmc.ctlClass = 0;//gParams->ctl[n].class;
        
		//Catch any longjmp's back...
		if (setjmp(Mde_error_exit_jb)) {
			//someone longjmp'ed back to us...
			return;   //abort
		}
		//Invoke the On Control handler
        X86_call_fmf1(gParams->CB_onCtl, (INT_PTR)&fmc);

}*/


//function added by Harald Heim on Dec 13, 2002
static UINT_PTR fm_setTimerEvent(INT_PTR nr, INT_PTR t, INT_PTR state)
{
    //nr -> timer number 0 - 9
    //t -> trigger time
    //state==0 -> kill timer
    //state==1 -> set timer
    UINT_PTR Retval;

    if (nr<0 && nr>9) return 0;

    KillTimer(fmc.hDlg, 3 + nr); //Cancel any previous trigger.
    
    if (state==1){
        //Retval = SetTimer(fmc.hDlg, 3 + nr, t, (TIMERPROC) MyTimerProc); //Don't use message cue
        Retval = SetTimer(fmc.hDlg, 3 + nr, (UINT)t, NULL);
    }

    return Retval;
}



static INT_PTR fm_getDisplaySettings(INT_PTR s)
{
   DEVMODE dm;
   dm.dmSize = sizeof(DEVMODE);
   dm.dmDriverExtra = 0;

   EnumDisplaySettings(NULL,ENUM_CURRENT_SETTINGS,&dm);
   
   switch (s) {   
        case 0:
            return (INT_PTR) dm.dmBitsPerPel;
        case 1:
            return (INT_PTR) dm.dmPelsWidth;
        case 2:
            return (INT_PTR) dm.dmPelsHeight;
        case 3:
            return (INT_PTR) dm.dmDisplayFrequency; 
   }
    return 0;

}

//function added by Harald Heim on Dec 14, 2002
static INT_PTR fm_scrollPreview(INT_PTR mode, INT_PTR ox, INT_PTR oy)
{

    INT_PTR RetVal;
    
    if (mode==1){
        ox = ox - (fmc.X-1)/2;
        oy = oy - (fmc.Y-1)/2;
    }

    RetVal = scrollPreview(fmc.hDlg, mode, ox, oy);
    return RetVal;

}

//////////////////////// BEGIN COLOR CONVERSIONS //////////////////////////

//
//	int rgb2hsl(int r, int g, int b, int z)
//
//	Converts an (r,g,b) triple to HSL, and
//	returns one requested component (h, s, or l)
//	per call, where (z mod 4) specifies which 
//	value will be returned: 'h' for z=0 (mod 4),
//	's' for z=1 (mod 4), or 'l' for z=2 (mod 4).
//	Returns 0 if z=3 (mod 4).
//
//	Calculated (h,s,l) components are cached
//	across each call to rgb2hsl(), to speed 
//	things up by avoiding recalculating all
//	components if they have already been
//	calculated for the same (r,g,b) color
//	by the previous call.
//
//	The (r,g,b) and (h,s,l) components will be	
//	treated as 8 bpc or 16 bpc, as determined by
//	the last call to the set_bitdepth_mode() routine.
//
#if 0
//function added by Harald Heim on Dec 23, 2002
//function as modified by Alex Hunter on Apr 24, 2014
static INT_PTR fm_rgb2hsl(INT_PTR r, INT_PTR g, INT_PTR b, INT_PTR z)
{
    int h,s,l;
    double rf,gf,bf,hf,sf,lf, cmax,cmin,diff;
    
	//Note that TLS access is relatively expensive,
	//so we try to minimize the number of accesses,
	//such as by combining (r,g,b) into a single
	//key value.
	//Question: Is it really worth the extra TLS overhead
	//to cache thread-local values, or is it just as fast
	//to always calculate h/s/l for each call??
	THREAD_LOCAL static int Ch,Cs,Cl;
	THREAD_LOCAL static QWORD cachedKey = -1;
	QWORD key;

	//compute a unique 64-bit key from r, g, b
	key = ( (QWORD)(((DWORD)r << 16) | (DWORD)g) << 16) | (DWORD)b;

	if (key == cachedKey) {
		//retrieve cached result
		switch (z) {
		case 0: return Ch;
		case 1: return Cs;
		case 2: return Cl;
		}
		return -1;	//error
	}
	else {
		//Not in cache -- must calculate

        //Scale rgb to float
#if 0
		//would multiplication by inverse be faster?? (see Agner Fog)
        rf=(double)r / bitMultiply255;
        gf=(double)g / bitMultiply255;
        bf=(double)b / bitMultiply255;
#else
		//Should be faster. But what about roundoff error??
        rf=(double)r * bitDivide255;
        gf=(double)g * bitDivide255;
        bf=(double)b * bitDivide255;
#endif

        //Convert RGB to HSL
#if 0
        cmax = bf;
        cmin = bf;
        if (rf > cmax) {cmax = rf;}
        if (gf > cmax) {cmax = gf;}
        if (rf < cmin) {cmin = rf;}
        if (gf < cmin) {cmin = gf;}
#elif 1
        cmax = bf;
        cmin = bf;
        if      (rf > cmax) cmax = rf;
        else if (rf < cmin) cmin = rf;
        if      (gf > cmax) cmax = gf;
        else if (gf < cmin) cmin = gf;
#else
		cmax = max(max(bf,rf),gf);
		cmin = min(min(bf,rf),gf);
#endif

        lf = (cmax + cmin)/2.0;

        if (cmax == cmin){
            sf = 0.0;
            hf = 0.0;
        } else {
			
			diff = cmax - cmin;
			if (lf < 0.5) {
				sf = diff/(cmax + cmin); 
			} else {
				sf = diff/(2.0 - (cmax + cmin));
			}

			if (rf == cmax) {
				hf = (gf - bf)/diff;
			} else if (gf == cmax) {
				hf = 2.0 + (bf - rf)/diff;
			} else {
				hf = 4.0 + (rf - gf)/diff;
			}
			hf = hf/6.0;
			if (hf < 0.0) hf = hf + 1.0;
        }   
        
        //Scale hsl to int (with rounding!)
		//and cache the results
        Ch = h = (int)(hf*bitMultiply255 + 0.5);
        Cs = s = (int)(sf*bitMultiply255 + 0.5);
        Cl = l = (int)(lf*bitMultiply255 + 0.5);

		cachedKey = key;

    } //end of calculation

	//Return the desired result
	switch (z) {
	case 0: return h;
	case 1: return s;
	case 2: return l;
	}
	return -1;	//error
}//fm_rgb2hsl
#else
//function added by Harald Heim on Dec 23, 2002
//function as modified by Alex Hunter on Apr 24, 2014
static INT_PTR fm_rgb2hsl(INT_PTR r, INT_PTR g, INT_PTR b, INT_PTR z)
{
    double rf,gf,bf,hf,sf,lf;
	double cmax,cmin,diff;
    
	//Note that TLS access is relatively expensive,
	//so we try to minimize the number of accesses,
	//such as by combining (r,g,b) into a single
	//key value.
	//Question: Is it really worth the extra TLS overhead
	//to cache thread-local values, or is it just as fast
	//to always calculate h/s/l for each call??
	typedef union {
		__int64 key;
		WORD wrd[4];
	} key_t;
	key_t rgb, hsl;
	THREAD_LOCAL static key_t cached_rgb = {-1};
	THREAD_LOCAL static key_t cached_hsl;

	//create a unique 64-bit key from r, g, b
	rgb.wrd[0] = (WORD)r;
	rgb.wrd[1] = (WORD)g;
	rgb.wrd[2] = (WORD)b;
	rgb.wrd[3] = 0;	//last word must be determinate!

	if (rgb.key == cached_rgb.key) {
		//return the previously cached result
		return cached_hsl.wrd[z & 3];
	}
	else {
		//Not in the cache -- must calculate

        //Scale rgb to float
#if 0
		//would multiplication by inverse be faster?? (see Agner Fog)
        rf=(double)r / bitMultiply255;
        gf=(double)g / bitMultiply255;
        bf=(double)b / bitMultiply255;
#else
		//Should be faster. But what about roundoff error??
        rf=(double)r * bitDivide255;
        gf=(double)g * bitDivide255;
        bf=(double)b * bitDivide255;
#endif

        //Convert RGB to HSL

		//cmax = max(bf,rf,gf);
		//cmin = min(bf,rf,gf);
        cmax = bf;
        cmin = bf;
        if      (rf > cmax) cmax = rf;
        else if (rf < cmin) cmin = rf;
        if      (gf > cmax) cmax = gf;
        else if (gf < cmin) cmin = gf;

        lf = (cmax + cmin)/2.0;

        if (cmax == cmin){
            sf = 0.0;
            hf = 0.0;
        } else {
			
			diff = cmax - cmin;
			if (lf < 0.5) {
				sf = diff/(cmax + cmin); 
			} else {
				sf = diff/(2.0 - (cmax + cmin));
			}

			if (rf == cmax) {
				hf = (gf - bf)/diff;
			} else if (gf == cmax) {
				hf = 2.0 + (bf - rf)/diff;
			} else {
				hf = 4.0 + (rf - gf)/diff;
			}
			hf = hf/6.0;
			if (hf < 0.0) hf = hf + 1.0;
        }   
        
        //Scale hsl to int (with rounding!)
        hsl.wrd[0] = (int)(hf*bitMultiply255 + 0.5);
        hsl.wrd[1] = (int)(sf*bitMultiply255 + 0.5);
        hsl.wrd[2] = (int)(lf*bitMultiply255 + 0.5);
		hsl.wrd[3] = 0; //unused

		//and cache the results
		cached_rgb = rgb;	//lookup key
		cached_hsl = hsl;	//result key

    } //end of calculation

	//Return the desired result
	return hsl.wrd[z & 3];
}//fm_rgb2hsl
#endif



//function added by Harald Heim on Dec 7, 2002
static INT_PTR fm_hsl2rgb(INT_PTR h, INT_PTR s, INT_PTR l, INT_PTR z)
{

    //int r,g,b;
    double rf,gf,bf,hf,sf,lf, hf2, m1,m2;

    THREAD_LOCAL static int Cr,Cg,Cb;
    THREAD_LOCAL static int Ch=-1,Cs=-1,Cl=-1;	//force initial cache miss with -1's


    if (h==Ch && s==Cs && l==Cl){ //Read from Cache

        if (z==0)
            return Cr;
        else if (z==1)
            return Cg;
        else
            return Cb;    


    } else { //Calculate

        Ch=(int)h;
        Cs=(int)s;
        Cl=(int)l;

        //Scale to float
#if 0
		//would multiplication by inverse be faster?? (see Agner Fog)
        hf=(double)h / bitMultiply255;
        sf=(double)s / bitMultiply255;
        lf=(double)l / bitMultiply255;
#else
		//Should be faster. But what about roundoff error??
        hf=(double)h * bitDivide255;
        sf=(double)s * bitDivide255;
        lf=(double)l * bitDivide255;
#endif

        //Convert HSL to RGB
        if (sf == 0.0) {
            rf = lf;
            gf = lf;
            bf = lf;
        } else {
			if (lf <= 0.5) {m2 = lf*(1.0+sf);} else {m2 = lf+sf-(lf*sf);}
			m1 = 2.0 * lf - m2;

            //Calculate Red
            hf2=hf+1.0/3.0;             
            if (hf2 < 0.0) {hf2 = hf2 + 1.0;}
            if (hf2 > 1.0) {hf2 = hf2 - 1.0;}
            if (6.0 * hf2 < 1.0){
                rf = (m1+(m2-m1)*hf2*6.0);}
            else{
                if (2.0 * hf2 < 1.0){
					rf = m2;
                }
                else {
					if (3.0*hf2 < 2.0) {rf = (m1+(m2-m1)*((2.0/3.0)-hf2)*6.0);} else {rf = m1;}
                } 
            }

            //Calculate Green               
            if (hf < 0.0) {hf = hf + 1.0;}
            if (hf > 1.0) {hf = hf - 1.0;}
            if (6.0 * hf < 1.0){
                gf = (m1+(m2-m1)*hf*6.0);}
            else {
                if (2.0 * hf < 1.0){
					gf = m2;	
                }
                else {
					if (3.0*hf < 2.0) {gf = (m1+(m2-m1)*((2.0/3.0)-hf)*6.0);} else {gf = m1;}
                } 
            }

            //Calculate Blue
            hf2=hf-1.0/3.0;             
            if (hf2 < 0.0) {hf2 = hf2 + 1.0;}
            if (hf2 > 1.0) {hf2 = hf2 - 1.0;}
            if (6.0 * hf2 < 1.0) {
                bf = (m1+(m2-m1)*hf2*6.0);}
            else {
                if (2.0 * hf2 < 1.0){
					bf = m2;
                }
                else {
					if (3.0*hf2 < 2.0) {bf = (m1+(m2-m1)*((2.0/3.0)-hf2)*6.0);} else {bf = m1;}
                } 
            }

        }

        //Scale rgb to int (with rounding!)
        Cr = (int)(rf*bitMultiply255 + 0.5);
        Cg = (int)(gf*bitMultiply255 + 0.5);
        Cb = (int)(bf*bitMultiply255 + 0.5);

        if (z==0)
            return Cr;
        else if (z==1)
            return Cg;
        else
            return Cb;
    }
}



//function added by Harald Heim on Feb 6, 2002
static INT_PTR fm_rgb2ycbcr(INT_PTR r, INT_PTR g, INT_PTR b, INT_PTR z)
{

    const INT_PTR bitMultiply128 = (bitDepthMode == 16 ? 128*128 : 1*128);

    // [  Y'  601 ] [ 0.299     0.587     0.114    ] [ R' ] 
    // [  PB  601 ]=[-0.168736 -0.331264  0.5      ]*[ G' ] 
    // [  PR  601 ] [ 0.5      -0.418688 -0.081312 ] [ B' ]   
    
#if INT_MAX	> 3*255*587000	//if numerator can't overflow
	//more precise -- should use in the next release??
    if (z == 0)
        return ( 299000*r + 587000*g + 114000*b)/1000000;	//y
    else if (z == 1)
        return (-168736*r - 331264*g + 500000*b)/1000000 + bitMultiply128;	//cb
    else if (z == 2)
        return ( 500000*r - 418688*g -  81312*b)/1000000 + bitMultiply128;	//cr
    else
        return 0;	//error

#else	
	//less accurate, but backwards compatible (and less prone to overflow?)
    if (z == 0)
        return ( 299*r + 587*g + 114*b)/1000;	//y
    else if (z == 1)
        return (-169*r - 331*g + 500*b)/1000 + bitMultiply128;	//cb
    else if (z == 2)
        return ( 500*r - 419*g -  81*b)/1000 + bitMultiply128;	//cr
    else
        return 0;	//error

#endif
}

//function added by Harald Heim on Feb 8, 2002
static INT_PTR fm_ycbcr2rgb(INT_PTR y, INT_PTR cb, INT_PTR cr, INT_PTR z)
{
    
    const INT_PTR bitMultiply128 = (bitDepthMode == 16 ? 128*128 : 1*128);

    cb = cb - bitMultiply128;
    cr = cr - bitMultiply128;
    
    //[ R' ] [ 1.        0.        1.402    ] [  Y'  601 ] 
    //[ G' ]=[ 1.       -0.344136 -0.714136 ]*[  PB  601 ] 
    //[ B' ] [ 1.        1.772     0.       ] [  PR  601 ]  

#if INT_MAX	> 2*255*1772000	//if numerator can't overflow
	//more precise -- should use in the next release??
    if (z == 0)
        return y + (              1402000*cr)/1000000;	//r
    else if (z == 1)
        return y - (  344136*cb +  714136*cr)/1000000;	//g
    else if (z == 2)
        return y + ( 1772000*cb             )/1000000;	//b
    else
        return 0;	//error

#else
	//less accurate, but backwards compatible (and less prone to overflow?)
    if (z == 0)
        return y + (           1402*cr)/1000;	//r
    else if (z == 1)
        return y - (  344*cb +  714*cr)/1000;	//g
    else if (z == 2)
        return y + ( 1772*cb          )/1000;	//b
    else
        return 0;	//error

#endif
}


//function added by Harald Heim on Jun 17, 2003
static INT_PTR fm_rgb2ypbpr(INT_PTR r, INT_PTR g, INT_PTR b, INT_PTR z)		//not referenced???
{
    const INT_PTR bitMultiply128 = (bitDepthMode == 16 ? 128*128 : 1*128);

    //Y=  0.2122*Red+0.7013*Green+0.0865*Blue    
    //Pb=-0.1162*Red-0.3838*Green+0.5000*Blue    
    //Pr= 0.5000*Red-0.4451*Green-0.0549*Blue
    
#if INT_MAX	> 3*255*7013	//if numerator can't overflow
	//more accurate -- should use in the next release??
    if (z == 0)
        return ( 2122*r + 7013*g +  865*b)/10000;	//y
    else if (z == 1)
        return (-1162*r - 3838*g + 5000*b)/10000 + bitMultiply128;	//pb
    else if (z == 2)
        return ( 5000*r - 4451*g -  549*b)/10000 + bitMultiply128;	//pr
    else
        return 0;	//error

#else
	//less accurate, but backwards compatible (and less prone to overflow?)
    if (z == 0)
        return ( 212*r + 701*g +  87*b)/1000;	//y
    else if (z == 1)
        return (-116*r - 384*g + 500*b)/1000 + bitMultiply128;	//pb
    else if (z == 2)
        return ( 500*r - 445*g -  55*b)/1000 + bitMultiply128;	//pr
    else
        return 0;	//error

#endif
}

//function added by Harald Heim on Jun 17, 2003
static INT_PTR fm_ypbpr2rgb(INT_PTR y, INT_PTR pb, INT_PTR pr, INT_PTR z)	//not referenced???
{
    const INT_PTR bitMultiply128 = (bitDepthMode == 16 ? 128*128 : 1*128);
    
    pb = pb - bitMultiply128;
    pr = pr - bitMultiply128;
    
    //Red  =1*Y+0.0000*Pb+1.5756*Pr
    //Green=1*Y-0.2253*Pb+0.5000*Pr		//correct??, if so fix bug below
    //Blue =1*Y+1.8270*Pb+0.0000*Pr

#if INT_MAX	> 2*255*18270	//if numerator can't overflow
	//more accurate -- should use in the next release??
    if (z == 0)
        return y + (            15756*pr)/10000; //r 
    else if (z == 1)
        return y - (  2253*pb -  5000*pr)/10000; //g	//bug fixed???
    else if (z == 2)
        return y + ( 18270*pb           )/10000; //b
    else
        return 0;	//error

#else
	//less accurate, but backwards compatible (and less prone to overflow?)
    if (z == 0)
        return y + (           1576*pr)/1000; //r
    else if (z == 1)
        return y - (  225*pb -  500*pr)/1000; //g	//bug fixed???
    else if (z == 2)
        return y + ( 1827*pb          )/1000; //b
    else
        return 0;	//error

#endif
}

//////////////////////// BEGIN COLOR CONVERSIONS //////////////////////////


//// Optimizations: Pluggable variants based on Z, bit depth, etc.?
////                Use aligned 1 byte/2 byte or 2 byte/1 byte accesses?
////                Factor out subscript expression if VC doesn't.

//// How about  srcRGB(), srcRGBA(), srcGrey(), srcDuo(), srcCMYK(), etc.
////         or src3(), src4(), src1(), src2(), src4(), ...
//// Ditto for  pgetRGB(), pgetRGBA(), etc.?

//Added by Harald Heim, January 2003

// srcp() -- get entire pixel from source image buffer
static INT_PTR fm_srcp(INT_PTR x, INT_PTR y)
{   
	register BYTE *pixelp;	//pixel pointer

	//Clamp coordinates to border of image -- or
	//should it be to border of curent tile, in case
	//isTileable is true and the entire image is
	//not accessible???
	//(Also, should we use edge mode instead??)
#if 1
    if (x < 0) x = 0;
    else if (x >= fmc.X) x = fmc.X - 1;
    if (y < 0) y = 0;
    else if (y >= fmc.Y) y = fmc.Y - 1;
#else
	//or is this faster when optimized??...
	x = max(x, 0);
	x = min(x, fmc.X - 1);
	y = max(y, 0);
	y = min(y, fmc.Y - 1);
#endif

	//compute address of current pixel
	pixelp = ((BYTE *)MyAdjustedInData) 
                             +  (x-fmc.x_start)*fmc.srcColDelta
                             +  (y-fmc.y_start)*fmc.srcRowDelta;

	// Would a switch(fmc.Z) statement be faster than the chain of if-clauses???
	if (fmc.Z >= 4) {
		// At least 4 bytes per pixel, so it's safe to load 32-bits
		// at a time without risking a #GP trap...
		// Return rgba or cmyk or ???
        return *(unsigned32 *)pixelp;
	}
	else if (fmc.Z == 3) {
		// 3 bytes per pixel (rgb with no alpha, or cmy, or??)
#if 0 //Is the following a speedup, or is it actually a slowdown???
		//The additional testing for end-of-buffer may actually be
		//more expensive than the time saved in using a single 32-bit
		//access to the pixel buffer!!
		if (x != fmc.x_end-1 || y != fmc.y_end-1){
			//We are not at the last pixel of the current tile, so
			//we can safely access 4 bytes at once without a #GP trap.
			//Note that this loads garbage into the a component, so
			//we force the high byte 3 (a) to 255 (fully opaque).
			//But is this correct for cmy, in which case we set the
			//missing k component to 255??
			return *(unsigned32 *)pixelp | 0xFF000000;	//set a = 255
		}
		else
#endif //speedup???
		{
			// We're either at the very last pixel in the current tile, or
			// the last pixel in the image if the image is not tiled (in
			// which case the current tile IS the entire image), and
			// there are only 3 bytes per pixel.
			// If so, trying to load the 3 bytes as a single 32-bit word may
			// cause a memory access violation at the end of the image/tile buffer.
			// (Note that some hosts, such as PSP, are much more likely to have
			// tight access bounds on the buffer, whereas other hosts, such as
			// Photoshop, are more likely to have relaxed access bounds to
			// allow for "sloppy" pixel accesses which cross over the end of
			// the buffer.)

			//We set the high byte 3 (a or k) to 255 for fully opaque alpha,
			//or for full black -- is the latter correct???
#if 0
			register int rg, b;
			rg = *(unsigned16 *)pixelp;
			b = pixelp[2];
			return b << 16 | rg | 0xFF000000;	//set a=255
#elif 1
			//Same as above when optimized by MSVC...
			return pixelp[2] << 16 | *(unsigned16 *)pixelp | 0xFF000000;	//set a=255
#endif
		}
	}
	else if (fmc.Z == 2) {
		//duotone??? return 2 bytes
		return *(unsigned16 *)pixelp;
	}
	else { //fmc.Z must be 1 (< 1 is not allowed!!)
		//greyscale, return 1 byte
		return *(unsigned8 *)pixelp;
    }
} /*fm_srcp*/


// pgetp() -- get pixel from output buffer
// (same logic as srcp() above)
static INT_PTR fm_pgetp(INT_PTR x, INT_PTR y)
{   
	register BYTE *pixelp;	//pixel pointer

	//clamp coordinates to border of current tile
	//(or should we use edge mode??)
#if 1
    if (x < fmc.x_start) x = fmc.x_start;
    else if (x >= fmc.x_end) x = fmc.x_end - 1;
    if (y < fmc.y_start) y = fmc.y_start;
    else if (y >= fmc.y_end) y = fmc.y_end - 1;
#else
	//or is this faster when optimized??...
	x = max(x, fmc.x_start);
	x = min(x, fmc.x_end - 1);
	y = max(y, fmc.y_start);
	y = min(y, fmc.y_end - 1);
#endif

	//Compute address of current pixel
	pixelp = (unsigned8 *)gStuff->outData
					+ (x-fmc.x_start)*fmc.dstColDelta
					+ (y-fmc.y_start)*fmc.dstRowDelta;

	// Would a switch(fmc.Z) statement be faster than the chain of if-clauses???
	if (fmc.Z >= 4) {
		// At least 4 bytes per pixel, so it's safe to load 32-bits
        return *(unsigned32 *)pixelp;	// Return rgba or cmyk or ???
	}
	else if (fmc.Z == 3) {
		// 3 bytes per pixel (rgb with no alpha, or cmy, or??)
#if 0 //Is the following a speedup, or is it actually a slowdown???
		if (x != fmc.x_end-1 || y != fmc.y_end-1){
			//We are not at the last pixel of the current tile, so
			//we can safely access 4 bytes at once without a #GP trap.
			return *(unsigned32 *)pixelp | 0xFF000000;	//set a = 255
		}
		else
#endif //speedup???
		{
			// We're at the very last pixel in the current tile, and
			// there are only 3 bytes per pixel.
			// We set the high byte 3 to 255 for fully opaque alpha,
			// or for full black with cmyk -- is the latter correct???
			return pixelp[2] << 16 | *(unsigned16 *)pixelp | 0xFF000000;	//set a=255
		}
	}
	else if (fmc.Z == 2) {
		//duotone??? return 2 bytes
		return *(unsigned16 *)pixelp;
	}
	else { //fmc.Z must be 1
		//greyscale, return 1 byte
		return *(unsigned8 *)pixelp;
    }//fmc.Z
} /*fm_pgetp*/


static INT_PTR fm_tgetp(INT_PTR x, INT_PTR y)
{   
	//Clamp coordinates to image size...
    //Added by Harald Heim, April 20, 2003
    if (x < 0) x = 0;
    else if (x >= fmc.X) x = fmc.X - 1;
    if (y < 0) y = 0;
    else if (y >= fmc.Y) y = fmc.Y - 1;

#if 0 //Should we check for buffer allocation?? -- Shouldn't FM allocate tbufs automatically??
    if (!fmc.tbuf) {
        // tbuf wasn't allocated????
        ErrorOk("No tile buffer!");
        return 0;
    }
#endif

	//return 4 bytes at current pixel
	//no worry about #GP at end of buffer (it's padded + 3 bytes)
	//Caller is responsible to clear unused bytes
	//(or should we do that here??)
    return *(unsigned32 *) (fmc.tbuf + (y*fmc.X + x)*fmc.Z );
} /*fm_tgetp*/

static INT_PTR fm_t2getp(INT_PTR x, INT_PTR y)
{   
	//Clamp coordinates to image size...
    //Added by Harald Heim, April 20, 2003
    if (x < 0) x = 0;
    else if (x >= fmc.X) x = fmc.X - 1;
    if (y < 0) y = 0;
    else if (y >= fmc.Y) y = fmc.Y - 1;

	//return 4 bytes at current pixel
	//no worry about #GP at end of buffer (it's padded + 3 bytes)
	//Caller is responsible to clear unused bytes
	//(or should we do that here??)
    return *(unsigned32 *) (fmc.t2buf + (y*fmc.X + x)*fmc.Z );
} /*fm_t2getp*/

static INT_PTR fm_t3getp(INT_PTR x, INT_PTR y)
{   
	//Clamp coordinates to image size...
    //Added by Harald Heim, April 20, 2003
    if (x < 0) x = 0;
    else if (x >= fmc.X) x = fmc.X - 1;
    if (y < 0) y = 0;
    else if (y >= fmc.Y) y = fmc.Y - 1;

	//return 4 bytes at current pixel
	//no worry about #GP at end of buffer (it's padded + 3 bytes)
	//Caller is responsible to clear unused bytes
	//(or should we do that here??)
    return *(unsigned32 *) (fmc.t3buf + (y*fmc.X + x)*fmc.Z );
} /*fm_t3getp*/

static INT_PTR fm_t4getp(INT_PTR x, INT_PTR y)
{   
	//Clamp coordinates to image size...
    //Added by Harald Heim, April 20, 2003
    if (x < 0) x = 0;
    else if (x >= fmc.X) x = fmc.X - 1;
    if (y < 0) y = 0;
    else if (y >= fmc.Y) y = fmc.Y - 1;

	//return 4 bytes at current pixel
	//no worry about #GP at end of buffer (it's padded + 3 bytes)
	//Caller is responsible to clear unused bytes
	//(or should we do that here??)
    return *(unsigned32 *) (fmc.t4buf + (y*fmc.X + x)*fmc.Z );
} /*fm_t4getp*/


//// Use "pluggable" versions of tsetp: tsetp4 when Z >=4;
///  tsetp1, tsetp2, tsetp3 when Z==1, 2, 3, resp.

//// Overlaps when Z<4???
static INT_PTR fm_tsetp(INT_PTR x, INT_PTR y, INT_PTR val)
{   
    /// WHY CLAMP COORDINATES?  Shouldn't the tsetp be ignored if out of bounds????
#if 0
    if (x < 0) x = 0;
    else if (x >= fmc.X) x = fmc.X - 1;
    if (y < 0) y = 0;
    else if (y >= fmc.Y) y = fmc.Y - 1;
#else
	//abort if outside image buffer (should padding be considered??)
    if (x < 0 || x >= fmc.X ||
		y < 0 || y >= fmc.Y)  return -1;	//failure
#endif
    
	//For now, we always store 4 bytes.  This means 1 or more bytes will
	//overlap on each subsequent store when Z<4.  Be sure you call tsetp
	//in ascending coordinate order so everything works out okay!!
    *(unsigned32 *) ( fmc.tbuf + (y*fmc.X + x)*fmc.Z ) = (unsigned32)val;

    return val;

} /*fm_tsetp*/

static INT_PTR fm_t2setp(INT_PTR x, INT_PTR y, INT_PTR val)
{   
    /// WHY CLAMP COORDINATES?  Shouldn't the tset2p be ignored if out of bounds????
#if 0
    if (x < 0) x = 0;
    else if (x >= fmc.X) x = fmc.X - 1;
    if (y < 0) y = 0;
    else if (y >= fmc.Y) y = fmc.Y - 1;
#else
	//abort if outside image buffer (should padding be considered??)
    if (x < 0 || x >= fmc.X ||
		y < 0 || y >= fmc.Y)  return -1;	//failure
#endif
    
	//For now, we always store 4 bytes.  This means 1 or more bytes will
	//overlap on each subsequent store when Z<4.  Be sure you call t2setp
	//in ascending coordinate order so everything works out okay!!
    *(unsigned32 *) ( fmc.t2buf + (y*fmc.X + x)*fmc.Z ) = (unsigned32)val;

    return val;

} /*fm_t2setp*/

static INT_PTR fm_t3setp(INT_PTR x, INT_PTR y, INT_PTR val)
{   
    /// WHY CLAMP COORDINATES?  Shouldn't the t3setp be ignored if out of bounds????
#if 0
    if (x < 0) x = 0;
    else if (x >= fmc.X) x = fmc.X - 1;
    if (y < 0) y = 0;
    else if (y >= fmc.Y) y = fmc.Y - 1;
#else
	//abort if outside image buffer (should padding be considered??)
    if (x < 0 || x >= fmc.X ||
		y < 0 || y >= fmc.Y)  return -1;	//failure
#endif
    
	//For now, we always store 4 bytes.  This means 1 or more bytes will
	//overlap on each subsequent store when Z<4.  Be sure you call t3setp
	//in ascending coordinate order so everything works out okay!!
    *(unsigned32 *) ( fmc.t3buf + (y*fmc.X + x)*fmc.Z ) = (unsigned32)val;

    return val;

} /*fm_t3setp*/

static INT_PTR fm_t4setp(INT_PTR x, INT_PTR y, INT_PTR val)
{   
    /// WHY CLAMP COORDINATES?  Shouldn't the t4setp be ignored if out of bounds????
#if 0
    if (x < 0) x = 0;
    else if (x >= fmc.X) x = fmc.X - 1;
    if (y < 0) y = 0;
    else if (y >= fmc.Y) y = fmc.Y - 1;
#else
	//abort if outside image buffer (should padding be considered??)
    if (x < 0 || x >= fmc.X ||
		y < 0 || y >= fmc.Y)  return -1;	//failure
#endif
    
	//For now, we always store 4 bytes.  This means 1 or more bytes will
	//overlap on each subsequent store when Z<4.  Be sure you call t4setp
	//in ascending coordinate order so everything works out okay!!
    *(unsigned32 *) ( fmc.t4buf + (y*fmc.X + x)*fmc.Z ) = (unsigned32)val;

    return val;

} /*fm_t4setp*/


//// Use "pluggable" versions of psetp: psetp4 when Z >=4;
///  psetp1, psetp2, psetp3 when Z==1, 2, 3, resp.

// psetp() -- Store entire pixel at once. (Pixel should be rgb or rgba.)
// Returns true if succeeded, or false if out of tile bounds.
// ***Should make this pluggable for efficiency!!***
static INT_PTR fm_psetp(INT_PTR x, INT_PTR y, INT_PTR val)
{   
	register BYTE *pixelp;	//pixel pointer

    /// WHY CLAMP COORDINATES?  Shouldn't the setp be ignored if out of bounds????
#if 0
	//clamp coordinates
    if (x < fmc.x_start) x = fmc.x_start;
    else if (x >= fmc.x_end) x = fmc.x_end - 1;
    if (y < fmc.y_start) y = fmc.y_start;
    else if (y >= fmc.y_end) y = fmc.y_end - 1;
#else
	//abort if outside current tile (should padding be considered??)
    if (x < fmc.x_start || x >= fmc.x_end ||
		y < fmc.y_start || y >= fmc.y_end)  return 0;	//failure
#endif

	//Compute address of current pixel
	pixelp = (unsigned8 *)gStuff->outData
					+ (x-fmc.x_start)*fmc.dstColDelta
					+ (y-fmc.y_start)*fmc.dstRowDelta;

	//Decide whether to store 3 bytes or 4 bytes...
    if (fmc.Z>3 && (psetpMode || (val & 0xFF000000))) {
		//RGBA mode, and either psetpMode is 1 or A is nonzero: Store RGBA.
		//Avoid messup of alpha channel and being forced to read it out,
		//only if set_psetp_mode(1) was used.
        *(unsigned32 *)pixelp = (unsigned32)val;	//RGBA
    }
	else {
		//RGB mode, or RGBA mode and psetpMode is 0 and A is 0
		// -- Store only the 3 RGB bytes (avoids inadvertently
		//setting alpha channel to 0 when storing RGB-only values).
		*(unsigned16 *)pixelp = (unsigned16)val;	//rg
		pixelp[2] = (unsigned8)(val >> 16);			//b
	}
    return 1;	//success
} /*fm_setp*/


static INT_PTR fm_getWindowsVersion(void)
{
    //#define BUFSIZE 80
    
    OSVERSIONINFO osvi;
    BOOL bOsVersionInfoEx;


    osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
    

    if( !(bOsVersionInfoEx = GetVersionEx ((OSVERSIONINFO *) &osvi)) )
    {
      osvi.dwOSVersionInfoSize = sizeof (OSVERSIONINFO);
      if (! GetVersionEx ( (OSVERSIONINFO *) &osvi) ) 
         return FALSE;
    }


    switch (osvi.dwPlatformId)
   {
      // Test for the Windows NT product family.
      case VER_PLATFORM_WIN32_NT:

		// Test for the specific product family.
         if ( osvi.dwMajorVersion == 6) //&& osvi.dwMinorVersion == 0
            //printf ("Microsoft Windows Vista");
            return 11;

         if ( osvi.dwMajorVersion == 5 && osvi.dwMinorVersion == 2 )
            //printf ("Microsoft Windows&nbsp;Windows Server 2003&nbsp;2003 family, ");
            return 10;

         if ( osvi.dwMajorVersion == 5 && osvi.dwMinorVersion == 1 )
            //printf ("Microsoft Windows XP ");
            return 9;

         if ( osvi.dwMajorVersion == 5 && osvi.dwMinorVersion == 0 )
            //printf ("Microsoft Windows 2000 ");
            return 8;

         if ( osvi.dwMajorVersion <= 4 )
            //printf("Microsoft Windows NT ");
            return 7;

         /*
         // Test for specific product on Windows NT 4.0 SP6 and later.
         if( bOsVersionInfoEx )
         {
            // Test for the workstation type.
            if ( osvi.wProductType == VER_NT_WORKSTATION )
            {
               if( osvi.dwMajorVersion == 4 )
                  printf ( "Workstation 4.0 " );
               else if( osvi.wSuiteMask & VER_SUITE_PERSONAL )
                  printf ( "Home Edition " );
               else
                  printf ( "Professional " );
            }
            
            // Test for the server type.
            else if ( osvi.wProductType == VER_NT_SERVER )
            {
               if( osvi.dwMajorVersion == 5 && osvi.dwMinorVersion == 2 )
               {
                  if( osvi.wSuiteMask & VER_SUITE_DATACENTER )
                     printf ( "Datacenter Edition " );
                  else if( osvi.wSuiteMask & VER_SUITE_ENTERPRISE )
                     printf ( "Enterprise Edition " );
                  else if ( osvi.wSuiteMask == VER_SUITE_BLADE )
                     printf ( "Web Edition " );
                  else
                     printf ( "Standard Edition " );
               }

               else if( osvi.dwMajorVersion == 5 && osvi.dwMinorVersion == 0 )
               {
                  if( osvi.wSuiteMask & VER_SUITE_DATACENTER )
                     printf ( "Datacenter Server " );
                  else if( osvi.wSuiteMask & VER_SUITE_ENTERPRISE )
                     printf ( "Advanced Server " );
                  else
                     printf ( "Server " );
               }

               else  // Windows NT 4.0 
               {
                  if( osvi.wSuiteMask & VER_SUITE_ENTERPRISE )
                     printf ("Server 4.0, Enterprise Edition " );
                  else
                     printf ( "Server 4.0 " );
               }
            }
         }
         else  // Test for specific product on Windows NT 4.0 SP5 and earlier
         {
            HKEY hKey;
            char szProductType[BUFSIZE];
            DWORD dwBufLen=BUFSIZE;
            LONG lRet;

            lRet = RegOpenKeyEx( HKEY_LOCAL_MACHINE,
               "SYSTEM\\CurrentControlSet\\Control\\ProductOptions",
               0, KEY_QUERY_VALUE, &hKey );
            if( lRet != ERROR_SUCCESS )
               return FALSE;

            lRet = RegQueryValueEx( hKey, "ProductType", NULL, NULL,
               (LPBYTE) szProductType, &dwBufLen);
            if( (lRet != ERROR_SUCCESS) || (dwBufLen > BUFSIZE) )
               return FALSE;

            RegCloseKey( hKey );

            if ( lstrcmpi( "WINNT", szProductType) == 0 )
               printf( "Workstation " );
            if ( lstrcmpi( "LANMANNT", szProductType) == 0 )
               printf( "Server " );
            if ( lstrcmpi( "SERVERNT", szProductType) == 0 )
               printf( "Advanced Server " );

            printf( "%d.%d ", osvi.dwMajorVersion, osvi.dwMinorVersion );
         }

      // Display service pack (if any) and build number.

         if( osvi.dwMajorVersion == 4 && 
             lstrcmpi( osvi.szCSDVersion, "Service Pack 6" ) == 0 )
         {
            HKEY hKey;
            LONG lRet;

            // Test for SP6 versus SP6a.
            lRet = RegOpenKeyEx( HKEY_LOCAL_MACHINE,
               "SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\Hotfix\\Q246009",
               0, KEY_QUERY_VALUE, &hKey );
            if( lRet == ERROR_SUCCESS )
               printf( "Service Pack 6a (Build %d)\n", osvi.dwBuildNumber & 0xFFFF );         
            else // Windows NT 4.0 prior to SP6a
            {
               printf( "%s (Build %d)\n",
                  osvi.szCSDVersion,
                  osvi.dwBuildNumber & 0xFFFF);
            }

            RegCloseKey( hKey );
         }
         else // Windows NT 3.51 and earlier or Windows 2000 and later
         {
            printf( "%s (Build %d)\n",
               osvi.szCSDVersion,
               osvi.dwBuildNumber & 0xFFFF);
         }
        */


         break;

      // Test for the Windows 95 product family.
      case VER_PLATFORM_WIN32_WINDOWS:

         if (osvi.dwMajorVersion == 4 && osvi.dwMinorVersion == 0)
         {
             //printf ("Microsoft Windows 95 ");
             return 2;
             if ( osvi.szCSDVersion[1] == 'C' || osvi.szCSDVersion[1] == 'B' )
                return 3;
                 //printf("OSR2 " );
         } 

         if (osvi.dwMajorVersion == 4 && osvi.dwMinorVersion == 10)
         {
             //printf ("Microsoft Windows 98 ");
             return 4;
             if ( osvi.szCSDVersion[1] == 'A' )
                return 5;
                 //printf("SE " );
         } 

         if (osvi.dwMajorVersion == 4 && osvi.dwMinorVersion == 90)
         {
             //printf ("Microsoft Windows Millennium Edition\n");
             return 6;
         } 
         break;

      case VER_PLATFORM_WIN32s:

         //printf ("Microsoft Win32s\n");
         return 1;
         break;
    }


    return 0; 

}


static INT_PTR fm_setPreviewCursor(INT_PTR res)
{

    if (GetForegroundWindow() == MyHDlg)
        return setPreviewCursor(res);
    else
        return 0;
        
}

static INT_PTR fm_getPreviewCursor(void)
{

    return getPreviewCursor();

}


//Added by Harald Heim, March 12, 2003
static INT_PTR fm_deleteRgn(INT_PTR Rgn)
{

    return DeleteObject ((HGDIOBJ) Rgn);

}


//Added by Harald Heim, May 17, 2003
/*
int comp(const void *a, const void *b )
{
    return *(int *)a - * (int *)b;
}

static INT_PTR fm_quickSort(INT_PTR first, INT_PTR last)
{

    if (!IS_OK_CELL_NR(first)) return 0;
    if (!IS_OK_CELL_NR(last)) return 0;
    ///if (first > last || first == last) return 0;
    if (first >= last) return 0;

    //qsort( (void*) &fmc.cell[first] , (size_t) (last - first), sizeof(int), (int(*)(const void*, const void*)) comp );
    qsort( (void*) &fmc.cell[first] , (size_t) (last - first + 1), sizeof(int),  comp );

    return 1;

}
*/




//Added by Harald Heim, May 17, 2003		//Why not just use C's qsort as above???
//CAVEAT-- Not multithreaded, operates on global fmc.
//N.B. Recursive implementation; would iterative be better??
static INT_PTR fm_quickSort(INT_PTR first, INT_PTR last)
{
	// Should this nassive 'if' be applied only in a level-0 driver???
    ///if (first<0 || first> N_CELLS-1 || last<0 || last> N_CELLS-1 || first > last || first == last) return 0;
    if (!IS_OK_CELL_NR(first) || !IS_OK_CELL_NR(last) || first >= last) return 0;

    ///if (first < last)	//already known to be true at this point
	{
        int pivot = fmc.cell[first];
        int i = (int)first;
        int j = (int)last;
        
        while (i < j){

            while (fmc.cell[i] <= pivot && i < last){
				i += 1;
            }

            while (fmc.cell[j] >= pivot && j > first){
                j -= 1;
            }
                    
            if (i < j){
                int temp = fmc.cell[i];
                fmc.cell[i] = fmc.cell[j];
                fmc.cell[j] = temp;
            }
        }

        //temp = fmc.cell[first];
        fmc.cell[first] = fmc.cell[j];
        fmc.cell[j] = pivot; //temp;

        fm_quickSort(first, j-1);
        fm_quickSort(j+1, last);
    }

    return 1;

} /*fm_quickSort*/


// Quick fill put/get cells from a square region in an Array
//
// quickFillArray(x, y, z, nr, radius, xstart, ystart, xend, yend)
//
// where:	(x, y, z) are the coordinates of the center of the square
//			nr	   = number of Array to use as source (0 thru N_ARRAYS-1)
//			radius = distance from center to nearest points on top, bottom, left,
//					 and right edges of the square region
//			xstart = leftmost valid pixel x-coordinate in current tile
//			ystart = topmost valid pixel y-coordinate in current tile
//			xend   = rightmost valid pixel x-coordinate + 1 in tile
//			yend   = bottom-most valid pixel y-coordinate + 1 in tile
//	Returns 1 if successful, or 0 if error (e.g., not enough put/get cells)
//
static INT_PTR fm_quickFillArray(const INT_PTR x, const INT_PTR y, const INT_PTR z,
								 const INT_PTR nr, const INT_PTR radius,
								 const INT_PTR xstart, const INT_PTR ystart,
								 const INT_PTR xend,   const INT_PTR yend)
{
    INT_PTR i, j;
    INT_PTR yCoord, xCoord;
    int count = 0;
    const INT_PTR n = radius*2 + 1;

	//Compute limits of the padded square
	const INT_PTR xmin = xstart - fmc.needPadding;	//leftmost limit
	const INT_PTR ymin = ystart - fmc.needPadding;	//topmost limit
	const INT_PTR xmax = xend-1 + fmc.needPadding;	//rightmost limit
	const INT_PTR ymax = yend-1 + fmc.needPadding;	//bottom-most limit

	//A few sanity checks...
	if (!IS_ALLOCATED_ARRAY(nr)) return 0;	//error, invalid Array # or unallocated
	if (z >= ArrayZ[nr])		 return 0;	//error, bad z coordinate

    if (n*n > N_CELLS)		return 0;	//error, not enough put/get cells

	for (i = x-radius; i <= x+radius; i++) {	//VC/rel precomputes x+radius
		//xCoord runs from x - radius to x + radius
		xCoord = i;

		//clamp x-coordinate to border of padded tile
		if		(xCoord < xmin) xCoord = xmin;
		else if (xCoord > xmax) xCoord = xmax;

		for (j = y-radius; j <= y+radius; j++) { //VC/rel precomputes y+radius
			//yCoord runs from y - radius to y + radius
			yCoord = j;

			//clamp y-coordinate to border of padded tile
			if		(yCoord < ymin) yCoord = ymin;
			else if (yCoord > ymax) yCoord = ymax;

			//////////////////////////////
			if (ArrayBytes[nr] == 1)			//bit depth 8
				fmc.cell[count++] = ((unsigned8 *)Array[nr])
										[((yCoord+fmc.needPadding) *ArrayX[nr] +
										  (xCoord+fmc.needPadding))*ArrayZ[nr] + z];
			else if (ArrayBytes[nr] == 2)		//bit depth 16
				fmc.cell[count++] = *(unsigned16 *)(((unsigned8 *)Array[nr]) +
										(((yCoord+fmc.needPadding) *ArrayX[nr] +
										  (xCoord+fmc.needPadding))*ArrayZ[nr] + z)
																   *ArrayBytes[nr] );
			else if (ArrayBytes[nr] == 4)		//bit depth 32
				fmc.cell[count++] = *(unsigned32 *)(((unsigned8 *)Array[nr]) +
										(((yCoord+fmc.needPadding) *ArrayX[nr] +
										  (xCoord+fmc.needPadding))*ArrayZ[nr] + z)
																   *ArrayBytes[nr] );
			//////////////////////////////
		}//for j
	}//for i

	return 1;

} /*fm_quickFillArray*/

//Added by Harald Heim, May 17, 2003
//
//Seems to be a special-purpose routine by Harald to quickly load
//a square region from a designated buffer into the put/get cells,
//possibly as a prelude to runnng quickSort or quickMedian. Since
//I don't fully understand the purpose of this routine, I will
//restrain myself to performing only a few of the more obvious
//speedups (afterall, this routine is eponymously "quick"); and
//refrain from adding any new functionality. -Alex Hunter 8/23/13
//
// quickFill(x, y, z, buffer, radius, xstart, ystart, xend, yend)
//
// where:	(x, y, z) are the coordinates of the center of the square
//			buffer designates the buffer containing the sqfuare:
//				0 = src (image input buffer)
//				1 = tbuf (temp buffer 1)
//				2 = t2buf (temp buffer 2)
//				3 = output buffer
//				4 = (reserved for future use to mean tbuf3)
//				5 = (reserved for future use to mean tbuf4)
//				6 - 9 are unused
//				10 - 109 = use Array[buffer-10]
//			radius = distance from center to nearest points on top, bottom, left,
//					 and right edges of the square region
//			xstart = leftmost valid pix-y x-coordinate in buffer tile
//			ystart = topmost valid pixel -ycoordinate in buffer tile
//			xend   = rightmost valid pixel x-coordinate + 1 in tile
//			yend   = bottom-most valid pixel y-coordinate + 1 in tile
//	Returns 1 if successful, or 0 if error (e.g., not enough put/get cells)
//	Notes:
//	1)	Fills cells with 0 if an unused buffer index is specified. No
//		longer true, just returns 0 in that case!!
//	2)	Does not currently support temporary buffers 3 and 4 (could add
//		later).
//	3)	Originally supported only the first 10 Arrays (even though 100
//		Arrays are allowed). Now supports all Arrays 0 thru N_ARRAYS-1.
//	4)	Clamps the x and  y coordinates to the borders of the input image
//		when using the input, output, or temp buffers (to avoid a #GP??).
//	5)	Additionally, always clamps x and y coordinates to the (possibly
//		padded) borders of the current tile (defined by [xstart,xend) and
//		[ystart,yend) +/- any padding).
//	6)	The radius extends equally in all 4 directions from the center
//		of the square, so the 'diameter' of the square is n = radius*2 + 1.
//	7)	The current maximum value of n is 32, which is the square rooot
//		of N_CELLS (1024); hence the maximum value of the radius is
//		(32 - 1)/2, or 15 pixels for now.
//	8)	For the input image buffer (0), both 8-bit and 16-bit pixel
//		depths are supported.
//	9)	For the temporary (1, 2) and output buffers (3), only 8-bit
//		depth is supported.
//	10)	For the Arrays, depths of 8-, 16-, and 32-bits are supported,
//		depending on the ArrayBytes setting (1, 2, or 4, resp.). The
//		depth of 64-bits (ArrayBytes = 8) is not yet supported, since
//		the put/get cells currently hold only 32-bit values.
// Changes:
//	08/23/2013 Alex Hunter 1. Fixed missing break in case 3 (gave wrong
//							  values when reading from output buffer).
//
static INT_PTR fm_quickFill(const INT_PTR x, const INT_PTR y, const INT_PTR z,
							const INT_PTR buffer, const INT_PTR radius,
							const INT_PTR xstart, const INT_PTR ystart,
							const INT_PTR xend,   const INT_PTR yend)
{
  if (buffer >= 10) {
	//Must be an Array index + 10;
	return fm_quickFillArray(x, y, z, buffer - 10, radius, xstart, ystart, xend, yend);
  }

  {//new scope...
    INT_PTR i, j;
    INT_PTR yCoord, xCoord;
    int count = 0;
    const INT_PTR n = radius*2 + 1;

	//Compute limits of the padded square and/or entire image [sic...]
	const INT_PTR xmin = max(0, xstart - fmc.needPadding); 			//leftmost limit
	const INT_PTR ymin = max(0, ystart - fmc.needPadding);			//topmost limit
	const INT_PTR xmax = min(fmc.X, xend + fmc.needPadding) - 1;	//rightmost limit
	const INT_PTR ymax = min(fmc.Y, yend + fmc.needPadding) - 1;	//bottom-most limit

    if (n*n > N_CELLS) return 0; //error, not enough put/get cells

	//In the following loops, it would increase locality to
	//have the outer loop run through y-coordinates, while
	//the inner loop runs through x-coordinates.
	//But the square region is quite small in this case, so
	//it probably doesn't really matter...

	switch (buffer){
		case 0: //input buffer
			for (i = x-radius; i <= x+radius; i++) {	//VC/rel precomputes x+radius
				//xCoord runs from x - radius to x + radius
				xCoord = i;

				//Clamp x-coordinate to border of image and/or padded tile
				//(Needs to be done otherwise Memory Access Error)
				if		(xCoord < xmin) xCoord = xmin;
				else if (xCoord > xmax) xCoord = xmax;

				for (j = y-radius; j <= y+radius; j++) { //VC/rel precomputes y+radius
					//yCoord runs from y - radius to y + radius
					yCoord = j;

					//Clamp y-coordinate to border of image and/or padded tile
					//(Needs to be done otherwise Memory Access Error)
					if		(yCoord < ymin) yCoord = ymin;
					else if (yCoord > ymax) yCoord = ymax;

					//////////////////////////////
					if (gStuff->imageMode < 10){	//bit depth 8
						fmc.cell[count++] = ((unsigned8 *)MyAdjustedInData)
												[(xCoord - fmc.x_start)*fmc.srcColDelta +
												 (yCoord - fmc.y_start)*fmc.srcRowDelta + z];
					} else {						//bit depth 16
						fmc.cell[count++] =  *(unsigned16 *)(((unsigned8 *)MyAdjustedInData) +
												 (((xCoord - fmc.x_start)*fmc.srcColDelta + z) << 1) +
												   (yCoord - fmc.y_start)*fmc.srcRowDelta);
					}
					//////////////////////////////
				}//for j
			}//for i
			break;

		case 1: //temp buffer 1
			for (i = x-radius; i <= x+radius; i++) {	//VC/rel precomputes x+radius, no need to hand-opt
				//xCoord runs from x - radius to x + radius
				xCoord = i;

				//Clamp x-coordinate to border of image and/or padded tile
				//(Needs to be done otherwise Memory Access Error)
				if		(xCoord < xmin) xCoord = xmin;
				else if (xCoord > xmax) xCoord = xmax;

				for (j = y-radius; j <= y+radius; j++) { //VC/rel precomputes y+radius, no need to hand-opt
					//yCoord runs from y - radius to y + radius
					yCoord = j;

					//Clamp y-coordinate to border of image and/or padded tile
					//(Needs to be done otherwise Memory Access Error)
					if		(yCoord < ymin) yCoord = ymin;
					else if (yCoord > ymax) yCoord = ymax;

					//////////////////////////////
					fmc.cell[count++] = fmc.tbuf[( yCoord*fmc.X + xCoord )*fmc.Z + z];
					//////////////////////////////
				}//for j
			}//for i
			break;

		case 2: //temp buffer 2
			for (i = x-radius; i <= x+radius; i++) {	//VC/rel precomputes x+radius, no need to hand-opt
				//xCoord runs from x - radius to x + radius
				xCoord = i;

				//Clamp x-coordinate to border of image and/or padded tile
				//(Needs to be done otherwise Memory Access Error)
				if		(xCoord < xmin) xCoord = xmin;
				else if (xCoord > xmax) xCoord = xmax;

				for (j = y-radius; j <= y+radius; j++) { //VC/rel precomputes x+radius, no need to hand-opt
					//yCoord runs from y - radius to y + radius
					yCoord = j;

					//Clamp y-coordinate to border of image and/or padded tile
					//(Needs to be done otherwise Memory Access Error)
					if		(yCoord < ymin) yCoord = ymin;
					else if (yCoord > ymax) yCoord = ymax;

					//////////////////////////////
					fmc.cell[count++] = fmc.t2buf[( yCoord*fmc.X + xCoord )*fmc.Z + z];
					//////////////////////////////
				}//for j
			}//for i
			break;

		case 3: //output buffer
			for (i = x-radius; i <= x+radius; i++) {	//VC/rel precomputes x+radius, no need to hand-opt
				//xCoord runs from x - radius to x + radius
				xCoord = i;

				//Clamp x-coordinate to border of image and/or padded tile
				//(Needs to be done otherwise Memory Access Error)
				if		(xCoord < xmin) xCoord = xmin;
				else if (xCoord > xmax) xCoord = xmax;

				for (j = y-radius; j <= y+radius; j++) { //VC/rel precomputes y+radius, no need to hand-opt
					//yCoord runs from y - radius to y + radius
					yCoord = j;

					//Clamp y-coordinate to border of image and/or padded tile
					//(Needs to be done otherwise Memory Access Error)
					if		(yCoord < ymin) yCoord = ymin;
					else if (yCoord > ymax) yCoord = ymax;

					//////////////////////////////
					fmc.cell[count++] = ((unsigned8 *)gStuff->outData)
											[(xCoord - fmc.x_start)*fmc.dstColDelta +
											 (yCoord - fmc.y_start)*fmc.dstRowDelta + z];
					//////////////////////////////
				}//for j
			}//for i
			break;

		default:
			return 0;	//error, not a valid buffer #
	}//switch

	return 1;	//success

  }//end scope
} /*fm_quickFill*/



/*
 *  This Quickselect routine is based on the algorithm described in
 *  "Numerical recipes in C", Second Edition,
 *  Cambridge University Press, 1992, Section 8.5, ISBN 0-521-43108-5
 *  This code by Nicolas Devillard - 1998. Public domain.
 */ 

#define ELEM_SWAP(a,b) { register int t=(a);(a)=(b);(b)=t; }


static INT_PTR fm_quickMedian(INT_PTR low, INT_PTR high)//(elem_type arr[], int n) 
{
    INT_PTR median = (low + high) / 2;
    INT_PTR middle, ll, hh;

    ///if (low<0 || low> N_CELLS-1 || high<0 || high> N_CELLS-1 || low > high || low == high) return 0;
    if (!IS_OK_CELL_NR(low) || !IS_OK_CELL_NR(high) || low >= high) return 0;


    for (;;) {
        if (high <= low) /* One element only */
            return fmc.cell[median] ;

        if (high == low + 1) {  /* Two elements only */
            if (fmc.cell[low] > fmc.cell[high])
                ELEM_SWAP(fmc.cell[low], fmc.cell[high]) ;
            return fmc.cell[median] ;
        }

        /* Find median of low, middle and high items; swap into position low */
        middle = (low + high) / 2;
        if (fmc.cell[middle] > fmc.cell[high])    ELEM_SWAP(fmc.cell[middle], fmc.cell[high]) ;
        if (fmc.cell[low] > fmc.cell[high])       ELEM_SWAP(fmc.cell[low], fmc.cell[high]) ;
        if (fmc.cell[middle] > fmc.cell[low])     ELEM_SWAP(fmc.cell[middle], fmc.cell[low]) ;

        /* Swap low item (now in position middle) into position (low+1) */
        ELEM_SWAP(fmc.cell[middle], fmc.cell[low+1]) ;

        /* Nibble from each end towards middle, swapping items when stuck */
        ll = low + 1;
        hh = high;
        for (;;) {
            do ll++; while (fmc.cell[low] > fmc.cell[ll]) ;
            do hh--; while (fmc.cell[hh]  > fmc.cell[low]) ;

            if (hh < ll)
            break;

            ELEM_SWAP(fmc.cell[ll], fmc.cell[hh]) ;
        }

        /* Swap middle item (in position low) back into correct position */
        ELEM_SWAP(fmc.cell[low], fmc.cell[hh]) ;

        /* Re-set active partition */
        if (hh <= median)
            low = ll;
        if (hh >= median)
            high = hh - 1;
    }

} /*fm_quickMedian*/
#undef ELEM_SWAP



INT_PTR fm_getCtlText(INT_PTR n)
{
	static char dummystr[20];

    if (!IS_INUSE_CTL(n)) {        
			strcpy(dummystr, "Not Available"); //Return this string to avoid memory access violation
			// Does the dummystr string need to be in RAM, or could it be in readonly memory, so we
			// could delete dummystr and simply use:
			//	return "Not Available";  //<== which might be used in common elsewhere
			return (INT_PTR)dummystr;
    }

	return (INT_PTR)gParams->ctl[n].label;

} /*fm_getCtlText*/



INT_PTR fm_getCtlClass(INT_PTR n)
{

	if (n == CTL_PREVIEW)

		return CC_PREVIEW;
	
	else if (n == CTL_PROGRESS)

		return CC_PROGRESSBAR;

	else if (n == CTL_ZOOM)

		return CC_ZOOM;

	else if (IS_INUSE_CTL(n)){
        
        return gParams->ctl[n].class;
    }

    return 0;  

} /*fm_getCtlClass*/



INT_PTR fm_getCtlItemCount(INT_PTR n)
{

	if (!IS_INUSE_CTL(n)) {
        return FALSE;
    }

	if (gParams->ctl[n].class == CC_LISTBOX)
		return (INT_PTR) SendMessage(gParams->ctl[n].hCtl, LB_GETCOUNT, 0, 0);
	else if (gParams->ctl[n].class == CC_COMBOBOX)
		return (INT_PTR) SendMessage(gParams->ctl[n].hCtl, CB_GETCOUNT, 0, 0);
	else if (gParams->ctl[n].class == CC_TAB)
#if 0   //following are identical... Except TabCtrl_GetItemCount casts the
		//result of SendMessage to (int), which might matter someday...
		return (INT_PTR) SendMessage(gParams->ctl[n].hCtl, TCM_GETITEMCOUNT, 0, 0);
#else
		return TabCtrl_GetItemCount(gParams->ctl[n].hCtl);
#endif
	else 
		return -1;

}

INT_PTR fm_getCtlItemTop(INT_PTR n)
{

	if (!IS_INUSE_CTL(n)) {
        return FALSE;
    }

	if (gParams->ctl[n].class == CC_LISTBOX)
		return (INT_PTR) SendMessage(gParams->ctl[n].hCtl, LB_GETTOPINDEX, 0, 0);
	else if (gParams->ctl[n].class == CC_COMBOBOX)
		return (INT_PTR) SendMessage(gParams->ctl[n].hCtl, CB_GETTOPINDEX, 0, 0);
	else 
		return -1;

}

INT_PTR fm_setCtlItemTop(INT_PTR n, INT_PTR item)
{
	INT_PTR retval;

	if (!IS_INUSE_CTL(n)) {
        return FALSE;
    }

	if (gParams->ctl[n].class == CC_LISTBOX)
		retval = SendMessage(gParams->ctl[n].hCtl, LB_SETTOPINDEX, item, 0);
	else if (gParams->ctl[n].class == CC_COMBOBOX)
		retval = SendMessage(gParams->ctl[n].hCtl, CB_SETTOPINDEX, item, 0);
	else
		return false;

	if (retval == LB_ERR) //|| retval == CB_ERR
		return false;
	else
		return true;

}

INT_PTR fm_getCtlItemText(INT_PTR n, INT_PTR item, INT_PTR str)
{
	INT_PTR retval;

	if (!IS_INUSE_CTL(n)) {
        return FALSE;
    }

	if (gParams->ctl[n].class == CC_LISTBOX)
		retval = (INT_PTR) SendMessage(gParams->ctl[n].hCtl, LB_GETTEXT, item, str);
	
	else if (gParams->ctl[n].class == CC_COMBOBOX)
		retval = (INT_PTR) SendMessage(gParams->ctl[n].hCtl, CB_GETLBTEXT, item, str);

	else if (gParams->ctl[n].class == CC_TAB){
		TCITEM tabItem; 
		
		tabItem.mask = TCIF_TEXT;
		tabItem.pszText = (char *)str;
		tabItem.cchTextMax = 256;

		if (TabCtrl_GetItem(gParams->ctl[n].hCtl, item, &tabItem)){
			retval = LB_OKAY;
		} else
			retval = LB_ERR;
	}

	if (retval == LB_ERR) //|| retval == CB_ERR
		return false;
	else
		return true;

}
 
INT_PTR fm_setCtlItemText(INT_PTR n, INT_PTR item, INT_PTR str)
{
	
	if (!IS_INUSE_CTL(n)) {
        return FALSE;
    }
	
	if (gParams->ctl[n].class == CC_LISTBOX)
		return (INT_PTR) SendMessage(gParams->ctl[n].hCtl, LB_INSERTSTRING, item, str);
	
	else if (gParams->ctl[n].class == CC_COMBOBOX)
		return (INT_PTR) SendMessage(gParams->ctl[n].hCtl, CB_INSERTSTRING, item, str);
	
	else if (gParams->ctl[n].class == CC_TAB){

		INT_PTR tabCount = TabCtrl_GetItemCount(gParams->ctl[n].hCtl);
		TCITEM tabItem; 
		
		tabItem.mask = TCIF_TEXT | TCIF_IMAGE; 
		tabItem.iImage = -1; 
		tabItem.pszText = (char *)str; 
		
		if (item<0) item = tabCount+1;
		
		if (item>tabCount)
			return (INT_PTR)TabCtrl_InsertItem(gParams->ctl[n].hCtl, item, &tabItem);
		else
			return (INT_PTR)TabCtrl_SetItem(gParams->ctl[n].hCtl, item, &tabItem);


	}
	
	return -1;

}

INT_PTR fm_deleteCtlItem(INT_PTR n, INT_PTR item)
{

	if (!IS_INUSE_CTL(n)) {
        return FALSE;
    }

	if (gParams->ctl[n].class == CC_LISTBOX)
		return (INT_PTR) SendMessage(gParams->ctl[n].hCtl, LB_DELETESTRING, item, 0);
	
	else if (gParams->ctl[n].class == CC_COMBOBOX)
		return (INT_PTR) SendMessage(gParams->ctl[n].hCtl, CB_DELETESTRING, item, 0);
	
	else if (gParams->ctl[n].class == CC_TAB)
		return (INT_PTR)TabCtrl_DeleteItem(gParams->ctl[n].hCtl, item);

	return false;

}

INT_PTR fm_deleteCtlItems(INT_PTR n)
{
	if (!IS_INUSE_CTL(n)) {
        return FALSE;
    }
	
	if (gParams->ctl[n].class == CC_TAB){
		return (INT_PTR)TabCtrl_DeleteAllItems (gParams->ctl[n].hCtl);

	} else if (gParams->ctl[n].class == CC_LISTBOX || gParams->ctl[n].class == CC_COMBOBOX) {
		return fm_setCtlText(n,(char)"");
	}
	
	return false;
}


INT_PTR fm_fillDir(INT_PTR n, INT_PTR attr, INT_PTR wildcard)
{
	if (!IS_INUSE_CTL(n)) {
        return FALSE;
    }

	if (gParams->ctl[n].class == CC_LISTBOX)
		return (INT_PTR) SendMessage(gParams->ctl[n].hCtl, LB_DIR, attr, wildcard);
	else if (gParams->ctl[n].class == CC_COMBOBOX)
		return (INT_PTR) SendMessage(gParams->ctl[n].hCtl, CB_DIR, attr, wildcard);
	else 
		return -1;

}



static INT_PTR fm_getResAddress(INT_PTR restype, INT_PTR resname){

	HRSRC hRes;         // handle/ptr. to res. info.
    INT_PTR cb;         // Size of resource (bytes)
    HRSRC hResLoad;     // handle to loaded resource  
    char *hResPointer;  // pointer to resource data 
    
    
    // Locate the resource. 
    //hRes = FindResource(hDllInstance,MAKEINTRESOURCE(resname),(LPCTSTR)restype);
	hRes = FindResource(hDllInstance,(LPCTSTR)resname,(LPCTSTR)restype);
    if (hRes == NULL) { 
        return FALSE;
    } 

    // Check for size of resource
    if ((cb=SizeofResource(hDllInstance, hRes)) <= 0)
		return FALSE;

	// Load the resource into global memory.
    hResLoad = LoadResource(hDllInstance, hRes); 
    if (hResLoad == NULL) { 
        return FALSE;
    } 

    // Lock the resource into global memory.
    hResPointer = LockResource(hResLoad);
    if (hResPointer == NULL) {
        return FALSE;
    } else 
		return (INT_PTR)hResPointer;

}

static INT_PTR fm_getResSize(INT_PTR restype, INT_PTR resname){

	HRSRC hRes;         // handle/ptr. to res. info.
    INT_PTR cb;         // Size of resource (bytes)
    
    // Locate the resource.
    //hRes = FindResource(hDllInstance,MAKEINTRESOURCE(resname),(LPCTSTR)restype);
	hRes = FindResource(hDllInstance,(LPCTSTR)resname,(LPCTSTR)restype);
    if (hRes == NULL) { 
        return FALSE;
    } 

	// Check for size of resource and allocate array
    if ((cb=SizeofResource(hDllInstance, hRes)) > 0)
		return (INT_PTR)cb;
	else 
		return 0;

}

static INT_PTR fm_copyResToArray (INT_PTR restype, INT_PTR resname, INT_PTR arraynr)
{
    HRSRC hRes;         // handle/ptr. to res. info.
    int cb;             // Size of resource (bytes)
    HRSRC hResLoad;     // handle to loaded resource  
    char *hResPointer;  // pointer to resource data 
    
    
    // Locate the resource. 
    //hRes = FindResource(hDllInstance,(LPCTSTR)resname,(LPCTSTR)restype);
    //hRes = FindResource(hDllInstance,MAKEINTRESOURCE(resname),(LPCTSTR)restype);
	hRes = FindResource(hDllInstance,(LPCTSTR)resname,(LPCTSTR)restype);
    if (hRes == NULL) { 
        
        //Instead try to find a file with the same name as resname
        FILE *file;
        int cb;
        
        file = fopen( (LPCTSTR)resname, "r" );
        if (file==NULL) return FALSE;

        // Determine file length...
		if (fseek(file, 0, SEEK_END) != 0) {
			//fseek error
			fclose(file);
			return FALSE;	//error
		}
        cb = ftell(file);	//get size of file
		if (cb <= 0) {
			//-1 if error, 0 if empty
			fclose(file);
			return FALSE;	//error
		}
        rewind(file);

		if (fm_allocArray (arraynr,cb,0,0,1)) {
			//read file contents into Array
            size_t cb2 = fread((void *)fm_getArrayAddress(arraynr), 1, cb, file);
			if (cb2 != cb) {
				//error reading file
				fclose(file);
				return FALSE;	//error
			}
		}
		else {
			//failed to allocate array
			fclose(file);
			return FALSE;	//error
		}

        fclose(file);
        return true;	//success

    } 

    // Check for size of resource and allocate Array
    if ((cb=SizeofResource(hDllInstance, hRes)) > 0)
    {
        if (!fm_allocArray (arraynr,cb,0,0,1))
			return FALSE;	//error allocating Array
    } else 
        return FALSE;	//error


    // Load the resource into global memory.
    hResLoad = LoadResource(hDllInstance, hRes); 
    if (hResLoad == NULL) { 
        return FALSE;	//error
    } 

    // Lock the resource into global memory.
    hResPointer = LockResource(hResLoad);
    if (hResPointer == NULL) {
        return FALSE;	//error
    }

    //Copy the resource data into the Array
    memcpy (Array[arraynr],hResPointer,cb);


    return TRUE;	//success
}




static INT_PTR fm_setDialogShowState(INT_PTR state)
{
    
    return ShowWindow (fmc.hDlg, (int)state);

}


static INT_PTR fm_setDialogSizeGrip(INT_PTR state)
{
    
    if (sizeGrip != (int)state){
        sizeGrip = (int)state;
        if (sizeGrip == 0) fm_refreshWindow(); //Make sure the grip vanishes
    }

    return true;

}


INT_PTR fm_PixelsToHDBUs(INT_PTR h) 
{
    int buX;
    //int buY;
    //TEXTMETRIC tm;
    SIZE  size;
    HDC hdc = GetDC(fmc.hDlg);
    HFONT hDlgFont = (HFONT) SendMessage(fmc.hDlg, WM_GETFONT, 0, 0L);
    HFONT hFontOld = SelectObject(hdc,hDlgFont);

    //GetTextMetrics(hdc,&tm);
    //buY = (WORD)tm.tmHeight;

    GetTextExtentPoint32(hdc,"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrst"
                "uvwxyz",52,&size);
    buX = (size.cx/26+1)/2;

    SelectObject(hdc,hFontOld);
    ReleaseDC(fmc.hDlg, hdc);
    
    //Convert to DBU
    return (h * 4) / buX;

}


INT_PTR fm_PixelsToVDBUs(INT_PTR v) 
{
    //int buX;
    int buY;
    TEXTMETRIC tm;
    //SIZE  size;
    HDC hdc = GetDC(fmc.hDlg);
    HFONT hDlgFont = (HFONT) SendMessage(fmc.hDlg, WM_GETFONT, 0, 0L);
    HFONT hFontOld = SelectObject(hdc,hDlgFont);

    GetTextMetrics(hdc,&tm);
    buY = (WORD)tm.tmHeight;

    //GetTextExtentPoint32(hdc,"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrst"
    //            "uvwxyz",52,&size);
    //buX = (size.cx/26+1)/2;

    SelectObject(hdc,hFontOld);
    ReleaseDC(fmc.hDlg, hdc);


    //Convert to DBU
    return (v * 8) / buY;

}


INT_PTR fm_setCtlTab(INT_PTR n, INT_PTR t, INT_PTR s)
{
	if (!IS_INUSE_CTL(n)) {
        return FALSE; //not a valid control
    }
    
    if (t == -1) {
        // disassociate this control from any Tab control
	    gParams->ctl[n].tabControl = (int)t;
	    gParams->ctl[n].tabSheet = (int)s;
        //fm_setCtlColor(n, -1); //set background color back to transparent???
        enableCtl((int)n); //automatically enable or disable the control
        return true;
    }

	if (!IS_INUSE_CTL(t) || gParams->ctl[t].class != CC_TAB) {
        return FALSE; // t is not a valid Tab control
    }
	
	gParams->ctl[n].tabControl = (int)t;
	gParams->ctl[n].tabSheet = (int)s;
    
	//Change background color to tab color if theme is active
	if (gParams->ctl[n].class == CC_CHECKBOX || gParams->ctl[n].class == CC_RADIOBUTTON || gParams->ctl[n].class == CC_OWNERDRAW) {
		if (appTheme == -1)
			appTheme = (int)fm_getAppTheme();
		if ( appTheme && (gParams->ctl[t].theme==1  || (gParams->ctl[t].theme==-1 && gParams->gDialogTheme==1)) )
			fm_setCtlColor(n,fm_getCtlColor(t));
	}

	//Info ("%d,%d\n%d, %d",t,s,gParams->ctl[t].val,gParams->ctl[t].enabled);

	//Automatically enable or disable control
	enableCtl ((int)n);	//MWVDL 2008-06-17

	return true;

}

INT_PTR fm_getCtlTab(INT_PTR n, INT_PTR t){

    if (!IS_OK_CTL_NR(n)) return -1;

    if (t==0) 
        return gParams->ctl[n].tabControl;
    else 
        return gParams->ctl[n].tabSheet;

}


INT_PTR fm_setCtlAnchor(INT_PTR n, INT_PTR flags)
{
	INT_PTR dialogWidth,dialogHeight;
	RECT rcParent;
	
	if ( !IS_INUSE_CTL(n) && n != CTL_PREVIEW && n != CTL_PROGRESS && n != CTL_ZOOM ) {
        return FALSE;
    }

	gParams->ctl[n].anchor = (int)flags;

	if (flags==0) return true;


	GetClientRect(fmc.hDlg, &rcParent);
	dialogWidth = fm_PixelsToHDBUs(rcParent.right);
	dialogHeight = fm_PixelsToVDBUs(rcParent.bottom);


	if (n == CTL_PREVIEW || n == CTL_PROGRESS || n == CTL_ZOOM){
		RECT rcCtl;
    
        if (n == CTL_PREVIEW)
			GetWindowRect( GetDlgItem(fmc.hDlg, 101), &rcCtl );
		else if (n == CTL_PROGRESS)
			GetWindowRect( GetDlgItem(fmc.hDlg, IDC_PROGRESS1), &rcCtl );
		else if (n == CTL_ZOOM)
			GetWindowRect( GetDlgItem(fmc.hDlg, IDC_BUTTON1), &rcCtl );
		
		ScreenToClient (fmc.hDlg, (LPPOINT)&rcCtl);
		ScreenToClient (fmc.hDlg, (LPPOINT)&(rcCtl.right));

		gParams->ctl[n].distWidth = (int)(dialogWidth - fm_PixelsToHDBUs(rcCtl.right));
		gParams->ctl[n].distHeight = (int)(dialogHeight - fm_PixelsToVDBUs(rcCtl.bottom));
	

		//Make sure the original coordinates are set
		//if (gParams->ctl[n].xPos == -1 || gParams->ctl[n].yPos == -1 || gParams->ctl[n].width == -1
		//	|| gParams->ctl[n].height == -1){

			gParams->ctl[n].width = (int)fm_PixelsToHDBUs(rcCtl.right - rcCtl.left);
			gParams->ctl[n].height = (int)fm_PixelsToVDBUs(rcCtl.bottom - rcCtl.top);

			/*if (n == CTL_ZOOM){ //Get new top and left for zoom control otherwise zoom control jumps
				GetWindowRect( GetDlgItem(fmc.hDlg, IDC_BUTTON2), &rcCtl );
				ScreenToClient (fmc.hDlg, (LPPOINT)&rcCtl);
				ScreenToClient (fmc.hDlg, (LPPOINT)&(rcCtl.right));
			}*/

			gParams->ctl[n].xPos = (int)fm_PixelsToHDBUs(rcCtl.left);
			gParams->ctl[n].yPos = (int)fm_PixelsToVDBUs(rcCtl.top);
		//}


	} else {
		gParams->ctl[n].distWidth = (int)dialogWidth - (gParams->ctl[n].xPos + gParams->ctl[n].width);
		gParams->ctl[n].distHeight = (int)dialogHeight - (gParams->ctl[n].yPos + gParams->ctl[n].height);
	}


	return true;
}


INT_PTR fm_updateAnchors(INT_PTR dialogWidth, INT_PTR dialogHeight){

    int n;
    
    //Get dialog size if necessary
    if (dialogWidth == -1 || dialogHeight == -1){
        RECT rcParent;
        GetClientRect(fmc.hDlg, &rcParent);
        if (dialogWidth == -1) dialogWidth = fm_PixelsToHDBUs(rcParent.right);
	    if (dialogHeight == -1) dialogHeight = fm_PixelsToVDBUs(rcParent.bottom);
    }


    for (n = 0;  n < N_CTLS;  n++)
    {
		if ( (gParams->ctl[n].inuse || n==CTL_PREVIEW || n==CTL_PROGRESS || n==CTL_ZOOM) && 
		  gParams->ctl[n].anchor > 0){


	        if (n == CTL_PREVIEW || n == CTL_PROGRESS || n == CTL_ZOOM){
		        RECT rcCtl;
    
                if (n == CTL_PREVIEW)
			        GetWindowRect( GetDlgItem(fmc.hDlg, 101), &rcCtl );
		        else if (n == CTL_PROGRESS)
			        GetWindowRect( GetDlgItem(fmc.hDlg, IDC_PROGRESS1), &rcCtl );
		        else if (n == CTL_ZOOM)
			        GetWindowRect( GetDlgItem(fmc.hDlg, IDC_BUTTON1), &rcCtl );
		        
		        ScreenToClient (fmc.hDlg, (LPPOINT)&rcCtl);
		        ScreenToClient (fmc.hDlg, (LPPOINT)&(rcCtl.right));

		        gParams->ctl[n].distWidth = (int)(dialogWidth - fm_PixelsToHDBUs(rcCtl.right));
		        gParams->ctl[n].distHeight = (int)(dialogHeight - fm_PixelsToVDBUs(rcCtl.bottom));

	        } else {
		        gParams->ctl[n].distWidth = (int)dialogWidth - (gParams->ctl[n].xPos + gParams->ctl[n].width);
		        gParams->ctl[n].distHeight = (int)dialogHeight - (gParams->ctl[n].yPos + gParams->ctl[n].height);
	        }
		}
	}

    return true;

}


INT_PTR fm_lockCtlScaling(INT_PTR lock){

    scaleLock = (BOOL)lock;

    return true;
}


INT_PTR fm_scaleCtls(INT_PTR dialogWidth, 
                 INT_PTR dialogHeight){

    int n;

    //Get dialog size if necessary
    if (dialogWidth == -1 || dialogHeight == -1){
        RECT rcParent;
        GetClientRect(fmc.hDlg, &rcParent);
        if (dialogWidth == -1) dialogWidth = fm_PixelsToHDBUs(rcParent.right);
	    if (dialogHeight == -1) dialogHeight = fm_PixelsToVDBUs(rcParent.bottom);
    }

    for (n = 0;  n < N_CTLS;  n++)
    {
		if ( (gParams->ctl[n].inuse || n==CTL_PREVIEW || n==CTL_PROGRESS || n==CTL_ZOOM) && 
		  gParams->ctl[n].anchor > 0){
			int newx = gParams->ctl[n].xPos;
			int newy = gParams->ctl[n].yPos;
			int neww = gParams->ctl[n].width;
			int newh = gParams->ctl[n].height;

			//Scale horizontal
			if ( (gParams->ctl[n].anchor & 1) && (gParams->ctl[n].anchor & 2)) 
				neww = neww + (((int)dialogWidth - (gParams->ctl[n].xPos + neww)) - gParams->ctl[n].distWidth);
			//Move horizontal
			else if ( !(gParams->ctl[n].anchor & 1) && (gParams->ctl[n].anchor & 2)) 
				newx  = newx + (((int)dialogWidth - (gParams->ctl[n].xPos + neww)) - gParams->ctl[n].distWidth);

			//Scale vertical
			if ( (gParams->ctl[n].anchor & 4) && (gParams->ctl[n].anchor & 8)) 
				newh = newh + (((int)dialogHeight - (gParams->ctl[n].yPos + newh)) - gParams->ctl[n].distHeight);
			//Move vertical
			else if ( !(gParams->ctl[n].anchor & 4) && (gParams->ctl[n].anchor & 8))
				newy  = newy + (((int)dialogHeight - (gParams->ctl[n].yPos + newh)) - gParams->ctl[n].distHeight);
			
			//Do not make preview smaller to avoid crash
			if (n==CTL_PREVIEW){
				if (neww<10) neww=10;
				if (newh<10) newh=10;
			}

			fm_setCtlPos(n,newx,newy,neww,newh);
		}
	}

    return true;

}






/*
HDC PixelhDC=0;
HDC BufferhDC=0;
HBITMAP hbmBuffer=0;
HFONT fontDB=0;


static int fm_startSetPixel (int n)
{
        //RECT rcCtl;
        //GetWindowRect(gParams->ctl[n].hCtl, &rcCtl);

        if (n==-1)
            PixelhDC = GetDC(fmc.hDlg);
        else
            PixelhDC = GetDC(gParams->ctl[n].hCtl);
        
        //BufferhDC = CreateCompatibleDC(PixelhDC);
        //hbmBuffer = CreateCompatibleBitmap(BufferhDC, rcCtl.right-rcCtl.left, rcCtl.bottom-rcCtl.top);
        //SelectObject(BufferhDC, hbmBuffer);

        return true;

}

static int fm_endSetPixel (int n)
{
        //RECT rcCtl;
        //GetWindowRect(gParams->ctl[n].hCtl, &rcCtl);

        //BitBlt(PixelhDC, 0, 0, rcCtl.right-rcCtl.left, rcCtl.bottom-rcCtl.top, BufferhDC, 0, 0, SRCCOPY);
        //DeleteObject (hbmBuffer);
        //DeleteDC (BufferhDC);

        if (n==-1)
            ReleaseDC (fmc.hDlg, PixelhDC);
        else    
            ReleaseDC (gParams->ctl[n].hCtl, PixelhDC);
        
        return true;
        
}

static int fm_setPixel (int x, int y, int value)
{   
            
        if (PixelhDC==0) return false;

        return SetPixelV (PixelhDC,x,y,(COLORREF) value);
        //return SetPixelV (BufferhDC,x,y,(COLORREF) value);

}
*/



static INT_PTR fm_countProcessors(void){

	SYSTEM_INFO SystemInfo;
	
	GetSystemInfo(&SystemInfo);
 
	return (INT_PTR)SystemInfo.dwNumberOfProcessors;
}


/****** THREAD APIs ******/

INT_PTR fm_triggerThread(INT_PTR n, INT_PTR event, INT_PTR previous ){

    HANDLE hThread;
    DWORD dwThread;
    FMcontext *pfmc_copy;

    if ( !IS_OK_CTL_NR(n) && event != FME_CUSTOMEVENT ) return false;

    pfmc_copy = malloc(sizeof(FMcontext));		//X64???
    if (!pfmc_copy) return false;   //couldn't allocate FMC as TLS
	

    //could do this in the thread, so it can be multitasked? 
    //but then it's hard to pass other params	
#if 0
    memcpy(pfmc_copy,&fmc,sizeof(FMcontext)); //Copy values of original FMC
#else
    *pfmc_copy = fmc; //Copy values of original FMC 
#endif

    //Set new values
    pfmc_copy->n = (int)n;
    pfmc_copy->e = (int)event;
    pfmc_copy->previous = (int)previous;
    pfmc_copy->ctlMsg = (int)event;
    pfmc_copy->ctlCode = 0;

    if (event==FME_CUSTOMEVENT)
    	pfmc_copy->ctlClass = CC_UNUSED;
    else
        pfmc_copy->ctlClass = gParams->ctl[n].class;

    
	/************ CAVEATS FROM MSDN: Creating Threads ****************************

	The ThreadProc function avoids the use of the C run-time library (CRT), 
	as many of its functions are not thread-safe, particularly if you are not 
	using the multithreaded CRT. If you would like to use the CRT in ThreadProc, 
	use the _beginthreadex function instead.
	(Should we be calling _beginthreadex instead of CreateThread?)

	It is risky to pass the address of a local variable if the creating thread 
	exits before the new thread, because the pointer becomes invalid. Instead, 
	either pass a pointer to dynamically allocated memory or make the creating 
	thread wait for the new thread to terminate. Data can also be passed from 
	the creating thread to the new thread using global variables. With global 
	variables, it is usually necessary to synchronize access by multiple threads
	(We are okay here, since we are passing the address of malloc'ed memory
	via the contents of pfmc_copy.)

	(Should we call a stub which calls X86_call_fmf1() to try/catch exceptions?)

	*****************************************************************************/
    //Do call
    //RetVal = X86_call_fmf1(gParams->CB_onCtl, (INT_PTR)&fmc);
    hThread = CreateThread( NULL,
                            0, 
                            (LPTHREAD_START_ROUTINE)gParams->CB_onCtl, //or call stub, which then calls CB_onCtl!
                            (LPVOID)pfmc_copy, 
                            0, 
                            &dwThread);

	if (hThread) { //Success
		ThreadCount++;
		ThreadHandle = realloc(ThreadHandle, ThreadCount*sizeof(*ThreadHandle));	//X64???
		ThreadHandle[ThreadCount-1]	= (INT_PTR)hThread;
		ThreadMem = realloc(ThreadMem, ThreadCount*sizeof(*ThreadMem));	//X64???
		ThreadMem[ThreadCount-1] = (INT_PTR)pfmc_copy;
		return (INT_PTR)hThread;
	} else //Failure
		free(pfmc_copy);


    return false;
}


INT_PTR fm_waitForThread(INT_PTR hThread, int ms, INT_PTR userinput){

	int retval = 0; 
	
	if (ThreadCount==0) return false; // No Threads available

	if (hThread==0){ //Wait for all threads
		
		if (ms==0){

			retval = WaitForMultipleObjects( 
						ThreadCount, 
						(HANDLE *)ThreadHandle, 
						TRUE, 
						0); // 0 => immediate return

		} else {

			MSG msg;
			int count=0;
		
			while (WaitForMultipleObjects( 
						ThreadCount, 
						(HANDLE *)ThreadHandle, 
						TRUE, 
						10) == WAIT_TIMEOUT){ //Use a 10 ms time-out

				if (ms!=INFINITE){
					count = count + 10;
					if (count >= ms) {
						retval = WAIT_TIMEOUT;
						break;
					}
				}
			
                if (userinput != -1){ //Use -1 to deactivate the message pump
				    //Flush key and mouse messages
				    if (userinput==0){
					    while( PeekMessage(&msg, NULL, WM_KEYFIRST, WM_KEYLAST, PM_REMOVE ) );
					    while( PeekMessage(&msg, NULL, WM_MOUSEFIRST, WM_MOUSELAST, PM_REMOVE ) );
				    }

				    //Message Pump
				    while (PeekMessage (&msg, NULL, 0, 0, PM_REMOVE)) {
					    TranslateMessage(&msg);
					    DispatchMessage(&msg);
				    }
                }
                
			}

		}

		if (retval != WAIT_TIMEOUT) {
			int i;

			for (i=0; i<ThreadCount; i++) 
				if (ThreadHandle[i]!=0) CloseHandle((HANDLE)ThreadHandle[i]);

			for (i=0; i<ThreadCount; i++) 
				if (ThreadMem[i]!=0) free ((void *)ThreadMem[i]);

			free(ThreadHandle);
			ThreadHandle = NULL;

			free(ThreadMem);
			ThreadMem = NULL;

			ThreadCount = 0;
		}

	} /*else { //Wait for specific thread					//EH??
		
		retval = WaitForSingleObject( 
						(HANDLE)hThread, 
						(DWORD) ms );
		CloseHandle((HANDLE)hThread);
	}*/	


	if (retval==WAIT_TIMEOUT)
		return false;
	else
		return true;
}


INT_PTR fm_isThreadActive(INT_PTR hThread){

	///LPDWORD lpExitCode=0;	//This was wrong!!!
	DWORD dwExitCode = 0;

	if (ThreadCount==0) return false;

	if (hThread==0){
		int i;
		for (i=0; i<ThreadCount; i++){ 
			if (GetExitCodeThread((HANDLE)ThreadHandle[i], &dwExitCode)){	//???
				if (dwExitCode == STILL_ACTIVE) return true;
			}
		}

		// Cleanup Threads
		for (i=0; i<ThreadCount; i++) 
			if (ThreadHandle[i]!=0) CloseHandle((HANDLE)ThreadHandle[i]);

		for (i=0; i<ThreadCount; i++) 
			if (ThreadMem[i]!=0) free ((void *)ThreadMem[i]);

		free(ThreadHandle);
		ThreadHandle = NULL;

		free(ThreadMem);
		ThreadMem = NULL;

		ThreadCount = 0;

	} else {
		if (GetExitCodeThread((HANDLE)hThread, &dwExitCode)){	
			if (dwExitCode == STILL_ACTIVE) return true;
		}
	}

	return false;
}


INT_PTR fm_getThreadRetVal(INT_PTR hThread){

	DWORD dwExitCode=0;

	if (ThreadCount==0) return false;

	if (hThread!=0){
		if (GetExitCodeThread((HANDLE)hThread,&dwExitCode)){	
			//if (dwExitCode != STILL_ACTIVE) 
				return dwExitCode;
		}
	}

	return 0;

}



INT_PTR fm_terminateThread(INT_PTR hThread){

	if (ThreadCount==0) return true;

	if (hThread==0){
		int i;
				
		for (i=0; i<ThreadCount; i++){ 
			if (ThreadHandle[i]!=0) {
				TerminateThread ((HANDLE)ThreadHandle[i], 0);
				CloseHandle((HANDLE)ThreadHandle[i]);
			}
		}

		for (i=0; i<ThreadCount; i++) 
			if (ThreadMem[i]!=0) free ((void *)ThreadMem[i]);

		free(ThreadHandle);
		ThreadHandle = NULL;

		free(ThreadMem);
		ThreadMem = NULL;

		ThreadCount = 0;
	
	} /*else {										//EH??

		TerminateThread (hThread, 0);
		CloseHandle(hThread);

	}*/

	return true;

}

/***** Critical Section APIs *****/

INT_PTR fm_createCriticalSection(void) {
    //returns handle to CS, or 0 if failed
    LPCRITICAL_SECTION lpcs = malloc(sizeof(CRITICAL_SECTION));
    if (lpcs) {
        // CS was successfully allocated, now initialize it
        InitializeCriticalSection(lpcs);
    }
    return (INT_PTR)lpcs;    //returns 0 if couldn't be allocated
}/*fm_createCriticalSection*/


#ifdef HAVE_XP  //requires XP or later
int fm_createCriticalSectionAndSpinCount(DWORD dwSpinCount) {
    //returns handle to CS, or 0 if failed
    LPCRITICAL_SECTION lpcs = malloc(sizeof(CRITICAL_SECTION));
    if (lpcs) {
        // CS was successfully allocated, now initialize it..
        // Returns 0 if error.
        if (InitializeCriticalSectionAndSpinCount(lpcs, dwSpinCount))
            return (int)lpcs;   //success -- return handle to CS
        else
            return 0;           //failure -- return 0
    }
    return (int)lpcs;    //returns 0 if couldn't be allocated
}/*fm_createCriticalSection*/
#endif //XP


#ifdef HAVE_VISTA //requires Vista or later
int fm_createCriticalSectionEx(DWORD dwSpinCount, DWORD flags) {
    //flags should be 0 or CRITICAL_SECTION_NO_DEBUG_INFO
    //returns handle to CS, or 0 if failed
    LPCRITICAL_SECTION lpcs = malloc(sizeof(CRITICAL_SECTION));
    if (lpcs) {
        // CS was successfully allocated, now initialize it..
        // Returns 0 if error.
        if (InitializeCriticalSectionEx(lpcs, dwSpinCount, flags))
            return (int)lpcs;   //success -- return handle to CS
        else
            return 0;           //failure -- return 0
    }
    return (int)lpcs;    //returns 0 if couldn't be allocated
}/*fm_createCriticalSection*/
#endif //Vista


#ifdef HAVE_XP  //requires XP or later
int fm_setCriticalSectionSpinCount(int hCS, DWORD spinCount) {
    //returns previous spin count, or 0 if hCS is 0
    //if hCS is 0, SetCriticalSectionSpinCount will GP fault!
    if (!hCS) return 0;
    return SetCriticalSectionSpinCount((LPCRITICAL_SECTION)hCS, spinCount);
}/*fm_setCriticalSectionSpinCount*/
#endif //XP


BOOL fm_enterCriticalSection(INT_PTR hCS) {
    //returns true upon successful entry, false if hCS is 0.
    //if hCS is 0, EnterCriticalSection will GP fault!
    if (!hCS) return false;
    EnterCriticalSection((LPCRITICAL_SECTION)hCS);
    return true;
}/*fm_enterCriticalSection*/


#if _WIN32_WINNT >= 0x0501 // Only available on Windows XP & newer 
INT_PTR fm_tryEnterCriticalSection(INT_PTR hCS) {
    //returns true if entered CS, false if not (or hCS is 0)
    //if hCS is 0, TryEnterCriticalSection will GP fault!
    if (!hCS) return false;
    return TryEnterCriticalSection((LPCRITICAL_SECTION)hCS);
}/*fm_tryEnterCriticalSection*/
#endif


BOOL fm_leaveCriticalSection(INT_PTR hCS) {
    //returns true upon successful exit, false if hCS is 0.
    //if hCS is 0, LeaveCriticalSection will GP fault!
    if (!hCS) return false;
    LeaveCriticalSection((LPCRITICAL_SECTION)hCS);
    return true;
}/*fm_leaveCriticalSection*/


BOOL fm_deleteCriticalSection(INT_PTR hCS) {
    //DE-initialize the CS...
    //returns true if successful, false if hCS is 0.
    //if hCS is 0, DeleteCriticalSection will GP fault!
    if (!hCS) return false;
    DeleteCriticalSection((LPCRITICAL_SECTION)hCS);
    //now delete the CS object itself
    free((void *)hCS);
    hCS = 0;    //Mark the CS handle no longer valid
    return true;
}/*fm_deleteCriticalSection*/


/****** Interlocked Variable API's ******/

INT_PTR fm_interlockedIncrement(INT_PTR pvar) {
    // Atomically increments a variable,
    // returning the resulting incremented value.
    // 486+ only? Win98+/NT4.0+?
    return InterlockedIncrement((int *)pvar);
}

INT_PTR fm_interlockedDecrement(INT_PTR pvar) {
    // Atomically decrements a variable,
    // returning the resulting decremented value.
    // 486+ only? Win98+/NT4.0+?
    return InterlockedDecrement((int *)pvar);
}

INT_PTR fm_interlockedExchange(INT_PTR ptarget, INT_PTR value) {
    // Atomically sets the target int variable to the 
    // specified value, and returns the previous value
    // of the target variable.
    return InterlockedExchange((int *)ptarget, (int)value);
}

INT_PTR fm_interlockedExchangeAdd(INT_PTR ptarget, INT_PTR value) {
    // Atomically adds the specified value to 
    // the target int variable, and returns the previous
    // value of the target variable.
    return InterlockedExchangeAdd((int *)ptarget, (int)value);
}

INT_PTR fm_interlockedCompareExchange(INT_PTR pdest, INT_PTR exchange, INT_PTR comparand) {
    // Atomically compares the 4-byte value of the destination variable
    // with the 4-byte comparand value.  If equal, then the exchange
    // value is stored in the destination variable;
    // otherwise the destination variable is not modified.
    // Returns the initial value of the destination variable.
    return InterlockedCompareExchange((PVOID)pdest, (int)exchange, (int)comparand);
}


/****** Thread Synchronization API's ******/

#if 0   //#1: Harry's first version (+ Alex's mods)

int NumberOfThreads=0;
int *lpThreadArray=NULL;
LPCRITICAL_SECTION lpcs1=NULL;


int fm_createSync(int number){

    if (number<2) return false;
        
    NumberOfThreads = number;
    lpThreadArray = calloc(number*2,4);
    if (!lpThreadArray) return false;

    lpcs1 = malloc(sizeof(CRITICAL_SECTION));
    if (!lpcs1) {
        free(lpThreadArray);
        lpThreadArray = NULL;
        return false;
    }
    InitializeCriticalSection(lpcs1);

    return true;
}


int fm_waitForSync(int syncnr, int timeout){

    int val,i, count;
    BOOL success=false;
    int ThreadID = (int)GetCurrentThreadId();
    int timeelapsed=false;
    int starttime = clock(); 
    
    
    if (NumberOfThreads<2 || lpThreadArray==NULL || lpcs1==NULL) return false; //In case Sync was not initialized correctly


    EnterCriticalSection(lpcs1);
        for (i=0;i<NumberOfThreads;i++){

            val = lpThreadArray[i*2];

            if (val==0){
                lpThreadArray[i*2] = ThreadID; 
                lpThreadArray[i*2+1] = syncnr;
                success = true;
                break;
        
            } else if (val==ThreadID){
                lpThreadArray[i*2+1] = syncnr;
                success = true;
                break;
            }

        }
    LeaveCriticalSection(lpcs1);

    if (!success) return false; //because an unknown thread entered waitForSync



    while (true){
        count=0;
        for (i=0;i<NumberOfThreads;i++){
            if (lpThreadArray[i*2]!=0 && lpThreadArray[i*2+1] >= syncnr) count++;
        }

        if (count>=NumberOfThreads){
            timeelapsed=false;
            break;
        }

        if (timeout!=0xffffffff && clock()-starttime >= timeout) {
            timeelapsed=true; 
            break;
        }

        Sleep(0);
    }  


    if (timeelapsed)
        return false;
    else
        return true;
}


int fm_deleteSync(void){

    NumberOfThreads = 0;

    if (lpThreadArray!=NULL){
        free (lpThreadArray);
        lpThreadArray=NULL;
    }

    if (lpcs1 != NULL){
        DeleteCriticalSection(lpcs1);
        free (lpcs1);
        lpcs1=NULL;
    }

    return true;
}

#elif 0 //#2: Tighten up the Critical Section *******************************************

int NumberOfThreads=0;
int *lpThreadArray=NULL;
LPCRITICAL_SECTION lpcs1=NULL;


int fm_createSync(int number){

    if (number<2) return false;
        
    NumberOfThreads = number;
    lpThreadArray = calloc(number*2,4);
    if (!lpThreadArray) return false;

    lpcs1 = malloc(sizeof(CRITICAL_SECTION));
    if (!lpcs1) {
        free(lpThreadArray);
        lpThreadArray = NULL;
        return false;
    }
    InitializeCriticalSection(lpcs1);

    return true;
}


int fm_waitForSync(int syncnr, int timeout){

    int val,i, count;
    BOOL success=false;
    int ThreadID = (int)GetCurrentThreadId();
    int timeelapsed=false;
    int starttime = clock(); 
    
    
    if (NumberOfThreads<2 || lpThreadArray==NULL || lpcs1==NULL) return false; //In case Sync was not initialized correctly


    //search for an empty slot (or one we already own)...

    for (i=0;i<NumberOfThreads;i++){

        //fetching, testing, and claiming slot i
        //must be atomic...
        EnterCriticalSection(lpcs1);        /****** enter CS ******/
        val = lpThreadArray[i*2];
        if (val==0){
            //take ownership of this slot
            lpThreadArray[i*2] = ThreadID;
            //now no one else can claim slot i
            LeaveCriticalSection(lpcs1);    /****** exit CS ******/

            //record our current sync number
            lpThreadArray[i*2+1] = syncnr;
            success = true;
            break;
    
        }
        LeaveCriticalSection(lpcs1);         /****** exit CS ******/

        //someone already owns slot i, but who?
        //(once slot i is claimed, it never gets
        //changed until deletion time)
        if (val==ThreadID){
            //we already own this slot...
            //record our current sync number
            lpThreadArray[i*2+1] = syncnr;
            success = true;
            break;
        }

    }//for i

    if (!success) return false; //because an unknown thread entered waitForSync



    while (true){
        count=0;
        for (i=0;i<NumberOfThreads;i++){
            if (lpThreadArray[i*2]!=0 && lpThreadArray[i*2+1] >= syncnr) count++;
        }

        if (count>=NumberOfThreads){
            timeelapsed=false;
            break;
        }

        if (timeout!=0xffffffff && clock()-starttime >= timeout) {
            timeelapsed=true; 
            break;
        }

        Sleep(0);
    }  


    if (timeelapsed)
        return false;
    else
        return true;
}


int fm_deleteSync(void){

    NumberOfThreads = 0;

    if (lpThreadArray!=NULL){
        free (lpThreadArray);
        lpThreadArray=NULL;
    }

    if (lpcs1 != NULL){
        DeleteCriticalSection(lpcs1);
        free (lpcs1);
        lpcs1=NULL;
    }

    return true;
}

#elif 0 //#3: Using Interlocked Instruction instead of Critical Section ******************************

int NumberOfThreads=0;
int *lpThreadArray=NULL;


int fm_createSync(int number){

    if (number<2) return false;
        
    NumberOfThreads = number;
    lpThreadArray = calloc(number*2,4);
    if (!lpThreadArray) return false;

    return true;
}


int fm_waitForSync(int syncnr, int timeout){

    int val,i, count;
    BOOL success=false;
    int ThreadID = (int)GetCurrentThreadId();
    int timeelapsed=false;
    int starttime = clock(); 
    
    
    if (NumberOfThreads<2 || lpThreadArray==NULL) return false; //In case Sync was not initialized correctly


    //search for an empty slot (or one we already own)...

    for (i=0;i<NumberOfThreads;i++){

        //fetching, testing, and claiming slot i
        //must be atomic...
        val = (int)InterlockedCompareExchange(&lpThreadArray[i*2], ThreadID, 0);
        //val is the *previous* value of this slot
        if (val==0){
            //we took ownership of this slot
            //now no one else can claim slot i
            //record our current sync number
            lpThreadArray[i*2+1] = syncnr;
            success = true;
            break;
    
        }

        //someone already owns slot i, but who?
        //(once slot i is claimed, it never gets
        //changed until deletion time)
        if (val==ThreadID){
            //we already own this slot...
            //record our current sync number
            lpThreadArray[i*2+1] = syncnr;
            success = true;
            break;
        }

    }//for i

    if (!success) return false; //because an unknown thread entered waitForSync



    while (true){
        count=0;
        for (i=0;i<NumberOfThreads;i++){
            if (lpThreadArray[i*2]!=0 && lpThreadArray[i*2+1] >= syncnr) count++;
        }

        if (count>=NumberOfThreads){
            timeelapsed=false;
            break;
        }

        if (timeout!=0xffffffff && clock()-starttime >= timeout) {
            timeelapsed=true; 
            break;
        }

        Sleep(0);
    }  


    if (timeelapsed)
        return false;
    else
        return true;
}


int fm_deleteSync(void){

    NumberOfThreads = 0;

    if (lpThreadArray!=NULL){
        free (lpThreadArray);
        lpThreadArray=NULL;
    }

    return true;
}

#elif 1 //#4: Tighten up the While-loop ******************************

int NumberOfThreads=0;
int *lpThreadArray=NULL;


INT_PTR fm_createSync(INT_PTR number){

    if (number<2) return false;
        
    NumberOfThreads = (int)number;
    lpThreadArray = calloc(number*2,4);
    if (!lpThreadArray) return false;

    return true;
}//createSync


INT_PTR fm_waitForSync(int syncnr, int timeout){

    int val, i;
    BOOL success=false;
    int ThreadID = (int)GetCurrentThreadId();
    BOOL timeelapsed=false;
    int starttime = clock(); 
    
    if (NumberOfThreads<2 || lpThreadArray==NULL) return false; //In case Sync was not initialized correctly

    //search for an empty slot (or one we already own)...

    for (i=0; i<NumberOfThreads; i++){

        //fetching, testing, and claiming slot i
        //must be atomic...
        val = (int)InterlockedCompareExchange(&lpThreadArray[i*2], ThreadID, 0);
        //val is the *previous* value of this slot
        if (val==0){
            //we took ownership of this slot
            //now no one else can claim slot i
            //record our current sync number
            lpThreadArray[i*2+1] = (int)syncnr;
            success = true;
            break;
    
        }

        //someone already owns slot i, but who?
        //(once slot i is claimed, it never gets
        //changed until deletion time)
        if (val==ThreadID){
            //we already own this slot...
            //record our current sync number
            lpThreadArray[i*2+1] = (int)syncnr;
            success = true;
            break;
        }

    }//for i

    if (!success) return false; //because an unknown thread entered waitForSync


    // Wait until all threads are at the current syncnr or greater.

    for (i=0; i<NumberOfThreads && !timeelapsed; i++) {
        //wait until thread i reaches syncnr or greater...
        while(lpThreadArray[i*2]==0 || lpThreadArray[i*2+1] < syncnr) {
            //check for timeout
            if (timeout!=INFINITE && clock()-starttime >= timeout) {
                timeelapsed=true; 
                break;
            }//if timeout
            //yield this timeslice
            Sleep(0);
        }//while
    }//for

    //Either we timed out, or all threads are at current syncnr or greater...
    return !timeelapsed;

}//waitForSync


INT_PTR fm_deleteSync(void){

    NumberOfThreads = 0;

    if (lpThreadArray!=NULL){
        free (lpThreadArray);
        lpThreadArray=NULL;
    }

    return true;
}//deleteSync

#elif 0 //#5: Cache mySlot ******************************

int NumberOfThreads=0;
int *lpThreadArray=NULL;


int fm_createSync(int number){

    if (number<2) return false;
        
    NumberOfThreads = number;
    lpThreadArray = calloc(number*2,4);
    if (!lpThreadArray) return false;

    return true;
}//createSync


int fm_waitForSync(int syncnr, int timeout){

    THREAD_LOCAL static int mySlot = -1;
    int val, i;
    BOOL success=false;
    int ThreadID = (int)GetCurrentThreadId();
    BOOL timeelapsed=false;
    int starttime = clock(); 
    
    if (NumberOfThreads<2 || lpThreadArray==NULL) return false; //In case Sync was not initialized correctly

    //search for an empty slot (if we don't already own one)...

    if (mySlot == -1) {
        //not yet assigned a slot...
        for (i=0; i<NumberOfThreads; i++){
            //fetching, testing, and claiming slot i
            //must be atomic...
            val = (int)InterlockedCompareExchange(&lpThreadArray[i*2], ThreadID, 0);
            //val is the *previous* value of this slot
            if (val==0){
                //we took ownership of this slot
                //now no one else can claim slot i
                //cache our assigned slot#
                mySlot = i;
                success = true;
                break;
            }//if
        }//for i
        if (!success) return false; //because an unknown thread entered waitForSync
    }//if

    lpThreadArray[mySlot*2+1] = syncnr;  //record our current syncnr


    // Wait until all threads are at the current syncnr or greater.

    for (i=0; i<NumberOfThreads && !timeelapsed; i++) {
        //wait until thread i reaches syncnr or greater...
        while(lpThreadArray[i*2]==0 || lpThreadArray[i*2+1] < syncnr) {
            //check for timeout
            if (timeout!=INFINITE && clock()-starttime >= timeout) {
                timeelapsed=true; 
                break;
            }//if timeout
            //yield this timeslice
            Sleep(0);
        }//while
    }//for

    //Either we timed out, or all threads are at current syncnr or greater...
    return !timeelapsed;

}//waitForSync


int fm_deleteSync(void){

    NumberOfThreads = 0;

    if (lpThreadArray!=NULL){
        free (lpThreadArray);
        lpThreadArray=NULL;
    }

    return true;
}//deleteSync

#elif 0 //#6: Assign mySlot directly ******************************

int NumberOfThreads=0;
int *lpThreadArray=NULL;
int NextSlot=0;


int fm_createSync(int number){

    if (number<2) return false;
        
    NumberOfThreads = number;
    NextSlot = 0;
    lpThreadArray = calloc(number,4);
    if (!lpThreadArray) return false;

    return true;
}//createSync


int fm_waitForSync(int syncnr, int timeout){

    THREAD_LOCAL static int mySlot = -1;
    int i;
    BOOL timeelapsed=false;
    int starttime = clock(); 
    
    if (NumberOfThreads<2 || lpThreadArray==NULL) return false; //In case Sync was not initialized correctly
    if (syncnr < 1) return false; //syncnr must be 1 or greater

    //first time around, assign next slot in ThreadArray to this thread...

    if (mySlot == -1) {
        //not yet assigned a slot...
#if 0 //InterlockedIncrement requires Win98+ or NT 4.0+ for proper return value
        mySlot = InterlockedIncrement(&NextSlot) - 1;
#else
        mySlot = InterlockedExchangeAdd(&NextSlot, 1); //should work on all cpus and Win versions?
#endif
        if (mySlot >= NumberOfThreads) return false; //an unknown thread entered waitForSync??
    }//if

    lpThreadArray[mySlot] = syncnr;  //record our current syncnr


    // Wait until all threads are at the current syncnr or greater.

    for (i=0; i<NumberOfThreads && !timeelapsed; i++) {
        //wait until thread i reaches syncnr or greater...
        //(note that an as-yet empty slot will have a syncnr value of 0)
        while(lpThreadArray[i] < syncnr) {
            //check for timeout
            if (timeout!=INFINITE && clock()-starttime >= timeout) {
                timeelapsed=true; 
                break;
            }//if timeout
            //yield this timeslice
            Sleep(0);
        }//while
    }//for

    //Either we timed out, or all threads are at current syncnr or greater...
    return !timeelapsed;

}//waitForSync


int fm_deleteSync(void){

    NumberOfThreads = 0;

    if (lpThreadArray!=NULL){
        free (lpThreadArray);
        lpThreadArray=NULL;
    }

    return true;
}//deleteSync

#elif 1 //#7: Let syncnr be TLS rather than a param ******************************

int NumberOfThreads=0;
int *lpThreadArray=NULL;
int NextSlot=0;


int fm_createSync(int number){

    if (number<2) return false;
        
    NumberOfThreads = number;
    NextSlot = 0;
    lpThreadArray = calloc(number,sizeof(*lpThreadArray));
    if (!lpThreadArray) return false;

    return true;
}//createSync


int fm_waitForSync(int unused, int timeout){

    THREAD_LOCAL static int mySlot = -1;
    THREAD_LOCAL static int syncnr = 0;
    int i;
    BOOL timeelapsed=false;
    int starttime = clock(); 
    
    if (NumberOfThreads<2 || lpThreadArray==NULL) return false; //In case Sync was not initialized correctly

    syncnr++;   //increment current syncnr level (must always be > 0)
    if(syncnr <= 0) return false;   //oops

    //first time around, assign next slot in ThreadArray to this thread...

    if (mySlot == -1) {
        //not yet assigned a slot...
#if 0 //InterlockedIncrement requires Win98+ or NT 4.0+ for proper return value
        mySlot = InterlockedIncrement(&NextSlot) - 1;
#else
        mySlot = InterlockedExchangeAdd(&NextSlot, 1); //should work on all cpus and Win versions?
#endif
        if (mySlot >= NumberOfThreads) return false; //an unknown thread entered waitForSync??
    }//if

    lpThreadArray[mySlot] = syncnr;  //record our current syncnr


    // Wait until all threads are at the current syncnr or greater.

    for (i=0; i<NumberOfThreads && !timeelapsed; i++) {
        //wait until thread i reaches syncnr or greater...
        //(note that an as-yet empty slot will have a syncnr value of 0)
        while(lpThreadArray[i] < syncnr) {
            //check for timeout
            if (timeout!=INFINITE && clock()-starttime >= timeout) {
                timeelapsed=true; 
                break;
            }//if timeout
            //yield this timeslice
            Sleep(0);
        }//while
    }//for

    //Either we timed out, or all threads are at current syncnr or greater...
    return !timeelapsed;

}//waitForSync


int fm_deleteSync(void){

    NumberOfThreads = 0;

    if (lpThreadArray!=NULL){
        free (lpThreadArray);
        lpThreadArray=NULL;
    }

    return true;
}//deleteSync

#endif //Sync API versions ********************************************



/****** EVENT API'S ******/

INT_PTR fm_triggerEvent(INT_PTR n, INT_PTR event, INT_PTR previous ){

        INT_PTR RetVal=0;

        //Save values
        int n2 = fmc.n;
        int e2 = fmc.e;
        int previous2 = fmc.previous;
        int ctlMsg2 = fmc.ctlMsg;
        int ctlCode2 = fmc.ctlCode;
        int ctlClass2 = fmc.ctlClass;
    

		if ( !IS_OK_CTL_NR(n) && event != FME_CUSTOMEVENT ) return false;


        //Set new values
        fmc.n = (int)n;
        fmc.e = (int)event;
        fmc.previous = (int)previous;
        fmc.ctlMsg = (int)event;//WM_SIZE;
        fmc.ctlCode = 0;

		if (event==FME_CUSTOMEVENT)
			fmc.ctlClass = CC_UNUSED;
		else
			fmc.ctlClass = gParams->ctl[n].class;
        
		//Catch any longjmp's back...
		if (setjmp(Mde_error_exit_jb)) {
			//someone longjmp'ed back to us...
			return -1;   //abort -- or restore old values first???
		}

		//Do call
        RetVal = X86_call_fmf1(gParams->CB_onCtl, (INT_PTR)&fmc);

        //Restore old values
        fmc.n = n2;
        fmc.e = e2;
        fmc.previous = previous2;
        fmc.ctlMsg = ctlMsg2;
        fmc.ctlCode = ctlCode2;
        fmc.ctlClass = ctlClass2;

        return RetVal;
}


INT_PTR fm_doEvents(void) {  

        MSG msg; 
        
        while ( PeekMessage(&msg, fmc.hDlg, 0, 0, PM_REMOVE ) )  {  
            TranslateMessage(&msg);
            DispatchMessage(&msg);
            if (msg.message == WM_QUIT) break; // controlled by windows
            if (msg.message == WM_CLOSE && msg.hwnd == fmc.hDlg) break;
            //if( m_bShutDown != FALSE ) break; // controlled by program
        }

        return true;
}



static INT_PTR fm_restoreRect (void){

    //Only run in proxy mode
    if (!fmc.doingProxy) return false;

    //imageOffset.x += 1;
    //imageOffset.y += 1;

    GetProxyItemRect (fmc.hDlg, &gProxyRect);
    SetupFilterRecordForProxy (globals, &fmc.scaleFactor, fmc.scaleFactor, &imageOffset);

    //Restore Values

{ //scope
    VRect scaledFilterRect;
    OSErr err = noErr;

	fmc.outRect32 = GetOutRect();	//needed??

    fmc.scaleFactor = gStuff->inputRate >> 16;
    fmc.srcPtr = (unsigned8 *) MyAdjustedInData;
    fmc.dstPtr = (unsigned8 *) gStuff->outData;
    fmc.mskData = (unsigned8 *) MyAdjustedMaskData;
    fmc.rows = fmc.outRect32.bottom - fmc.outRect32.top;
    fmc.columns = fmc.outRect32.right - fmc.outRect32.left;
    fmc.srcRowDelta = gStuff->inRowBytes;
    fmc.srcColDelta = gStuff->inHiPlane - gStuff->inLoPlane + 1;
    fmc.inPlanes = fmc.srcColDelta;
    fmc.dstRowDelta = gStuff->outRowBytes;
    fmc.dstColDelta = gStuff->planes;
    fmc.mskRowDelta = gStuff->maskRowBytes;
    fmc.mskColDelta = 1;

#ifndef DONT_USE_MASK
    if (gStuff->haveMask)
    {   // mask stuff already set up for selection mask...
        gNoMask = FALSE;
    }
    else
#endif
    if (fmc.inPlanes >= 4)
    {   // Channel 3 is probably an alpha channel, so
        // use the alpha channel (layer transparency mask)
        // instead of the selection mask
        fmc.mskData = (unsigned8 *) MyAdjustedInData + 3;
        fmc.mskRowDelta = fmc.srcRowDelta;
        fmc.mskColDelta = fmc.srcColDelta;
        gNoMask = FALSE;
    }
    else
    {   // No selection mask and no transparency channel, so
        // we'll just fake a mask that includes the entire
        // filterRect (which will give us a nice rectangular
        // bevel).
        gNoMask = TRUE;
    }

	fmc.filterRect32 = GetFilterRect();	//needed??
	scaledFilterRect = fmc.filterRect32;
    scaleRect32(&scaledFilterRect, 1, fmc.scaleFactor);
    X0 = scaledFilterRect.left;
    Y0 = scaledFilterRect.top;

	fmc.inRect32 = GetInRect();	//needed??
    fmc.x_start = fmc.inRect32.left - X0;
    fmc.y_start = fmc.inRect32.top - Y0;

    fmc.x_end = fmc.x_start + fmc.columns;
    fmc.y_end = fmc.y_start + fmc.rows;

}//scope

    return true;

}


static INT_PTR fm_requestRect (INT_PTR inLeft, INT_PTR inTop,
							   INT_PTR inRight, INT_PTR inBottom, INT_PTR scaleFactor)
{
    long    maxScaleFactor;
    VRect   scaledImage;

    //Only run in proxy mode
    if (!fmc.doingProxy) return false;

    fmc.inRect32.left   = (int)inLeft;
    fmc.inRect32.right  = (int)inRight;
    fmc.inRect32.top    = (int)inTop;
    fmc.inRect32.bottom = (int)inBottom;
	SetInRect(fmc.inRect32);	//needed??

	fmc.filterRect32 = GetFilterRect();
    CalcProxyScaleFactor (&fmc.filterRect32, &gProxyRect, &maxScaleFactor); //Calc maxScaleFactor
    if (scaleFactor == -888)
    {   // -888 means use calculated optimal scale factor...
        scaleFactor = maxScaleFactor;
    }
    // Make sure scale factor is in range [1,MAX_SCALE_FACTOR].
    if (scaleFactor < 1) scaleFactor = 1;
    if (scaleFactor > MAX_SCALE_FACTOR) scaleFactor = MAX_SCALE_FACTOR;

#if 0
    // Skip scalefactor 13 or 15...
    if (scaleFactor == 13 || scaleFactor == 15)
    {
        scaleFactor++;  //use next higher scaleFactor (14 or 16)
    }
#else
	scaleFactor = roundUpScaleFactor((int)scaleFactor);
#endif

	fmc.inRect32 = GetInRect();	//needed??
    scaleRect32(&fmc.inRect32, 1, (long)scaleFactor);
	scaledImage = fmc.inRect32;
	SetInRect(fmc.inRect32);	//needed??

    if ((gParams->flags & (AFH_DRAGGING_PROXY|AFH_ZERO_OUTPUT_DRAG|AFH_DRAG_FILTERED_OUTPUT))
                       == (AFH_DRAGGING_PROXY|AFH_ZERO_OUTPUT_DRAG))
    {   // Don't need to set up output Rect.  Could save some time.
        PISetRect32(&fmc.outRect32, 0, 0, 0, 0);
    }
    else
    {
		fmc.outRect32 = fmc.inRect32;
    }
	SetOutRect(fmc.outRect32);	//needed??

    gStuff->inLoPlane = gStuff->outLoPlane = 0;
    gStuff->inHiPlane = gStuff->outHiPlane = gStuff->planes - 1;

#ifndef DONT_USE_MASK
	if (gStuff->haveMask) {
		///fmc.inRect32 = GetInRect();	//needed??
		fmc.maskRect32 = fmc.inRect32;
	}
    else
#endif
	{
        PISetRect32(&fmc.maskRect32, 0, 0, 0, 0);
	}
	SetMaskRect(fmc.maskRect32); //needed??

    gStuff->inputRate = gStuff->maskRate = long2fixed((long)scaleFactor);
    
    if (gParams->flags & AFH_USE_HOST_PADDING)
    {
        gStuff->inputPadding = 0;   // 0=black.
        gStuff->outputPadding = 0;  // 0=black.
        gStuff->maskPadding = 0;    // 0=black (non-selected).
    }
    else
    {
        gStuff->inputPadding = plugInDoesNotWantPadding;
        gStuff->outputPadding = plugInDoesNotWantPadding;
        gStuff->maskPadding = plugInDoesNotWantPadding;
    }

    // Now call AdvanceState() for real, to get the actual proxy pixels.
    addPadding(globals);

    gResult = AdvanceState();

    if (gResult != noErr) {
        // We use the special combo MB_ICONHAND|MB_SYSTEMMODAL
        // to ensure that the message box will be displayed
        // regardless of low memory...
        if (gResult == memFullErr) {
            msgBox( MB_ICONHAND|MB_SYSTEMMODAL,
                    fmc.filterTitleText,//"FilterMeister",
                    "Not enough memory to perform requestRect()\n"
                    "at this zoom level."
                    );
            return false;
        }
        else {
            msgBox( MB_ICONHAND|MB_SYSTEMMODAL,
                    fmc.filterTitleText,//"FilterMeister",
                    "Unknown error (%d) while perfoming requestRect()\n",
                    gResult
                    );
            return false;
        }
    }
    
    subtractPadding(globals);

    // Need to set outRect correctly for ShowOutput();
	fmc.inRect32 = GetInRect();	//needed??
	fmc.outRect32 = fmc.inRect32;
	SetOutRect(fmc.outRect32);

    // Now gStuff->inData should have the (possibly padded) subsampled version of input data,
    // gStuff->maskData may have the (possibly padded) subsampled version of the mask data,
    // and gStuff->outData should be a buffer the size of input data that
    // could be used as a temporary buffer to do intermediate filtering
    // to display proxies

    // MyAdjustedInData and MyAdjustedMaskData will point to the start of the NON-padded
    // data.

    //Setup Values

    //fmc.scaleFactor = gStuff->inputRate >> 16; //Don't use because it confuses restoreImageData()
    fmc.srcPtr = (unsigned8 *) MyAdjustedInData;
    fmc.dstPtr = (unsigned8 *) gStuff->outData;
    fmc.mskData = (unsigned8 *) MyAdjustedMaskData;
	fmc.outRect32 = GetOutRect();	//needed??
    fmc.rows = fmc.outRect32.bottom - fmc.outRect32.top;
    fmc.columns = fmc.outRect32.right - fmc.outRect32.left;
    fmc.srcRowDelta = gStuff->inRowBytes;
    fmc.srcColDelta = gStuff->inHiPlane - gStuff->inLoPlane + 1;
    fmc.inPlanes = fmc.srcColDelta;
    fmc.dstRowDelta = gStuff->outRowBytes;
    fmc.dstColDelta = gStuff->planes;
    fmc.mskRowDelta = gStuff->maskRowBytes;
    fmc.mskColDelta = 1;

#ifndef DONT_USE_MASK
    if (gStuff->haveMask)
    {   // mask stuff already set up for selection mask...
        gNoMask = FALSE;
    }
    else
#endif
    if (fmc.inPlanes >= 4)
    {   // Channel 3 is probably an alpha channel, so
        // use the alpha channel (layer transparency mask)
        // instead of the selection mask
        fmc.mskData = (unsigned8 *) MyAdjustedInData + 3;
        fmc.mskRowDelta = fmc.srcRowDelta;
        fmc.mskColDelta = fmc.srcColDelta;
        gNoMask = FALSE;
    }
    else
    {   // No selection mask and no transparency channel, so
        // we'll just fake a mask that includes the entire
        // filterRect (which will give us a nice rectangular
        // bevel).
        gNoMask = TRUE;
    }

	fmc.inRect32 = GetInRect();	//needed??
    fmc.x_start = fmc.inRect32.left;
    fmc.y_start = fmc.inRect32.top;
    fmc.x_end = fmc.inRect32.right;
    fmc.y_end = fmc.inRect32.bottom;

    return true;

} /*fm_requestRect*/





/********************************************************************/
/*                 DLL Access Routines                              */
/*                                                                  */
/*  1. Original code contributed by Florian Xhumari, as transmitted */
/*     by Harald Heim.                                              */
/*  2. Remove commented-out code (Alex Hunter).                     */
/*                                                                  */
/********************************************************************/


////typedef void (*FMLIB_FN)();	//eh?? used??


static INT_PTR fm_loadLib(INT_PTR libname) 
{
    HINSTANCE retval;
	
	if (!strrchr((char *)libname, 0x5C)){	//if contains NO '\' char??? Is that what we wanted??? Why rchr???
		char string[512];		/// ACK!!!! POSSIBLE BUFFER OVERFLOW ERROR!!!

		strcpy(string, fmc.filterInstallDir); //Always has an ending \ sign
		strcat(string,(char *)libname);
		retval = LoadLibrary((LPCTSTR)string);	//try to locate DLL in FM install dir
		if (retval != 0) return (INT_PTR) retval;	//return if located, else fall though
												//to search for unmodified lib name
	}

	return (INT_PTR)LoadLibrary((char *)libname); 
}


static INT_PTR fm_freeLib(INT_PTR hinstLib)
{
	return FreeLibrary((HINSTANCE)hinstLib);
}


static INT_PTR fm_getLibFn(INT_PTR hinstLib, INT_PTR fnName)
{
	return (INT_PTR)GetProcAddress((HINSTANCE)hinstLib, (char *)fnName);
}


// Disable "no return value" warnings
#pragma warning(disable:4035)

#ifndef _WIN64	//Win64 version is in ML64 asm module.
__declspec(naked)
static int fm_callLibFmc(int fn, ...)
{
	/**
	 * Calls a function fn that takes exactly the same parameters as used
	 * to call this function, except the first argument (fn) is replaced by
     * a pointer to the FM Context record (fmcp).
     * This function is agnostic about the return value, which may be void,
     * int (returned in EAX), double (returned on top of NDP stack ST(0)),
     * etc., depending on an appropriate cast of the function signature.
	 *
	 * N.B. These functions assume the called function fn uses the CDECL
	 *      calling convention.  To use these functions to call a STDCALL
	 *		function fn, they themselves must be compiled with the STDCALL
	 *		convention.
	 * N.B.#2	This function will be deprecated when FM has the ability to
	 *			call a function indirectly through a pointer, in which case
	 *			simply replace:
	 *				ival = callLibFmc(fn, a, b, c,...);
	 *			with:
	 *				ival = fn(fmcp, a, b, c,...);	//X64 THREAD PROBLEM: must be task-local fmcp!!! (e.g. fmcp->fmcp??? or fn(eBX,a,b,c...)???)
	 *			etc.
     *
	 * Example of called functions in a DLL: 
     *
     *      __declspec(dllexport)
     *      int MyForEveryTile(FMcontext *fmcp, int X, int Y) { return false; }
     *
     *      __declspec(dllexport)
     *      double MyFloat(FMcontext *fmcp, int a, double x, double y) {
     *          return a*(x - y);
     *      }
     *
     * To invoke these functions from FM:
     *
     *      int g_hMyDll = loadLib("MyDll");
     *      if (!g_hMyDll) ERROR...
     *
     *      int g_pfnMyForEveryTile = getLibFn(g_hMyDll, "MyForEveryTile");
     *      if (!g_pfnMyForEveryTile) ERROR...
     *
     *      int g_pfnMyFloat = getLibFn(g_hMyDll, "MyFloat");
     *      if (!g_pfnMyFloat) ERROR...
     *
     *      int iRes = callLibFmc(g_pfnMyForEveryTile, 100, 200);
     *
     *      double fRes = fcallLibFmc(g_pfnMyFloat, 10, 1.2, 3.14);
     *
     * where FM effectively defines callLibFmc and fcallLibFmc as:
     *
     *      typedef int (*ifn1v) (int, ...);
     *      typedef double (*dfn1v) (int, ...);
     *      #define callLibFmc    ((ifn1v)callLibFmc)
     *      #define fcallLibFmc   ((dfn1v)callLibFmc)
	 */

	__asm {
#if 1	//assumes ebx already loaded with fmcp...
        mov eax, [esp+4]        ; eax <- fn
        mov [esp+4], ebx        ; replace fn with ebx (fmcp) in arg list
        jmp eax                 ; jmp to fn, which returns to caller
#elif 0	//same as above but uses XCHG instruction, which was deprecated by Fog...
        mov eax, ebx            ; eax <- ebx (=fmcp)
        xchg eax, [esp+4]       ; swap with fn in arg list; eax now has fn  //don't use XCHG per Fog!!!
        jmp eax                 ; jmp to fn, which returns to caller
#elif 1	//loads fmcp explicitly in case ebx was not already loaded...
        mov eax, [esp+4]        ; eax <- fn
        mov ebx, gFmcp          ; (In case we weren't called from FM-generated code!)	//MULTI-THREAD PROBLEM!!!
        mov [esp+4], ebx        ; replace fn with ebx (fmcp) in arg list
        jmp eax                 ; jmp to fn, which returns to caller
#else
        ; old version, minimal case, in case fmcp argument is unused...
        jmp dword ptr [esp+4]   ; jump to the function; it will return to caller of current function
#endif
        //
        // return result will be in eax (for callLibFmc) or
        // on top of NDP stack ST(0) (for fcallLibFmc) --
        // only the caller cares which.
	}
} /*fm_callLibFmc*/
#endif //!WIN64

#if 0
////////////// TEST PATCH /////////////////////
     
           typedef int (*ifn1v) (int, ...);
           typedef double (*dfn1v) (int, ...);
           #define fm_callLibraryIntFast      ((ifn1v)fm_callLibFmc)
           #define fm_callLibraryDoubleFast   ((dfn1v)fm_callLibFmc)

void xyxxy() {
           int g_hMyDll = gFmcp->loadLibrary((int)"MyDll");
           if (!g_hMyDll) MessageBeep(MB_OK);
           {
           int g_pfnMyForEveryTile = gFmcp->getLibraryFunction(g_hMyDll, (int)"MyForEveryTile");
           if (!g_pfnMyForEveryTile) MessageBeep(MB_OK);
           {    
           int g_pfnMyFloat = gFmcp->getLibraryFunction(g_hMyDll, (int)"MyFloat");
           if (!g_pfnMyFloat) MessageBeep(MB_OK);
           {
           int iRes = fm_callLibraryIntFast(g_pfnMyForEveryTile, 100, 200);
           double x1,x2,x3,x4,x5,x6,x7,x8,x9;
           double fRes = fm_callLibraryDoubleFast(g_pfnMyFloat, 10, 1.2, 3.14);

           x1 = sin(x1)+cos(x2)+sin(x3+x4)*cos(x5-x6);
           }}}
}
///////////////////////////////////
#endif

#ifndef _WIN64	//Win64 version is in ML64 asm module.
__declspec(naked)
static int fm_callLib(int fn, ...)
{
	/**
	 * Calls a function that takes the same parameters as used
	 * to call this function with the exception of the first one, fn.
	 *
	 * Example of called function in a DLL:
	 *
	 *		int myFunction(double a, double b) { return a + b; }
	 *
	 * N.B.	This function will be deprecated when FM has the ability to
	 *		call a function indirectly through a pointer, in which case
	 *		simply replace:
	 *			ival = callLib(fn, a, b, c,...);
	 *		with:
	 *			ival = fn(a, b, c,...);
	 *		etc.
	 */

	/**
	 * Use all data in previous function's frame (including all arguments passed
	 * to the current function) as arguments to the function fn. Actually,
	 * not all data, but all data excluding last argument (fn).
	 *
	 * Then, call function fn and remove added arguments.
	 *
	 * It is assumed that called function fn follows CDECL (or STDCALL) conventions,
	 * in particular it preserves register ebx.
	 */
#if 0										//CHECK -- WHAT DID 109f USE????
	//This was Florian Xhumari's original code, with some optimizations...
	__asm {
        ; prolog
        push ebp            ; save must-save registers
        ;;push ebx          ; not used
        push esi
        push edi
        mov ebp, esp        ; our new stack frame

        dec ecx             ; upon entry, ecx contains # of actual args (including fn)
        shl ecx, 3          ; subtract the fn argument and convert to (worst case) arg byte count
        lea esi, [ebp+5*4]	; skip 5 dwords (pushed edi, esi, ebp, eip and first argument (fn))

		sub esp, ecx		; allocate space on stack
		mov edi, esp		; destination: end of added space in stack
		shr	ecx, 2			; will be moving dwords, so divide count by 4
		rep movs dword ptr [edi]	; copy argument list, excluding fn
		call dword ptr [ebp+4*4] ; call the function

		;;;add esp, XXX		; remove arguments (but STDCALL may have already removed them!!!)

        ; epilog
        mov esp, ebp        ; restore our stack frame
        pop edi             ; restore callee-save registers
        pop esi
        ;;pop ebx           ; wasn't used
        pop ebp             ; restore caller's stack frame
        ret                 ; caller will pop actual args
	} //__asm

/////Is it really necessary to copy the arg list as above, instead of reducing it in situ as below???
#elif 1	//CAN USE NOW THAT FM GENERATES UNIVERSAL CALL???
	//Only if fn and callLib are both STDCALL, or if UNIVERSAL call to callLib???
	__asm {
		pop	eax				; eax = return ip
		xchg eax, [esp]		; eax = fn, return ip overwrites fn on stack	//DEPRECATE XCHG to mem!!!
		jmp	eax				; jump to fn.  We already removed one "arg" from stack,
							; fn will remove the remaining n-1 args when it returns
							; via a RET (n-1).  Our caller will remove no more args,
							; since it called us as STDCALL.
							; X64 PROBLEM: Pre-allocated arg stack will shrink by 1
							; each time we make such a call!  NFG!!!  But there are
							; no STDCALLs in X64!!!!
	} //__asm
#elif 1	//NOT VALID -- FOR TESTING ONLY!!!
	//Restore arg stack to original size before returning, call to callLib is not STDCALL..
	//dummy frame to allow compilation:
	typedef struct {int tmps[MAX_TEMPS]; int tmp0; int tmp1; int tmp2;} framex;	//NOT VALID -- FOR TESTING ONLY!!!
	__asm {
		; allocate tmp0, tmp1, and tmp2 in callLib caller's stackframe
	    ; or in FMC record (any problem with nested calls?)
		pop [ebp].tmp1			; tmp1 = return eip
		pop [ebp].tmp2			; tmp2 = fn
								; esp now points to first arg (if any) past fn
		mov [ebp].tmp0, esp		; save current esp in tmp0
		call [ebp].tmp2			; universal call to fn indirectly thru tmp2
		mov esp, [ebp].tmp0		; restore esp in case fn is STDCALL (else optional?)
		push [ebp].tmp2			; restore fn arg (not needed??)
		push [ebp].tmp1			; restore original return eip
		ret						; assumes original call is CDECL or universal
	} //__asm
#elif 0	//Alternate of above, doesn't restore the fn arg.	//NOT VALID -- FOR TESTING ONLY!!!
	__asm {
		; allocate tmp0 and tmp1 at top of tmp stack
		; note 1 or 2 of the tmp's might be regs, for extra efficiency!
		mov tmp0, esp		; save current sp in tmp0
		pop tmp1			; tmp1 = return ip
		pop eax				; eax = fn
		call eax			; universal call to fn indirectly thru eax
		mov esp, tmp0		; restore esp in case altered by STDCALL fn (else optional?)
		mov [esp+0], tmp1	; restore original return ip
		ret					; assumes original call is CDECL or universal
	} //__asm
#elif 0	//NFG!!
	//Alternate of above, assumes fn is CDECL (i.e., preserves sp).
	//would it be easier to gen this into a code buffer at startup???
	//no, tos may vary from point to point

	////temporary defs to allow compilation...
#define REGOBJECTKIND 123
	static int tos = -1;

	//allocate tmp1 at top of tmp stack
	//note tmp1 might be a reg, for extra efficiency!
	push(REGOBJECTKIND);	//allocate a tmp at top of stack

	// NFG -- tos, push(), pop() are compile-time, not run-time!!!
	// use dedicated tmp(s) in stack frame!!!

	if (isReg(tos)) {
		switch (tos) {
			case 0: __asm {pop edi}; break;
			case 1: __asm {pop esi}; break;
#if 0	//can't use x64 regs for x86 assembly!!!
			case 2: __asm {pop r12}; break;
			case 3: __asm {pop r13}; break;
			case 4: __asm {pop r14}; break;
			case 5: __asm {pop r15}; break;
#endif
			default: bug(975);
		}//switch
	} else {
		switch (tos) {
			case 0: __asm {pop dword ptr[ebp+0]}; break;
			case 1: __asm {pop dword ptr[ebp+4]}; break;
			case 2: __asm {pop dword ptr[ebp+8]}; break;
			case 3: __asm {pop dword ptr[ebp+12]}; break;
			case 4: __asm {pop dword ptr[ebp+16]}; break;
			case 5: __asm {pop dword ptr[ebp+20]}; break;
			default: bug(976);
		}//switch
	}//if
	__asm {
		;;;pop tmp1			; tmp1 = return ip
		pop eax				; eax = fn
		call eax			; call fn indirectly thru eax
							; fn is CDECL, hence restores esp w/o popping args
		sub esp, 4			; esp points to old fn arg (which may be clobbered by now)
	}//__asm
	if (isReg(tos)) {
		switch (tos) {
			case 0: __asm {push edi}; break;
			case 1: __asm {push esi}; break;
#if 0	//can't use x64 regs for x86 assembly!!!
			case 2: __asm {push r12}; break;
			case 3: __asm {push r13}; break;
			case 4: __asm {push r14}; break;
			case 5: __asm {push r15}; break;
#endif
			default: bug(975);
		}//switch
	} else {
		switch (tos) {
			case 0: __asm {push dword ptr [ebp+0]}; break;
			case 1: __asm {push dword ptr [ebp+4]}; break;
			case 2: __asm {push dword ptr [ebp+8]}; break;
			case 3: __asm {push dword ptr [ebp+12]}; break;
			case 4: __asm {push dword ptr [ebp+16]}; break;
			case 5: __asm {push dword ptr [ebp+20]}; break;
			default: bug(976);
		}//switch
		pop();				//deallocate tmp1
	}//if
	__asm {
		;push tmp1			; restore original return ip
		ret					; assumes original call is CDECL or universal
	} //__asm
#elif 1
	//Alternate of above, assumes fn is CDECL (i.e., preserves sp).
	//would it be easier to gen this into a code buffer at startup???
	//no, tos may vary from point to point

	////temporary defs to allow compilation...
	static int tos = -1;
/////#include "mdd.h"
#include "md386def.hc"
/////#include "md386proto.h"

	// NFG -- tos, push(), pop() are compile-time, not run-time!!!
	// Provide separate entry points, callLib_n, where n is # of
	// available temps available to caller at point of call.

	//allocate tmp1 at top of tmp stack
	//note tmp1 might be a reg, for extra efficiency!
     if (tos >= MAX_TEMPS - 1)
         Mde_abort_command(MD386__EVALOV, 0, 0, 0);
     ++tos;

	//assumes highest reg used for temps is esi... (for x86)
	assert( tmpReg(nTmps-1)==esi );

	if (isReg(tos)) {
		__asm {
			pop	esi;			; tmp1 = return ip
			pop eax				; eax = fn
			call eax			; call fn indirectly thru eax
								; fn is CDECL, hence restores esp w/o popping args
			sub esp, 4			; esp points to old fn arg (which may be clobbered by now)
			push esi			; restore original return ip from tmp1
			dec tos				; deallocate tmp1
			ret					; assumes original call is CDECL or UNIVERSAL
		}//__asm
	} else {
		__asm {
			pop	dword ptr [ebp+4*(MAX_TEMPS - 1)]; tmp1 = return ip
			pop eax				; eax = fn
			call eax			; call fn indirectly thru eax
								; fn is CDECL, hence restores esp w/o popping args
			sub esp, 4			; esp points to old fn arg (which may be clobbered by now)
			push dword ptr [ebp+4*(MAX_TEMPS - 1)]; restore original return ip from tmp1
			dec tos				; deallocate tmp1
			ret					; assumes original call is CDECL or UNIVERSAL
		}//__asm
	}//if
#endif
} /*fm_callLib*/
#endif //!WIN64

#ifndef _WIN64	//Not used in WIN64 implementation
#if 1
	// Provide separate entry points, callLib_n, where n is # of
	// temp registers available to caller at point of call;
	// assumes sufficient nonreg temps are available.
__declspec(naked)
/*static*/ int fm_callLib_0(int fn, ...)
{	//no temp regs available, but at least one temp loc is available
	__asm {
		pop	dword ptr [ebp+4*(MAX_TEMPS - 1)]; tmp1 = return ip
		pop eax				; eax = fn
		call eax			; call fn indirectly thru eax
							; fn is CDECL, hence restores esp w/o popping args
#if 0	//use whichever is shorter and/or faster...
		push eax			; push garbage over old fn arg, esp points to old fn arg loc
#else
		sub esp, 4			; esp points to old fn arg (which may be clobbered by now)
#endif
#if 0
		jmp dword ptr [ebp+4*(MAX_TEMPS-1)]	; quick return, but unmatched calls/rets may defeat cpu optimization
#else
		push dword ptr [ebp+4*(MAX_TEMPS - 1)]; restore original return ip from tmp1
		ret					; assumes original call is CDECL or universal
#endif
	}//__asm
}/*fm_callLib_0*/
__declspec(naked)
/*static*/ int fm_callLib_1(int fn, ...)
{	//at least 1 temp reg available (assumes x86; not used in X64gen)
	__asm {
		pop	esi;			; tmp1 = return ip
		pop eax				; eax = fn
		call eax			; call fn indirectly thru eax
							; fn is CDECL, hence restores esp w/o popping args
#if 1	//use whichever is shorter and/or faster...
		push eax			; push garbage over old fn arg, esp points to old fn arg loc
#else
		sub esp, 4			; esp points to old fn arg (which may be clobbered by now)
#endif
#if 0
		jmp esi				; quick return, but unmatched calls/rets may defeat cpu optimization
#else
		push esi			; restore original return ip from tmp1
		ret					; assumes original call is CDECL or universal
#endif
	}//__asm
}/*fm_callLib_1*/
#endif
#endif //!WIN64

#ifndef _WIN64	//Win64 version is in ML64 asm module.
__declspec(naked)
static int __cdecl fm_getCpuReg(int nr)
{
    // Get value of specified CPU register:
    // nr: 0 = edi, 1 = esi, 2 = ebp, 3 = esp, 4 = ebx, 5 = edx, 6 = ecx, 7 = eax, 8 = eip
    __asm {                 ;        +0    1    2        3    4    5    6    7    8    9
        pushad              ; esp -> edi, esi, ebp, old esp, ebx, edx, ecx, eax, eip, nr
        add dword ptr [esp+3*4], 2*4  ; correct saved esp for push nr, push eip
        mov eax, [esp+9*4]  ; eax <- nr
        and eax, 0x0f       ; prevent GPF on out-of-range arg.
        mov eax, [esp+eax*4]    ; load desired register
        add esp, 8*4        ; restore esp
        ret
    }
}/*fm_getCpuReg*/
#endif //!WIN64

#ifndef _WIN64	//Win64 version is in ML64 asm module.
__declspec(naked)
static void __cdecl fm_finit(void)
{
    __asm {
        finit               ; init cw to 0x037f, sw to 0, tag word to 0xffff, rest to 0
        ret
    }
}
#endif //!WIN64

#ifndef _WIN64	//Win64 version is in ML64 asm module.
__declspec(naked)
static int __cdecl fm_fstsw(void)
{
    __asm {
        xor eax, eax        ; clear 32-bit eax
        fstsw ax            ; return 16-bit status word in ax
        ret
    }
}
#endif //!WIN64

#ifndef _WIN64	//Win64 version is in ML64 asm module.
__declspec(naked)
static int __cdecl fm_fstcw(void)
{
    __asm {
        push 0              ; clear a dword on the stack
        fstcw word ptr[esp] ; store control word in low 16 bits of dword
        pop eax             ; pop zero-extended control word into eax
        ret
    }
}
#endif //!WIN64

#ifndef _WIN64	//Win64 version is in ML64 asm module.
__declspec(naked)
static void __cdecl fm_fldcw(INT_PTR cw)
{
    __asm {
        fldcw word ptr[esp+4]   ; load control word from low 16 bits of argument
        ret
    }
}
#endif //!WIN64

// restore "no return value" warnings
#pragma warning (default: 4035)


/****************** End DLL Access Routines *************************/




static INT_PTR fm_setPreviewDragMode (INT_PTR mode)
{
    //0
    //1 = ZeroDrag //Default!
    //2 = OutputDrag
    //3 = Zero + Output
    
    if (mode==0){
        gParams->flags &= ~AFH_ZERO_OUTPUT_DRAG;
        gParams->flags &= ~AFH_DRAG_FILTERED_OUTPUT;
		//gParams->flags |= AFH_QUICK_TOGGLE;
    } else if (mode==1){
        gParams->flags |= AFH_ZERO_OUTPUT_DRAG;
        gParams->flags &= ~AFH_DRAG_FILTERED_OUTPUT;
		//gParams->flags |= AFH_QUICK_TOGGLE;
    } else if (mode==2){
        gParams->flags &= ~AFH_ZERO_OUTPUT_DRAG;
        gParams->flags |= AFH_DRAG_FILTERED_OUTPUT;
		//gParams->flags |= AFH_QUICK_TOGGLE;
    } else if (mode==3){
        gParams->flags |= AFH_ZERO_OUTPUT_DRAG;
        gParams->flags |= AFH_DRAG_FILTERED_OUTPUT;
		//gParams->flags |= AFH_QUICK_TOGGLE;

	}/*else if (mode==4){
        gParams->flags &= ~AFH_ZERO_OUTPUT_DRAG;
        gParams->flags &= ~AFH_DRAG_FILTERED_OUTPUT;
		gParams->flags &= ~AFH_QUICK_TOGGLE;
    } else if (mode==5){
        gParams->flags |= AFH_ZERO_OUTPUT_DRAG;
        gParams->flags &= ~AFH_DRAG_FILTERED_OUTPUT;
		gParams->flags &= ~AFH_QUICK_TOGGLE;
    } else if (mode==6){
        gParams->flags &= ~AFH_ZERO_OUTPUT_DRAG;
        gParams->flags |= AFH_DRAG_FILTERED_OUTPUT;
		gParams->flags &= ~AFH_QUICK_TOGGLE;
    } else if (mode==7){
        gParams->flags |= AFH_ZERO_OUTPUT_DRAG;
        gParams->flags |= AFH_DRAG_FILTERED_OUTPUT;
		gParams->flags &= ~AFH_QUICK_TOGGLE;
    }*/

    return true;
    
}



static INT_PTR fm_linearInterpolate (INT_PTR v1, INT_PTR v2, double x)
{

    return  (INT_PTR)(v1*(1.0-x) + v2*x);
}

static INT_PTR fm_cosineInterpolate (INT_PTR v1, INT_PTR v2, double x)
{

    double ft = x * 3.1415927;			//precision??? Use M_PI or ???
	double f = (1.0 - cos(ft)) * 0.5;

	return  (INT_PTR) (v1*(1.0-f) + v2*f);

}

static INT_PTR fm_cubicInterpolate (INT_PTR v0, INT_PTR v1, INT_PTR v2, INT_PTR v3, double x)
{

    INT_PTR P = (v3 - v2) - (v0 - v1);
	INT_PTR Q = (v0 - v1) - P;
	INT_PTR R = v2 - v0;
	INT_PTR S = v1;

    return (INT_PTR)(P*x*x*x + Q*x*x + R*x + S);

}


static INT_PTR fm_hermiteInterpolate (INT_PTR T1, INT_PTR P1, INT_PTR P2, INT_PTR T2, double s)
{
  double s2 = s*s;
  double s3 = s2*s;

  double h1 =  2*s3 - 3*s2 + 1;			// calculate basis function 1
  double h2 = -2*s3 + 3*s2;				// calculate basis function 2
  double h3 =    s3 - 2*s2 + s;			// calculate basis function 3
  double h4 =    s3 -  s2;				// calculate basis function 4

  return (INT_PTR)(h1*P1 +              // multiply and sum all funtions
				   h2*P2 +              // together to build the interpolated
				   h3*T1 +              // point along the curve.
				   h4*T2);
}


static int fm_getImageTitle (LPSTR n)
{
    GetPropertyProc pGPP;
    OSErr err;
    int32 simpleProperty;
    Handle complexProperty;
    int success = false;


    pGPP = gStuff->getPropertyObsolete;
    if (gStuff->propertyProcs) {
        if (gStuff->propertyProcs->propertyProcsVersion >= kCurrentPropertyProcsVersion &&
            gStuff->propertyProcs->numPropertyProcs >= 1 &&
            gStuff->propertyProcs->getPropertyProc) {
            //prefer this version of getPropertyProc over the obsolete one
            pGPP = gStuff->propertyProcs->getPropertyProc;
        }
    } //propertyProcs != 0

    if (pGPP) {

        err = pGPP('8BIM', 'titl', 0, &simpleProperty, &complexProperty);
        if (err == noErr) {
            int size;
            LPSTR ptr;
            size = HostGetHandleSize(gStuff->handleProcs, complexProperty);
            
            if (size>0){
                size = min(size, 255);  //limit to avoid buff overflow  //BUG, was 256
                ptr = HostLockHandle(gStuff->handleProcs, complexProperty, FALSE);
           
                if (ptr != NULL){
                
                    //strcpy (n, ptr); // Title is not null-terminated.
                    memcpy (n, ptr, size);
                    memset (n+size,0,1);
                
                    success = true;
                }

                HostUnlockHandle(gStuff->handleProcs, complexProperty);
                HostDisposeHandle(gStuff->handleProcs, complexProperty);

				return success;
            
            } 

        }

    }   


    return false;

}

static BOOL fm_destroyMenu(HMENU hMenu)
{
    // cdecl -> stdcall thunk
    return DestroyMenu(hMenu);
}


static int fm_insertMenuItem (INT_PTR hMenu, int uItem, LPSTR itemName, int fState, INT_PTR subMenu)
{

    MENUITEMINFO mii;

    if (uItem<1){ //Not smaller than 1
        //uItem=1;
		return false;
	}

    mii.cbSize = sizeof(MENUITEMINFO);

    mii.fMask = MIIM_STATE | MIIM_ID | MIIM_TYPE;
    if (subMenu != 0){ 
            mii.fMask = mii.fMask | MIIM_SUBMENU; 
            mii.hSubMenu = (HMENU) subMenu;
    }

    if (strcmp (itemName, "---") == 0)
        mii.fType = MFT_SEPARATOR;
    else 
        mii.fType = MFT_STRING;

    mii.fState = fState;
    mii.wID = uItem;
    mii.dwTypeData = (LPTSTR)itemName;
    mii.cch = (UINT)strlen(itemName);


    InsertMenuItem ( (HMENU)hMenu, uItem-1, MF_BYPOSITION, &mii);


    return true;
}


static INT_PTR fm_trackPopupMenu (INT_PTR hMenu, INT_PTR type, INT_PTR x, INT_PTR y, INT_PTR style)
{
    RECT r;
    POINT pt;
    
    if (type==1){ //Display at cursor coordinates
        
        GetCursorPos(&pt);
        x = pt.x;
        y = pt.y;
        
    } else { //Use provided coordinates
    
        //Convert DBUs to Pixel
        x = fm_HDBUsToPixels(x);
        y = fm_VDBUsToPixels(y);

        //Map to Screen coordinates
        GetWindowRect(fmc.hDlg, &r);
        x += r.left;
        y += r.top;
    }

    if (style==0) style = TPM_TOPALIGN | TPM_LEFTALIGN | TPM_NONOTIFY | TPM_RETURNCMD | TPM_LEFTBUTTON;

    return TrackPopupMenuEx((HMENU)hMenu, (UINT)style , 
							(int)x, (int)y, fmc.hDlg, NULL);
}


static INT_PTR fm_setMenu (INT_PTR hMenu){


    return SetMenu (fmc.hDlg, (HMENU)hMenu);

}


static INT_PTR fm_setCtlFocus (INT_PTR n)
{

    if (!IS_INUSE_CTL(n) && n != CTL_PREVIEW) {
        return FALSE;
    }

	return (INT_PTR)SetFocus (gParams->ctl[n].hCtl);

}

static INT_PTR fm_checkCtlFocus (INT_PTR n)
{
    
	if (!IS_INUSE_CTL(n) && n!= CTL_PREVIEW) {
        return FALSE;
    }

    if (GetFocus() == gParams->ctl[n].hCtl)
        return true;
    else
        return false;

}

static INT_PTR fm_checkDialogFocus (void)
{
    
	if (GetForegroundWindow() == MyHDlg)
        return true;
    else
        return false;

}


    //// WARNING!! findFirstFile() may modify the lpFileName string!!! /////

static INT_PTR fm_findFirstFile (LPCTSTR lpFileName, LPSTR foundItem, int *dwFileAttributes)
{
    HANDLE hFindFile;
    WIN32_FIND_DATA FindFileData = {0};
    
    //Check if no file pattern and add one
    if (strcmp(lpFileName+strlen(lpFileName)-1,"\\")==0)
        strncat ((char *)lpFileName, "*", 1);//strncat ((const char *)lpFileName, "*", 1);

    hFindFile = FindFirstFile(lpFileName, &FindFileData);

    if (hFindFile != INVALID_HANDLE_VALUE){
        strcpy (foundItem, FindFileData.cFileName);//Pass back item
        *dwFileAttributes = (int) FindFileData.dwFileAttributes;
    }

    return (INT_PTR)hFindFile;

}


static int fm_findNextFile (INT_PTR hFindFile, LPSTR foundItem, int *dwFileAttributes)
{
    BOOL val;
    WIN32_FIND_DATA FindFileData = {0};
 
    val = FindNextFile ((HANDLE) hFindFile, &FindFileData);

    if (val!=0){
        strcpy (foundItem, FindFileData.cFileName);//Pass back item
        *dwFileAttributes = (int) FindFileData.dwFileAttributes;
    }

    return (int) val;

}

static BOOL fm_findClose(HANDLE hFindFile)
{
    // cdecl -> stdcall thunk
    return FindClose(hFindFile);
}


static INT_PTR fm_getLocaleInfo(INT_PTR Locale,
								INT_PTR LCType,
								INT_PTR lpLCData,
								INT_PTR cchData)
{
    // cdecl -> stdcall thunk
    return GetLocaleInfo((LCID)Locale, (LCTYPE)LCType, (LPSTR)lpLCData, (int)cchData);
}



static INT_PTR fm_getSpecialFolder (INT_PTR val, INT_PTR str)
{

	LPITEMIDLIST ppidl; //LPITEMIDLIST 

	if (SHGetSpecialFolderLocation (fmc.hDlg, (int)val, &ppidl) == NOERROR) {
			if (SHGetPathFromIDList(ppidl, (char *)str))
				return true;
			else
				return false;
	} else {

		return false;
	}

}; 


static INT_PTR fm_setCtlStyle(INT_PTR n, INT_PTR flags)
{
    
    if ( !IS_INUSE_CTL(n) ) {
        return FALSE;
    }
    
    gParams->ctl[n].style |= flags;

	SetWindowLong(gParams->ctl[n].hCtl, GWL_STYLE,
        GetWindowLong(gParams->ctl[n].hCtl, GWL_STYLE) | (int)flags);
    //use SWP_FRAMECHANGED to force nonclient area update...
    return SetWindowPos(gParams->ctl[n].hCtl, NULL,
                        0, 0, 0, 0,
                        SWP_FRAMECHANGED|SWP_NOZORDER|SWP_NOACTIVATE|SWP_NOSIZE|SWP_NOMOVE);
} /*fm_setCtlStyle*/

static INT_PTR fm_clearCtlStyle(INT_PTR n, INT_PTR flags)
{

    if ( !IS_INUSE_CTL(n) ) {
        return FALSE;
    }

    gParams->ctl[n].style &= ~flags;
	
	SetWindowLong(gParams->ctl[n].hCtl, GWL_STYLE,
        GetWindowLong(gParams->ctl[n].hCtl, GWL_STYLE) & ~(int)flags);
    //use SWP_FRAMECHANGED to force nonclient area update...
    return SetWindowPos(gParams->ctl[n].hCtl, NULL,
                        0, 0, 0, 0,
                        SWP_FRAMECHANGED|SWP_NOZORDER|SWP_NOACTIVATE|SWP_NOSIZE|SWP_NOMOVE);
} /*fm_clearCtlStyle*/

static INT_PTR fm_setCtlStyleEx(INT_PTR n, INT_PTR flags)
{
    
    if ( !IS_INUSE_CTL(n) ) {
        return FALSE;
    }
    
    gParams->ctl[n].styleEx |= flags;
	
	SetWindowLong(gParams->ctl[n].hCtl, GWL_EXSTYLE,
        GetWindowLong(gParams->ctl[n].hCtl, GWL_EXSTYLE) | (int)flags);
    //use SWP_FRAMECHANGED to force nonclient area update...
    return SetWindowPos(gParams->ctl[n].hCtl, NULL,
                        0, 0, 0, 0,
                        SWP_FRAMECHANGED|SWP_NOZORDER|SWP_NOACTIVATE|SWP_NOSIZE|SWP_NOMOVE);
} /*fm_setCtlStyleEx*/

static INT_PTR fm_clearCtlStyleEx(INT_PTR n, INT_PTR flags)
{
    
    if ( !IS_INUSE_CTL(n) ) {
        return FALSE;
    }
    
    gParams->ctl[n].styleEx &= ~flags;

	SetWindowLong(gParams->ctl[n].hCtl, GWL_EXSTYLE,
        GetWindowLong(gParams->ctl[n].hCtl, GWL_EXSTYLE) & ~(int)flags);
    //use SWP_FRAMECHANGED to force nonclient area update...
    return SetWindowPos(gParams->ctl[n].hCtl, NULL,
                        0, 0, 0, 0,
                        SWP_FRAMECHANGED|SWP_NOZORDER|SWP_NOACTIVATE|SWP_NOSIZE|SWP_NOMOVE);
} /*fm_clearCtlStyleEx*/



INT_PTR fm_setCtlBuddyStyle(INT_PTR n, INT_PTR buddy, INT_PTR flags)
{
    HWND hBuddy;

    if ( !IS_INUSE_CTL(n) ) {
        return FALSE;
    }

	if (buddy==1){
		hBuddy = gParams->ctl[n].hBuddy1;
		if (hBuddy==NULL) return false;
		gParams->ctl[n].buddy1Style |= flags;
	} else {
		hBuddy = gParams->ctl[n].hBuddy2;
		if (hBuddy==NULL) return false;
		gParams->ctl[n].buddy2Style |= flags;
	}
	

	SetWindowLong(hBuddy, GWL_STYLE,
        GetWindowLong(hBuddy, GWL_STYLE) | (int)flags);
    //use SWP_FRAMECHANGED to force nonclient area update...
    return SetWindowPos(hBuddy, NULL,
                        0, 0, 0, 0,
                        SWP_FRAMECHANGED|SWP_NOZORDER|SWP_NOACTIVATE|SWP_NOSIZE|SWP_NOMOVE);
} /*fm_setCtlBuddyStyle*/

INT_PTR fm_clearCtlBuddyStyle(INT_PTR n, INT_PTR buddy, INT_PTR flags)
{
    HWND hBuddy;

    if ( !IS_INUSE_CTL(n) ) {
        return FALSE;
    }

	if (buddy==1){
		hBuddy = gParams->ctl[n].hBuddy1;
		if (hBuddy==NULL) return false;
		gParams->ctl[n].buddy1Style &= ~flags;
	} else {
		hBuddy = gParams->ctl[n].hBuddy2;
		if (hBuddy==NULL) return false;
		gParams->ctl[n].buddy2Style &= ~flags;
	}
	

	SetWindowLong(hBuddy, GWL_STYLE,
        GetWindowLong(hBuddy, GWL_STYLE) & ~(int)flags);
    //use SWP_FRAMECHANGED to force nonclient area update...
    return SetWindowPos(hBuddy, NULL,
                        0, 0, 0, 0,
                        SWP_FRAMECHANGED|SWP_NOZORDER|SWP_NOACTIVATE|SWP_NOSIZE|SWP_NOMOVE);
} /*fm_clearBuddyCtlStyle*/


INT_PTR fm_setCtlBuddyStyleEx(INT_PTR n, INT_PTR buddy, INT_PTR flags)
{
    HWND hBuddy;

    if ( !IS_INUSE_CTL(n) ) {
        return FALSE;
    }

	if (buddy==1){
		hBuddy = gParams->ctl[n].hBuddy1;
		if (hBuddy==NULL) return false;
		gParams->ctl[n].buddy1StyleEx |= flags;
	} else {
		hBuddy = gParams->ctl[n].hBuddy2;
		if (hBuddy==NULL) return false;
		gParams->ctl[n].buddy2StyleEx |= flags;
	}
	

	SetWindowLong(hBuddy, GWL_EXSTYLE,
        GetWindowLong(hBuddy, GWL_EXSTYLE) | (int)flags);
    //use SWP_FRAMECHANGED to force nonclient area update...
    return SetWindowPos(hBuddy, NULL,
                        0, 0, 0, 0,
                        SWP_FRAMECHANGED|SWP_NOZORDER|SWP_NOACTIVATE|SWP_NOSIZE|SWP_NOMOVE);
} /*fm_setCtlBuddyStyleEx*/

INT_PTR fm_clearCtlBuddyStyleEx(INT_PTR n, INT_PTR buddy, INT_PTR flags)
{
    HWND hBuddy;

    if ( !IS_INUSE_CTL(n) ) {
        return FALSE;
    }

	if (buddy==1){
		hBuddy = gParams->ctl[n].hBuddy1;
		if (hBuddy==NULL) return false;
		gParams->ctl[n].buddy1StyleEx &= ~flags;
	} else {
		hBuddy = gParams->ctl[n].hBuddy2;
		if (hBuddy==NULL) return false;
		gParams->ctl[n].buddy2StyleEx &= ~flags;
	}
	

	SetWindowLong(hBuddy, GWL_EXSTYLE,
        GetWindowLong(hBuddy, GWL_EXSTYLE) & ~(int)flags);
    //use SWP_FRAMECHANGED to force nonclient area update...
    return SetWindowPos(hBuddy, NULL,
                        0, 0, 0, 0,
                        SWP_FRAMECHANGED|SWP_NOZORDER|SWP_NOACTIVATE|SWP_NOSIZE|SWP_NOMOVE);
} /*fm_clearCtlBuddyStyleEx*/



static INT_PTR fm_getSysColor(INT_PTR n)
{	

    if ( !IS_INUSE_CTL(n) ) {
        return FALSE;
    }

	return GetSysColor((int)n);
 
}

static SIZE_T fm_getSysMem(INT_PTR n)
{	
	SIZE_T val;

	//Win 95, 98, ME, NT
//	if (fm_getWindowsVersion() <= 4){
	
		MEMORYSTATUS statex;

	  	
	  	statex.dwLength = sizeof (statex);
	  	GlobalMemoryStatus (&statex);
	  	
		switch (n){
			case 0:
				val = statex.dwMemoryLoad; // percent of memory in use 	
				break;
			case 1:
				val = statex.dwTotalPhys / 1024;     // bytes of physical memory 
				break; 
			case 2:
				val = statex.dwAvailPhys / 1024;     // free physical memory bytes 	
				break; 
			case 3:
				val = statex.dwTotalPageFile / 1024; // bytes of paging file 
				break; 
			case 4:
				val = statex.dwAvailPageFile / 1024; // free bytes of paging file 
				break; 
			case 5:
				val = statex.dwTotalVirtual / 1024;  // user bytes of address space 
				break; 
			case 6:
				val = statex.dwAvailVirtual / 1024;  // free user bytes 
				break; 
			default:
				val = -1;	//bad parameter
				break;
		}
	
	//Win 2000, XP, Vista	
/*	} else {
		
		MEMORYSTATUSEX statex;


  		statex.dwLength = sizeof (statex);
  		GlobalMemoryStatusEx (&statex);

		switch (n){
			case 0:
				val = statex.dwMemoryLoad; // percent of memory in use 	
				break;
			case 1:
				val = (int)(statex.ullTotalPhys / 1024);     // bytes of physical memory 
				break; 
			case 2:
				val = (int)(statex.ullAvailPhys / 1024);     // free physical memory bytes 	
				break; 
			case 3:
				val = (int)(statex.ullTotalPageFile / 1024); // bytes of paging file 
				break; 
			case 4:
				val = (int)(statex.ullAvailPageFile / 1024); // free bytes of paging file 
				break; 
			case 5:
				val = (int)(statex.ullTotalVirtual / 1024);  // user bytes of address space 
				break; 
			case 6:
				val = (int)(statex.ullAvailVirtual / 1024);  // free user bytes 
				break; 
			default:
				val = -1;	//bad parameter
				break;
		}
	}*/

	return val;
}
 
 
static INT_PTR fm_checkArrayAlloc(INT_PTR nr){

	int val;

	if (!IS_OK_ARRAY_NR(nr)) val=0;
	else if (Array[nr]==NULL) val=0;
	else if (ArrayID[nr]!=0) val=2;
	else val=1;
	
	return val;
}


static INT_PTR fm_getArrayAddress(INT_PTR nr){

	if (IS_ALLOCATED_ARRAY(nr))
		return (INT_PTR)Array[nr];
	else
		return 0;
}

static INT_PTR fm_getBufferAddress(INT_PTR nr){

	if (nr == 0) //Source Buffer
		return (INT_PTR)MyAdjustedInData;
	else if (nr == 1) //T1
		return (INT_PTR)fmc.tbuf;
	else if (nr == 2) //T2
		return (INT_PTR)fmc.t2buf;
	else if (nr == 3) //Output Buffer
		return (INT_PTR)gStuff->outData;
	else if (nr == 4) //T3
		return (INT_PTR)fmc.t3buf;
	else if (nr == 5) //T4
		return (INT_PTR)fmc.t4buf;
	else
		return 0;
}


INT_PTR fm_setCtlScripting (INT_PTR n, INT_PTR state)
{
    
    if ( !IS_INUSE_CTL(n) ) {
        return FALSE;
    }

    if (state != 0) state = -1;
    gParams->ctl[n].scripted = (int)state;

    return true;
}


static INT_PTR fm_checkScriptVal(INT_PTR n)
{
    
    if (n < 0 || n >= 256) {
        return false;
    }
#ifdef SCRIPTABLE
    return ScriptParamUsed[n];
#else
    return 0;
#endif

}

static INT_PTR fm_enableScriptVal(INT_PTR n, INT_PTR state)
{
    
    if (n < 0 || n >= 256) {
        return false;
    }

    if (state !=0 ) state = -1;
#ifdef SCRIPTABLE
    ScriptParamUsed[n] = (BOOL)state;
#endif

    return true;
}

static INT_PTR fm_getScriptVal(INT_PTR n)
{
    
    if (n < 0 || n >= 256) {
        return 0;
    }
#ifdef SCRIPTABLE
    return ScriptParam[n];
#else
    return 0;
#endif

}

static INT_PTR fm_setScriptVal(INT_PTR n, INT_PTR val)
{
    
    if (n < 0 || n >= 256) {
        return false;
    }
#ifdef SCRIPTABLE
    ScriptParam[n] = (int)val;
#endif
    return true;
}


static INT_PTR fm_getDialogHandle(void)
{
    return (INT_PTR)fmc.hDlg;
}

static INT_PTR fm_getCtlHandle(INT_PTR n)
{
    if ( !IS_INUSE_CTL(n) ) {
        return 0;
    }

    return (INT_PTR)gParams->ctl[n].hCtl;
}




#ifdef THEPLUGINSITE
	#include "ThePluginSiteFunctions.h"
#endif

#ifdef IMGSOURCE
	//#include "ImgSource22/_ISource.h"
	#include "ThePluginSiteImgSrc.h"
#endif


#ifdef WATERMARK
	#include "ThePluginSiteWatermark.h"
#endif //WATERMARK


//Martijn's setPixel functions
#include "fmdib.h"










//*******************************************************



static INT_PTR fm_cnvX(INT_PTR k, INT_PTR off, INT_PTR d,
                       INT_PTR pGetf, INT_PTR x, INT_PTR y, INT_PTR z)
{
    INT_PTR i;
    INT_PTR sum = 0;
    INT_PTR n = k*2 + 1;

#if 0
    //check for valid indexing into cell[]...
    if (n + off > N_CELLS) {
        return -1;  //invalid
    }
    else if (n + off < 0) {
        return -1;  //invalid
    }
#endif

#if 0
    if (d == 0)
    {   // default divisor is sum of weights...
        for (i = 0; i < n; i++) {
            d += fmc.cell[(off + i) & (N_CELLS-1)];
        }
        if (d == 0)
            return -1; // give up if divisor still 0
    }
#endif

    for (i = 0; i < n; i++) {
        sum += fmc.cell[(off + i) & (N_CELLS-1)] * ((fmf3)pGetf)(x - k + i, y, z);
    }
    return sum / d;
} /*fm_cnvX*/

static INT_PTR fm_cnvY(INT_PTR k, INT_PTR off, INT_PTR d,
					   INT_PTR pGetf, INT_PTR x, INT_PTR y, INT_PTR z)
{
    INT_PTR i;
    INT_PTR sum = 0;
    INT_PTR n = k*2 + 1;

#if 0
    //check for valid indexing into cell[]...
    if (n + off > N_CELLS) {
        return -1;  //invalid
    }
    else if (n + off < 0) {
        return -1;  //invalid
    }
#endif

#if 0
    if (d == 0)
    {   // default divisor is sum of weights...
        for (i = 0; i < n; i++) {
            d += fmc.cell[(off + i) & (N_CELLS-1)];
        }
        if (d == 0)
            return -1; // give up if divisor still 0
    }
#endif

    for (i = 0; i < n; i++) {
        sum += fmc.cell[(off + i) & (N_CELLS-1)] * ((fmf3)pGetf)(x, y - k + i, z);
    }
    return sum / d;
} /*fm_cnvY*/

#if 0 //old version of fm_abort()
INT_PTR fm_abort(void) {
#ifdef _WIN64	//delete this workaround when SEH is implemented in X64
#if 1
	//Structured exception handling is not yet implemented for X64,
	//so we call X86_abort() instead to do a longjmp...
	///X86_abort("Processing aborted...");
	//But X86_abort() is not defined in STANDALONE, so we replicate
	//the body of X86_abort here...
#if 0	//no error message
	char *msg = "Processing aborted...";
#ifdef FULL
	PRINTF2("\r\nX86 Fatal Error: - %s\r\n", msg);
#endif
    msgBox(MB_ICONERROR, "X86 Fatal Error", msg);
#endif //error message
    ///gParams->fValidCompile = FALSE;
    longjmp(Mde_error_exit_jb, 1);
#else	//Harry's version...
    gResult = userCanceledErr;
#endif
#else ///Win32
    RaiseException(STATUS_CONTROL_C_EXIT,         // exception code 
        0,                    // continuable exception 
        0, NULL);             // no arguments 
#endif
    return 0;  //no return expected
} /*fm_abort*/

#else //new version of fm_abort()
INT_PTR fm_abort(void) {
	switch (fmc.abort_mode) {
#ifndef _WIN64
	default:
		//Win32: default is case 0
#endif
	case 0:
		//Original version: Throw an exception (requires SEH)
		RaiseException(STATUS_CONTROL_C_EXIT,         // exception code 
			0,                    // continuable exception 
			0, NULL);             // no arguments 
		break;
	case 1:
		//Harry's version...
		gResult = userCanceledErr;
		break;
	case 2:
		//display message box "Processing aborted..."; then
		//longjmp back to most recent setjmp;
		{//scope
			char *msg = "Processing aborted...";
#ifdef FULL
			PRINTF2("\r\nX86 Fatal Error: - %s\r\n", msg);
#endif
			msgBox(MB_ICONERROR, "X86 Fatal Error", msg);
		}//scope
		//fall through to case 3...
#ifdef _WIN64
	default:
		//Win64: default is case 3
#endif
	case 3:
		//longjmp back to most recent setjmp;
		longjmp(Mde_error_exit_jb, 1);
		break;
	}//switch
    return userCanceledErr;  //no return expected
} /*fm_abort*/
#endif //new version of fm_abort

static INT_PTR fm_testAbort(void) {
    if (fmc.doingProxy) {
        //check for any character typed...
        MSG msg;
        if (PeekMessage(
            &msg,   // pointer to structure for message
            NULL,   // handle to window (NULL => any window belonging to this thread)
            WM_KEYDOWN,     // first message
            WM_KEYDOWN,     // last message
            PM_REMOVE       // removal flags
           ))
        { 
            if (msg.wParam == VK_ESCAPE) {
#if 0
                // User pressed escape -- ask if he really wants
                // to abort.
                if (YesNo("Abort filter operation?") == IDYES)
                    return userCanceledErr;
                else
                    return 0;
#else
                // Unconditional abort.
                return userCanceledErr;
#endif
            }
            else if (msg.wParam == VK_PAUSE) {
                // User pressed pause -- ask if he wants
                // to continue or abort.
#if 0
                if (Warn("Filter operation paused...\n\n"
                         "Select OK to continue\n"
                         "or Cancel to abort.") == IDCANCEL)
#else
                if (msgBox(MB_ICONWARNING|MB_YESNO,
                           "Pause",//"FilterMeister Pause",
                           "Continue?") == IDNO)
#endif
                    return userCanceledErr;
                else
                    return 0;
            }
#if 0
            else if (msg.wParam == VK_APPS) {
                // Context menu key ==> unconditional abort...
                return userCanceledErr;
            }
#endif
            // Ignore any other key-down
            return 0;
        };
    }
    else {
        // Running filter on main image.
        if (TestAbort ())
        {
            gResult = userCanceledErr;
            //longjmp()
        }
    }
    return gResult;
} //fm_testAbort

INT_PTR fm_updateProgress(INT_PTR n, INT_PTR d) {
	// Set progress bar to indicate (n*100)/(d-1) percent complete.
	// We assume that n has values 0, 1,...,d-1.
	// (The -1 makes a difference when d is very small, such as
	// when the image height is < 10 or so ... in which case
	// the bar would only reach 90% or less at maximum without
	// the -1).  Of course, we need to make sure d doesn't get
	// TOO small!
	d = d - 1; //so max range will reach 100%
    if (d <= 0) d = 1; //but don't let d get TOO small!
    if (fmc.doingProxy) {
        //update progress bar for proxy....
		//default progress bar range is 0..100
        SendMessage(GetDlgItem(MyHDlg, IDC_PROGRESS1), PBM_SETPOS, (n*100)/d, 0);
        fmc.pre_ctl[CTL_PROGRESS] = (int)((n*100)/d); //update cached values for reading progress bar value with ctl()
        return fm_testAbort();
    }
    else {
        UpdateProgress((int)n, (int)d);
        if (TestAbort ())
        {
            gResult = userCanceledErr;
            //longjmp()
        }
    }
    return gResult;
} //fm_updateProgress

static INT_PTR fm_updatePreview(INT_PTR n) {
    //n is ignored for now ,,, it will be the id of the preview
    //control to be updated.
    if (fmc.doingProxy) {
        //update progress bar for proxy....
        // thread is tied up computing the filter, so
        // WM_PAINT doesn't get processed in message loop.
        UpdateProxy(MyHDlg, FALSE);
        // So update the proxy immediately!
        ShowOutputBuffer(globals, MyHDlg, FALSE /*fDragging*/);
        return 1;  //doingProxy
    }
    return 0;   //not doingProxy, so no update
} //fm_updatePreview

/**********************************************************************/
/* getFilterInstallDir()
/*
/*  Returns a string containing the full path of the directory from
/*  which this filter was loaded (which is presumably the installation
/*  directory).
/*
/*  As a side effect, this function also copies the path name into
/*  the predefined string filterInstallDir[256].
/*
/**********************************************************************/
char *fm_getFilterInstallDir(void)
{
    int i;
    DWORD ret = GetModuleFileName(hDllInstance,
                                  fmc.filterInstallDir,
                                  sizeof(fmc.filterInstallDir));
    // Locate the final '\', if any...
    for (i = ret - 1; i >= 0 && fmc.filterInstallDir[i] != '\\'; --i)
        ;
    // Truncate string after final '\' (or set to null string if none).
    fmc.filterInstallDir[i+1] = '\0';
    return fmc.filterInstallDir;
} /*fm_getFilterInstallDir*/

/**********************************************************************/
/* appendEllipsis(s)
/*
/*  Appends an ellipsis (...) to a string.
/*
/*  Caution: The string returned by appendEllipsis() resides in static
/*           storage and will be overwritten by the next call to
/*           appendEllipsis().  If you need the appended string to
/*           persist, copy it to your own storage.
/*
/**********************************************************************/
char *fm_appendEllipsis(const char *s)
{
    THREAD_LOCAL static char f[512];        //Warning: static!!!!

    strncpy(f, s, sizeof(f)-4); //leave room for ellipsis
    f[sizeof(f)-4] = '\0';      //ensure f is null-terminated.
    return strcat(f, "...");
} /*fm_appendEllipsis*/

/**********************************************************************/
/* stripEllipsis(s)
/*
/*  Strips a trailing ellipsis (if any) from a string.
/*
/*  Caution: The string returned by stripEllipsis() resides in static
/*           storage and will be overwritten by the next call to
/*           stripEllipsis().  If you need the stripped string to
/*           persist, copy it to your own storage.
/*
/**********************************************************************/
char *fm_stripEllipsis(const char *s)
{
    THREAD_LOCAL static char f[512];
    int n = (int)strlen(s);
    if (n >= 3 && strcmp(&s[n-3], "...") == 0) {
        // strip ellipsis from end of s
        n -= 3;
    }
    n = min(n, sizeof(f)-1);    //truncate to fit in f if necessary
    strncpy(f, s, n);           //copy n chars to f
    f[n] = '\0';                //add terminating null
    return f;
} /*fm_stripEllipsis*/

/**********************************************************************/
/* formatString(s)
/*
/*  Formats a string by interpolating the following special codes:
/*
/*       !A = Author (filterAuthorText)
/*       !a = About (filterAboutText)
/*       !C = Category (filterCategoryText)
/*       !c = Copyright (filterCopyrightText)
/*       !D = Description (filterDescriptionText)
/*   **  !d = Document title (from propTitle (titl)) -> documentTitleText
/*   XX  !d = current date (oops, used above)
/*       !F = Filename (filterFilenameText)
/*       !f = Filter case (filterCaseText)
/*       !H = Host (filterHostText)
/*       !h = Image height (imageHeight)
/*   **  !I = filterInstallDir
/*   **  !i = filterImageTitle (same as documentTitleText ??)
/*       !M = Image Mode (filterImageModeText)
/*       !m = Image Mode (numeric, from imageMode)
/*       !O = Organization (filterOrganizationText)
/*   **  !S = Serial string (hostSerialString)
/*   **  !s = Binary serial number (hostSerialNumber)
/*       !T = Title (filterTitleText)
/*       !t = Title with any trailing ellipsis stripped
/*   XX  !t = current time (oops, used above)
/*       !U = URL (filterURLText)
/*       !V = Version (filterVersionText)
/*   **  !v = FilterMeister Version string???
/*       !w = Image width (imageWidth)
/*		 !X = Host architecture ("32" or "64")
/*		 !x = Target architecture ("32" or "64")
/*       !Y - current year (e.g. 2003) -- for use in Copyright
/*       !z = Proxy zoom factor (zoomFactor)
/*       !! = !
/*
/* ** - Not yet implemented.
/*
/*  Also translates many of the latest HTML entity strings to
/*  their corresponding ISO codes (or sometimes fakes the
/*  translation by using low-valued ASCII codes or a combination
/*  of ASCII characters).  An HTML entity string has the general
/*  format '&entity;'.  It must begin with the ampersand character (&)
/*  and end with a semicolon (;). The string 'entity' between the
/*  & and the ; must be a sequence of two or more ASCII lowercase
/*  or uppercase letters (the string is case sensitive).
/*
/*  HTML numeric character references (both decimal and hexadecimal)
/*  are also recognized and translated.  The decimal format is
/*  '&#D;', where D is a decimal number which refers to the ISO 10646 
/*  decimal character number D.  The hexadecimal syntax is '&#xH;' or 
/*  '&#XH;', where H is a hexadecimal number which refers to the ISO 
/*  10646 hexadecimal character number H.  Hexadecimal numbers in 
/*  numeric character references are case-insensitive. 
/*
/*  Here are some examples of numeric character references:
/*
/*	&#229;		(in decimal) represents the letter "a" with a small circle 
/*				above it (used, for example, in Norwegian). 
/*	&#xE5;		(in hexadecimal) represents the same character. 
/*	&#Xe5;		(in hexadecimal) represents the same character as well. 
/*	&#1048;		(in decimal) represents the Cyrillic capital letter "I". 
/*	&#x6C34;	(in hexadecimal) represents the Chinese character for water. 
/*
/*  As examples of HTML character entity references, FM will translate the 
/*  following HTML entities as indicated:
/*
/*      Entity  Translation Meaning
/*      ------  ----------- ------------------------------------
/*		&quot;	    "       double quote
/*		&amp;	    &       ampersand
/*		&apos;	    '       apostrophe
/*		&lt;	    <       less than
/*		&gt;	    >       greater than
/*		&nbsp;		        non-breaking space
/*		&iexcl;		       inverted exclamation mark
/*      &iquest;           inverted question mark
/*		&cent;		       cent sign
/*		&pound;		       pound sign
/*		&copy;		       copyright
/*      &reg;              registered trademark
/*		&trade;     tm      trademark
/*		&yen;		       Yen sign
/*      &euro;             Euro sign
/*      
/*  For a complete list of all HTML entities currently translated by 
/*  FM, see the code below.
/*
/*  Note that, in order to represent a bare ampersand (&), you can
/*  code the entity string '&amp;'.  Since this is rather cumbersome,
/*  FM will also recognize and translate two successive ampersands
/*  (&&) into a single ampersand (&) in the formatted string.
/*  Furthermore, any occurrence of an ampersand (&) that does not
/*  initiate a recognized HTML entity will also be translated as
/*  a bare ampersand (&).
/*
/*  For example, the following strings:
/*
/*      "This is an ampersand:'&amp;'; &amp; so is this: &amp;."
/*      "This is an ampersand:'&&'; && so is this: &&."
/*      "This is an ampersand:'&'; & so is this: &."
/*
/*  will all be formatted as:
/*
/*      "This is an ampersand:'&'; & so is this: &."
/*
/*  For possible future implementation:
/*
/*       !<n> = value of ctl(<n>) where <n> is a decimal number
/*              (e.g., for updating a static text control?)
/*
/*  Note: The formatting is nonrecursive, to prevent inadvertent
/*        infinite loops.
/*
/*  Caution: The string returned by formatString() has a maximum
/*           length of 1024 chars (not including the terminating
/*           NUL), and will be truncated if necessary.  It resides 
/*           in (thread-local) static storage and will be overwritten
/*           by the next call to formatString() in the current thread.
/*           If you need the formatted string to persist, copy it to 
/*           your own storage.
/*
/**********************************************************************/

char *fm_formatString(const char *s)
{
    THREAD_LOCAL static char f[1024];
    unsigned int cbLeft = sizeof(f) - 1;
    unsigned int cbCopied;
	int n;
    char *p = f;
    char *q;
    char buffer[32];

    while (*s) {
        if (*s == '!') {
            s++;
            switch (*s) {
            case 'A':
                q = fmc.filterAuthorText;
                break;
            case 'a':
                q = gParams->filterAboutText;
                break;
            case 'C':
                q = fmc.filterCategoryText;
                break;
            case 'c':
                q = fmc.filterCopyrightText;
                break;
            case 'D':
                q = fmc.filterDescriptionText;
                break;
            case 'F':
                q = fmc.filterFilenameText;
                break;
            case 'f':
                q = fmc.filterCaseText;
                break;
            case 'H':
                q = fmc.filterHostText;
                break;
            case 'h':
                q = _itoa(fmc.imageSize32.v, buffer, 10);	//image height
                break;
            case 'M':
                q = fmc.filterImageModeText;
                break;
            case 'm':
                q = _itoa(fmc.imageMode, buffer, 10);
                break;
            case 'O':
                q = fmc.filterOrganizationText;
                break;
            case 'T':
                q = fmc.filterTitleText;
                break;
            case 't':
                q = fm_stripEllipsis(fmc.filterTitleText);
                break;
            case 'U':
                q = fmc.filterURLText;
                break;
            case 'V':
                q = fmc.filterVersionText;
                break;
            case 'w':
                q = _itoa(fmc.imageSize32.h, buffer, 10);	//image width
                break;
			case 'X':
#ifdef _WIN64
				q = "64";		//Host app running on Win64
#elif defined(WIN32)
				q = "32";		//Host app running on Win32
#else
				q = "";			//Host architecture unknown.
#endif
				break;
			case 'x':
				if (X86GEN) {
					q = "32";	//generating 32-bit x86 code
				}
				else if (X64GEN) {
					q = "64";	//generating 64-bit x64 code
				}
				else {
					q = "86";	//x86 code size not specified??
				}
				break;
            case 'Y':
                {   time_t tm = time(NULL);
                    q = _itoa(localtime(&tm)->tm_year + 1900, buffer, 10);
                }
                break;
            case 'z':
                q = _itoa(fmc.zoomFactor, buffer, 10);
                break;
            case '!':
                q = "!";
                break;
            default:
                //copy a '!' to the formatted string
                q = "!";
                s--;    //back up one char to rescan the char after the '!'
                break;
            } //switch
            //copy the portion of the interpolated string
            //that will fit in the formatted string...
            cbCopied = min(cbLeft, (UINT)strlen(q));
            strncpy(p, q, cbCopied);
            p += cbCopied;
            cbLeft -= cbCopied;
        }
		else if (*s == '&') {
			//possible HTML entity...

			//define a macro to compare against a named entity
			#define ENTITY(entity, subst) \
				if (!strncmp(s, (entity), n = sizeof(entity)-1)) {\
					q = (subst); goto subst_entity;\
				}

            /* Cases are sorted by (first char of) entity name for faster lookup */
            switch (s[1]) {
			case '#':
				/* Numeric character reference */
				//we don't accept UNICODE numeric entities above 0x00ff for now...
				#define MAX_NUMERIC_ENTITY_VAL	255
				{
					unsigned int iVal = 0;
					const char *s2 = s + 2; // point past '#'
					n = 2;	// chars eaten so far
					if (*s2 == 'x' || *s2 == 'X') {
						/* hexadecimal numeric entity */
						s2++;
						n++;
						if (!isxdigit(*s2)) break;	// error, must be at least one digit
						while (isxdigit(*s2)) {
							iVal = iVal * 16 + ( isdigit(*s2) ? *s2 - '0' : (*s2 & 7) + 9 );
							// check for overflow
							if (iVal > MAX_NUMERIC_ENTITY_VAL) break;	// error, value too large
							s2++;
							n++;
						}//while hex digit
					}
					else {
						/* decimal numeric entity */
						if (!isdigit(*s2)) break;	// error, must be at least one digit
						while (isdigit(*s2)) {
							iVal = iVal * 10 + (*s2 - '0');
							// check for overflow
							if (iVal > MAX_NUMERIC_ENTITY_VAL) break;	// error, value too large
							s2++;
							n++;
						}//while decimal digit
					}
					if (*s2 != ';') break;	// Error, no final ';'
					n++;
					// disallow certain graphic values
					if (iVal < 0x00 ||		// NUL allowed?
						iVal == 0x7f ||		// 0x7f reserved for multibyte leader?
						iVal == 0x81 ||		// 0x81 reserved for multibyte????
						iVal > 0xff )		// only allow 8-bit codes? TRANSLATE other ISO-xxxx to Windows charset???
											// Could normalize by stripping leading 0's, then lookup e.g. "&#8084;"
											//  -- nahh, might as well lookup binary val directly.
					{
						iVal = 0x81;		// graphic to represent non-displayable coe?
					}
					q = (char *)&iVal;		// treat as NUL-terminated string. (tricky cast!!!)
				}
				cbCopied = 1;				// copy exactly 1 byte, which may be a NUL byte
				goto subst_entity2;
				#undef MAX_NUMERIC_ENTITY_VAL
				break;
            case '&':
			    ENTITY("&&",		"&");		// & (&& => verbatim &)
                break;
            case 'A':
			    ENTITY("&AElig;",	"\xC6");	// 
			    ENTITY("&Aacute;",	"\xC1");	// 
			    ENTITY("&Acirc;",	"\xC2");	// 
			    ENTITY("&Agrave;",	"\xC0");	// 
			    ENTITY("&Alpha;",	"A");		// A (Greek Alpha, use Latin A)
			    ENTITY("&Aring;",	"\xC5");	// 
			    ENTITY("&Atilde;",	"\xC3");	// 
			    ENTITY("&Auml;",	"\xC4");	// 
                break;
            case 'B':
				ENTITY("&BEL;",		"\x07");	// BEL (ASCII Control code, not ISO)
				ENTITY("&BS;",		"\x08");	// BS (ASCII Control code, not ISO)
			    ENTITY("&Beta;",	"B");		// B (Greek Beta, use Latin B)
                break;
            case 'C':
				ENTITY("&CR;",		"\x0D");	// CR (ASCII Control code, not ISO)
				ENTITY("&CRLF;",	"\r\n");	// CR/LF (ASCII Control code sequence, not ISO)
			    ENTITY("&Ccedil;",	"\xC7");	// 
			    ENTITY("&Chi;",	    "X");		// X (Greek Chi, use Latin X)
			    ENTITY("&Colon;",	"::");		// :: (double colon)
                break;
            case 'D':
			    ENTITY("&Dagger;",	"\x87");	//  (double dagger)
                break;
            case 'E':
			    ENTITY("&ESC;",		"\x1B");	// ESC (ASCII control code, not ISO entity)
			    ENTITY("&ETH;",		"\xD0");	// 
			    ENTITY("&Eacute;",	"\xC9");	// 
			    ENTITY("&Ecirc;",	"\xCA");	// 
			    ENTITY("&Egrave;",	"\xC8");	// 
			    ENTITY("&Epsilon;",	"E");		// E (Greek Epsilon, use Latin E)
			    ENTITY("&Eta;",	    "H");		// H (Greek Eta, use Latin H)
			    ENTITY("&Euml;",	"\xCB");	// 
			    ENTITY("&Exclam;",	"\x13");	// OEM (double exclamation mark !!)
                break;
			case 'F':
				ENTITY("&FF;",		"\x0c");	// FF (ASCII Control code, not ISO)
                break;
			case 'H':
				ENTITY("&HT;",		"\x09");	// HT (ASCII Control code, not ISO)
				break;
            case 'I':
			    ENTITY("&Iacute;",	"\xCD");	// 
			    ENTITY("&Icirc;",	"\xCE");	// 
			    ENTITY("&Igrave;",	"\xCC");	// 
			    ENTITY("&Iota;",	"I");		// I (Greek Iota, use Latin I)
			    ENTITY("&Iuml;",	"\xCF");	// 
                break;
            case 'K':
			    ENTITY("&Kappa;",	"K");		// K (Greek Kappa, use Latin K)
                break;
            case 'L':
				ENTITY("&LF;",		"\x0A");	// LF (ASCII Control code, not ISO)
                break;
            case 'M':
			    ENTITY("&Mu;",	    "M");		// M (Greek Mu, use Latin M)
                break;
            case 'N':
				///ENTITY("&NUL;",		"\x00");	// NUL (ASCII Control code, not ISO)	/// NFG!!!
			    ENTITY("&Ntilde;",	"\xD1");	// 
			    ENTITY("&Nu;",	    "N");		// N (Greek Nu, use Latin N)
                break;
            case 'O':
			    ENTITY("&OElig;",	"\x8C");	//  (capital ligature OE)
			    ENTITY("&Oacute;",	"\xD3");	// 
			    ENTITY("&Ocirc;",	"\xD4");	// 
			    ENTITY("&Ograve;",	"\xD2");	// 
			    ENTITY("&Omicron;",	"O");		// O (Greek Omicron, use Latin O)
			    ENTITY("&Oslash;",	"\xD8");	// 
			    ENTITY("&Otilde;",	"\xD5");	// 
			    ENTITY("&Ouml;",	"\xD6");	// 
                break;
            case 'P':
			    ENTITY("&Prime;",	"\"");		// " (seconds, inches; use &quot;)
                break;
            case 'R':
			    ENTITY("&Rho;",	    "P");		// P (Greek Rho, use Latin P)
                break;
            case 'S':
			    ENTITY("&Scaron;",	"\x8A");	//  (capital S with caron)
                break;
            case 'T':
			    ENTITY("&THORN;",	"\xDE");	// 
			    ENTITY("&Tau;",	    "T");		// T (Greek Tau, use Latin T)
                break;
            case 'U':
			    ENTITY("&Uacute;",	"\xDA");	// 
			    ENTITY("&Ucirc;",	"\xDB");	// 
			    ENTITY("&Ugrave;",	"\xD9");	// 
			    ENTITY("&Upsilon;",	"Y");		// Y (Greek Upsilon, use Latin Y)
			    ENTITY("&Uuml;",	"\xDC");	// 
                break;
            case 'V':
				ENTITY("&VT;",		"\x0B");	// VT (ASCII Control code, not ISO)
                break;
            case 'Y':
			    ENTITY("&Yacute;",	"\xDD");	// 
			    ENTITY("&Yuml;",	"\x9F");	//  (capital Y with umlaut/diaeresis)
                break;
            case 'Z':
			    ENTITY("&Zcaron;",	"\x8E");	//  (capital Z with caron, not official?)
			    ENTITY("&Zeta;",	"Z");		// Z (Greek Zeta, use Latin Z)
                break;
            case 'a':
			    ENTITY("&aacute;",	"\xE1");	// 
			    ENTITY("&acirc;",	"\xE2");	// 
			    ENTITY("&acute;",	"\xB4");	// 
			    ENTITY("&aelig;",	"\xE6");	// 
			    ENTITY("&agrave;",	"\xE0");	// 
			    ENTITY("&amp;",		"\x26");	// &
			    ENTITY("&apos;",	"\x27");	// ' (not official HTML)
			    ENTITY("&aring;",	"\xE5");	// 
			    ENTITY("&atilde;",	"\xE3");	// 
			    ENTITY("&auml;",	"\xE4");	// 
                break;
            case 'b':
			    ENTITY("&bdquo;",	"\x84");	//  (double low-9 quotation mark)
#if 0
			    ENTITY("&beamedsixteenthnotes;", "\x0E");   // OEM (beamed 16th notes)
#endif
			    ENTITY("&beta;",	"\xDF");	//  (Greek beta, use &szlig; -- yecch!)
			    ENTITY("&blk14;",	"\x0F");	// OEM LIGHT SHADE  
			    ENTITY("&boxdl;",	"\x02");	// OEM BOX DRAWINGS LIGHT DOWN AND LEFT  
			    ENTITY("&boxdr;",	"\x01");	// OEM BOX DRAWINGS LIGHT DOWN AND RIGHT  
			    ENTITY("&boxh;",	"\x06");	// OEM BOX DRAWINGS LIGHT HORIZONTAL  
			    ENTITY("&boxhd;",	"\x16");	// OEM BOX DRAWINGS LIGHT HORIZONTAL AND DOWN
			    ENTITY("&boxhu;",	"\x15");	// OEM BOX DRAWINGS LIGHT HORIZONTAL AND UP
			    ENTITY("&boxul;",	"\x04");	// OEM BOX DRAWINGS LIGHT UP AND LEFT  
			    ENTITY("&boxur;",	"\x03");	// OEM BOX DRAWINGS LIGHT UP AND RIGHT  
			    ENTITY("&boxv;",	"\x05");	// OEM BOX DRAWINGS LIGHT VERTICAL  
			    ENTITY("&boxvh;",	"\x10");	// OEM BOX DRAWINGS LIGHT VERTICAL AND HORIZONTAL  
			    ENTITY("&boxvl;",	"\x17");	// OEM BOX DRAWINGS LIGHT VERTICAL AND LEFT  
			    ENTITY("&boxvr;",	"\x19");	// OEM BOX DRAWINGS LIGHT VERTICAL AND RIGHT  
			    ENTITY("&brvbar;",	"\xA6");	// 
			    ENTITY("&bull;",	"\x95");	//  (use  for bullet)
			    ENTITY("&bull2;",	"\x07");	// OEM (bullet style 2)
				ENTITY("&bull3;",	"\x7F");	//  (unofficial; narrow unfilled rectangle)
				ENTITY("&bull4;",	"\x81");	//  (unofficial; wide unfilled rectangle)
                break;
            case 'c':
			    ENTITY("&ccedil;",	"\xE7");	// 
			    ENTITY("&cedil;",	"\xB8");	// 
			    ENTITY("&cent;",	"\xA2");	// 
			    ENTITY("&circ;",	"\x88");	//  (modifier letter circumflex accent)
			    ENTITY("&colone;",	":=");		// := (colon equal)
			    ENTITY("&copy;",	"\xA9");	// 
			    ENTITY("&crarr;",	"\x11\x04");	// OEM (subst. for carriage return arrow)
			    ENTITY("&curren;",	"\xA4");	// 
                break;
            case 'd':
			    ENTITY("&dagger;",	"\x86");	//  (dagger)
			    ENTITY("&deg;",		"\xB0");	// 
			    ENTITY("&divide;",	"\xF7");	// 
                break;
            case 'e':
			    ENTITY("&eacute;",	"\xE9");	// 
			    ENTITY("&ecirc;",	"\xEA");	// 
			    ENTITY("&egrave;",	"\xE8");	// 
			    ENTITY("&empty;",	"\xD8");	//  (use capital O-slash for empty set)
			    ENTITY("&emsp;",	" ");		//   (em-space, use normal space)
			    ENTITY("&ensp;",	" ");		//   (en-space, use normal space)
			    ENTITY("&eth;",		"\xF0");	// 
			    ENTITY("&euml;",	"\xEB");	// 
			    ENTITY("&euro;",	"\x80");	//  (use  for euro sign)
                break;
            case 'f':
			    ENTITY("&female;",	"\x0C");	// OEM (female)
			    ENTITY("&fnof;",	"\x83");	//  (f with hook)
			    ENTITY("&frac12;",	"\xBD");	// 
			    ENTITY("&frac14;",	"\xBC");	// 
			    ENTITY("&frac34;",	"\xBE");	// 
			    ENTITY("&frasl;",	"/");		// / (use / for fraction slash)
                break;
            case 'g':
			    ENTITY("&ge;",		">=");		// >= (synthetic)
			    ENTITY("&gt;",		"\x3E");	// >
                break;
            case 'h':
			    ENTITY("&hellip;",	"\x85");	//  (horizontal ellipsis)
                break;
            case 'i':
			    ENTITY("&iacute;",	"\xED");	// 
			    ENTITY("&ibull2;",	"\x08");	// OEM (inverse bullet style 2)
			    ENTITY("&icirc;",	"\xEE");	// 
			    ENTITY("&iexcl;",	"\xA1");	// 
			    ENTITY("&igrave;",	"\xEC");	// 
			    ENTITY("&image;",	"I");		// I imaginary part (blackletter I)
			    ENTITY("&iquest;",	"\xBF");	// 
			    ENTITY("&iuml;",	"\xEF");	// 
                break;
            case 'l':
			    ENTITY("&laquo;",	"\xAB");	// 
			    ENTITY("&larr;",	"\x1B");	// OEM (left arrow)
			    ENTITY("&ldquo;",	"\x93");	//  (left double quotation mark)
			    ENTITY("&le;",		"<=");		// <= (synthetic)
			    ENTITY("&lowast;",	"*");		// * (use asterisk for low asterisk)
			    ENTITY("&lsaquo;",	"\x8B");	//  (single left angle quotation)
			    ENTITY("&lsquo;",	"\x91");	//  (left single quotation mark)
			    ENTITY("&lt;",		"\x3C");	// <
			    ENTITY("&ltrif;",	"\x11");	// OEM (left-pointing triangle, filled)
                break;
            case 'm':
			    ENTITY("&macr;",	"\xAF");	//  (spacing macron)
			    ENTITY("&male;",	"\x0B");	// OEM (male)
			    ENTITY("&mdash;",	"\x97");	//  (em dash)
			    ENTITY("&micro;",	"\xB5");	// 
			    ENTITY("&middot;",	"\xB7");	// 
			    ENTITY("&minus;",	"\x96");	// - (use en-dash for minus symbol)
			    ENTITY("&mu;",	    "\xB5");	//  (mu, use &micro;)
                break;
            case 'n':
			    ENTITY("&nbsp;",	"\xA0");	//	 (no-break space)
			    ENTITY("&ndash;",	"\x96");	//  (en dash)
			    ENTITY("&ne;",		"!=");		// ? (not equal, use !=)
			    ENTITY("&not;",		"\xAC");	// 
			    ENTITY("&ntilde;",	"\xF1");	// 
                break;
            case 'o':
			    ENTITY("&oacute;",	"\xF3");	// 
			    ENTITY("&ocirc;",	"\xF4");	// 
			    ENTITY("&oelig;",	"\x9C");	//  (small ligature oe)
			    ENTITY("&ograve;",	"\xF2");	// 
			    ENTITY("&oline;",	"\xAF");	//  (overline, use macron)
			    ENTITY("&omicron;",	"o");	    // o (Greek omicron, use Latin o)
			    ENTITY("&ordf;",	"\xAA");	// 
			    ENTITY("&ordm;",	"\xBA");	// 
			    ENTITY("&oslash;",	"\xF8");	// 
			    ENTITY("&otilde;",	"\xF5");	// 
			    ENTITY("&ouml;",	"\xF6");	// 
                break;
            case 'p':
			    ENTITY("&para;",	"\xB6");	// 
			    ENTITY("&permil;",	"\x89");	//  (per mille)
			    ENTITY("&pilcrow;",	"\x14");	// OEM (pilcrow, unfilled paragaph mark)
			    ENTITY("&plusmn;",	"\xB1");	// 
			    ENTITY("&pound;",	"\xA3");	// 
			    ENTITY("&prime;",	"'");		// ' (minutes, feet; use &apos;)
                break;
            case 'q':
			    ENTITY("&quot;",	"\x22");	// "
                break;
            case 'r':
			    ENTITY("&raquo;",	"\xBB");	// 
			    ENTITY("&rarr;",	"\x1A");	// OEM (right arrow)
			    ENTITY("&rdquo;",	"\x94");	//  (right double quotation mark)
			    ENTITY("&real;",	"R");		// R real part (blackletter R)
			    ENTITY("&reg;",		"\xAE");	// 
			    ENTITY("&rsaquo;",	"\x9B");	//  (single right angle quotation)
			    ENTITY("&rsquo;",	"\x92");	//  (right single quotation mark)
                break;
            case 's':
			    ENTITY("&sbquo;",	"\x82");	//  (single low-9 quotation mark)
			    ENTITY("&scaron;",	"\x9A");	//  (small s with caron)
			    ENTITY("&sdot;",	"");		//  (dot operator)
			    ENTITY("&sect;",	"\xA7");	// 
#if 0
			    ENTITY("&shy;",		"\xAD");	//  (soft hyphen, doesn't work in FM)
#else
			    ENTITY("&shy;",		"");		//  (soft hyphen doesn't break in FM, always use "")
#endif
			    ENTITY("&sim;",		"~");		// ~ (use tilde for 'similar to')
#if 0
				ENTITY("&sup0;",	"\xB0");	//  (unofficial; use &deg; in MS Sans Serif, Courier, Courier New, Fixedsys, System, most fixed-width fonts)
#else
				ENTITY("&sup0;",	"\xBA");	//  (unofficial; use &ordm; in Arial, Lucida Console, most propertional fonts.)
#endif
			    ENTITY("&sup1;",	"\xB9");	// 
			    ENTITY("&sup2;",	"\xB2");	// 
			    ENTITY("&sup3;",	"\xB3");	// 
			    ENTITY("&szlig;",	"\xDF");	// 
                break;
            case 't':
			    ENTITY("&thinsp;",	" ");		//   (thin space, use normal space)
			    ENTITY("&thorn;",	"\xFE");	// 
			    ENTITY("&tilde;",	"\x98");	//  (small tilde)
			    ENTITY("&times;",	"\xD7");	// 
#if 1
			    ENTITY("&trade;",	"\x99");	//  (use  for trademark)
#else
			    ENTITY("&trade;",	"(tm)");	//  (use (tm) for trademark)
#endif
			    ENTITY("&twonotes;", "\x0E");	// OEM (two beamed 16th notes)
                break;
            case 'u':
			    ENTITY("&uacute;",	"\xFA");	// 
			    ENTITY("&uarr;",	"\x18");	// OEM (up arrow)
			    ENTITY("&ucirc;",	"\xFB");	// 
			    ENTITY("&ugrave;",	"\xF9");	// 
			    ENTITY("&uml;",		"\xA8");	// 
			    ENTITY("&uuml;",	"\xFC");	// 
                break;
            case 'v':
			    ENTITY("&varr;",	"\x12");	// OEM (up/down arrow; not ISO)
                break;
            case 'w':
			    ENTITY("&weierp;",	"P");		// P powerset (script P)
                break;
            case 'y':
			    ENTITY("&yacute;",	"\xFD");	// 
			    ENTITY("&yen;",		"\xA5");	// 
			    ENTITY("&yuml;",	"\xFF");	// 
                break;
            case 'z':
			    ENTITY("&zcaron;",	"\x9E");	//  (small z with caron, not official?)
			    ENTITY("&zwj;",		"");		//  zero width joiner (no char)
			    ENTITY("&zwnj;",	"");		//  zero width non-joiner (no char)
                break;
            }//switch
			assert(*s == '&');
			// no entity match: copy the single '&'...
			q = "&"; n = 1;

subst_entity:
            //copy the portion of the entity string value
            //that will fit in the formatted string...
            cbCopied = min(cbLeft, (UINT)strlen(q));
subst_entity2:	//entry when cbCopied has already been computed (to copy NULs)
            memcpy(p, q, cbCopied); //use memcpy, not str(n)cpy, to copy internal NULs.
            p += cbCopied;
            cbLeft -= cbCopied;
			s += n - 1;
			#undef ENTITY
		}
        else if (cbLeft > 0) {
            //copy this character verbatim to the formatted string
            *p++ = *s;
            cbLeft--;
        }
        s++;
    } //while
    *p = '\0';  //terminate the formatted string
    return f;
} /*fm_formatString*/


/**********************************************************************/
/* setCtlText(n, "text")
/*
/*  Sets the text label or content for control n to the given text string.
/*
/**********************************************************************/
INT_PTR fm_setCtlText(INT_PTR n, INT_PTR iText)
{
    // Could check nargs and pass to setCtlTextv() if nargs > 2 ????
    RECT    wRect;

    if (!IS_INUSE_CTL(n)) {
        return FALSE;
    }

#ifndef STANDALONE
    //Override text for CTL_EDIT...
    if (n == CTL_EDIT) {
        iText = (gParams->flags & AFH_EDITMODE) ?
                     (INT_PTR)"&Edit <<<" :
                     (INT_PTR)"&Edit >>>";
    }
#endif

    if (iText) {
        strncpy(gParams->ctl[n].label, (char *)iText, MAX_LABEL_SIZE+1);
        gParams->ctl[n].label[MAX_LABEL_SIZE] = '\0';   //ensure null-terminated.
    }
    else {
        //No text string
        gParams->ctl[n].label[0] = '\0';
    }

    switch (gParams->ctl[n].class) {
    case CC_STATICTEXT:
    case CC_CHECKBOX:
    case CC_RADIOBUTTON: //did someone make the following change to try to allow
						//more than 1023 chars of text in a control??  Doesn't
						//work because (a) fm_formatString will truncate it to
						//1023 anyway, and (b) whenever the text gets refreshed
						//from the .label field, it will be truncated to 1023!!!
        //SetWindowText(gParams->ctl[n].hCtl, fm_formatString(gParams->ctl[n].label));
		SetWindowText(gParams->ctl[n].hCtl, fm_formatString((char *)iText));
        if (gParams->ctl[n].bkColor == (COLORREF) -1) {
            //It seems these controls with a transparent
            //background color do not erase the background
            //when they receive a WM_SETTEXT message
            //(maybe because we are handling WM_CTLCOLOR*
            //and setting the background to transparent??), so
            //we brute-force it... (this may be overkill??)
            if (gParams->ctl[n].enabled){
				GetWindowRect(gParams->ctl[n].hCtl, &wRect); 
				ScreenToClient (fmc.hDlg, (LPPOINT)&wRect);
				ScreenToClient (fmc.hDlg, (LPPOINT)&(wRect.right));
				return RedrawWindow(
						fmc.hDlg,   // handle of entire dialog window
						&wRect, // address of structure with update rectangle
						NULL,   // handle of update region
						RDW_ERASE|RDW_INVALIDATE|RDW_ERASENOW|RDW_UPDATENOW // array of redraw flags
				   );
			}
        }
        return TRUE;
        break;

    case CC_STANDARD:
	case CC_SLIDER:
        //specific to default scrollbars...
        {//scope
            //Resize the static text buddy to exactly fit the text,
            //to minimize problems with overlapping controls...
            SIZE  size;
            RECT  r;
            HWND  hWnd = GetDlgItem(fmc.hDlg, (int)(IDC_BUDDY2+n));
            HDC   hDC = GetDC(hWnd);
            HFONT hOldFont = SelectObject(hDC, gParams->ctl[n].hFont);
            LPSTR p = fm_formatString(gParams->ctl[n].label);
            //CAUTION: p points to volatile storage (which will
            //be overwritten by calls to Info(), etc.)!!
            GetTextExtentPoint32(hDC, p, (int)strlen(p), &size);
            gParams->ctl[n].sb.w2 = size.cx;    //text width in pixels
            gParams->ctl[n].sb.h2 = size.cy;    //text height in pixels
            computeBuddy2Pos(n,
                             gParams->ctl[n].xPos,
                             gParams->ctl[n].yPos,
                             gParams->ctl[n].width,
                             gParams->ctl[n].height,
                             &r);
            SetWindowPos(gParams->ctl[n].hBuddy2, NULL,
                         r.left,    //x
                         r.top,     //y
                         r.right,   //w
                         r.bottom,  //h
                         SWP_NOZORDER | SWP_NOACTIVATE | SWP_NOCOPYBITS);
            //Cleanup...
            SelectObject(hDC, hOldFont);
            ReleaseDC(hWnd, hDC);
            //Set the actual control text...
            SetWindowText(hWnd, p);

            //It seems a static text control does not erase the
            //background when it receives a WM_SETTEXT message
            //(maybe because we are handling WM_CTLCOLORSTATIC
            //and setting the background to transparent??), so
            //we brute-force it... (this may be overkill??)
            if (gParams->ctl[n].enabled){
				GetWindowRect(hWnd, &wRect); 
				ScreenToClient (fmc.hDlg, (LPPOINT)&wRect);
				ScreenToClient (fmc.hDlg, (LPPOINT)&(wRect.right));
				return RedrawWindow(
	#if 0
						//just updating the control window doesn't work :(
						hWnd,   // handle of control window
	#else
						fmc.hDlg,   // handle of entire dialog window
	#endif
	#if 1
						&wRect, // address of structure with update rectangle
	#else
						NULL,   // address of structure with update rectangle, NULL=>entire client area
	#endif
						NULL,   // handle of update region
						RDW_ERASE|RDW_INVALIDATE|RDW_ERASENOW|RDW_UPDATENOW     // array of redraw flags
				   );
			}

        }//scope
        break;

    case CC_EDIT:		//see note above at CC_RADIOBUTTON!!!
        //SetWindowText(gParams->ctl[n].hCtl, fm_formatString(gParams->ctl[n].label));
		SetWindowText(gParams->ctl[n].hCtl, fm_formatString((char *)iText));

        //for an Edit control, we need to call UpdateWindow() to force a
        // WM_PAINT if we want an immediate update
        if (gParams->ctl[n].enabled){
			return UpdateWindow(gParams->ctl[n].hCtl);
		} else
			return TRUE;
        break;

    case CC_COMBOBOX:		//see note above at CC_RADIOBUTTON!!!
        SendMessage(gParams->ctl[n].hCtl, CB_RESETCONTENT, 0, 0);
        {
            char *p, *q, q_save;
            //p = gParams->ctl[n].label;
			p = (char *)iText; //Allow combo box text that is longer than 1023 bytes //// AARGH NFG when reloaded??? ///and formatString below might truncate anyway!!!
            while (*p) {
                for (q = p; *q != '\0' && *q != '\n'; q++)
                    ;
                q_save = *q;
                *q = '\0';	//aaargh!  *q may not be writeable?  (.label was okay...)
                SendMessage(gParams->ctl[n].hCtl, CB_ADDSTRING, 0, (LPARAM)fm_formatString(p));	//note formatString may truncate item
				*q = q_save;
                p = q;
                if (*p) p++;
            } //while
        }
        // set the control value to set the initial selection
        fm_setCtlVal(n, gParams->ctl[n].val);
        break;

    case CC_LISTBOX:		//see note above at CC_RADIOBUTTON!!!
        SendMessage(gParams->ctl[n].hCtl, LB_RESETCONTENT, 0, 0);
        {
            char *p, *q, q_save;
            //p = gParams->ctl[n].label;
			p = (char *)iText; //Allow list box text that is longer than 1023 bytes  //// AARGH NFG when reloaded???
            while (*p) {
                for (q = p; *q != '\0' && *q != '\n'; q++)
                    ;
                q_save = *q;
                *q = '\0';	//aaargh!  *q may not be writeable?  (.label was okay...)
                SendMessage(gParams->ctl[n].hCtl, LB_ADDSTRING, 0, (LPARAM)fm_formatString(p));	//note formatString may truncate item
                *q = q_save;
                p = q;
                if (*p) p++;
            } //while
        }
        // set the control value to set the initial selection
        fm_setCtlVal(n, gParams->ctl[n].val);
        break;
	
	case CC_TAB:
		{
			TCITEM tabItem; 
			int i;
			char *a;
			char copy[MAX_LABEL_SIZE+1];

			TabCtrl_DeleteAllItems (gParams->ctl[n].hCtl); 
			
			strncpy(copy, (char *)iText, MAX_LABEL_SIZE+1);

			tabItem.mask = TCIF_TEXT | TCIF_IMAGE; 
			tabItem.iImage = -1; 

			a = strtok(copy,"\n");
			for (i=0;i>=0;i++){
			  	
				tabItem.pszText = a; 
				TabCtrl_InsertItem(gParams->ctl[n].hCtl, i, &tabItem);

				a = strtok(NULL,"\n");
				if (a==0) break;
			}


		}	
		break;

    default:
        //most other controls....
        return SetWindowText(gParams->ctl[n].hCtl, fm_formatString(gParams->ctl[n].label));
        break;
    } //switch
    return FALSE;   //failed			///EH??? IS THIS ALWAYS TRUE???
} /*fm_setCtlText*/

/**********************************************************************/
/* setCtlTextv(n, "format",...)
/*
/*  Sets the text label or content for control n to the expanded
/*  printf-style format string.
/*
/**********************************************************************/
INT_PTR fm_setCtlTextv(INT_PTR n, INT_PTR iFmt,...)
{
    char szBuffer[1024];
    va_list ap;

    va_start(ap, iFmt);
    _vsnprintf(szBuffer, sizeof(szBuffer), (char *)iFmt, ap);
    szBuffer[sizeof(szBuffer)-1] = 0;   //ensure NUL-terminated
    va_end(ap);
    return fm_setCtlText(n, (INT_PTR)szBuffer);
} /*fm_setCtlTextv*/


/**********************************************************************/
/* enableToolTipBalloon(enable)
/*
/*  Enables or disables the balloon shape of all tool tips, based on
/*  the value of boolean value "enable".
/*
/**********************************************************************/
// For older versions of the SDK include files...
#ifndef TTS_BALLOON

#define TTS_NOANIMATE           0x10
#define TTS_NOFADE              0x20
#define TTS_BALLOON             0x40
#define TTS_CLOSE               0x80

//flags
#define TTF_PARSELINKS          0x1000


// ToolTip Icons (Set with TTM_SETTITLE)
#define TTI_NONE                0
#define TTI_INFO                1
#define TTI_WARNING             2
#define TTI_ERROR               3

#define TTM_SETTITLEA           (WM_USER + 32)  // wParam = TTI_*, lParam = char* szTitle
#define TTM_SETTITLEW           (WM_USER + 33)  // wParam = TTI_*, lParam = wchar* szTitle

#ifdef UNICODE
#define TTM_SETTITLE            TTM_SETTITLEW
#else
#define TTM_SETTITLE            TTM_SETTITLEA
#endif

#endif  //old include file


INT_PTR fm_enableToolTipBalloon(INT_PTR enable) {
	LONG dwStyle = GetWindowLong(ghTT, GWL_STYLE);

	int oldStyle = (dwStyle & TTS_BALLOON)? TRUE : FALSE;

	if (enable != FALSE) {
        dwStyle = WS_POPUP | TTS_ALWAYSTIP | TTS_NOPREFIX | TTS_BALLOON;

	} else {
        dwStyle = WS_POPUP | TTS_ALWAYSTIP | TTS_NOPREFIX | WS_BORDER;
	}

	SetWindowLong(ghTT, GWL_STYLE, dwStyle);

	return oldStyle;
}

// NOTE: Shouldn't we have a version of SetCtlToolTip with printf-style
//       formatting?  E.g., setCtlToolTipv(n, s, "format",...)

/**********************************************************************/
/* setCtlToolTip(n, "Tool Tip Text", s)
/*
/*  Sets the tool tip text for control n to the specified string, with
/*  style s (TTF_CENTERTIP, TTF_RTLREADING, TTF_TRACK, TTF_ABSOLUTE,
/*  TTF_TRANSPARENT), or deletes the tool tip for this control if the
/*  text argument is NULL or "".
/*
/**********************************************************************/
INT_PTR fm_setCtlToolTip(INT_PTR n, INT_PTR iText, INT_PTR s)
{
    INT_PTR iss = TRUE;
    TOOLINFO ti;
    int safe_style = (int)s & (TTF_CENTERTIP | TTF_RTLREADING | TTF_TRACK |
							   TTF_ABSOLUTE | TTF_TRANSPARENT);

    if (!IS_INUSE_CTL(n)) {
        return FALSE;
    }

    if (gParams->ctl[n].tooltip[0] == '\0') {
        //no current tool tip for this control
        if (iText != 0 && ((char *)iText)[0] != '\0') {
           
			if (n==CTL_PREVIEW){
				ti.cbSize = sizeof(TOOLINFO);
                ti.uFlags = safe_style | TTF_IDISHWND | TTF_SUBCLASS;//0;                      // or TTF_SUBCLASS???
                ti.hwnd = MyHDlg;
                ti.uId = (UINT_PTR) CTL_PREVIEW;         // our id for this control
                ti.rect.left = gProxyRect.left;
                ti.rect.top = gProxyRect.top;
                ti.rect.right = gProxyRect.right;
                ti.rect.bottom = gProxyRect.bottom;
                ti.hinst = hDllInstance;
			} else {
				//register a tooltip for this control
				ti.cbSize = sizeof(TOOLINFO);
				ti.uFlags = safe_style | TTF_IDISHWND | TTF_SUBCLASS;
				ti.hwnd = MyHDlg;
				ti.uId = (UINT_PTR) gParams->ctl[n].hCtl;
                // IIRC: leaving the rect undefined results in no tooltips under
                //       NT 4.0+ and XP.  Setting to 0 seems to fix the problem.
                ti.rect.left = ti.rect.top = ti.rect.right = ti.rect.bottom = 0;
				ti.hinst = hDllInstance;
			}
            ti.lpszText = fm_formatString((char *)iText);
            iss = SendMessage(ghTT, TTM_ADDTOOL, 0, (LPARAM)&ti);
            if (gParams->ctl[n].class == CC_STANDARD || gParams->ctl[n].class == CC_SLIDER) {
                //set tooltips for buddies
                ti.uId = (UINT_PTR) gParams->ctl[n].hBuddy1;
                iss = SendMessage(ghTT, TTM_ADDTOOL, 0, (LPARAM)&ti);
#if 1
                ti.uId = (UINT_PTR) gParams->ctl[n].hBuddy2;
                iss = SendMessage(ghTT, TTM_ADDTOOL, 0, (LPARAM)&ti);
#endif
            }
        }
    }
    else {
        //already have a tool tip for this control
        if (iText != 0 && ((char *)iText)[0] != '\0') {
            //update text for this tool tip...
            ti.cbSize = sizeof(TOOLINFO);
            ti.uFlags = safe_style | TTF_IDISHWND;
            ti.hwnd = MyHDlg;
            ti.uId = (UINT_PTR) gParams->ctl[n].hCtl;
            //ti.rect = xxx;
            ti.hinst = hDllInstance;
            ti.lpszText = fm_formatString((char *)iText);
            iss = SendMessage(ghTT, TTM_UPDATETIPTEXT, 0, (LPARAM)&ti);
            if (gParams->ctl[n].class == CC_STANDARD  || gParams->ctl[n].class == CC_SLIDER) {
                //update tooltips for buddies
                ti.uId = (UINT_PTR) gParams->ctl[n].hBuddy1;
                iss = SendMessage(ghTT, TTM_UPDATETIPTEXT, 0, (LPARAM)&ti);
#if 1
                ti.uId = (UINT_PTR) gParams->ctl[n].hBuddy2;
                iss = SendMessage(ghTT, TTM_UPDATETIPTEXT, 0, (LPARAM)&ti);
#endif
            }
        }
        else {
            //delete this tool tip
            ti.cbSize = sizeof(TOOLINFO);
            ti.uFlags = safe_style | TTF_IDISHWND;
            ti.hwnd = MyHDlg;
            ti.uId = (UINT_PTR) gParams->ctl[n].hCtl;
            //ti.rect = xxx;
            ti.hinst = hDllInstance;
            ti.lpszText = NULL;
            iss = SendMessage(ghTT, TTM_DELTOOL, 0, (LPARAM)&ti);
            if (gParams->ctl[n].class == CC_STANDARD  || gParams->ctl[n].class == CC_SLIDER) {
                //delete tooltips for buddies
                ti.uId = (UINT_PTR) gParams->ctl[n].hBuddy1;
                iss = SendMessage(ghTT, TTM_DELTOOL, 0, (LPARAM)&ti);
                ti.uId = (UINT_PTR) gParams->ctl[n].hBuddy2;
                iss = SendMessage(ghTT, TTM_DELTOOL, 0, (LPARAM)&ti);
            }
        }
    }

    //save tool tip text for this control
    if (iText == 0) {
        gParams->ctl[n].tooltip[0] = '\0';   //null string
    }
    else {
        strncpy(gParams->ctl[n].tooltip, (char *)iText, MAX_TOOLTIP_SIZE+1);
        gParams->ctl[n].tooltip[MAX_TOOLTIP_SIZE] = '\0';   //ensure null-terminated
    }

    return iss;    //success/failure
} /*fm_setCtlToolTip*/


INT_PTR fm_setToolTipDelay(INT_PTR m, INT_PTR iTime)
{

	WPARAM w;

	if (m==1)
		w = TTDT_INITIAL;
	else if (m==2)
		w = TTDT_AUTOPOP;
	else if (m==3)
		w = TTDT_RESHOW;
	else //m==0
		w = TTDT_AUTOMATIC;

	if (iTime==-1){
		// TTDT_AUTOMATIC with initial < 0 sets all values to default...
        SendMessage(ghTT, TTM_SETDELAYTIME, w, MAKELONG(-1, 0));

	} else {
		// TTDT_AUTOMATIC sets autopop = initial*10, reshow = initial/5...
        SendMessage(ghTT, TTM_SETDELAYTIME, w, MAKELONG(iTime, 0));
	}

	return true;
}


INT_PTR fm_getCtlImage(INT_PTR n)
{

	if (!IS_INUSE_CTL(n)) {
        return FALSE;
    }

	//if (gParams->ctl[n].image[0] == '\0')

	return (INT_PTR)(gParams->ctl[n].image);

}


/**********************************************************************/
/* setCtlImage(n, "Image Name", 'X')
/*
/*  Sets the image (if any) associated with control n to the specified
/*  filename or embedded resource name, with image type 'X', where
/*  'X' is one of the following character constants:
/*
/*      'B' -   bitmap file (.bmp)
/*      'W' -   Windows (old-style) metafile (.wmf)
/*      'E' -   enhanced metafile (.emf)
/*      'I' -   icon file (.ico)
/*      'C' -   cursor file (.cur)
/*      'J' -   JPEG file (.jpg)
/*      'G' -   GIF file (.gif)
/*      'M' -   MIG (mouse-ivo graphics) file (.mig)
/*       0  -   unspecified file type
/*
/**********************************************************************/
INT_PTR fm_setCtlImage(INT_PTR n, INT_PTR iName, INT_PTR iType)
{
    int iss = TRUE;

    if (!IS_INUSE_CTL(n)) {
        return FALSE;
    }

    if (gParams->ctl[n].image[0] == '\0') {
        //no current image for this control
        if (iName != 0 && ((char *)iName)[0] != '\0') {
            //set image for this control
        }
    }
    else {
        //already have an image for this control
        if (gParams->ctl[n].class == CC_IMAGE) {
            //select old bitmaps back into DC's, then
            //delete all image bitmaps and DC's.

            if (gParams->ctl[n].im.hbmOld) {
                SelectObject(gParams->ctl[n].im.hdcMem, gParams->ctl[n].im.hbmOld);
                gParams->ctl[n].im.hbmOld = 0;
            }
            if (gParams->ctl[n].im.hbm) {
                DeleteObject(gParams->ctl[n].im.hbm);
                gParams->ctl[n].im.hbm = 0;
            }
            if (gParams->ctl[n].im.hdcMem) {
                DeleteDC(gParams->ctl[n].im.hdcMem);
                gParams->ctl[n].im.hdcMem = 0;
            }

            if (gParams->ctl[n].im.hbmOldAnd) {
                SelectObject(gParams->ctl[n].im.hdcAnd, gParams->ctl[n].im.hbmOldAnd);
                gParams->ctl[n].im.hbmOldAnd = 0;
            }
            if (gParams->ctl[n].im.hbmAnd) {
                DeleteObject(gParams->ctl[n].im.hbmAnd);
                gParams->ctl[n].im.hbmAnd = 0;
            }
            if (gParams->ctl[n].im.hdcAnd) {
                DeleteDC(gParams->ctl[n].im.hdcAnd);
                gParams->ctl[n].im.hdcAnd = 0;
            }
#if 0
            //force update to reset background????
            InvalidateRect(gParams->ctl[n].hCtl, NULL, TRUE /*erase background*/);
            UpdateWindow(gParams->ctl[n].hCtl);
            //Sleep(1000);
#endif
        }
        else if (gParams->ctl[n].class == CC_METAFILE) {
            //set no image, and delete previous image
            HENHMETAFILE hPrevImage;
            hPrevImage = (HENHMETAFILE)SendMessage(gParams->ctl[n].hCtl,
                                                   STM_SETIMAGE,
                                                   IMAGE_ENHMETAFILE,
                                                   (LPARAM)NULL);
            //delete previous image, if any
            if (hPrevImage) {
                if (!DeleteEnhMetaFile(hPrevImage)) {
#ifndef STANDALONE
                    ErrorOk("DeleteEnhMetaFile failed, image = %8.8x", hPrevImage);
#endif
                }
                else {
                    //Info("Deleted metafile image = %8.8x", hPrevImage);
                }
            }
        }
        else if (gParams->ctl[n].class == CC_BITMAP) {
            //set no image, and delete previous bitmap
            HBITMAP hPrevImage;
            hPrevImage = (HBITMAP)SendMessage(gParams->ctl[n].hCtl,
                                              STM_SETIMAGE,
                                              IMAGE_BITMAP,
                                              (LPARAM)NULL);
            //delete previous bitmap, if any
            if (hPrevImage) {
                if (!DeleteObject(hPrevImage)) {
#ifndef STANDALONE
                    ErrorOk("DeleteObject failed, bitmap = %8.8x", hPrevImage);
#endif
                }
                else {
                    //Info("Deleted bitmap = %8.8x", hPrevImage);
                }
            }
        }
        else if (gParams->ctl[n].class == CC_ICON) {
            //set no image (icons don't get deleted)
            HICON hPrevImage;
            hPrevImage = (HICON)SendMessage(gParams->ctl[n].hCtl,
                                            STM_SETIMAGE,
                                            IMAGE_ICON,
                                            (LPARAM)NULL);
            //no need to delete previous icon (in fact, can't!)
        }
        gParams->ctl[n].image[0] = '\0';   //set no image
    }

    //save new image name for this control
    if (iName == 0) {
        gParams->ctl[n].image[0] = '\0';   //null string
    }
    else {
        strncpy(gParams->ctl[n].image, (char *)iName, _MAX_PATH+1);
        gParams->ctl[n].image[_MAX_PATH] = '\0';   //ensure null-terminated
    }
    gParams->ctl[n].imageType = (int)iType;


    if (gParams->ctl[n].class == CC_BITMAP) {
        if (gParams->ctl[n].image[0] != '\0') {
            HBITMAP hBmp;

            // First try loading the bitmap from an embedded BITMAP resource...
            hBmp = LoadImage(hDllInstance,
                             gParams->ctl[n].image,
                             IMAGE_BITMAP,
                             0, 0,
                             LR_DEFAULTCOLOR);
            //if (hBmp) Info("Loaded bitmap from resource: %s", gParams->ctl[n].image);

            // If it's not embedded, try searching for it as an
            // external bitmap file...
            if (!hBmp) {
                // Load bitmap from file...
                hBmp = LoadImage(hDllInstance,
                                FmSearchFile(gParams->ctl[n].image),
                                IMAGE_BITMAP,
                                0, 0,
                                LR_LOADFROMFILE);
            }
            if (!hBmp) {
                ErrorOk("Could not load bitmap %s", gParams->ctl[n].image);
                return FALSE;
            }
            else {
                HBITMAP hPrevImage;
                //Info("hBmp = %x", hBmp);
                hPrevImage = (HBITMAP)SendMessage(gParams->ctl[n].hCtl,
                                                  STM_SETIMAGE,
                                                  IMAGE_BITMAP,
                                                  (LPARAM)hBmp);
                //delete previous bitmap
                if (hPrevImage) {
                    if (!DeleteObject(hPrevImage)) {
#ifndef STANDALONE
                        ErrorOk("DeleteObject failed, bitmap = %8.8x", hPrevImage);
#endif
                    }
                    else {
                        //Info("Deleted bitmap = %8.8x", hPrevImage);
                    }
                }
                //DeleteObject(hBmp);    //do this at end of dialog???
#if 0
                return TRUE;        //unless need to force window update below...
#endif
            }
        }
    }//if CC_BITMAP

    else if (gParams->ctl[n].class == CC_ICON) {
        if (gParams->ctl[n].image[0] != '\0') {
            HICON   hIcon;

            // First try loading the icon from an embedded RT_ICON resource...
            hIcon = LoadImage(hDllInstance,
                              gParams->ctl[n].image,
                              IMAGE_ICON,
                              0, 0,
                              LR_DEFAULTCOLOR);
            //if (hIcon) Info("Loaded icon from resource: %s", gParams->ctl[n].image);
            //Info("Loaded icon from resource: %8.8x", hIcon);

            // If it's not embedded, try searching for it as an
            // external icon file...
            if (!hIcon) {
                // Load icon from file...
                hIcon = LoadImage(hDllInstance,
                                 FmSearchFile(gParams->ctl[n].image),
                                 IMAGE_ICON,
                                 0, 0,
                                 LR_LOADFROMFILE);
                //Info("Loaded icon from file: %8.8x", hIcon);
            }
            if (!hIcon) {
                ErrorOk("Could not load icon %s", gParams->ctl[n].image);
                return FALSE;
            }
            else {
                HICON hPrevImage;
                //Info("hIcon = %x", hIcon);
                hPrevImage = (HICON)SendMessage(gParams->ctl[n].hCtl,
                                                STM_SETIMAGE,
                                                IMAGE_ICON,
                                                (LPARAM)hIcon);
                //no need to delete previous icon (in fact, can't!)
#if 0
                return TRUE;        //unless need to force window update below...
#endif
            }
        }
    }//if CC_ICON

    else if (gParams->ctl[n].class == CC_METAFILE) {
      if (gParams->ctl[n].image[0] != '\0') {
        //enhanced metafile static control....
        HENHMETAFILE hImage = NULL;
        FILE *file;

        // Read from resource
        // Could also use LoadIMage (  IMAGE_ENHMETAFILE... ) ???
        HRSRC hRes;         // handle/ptr. to res. info.
        HRSRC hResLoad;     // handle to loaded resource  
        char *pFmMf;        // pointer to FMMETAFILE resource data 
        int cb;             // Size of resource (bytes)

        // Try to locate FMMETAFILE resource first...
        hRes = FindResource(hDllInstance, gParams->ctl[n].image, "FMMETAFILE"); 
        //Info("hRes = %x", hRes);
        if (hRes != NULL) { 
            // Get size of resource.
            cb = SizeofResource(hDllInstance, hRes);
            if (cb < 100) {
                //Implausible size
                ErrorOk("FMMETAFILE resource is too short (%d bytes)", cb);
            }
            else {
                // Load the FMMETAFILE resource into global memory. 
                hResLoad = LoadResource(hDllInstance, hRes); 
                //Info("hResLoad = %x", hResLoad);
                if (hResLoad == NULL) { 
                    ErrorOk("Could not load FMMETAFILE resource.");
                } 
                else {
                    // Lock the FMMETAFILE resource into global memory. 
                    pFmMf = LockResource(hResLoad); 
                    //Info("pFmMf = %x", pFmMf);
                    if (pFmMf == NULL) { 
                        ErrorOk("Could not lock FMMETAFILE resource.");
                    }
                    else {
                        // Determine what kind of metafile..,
                        if (((LPENHMETAHEADER)pFmMf)->iType == EMR_HEADER &&
                            ((LPENHMETAHEADER)pFmMf)->dSignature == ENHMETA_SIGNATURE) {
                            //It's an enhanced metafile
                            hImage = SetEnhMetaFileBits(cb, pFmMf);
                        }
                        else {
                            //Probably a Windows metafile
                            if (*(DWORD *)pFmMf == 0x9AC6CDD7) {
                                //metafile has an Aldus placeable header
                                //--skip over it...
                                pFmMf += 22;
                                cb -= 22;
                                //Info("WMF adjusted 1st dword = %x", *(DWORD *)pFmMf);
                            }
#if 0
                            if (((LPMETAHEADER)pFmMf)->mtSize * 2 != (size_t)cb) {
                                ErrorOk("FMMETAFILE resource is wrong size (%d instead of %d)",
                                    cb, ((LPMETAHEADER)pFmMf)->mtSize * 2);
                            }
#else
                            if (((LPMETAHEADER)pFmMf)->mtSize * 2 > (size_t)cb) {
                                ErrorOk("FMMETAFILE resource is too small (%d instead of %d)",
                                    cb, ((LPMETAHEADER)pFmMf)->mtSize * 2);
                            }
#endif
                            else {
                                hImage = SetWinMetaFileBits(cb, pFmMf,
                                                            NULL /*hdcRef*/,
                                                            NULL /*lpmfp*/);
                            }
                        }
                    }
                }
            }
        } //if hRes != NULL
        else {
            /* Search for metafile file in PATH and FM_PATH environment variables: */
            if ((file = fopen(FmSearchFile(gParams->ctl[n].image), "rb"))) {
                // Read from *.wmf (Plain Windows or Aldus Placeable) file
                // or *.emf (enhanced metafile).

                // Determine file length...
                fseek(file, 0, SEEK_END);
                cb = ftell(file);
                rewind(file);

                if (!(pFmMf = malloc(cb)))
                {
                    fclose(file);
                    ErrorOk("Could not allocate %d bytes for metafile", cb);
                }
                else {
                    //Now get the actual data bits...
                    cb = (int)fread(pFmMf, 1, cb, file);
                    //check for correct read length???
                    if (cb < 100) {
                        //implausible file size
                        ErrorOk("Metafile is too short (%d bytes)", cb);
                    }
                    // Determine what kind of metafile..,
                    else if (((LPENHMETAHEADER)pFmMf)->iType == EMR_HEADER &&
                        ((LPENHMETAHEADER)pFmMf)->dSignature == ENHMETA_SIGNATURE) {
                        //It's an enhanced metafile
                        hImage = SetEnhMetaFileBits(cb, pFmMf);
                    }
                    else {
                        //Info("WMF 1st dword = %x", *(DWORD *)pFmMf);
                        if (*(DWORD *)pFmMf == 0x9AC6CDD7) {
                            //file has an Aldus placeable header
                            //--skip over it...
                            pFmMf += 22;
                            cb -= 22;
                            //Info("WMF adjusted 1st dword = %x", *(DWORD *)pFmMf);
                        }
#if 0
                        if (((LPMETAHEADER)pFmMf)->mtSize * 2 != (size_t)cb) {
                            ErrorOk("Metafile is wrong size (%d instead of %d)",
                                cb, ((LPMETAHEADER)pFmMf)->mtSize * 2);
                        }
#elif 1
                        if (((LPMETAHEADER)pFmMf)->mtSize * 2 > (size_t)cb) {
                            ErrorOk("Metafile is too small (%d instead of %d)",
                                cb, ((LPMETAHEADER)pFmMf)->mtSize * 2);
                        }
#else
                        if (((LPMETAHEADER)pFmMf)->mtSize * 2 != (size_t)cb &&
                            IDOK != Warn("Metafile is wrong size (%d instead of %d)",
                                         cb, ((LPMETAHEADER)pFmMf)->mtSize * 2)) {
                            //bad size and user pressed Cancel...
                            ; //do nothiing (so hImage == NULL)
                        }
#endif
                        else {
                            hImage = SetWinMetaFileBits(cb, pFmMf,
                                                        NULL /*hdcRef*/,
                                                        NULL /*lpmfp*/);
                        }
                    }

                    free(pFmMf);
                }//if malloc

                fclose(file);
            }//if fopen
        }

        if (!hImage) {
            ErrorOk("Could not load metafile %s", gParams->ctl[n].image);
            return FALSE;
        }
        else {
            HENHMETAFILE hPrevImage;
            //Info("hImage = %x", hImage);
            //Info("EMF 1st dword = %x", *(DWORD *)hImage);
            hPrevImage = (HENHMETAFILE)SendMessage(gParams->ctl[n].hCtl,
                                                   STM_SETIMAGE,
                                                   IMAGE_ENHMETAFILE,
                                                   (LPARAM)hImage);
            //delete previous image
            if (hPrevImage) {
                if (!DeleteEnhMetaFile(hPrevImage)) {
#ifndef STANDALONE
                    ErrorOk("DeleteEnhMetaFile failed, image = %8.8x", hPrevImage);
#endif
                }
                else {
                    //Info("Deleted metafile image = %8.8x", hPrevImage);
                }
            }
            //DeleteEnhMetaFile(hImage);    //do this at end of dialog???
        }
      }//if image not null
      ///////return TRUE;  //unless need to force refresh below...
      //fall through to update control...
    }//if CC_METAFILE


    //FORCE CONTROL UPDATE IF NEEDED!!!
#if 0
    //This doesn't work...
    InvalidateRect(gParams->ctl[n].hCtl, NULL, TRUE /*erase background*/);
    return iss && UpdateWindow(gParams->ctl[n].hCtl);
#elif 0
    //Try updating entire dialog!
    InvalidateRect(fmc.hDlg, NULL, TRUE /*erase background*/);
    return iss && UpdateWindow(fmc.hDlg);
#elif 0
    //This doesn't do it either...
    return iss && fm_setCtlPos(n, -1, -1, -1, -1);
#elif 1
    //This works, but is pretty heavy-handed!!!
    fm_setCtlPos(n, gParams->ctl[n].xPos+1, -1, -1, -1);
    fm_setCtlPos(n, gParams->ctl[n].xPos-1, -1, -1, -1);
    InvalidateRect(gParams->ctl[n].hCtl, NULL, TRUE /*erase background*/);
    return iss && UpdateWindow(gParams->ctl[n].hCtl);
#elif 0
    //This works, but is pretty heavy-handed!!!
    fm_setCtlPos(n, gParams->ctl[n].xPos+1, -1, -1, -1);
    fm_setCtlPos(n, gParams->ctl[n].xPos-1, -1, -1, -1);
    return iss;
#else
    return iss;    //success/failure
#endif
} /*fm_setCtlImage*/

/**********************************************************************/
/* setCtlTicFreq(n, m)
/*
/*  Sets the frequency with which tick marks are displayed for slider
/*  control n. For example, if the frequency is set to 2, a tick mark
/*  is displayed for every other increment in the sliders range. The
/*  default setting for the frequency is 1 (that is, every increment 
/*  in the range is associated with a tick mark).
/*
/**********************************************************************/
INT_PTR fm_setCtlTicFreq(INT_PTR n, INT_PTR m)
{
    if (!IS_INUSE_CTL(n)) {
        return -1;
    }

    if (gParams->ctl[n].class == CC_TRACKBAR || gParams->ctl[n].class == CC_SLIDER) {
        gParams->ctl[n].tb.ticFreq = (int)m;
        SendMessage(gParams->ctl[n].hCtl, TBM_SETTICFREQ, m /*freq*/, 0);
        return TRUE;
    }
    else {
        //not a TRACKBAR
        return FALSE;
    }
} /*fm_setCtlTicFreq*/

/**********************************************************************/
/* setWindowTheme(hwnd, pszSubAppName, pszSubIdList)
/*
/*  Sets the Visual Theme for a given window to the specified theme.
/*
/*  Calls the following XP API function (if it exists):
/*
/*  THEMEAPI SetWindowTheme(HWND hwnd, LPCWSTR pszSubAppName, 
/*    LPCWSTR pszSubIdList);
/*
/*  N.B.  The 2nd and 3rd args must be UNICODE strings!
/**********************************************************************/
typedef HRESULT (STDAPICALLTYPE *LPSETWINDOWTHEME) (HWND, LPCWSTR, LPCWSTR);

static LPSETWINDOWTHEME lpSetWindowTheme;
static HINSTANCE hinstUXTHEMEDLL;

HRESULT setWindowTheme(HWND hwnd, LPCWSTR pszSubAppName, LPCWSTR pszSubIdList)
{
    HRESULT hres = -1;     //assume error
//#if 0                                                   // temporarily disable
    BOOL hres2 = 0;
    hinstUXTHEMEDLL = LoadLibrary("uxtheme.dll");
    if (hinstUXTHEMEDLL) {
        lpSetWindowTheme = (LPSETWINDOWTHEME) GetProcAddress(hinstUXTHEMEDLL, "SetWindowTheme");
        if (lpSetWindowTheme) {
            hres = lpSetWindowTheme( hwnd, pszSubAppName, pszSubIdList);
            //if (hres) Info("hres = %8.8x", hres);
            hres2 = FreeLibrary(hinstUXTHEMEDLL);
            //if (!hres2) Info("hres2 = %8.8x", hres2);
        }
        else {
            //Info("lpSetWindowTheme = %8.8x", lpSetWindowTheme);
        }
    }
    else {
        //Info("hinstUXTHEMEDLL = %8.8x", hinstUXTHEMEDLL);
    }
//#endif

    return hres;
} /*setWindowTheme*/


//Internal Theme Variable -> theme off by default -> replaced by gParams->gDialogTheme
//int DefaultTheme=0;


/**********************************************************************/
/* setDefaultWindowTheme(hwnd)
/*
/*  Sets the Visual Theme for a given window to the default visual
/*  theme, which for now means to turn OFF any XP Visual Styles.
/*
/**********************************************************************/
HRESULT setDefaultWindowTheme(HWND hwnd)
{
    // Turn off XP Visual Styles for now.
    if (gParams->gDialogTheme==0) //DefaultTheme
		return setWindowTheme(hwnd, L"", L""); //disable theme
	else
		return setWindowTheme(hwnd, NULL, NULL); //enable theme

} /*setDefaultWindowTheme*/


/**********************************************************************/
/* setCtlTheme(n, pszSubAppName, pszSubIdList)
/*
/*  Sets the Visual Theme for control n to the specified theme.
/*
/**********************************************************************/
/*int fm_setCtlTheme(int n, int pszSubAppName, int pszSubIdList) //Removed fm_ infront of setCtlTheme
{
    if (!IS_INUSE_CTL(n)) {
        return -1;
    }

    // todo: convert ascii strings to unicode

    // todo: setWindowTheme for buddies...

    return setWindowTheme(gParams->ctl[n].hCtl, L"", L"" );  //for now
}
*/


//New Styles Functions

typedef BOOL (STDAPICALLTYPE *LPISTHEMEACTIVE) (VOID);
static LPISTHEMEACTIVE lpIsThemeActive;

typedef struct _DllVersionInfo {
    DWORD cbSize;
    DWORD dwMajorVersion;
    DWORD dwMinorVersion;
    DWORD dwBuildNumber;
    DWORD dwPlatformID;
} DLLVERSIONINFO;

//
// Following code is cribbed from:
//
//      http://www.codeproject.com/KB/tips/DetectTheme.aspx
//
// q.v. for possible errors and updates!
//
INT_PTR fm_getAppTheme (void){

	INT_PTR ret = 0;
    OSVERSIONINFO ovi = {0};
    ovi.dwOSVersionInfoSize = sizeof ovi;
    GetVersionEx(&ovi);
    
	if (ovi.dwMajorVersion>=5) // && ovi.dwMinorVersion==1)
    {
        //Windows XP or higher detected
        typedef BOOL WINAPI ISAPPTHEMED(void);
        typedef BOOL WINAPI ISTHEMEACTIVE(void);
        ISAPPTHEMED* pISAPPTHEMED = NULL;
        ISTHEMEACTIVE* pISTHEMEACTIVE = NULL;
        HMODULE hMod = LoadLibrary("uxtheme.dll");
        if(hMod)
        {
            pISAPPTHEMED = (ISAPPTHEMED*) GetProcAddress(hMod, "IsAppThemed");				
            pISTHEMEACTIVE = (ISTHEMEACTIVE*) GetProcAddress(hMod,"IsThemeActive");

            if(pISAPPTHEMED && pISTHEMEACTIVE)
            {
                if(pISAPPTHEMED() && pISTHEMEACTIVE())                
                {                
                    typedef HRESULT CALLBACK DLLGETVERSION(DLLVERSIONINFO*);
                    DLLGETVERSION* pDLLGETVERSION = NULL;

                    HMODULE hModComCtl = LoadLibrary("comctl32.dll");
                    if(hModComCtl)
                    {
                        pDLLGETVERSION = (DLLGETVERSION*) GetProcAddress(hModComCtl,"DllGetVersion");
                        if(pDLLGETVERSION)
                        {
                            DLLVERSIONINFO dvi = {0};
                            dvi.cbSize = sizeof dvi;
                            if(pDLLGETVERSION(&dvi) == NOERROR )
                            {
                                //ret = dvi.dwMajorVersion >= 6;
								if (dvi.dwMajorVersion >= 6) 
									ret = 1;	
                            }
                        }
                        FreeLibrary(hModComCtl);                    
                    }
                }
            }
            FreeLibrary(hMod);
        }
    }    
    return ret;

}


INT_PTR fm_setDialogTheme (INT_PTR state){

	if (state==0){
		//DefaultTheme=0;
		gParams->gDialogTheme=0;
		return setWindowTheme(fmc.hDlg, L"", L""); //disable theme
	} else {
		
		//These two lines make sure that the title bar does not lose its theme at the second invocation
	    SetWindowRgn(fmc.hDlg, NULL, TRUE /*redraw*/);
		gParams->gDialogRegion = NULL;

		//DefaultTheme=1;
		gParams->gDialogTheme=1;
		return setWindowTheme(fmc.hDlg, NULL, NULL); //enable theme
	}
}

INT_PTR fm_setCtlTheme(INT_PTR n, INT_PTR state)
{
    int nIDDlgItem;
	LPCWSTR pszSubAppName; 
	LPCWSTR pszSubIdList;
	INT_PTR RetVal=0;

	//if (n < 0 || n >= N_CTLS) { //|| !gParams->ctl[n].inuse
	if ( !IS_INUSE_CTL(n) && n != CTL_PROGRESS && n != CTL_FRAME && n != CTL_ZOOM) {
	    return -1;
    }

    // todo: setWindowTheme for buddies...

	if (n == CTL_PROGRESS) nIDDlgItem = IDC_PROGRESS1;
    else if (n == CTL_FRAME) nIDDlgItem = IDC_PROXY_GROUPBOX;
    else if (n == CTL_ZOOM) nIDDlgItem = 498; //Zoom Label
    else  nIDDlgItem = (int)(IDC_CTLBASE+n);


	if (state==-1){ //DefaultTheme
		if (gParams->gDialogTheme==0) {
			pszSubAppName = L""; //disable theme
			pszSubIdList = L"";
		} else if (gParams->gDialogTheme==1){
			pszSubAppName = NULL; //enable theme
			pszSubIdList = NULL;
		}
		gParams->ctl[n].theme = -1;
	} else if (state==0){
		pszSubAppName = L""; //disable theme
		pszSubIdList = L"";
		gParams->ctl[n].theme = 0;
	} else {
		pszSubAppName = NULL; //enable theme
		pszSubIdList = NULL;
		gParams->ctl[n].theme = 1;
	}


	if (n==CTL_ZOOM){
		setWindowTheme(GetDlgItem(fmc.hDlg, IDC_BUTTON1), pszSubAppName, pszSubIdList);
		setWindowTheme(GetDlgItem(fmc.hDlg, IDC_BUTTON2), pszSubAppName, pszSubIdList);
	}
	if (gParams->ctl[n].class == CC_STANDARD || gParams->ctl[n].class == CC_SLIDER) {
		setWindowTheme(GetDlgItem(fmc.hDlg, (int)(IDC_BUDDY1+n)), pszSubAppName, pszSubIdList);
		setWindowTheme(GetDlgItem(fmc.hDlg, (int)(IDC_BUDDY2+n)), pszSubAppName, pszSubIdList);
	}
	
	RetVal = setWindowTheme(GetDlgItem(fmc.hDlg, nIDDlgItem), pszSubAppName, pszSubIdList); 


	//Avoid black check box text under XP
	if (gParams->ctl[n].class == CC_CHECKBOX) { 
		if (appTheme==-1) appTheme = (int)fm_getAppTheme();
		if ( appTheme && (state==1  || (state==-1 && gParams->gDialogTheme==1)) && fm_getCtlColor(n)==-1 )
			fm_setCtlColor(n,GetSysColor(COLOR_BTNFACE));
		else if ( (state==0  || (state==-1 && gParams->gDialogTheme==0)) && fm_getCtlColor(n)==(int)GetSysColor(COLOR_BTNFACE) ) //Switch back to transparent
			fm_setCtlColor(n,-1);
	}

	return RetVal;
	
}


INT_PTR fm_setDialogEvent (INT_PTR state){

	/*
	0 = None
	1 = Init event
	2 = Cancel event
	4 = Keydown event
	*/

	if (state < 0) state = 0; else if (state > 7) state = 7;

	if (state & 4) {		//not thread-safe???
		if (KeyHook != NULL) UnhookWindowsHookEx(KeyHook); //Avoid Multiple Hooks
		KeyHook = SetWindowsHookEx(WH_KEYBOARD, KeyHookProc,
#ifdef _WIN64
								   (HINSTANCE) GetWindowLongPtr(fmc.hDlg, GWLP_HINSTANCE), 
#else //Win32
								   (HINSTANCE) GetWindowLong(fmc.hDlg, GWL_HINSTANCE), 
#endif
								   GetCurrentThreadId());//(DWORD)NULL);
	}

	gParams->gDialogEvent |= state;
	
	return true;
}

INT_PTR fm_clearDialogEvent (INT_PTR state){

	/*
	0 = None
	1 = Init event
	2 = Cancel event
	4 = Keydown event
	*/

	if (state<0) state=0; else if (state>7) state=7;

	if (state & 4) {
		if (KeyHook != NULL) UnhookWindowsHookEx(KeyHook);
	}

	gParams->gDialogEvent &= ~state;
	
	return true;
}

INT_PTR fm_createFont(INT_PTR i, INT_PTR size, INT_PTR bold, INT_PTR italic, INT_PTR fontname)
{
    HFONT hfont;

    if (i < 0 || i >= N_FONTS) {
        return false;
    }

    if (Font[i] != 0) DeleteObject(Font[i]);

	hfont = CreateFont((int)size, 0, 0, 0, (bold? 700 : 400), (DWORD)italic,
						false, false,
						ANSI_CHARSET, OUT_DEFAULT_PRECIS,
						CLIP_DEFAULT_PRECIS, ANTIALIASED_QUALITY, DEFAULT_PITCH,
						(LPCTSTR)fontname);

    if (hfont){
        Font[i] = hfont;
        return true;
    }

    return false;
}

INT_PTR fm_deleteFont (INT_PTR i){

     if (i < 0 || i >= N_FONTS) {
        return false;
    }

     if (Font[i] != 0) {
        DeleteObject(Font[i]);
        Font[i] = 0;
        return true;
    }

    return false;
}


INT_PTR fm_setCtlFont(INT_PTR n, INT_PTR i)
{

    if (!IS_INUSE_CTL(n)) {
        return false;
    }


    if (i==-1){

        HFONT hfont = GetStockObject(DEFAULT_GUI_FONT);

        if (gParams->ctl[n].class == CC_STANDARD || gParams->ctl[n].class == CC_SLIDER){
            SendMessage(gParams->ctl[n].hBuddy1, WM_SETFONT, (WPARAM)hfont, TRUE);
            SendMessage(gParams->ctl[n].hBuddy2, WM_SETFONT, (WPARAM)hfont, FALSE);
        } else
            SendMessage(gParams->ctl[n].hCtl, WM_SETFONT, (WPARAM)hfont, FALSE);

        gParams->ctl[n].hFont = hfont;

        //Kludge to force a window update...
        fm_setCtlText(n, (INT_PTR)gParams->ctl[n].label); 

        return true;

    } else if (i >= 0 && i < N_FONTS && Font[i]!=0) {
        
        if (gParams->ctl[n].class ==CC_STANDARD || gParams->ctl[n].class ==CC_SLIDER){
            SendMessage(gParams->ctl[n].hBuddy1, WM_SETFONT, (WPARAM)Font[i], TRUE);
            SendMessage(gParams->ctl[n].hBuddy2, WM_SETFONT, (WPARAM)Font[i], FALSE);
        } else
            SendMessage(gParams->ctl[n].hCtl, WM_SETFONT, (WPARAM)Font[i], FALSE);
        
        gParams->ctl[n].hFont = Font[i];

        //Kludge to force a window update...
        fm_setCtlText(n, (INT_PTR)gParams->ctl[n].label);        

        return true;
    }

	
    return false;
}


/**********************************************************************/
/* setCtlFontColor(n, color)
/*
/*  Sets the text color for control n to the specified RGB-triple value.
/*
/**********************************************************************/
INT_PTR fm_setCtlFontColor(INT_PTR n, INT_PTR color)
{
    if (!IS_INUSE_CTL(n)) {
        return -1;
    }
    gParams->ctl[n].textColor = (COLORREF)color;
    //Kludge to force a window update...
    return fm_setCtlText(n, (INT_PTR)gParams->ctl[n].label);
} /*fm_setCtlFontColor*/

/**********************************************************************/
/* setCtlColor(n, color)
/*
/*  Sets the background color of control n to the specified RGB-triple.
/*  (This actually has no effect for many control styles; the effect is
/*  specific to each control style, and will be specified under the
/*  description for each style.)
/*
/*        Color = -1 means transparent.
/*
/**********************************************************************/
INT_PTR fm_setCtlColor(INT_PTR n, INT_PTR color)
{
	HWND hCtl = gParams->ctl[n].hCtl;

	if (n == CTL_PREVIEW) {
		hCtl = fmc.hDlg;//GetDlgItem(fmc.hDlg, 101);
		gParams->ctl[n].bkColor = (COLORREF)color;
		InvalidateRect(fmc.hDlg, NULL, TRUE /*erase background*/);
        return UpdateWindow(fmc.hDlg);
	}


#if 0
    RECT    wRect;
#endif

    if (!IS_INUSE_CTL(n)) {
        return -1;
    }
    gParams->ctl[n].bkColor = (COLORREF)color;
    DeleteObject(gParams->ctl[n].hBrush);
    gParams->ctl[n].hBrush = ((COLORREF)color == -1) ?
                                GetStockObject(HOLLOW_BRUSH) : 
                                CreateSolidBrush((COLORREF)color);
    // Force a window update...
#if 1
    if (gParams->ctl[n].class == CC_TRACKBAR || gParams->ctl[n].class == CC_SLIDER) {
        // Do a dummy SETRANGEMAX with the REDRAW flag set...
        SendMessage(hCtl, TBM_SETRANGEMAX, TRUE /*redraw*/,
                    gParams->ctl[n].maxval);
        return TRUE;
    }
    else {
        //following doesn't work for trackbars
        //but is OK for checkbox, radiobutton, statictext, groupbox, etc.
        InvalidateRect(hCtl, NULL, TRUE /*erase background*/);
        return UpdateWindow(hCtl);
    }
#elif 0
    return TRUE; //NFG for checkbox, radiobutton, statictext, etc.
#elif 0
    // This doesn't work for anything...
    return SendMessage(hCtl, WM_ERASEBKGND, (WPARAM)GetDC(hCtl), 0);
#else
    //Bring out the big guns...
    //Well, even this doesn't work for trackbars, and it's
    //probably vast overkill for other controls...
    GetWindowRect(hCtl, &wRect); 
    ScreenToClient (fmc.hDlg, (LPPOINT)&wRect);
    ScreenToClient (fmc.hDlg, (LPPOINT)&(wRect.right));
    return RedrawWindow(
#if 0
            //just updating the control window doesn't work :(
            hCtl,   // handle of control window
#else
            fmc.hDlg,   // handle of entire dialog window
#endif
#if 1
            &wRect, // address of structure with update rectangle
#else
            NULL,   // address of structure with update rectangle, NULL=>entire client area
#endif
            NULL,   // handle of update region
            RDW_ERASE|RDW_INVALIDATE|RDW_ERASENOW|RDW_UPDATENOW     // array of redraw flags
       );

#endif
    // REMEMBER TO DELETE THE BRUSH WHEN EXITING????
} /*fm_setCtlColor*/


INT_PTR fm_setCtlSysColor(INT_PTR n, INT_PTR con){

	return fm_setCtlColor (n, GetSysColor((int)con));

}

INT_PTR fm_setCtlFontSysColor(INT_PTR n, INT_PTR con){

	return fm_setCtlFontColor (n, GetSysColor((int)con));

}


/**********************************************************************/
/* getCtlColor(n)
/*
/*  Returns the current background color of control n as an RGB-triple,
/*  or -1 if n is not a valid control.
/*
/**********************************************************************/
INT_PTR fm_getCtlColor(INT_PTR n)
{
    if (!IS_INUSE_CTL(n)) {
        return -1;
    }

	if (gParams->ctl[n].class == CC_TAB){

			if (appTheme==-1) appTheme = (int)fm_getAppTheme();
			if ( appTheme && (gParams->ctl[n].theme==1  || (gParams->ctl[n].theme==-1 && gParams->gDialogTheme==1)) ){
				typedef HANDLE (STDAPICALLTYPE *LPOpenThemeData) (HWND, LPCWSTR);
				typedef HRESULT (STDAPICALLTYPE *LPGetThemeColor) (HTHEME,INT,INT,INT,COLORREF);
				typedef HRESULT (STDAPICALLTYPE *LPCloseThemeData) (HTHEME);
				static LPOpenThemeData lpOpenThemeData = NULL;
				static LPGetThemeColor lpGetThemeColor = NULL;
				static LPCloseThemeData lpCloseThemeData = NULL;
				HANDLE vsTheme;

				HMODULE hMod = LoadLibrary("uxtheme.dll");
				if (hMod){
					// Should error check the following!!!
					lpOpenThemeData = (LPOpenThemeData) GetProcAddress(hMod, "OpenThemeData");
					lpGetThemeColor = (LPGetThemeColor) GetProcAddress(hMod,"GetThemeColor");
					lpCloseThemeData = (LPCloseThemeData) GetProcAddress(hMod,"CloseThemeData");

					vsTheme = lpOpenThemeData(gParams->ctl[n].hCtl,L"TAB");
					if (vsTheme){
						COLORREF pColor = 0;
						lpGetThemeColor(vsTheme, 9, 1, 3821, &pColor); //3802
						lpCloseThemeData(vsTheme);
						FreeLibrary(hMod);
						return (INT_PTR)pColor;
					} else
						return (INT_PTR)GetSysColor(COLOR_BTNFACE);

					FreeLibrary(hMod);

				} else
					return (INT_PTR)GetSysColor(COLOR_BTNFACE);
			} else
				return (INT_PTR)GetSysColor(COLOR_BTNFACE);
	}

	return gParams->ctl[n].bkColor;

} /*fm_getCtlColor*/


INT_PTR fm_setCtlDefVal(INT_PTR n, INT_PTR defval)
{
    if (!IS_INUSE_CTL(n)) {
        return false;
    }

    gParams->ctl[n].defval = (int)defval;

    return true;
}

/**********************************************************************/
/* setCtlVal(n, val)
/*
/*  Sets the value of control n to val.
/*
/*  Returns the previous value of control n, or -1 if n is not a
/*  valid control.
/*
/**********************************************************************/
INT_PTR fm_setCtlVal(INT_PTR n, INT_PTR val)
{
    INT_PTR originalValue = val;    //Is this really needed???
    INT_PTR prevValue;

    INT_PTR val2;


    //Info("setCtlVal(%d, %d)", n, val);
    if (!IS_INUSE_CTL(n)) {
        return -1;
    }

    prevValue = gParams->ctl[n].val;


    //Only manipulate the controls if dialog was displayed, otherwise we get problems
    if (fmc.displayDialog){ //eh???  Why was this added??? Should displayDialog be set false when updating main image???

        switch (gParams->ctl[n].class) {
        case CC_STANDARD:
	    case CC_SLIDER:
        case CC_SCROLLBAR:
            if (gParams->ctl[n].minval > gParams->ctl[n].maxval) {
                //range is inverted...              //NFG for scrollbar -- thumb stays at left
                                                    // but we could fake it by using -val internally...
                if (val > gParams->ctl[n].minval) val = gParams->ctl[n].minval;
                else if (val < gParams->ctl[n].maxval) val = gParams->ctl[n].maxval;
                //make sure lineSize and pageSize are negative...       //// why not do this (only) in setCtlRange()
                                                                        //// and setCtlLineSize()/setCtlPageSize()
                                                                        //// i.e., only when linesize, pagesize, 
                                                                        //// minval, or maxval is modified ?
                if (gParams->ctl[n].lineSize > 0)
                    gParams->ctl[n].lineSize = -gParams->ctl[n].lineSize;
                if (gParams->ctl[n].pageSize > 0)
                    gParams->ctl[n].pageSize = -gParams->ctl[n].pageSize;
            }
            else {
                //normal range (min <= max)
                if (val < gParams->ctl[n].minval) val = gParams->ctl[n].minval;
                else if (val > gParams->ctl[n].maxval) val = gParams->ctl[n].maxval;
                //make sure lineSize and pageSize are positive...       ///// DITTO
                if (gParams->ctl[n].lineSize < 0)
                    gParams->ctl[n].lineSize = -gParams->ctl[n].lineSize;
                if (gParams->ctl[n].pageSize < 0)
                    gParams->ctl[n].pageSize = -gParams->ctl[n].pageSize;
            }
            //if (prevValue != originalValue || prevValue != val) {	//eh??? Why not???
                //update the actual control if it changed

			    if ( gParams->ctl[n].thumbSize > 0){
				    int range = abs(gParams->ctl[n].maxval - gParams->ctl[n].minval);
				    val2 = val * (range - gParams->ctl[n].thumbSize-1) / range;
                } else {
				    val2 = val;
			    }
            
                if (gParams->ctl[n].gamma != 100){
                
                     double r = 1.0/(gParams->ctl[n].gamma/100.0);

                    if (gParams->ctl[n].minval >= 0){
                        double absRange = (double) IABS(gParams->ctl[n].maxval - gParams->ctl[n].minval) - 1.0;
                        val2 = gParams->ctl[n].minval + (int)(pow( (val2 - gParams->ctl[n].minval) / absRange, r) * absRange + 0.5);
                    } else {
                        if (val >= 0){                    
                            double absRange = (double)gParams->ctl[n].maxval - 1.0;
                            val2 = (int)(pow( (val2) / absRange, r) * absRange + 0.5);
                        }else {
                            double absRange = (double) IABS(gParams->ctl[n].minval) - 1.0;
                            val2 = - (int)(pow( IABS(val2) / absRange, r) * absRange + 0.5);
                        }
                    }

                } 


                if (gParams->ctl[n].minval > gParams->ctl[n].maxval) {
                    //range is inverted...
                    if (gParams->ctl[n].class == CC_SLIDER)
					    SendMessage(gParams->ctl[n].hCtl, TBM_SETPOS, TRUE, gParams->ctl[n].minval + gParams->ctl[n].maxval - val2);
				    else
					    SetScrollPos(gParams->ctl[n].hCtl, SB_CTL,
                                 (int)(gParams->ctl[n].minval + gParams->ctl[n].maxval - val2),
                                 TRUE /*redraw*/);	
                }
                else {
                    //normal range (min <= max)
                    if (gParams->ctl[n].class == CC_SLIDER)
					    SendMessage(gParams->ctl[n].hCtl, TBM_SETPOS, TRUE, val2);
				    else
					    SetScrollPos(gParams->ctl[n].hCtl, SB_CTL,
                                 (int)val2,
                                 TRUE /*redraw*/);
                }
                if ( (gParams->ctl[n].class == CC_STANDARD || gParams->ctl[n].class == CC_SLIDER)
					&& (EditBoxUpdate==0 || GetFocus() != GetDlgItem(fmc.hDlg, (int)(IDC_BUDDY1+n))) ) {
                    //update the numeric readout field if it changed
                
    #if 1
                    if (gParams->ctl[n].divisor == 1) { // Set Integer Value

                        SetDlgItemInt(fmc.hDlg, (int)(IDC_BUDDY1+n),
                                      (UINT)val,
                                      TRUE /*signed*/);

                    } else { // Set Double Value
                        char szBuffer[32];
                    
                        if (gParams->ctl[n].divisor<=10)
                            sprintf(szBuffer, "%.1f", (double) val / gParams->ctl[n].divisor );
                        else if (gParams->ctl[n].divisor<=100)
                            sprintf(szBuffer, "%.2f", (double) val / gParams->ctl[n].divisor );
                        else ///// if (gParams->ctl[n].divisor<=1000)
                            sprintf(szBuffer, "%.3f", (double) val / gParams->ctl[n].divisor );
                    
                        SetDlgItemText (fmc.hDlg, (int)(IDC_BUDDY1+n),(LPCTSTR) &szBuffer ) ;
                    }
    #else
                    char szBuffer[32];

                    sprintf(szBuffer, "%.3g", (double) val / gParams->ctl[n].divisor );
                    SetDlgItemText (fmc.hDlg, IDC_BUDDY1+n, (LPCTSTR) &szBuffer ) ;
    #endif


    #if 1
                    //if the text buddy is an Edit control, we need to
                    //call UpdateWindow() if we want an immediate update
                    UpdateWindow(GetDlgItem(fmc.hDlg, (int)(IDC_BUDDY1+n)));
    #endif
                } //if CC_STANDARD
            //}//if
            break;

        case CC_TRACKBAR:
            if (prevValue != originalValue) {
                //update the actual control if it changed
            
                SendMessage(gParams->ctl[n].hCtl, TBM_SETPOS, TRUE /*set*/, val);
            }
            break;

        case CC_CHECKBOX:
            val &= 3;
            if (prevValue != val) {
                //update the control if it changed
                if (SendMessage(gParams->ctl[n].hCtl, BM_GETCHECK, 0, 0) != val) {
                    //update the control itself if it was not already
                    //automatically updated...
                    SendMessage(gParams->ctl[n].hCtl, BM_SETCHECK, val, 0);
                }
            }
            break;

        case CC_PUSHBUTTON:
        case CC_RADIOBUTTON:
            val &= 1;
            if (prevValue != val) {
                //update the control if it changed
                HWND nextCtl;
                int next;
                HWND thisCtl = gParams->ctl[n].hCtl;
                int nIter;
                switch (gParams->ctl[n].style & BS_TYPEMASK) {
                case BS_AUTORADIOBUTTON:
                    if (SendMessage(thisCtl, BM_GETCHECK, 0, 0) != val) {
                        //update the control itself if it was not already
                        //automatically updated...
                        SendMessage(thisCtl, BM_SETCHECK, val, 0);
                    }
                    //for an auto radio button, set all other radio buttons
                    //in the current group to 0 if this button is being set
                    //to non-zero.		///eh?? Also sets if this button is being set to 0?????
                    nIter = 0;
                    nextCtl = GetNextDlgGroupItem(fmc.hDlg, thisCtl, FALSE /*next*/);
                    while (nextCtl != NULL && nextCtl != thisCtl) {
                        if (++nIter > N_CTLS) {
                            // can't be more than N_CTLS user controls --
                            // we must be looping, so exit fast...
                            break;
                        }
                        next = GetDlgCtrlID(nextCtl) - IDC_CTLBASE;
                        if (next >= 0 && next < N_CTLS) {
                            if (gParams->ctl[next].class == CC_RADIOBUTTON) {
                                if (gParams->ctl[next].val != 0) {
                                    //set it to zero... (don't call setCtlVal to do
                                    //this, to avoid infinite recursion!!!!)
                                    gParams->ctl[next].val = 0;
                                    fmc.pre_ctl[next] = 0; //update cached value
                                    if (SendMessage(nextCtl, BM_GETCHECK, 0, 0) != 0) {
                                        //update the control itself if it was not already
                                        //automatically updated...
                                        SendMessage(nextCtl, BM_SETCHECK, 0, 0);
                                    }
                                }//if
                            }//if
                        }
                        else {
                            //not a valid user control??? Get out quick!!
                            break;
                        }//if
                        nextCtl = GetNextDlgGroupItem(fmc.hDlg, nextCtl, FALSE /*next*/);
                    }//while
                    break;
                case BS_GROUPBOX:
                    //nothing to update for this control...
                    break;
                default:
                    //unconditionally update all other controls...
                    SendMessage(thisCtl, BM_SETCHECK, val, 0);
                    break;
                }//switch button subtype
            }//if value changed
            break;

        case CC_LISTBOX:
            if (gParams->ctl[n].style & (LBS_MULTIPLESEL | LBS_EXTENDEDSEL)) {
                //Must use LB_SELITEMRANGEEX for multiple selection list boxes...
                SendMessage(gParams->ctl[n].hCtl, LB_SELITEMRANGEEX, val, val);
            }
            else {
                //Use LB_SETCURSEL for single-selection list boxes...
                /*val =*/ SendMessage(gParams->ctl[n].hCtl, LB_SETCURSEL, val, 0);	//don't set readback value -- not always useful!!
            }
            break;

        case CC_COMBOBOX:
            /*val =*/ SendMessage(gParams->ctl[n].hCtl, CB_SETCURSEL, val, 0);	//don't set readback value -- not always useful!! 
            break;

	    case CC_TAB:
		    
		    TabCtrl_SetCurSel (gParams->ctl[n].hCtl,val);
		    
		    gParams->ctl[n].val = (int)val;		//MWVDL Uncommented	//eh???
		    fmc.pre_ctl[n] = (int)val; //update cached value
		    //fm_triggerEvent(n, FME_CLICKED, 0);

		    //Info ("%d",val);

		    if (val!=-1) 
		    {
			    int i;
			    for (i = 0; i < N_CTLS; ++i) {
				    if (gParams->ctl[i].inuse && gParams->ctl[i].tabControl == n &&  gParams->ctl[i].tabSheet != -1){
					    enableCtl(i);	//MWVDL 2008-07-18
				    }	
			    }
		    }
	    
		    break;

        case CC_OWNERDRAW:
        case CC_STATICTEXT:
        case CC_FRAME:
        case CC_RECT:
        case CC_BITMAP:
        case CC_ICON:
        case CC_METAFILE:
        case CC_GROUPBOX:
        default:
            break;
        } //switch class

    }//if fmc.dialogDisplayed


    //CHECK FOR IN RANGE FOR ALL CONTROLS!!!

    gParams->ctl[n].val = (int)val;
    fmc.pre_ctl[n] = (int)val; //update cached value
    return prevValue;
} /*fm_setCtlVal*/



INT_PTR fm_setCtlDivisor(INT_PTR n, INT_PTR div)
{
    if (!IS_INUSE_CTL(n)) {
        return -1;
    }

    if (gParams->ctl[n].class == CC_STANDARD || gParams->ctl[n].class == CC_SLIDER) {
    
        if (div >= 1  && div <= 1000){
            gParams->ctl[n].divisor = (int)div;
            // Update the control display if necessary
            fm_setCtlVal(n, gParams->ctl[n].val);
            return true;
        }
    }

    return false;

} /*fm_setCtlDivisor*/



INT_PTR fm_getCtlDivisor(INT_PTR n)
{
    if (!IS_INUSE_CTL(n)) {
        return 1;
    }

    if (gParams->ctl[n].class == CC_STANDARD || gParams->ctl[n].class == CC_SLIDER) {
		return gParams->ctl[n].divisor;
    }

    return 1;

} /*fm_getCtlDivisor*/



INT_PTR fm_setCtlGamma(INT_PTR n, INT_PTR gamma)
{
    if (!IS_INUSE_CTL(n)) {
        return -1;
    }

    if (gParams->ctl[n].class == CC_STANDARD || gParams->ctl[n].class == CC_SLIDER) {
    
        if (gamma >= 1  && gamma <= 1000){
            gParams->ctl[n].gamma = (int)gamma;
            // Update the control display if necessary
            fm_setCtlVal(n, gParams->ctl[n].val);
            return true;
        }
    }

    return false;

} /*fm_setCtlGamma*/



INT_PTR fm_setCtlProperties(INT_PTR n, INT_PTR props)
{
    int prevProps;
    if (n != CTL_PREVIEW && (n < 0 || n >= sizeof(gParams->ctl)/sizeof(gParams->ctl[0]) ||
        !gParams->ctl[n].inuse) ) {			// && !IS_INUSE_CTL(n)  ???
        return -1;
    }
    prevProps = gParams->ctl[n].properties;
    gParams->ctl[n].properties |= props;
    if (gParams->ctl[n].class == CC_COMBOBOX &&
        gParams->ctl[n].properties & CTP_EXTENDEDUI) {
        // Set extended keyboard UI
        SendMessage(gParams->ctl[n].hCtl, CB_SETEXTENDEDUI, TRUE, 0);
    }
    return prevProps;
} /*fm_setCtlProperties*/

INT_PTR fm_clearCtlProperties(INT_PTR n, INT_PTR props)
{
    int prevProps;
    if (!IS_INUSE_CTL(n)) {
        return -1;
    }
    prevProps = gParams->ctl[n].properties;
    gParams->ctl[n].properties &= ~props;
    if (gParams->ctl[n].class == CC_COMBOBOX &&
        !(gParams->ctl[n].properties & CTP_EXTENDEDUI)) {
        // Reset extended keyboard UI
        SendMessage(gParams->ctl[n].hCtl, CB_SETEXTENDEDUI, FALSE, 0);
    }
    return prevProps;
} /*fm_clearCtlProperties*/

INT_PTR fm_setCtlRange(INT_PTR n, INT_PTR lo, INT_PTR hi)
{   /* returns TRUE if success, else FALSE */
    INT_PTR retVal = 0; //assume the worst
    if (!IS_INUSE_CTL(n)) {
        return retVal;
    }
    gParams->ctl[n].minval = (int)lo;
    gParams->ctl[n].maxval = (int)hi;

    //update the actual control range
    switch (gParams->ctl[n].class) {
    case CC_STANDARD:
    case CC_SCROLLBAR:
        if (gParams->ctl[n].minval > gParams->ctl[n].maxval) {
            //range is inverted...
            retVal= SetScrollRange(gParams->ctl[n].hCtl,
                                   SB_CTL,
                                   gParams->ctl[n].maxval,
                                   gParams->ctl[n].minval,
                                   TRUE /*redraw*/);
        }
        else {
            //normal range (min <= max)
            retVal= SetScrollRange(gParams->ctl[n].hCtl,
                                   SB_CTL,
                                   gParams->ctl[n].minval,
                                   gParams->ctl[n].maxval,
                                   TRUE /*redraw*/);
        }
        break;

	case CC_SLIDER:
	case CC_TRACKBAR:
        SendMessage(gParams->ctl[n].hCtl, TBM_SETRANGEMIN, FALSE /*redraw*/,
                    gParams->ctl[n].minval);
        SendMessage(gParams->ctl[n].hCtl, TBM_SETRANGEMAX, TRUE /*redraw*/,
                    gParams->ctl[n].maxval);
        break;

    case CC_PUSHBUTTON:
    case CC_RADIOBUTTON:
    case CC_GROUPBOX:
        if (gParams->ctl[n].minval < 0) gParams->ctl[n].minval = 0;
        if (gParams->ctl[n].maxval > 1) gParams->ctl[n].maxval = 1;
        break;

    case CC_CHECKBOX:
        if (gParams->ctl[n].minval < 0) gParams->ctl[n].minval = 0;
        if (gParams->ctl[n].maxval > 2) gParams->ctl[n].maxval = 2;
        break;

    case CC_LISTBOX:
    case CC_COMBOBOX:
        if (gParams->ctl[n].minval < -1) gParams->ctl[n].minval = -1;
        if (gParams->ctl[n].maxval > 0xFFFF) gParams->ctl[n].maxval = 0xFFFF;
        break;


    case CC_STATICTEXT:
    case CC_FRAME:
    case CC_RECT:
    case CC_BITMAP:
    case CC_ICON:
    case CC_METAFILE:
    case CC_OWNERDRAW:
        //allow any arbitrary range for these controls
    default:
        break;
    } //switch class

    //force value to lie within new range, and update position
    //and numeric readout if necessary (must be done AFTER
    //setting the scrollbar range)
    fm_setCtlVal(n, gParams->ctl[n].val);
    return retVal;
} /*fm_setCtlRange*/

/**********************************************************************/
/* setCtlLineSize(n, a)
/*
/*  Sets the small step ("line size") value for control n to a, and
/*  returns the previous small step value.
/*
/**********************************************************************/
INT_PTR fm_setCtlLineSize(INT_PTR n, INT_PTR a)
{
    int prevValue;
    if (!IS_INUSE_CTL(n)) {
        return -1;
    }
    prevValue = gParams->ctl[n].lineSize;
    gParams->ctl[n].lineSize = (int)a;
    switch (gParams->ctl[n].class) {
    case CC_STANDARD:
	case CC_SLIDER:
    case CC_SCROLLBAR:
        if (gParams->ctl[n].minval > gParams->ctl[n].maxval) {
            //range is inverted...
            //make sure lineSize is negative...
            if (gParams->ctl[n].lineSize > 0)
                gParams->ctl[n].lineSize = -gParams->ctl[n].lineSize;
        }
        else {
            //normal range (min <= max)
            //make sure lineSize is positive...
            if (gParams->ctl[n].lineSize < 0)
                gParams->ctl[n].lineSize = -gParams->ctl[n].lineSize;
        }
		if (gParams->ctl[n].class == CC_SLIDER) 
			SendMessage(gParams->ctl[n].hCtl, TBM_SETLINESIZE, 0 /*unused*/, a);
		break; 
    case CC_TRACKBAR:
        SendMessage(gParams->ctl[n].hCtl, TBM_SETLINESIZE, 0 /*unused*/, a);
        break;
    default:
        break;
    } //switch class
    return prevValue;
} /*fm_setCtlLineSize*/

/**********************************************************************/
/* setCtlPageSize(n, b)
/*
/*  Sets the large step ("page size") value for control n to b, and
/*  returns the previous large step value.
/*
/**********************************************************************/
INT_PTR fm_setCtlPageSize(INT_PTR n, INT_PTR b)
{
    int prevValue;
    if (n < 0 || n >= sizeof(gParams->ctl)/sizeof(gParams->ctl[0]) ||
        !gParams->ctl[n].inuse) {
        return -1;
    }
    prevValue = gParams->ctl[n].pageSize;
    gParams->ctl[n].pageSize = (int)b;
    switch (gParams->ctl[n].class) {
    case CC_STANDARD:
	case CC_SLIDER:
    case CC_SCROLLBAR:
        if (gParams->ctl[n].minval > gParams->ctl[n].maxval) {
            //range is inverted...
            //make sure pageSize is negative...
            if (gParams->ctl[n].pageSize > 0)
                gParams->ctl[n].pageSize = -gParams->ctl[n].pageSize;
        }
        else {
            //normal range (min <= max)
            //make sure pageSize is positive...
            if (gParams->ctl[n].pageSize < 0)
                gParams->ctl[n].pageSize = -gParams->ctl[n].pageSize;
        }
		if (gParams->ctl[n].class == CC_SLIDER) 
			SendMessage(gParams->ctl[n].hCtl, TBM_SETPAGESIZE, 0 /*unused*/, b);
        break;
    case CC_TRACKBAR:
        SendMessage(gParams->ctl[n].hCtl, TBM_SETPAGESIZE, 0 /*unused*/, b);
        break;
    default:
        break;
    } //switch class
    return prevValue;
} /*fm_setCtlPageSize*/


INT_PTR fm_setCtlThumbSize(INT_PTR n, INT_PTR a)
{
    
    if (a > 0 && gParams->ctl[n].class == CC_SCROLLBAR || gParams->ctl[n].class == CC_STANDARD){
    	
		SCROLLINFO si;
    	LPSCROLLINFO lpsi;

		if (a > gParams->ctl[n].maxval - gParams->ctl[n].minval - 2) 
			a = gParams->ctl[n].maxval - gParams->ctl[n].minval - 2; //avoid crash
    	
    	si.cbSize = sizeof(si);
    	si.fMask = SIF_PAGE; //| SIF_RANGE;
    	si.nPage = (int)a;
    	//si.nMin = gParams->ctl[n].minval;
    	//si.nMax = gParams->ctl[n].maxval;

		lpsi = &si;
    	
    	SendMessage(gParams->ctl[n].hCtl, SBM_SETSCROLLINFO, TRUE, (LPARAM) lpsi);

    	gParams->ctl[n].thumbSize = (int)a;
		//fm_setCtlPageSize (n,a);

    	return true;
    
	} else if (a > 0 && gParams->ctl[n].class == CC_TRACKBAR || gParams->ctl[n].class == CC_SLIDER){

		 SendMessage (gParams->ctl[n].hCtl, TBM_SETTHUMBLENGTH, (WPARAM) fm_PixelsToVDBUs(a), 0);

    }
   
   return false;
   
} /*fm_setCtlThumbSize*/


/*************************************************************/
//
//  A couple of local helper functions...
//
/*************************************************************/

static void computeBuddy1Pos(INT_PTR n, INT_PTR x, INT_PTR y, INT_PTR w, INT_PTR h, RECT *pr)
{
    // Compute position of the numedit buddy for a
    // STANDARD control.
    
    //Changed by Harald Heim, May 1, 2003
    pr->right = gParams->ctl[n].editwidth; //24; //20;               //width


    pr->bottom = gParams->ctl[n].editheight; //10;              //height
    if (gParams->ctl[n].style & SBS_VERT) {
        //vertical scrollbar
        pr->left = (int)(x + (w - pr->right)/2);
#ifdef LABEL_ON_TOP
        pr->top = (int)(y + h + 2);
#else
        pr->top = (int)(y - pr->bottom - 2);
#endif
    }
    else {
        //horizontal scrollbar
        pr->left = (int)(x + w + 5);
        pr->top = (int)(y + (h - pr->bottom)/2);
    }
    //map DBUs to pixels...
    MapDialogRect(fmc.hDlg, pr);
} /*computeBuddy1Pos*/

static void computeBuddy2Pos(INT_PTR n, INT_PTR x, INT_PTR y, INT_PTR w, INT_PTR h, RECT *pr)
{
    // Compute position of the static text buddy for a
    // STANDARD control.
#if 0
    pr->right = gParams->ctl[n].sb.w2;               //width (pixels)
    pr->bottom = gParams->ctl[n].sb.h2;              //height (pixels)
    if (gParams->ctl[n].style & SBS_VERT) {
        //vertical scrollbar
        pr->left = x + (w - pr->right)/2;
#ifdef LABEL_ON_TOP
        pr->top = y - 10;
#else
        pr->top = y + h + 2;
#endif
    }
    else {
        //horizontal scrollbar
        pr->left = x - pr->right - 5;
#if 1
        pr->top = y + (h - pr->bottom)/2 + 1;
#else
        pr->top = y + (h - pr->bottom)/2;
#endif
    }
    //map DBUs to pixels...
    MapDialogRect(fmc.hDlg, pr);
#else
    pr->right = (int)w;
    pr->bottom = (int)h;
    if (gParams->ctl[n].style & SBS_VERT) {
        //vertical scrollbar
        pr->left = (int)x;
#ifdef LABEL_ON_TOP
        pr->top = (int)(y - 10);
#else
        pr->top = (int)(y + h + 2);
#endif
        //map DBUs to pixels...
        MapDialogRect(fmc.hDlg, pr);
        pr->left += (pr->right - gParams->ctl[n].sb.w2)/2;
    }
    else {
        //horizontal scrollbar
        pr->left = (int)(x - 5);
        pr->top = (int)y;
        //map DBUs to pixels...
        MapDialogRect(fmc.hDlg, pr);
        pr->left -= gParams->ctl[n].sb.w2;
        pr->top += (pr->bottom - gParams->ctl[n].sb.h2)/2;
    }
    pr->right = gParams->ctl[n].sb.w2;               //width (pixels)
    pr->bottom = gParams->ctl[n].sb.h2;              //height (pixels)
#endif
    //Info("Buddy2 pos = %d, %d, %d, %d", pr->left, pr->top, pr->right, pr->bottom);
} /*computeBuddy2Pos*/



int pixelUnit=0;	//not thread-safe!!!

/**********************************************************************/
/* setCtlPos(n, x, y, w, h)
/*
/*  Sets the position of the upper lefthand corner of control n to 
/*  coordinates (x,y) within the client area of the dialog box; sets 
/*  the width of the control to w and the height to h.  All measurements 
/*  are in dialog box units (DBUs).  To leave a particular parameter
/*  unchanged, specify a value of -1 for that parameter.
/*
/**********************************************************************/
INT_PTR fm_setCtlPos(INT_PTR n, INT_PTR x, INT_PTR y, INT_PTR w, INT_PTR h)
{
    RECT r;
    UINT flags = SWP_NOZORDER | SWP_NOACTIVATE | SWP_NOCOPYBITS;

    //added by Harald Heim, Dec 24, 2002
    if ( !IS_INUSE_CTL(n) &&
         n != CTL_PREVIEW  && n != CTL_PROGRESS && n != CTL_FRAME && n != CTL_ZOOM) {	//EH??
        return FALSE;
    }

	
	if (n == CTL_PREVIEW || n == CTL_PROGRESS || n == CTL_FRAME || n == CTL_ZOOM){
		if (x < 0 || y < 0) {
			flags |= SWP_NOMOVE;
		}
		if (w < 0 || h < 0) {
			flags |= SWP_NOSIZE;
		}
	}


if (pixelUnit==0) {

#if 0
    if (x < 0 || y < 0) {
        flags |= SWP_NOMOVE;
    }
    else {
        gParams->ctl[n].xPos = (int)x;
        gParams->ctl[n].yPos = (int)y;
    }
    if (w < 0 || h < 0) {
        flags |= SWP_NOSIZE;
    }
    else {
        gParams->ctl[n].width = (int)w;
        gParams->ctl[n].height = (int)h;
    }
#else
    if (x >= 0) gParams->ctl[n].xPos = (int)x;
    if (y >= 0) gParams->ctl[n].yPos = (int)y;
    if (w >= 0) gParams->ctl[n].width = (int)w;
    if (h >= 0) gParams->ctl[n].height = (int)h;
#endif

    /* Convert dialog units to pixels */
#if 0
    r.left = (int)x;
    r.top = (int)y;
    r.right = (int)w;
    r.bottom = (int)h;
#else
    x = r.left = gParams->ctl[n].xPos;
    y = r.top = gParams->ctl[n].yPos;
    w = r.right = gParams->ctl[n].width;
    h = r.bottom = gParams->ctl[n].height;
#endif

    MapDialogRect(fmc.hDlg, &r);

} else { //setCtlPixelPos

    if (x<0)
        x = fm_HDBUsToPixels(gParams->ctl[n].xPos);
    else {
        gParams->ctl[n].xPos = (int)fm_PixelsToHDBUs(x);
    }
    r.left = (int)x;
    
    if (y<0)
        y = fm_VDBUsToPixels(gParams->ctl[n].yPos);
    else {
        gParams->ctl[n].yPos = (int)fm_PixelsToVDBUs(y);
    }
    r.top = (int)y;

    if (w<0)
        w = fm_HDBUsToPixels(gParams->ctl[n].width);
    else {
        gParams->ctl[n].width = (int)fm_PixelsToHDBUs(w);
    }
    r.right = (int)w;
    
    if (h<0)
        h = fm_VDBUsToPixels(gParams->ctl[n].height);
    else {
        gParams->ctl[n].height = (int)fm_PixelsToVDBUs(h);
    }
    r.bottom = (int)h;

}


    if (gParams->ctl[n].class == CC_STANDARD || gParams->ctl[n].class == CC_SLIDER) {
        RECT r2;

        //reposition the numedit control
        computeBuddy1Pos(n, x, y, w, h, &r2);
        SetWindowPos(gParams->ctl[n].hBuddy1, NULL,
                        r2.left,    //x
                        r2.top,     //y
                        r2.right,   //w
                        r2.bottom,  //h
                        flags);

        //reposition the static text control
        computeBuddy2Pos(n, x, y, w, h, &r2);
        SetWindowPos(gParams->ctl[n].hBuddy2, NULL,
                        r2.left,    //x
                        r2.top,     //y
                        r2.right,   //w
                        r2.bottom,  //h
                        flags);
    } //if CC_STANDARD


    //added by Harald Heim, Dec 24, 2002
    if (n == CTL_PREVIEW) {					//missing 'else' or intentional?? See below...
        HWND hProxy;
        RECT imageRect;
        //// Why not set .hCtl = 101 for PREVIEW???

        //Auto position frame around preview
        SetWindowPos(GetDlgItem(fmc.hDlg, IDC_PROXY_GROUPBOX), NULL,
                            r.left-3,     //x
                            r.top-9,      //y
                            r.right+7,   //w
                            r.bottom+13,   //h
                            flags);
        
        // Invalidate Proxy Item
        hProxy = GetDlgItem(fmc.hDlg, 101);
        GetWindowRect(hProxy, &imageRect);
        ScreenToClient (fmc.hDlg, (LPPOINT)&imageRect);
        ScreenToClient (fmc.hDlg, (LPPOINT)&(imageRect.right));
        InvalidateRect (fmc.hDlg, &imageRect, TRUE);

        //Position preview
        SetWindowPos(GetDlgItem(fmc.hDlg, 101), NULL,
                            r.left,     //x
                            r.top,      //y
                            r.right,    //w
                            r.bottom,   //h
                            flags);
        //Update Preview
        ShowOutputBuffer(globals, MyHDlg, FALSE /*fDragging*/);

        //Make FilterMeister aware of the changed proxy preview
        GetProxyItemRect (MyHDlg, &gProxyRect);

        // TODO: Update tooltip rect if in use...

		if (!fScaling) fm_setCtlAnchor (n, gParams->ctl[n].anchor);

        return 1;


    
    } else if (n == CTL_ZOOM){

        HWND Control;
        int ret;


        Control = GetDlgItem(fmc.hDlg, IDC_BUTTON1);
        ret = SetWindowPos(Control, NULL,
                        r.left,     //x
                        r.top,      //y
                        (int)fm_HDBUsToPixels(12),    //w
                        (int)fm_VDBUsToPixels(10),   //h
                        flags);
        
        Control = GetDlgItem(fmc.hDlg, 498);
        ret = SetWindowPos(Control, NULL,
                        r.left + (int)fm_HDBUsToPixels(12),     //x
                        r.top,      //y
                        (int)fm_HDBUsToPixels(30),    //w
                        (int)fm_VDBUsToPixels(10),   //h
                        flags);


        Control = GetDlgItem(fmc.hDlg, IDC_BUTTON2);
        ret = SetWindowPos(Control, NULL,
                        r.left + (int)fm_HDBUsToPixels(42),     //x
                        r.top,      //y
                        (int)fm_HDBUsToPixels(12),    //w
                        (int)fm_VDBUsToPixels(10),   //h
                        flags);

		if (!fScaling) fm_setCtlAnchor (n, gParams->ctl[n].anchor);

        return ret;


    } else {	//Are CC_STANDARD and CC_SLIDER also meant to come here??
				//They *will*, because of a possible missing 'else' above!!

        HWND Control;
		int ret;

        // TODO: Set .hCtl for all predefined controls...

        if (n == CTL_ZOOM) Control = GetDlgItem(fmc.hDlg, IDC_BUTTON1);
        else if (n == CTL_FRAME) Control = GetDlgItem(fmc.hDlg, IDC_PROXY_GROUPBOX);
        else if (n == CTL_PROGRESS) Control = GetDlgItem(fmc.hDlg, IDC_PROGRESS1);
        else Control = gParams->ctl[n].hCtl;
        
        ret = SetWindowPos(Control, NULL,
                            r.left,     //x
                            r.top,      //y
                            r.right,    //w
                            r.bottom,   //h
                            flags);

		if (!fScaling) fm_setCtlAnchor (n, gParams->ctl[n].anchor);
        else {
            //scaling the control.  Force it to update value
            // (needed for STANDARD and SCROLLBAR controls to
            // keep the flashing highlight in sync with the thumb;
            // probably not needed by other controls, but better
            // safe than sorry..).
            if (GetFocus() == gParams->ctl[n].hCtl && 
                (gParams->ctl[n].class == CC_SCROLLBAR || gParams->ctl[n].class == CC_STANDARD) ) 
                    fm_setCtlVal(n, fmc.pre_ctl[n]);
        }

		return ret;

    }

} /*fm_setCtlPos*/


INT_PTR fm_setCtlPixelPos(INT_PTR n, INT_PTR x, INT_PTR y, INT_PTR w, INT_PTR h)
{
    INT_PTR retval;

    pixelUnit=1;	//not thread-safe!!
    retval = fm_setCtlPos(n,x,y,w,h);
    pixelUnit=0;

    return retval;

} /*fm_setCtlPixelPos*/



INT_PTR fm_setCtlEditSize(INT_PTR n, INT_PTR w, INT_PTR h)
{

    if (gParams->ctl[n].class == CC_STANDARD || gParams->ctl[n].class == CC_SLIDER) {
        RECT r1,r2;
        UINT flags = SWP_NOZORDER | SWP_NOACTIVATE | SWP_NOCOPYBITS;

        r1.left = 0; //If these two lines are not used -> Crash under 98 and ME //eh?? garbage for MapDialogRect??-afh
        r1.top = 0;
        gParams->ctl[n].editwidth =  r1.right = (int)w;
        gParams->ctl[n].editheight = r1.bottom = (int)h;
        
        MapDialogRect(fmc.hDlg, &r1);

        //reposition the numedit control
        computeBuddy1Pos(n, gParams->ctl[n].xPos, gParams->ctl[n].yPos, gParams->ctl[n].width, gParams->ctl[n].height, &r2);

        //Assign new width and height
        r2.right = r1.right;
        r2.bottom = r1.bottom;

        SetWindowPos(gParams->ctl[n].hBuddy1, NULL,
                        r2.left,    //x
                        r2.top,     //y
                        r2.right,   //w
                        r2.bottom,  //h
                        flags);

        return true;

    } //if CC_STANDARD

    return false;

} /*fm_setCtlEditSize*/



INT_PTR fm_setEditBoxUpdate(INT_PTR ms)
{   

    EditBoxUpdate = (int)ms;    

    return true;
}



INT_PTR fm_getCtlRange(INT_PTR n, INT_PTR w)
{   
    
    if (n==CTL_PROGRESS){
        
        if (fmc.doingProxy) {
            if (w==0)
                return SendMessage(GetDlgItem(MyHDlg, IDC_PROGRESS1), PBM_GETRANGE, TRUE, 0);
            else
                return SendMessage(GetDlgItem(MyHDlg, IDC_PROGRESS1), PBM_GETRANGE, FALSE, 0);
        }

    } else if (IS_INUSE_CTL(n)){
        
        if (w==0)
            return gParams->ctl[n].minval;
        else
            return gParams->ctl[n].maxval;
    }

    return 0;            

}



/**********************************************************************/
/* setCtlAction(n, a)
/*
/*  Sets the default action for control n to a, where a is CA_NONE,
/*  CA_CANCEL, CA_APPLY, CA_PREVIEW, CA_EDIT, CA_ABOUT, or CA_RESET.
/*  Returns the previous default action for the control (or 0 if the
/*  control index is invalid).
/*
/*  The default action is the action that will be taken if the default
/*  OnCtl handler is invoked for control n.  The meaning of each action
/*  is:
/*
/*  CA_NONE     No action.  This is the default action for radio buttons.
/*
/*  CA_CANCEL   FM exits, leaving the original source image unaltered.
/*              This is the default action for the Cancel button,
/*
/*  CA_APPLY    The filter is applied to the original source image,
/*              and FM exits.  This is the default action for the OK
/*              button.
/*
/*  CA_PREVIEW  The filter is applied to the proxy image, and all previews
/*              are updated.  This is the default action for most user
/*              controls, including the STANDARD, SCROLLBAR, TRACKBAR,
/*              PUSHBUTTON, and CHECKBOX controls.
/*
/*  CA_EDIT     FM enters or leaves Edit Mode.  This is the default
/*              action for the Edit control.
/*
/*  CA_ABOUT    FM displays the ABOUT dialog box.
/*
/*  CA_RESET    Resets all user controls to their initial values.
/*
/**********************************************************************/
INT_PTR fm_setCtlAction(INT_PTR n, INT_PTR a)
{
    int prev_action;

    if (!IS_INUSE_CTL(n)) {
        return 0;
    }

    prev_action = gParams->ctl[n].action;
    gParams->ctl[n].action = (int)a;
    return prev_action;
} /*fm_setCtlAction*/

/**********************************************************************/
/* fm_enableCtl(n, level)
/*
/*  Determines whether control n is visible and whether it is enabled
/*  by specifying level = 0 (invisible and disabled), 1 (visible but
/*  disabled), or 3 (visible and enabled).  Returns the previous enable
/*  level, or 0 if n is not a valid control index.
/*
/*  NOTE: Due to an early documentation error, level=-1 is considered
/*        equivalent to level=3.  However, other bits in level are
/*        reserved for future use, and should not be randomly set!
/*
/**********************************************************************/
INT_PTR fm_enableCtl(INT_PTR n, INT_PTR level)
{
    int prev_level;

    if (!IS_OK_CTL_NR(n)) {						//no test for inuse???
        return 0;   //avoid array indexing error
    }

    if (level == -1) level = 3; // see documentation note above

    prev_level = gParams->ctl[n].enabled;
    gParams->ctl[n].enabled = (int)level & 3;    //clear reserved bits for now
	enableCtl((int)n);
    return prev_level;
} /*fm_enableCtl*/

/**********************************************************************/
/* enableCtl(n) support function
/*
/*  Determines whether a control should be visible, disabled or 
/*	invisible and sets it as such. Returns 0.
/*
/**********************************************************************/
int enableCtl(int n)
{
    int nIDDlgItem;

    //if (level == -1) level = 3; // see documentation note above

    if ( !IS_INUSE_CTL(n) && n != CTL_PROGRESS && n != CTL_FRAME && n != CTL_ZOOM) { //&& n != CTL_PREVIEW
        return 0;
    }

	if (n == CTL_PROGRESS) nIDDlgItem = IDC_PROGRESS1;
    else if (n == CTL_FRAME) nIDDlgItem = IDC_PROXY_GROUPBOX;
    else if (n == CTL_ZOOM) nIDDlgItem = 498; //Zoom Label
	//else if (n == CTL_PREVIEW) nIDDlgItem = 101;
    else  nIDDlgItem = IDC_CTLBASE+n;

	//Causes controls to appear when running FM a second time in Photoshop
	//if (level==0 && gParams->ctl[n].enabled == 0 && n != CTL_PROGRESS && n != CTL_FRAME && n != CTL_ZOOM){
	//if (level==0 && IsWindowVisible (GetDlgItem(fmc.hDlg, nIDDlgItem))==0 ){
	//	return 0;
	//}

	//MWVDL 2008-07-18
    {//scope
	int level = (int)fm_ctlEnabledAs(n);
#if 0
	if (level == -1) level = 3;
#endif
	
    // enable or disable the control
    EnableWindow(GetDlgItem(fmc.hDlg, nIDDlgItem), (level & 2) != 0);
    // show or hide the control
    ShowWindow(GetDlgItem(fmc.hDlg, nIDDlgItem), (level & 1)?SW_SHOW:SW_HIDE);

    if (n == CTL_ZOOM){
        // enable or disable the control
        EnableWindow(GetDlgItem(fmc.hDlg, IDC_BUTTON1), (level & 2) != 0);
        // show or hide the control
        ShowWindow(GetDlgItem(fmc.hDlg, IDC_BUTTON1), (level & 1)?SW_SHOW:SW_HIDE);

        // enable or disable the control
        EnableWindow(GetDlgItem(fmc.hDlg, IDC_BUTTON2), (level & 2) != 0);
        // show or hide the control
        ShowWindow(GetDlgItem(fmc.hDlg, IDC_BUTTON2), (level & 1)?SW_SHOW:SW_HIDE);
    }

    if (gParams->ctl[n].class == CC_STANDARD || gParams->ctl[n].class == CC_SLIDER) {
        // enable or disable the control edit field
        EnableWindow(GetDlgItem(fmc.hDlg, IDC_BUDDY1+n), (level & 2) != 0);
        // enable or disable the control text label (no, looks crappy)
        EnableWindow(GetDlgItem(fmc.hDlg, IDC_BUDDY2+n), (level & 2) != 0);
        //Avoid redraw problem // Doesn't work
        //InvalidateRect(GetDlgItem(fmc.hDlg, IDC_BUDDY2+n), NULL, TRUE );
        //UpdateWindow (GetDlgItem(fmc.hDlg, IDC_BUDDY2+n));

        // show or hide the control edit field
        ShowWindow(GetDlgItem(fmc.hDlg, IDC_BUDDY1+n), (level & 1)?SW_SHOW:SW_HIDE);
        // show or hide the control text label
        ShowWindow(GetDlgItem(fmc.hDlg, IDC_BUDDY2+n), (level & 1)?SW_SHOW:SW_HIDE);
    } //if CC_STANDARD

//	    else if (level != 1 && gParams->ctl[n].class == CC_TAB){		
	    else if (gParams->ctl[n].class == CC_TAB){						//MWVDL 2008-08-17: Without the level-check, disabling of controls on the tab should be possible.
		int i;
		for (i = 0; i < N_CTLS; ++i) {
				if (gParams->ctl[i].inuse								// Only process children of this tab.
				 && gParams->ctl[i].tabControl == n
				 && gParams->ctl[i].tabSheet == gParams->ctl[n].val) {
					enableCtl(i);	//MWVDL 2008-06-17
				}
		    }
	    }
    }//scope

	return 0;
} /*enableCtl*/

/**********************************************************************/
/* deleteCtl(n)
/*
/*  Deletes control n.
/*
/**********************************************************************/
static INT_PTR fm_deleteCtl(INT_PTR n)
{
    if (!IS_OK_CTL_NR(n)) {
        //invalid control id
        return FALSE;
    }

    if (n == CTL_PROGRESS || n == CTL_FRAME || n == CTL_ZOOM){
    
        int nIDDlgItem;

        if (n == CTL_PROGRESS) nIDDlgItem = IDC_PROGRESS1;
        else if (n == CTL_FRAME) nIDDlgItem = IDC_PROXY_GROUPBOX;
        else if (n == CTL_ZOOM) nIDDlgItem = 498; //Zoom Label

        DestroyWindow (GetDlgItem(fmc.hDlg, nIDDlgItem));
        
        if (n == CTL_ZOOM){
            DestroyWindow (GetDlgItem(fmc.hDlg, IDC_BUTTON1));
            DestroyWindow (GetDlgItem(fmc.hDlg, IDC_BUTTON2));
        }

		gParams->ctl[n].enabled = 0;
        
        
    } else if (gParams->ctl[n].inuse) {

        //delete any tool tip
        fm_setCtlToolTip(n, 0, 0);

        //delete any associated image resources
        fm_setCtlImage(n, 0, 0);

        if (gParams->ctl[n].hCtl) {
            //destroy the control
            DestroyWindow(gParams->ctl[n].hCtl);
            gParams->ctl[n].hCtl = NULL;
        }
        if (gParams->ctl[n].hBuddy1) {
            //destroy the first buddy control
            DestroyWindow(gParams->ctl[n].hBuddy1);
            gParams->ctl[n].hBuddy1 = NULL;
			gParams->ctl[n].buddy1Style = 0;
			gParams->ctl[n].buddy1StyleEx = 0;
        }
        if (gParams->ctl[n].hBuddy2) {
            //destroy the second buddy control
            DestroyWindow(gParams->ctl[n].hBuddy2);
            gParams->ctl[n].hBuddy2 = NULL;
			gParams->ctl[n].buddy2Style = 0;
			gParams->ctl[n].buddy2StyleEx = 0;
        }
        gParams->ctl[n].inuse = FALSE;
    }
    return TRUE;    //success
} /*fm_deleteCtl*/

/**********************************************************************/
/* createPredefinedCtls()
/*
/*  Creates the pre-defined user controls.
/*
/**********************************************************************/
void createPredefinedCtls(void)
{

    //OK button
    fm_createCtl(CTL_OK, CC_PUSHBUTTON, (INT_PTR)"OK", -1, -1, -1, -1, BS_DEFPUSHBUTTON, 0, 0, 3);
	fm_setCtlAnchor (CTL_OK,ANCHOR_BOTTOM|ANCHOR_RIGHT);
    fm_setCtlScripting (CTL_OK, false);
    
    //Cancel button
	fm_createCtl(CTL_CANCEL, CC_PUSHBUTTON, (INT_PTR)"Cancel", -1, -1, -1, -1, 0, 0, 0, 3);
	fm_setCtlAnchor (CTL_CANCEL,ANCHOR_BOTTOM|ANCHOR_RIGHT);
    fm_setCtlScripting (CTL_CANCEL, false);
	
    //Commented out by Harald Heim, March 9, 2003
    //FM Logo
    //fm_createCtl(CTL_LOGO, CC_IMAGE, 0, -1, -1, -1, -1, SS_NOTIFY, 0, 0, 3);
    //fm_setCtlImage(CTL_LOGO, (int)"logo2.bmp", 'B');

#ifndef STANDALONE
    //Commented out by Harald Heim, March 24, 2003
    //Edit button
    //fm_createCtl(CTL_EDIT, CC_PUSHBUTTON, 0, -1, -1, -1, -1, 0, 0, 0, 3);
#endif //STANDALONE

    setDefaultWindowTheme(GetDlgItem(MyHDlg, IDC_PROGRESS1));
    setDefaultWindowTheme(GetDlgItem(MyHDlg, IDC_BUTTON1));
    setDefaultWindowTheme(GetDlgItem(MyHDlg, IDC_BUTTON2));
    setDefaultWindowTheme(GetDlgItem(MyHDlg, IDC_PROXY_GROUPBOX));

    //Fill in fields for other predefined controls...
    gParams->ctl[CTL_PROGRESS].hCtl = GetDlgItem(MyHDlg, IDC_PROGRESS1);
    ///// gParams->ctl[CTL_PROGRESS].inuse = true;  //breaks PROGRESS bar (scaled twice, created dynamically
                                                    //as well as statically, anything else?)
    gParams->ctl[CTL_PROGRESS].enabled = 3;

	//Set values for existing controls
	{
		RECT rcCtl;
    
		GetWindowRect( GetDlgItem(fmc.hDlg, 101), &rcCtl );
		ScreenToClient (fmc.hDlg, (LPPOINT)&rcCtl);
		ScreenToClient (fmc.hDlg, (LPPOINT)&(rcCtl.right));
		gParams->ctl[CTL_PREVIEW].width = (int)fm_PixelsToHDBUs(rcCtl.right - rcCtl.left);
		gParams->ctl[CTL_PREVIEW].height = (int)fm_PixelsToVDBUs(rcCtl.bottom - rcCtl.top);
		gParams->ctl[CTL_PREVIEW].xPos = (int)fm_PixelsToHDBUs(rcCtl.left);
		gParams->ctl[CTL_PREVIEW].yPos = (int)fm_PixelsToVDBUs(rcCtl.top);

		GetWindowRect( GetDlgItem(fmc.hDlg, IDC_PROGRESS1), &rcCtl );
		ScreenToClient (fmc.hDlg, (LPPOINT)&rcCtl);
		ScreenToClient (fmc.hDlg, (LPPOINT)&(rcCtl.right));
		gParams->ctl[CTL_PROGRESS].width = (int)fm_PixelsToHDBUs(rcCtl.right - rcCtl.left);
		gParams->ctl[CTL_PROGRESS].height = (int)fm_PixelsToVDBUs(rcCtl.bottom - rcCtl.top);
		gParams->ctl[CTL_PROGRESS].xPos = (int)fm_PixelsToHDBUs(rcCtl.left);
		gParams->ctl[CTL_PROGRESS].yPos = (int)fm_PixelsToVDBUs(rcCtl.top);

		GetWindowRect( GetDlgItem(fmc.hDlg, IDC_BUTTON1), &rcCtl );
		ScreenToClient (fmc.hDlg, (LPPOINT)&rcCtl);
		ScreenToClient (fmc.hDlg, (LPPOINT)&(rcCtl.right));
		gParams->ctl[CTL_ZOOM].width = (int)fm_PixelsToHDBUs(rcCtl.right - rcCtl.left);
		gParams->ctl[CTL_ZOOM].height = (int)fm_PixelsToVDBUs(rcCtl.bottom - rcCtl.top);
		gParams->ctl[CTL_ZOOM].xPos = (int)fm_PixelsToHDBUs(rcCtl.left);
		gParams->ctl[CTL_ZOOM].yPos = (int)fm_PixelsToVDBUs(rcCtl.top);
	}

	fm_setCtlAnchor (CTL_PREVIEW,ANCHOR_LEFT|ANCHOR_RIGHT|ANCHOR_TOP|ANCHOR_BOTTOM);
	fm_setCtlAnchor (CTL_PROGRESS,ANCHOR_BOTTOM);
	fm_setCtlAnchor (CTL_ZOOM,ANCHOR_BOTTOM);

    fm_setCtlScripting (CTL_PREVIEW, false);
    fm_setCtlScripting (CTL_PROGRESS, false);
    fm_setCtlScripting (CTL_ZOOM, false);

} /*createPredefinedCtls*/

/**********************************************************************/
/* resetAllCtls()
/*
/*  Resets all controls (deletes them).
/*
/**********************************************************************/
INT_PTR fm_resetAllCtls(void)
{
    int i;

    //Delete any previously defined controls.
    for (i=0; i < N_CTLS; i++) {
        if (gParams->ctl[i].inuse) {
            fm_deleteCtl(i);
        }
    }

    //Now define the default controls:
    createPredefinedCtls();

    //This is also a good place (for now) to reinitialize
    //some dialog attributes...

    //set clipping region to NULL.
    SetWindowRgn(fmc.hDlg, NULL, TRUE /*redraw*/);
    gParams->gDialogRegion = NULL;

    //set other dialog parameters back to defaults
#if 0
    gParams->gDialogGradientColor1 = RGB(0,0,0);        //black
    gParams->gDialogGradientColor2 = RGB(0,0,0);        //black
#else
    //Added by Harald Heim, March 9, 2003
    gParams->gDialogGradientColor1 = GetSysColor(COLOR_BTNFACE); //RGB(0xD4,0xD0,0xC8);   //yellow grey
    gParams->gDialogGradientColor2 = GetSysColor(COLOR_BTNFACE); //RGB(0xD4,0xD0,0xC8);   //yellow grey
    //gParams->gDialogGradientColor1 = RGB(0x5F,0x9E,0xA0);   //X11.cadetblue
    //gParams->gDialogGradientColor2 = RGB(0x5F,0x9E,0xA0);   //X11.cadetblue
#endif
    gParams->gDialogGradientDirection = 0;              //vertical
    gParams->gDialogImage_save = gDialogImage = NULL;   //no background image
    gParams->gDialogImageFileName[0] = '\0';
    gParams->gDialogImageMode = DIM_TILED;              //TILED
    gParams->gDialogImageStretchMode = COLORONCOLOR;
    gParams->gDialogDragMode = 0;                       //drag by titlebar
#if 0
    gParams->gDialogTitle[0] = '\0';                    //default dialog title
#elif 1
    //default dialog title is "filter title (image mode, filter case)"
    strcpy(gParams->gDialogTitle, "!t (!M, !f)");
#else
    //default dialog title is "filter title (filter case) [image mode]"
    strcpy(gParams->gDialogTitle, "!t (!f) [!M]");
#endif

	gParams->gDialogTheme = 0; //Default Dialog Theme


    //initialize registry access...
    fm_setRegRoot(HKEY_CURRENT_USER);
    fm_setRegPath("Software\\!O\\!C\\!t");

    // Initialize built-in strings...
    gParams->filterAuthorText[0] = 0;
    //gParams->filterCategoryText[0] = 0;       //NO!! Clobbers val set at compile time
    //gParams->filterTitleText[0] = 0;          //NO!! Clobbers val set at compile time
    //gParams->filterFilenameText[0] = 0;       //NO!! Clobbers val set at compile time
    gParams->filterCopyrightText[0] = 0;
    gParams->filterDescriptionText[0] = 0;
    gParams->filterOrganizationText[0] = 0;
    gParams->filterURLText[0] = 0;
    gParams->filterVersionText[0] = 0;
    //gParams->filterAboutText[0] = 0;          //NO!! Clobbers val set at compile time
    //fmc.filterUniqueID[0] = 0;

    // Initialize built-in strings...
    fmc.filterAuthorText[0] = 0;
    //fmc.filterCategoryText[0] = 0;            //NO!! Clobbers val set at compile time
    //fmc.filterTitleText[0] = 0;               //NO!! Clobbers val set at compile time
    //fmc.filterFilenameText[0] = 0;            //NO!! Clobbers val set at compile time
    fmc.filterCopyrightText[0] = 0;
    fmc.filterDescriptionText[0] = 0;
    fmc.filterOrganizationText[0] = 0;
    fmc.filterURLText[0] = 0;
    fmc.filterVersionText[0] = 0;
    //fmc.filterUniqueID[0] = 0;

#ifndef DONT_SAVE_GLOBALS
    //Zero out the predefined globals (but not filterCategoryText, etc.!!!) */
    memset(&fmc.globvars, 0, sizeof(fmc.globvars));
#endif //DONT_SAVE_GLOBALS

    //Restore initial dialog styles...
    SetWindowLong(fmc.hDlg, GWL_STYLE, gParams->gDialogInitialStyle | WS_VISIBLE);
    SetWindowLong(fmc.hDlg, GWL_EXSTYLE, gParams->gDialogInitialStyleEx);

#if 1
    //turn this on only if needed to force update of dialog styles
    SetWindowPos(fmc.hDlg, NULL,
                 0, 0, 0, 0,
                 SWP_FRAMECHANGED|SWP_NOZORDER|SWP_NOACTIVATE|SWP_NOSIZE|SWP_NOMOVE);
#endif

#if 1
    //turn this on only if necessary to force update of dialog background and fonts
    return RedrawWindow(
            fmc.hDlg,   // handle of window
            NULL,   // address of structure with update rectangle, NULL=>entire client area
            NULL,   // handle of update region
            RDW_ERASE|RDW_INVALIDATE|RDW_ERASENOW|RDW_UPDATENOW     // array of redraw flags
       );
#else
    return TRUE;
#endif
} /*fm_resetAllCtls*/

/**********************************************************************/
/* createCtl(n, c, t, x, y, w, h, s, sx, p, e)
/*
/*  Dynamically creates a control with index n, class c, text t,
/*  coordinates (x,y), width w, height h, style s, extended style sx,
/*  properties p, and enable level e.
/*  All measurements are in DBUs.
/*  For x, y, w, and h, a value of -1 means use the default value.
/*
/*  *** NFG *** doesn't handle negative coords for multiple screens!!
/*
/**********************************************************************/
INT_PTR fm_createCtl(INT_PTR n, INT_PTR c, INT_PTR t, INT_PTR x, INT_PTR y,
					 INT_PTR w, INT_PTR h, INT_PTR s, INT_PTR sx, INT_PTR p,
					 INT_PTR e)
{
    if (!IS_OK_CTL_NR(n)) {
        //invalid control id
        return 0;
    }

#ifdef STANDALONE
    // Refuse to create CTL_EDIT in a standalone filter.
    if (n == CTL_EDIT) return 0;
#endif

    if (gParams->ctl[n].inuse) {
        // kill previous control with this id
        fm_deleteCtl(n);
    }

    gParams->ctl[n].class = (int)c;
    gParams->ctl[n].hCtl = NULL;
    gParams->ctl[n].hBuddy1 = NULL;
    gParams->ctl[n].hBuddy2 = NULL;

#ifndef STANDALONE
    //Override text for CTL_EDIT...
    if (n == CTL_EDIT) {
        t = (gParams->flags & AFH_EDITMODE) ?
                     (INT_PTR)"&Edit <<<" :
                     (INT_PTR)"&Edit >>>";
    }
#endif
    if (t) {
        strncpy(gParams->ctl[n].label, (char *)t, MAX_LABEL_SIZE+1);
        gParams->ctl[n].label[MAX_LABEL_SIZE] = '\0';   //ensure null-terminated.
    }
    else {
        //No text string
        gParams->ctl[n].label[0] = '\0';
    }

    // -1 means use default value for x, y, w, or h.

    // Default size depends on control class (and sometimes
    // control index, style, or properties)...
    switch (c) {
    case CC_PUSHBUTTON:
        if (w < 0) w = 34;
        if (h < 0) h = 14;
        break;
    case CC_RECT:
    case CC_FRAME:
    case CC_IMAGE:
    case CC_BITMAP:
    case CC_ICON:
    case CC_METAFILE:
    case CC_ANIMATION:
        if (w < 0) w = 32;
        if (h < 0) h = (n == CTL_LOGO) ? 43 : 32;
        break;
    case CC_GROUPBOX:
        if (w < 0) w = 100;
        if (h < 0) h = 32;
        break;
    case CC_LISTBOX:
    case CC_COMBOBOX:
        if (w < 0) w = 90;
        if (h < 0) h = 32;
        break;
    case CC_STANDARD:
    case CC_SCROLLBAR:
        if (s & SBS_VERT) {
            //vertical scrollbar
            if (w < 0) w = 8;
            if (h < 0) h = 90;
        }
        else {
            //horizontal scrollbar
            if (w < 0) w = 90;
            if (h < 0) h = 8;
        }
        break;
    case CC_TRACKBAR:
	case CC_SLIDER:
        if (s & TBS_VERT) {
            //vertical trackbar
            if (w < 0) w = 10;
            if (h < 0) h = 90;
        }
        else {
            //horizontal trackbar
            if (w < 0) w = 90;
            if (h < 0) h = 10;
        }
        break;
    case CC_PREVIEW:
        if (w < 0) w = 228;
        if (h < 0) h = 154;
        break;
    case CC_TAB:
		if (w < 0) w = 120;
        if (h < 0) h = 60;
		break;
	case CC_SPINNER:
    case CC_UPDOWN:
    case CC_CHECKBOX:
    case CC_RADIOBUTTON:
    case CC_OWNERDRAW:
    case CC_PROGRESSBAR:
    case CC_EDIT:
    case CC_STATICTEXT:
    case CC_TOOLTIP:
    default:
        if (w < 0) w = 90;
        if (h < 0) h = 8;
        break;
    }//switch class

    // Default position depends on whether the control is
    // a predefined control...
    switch (n) {

    //Changed by Harald Heim, March 9, 2003
    case CTL_OK:
        if (x < 0) x = 408;//306;
        if (y < 0) y = 160;//126;
        break;
    case CTL_CANCEL:
        if (x < 0) x = 446;//269;
        if (y < 0) y = 160;//126;
        break;
#ifndef STANDALONE
    case CTL_EDIT:
        if (x < 0) x = 370;//231;
        if (y < 0) y = 160;//126;
        break;
#endif
    //case CTL_LOGO:
    //    if (x < 0) x = 352;
    //    if (y < 0) y = 97;
    //    break;


    case CTL_PREVIEW:
        if (x < 0) x = 5;
        if (y < 0) y = 5;
        break;
	case CTL_PROGRESS:
    case CTL_HOST:
    default:

        //Changed by Harald Heim, March 9, 2003
        if (x < 0) x = 350 - (n/10 % 10)*5;
        if (y < 0) y = 10*(n % 10) + 5;

        break;
    }//switch n

    // Default action depends on control class (and sometimes
    // control index)...
    switch (c) {
    case CC_RADIOBUTTON:
    case CC_GROUPBOX:
    case CC_LISTBOX:
    case CC_COMBOBOX:
	case CC_EDIT:
	case CC_TAB:
        // Default action for these is NONE
        gParams->ctl[n].action = CA_NONE;
        break;
    case CC_PUSHBUTTON:
    case CC_STANDARD:
	case CC_SLIDER:
    case CC_SCROLLBAR:
    case CC_TRACKBAR:
    case CC_SPINNER:
    case CC_UPDOWN:
    case CC_CHECKBOX:
    case CC_OWNERDRAW:
    case CC_PROGRESSBAR:
    case CC_STATICTEXT:
    case CC_RECT:
    case CC_FRAME:
    case CC_IMAGE:
    case CC_BITMAP:
    case CC_ICON:
    case CC_METAFILE:
    case CC_TOOLTIP:
    case CC_PREVIEW:
    case CC_ANIMATION:
    default:
        //default action for most controls is PREVIEW
        gParams->ctl[n].action = CA_PREVIEW;
        break;
    }//switch class

    //Default actions for predefined controls are different...
    switch (n) {
    case CTL_OK:
        gParams->ctl[n].action = CA_APPLY;
        break;
    case CTL_CANCEL:
        gParams->ctl[n].action = CA_CANCEL;
        break;
#ifndef STANDALONE
    case CTL_EDIT:
        gParams->ctl[n].action = CA_EDIT;
        break;
#endif
    case CTL_LOGO:
        gParams->ctl[n].action = CA_ABOUT;
        break;
    case CTL_PREVIEW:
    case CTL_PROGRESS:
    case CTL_HOST:
    default:
        break;
    }//switch n


    gParams->ctl[n].xPos = (int)x;
    gParams->ctl[n].yPos = (int)y;
    gParams->ctl[n].width = (int)w;
    gParams->ctl[n].height = (int)h;
    gParams->ctl[n].style = (int)s;
    gParams->ctl[n].styleEx = (int)sx;
    gParams->ctl[n].properties = (int)p;
    gParams->ctl[n].hBrush = GetStockObject(HOLLOW_BRUSH);
#if 0
    gParams->ctl[n].bkColor = GetSysColor(COLOR_BTNFACE);
#else
    if (c == CC_GROUPBOX) //Avoid that text of a group box is stricken through
        gParams->ctl[n].bkColor = GetSysColor(COLOR_BTNFACE);
    else
        gParams->ctl[n].bkColor = (COLORREF) -1;    //transparent
#endif


#if 0
    gParams->ctl[n].textColor = GetSysColor(COLOR_BTNTEXT);
#else
    //Changed by Harald Heim, March 22, 2003
    //gParams->ctl[n].textColor = RGB(255,255,255);  //fontcolor=white
    gParams->ctl[n].textColor = GetSysColor(COLOR_WINDOWTEXT);

#endif
    gParams->ctl[n].hFont = GetStockObject(DEFAULT_GUI_FONT);
    gParams->ctl[n].fontSize = 8;

    gParams->ctl[n].enabled = (int)e;

    gParams->ctl[n].val = 0;
    gParams->ctl[n].initVal = 0;
    gParams->ctl[n].minval = 0;
    gParams->ctl[n].maxval = 255;
    gParams->ctl[n].lineSize = 1;
    gParams->ctl[n].pageSize = 10;
	gParams->ctl[n].thumbSize = 0;
    gParams->ctl[n].divisor = 1;
    gParams->ctl[n].tooltip[0] = '\0';
    gParams->ctl[n].image[0] = '\0';
    gParams->ctl[n].imageType = 0;

    switch (c) {
    case CC_STANDARD:
    case CC_SCROLLBAR:
	case CC_TRACKBAR:
	case CC_SLIDER:
        {                   ///////  BUG!!!!  Buddy2 (static text) should be created first
                            ///////           for correct tab order!!
            //create a scrollbar with a static text buddy
            //and a NUMEDIT buddy.
            HWND hCtl, hBuddy1, hBuddy2;//, hBuddy3;
            int alignStyle;
            /* Convert dialog units to pixels */
            RECT r;

			r.left = (int)x;
			r.top = (int)y;
			r.right = (int)w;
			r.bottom = (int)h;
			MapDialogRect(fmc.hDlg, &r);


			if (c==CC_TRACKBAR || c==CC_SLIDER){
				
				// Should link dynamically, so only load comctl32.dll
				// if necessary!
				InitCommonControls();

				//r.left = x;
				//r.top = y;
				//r.right = w;
				//r.bottom = h;
				//MapDialogRect(fmc.hDlg, &r);

				//Default tick frequency is 1.
				gParams->ctl[n].tb.ticFreq = 1;

				// force default styles (note TBS_HORZ==0 and TBS_BOTTOM==0).
				gParams->ctl[n].style |= WS_VISIBLE | WS_CHILD | WS_TABSTOP;

				// Transparent backgrounds don't work well for trackbars,
				// so change the default to opaque light gray.
				DeleteObject(gParams->ctl[n].hBrush); //superfluous for stock object
				gParams->ctl[n].hBrush = CreateSolidBrush(GetSysColor(COLOR_BTNFACE));//GetStockObject(LTGRAY_BRUSH);
				gParams->ctl[n].bkColor = GetSysColor(COLOR_BTNFACE);

				// create the trackbar control
				hCtl = CreateWindowEx(  
					gParams->ctl[n].styleEx,        // extended styles
					TRACKBAR_CLASS,                 // predefined class 
					"",                             // control text
					gParams->ctl[n].style,          // styles
					r.left,                         // starting x position 
					r.top,                          // starting y position 
					r.right,                        // control width 
					r.bottom,                       // control height 
					fmc.hDlg,                       // parent window 
					(HMENU)(INT_PTR)(IDC_CTLBASE+n),// control ID
					hDllInstance,
					NULL);                          // pointer not needed 

				if (!hCtl) return FALSE;    //failed to create window.
				gParams->ctl[n].hCtl = hCtl;

				// Turn off XP Visual Styles
				//setDefaultWindowTheme(hCtl);
				gParams->ctl[n].inuse = TRUE;
				fm_setCtlTheme(n, (int)gParams->ctl[n].theme);


				if (gParams->ctl[n].style & TBS_TOOLTIPS) {
					int side;
					//set side for automatic tooltips
					if (gParams->ctl[n].properties & CTP_BOTTOM) {
						side = TBTS_BOTTOM;
					}
					else if (gParams->ctl[n].properties & CTP_RIGHT) {
						side = TBTS_RIGHT;
					}
					else if (gParams->ctl[n].properties & CTP_LEFT) {
						side = TBTS_LEFT;
					}
					else {
						//default is TOP
						side = TBTS_TOP;
					}
					SendMessage(gParams->ctl[n].hCtl, TBM_SETTIPSIDE, side, 0);
				}


			} else {

				//r.left = x;
				//r.top = y;
				//r.right = w;
				//r.bottom = h;
				//MapDialogRect(fmc.hDlg, &r);

				// force default styles (note SBS_HORZ==0).
				gParams->ctl[n].style |= WS_VISIBLE | WS_CHILD | WS_TABSTOP;

				// create the scrollbar control
				hCtl = CreateWindowEx(  
					gParams->ctl[n].styleEx,        // extended styles
					"SCROLLBAR",                    // predefined class 
					"",                             // control text
					gParams->ctl[n].style           // styles
					& ~WS_BORDER,                   // ...except WS_BORDER (see below)
					r.left,                         // starting x position 
					r.top,                          // starting y position 
					r.right,                        // control width 
					r.bottom,                       // control height 
					fmc.hDlg,                       // parent window 
					(HMENU)(INT_PTR)(IDC_CTLBASE+n),// control ID
					hDllInstance,
					NULL);                          // pointer not needed 

				if (!hCtl) return FALSE;    //failed to create window.
				gParams->ctl[n].hCtl = hCtl;

				// Turn off XP Visual Styles
				//setDefaultWindowTheme(hCtl);
			}


			
            if (c != CC_STANDARD && c != CC_SLIDER) break;

            //create the NUMEDIT buddy control
            if (s & SBS_VERT || s & TBS_VERT) {
                //vertical scrollbar
                alignStyle = ES_CENTER;
            }
            else {
                //horizontal scrollbar
                alignStyle = ES_LEFT;
            }
            
			//gParams->ctl[n].buddy1StyleEx |= WS_EX_STATICEDGE; //staticedge replaces black border as default?
			
			computeBuddy1Pos(n, x, y, w, h, &r);

            hBuddy1 = CreateWindowEx(  
                gParams->ctl[n].buddy1StyleEx, //0 /*WS_EX_STATICEDGE*/,               // extended styles
                "EDIT",                         // predefined class 
                "",                            // control text
                alignStyle | WS_VISIBLE | WS_CHILD /*| WS_BORDER*/ | WS_TABSTOP
#if 1
                | ES_MULTILINE | ES_WANTRETURN  // to get EN_MAXTEXT when user hits return
#endif
                /*| ES_NUMBER*/,                // styles (ES_NUMBER doesn't allow negative numbers!)
                r.left,                         // starting x position 
                r.top,                          // starting y position 
                r.right,                        // button width 
                r.bottom,                       // button height 
                fmc.hDlg,                       // parent window 
                (HMENU)(INT_PTR)(IDC_BUDDY1+n), // control ID
                hDllInstance,
                NULL);                          // pointer not needed 

            if (!hBuddy1) return FALSE;    //failed to create window.
            gParams->ctl[n].hBuddy1 = hBuddy1;

			//Add border here, otherwise we cannot remove it anymore
			SetWindowLong(gParams->ctl[n].hBuddy1, GWL_STYLE,
				GetWindowLong(gParams->ctl[n].hBuddy1, GWL_STYLE) | WS_BORDER);
			gParams->ctl[n].buddy1Style = WS_BORDER;

			SetWindowPos(gParams->ctl[n].hBuddy1, NULL, 0, 0, 0, 0,
                        SWP_FRAMECHANGED|SWP_NOZORDER|SWP_NOACTIVATE|SWP_NOSIZE|SWP_NOMOVE);

            // Turn off XP Visual Styles
            //setDefaultWindowTheme(hBuddy1);

			/*
			 hBuddy3 = CreateWindowEx(  
                0,//gParams->ctl[n].buddy1StyleEx, 
                "msctls_updown32",				// predefined class 
                "",								// control text
                WS_VISIBLE | WS_CHILD | WS_TABSTOP | UDS_ALIGNRIGHT | UDS_ARROWKEYS,
                0,								// starting x position 
                0,								// starting y position 
                0,								// button width 
                0,								// button height 
                fmc.hDlg,                       // parent window 
                (HMENU)(INT_PTR)(IDC_BUDDY1+n), // control ID
                hDllInstance,
                NULL);                          // pointer not needed 

			SendMessage(hBuddy3, UDM_SETBUDDY, (WPARAM)gParams->ctl[n].hBuddy1, 0);
			*/

            //create the static text (label) buddy
            if (s & SBS_VERT || s & TBS_VERT) {
                //vertical scrollbar
                alignStyle = SS_CENTER;
            }
            else {
                //horizontal scrollbar
                alignStyle = SS_RIGHT;
            }
            alignStyle |= (s & WS_BORDER);  //OR in border style from main control
            gParams->ctl[n].sb.w2 = 0;      //dummy width
            gParams->ctl[n].sb.h2 = 0;      //dummy height
            computeBuddy2Pos(n, x, y, w, h, &r);

            hBuddy2 = CreateWindowEx(  
                gParams->ctl[n].buddy2StyleEx, //0,                              // extended styles
                "STATIC",                       // predefined class 
                "",                             // control text (set later...)
                WS_VISIBLE | WS_CHILD |
                SS_NOTIFY |                     // (So tooltips will work)
                gParams->ctl[n].buddy2Style | alignStyle,                     // styles
                r.left,                         // starting x position 
                r.top,                          // starting y position 
                r.right,                        // button width 
                r.bottom,                       // button height 
                fmc.hDlg,                       // parent window 
                (HMENU)(INT_PTR)(IDC_BUDDY2+n), // control ID
                hDllInstance,
                NULL);                          // pointer not needed 

            if (!hBuddy2) return FALSE;    //failed to create window.
            gParams->ctl[n].hBuddy2 = hBuddy2;

            // Turn off XP Visual Styles
            //setDefaultWindowTheme(hBuddy2);
			gParams->ctl[n].inuse = TRUE;
			fm_setCtlTheme(n, (int)gParams->ctl[n].theme);
			

            // Set control font to same as dialog font
            {
                HFONT hDlgFont;
                if ((hDlgFont = (HFONT)SendMessage(fmc.hDlg, WM_GETFONT, 0, 0L)) != NULL)
                {
                    gParams->ctl[n].hFont = hDlgFont;
                    SendMessage(hBuddy1, WM_SETFONT, (WPARAM)hDlgFont, 0L);  //redraw=FALSE
                    SendMessage(hBuddy2, WM_SETFONT, (WPARAM)hDlgFont, 0L);  //redraw=FALSE
                }
            }

            // Set the control text now, so the size of the static text control
            // can be computed correctly...
            if (gParams->ctl[n].label[0] != '\0') {
                gParams->ctl[n].inuse = TRUE;   //so setCtlText can work
                fm_setCtlText(n, (INT_PTR)fm_formatString(gParams->ctl[n].label));
            }
        }//scope
        break;

    case 33333333://CC_TRACKBAR:
        {
            //create a trackbar control
            //(and a NUMEDIT buddy.???)
            HWND hCtl;
            /* Convert dialog units to pixels */
            RECT r;

            // Should link dynamically, so only load comctl32.dll
            // if necessary!
            InitCommonControls();

            r.left = (int)x;
            r.top = (int)y;
            r.right = (int)w;
            r.bottom = (int)h;
            MapDialogRect(fmc.hDlg, &r);

            //Default tick frequency is 1.
            gParams->ctl[n].tb.ticFreq = 1;

            // force default styles (note TBS_HORZ==0 and TBS_BOTTOM==0).
            gParams->ctl[n].style |= WS_VISIBLE | WS_CHILD | WS_TABSTOP;

            // Transparent backgrounds don't work well for trackbars,
            // so change the default to opaque light gray.
            DeleteObject(gParams->ctl[n].hBrush); //superfluous for stock object
            gParams->ctl[n].hBrush = CreateSolidBrush(GetSysColor(COLOR_BTNFACE));//GetStockObject(LTGRAY_BRUSH);
            gParams->ctl[n].bkColor = GetSysColor(COLOR_BTNFACE);

            // create the trackbar control
            hCtl = CreateWindowEx(  
                gParams->ctl[n].styleEx,        // extended styles
                TRACKBAR_CLASS,                 // predefined class 
                "",                             // control text
                gParams->ctl[n].style,          // styles
                r.left,                         // starting x position 
                r.top,                          // starting y position 
                r.right,                        // control width 
                r.bottom,                       // control height 
                fmc.hDlg,                       // parent window 
                (HMENU)(INT_PTR)(IDC_CTLBASE+n),// control ID
                hDllInstance,
                NULL);                          // pointer not needed 

            if (!hCtl) return FALSE;    //failed to create window.
            gParams->ctl[n].hCtl = hCtl;

            // Turn off XP Visual Styles
            //setDefaultWindowTheme(hCtl);
			gParams->ctl[n].inuse = TRUE;
			fm_setCtlTheme(n, (int)gParams->ctl[n].theme);


            if (gParams->ctl[n].style & TBS_TOOLTIPS) {
                int side;
                //set side for automatic tooltips
                if (gParams->ctl[n].properties & CTP_BOTTOM) {
                    side = TBTS_BOTTOM;
                }
                else if (gParams->ctl[n].properties & CTP_RIGHT) {
                    side = TBTS_RIGHT;
                }
                else if (gParams->ctl[n].properties & CTP_LEFT) {
                    side = TBTS_LEFT;
                }
                else {
                    //default is TOP
                    side = TBTS_TOP;
                }
                SendMessage(gParams->ctl[n].hCtl, TBM_SETTIPSIDE, side, 0);
            }

#if 0
            // Set control font to same as dialog font
            {
                HFONT hDlgFont;
                if ((hDlgFont = (HFONT)SendMessage(fmc.hDlg, WM_GETFONT, 0, 0L)) != NULL)
                {
                    gParams->ctl[n].hFont = hDlgFont;
                    SendMessage(hCtl, WM_SETFONT, (WPARAM)hDlgFont, 0L);  //redraw=FALSE
                }
            }
#endif
        }
        break;

    case CC_PUSHBUTTON:
    case CC_CHECKBOX:
    case CC_RADIOBUTTON:
    case CC_GROUPBOX:
    case CC_OWNERDRAW:
        {
            //create a button control
            HWND hCtl;
            /* Convert dialog units to pixels */
            RECT r;

            r.left = (int)x;
            r.top = (int)y;
            r.right = (int)w;
            r.bottom = (int)h;
            MapDialogRect(fmc.hDlg, &r);

            // force default styles.
            gParams->ctl[n].style |= WS_VISIBLE | WS_CHILD;
            if (c != CC_RADIOBUTTON && c != CC_GROUPBOX) {
                //only first radio button in group should have TABSTOP.
                gParams->ctl[n].style |= WS_TABSTOP;
            }
#if 0
            if (c == CC_RADIOBUTTON && (gParams->ctl[n].style & WS_GROUP)) {
                //set TABSTOP on first radio button in GROUP???
                gParams->ctl[n].style |= WS_TABSTOP;
            }
#endif
#if 0
            if (c == CC_GROUPBOX) {
                //groupbox should always have GROUP style???
                gParams->ctl[n].style |= WS_GROUP;
            }
#endif
            gParams->ctl[n].maxval = 2;

            // create the button control
            hCtl = CreateWindowEx(  
                gParams->ctl[n].styleEx,        // extended styles
                "BUTTON",                       // predefined class 
                fm_formatString(gParams->ctl[n].label),  // control text
                gParams->ctl[n].style,          // styles
                r.left,                         // starting x position 
                r.top,                          // starting y position 
                r.right,                        // control width 
                r.bottom,                       // control height 
                fmc.hDlg,                       // parent window 
                (HMENU)(INT_PTR)(IDC_CTLBASE+n),// control ID
                hDllInstance,
                NULL);                          // pointer not needed 

            if (!hCtl) return FALSE;    //failed to create window.
            gParams->ctl[n].hCtl = hCtl;

            // Turn off XP Visual Styles
            //setDefaultWindowTheme(hCtl);
			gParams->ctl[n].inuse = TRUE;
			fm_setCtlTheme(n, (int)gParams->ctl[n].theme);

            // Set control font to same as dialog font
            {
                HFONT hDlgFont;
                if ((hDlgFont = (HFONT)SendMessage(fmc.hDlg, WM_GETFONT, 0, 0L)) != NULL)
                {
                    gParams->ctl[n].hFont = hDlgFont;
                    SendMessage(hCtl, WM_SETFONT, (WPARAM)hDlgFont, 0L);  //redraw=FALSE
                }
            }
        }
        break;

    case CC_IMAGE:
        gParams->ctl[n].im.hdcMem = 0;
        gParams->ctl[n].im.hdcAnd = 0;
        gParams->ctl[n].im.hbm = 0;
        gParams->ctl[n].im.hbmOld = 0;
        gParams->ctl[n].im.hbmOldAnd = 0;
        gParams->ctl[n].style |= SS_OWNERDRAW;  //force ownerdraw style
        //fall through...
    case CC_STATICTEXT:
    case CC_FRAME:
    case CC_RECT:
    case CC_BITMAP:
        //gParams->ctl[n].style |= SS_BITMAP;  //force style
    case CC_ICON:
        //gParams->ctl[n].style |= SS_ICON;  //force style
    case CC_METAFILE:
        //gParams->ctl[n].style |= SS_ENHMETAFILE;  //force style
        {
            //create a static control
            HWND hCtl;
            /* Convert dialog units to pixels */
            RECT r;

            r.left = (int)x;
            r.top = (int)y;
            r.right = (int)w;
            r.bottom = (int)h;
            MapDialogRect(fmc.hDlg, &r);

            // force default styles.
            gParams->ctl[n].style |= WS_VISIBLE | WS_CHILD;

            // create the static control
            hCtl = CreateWindowEx(  
                gParams->ctl[n].styleEx,        // extended styles
                "STATIC",                       // predefined class 
                fm_formatString(gParams->ctl[n].label),  // control text
                gParams->ctl[n].style,          // styles
                r.left,                         // starting x position 
                r.top,                          // starting y position 
                r.right,                        // control width 
                r.bottom,                       // control height 
                fmc.hDlg,                       // parent window 
                (HMENU)(INT_PTR)(IDC_CTLBASE+n),// control ID
                hDllInstance,
                NULL);                          // pointer not needed 

            if (!hCtl) return FALSE;    //failed to create window.
            gParams->ctl[n].hCtl = hCtl;

            // Turn off XP Visual Styles
            //setDefaultWindowTheme(hCtl);
			gParams->ctl[n].inuse = TRUE;
			fm_setCtlTheme(n, (int)gParams->ctl[n].theme);

            // Set control font to same as dialog font
            {
                HFONT hDlgFont;
                if ((hDlgFont = (HFONT)SendMessage(fmc.hDlg, WM_GETFONT, 0, 0L)) != NULL)
                {
                    gParams->ctl[n].hFont = hDlgFont;
                    SendMessage(hCtl, WM_SETFONT, (WPARAM)hDlgFont, 0L);  //redraw=FALSE
                }
            }
        }
        break;

    case CC_COMBOBOX:
        {
            //create a combo box control
            HWND hCtl;
            /* Convert dialog units to pixels */
            RECT r;

            r.left = (int)x;
            r.top = (int)y;
            r.right = (int)w;
            r.bottom = (int)h;
            MapDialogRect(fmc.hDlg, &r);

            // force default styles.
            gParams->ctl[n].style |= WS_VISIBLE | WS_CHILD | WS_TABSTOP;

            // Combo box-specific parameters
            gParams->ctl[n].val = -1;       //no item selected
            gParams->ctl[n].initVal = -1;   //no item selected
            gParams->ctl[n].minval = -1;
            gParams->ctl[n].maxval = 0xFFFF;    //max item index

            // create the combo box control
            hCtl = CreateWindowEx(  
                gParams->ctl[n].styleEx,        // extended styles
                "COMBOBOX",                     // predefined class 
                "",                             // control text
                gParams->ctl[n].style,          // styles
                r.left,                         // starting x position 
                r.top,                          // starting y position 
                r.right,                        // control width 
                r.bottom,                       // control height 
                fmc.hDlg,                       // parent window 
                (HMENU)(INT_PTR)(IDC_CTLBASE+n),// control ID
                hDllInstance,
                NULL);                          // pointer not needed 

            if (!hCtl) return FALSE;    //failed to create window.
            gParams->ctl[n].hCtl = hCtl;

            // Turn off XP Visual Styles
            //setDefaultWindowTheme(hCtl);
			gParams->ctl[n].inuse = TRUE;
			fm_setCtlTheme(n, (int)gParams->ctl[n].theme);


            if (gParams->ctl[n].properties & CTP_EXTENDEDUI) {
                // Set extended keyboard UI
                SendMessage(hCtl, CB_SETEXTENDEDUI, TRUE, 0);
            }

            // Set control font to same as dialog font
            {
                HFONT hDlgFont;
                if ((hDlgFont = (HFONT)SendMessage(fmc.hDlg, WM_GETFONT, 0, 0L)) != NULL)
                {
                    gParams->ctl[n].hFont = hDlgFont;
                    SendMessage(hCtl, WM_SETFONT, (WPARAM)hDlgFont, 0L);  //redraw=FALSE
                }
            }

            // Populate the list box from the control text string...
            gParams->ctl[n].inuse = TRUE;   //so setCtlText can work
            fm_setCtlText(n, (INT_PTR)gParams->ctl[n].label);
        }
        break;

    case CC_LISTBOX:
        {
            //create a list box control
            HWND hCtl;
            /* Convert dialog units to pixels */
            RECT r;

            r.left = (int)x;
            r.top = (int)y;
            r.right = (int)w;
            r.bottom = (int)h;
            MapDialogRect(fmc.hDlg, &r);

            // force default styles.
            gParams->ctl[n].style |= LBS_NOTIFY | WS_VISIBLE | WS_CHILD | WS_TABSTOP;

            // List box-specific parameters
            gParams->ctl[n].val = -1;       //no item selected
            gParams->ctl[n].initVal = -1;   //no item selected
            gParams->ctl[n].minval = -1;
            gParams->ctl[n].maxval = 0xFFFF;    //max item index

			DeleteObject(gParams->ctl[n].hBrush); //superfluous for stock object
			gParams->ctl[n].hBrush = CreateSolidBrush(GetSysColor(COLOR_BTNFACE));
			gParams->ctl[n].bkColor = GetSysColor(COLOR_BTNFACE);
			
            // create the list box control
            hCtl = CreateWindowEx(  
                gParams->ctl[n].styleEx,        // extended styles
                "LISTBOX",                      // predefined class 
                "",                             // control text
                gParams->ctl[n].style,          // styles
                r.left,                         // starting x position 
                r.top,                          // starting y position 
                r.right,                        // control width 
                r.bottom,                       // control height 
                fmc.hDlg,                       // parent window 
                (HMENU)(INT_PTR)(IDC_CTLBASE+n),// control ID
                hDllInstance,
                NULL);                          // pointer not needed 

            if (!hCtl) return FALSE;    //failed to create window.
            gParams->ctl[n].hCtl = hCtl;

            // Turn off XP Visual Styles
            //setDefaultWindowTheme(hCtl);
			gParams->ctl[n].inuse = TRUE;
			fm_setCtlTheme(n, (int)gParams->ctl[n].theme);

#if 0
            {//scope
                RECT rc;
                GetWindowRect(hCtl, &rc);
                Info("Set to (%d,%d,%d,%d)\n"
                     "Adjusted to (%d,%d,%d,%d)",
                     r.left, r.top, r.right, r.bottom,
                     rc.left, rc.top, rc.right, rc.bottom);
            }//scope
#endif

            // Set control font to same as dialog font
            {
                HFONT hDlgFont;
                if ((hDlgFont = (HFONT)SendMessage(fmc.hDlg, WM_GETFONT, 0, 0L)) != NULL)
                {
                    gParams->ctl[n].hFont = hDlgFont;
                    SendMessage(hCtl, WM_SETFONT, (WPARAM)hDlgFont, 0L);  //redraw=FALSE
                }
            }

            // Populate the list box from the control text string...
            gParams->ctl[n].inuse = TRUE;   //so setCtlText can work
            fm_setCtlText(n, (INT_PTR)gParams->ctl[n].label);
        }
        break;

    case CC_EDIT:
		{
            //create a edit control
            HWND hCtl;
            /* Convert dialog units to pixels */
            RECT r;

            r.left = (int)x;
            r.top = (int)y;
            r.right = (int)w;
            r.bottom = (int)h;
            MapDialogRect(fmc.hDlg, &r);

            // force default styles.
            gParams->ctl[n].style |= WS_VISIBLE | WS_CHILD | WS_TABSTOP;

            // create the static control
            hCtl = CreateWindowEx(  
                gParams->ctl[n].styleEx,        // extended styles
                "EDIT",							// predefined class 
                "",//fm_formatString(gParams->ctl[n].label), // control text
                gParams->ctl[n].style,          // styles
                r.left,                         // starting x position 
                r.top,                          // starting y position 
                r.right,                        // control width 
                r.bottom,                       // control height 
                fmc.hDlg,                       // parent window 
                (HMENU)(INT_PTR)(IDC_CTLBASE+n),// control ID
                hDllInstance,
                NULL);                          // pointer not needed 

            if (!hCtl) return FALSE;    //failed to create window.
            gParams->ctl[n].hCtl = hCtl;

            // Turn off XP Visual Styles
            gParams->ctl[n].inuse = TRUE;
			fm_setCtlTheme(n, (int)gParams->ctl[n].theme);

            // Set control font to same as dialog font
            {
                HFONT hDlgFont;
                if ((hDlgFont = (HFONT)SendMessage(fmc.hDlg, WM_GETFONT, 0, 0L)) != NULL)
                {
                    gParams->ctl[n].hFont = hDlgFont;
                    SendMessage(hCtl, WM_SETFONT, (WPARAM)hDlgFont, 0L);  //redraw=FALSE
                }
            }
        }
        break;

	case CC_TAB:
		{
            //create a tab control
            HWND hCtl;
            /* Convert dialog units to pixels */
            RECT r;
			//TCITEM tabItem; 

			
			InitCommonControls(); 

            r.left = (int)x;
            r.top = (int)y;
            r.right = (int)w;
            r.bottom = (int)h;
            MapDialogRect(fmc.hDlg, &r);

            // force default styles.
            gParams->ctl[n].style |= WS_CHILD | WS_CLIPSIBLINGS | WS_VISIBLE;

            // create the static control
            hCtl = CreateWindowEx(  
                gParams->ctl[n].styleEx,        // extended styles
                WC_TABCONTROL,                  // predefined class 
                "",//fm_formatString(gParams->ctl[n].label), // control text
                gParams->ctl[n].style,          // styles
                r.left,                         // starting x position 
                r.top,                          // starting y position 
                r.right,                        // control width 
                r.bottom,                       // control height 
                fmc.hDlg,                       // parent window 
                (HMENU)(INT_PTR)(IDC_CTLBASE+n),// control ID
                hDllInstance,
                NULL);                          // pointer not needed 

            if (!hCtl) return FALSE;    //failed to create window.
            gParams->ctl[n].hCtl = hCtl;

			//TabCtrl_SetMinTabWidth (gParams->ctl[n].hCtl, 20);

            // Turn off XP Visual Styles
            gParams->ctl[n].inuse = TRUE;
			fm_setCtlTheme(n, (int)gParams->ctl[n].theme);

            // Set control font to same as dialog font
            {
                HFONT hDlgFont;
                if ((hDlgFont = (HFONT)SendMessage(fmc.hDlg, WM_GETFONT, 0, 0L)) != NULL)
                {
                    gParams->ctl[n].hFont = hDlgFont;
                    SendMessage(hCtl, WM_SETFONT, (WPARAM)hDlgFont, 0L);  //redraw=FALSE
                }
            }

			//Add tab button
			//tabItem.mask = TCIF_TEXT | TCIF_IMAGE; 
			//tabItem.iImage = -1; 
			//tabItem.pszText = "Tab1"; 

			//TabCtrl_InsertItem(hCtl, 0, &tabItem);

        }
        break;

    case CC_UNUSED:
    case CC_SPINNER:
    case CC_UPDOWN:
    case CC_PROGRESSBAR:
	case CC_TOOLTIP:
    case CC_ANIMATION:
    case CC_PREVIEW:
    default:
        break;
    } //switch

    gParams->ctl[n].inuse = TRUE;
    gParams->ctl[n].tabControl = -1;    //no tab control
    gParams->ctl[n].tabSheet = -1;      //no tab sheet
    fm_setCtlRange(n, gParams->ctl[n].minval, gParams->ctl[n].maxval);
    fm_setCtlLineSize(n, gParams->ctl[n].lineSize);
    fm_setCtlPageSize(n, gParams->ctl[n].pageSize);
    fm_setCtlVal(n, gParams->ctl[n].initVal);
    fm_enableCtl(n, e); // visible/enabled
	fm_setCtlAnchor (n,ANCHOR_RIGHT);

    // Set default tooltips for predefined controls...
    switch (n) {

#if 0
    //Don't need them in my FocalBlade Plugin
    case CTL_OK:
        fm_setCtlToolTip(n, (int)"Apply filter to main image", 0);
        break;
    case CTL_CANCEL:
        fm_setCtlToolTip(n, (int)"Exit without applying filter", 0);
        break;
#endif

#ifndef STANDALONE
    case CTL_EDIT:
        fm_setCtlToolTip(n,
                         (gParams->flags & AFH_EDITMODE) ?
                         (INT_PTR)"Close source edit window" :
                         (INT_PTR)"Edit filter source code",
                         0);
        break;
#endif
    case CTL_LOGO:
        fm_setCtlToolTip(n, (INT_PTR)"About this filter", 0);
        break;
    case CTL_PREVIEW:
    case CTL_PROGRESS:
    case CTL_HOST:
    default:
        break;
    }//switch n

    // update cached control values...
#if 0
    fmc.pre_ctl[n] = gParams->ctl[n].val;   //already done by setCtlVal()	//eh??? Double-check this!!!
#endif
    return TRUE;    //success
} /*fm_createCtl*/

/**********************************************************************/
/* initCtl(n)
/*
/*  Initializes control n to default values.
/*
/**********************************************************************/
static INT_PTR fm_initCtl(INT_PTR n)
{
    char szBuffer[1024];

    if (!IS_OK_CTL_NR(n)) {
        return FALSE;
    }

    sprintf(szBuffer, "Control &%Id:", n);
    fm_createCtl(n,
                 CC_STANDARD,
                 (INT_PTR)szBuffer,
#if 0
                 270 - (n/10 % 10)*5,   //x
                 10*(n % 10) + 5,       //y
                 90,                    //w
                 8,                     //h
#else
                 -1,                    //x default
                 -1,                    //y default
                 -1,                    //h default
                 -1,                    //w default
#endif
                 0,                     //styles
                 0,                     //extended styles
                 0,                     //properties
                 3);                    //visible/enabled
    return TRUE;
} /*fm_initCtl*/


/////// BUGNOTE:
/////// Should not use negative x, y coords as flags, since they
/////// may legitimately be negative; e.g., in a multi-monitor
/////// environment.  !!!!!!! ?????? !!!!!!

/**********************************************************************/
/*  setDialogPos(fAbs, x, y, w, h)
/*
/*  Sets the position and size of the dialog window.  If fAbs is true,
/*  x and y are absolute screen coordinates; otherwise, x and y are
/*  relative to the upper-left corner of the client area in the host
/*  application's main window.  If x and y are set to -1, the dialog
/*  window will be centered within the host client area or the working
/*  area of the entire screen, depending on whether fAbs is false or
/*  true, respectively.  Otherwise, if either x or y is negative, the
/*  position of the dialog will not be changed. If w or h is negative,
/*  the size of the dialog window will not be changed.
/*
/*  All measurements are in dialog box units (DBUs).
/*
/**********************************************************************/
INT_PTR fm_setDialogPos(INT_PTR fAbs, INT_PTR x, INT_PTR y, INT_PTR w, INT_PTR h)
{
    UINT flags = SWP_NOZORDER | SWP_NOACTIVATE;
    HWND hParent;
    int  nHeight;
    int  nWidth;
    RECT rcDialog;
    RECT rcParent;
    RECT workArea;
    int  xScreen;
    int  yScreen;

#if 0
    Info("setDialogPos(%Id, %Id, %Id, %Id, %Id)", fAbs, x, y, w, h);
#endif

    /* Convert dialog units to pixels */
    {
        RECT r;
        r.left = (int)x;
        r.top = (int)y;
        r.right = (int)w;
        r.bottom = (int)h;
        MapDialogRect(fmc.hDlg, &r);
        // Need to add SM_CYCAPTION, SM_CXDLGFRAME, etc., depending on dialog box style?
        // Or could add the difference between dialog window size and dialog client size,
        // so it automatically adjusts for caption/no caption, border style, etc.?
        if (x >= 0) x = r.left;
        if (y >= 0) y = r.top;
#if 1
        GetWindowRect(fmc.hDlg, &rcDialog);
        GetClientRect(fmc.hDlg, &rcParent);
        if (w >= 0) w = r.right + (rcDialog.right-rcDialog.left) - (rcParent.right-rcParent.left);
        if (h >= 0) h = r.bottom + (rcDialog.bottom-rcDialog.top) - (rcParent.bottom-rcParent.top);

#else
        if (w >= 0) w = r.right + 2*GetSystemMetrics(SM_CXDLGFRAME);
        if (h >= 0) h = r.bottom + 2*GetSystemMetrics(SM_CYDLGFRAME) + GetSystemMetrics(SM_CYCAPTION);
#endif
    }

    if (w < 0 || h < 0) {
        // don't resize the dialog
        flags |= SWP_NOSIZE;
        GetWindowRect(fmc.hDlg, &rcDialog);
        nWidth  = rcDialog.right  - rcDialog.left;
        nHeight = rcDialog.bottom - rcDialog.top;
    }
    else {
        nWidth = (int)w;
        nHeight = (int)h;
        //could optimize following by doing it only for stretched images
        //and vertical gradients...
        flags |= SWP_NOCOPYBITS;  //so background will be redrawn when changing size
    }

    if (x == -1 && y == -1) {
        // center the dialog on screen if fAbs (or no parent);
        // else center in client area of parent.
        if  (fAbs || !(hParent = GetParent(fmc.hDlg)))
            hParent = GetDesktopWindow();

        GetClientRect(hParent, &rcParent);
        ClientToScreen(hParent, (LPPOINT)&rcParent.left);  // point(left,  top)
        ClientToScreen(hParent, (LPPOINT)&rcParent.right); // point(right, bottom)

        
        //Get Screen area without the task bar 
        //-> the bottom of the dialog isn't hidden by the taskbar anymore
        SystemParametersInfo(SPI_GETWORKAREA, 0, &workArea, 0);

        x = max(rcParent.right - rcParent.left - nWidth, 0) / 2
            + rcParent.left;
        //xScreen = GetSystemMetrics(SM_CXSCREEN);
        xScreen = workArea.right - workArea.left;
        if  (x + nWidth > xScreen)
            x = max (0, xScreen - nWidth);

        y = max(rcParent.bottom - rcParent.top - nHeight, 0) / 3
            + rcParent.top;        
        //yScreen = GetSystemMetrics(SM_CYSCREEN);
        yScreen = workArea.bottom - workArea.top;
        if  (y + nHeight > yScreen)
            y = max(0 , yScreen - nHeight);
    }
    else if (x < 0 || y < 0) {
        // don't move the dialog
        flags |= SWP_NOMOVE;
    }
    else if (!fAbs) {
        // Compute coords relative to client area of parent...
        if  ( ! (hParent = GetParent(fmc.hDlg)))
            hParent = GetDesktopWindow();

        GetClientRect(hParent, &rcParent);
        ClientToScreen(hParent, (LPPOINT)&rcParent.left);  // point(left,  top)

        x += rcParent.left;
        y += rcParent.top;

    }

    return SetWindowPos(fmc.hDlg, NULL,
                        (int)x, (int)y, nWidth, nHeight,
                        flags);
} /*fm_setDialogPos*/


/**********************************************************************/
/* setDialogMaxMin()
/*  Set the minum and maximum size of the dialog
/*
/**********************************************************************/
static INT_PTR fm_setDialogMinMax(INT_PTR mintrackX, INT_PTR mintrackY, INT_PTR maxtrackX, INT_PTR maxtrackY) //int maxsizeX,int maxsizeY,int maxposX, int maxposY,
{

    //Specifies the maximized width (point.x) and the maximized height (point.y) of the window. 
    //MaxMinInfo.ptMaxSize.x = maxsizeX;
    //MaxMinInfo.ptMaxSize.y = maxsizeY;
    //Specifies the position of the left side of the maximized window (point.x) and the position of the top of the maximized window (point.y). 
    //MaxMinInfo.ptMaxPosition.x = maxposX; 
    //MaxMinInfo.ptMaxPosition.y = maxposY; 

	if (mintrackX<0) mintrackX=0;
	if (mintrackY<0) mintrackY=0;
	if (maxtrackX<0) maxtrackX=0;
	if (maxtrackY<0) maxtrackY=0;

    //Specifies the minimum tracking width (point.x) and the minimum tracking height (point.y) of the window. 
    MaxMinInfo.ptMinTrackSize.x = (int)fm_HDBUsToPixels(mintrackX);
    MaxMinInfo.ptMinTrackSize.y = (int)fm_VDBUsToPixels(mintrackY);
    //Specifies the maximum tracking width (point.x) and the maximum tracking height (point.y) of the window. 
    MaxMinInfo.ptMaxTrackSize.x = (int)fm_HDBUsToPixels(maxtrackX); 
    MaxMinInfo.ptMaxTrackSize.y = (int)fm_VDBUsToPixels(maxtrackY);

    return true;

}/*fm_getDialogWidth*/



INT_PTR fm_getDialogPos(INT_PTR w, INT_PTR t)
{

	RECT rc;
	INT_PTR val;

	if (t==0){ //Full window
		
		GetWindowRect(fmc.hDlg, &rc);

	} else { //Client area

		GetClientRect(fmc.hDlg, &rc);
	}

	switch (w){
		case 0: val = fm_PixelsToHDBUs(rc.left); break;
		case 1: val = fm_PixelsToVDBUs(rc.top); break;
		case 2: val = fm_PixelsToHDBUs(rc.right - rc.left); break;
		case 3: val = fm_PixelsToVDBUs(rc.bottom - rc.top); break;
	}

	return val;

}


/**********************************************************************/
/* getDialogWidth()
/*  Gets the width of the client area of the dialog.
/*
/**********************************************************************/
static INT_PTR fm_getDialogWidth(void)
{
    RECT rcParent;
      
    //Get Pixel values
    GetClientRect(fmc.hDlg, &rcParent);
    //Convert to DBU
    return fm_PixelsToHDBUs(rcParent.right);


}/*fm_getDialogWidth*/

/**********************************************************************/
/* getDialogHeight()
/*  Gets the height of the client area of the dialog.
/*
/**********************************************************************/
static INT_PTR fm_getDialogHeight(void)
{
    RECT rcParent;

    //Get Pixel values
    GetClientRect(fmc.hDlg, &rcParent);
    //Convert to DBU
    return fm_PixelsToVDBUs(rcParent.bottom);

}/*fm_getDialogHeight*/



/**********************************************************************/
/* setDialogStyle(TitleBar |...)
/*  Sets various styles of the dialog.
/*
/**********************************************************************/
static INT_PTR fm_setDialogStyle(INT_PTR flags)
{
#if 1
    //disallow MaxBox for now -- it allows dialog to be maximized by
    //double-clicking the titlebar (NoSysMenu must also be specified).
    //The problem is that if you exit while maximized, then re-invoke,
    //it comes back in normal mode but with max size, so can't then
    //"unmaximize" the dialog.  Yecch.
    flags &= ~WS_MAXIMIZEBOX;
#endif
    SetWindowLong(fmc.hDlg, GWL_STYLE,
        GetWindowLong(fmc.hDlg, GWL_STYLE) | (int)flags);
    //use SWP_FRAMECHANGED to force nonclient area update...
    return SetWindowPos(fmc.hDlg, NULL,
                        0, 0, 0, 0,
                        SWP_FRAMECHANGED|SWP_NOZORDER|SWP_NOACTIVATE|SWP_NOSIZE|SWP_NOMOVE);
} /*fm_setDialogStyle*/

/**********************************************************************/
/* clearDialogStyle(TitleBar |...)
/*  Clears various styles of the dialog.
/*
/**********************************************************************/
static INT_PTR fm_clearDialogStyle(INT_PTR flags)
{
    SetWindowLong(fmc.hDlg, GWL_STYLE,
        GetWindowLong(fmc.hDlg, GWL_STYLE) & ~(int)flags);
    //use SWP_FRAMECHANGED to force nonclient area update...
    return SetWindowPos(fmc.hDlg, NULL,
                        0, 0, 0, 0,
                        SWP_FRAMECHANGED|SWP_NOZORDER|SWP_NOACTIVATE|SWP_NOSIZE|SWP_NOMOVE);
} /*fm_clearDialogStyle*/

// !!!! Should recalc Edit-mode dialog size when TOOLWINDOW is set/cleared
// !!!! and when dialog is resized!!!!


/**********************************************************************/
/* setDialogStyleEx(ToolWindow |...)
/*  Sets various extended styles of the dialog.
/*
/**********************************************************************/
static INT_PTR fm_setDialogStyleEx(INT_PTR flags)
{
    SetWindowLong(fmc.hDlg, GWL_EXSTYLE,
        GetWindowLong(fmc.hDlg, GWL_EXSTYLE) | (int)flags);
    //use SWP_FRAMECHANGED to force nonclient area update...
    return SetWindowPos(fmc.hDlg, NULL,
                        0, 0, 0, 0,
                        SWP_FRAMECHANGED|SWP_NOZORDER|SWP_NOACTIVATE|SWP_NOSIZE|SWP_NOMOVE);
} /*fm_setDialogStyleEx*/

/**********************************************************************/
/* clearDialogStyleEx(ToolWindow |...)
/*  Clears various extended styles of the dialog.
/*
/**********************************************************************/
static INT_PTR fm_clearDialogStyleEx(INT_PTR flags)
{
    SetWindowLong(fmc.hDlg, GWL_EXSTYLE,
        GetWindowLong(fmc.hDlg, GWL_EXSTYLE) & ~(int)flags);
    //use SWP_FRAMECHANGED to force nonclient area update...
    return SetWindowPos(fmc.hDlg, NULL,
                        0, 0, 0, 0,
                        SWP_FRAMECHANGED|SWP_NOZORDER|SWP_NOACTIVATE|SWP_NOSIZE|SWP_NOMOVE);
} /*fm_clearDialogStyleEx*/

/**********************************************************************/
/* setDialogText("title")
/*  Sets the caption in the title bar.
/*
/**********************************************************************/
INT_PTR fm_setDialogText(INT_PTR title)
{
    strncpy(gParams->gDialogTitle, (char *)title, MAX_LABEL_SIZE+1);
    gParams->gDialogTitle[MAX_LABEL_SIZE] = '\0';   //ensure null-terminated
    return SetWindowText(fmc.hDlg, fm_formatString(gParams->gDialogTitle));
} /*fm_setDialogText*/

/**********************************************************************/
/* setDialogTextv("format", ...)
/*  Sets the caption in the title bar with printf-style formatting.
/*
/**********************************************************************/
INT_PTR fm_setDialogTextv(INT_PTR iFmt,...)
{
    char szBuffer[1024];
    va_list ap;

    va_start(ap, iFmt);
    _vsnprintf(szBuffer, sizeof(szBuffer), (char *)iFmt, ap);
    szBuffer[sizeof(szBuffer)-1] = 0;   //ensure NUL-terminated
    va_end(ap);
    return fm_setDialogText((INT_PTR)szBuffer);
} /*fm_setDialogTextv*/

/**********************************************************************/
/* setDialogGradient(color1,color2,direction)
/*  Fills the background of the client area of the dialog box with
/*  a vertical (direction=0) or horizontal (direction=1) linear gradient.
/*
/**********************************************************************/
static INT_PTR fm_setDialogGradient(INT_PTR color1, INT_PTR color2, INT_PTR direction)
{
    if (gParams->gDialogGradientColor1 == (COLORREF)color1 &&
        gParams->gDialogGradientColor2 == (COLORREF)color2 &&
        gParams->gDialogGradientDirection == (int)direction) {
        //quick exit if no change...
        return TRUE;
    }

    gParams->gDialogGradientColor1 = (COLORREF)color1;
    gParams->gDialogGradientColor2 = (COLORREF)color2;
    gParams->gDialogGradientDirection = (int)direction;
    /* Force the background to be redrawn */
#if 0
    InvalidateRect(fmc.hDlg,
                          NULL, /*entire client area */
                          TRUE); /*erase background*/
    return UpdateWindow(fmc.hDlg);   //force a WM_PAINT
#elif 1
    return RedrawWindow(
            fmc.hDlg,   // handle of window
            NULL,   // address of structure with update rectangle, NULL=>entire client area
            NULL,   // handle of update region
            RDW_ERASE|RDW_INVALIDATE|RDW_ERASENOW|RDW_UPDATENOW     // array of redraw flags
       );
#else
    //What if just send WM_ERASEBKGND???
    return SendMessage(fmc.hDlg, WM_ERASEBKGND, 0, 0 );
#endif
} /*fm_setDialogGradient*/

/**********************************************************************/
/* setDialogColor(color)
/*  Sets the background color for the client area of the dialog box.
/*
/**********************************************************************/
static INT_PTR fm_setDialogColor(INT_PTR color)
{
    return fm_setDialogGradient(color, color, 0);
} /*fm_setDialogColor*/

static INT_PTR fm_setDialogSysColor(INT_PTR con)
{
	DWORD color = GetSysColor((int)con);
    return fm_setDialogGradient(color, color, 0);
} /*fm_setDialogSysColor*/

/**********************************************************************/
/* setDialogImage(filename)
/*  Applies an image from the specified file as the background image
/*  for the client area of the dialog box, tiling or stretching it to
/*  fit if requested by setDialogImageMode.
/*
/**********************************************************************/
static INT_PTR fm_setDialogImage(INT_PTR filename)
{
    strncpy(gParams->gDialogImageFileName, (char *)filename, _MAX_PATH+1);
    gParams->gDialogImageFileName[_MAX_PATH] = '\0';   //ensure null-terminated
    gParams->gDialogImage_save = gDialogImage = gParams->gDialogImageFileName;
    /* Force the background to be redrawn */
    return RedrawWindow(
            fmc.hDlg,   // handle of window
            NULL,   // address of structure with update rectangle, NULL=>entire client area
            NULL,   // handle of update region
            RDW_ERASE|RDW_INVALIDATE|RDW_ERASENOW|RDW_UPDATENOW     // array of redraw flags
       );
} /*fm_setDialogImage*/

/**********************************************************************/
/* setDialogImageMode({EXACT==0|TILED==1|STRETCHED==2}, <stretch_mode>)
/*  Specifies whether the background image is to the tiled or
/*  stretched to fill the entire dialog, or used exactly as is.
/*  If the image is to be stretched, then the stretch mode is
/*  also specified.
/*
/**********************************************************************/
static INT_PTR fm_setDialogImageMode(INT_PTR mode, INT_PTR stretchMode)
{
    gParams->gDialogImageMode = (int)mode;
    if (stretchMode != 0) gParams->gDialogImageStretchMode = (int)stretchMode;
    /* Force the background to be redrawn */
    return RedrawWindow(
            fmc.hDlg,   // handle of window
            NULL,   // address of structure with update rectangle, NULL=>entire client area
            NULL,   // handle of update region
            RDW_ERASE|RDW_INVALIDATE|RDW_ERASENOW|RDW_UPDATENOW     // array of redraw flags
       );
} /*fm_setDialogImageMode*/

/**********************************************************************/
/* setDialogDragMode({TitleBar==0|Background==1|None==2})
/*  Determines whether the filter dialog box can be dragged by the
/*  title bar only (0), by dragging the title bar or anywhere on the
/*  dialog background (1), or not at all (2).  The default mode is 0.
/**********************************************************************/
static INT_PTR fm_setDialogDragMode(INT_PTR mode)
{
    gParams->gDialogDragMode = (int)mode;
    return TRUE;
} /*fm_setDialogDragMode*/

/**********************************************************************/
/* setDialogRegion(<region_expression>)
/*  Sets the clipping region (outline) of the filter dialog.
/*
/**********************************************************************/
static INT_PTR fm_setDialogRegion(INT_PTR rgn)
{
    //gParams->gDialogRegion = (HRGN)rgn; //save across invocations
    return SetWindowRgn(
                fmc.hDlg,   // handle to window whose window region is to be set
                (HRGN)rgn,  // handle to region 
                TRUE		// window redraw flag 
                       );
    //delete the component regions with DeleteObj()???
} /*fm_setDialogRegion*/

/**********************************************************************/
/* REGION createRectRgn(xLeft, yTop, xRight, yBottom)
/*  Creates a simple rectangular region, with coordinates given in DBUs
/*  relative to the upper-left corner of the dialog box.  <left> is the
/*  x-coordinate of the left edge of the rectangle; <right> is the
/*  x-coordinate of the right edge.  <top> is the y-coordinate of the
/*  top edge of the rectangle; <bottom> is the y-coordinate of the
/*  bottom edge.  The <left> and <top> coordinates are inclusive (i.e.,
/*  pixels at these coordinates are included in the rectangle).  The
/*  <right> and <bottom> coordinates are exclusive (i.e., pixels up to
/*  but *not* including these coordinates are part of the rectangle).
/*
/**********************************************************************/
static INT_PTR fm_createRectRgn(INT_PTR xLeft, INT_PTR yTop, INT_PTR xRight, INT_PTR yBottom)
{
	int x1 = (int)xLeft;
	int y1 = (int)yTop;
	int x2 = (int)xRight;
	int y2 = (int)yBottom;

    Map4DBUsToPixels(&x1, &y1, &x2, &y2);
    return (INT_PTR)CreateRectRgn(x1, y1, x2, y2);
} /*fm_createRectRgn*/

/**********************************************************************/
/* REGION createRoundRectRgn(xLeft, yTop, xRight, yBottom, eWidth, eHeight)
/*  Creates a rectangular region with rounded corners.  <left>, <top>,
/*  <right>, and <bottom> are the same as for 'RECT' and describe the
/*  unrounded rectangle.  <width> and <height> specify the width and
/*  height of an ellipse; quarter sections of this ellipse are used to
/*  form the rounded corners of the rectangle.  All measurements are
/*  in DBUs.
/*
/**********************************************************************/
static INT_PTR fm_createRoundRectRgn(INT_PTR xLeft, INT_PTR yTop,
									 INT_PTR xRight, INT_PTR yBottom,
                                     INT_PTR eWidth, INT_PTR eHeight)
{
	int x1 = (int)xLeft;
	int y1 = (int)yTop;
	int x2 = (int)xRight;
	int y2 = (int)yBottom;
	int w1 = (int)eWidth;
	int h1 = (int)eHeight;

    Map4DBUsToPixels(&x1, &y1, &x2, &y2);
    Map2DBUsToPixels(&w1, &h1);
    return (INT_PTR)CreateRoundRectRgn(x1, y1, x2, y2, w1, h1);
} /*fm_createRoundRectRgn*/

/**********************************************************************/
/* REGION createCircularRgn(xLeft, yTop, iDiameter)
/*  Creates a simple circular region.  <left> and <top> give the x- and
/*  y-coordinates of the upper-left corner of the bounding box for the
/*  circle.  <diameter> gives the diameter of the circle.  All measurements
/*  are in DBUs.
/*
/**********************************************************************/
static INT_PTR fm_createCircularRgn(INT_PTR xLeft, INT_PTR yTop, INT_PTR iDiameter)
{
	int x1 = (int)xLeft;
	int y1 = (int)yTop;
    int x2 = (int)(xLeft + iDiameter);
    int y2 = (int)(yTop + iDiameter);
    Map4DBUsToPixels(&x1, &y1, &x2, &y2);
    return (INT_PTR)CreateEllipticRgn(x1, y1, x2, y2);
} /*fm_createCircularRgn*/

/**********************************************************************/
/* REGION createEllipticRgn(xLeft, yTop, xRight, yBottom)
/*  Creates an elliptical region.  <left>, <top>, <right>, and <bottom>
/*  define the coordinates of the bounding box for the ellipse.  All
/*  measurements are in DBUs.
/*
/**********************************************************************/
static INT_PTR fm_createEllipticRgn(INT_PTR xLeft, INT_PTR yTop, INT_PTR xRight, INT_PTR yBottom)
{
	int x1 = (int)xLeft;
	int y1 = (int)yTop;
	int x2 = (int)xRight;
	int y2 = (int)yBottom;

    Map4DBUsToPixels(&x1, &y1, &x2, &y2);
    return (INT_PTR)CreateEllipticRgn(x1, y1, x2, y2);
} /*fm_createEllipticRgn*/

/**********************************************************************/
/* REGION createPolyRgn(fillMode, nPoints, x1, y1, x2, y2, x3, y3, ...)
/*  Creates an arbitrary polygon from a list of pairs of vertices.  This
/*  list must consist of at least 3 pairs of x- and y-coordinates (in
/*  DBUs).  The total number of coordinates must be even.  The polygon is
/*  defined by drawing an imaginary line from the first vertex to the
/*  second, then to each subsequent vertex, and finally back to the
/*  original vertex.  The region is then defined by "filling" this polygon
/*  according to the fill mode.  The fill mode must be specified as either
/*  'ALTERNATE' (1) or 'WINDING' (2).
/*
/**********************************************************************/
static INT_PTR fm_createPolyRgn(INT_PTR fillMode, INT_PTR nPoints,
								INT_PTR x1, INT_PTR y1,
								INT_PTR x2, INT_PTR y2,
								INT_PTR x3, INT_PTR y3, ...)
{
#define stride (sizeof(INT_PTR)/2)	//2 for Win32, 4 for Win64

    int i;

    //Map 1st three mandatory points to pixels.
	//(In Win64, this forces x1 and y1 to be homed.
	//Note that x2 and beyond are always homed on
	//the call stack.)
    Map4DBUsToPixels((int *)&x1, (int *)&y1, (int *)&x2, (int *)&y2);
    Map2DBUsToPixels((int *)&x3, (int *)&y3);

    //Map any remaining optional points.
    //Following is non-portable!!!! (Depends on the ABI calling
    //convention to push args right-to-left as contiguous INT_PTRs
    //on the stack. Could use the VARARGS mechanism to do this
	//more portably.)
    for (i = 3; i < nPoints; i++) {
        Map2DBUsToPixels(&((int *)&x1)[i*stride], &((int *)&y1)[i*stride]);
    }

    //Following is non-portable!!!!  (Uses caller's argument
	//stack space to build an array of contiguous points.)

	//For Win64, we need to pack the 32-bit x- and y-coordinates
	//together on the stack to make an array of *adjacent* 32-bit
	//coordinates, the way CreatePolygonRgn() expects them. 
	//(For Win32, these int32 args on the stack are already
	//tightly packed, since each 32-bit arg fills exactly
	//one whole 32-bit stack slot. But for Win64, the 32-bit args
	//have 32-bit gaps between them, since each stack slot is
	//64 bits wide.)  Note that according to the x64 ABI, the
	//callee "owns" the argument space on the stack created by
	//the caller, so the callee is free to do what it wants with
	//this space. (A more portable implementation would use the
	//VARARGS mechanism to access the points, and would pack them
	//into a dynamically allocated array of size nPoints*2*sizeof(int)
	//bytes, but that consumes more memory and cpu resources than
	//the current implementation.)
#ifdef _WIN64
	for (i = 1; i < nPoints*2; i++) {	//index 0 is a no-op, start at i=1
		((int *)&x1)[i] = ((int *)&x1)[i*2];
	}
#endif

    return (INT_PTR)CreatePolygonRgn(
            //Following is non-portable!!!!  (Uses caller's argument
			//stack space to build an array of contiguous points.)
            (CONST POINT *)&x1, // pointer to array of points 
            (int)nPoints,    // number of points in array 
            (int)fillMode    // polygon-filling mode 
           );
#undef stride
} /*fm_createPolyRgn*/  

/**********************************************************************/
/* REGION createDialogRgn();
/*  Is a predefined rectangular region consisting of the default dialog
/*  box itself.
/*
/**********************************************************************/
static INT_PTR fm_createDialogRgn(void)
{
    // Should emit a warning or error !!!!
	ErrorOk("Function " __FUNCTION__ " is not yet implemented.");
    return 0;   //for now...
} /*fm_createDialogRgn*/

/**********************************************************************/
/* REGION createTitlebarRgn();
/*  Is a predefined rectangular region defining the area of the dialog
/*  title bar (if present).
/*
/**********************************************************************/
static INT_PTR fm_createTitlebarRgn(void)
{
    // Should emit a warning or error !!!!
	ErrorOk("Function " __FUNCTION__ " is not yet implemented.");
    return 0;   //for now...
} /*fm_createTitlebarRgn*/

/**********************************************************************/
/* REGION createMenubarRgn();
/*  Is a predefined rectangular region defining the area of the dialog
/*  menu bar (if any).
/*
/**********************************************************************/
static INT_PTR fm_createMenubarRgn(void)
{
    // Should emit a warning or error !!!!
	ErrorOk("Function " __FUNCTION__ " is not yet implemented.");
    return 0;   //for now...
} /*fm_createMenubarRgn*/

/**********************************************************************/
/* REGION createClientRgn();
/*  Is a predefined rectanglar region defining the client area of the
/*  dialog box.
/*
/**********************************************************************/
static INT_PTR fm_createClientRgn(void)
{
    // Should emit a warning or error !!!!
	ErrorOk("Function " __FUNCTION__ " is not yet implemented.");
    return 0;   //for now...
} /*fm_createClientRgn*/

/**********************************************************************/
/* REGION createStatusbarRgn();
/*  Is a predefined rectangular region defining the area of the dialog
/*  status bar (if any).
/*
/**********************************************************************/
static INT_PTR fm_createStatusbarRgn(void)
{
    // Should emit a warning or error !!!!
	ErrorOk("Function " __FUNCTION__ " is not yet implemented.");
    return 0;   //for now...
} /*fm_createStatusbarRgn*/

/**********************************************************************/
/* BOOL playSoundWave(szWaveFile);
/*  Plays the specified wave file asynchronously.  Looks first for
/*  an embedded WAVE resource with the specified name; else searches
/*  the usual paths for the specified wave file.
/*
/**********************************************************************/
INT_PTR fm_playSoundWave(INT_PTR szWaveFile)
{
    if (PlaySound((const char *)szWaveFile, hDllInstance,
                  SND_ASYNC | SND_RESOURCE | SND_NODEFAULT)) {
        //played from a WAVE resource
        return TRUE;
    }
    else if (szWaveFile != 0) {
        //try to locate the original wave file...
        return PlaySound(FmSearchFile((const char *)szWaveFile), NULL,
                         SND_ASYNC | SND_FILENAME);
    }
    return TRUE;
} /*fm_playSoundWave*/

/**********************************************************************/
/* BOOL playSoundWaveLoop(szWaveFile);
/*  Plays the specified wave file repeatedly.  Looks first for
/*  an embedded WAVE resource with the specified name; else searches
/*  the usual paths for the specified wave file.
/*
/**********************************************************************/
INT_PTR fm_playSoundWaveLoop(INT_PTR szWaveFile)
{
    if (PlaySound((const char *)szWaveFile, hDllInstance,
        SND_LOOP | SND_ASYNC | SND_RESOURCE | SND_NODEFAULT)) {
        //played from a WAVE resource
        return TRUE;
    }
    else if (szWaveFile != 0) {
        //try to locate the original wave file...
        return PlaySound(FmSearchFile((const char *)szWaveFile), NULL,
                         SND_LOOP | SND_ASYNC | SND_FILENAME);
    }
    return TRUE;
} /*fm_playSoundWaveLoop*/

/**********************************************************************/
/* BOOL playSoundWaveSync(szWaveFile);
/*  Plays the specified wave file synchronously.  Looks first for
/*  an embedded WAVE resource with the specified name; else searches
/*  the usual paths for the specified wave file.
/*
/**********************************************************************/
INT_PTR fm_playSoundWaveSync(INT_PTR szWaveFile)
{
    if (PlaySound((const char *)szWaveFile, hDllInstance,
                  SND_SYNC | SND_RESOURCE | SND_NODEFAULT)) {
        //played from a WAVE resource
        return TRUE;
    }
    else if (szWaveFile != 0) {
        //try to locate the original wave file...
        return PlaySound(FmSearchFile((const char *)szWaveFile), NULL,
                         SND_SYNC | SND_FILENAME);
    }
    return TRUE;
} /*fm_playSoundWaveSync*/

static INT_PTR fm_sleep(INT_PTR msecs)
{
    // interface to the __stdcall Sleep service
    Sleep((DWORD)msecs);
    return 0;
} /*fm_sleep*/

INT_PTR fm_RGB(INT_PTR r, INT_PTR g, INT_PTR b) {
    return ((b & 0xff) << 8 | (g & 0xff)) << 8 | (r & 0xff);
} /*fm_RGB*/

INT_PTR fm_RGBA(INT_PTR r, INT_PTR g, INT_PTR b, INT_PTR a) {
    return (((a & 0xff) << 8 | (b & 0xff)) << 8 | (g & 0xff)) << 8 | (r & 0xff);
} /*fm_RGBA*/

INT_PTR fm_Rval(INT_PTR rgba) {
    return GetRValue(rgba);
} /*fm_Rval*/

INT_PTR fm_Gval(INT_PTR rgba) {
    return GetGValue(rgba);
} /*fm_Gval*/

INT_PTR fm_Bval(INT_PTR rgba) {
    return GetBValue(rgba);
} /*fm_Bval*/

INT_PTR fm_Aval(INT_PTR rgba) {
    return (BYTE)(rgba >> 24);
} /*fm_Aval*/

/**********************************************************************/
/* iError = getOpenFileName( flags,
/*                     lpstrFile, nMaxFile,
/*                     [&nFIleOffset], [&nFileExtension],
/*                     [lpstrFileTitle], nMaxFileTitle,
/*                     [lpstrFilter], 
/*                     [lpstrCustomFilter], nMaxCustFilter,
/*                     [&nFilterIndex],
/*                     [lpstrInitialDir],
/*                     [lpstrDialogTitle],
/*                     [lpstrDefExt],
/*                     [&oflags]
/*                   );
/*
/*  Invokes the Open file common dialog.
/*
/**********************************************************************/
int fm_getOpenFileName( int flags,
                     LPSTR lpstrFile, int nMaxFile,
                     int *pnFileOffset, int *pnFileExtension,
                     LPSTR lpstrFileTitle, int nMaxFileTitle,
                     LPCSTR lpstrFilter,
                     LPSTR lpstrCustomFilter, int nMaxCustFilter,
                     int *pnFilterIndex,
                     LPCSTR lpstrInitialDir,
                     LPCSTR lpstrDialogTitle,
                     LPCSTR lpstrDefExt,
                     int *pOflags
                   )
{
    OPENFILENAME ofn;
    BOOL ok;
    int iError;

    ofn.lStructSize       = sizeof(ofn);
    ofn.hwndOwner         = fmc.doingProxy?fmc.hDlg:NULL;   // handle of owner window
    ofn.hInstance         = NULL;
    ofn.lpstrFilter       = lpstrFilter;
    ofn.lpstrCustomFilter = lpstrCustomFilter;
    ofn.nMaxCustFilter    = nMaxCustFilter;
    ofn.nFilterIndex      = pnFilterIndex ? *pnFilterIndex : 1;
    ofn.nMaxFile          = nMaxFile;
    ofn.nMaxFileTitle     = nMaxFileTitle;
    ofn.lpstrInitialDir   = lpstrInitialDir;
    ofn.lpstrTitle        = lpstrDialogTitle;
    ofn.nFileOffset       = 0;
    ofn.nFileExtension    = 0;
    ofn.lpstrDefExt       = lpstrDefExt;
    ofn.lCustData         = 0;
    ofn.lpfnHook          = NULL;
    ofn.lpTemplateName    = NULL;
    ofn.lpstrFile         = lpstrFile;
    ofn.lpstrFileTitle    = lpstrFileTitle;
    ofn.Flags             = flags | OFN_EXPLORER;   //force Explorer-style dialog
    // Don't allow user to set following flags...
    ofn.Flags &= ~(OFN_ENABLEHOOK | OFN_ENABLETEMPLATE | OFN_ENABLETEMPLATEHANDLE);

    ok = GetOpenFileName(&ofn);

#if 1
    if (pOflags) *pOflags = ofn.Flags;
#else
    //screen out non-return bits?
    if (pOflags) {
        *pOflags = ofn.Flags & (OFN_EXTENSIONDIFFERENT | OFN_NOREADONLYRETURN | OFN_READONLY);
    }
#endif
    if (pnFileOffset) *pnFileOffset = ofn.nFileOffset;
    if (pnFileExtension) *pnFileExtension = ofn.nFileExtension;
    if (pnFilterIndex) *pnFilterIndex = ofn.nFilterIndex;
    if (ok) {
        iError = 0;   //success
    }
    else {
        //error...
        iError = CommDlgExtendedError();
        if (iError == 0) {
            //User closed or canceled the dialog box...
            iError = -1;
        }
    }
    return iError;
} //fm_getOpenFileName

/**********************************************************************/
/* iError = getSaveFileName( flags,
/*                     lpstrFile, nMaxFile,
/*                     [&nFIleOffset], [&nFileExtension],
/*                     [lpstrFileTitle], nMaxFileTitle,
/*                     [lpstrFilter], 
/*                     [lpstrCustomFilter], nMaxCustFilter,
/*                     [&nFilterIndex],
/*                     [lpstrInitialDir],
/*                     [lpstrDialogTitle],
/*                     [lpstrDefExt],
/*                     [&oflags]
/*                   );
/*
/*  Invokes the Save As... file common dialog.
/*
/**********************************************************************/
int fm_getSaveFileName( int flags,
                     LPSTR lpstrFile, int nMaxFile,
                     int *pnFileOffset, int *pnFileExtension,
                     LPSTR lpstrFileTitle, int nMaxFileTitle,
                     LPCSTR lpstrFilter,
                     LPSTR lpstrCustomFilter, int nMaxCustFilter,
                     int *pnFilterIndex,
                     LPCSTR lpstrInitialDir,
                     LPCSTR lpstrDialogTitle,
                     LPCSTR lpstrDefExt,
                     int *pOflags
                   )
{
    OPENFILENAME ofn;
    BOOL ok;
    int iError;

    ofn.lStructSize       = sizeof(ofn);
    ofn.hwndOwner         = fmc.doingProxy?fmc.hDlg:NULL;   // handle of owner window
    ofn.hInstance         = NULL;
    ofn.lpstrFilter       = lpstrFilter;
    ofn.lpstrCustomFilter = lpstrCustomFilter;
    ofn.nMaxCustFilter    = nMaxCustFilter;
    ofn.nFilterIndex      = pnFilterIndex ? *pnFilterIndex : 1;
    ofn.nMaxFile          = nMaxFile;
    ofn.nMaxFileTitle     = nMaxFileTitle;
    ofn.lpstrInitialDir   = lpstrInitialDir;
    ofn.lpstrTitle        = lpstrDialogTitle;
    ofn.nFileOffset       = 0;
    ofn.nFileExtension    = 0;
    ofn.lpstrDefExt       = lpstrDefExt;
    ofn.lCustData         = 0;
    ofn.lpfnHook          = NULL;
    ofn.lpTemplateName    = NULL;
    ofn.lpstrFile         = lpstrFile;
    ofn.lpstrFileTitle    = lpstrFileTitle;
    ofn.Flags             = flags | OFN_EXPLORER;   //force Explorer-style dialog
    // Don't allow user to set following flags...
    ofn.Flags &= ~(OFN_ENABLEHOOK | OFN_ENABLETEMPLATE | OFN_ENABLETEMPLATEHANDLE | OFN_ALLOWMULTISELECT);

    ok = GetSaveFileName(&ofn);

#if 1
    if (pOflags) *pOflags = ofn.Flags;
#else
    //screen out non-return bits?
    if (pOflags) {
        *pOflags = ofn.Flags & (OFN_EXTENSIONDIFFERENT | OFN_NOREADONLYRETURN | OFN_READONLY);
    }
#endif
    if (pnFileOffset) *pnFileOffset = ofn.nFileOffset;
    if (pnFileExtension) *pnFileExtension = ofn.nFileExtension;
    if (pnFilterIndex) *pnFilterIndex = ofn.nFilterIndex;
    if (ok) {
        iError = 0;   //success
    }
    else {
        //error...
        iError = CommDlgExtendedError();
        if (iError == 0) {
            //User closed or canceled the dialog box...
            iError = -1;
        }
    }
    return iError;
} //fm_getSaveFileName

/**********************************************************************/
/* chooseColor(initialColor, szPrompt, ...)
/*
/*  Invokes the host app's color picker to choose a color.
/*  Returns the chosen color, or -1 if an error occurred.
/*
/**********************************************************************/
INT_PTR fm_chooseColor(INT_PTR initialColor, const char *szPrompt, ...)
{
    ColorServicesInfo csi;
    OSErr err;
    Str255 pickerPrompt;
    char szBuffer[1024];
    va_list ap;

    /* Interpret and interpolate the prompt string */
    va_start(ap, szPrompt);
    _vsnprintf(szBuffer, sizeof(szBuffer), szPrompt, ap);
    szBuffer[sizeof(szBuffer)-1] = 0;   //ensure NUL-terminated
    va_end(ap);
    strncpy(szBuffer, fm_formatString(szBuffer), sizeof(szBuffer));
    szBuffer[sizeof(szBuffer)-1] = 0;   //ensure NUL-terminated
    // convert to Pascal (byte-counted) string, 255 chars max
    strncpy(&pickerPrompt[1], szBuffer, 255);
    pickerPrompt[0] = (BYTE)min(strlen(szBuffer), 255);

    csi.infoSize = sizeof(csi);
    csi.selector = plugIncolorServicesChooseColor;
    csi.sourceSpace = plugIncolorServicesRGBSpace;
    csi.resultSpace = plugIncolorServicesRGBSpace;
    csi.reservedSourceSpaceInfo = NULL;
    csi.reservedResultSpaceInfo = NULL;
    csi.reserved = NULL;
    csi.colorComponents[0] = (int16)fm_Rval(initialColor);
    csi.colorComponents[1] = (int16)fm_Gval(initialColor);
    csi.colorComponents[2] = (int16)fm_Bval(initialColor);
    csi.colorComponents[3] = 0;
    csi.selectorParameter.pickerPrompt = &pickerPrompt;

    if (gStuff->colorServices) {
        //call the host's color picker
        err = gStuff->colorServices(&csi);
        if (err == noErr) {
            return fm_RGB(csi.colorComponents[0], csi.colorComponents[1], csi.colorComponents[2]);
        }
        else {
#ifdef VERBOSE
            if (err != userCanceledErr) {
                ErrorOk("Color picker error: %d", err);
            }
#endif //VERBOSE
			
			if (err == userCanceledErr) {
				return -1;  //error occurred
			}
			
			//If another error occured, display our own color dialog...
        }
    }

	//Display our own color dialog if host does not do it
    {
        //No host color service. Call the Windows native color picker.

        //Added by Harald Heim, January 2002
        CHOOSECOLOR cc;                 // common dialog box structure 
        static COLORREF acrCustClr[16]; // array of custom colors  NOT THREAD-SAFE!! (Should it be?? or just MT lock?)
        //HWND hwnd ;                      // owner window
        //static DWORD rgbCurrent;        // initial color selection

        // Initialize CHOOSECOLOR 
        ZeroMemory(&cc, sizeof(cc));
        cc.lStructSize = sizeof(cc);
        cc.hwndOwner = fmc.hDlg;
        cc.lpCustColors = (LPDWORD) acrCustClr;
        cc.rgbResult = (DWORD) initialColor;
        cc.Flags = CC_FULLOPEN | CC_RGBINIT;
 
        if (ChooseColor(&cc))
            return cc.rgbResult;
        else 
            return -1;

    }
} //fm_chooseColor

/**********************************************************************/
/* setRegRoot(int hkey)
/*  Sets the current registry root key.
/*
/**********************************************************************/
int fm_setRegRoot(HKEY hkey)
{
    gParams->gRegRoot = hkey;
    return ERROR_SUCCESS;
} /*fm_setRegRoot*/

/**********************************************************************/
/* getRegRoot(int *hkey)
/*  Retrieves the current registry root key.
/*
/**********************************************************************/
int fm_getRegRoot(HKEY *phkey)
{
    *phkey = gParams->gRegRoot;
    return ERROR_SUCCESS;
} /*fm_getRegRoot*/

/**********************************************************************/
/* setRegPath(lpsz szPath[, varargs]...)
/*  Sets the current registry path, with printf-style formatting.
/*  Returns ERROR_INVALID_DATA if expanded path string is too long;
/*  else returns ERROR_SUCCESS.
/*
/**********************************************************************/
int fm_setRegPath(LPCSTR szPath,...)
{
    va_list ap;
    int cb;

    va_start(ap, szPath);
    cb = _vsnprintf(gParams->gRegPath, sizeof(gParams->gRegPath), szPath, ap);
    gParams->gRegPath[sizeof(gParams->gRegPath)-1] = 0;   //ensure NUL-terminated
    va_end(ap);
    return (cb < 0) ? ERROR_INVALID_DATA : ERROR_SUCCESS;
} /*fm_setRegPath*/

/**********************************************************************/
/* getRegPath(lpsz szPath, int maxPathLen)
/*  Retrieves the current (not yet interpolated) registry path.
/*  Returns ERROR_INVALID_DATA if caller's buffer is too small;
/*  else returns ERROR_SUCCESS.
/*
/**********************************************************************/
int fm_getRegPath(LPSTR szPath, int maxPathLen)
{
    strncpy(szPath, gParams->gRegPath, maxPathLen);
    if (szPath[maxPathLen-1]) {
        //caller's buffer was too short...
        szPath[maxPathLen-1] = 0;   //force NUL-terminated string;
        return ERROR_INVALID_DATA;  //return error status
    }
    return ERROR_SUCCESS;
} /*fm_getRegPath*/

/**********************************************************************/
/* putRegInt(int iValue, lpsz szValueName[, varargs]...)
/*  Stores an integer value into the registry, with printf-style formatting
/*  for the value name.
/*
/**********************************************************************/
int fm_putRegInt(int iValue, LPCSTR szValueName,...)
{
    va_list ap;
    char szName[1024];
    int cb;
    HKEY hkey;

    va_start(ap, szValueName);
    cb = _vsnprintf(szName, sizeof(szName), szValueName, ap);
    szName[sizeof(szName)-1] = 0;   //ensure NUL-terminated
    va_end(ap);
    if (cb < 0) {
        //oops... expanded string is too long
        return ERROR_INVALID_DATA;
    }
    hkey = RCK(gParams->gRegRoot, fm_formatString(gParams->gRegPath));
    return RSD(hkey, fm_formatString(szName), (DWORD)iValue);
} /*fm_putRegInt*/

/**********************************************************************/
/* getRegInt(int *iValue, lpsz szValueName[, varargs]...)
/*  Gets an integer value from the registry, with printf-style formatting
/*  for the value name.
/*
/**********************************************************************/
int fm_getRegInt(int *iValue, LPCSTR szValueName,...)
{
    va_list ap;
    char szName[1024];
    int cb;
    HKEY hkey;

    va_start(ap, szValueName);
    cb = _vsnprintf(szName, sizeof(szName), szValueName, ap);
    szName[sizeof(szName)-1] = 0;   //ensure NUL-terminated
    va_end(ap);
    if (cb < 0) {
        //oops... expanded string is too long
        return ERROR_INVALID_DATA;
    }
    hkey = ROK(gParams->gRegRoot, fm_formatString(gParams->gRegPath));
    return RQD(hkey, fm_formatString(szName), (DWORD *)iValue);
} /*fm_getRegInt*/

/**********************************************************************/
/* putRegString(lpsz szString, lpsz szValueName[, varargs]...)
/*  Stores a string value into the registry, with printf-style formatting
/*  for the value name.
/*
/**********************************************************************/
int fm_putRegString(LPCSTR szString, LPCSTR szValueName,...)
{
    va_list ap;
    char szName[1024];
    int cb;
    HKEY hkey;

    va_start(ap, szValueName);
    cb = _vsnprintf(szName, sizeof(szName), szValueName, ap);
    szName[sizeof(szName)-1] = 0;   //ensure NUL-terminated
    va_end(ap);
    if (cb < 0) {
        //oops... expanded string is too long
        return ERROR_INVALID_DATA;
    }
#ifdef REG_DATA_LIMIT
    //check for data value too long...
    if (strlen(szString) > REG_DATA_LIMIT) {
        //limit amount of data user is allowed to store in registry
        return ERROR_INVALID_DATA;
    }
#endif //REG_DATA_LIMIT
    hkey = RCK(gParams->gRegRoot, fm_formatString(gParams->gRegPath));
    return RSS(hkey, fm_formatString(szName), szString);
} /*fm_putRegString*/

/**********************************************************************/
/* getRegString(lpsz szString, int iMaxlen, lpsz szValueName[, varargs]...)
/*  Gets a string value from the registry, with printf-style formatting
/*  for the value name.
/*
/**********************************************************************/
int fm_getRegString(LPSTR szString, int iMaxlen, LPCSTR szValueName,...)
{
    va_list ap;
    char szName[1024];
    int cb;
    HKEY hkey;

    va_start(ap, szValueName);
    cb = _vsnprintf(szName, sizeof(szName), szValueName, ap);
    szName[sizeof(szName)-1] = 0;   //ensure NUL-terminated
    va_end(ap);
    if (cb < 0) {
        //oops... expanded string is too long
        return ERROR_INVALID_DATA;
    }
    hkey = ROK(gParams->gRegRoot, fm_formatString(gParams->gRegPath));
    return RQS(hkey, fm_formatString(szName), szString, iMaxlen);
} /*fm_getRegString*/

/**********************************************************************/
/* putRegData(void *data, int dataLen, lpsz szValueName[, varargs]...)
/*  Stores a set of binary data into the registry, with printf-style
/*  formatting for the value name.
/*
/**********************************************************************/
int fm_putRegData(const void *pData, int dataLen, LPCSTR szValueName,...)
{
    va_list ap;
    char szName[1024];
    int cb;
    HKEY hkey;

    va_start(ap, szValueName);
    cb = _vsnprintf(szName, sizeof(szName), szValueName, ap);
    szName[sizeof(szName)-1] = 0;   //ensure NUL-terminated
    va_end(ap);
    if (cb < 0) {
        //oops... expanded string is too long
        return ERROR_INVALID_DATA;
    }
#ifdef REG_DATA_LIMIT
    //check for data value too long...
    if (dataLen > REG_DATA_LIMIT) {
        //limit amount of data user is allowed to store in registry
        return ERROR_INVALID_DATA;
    }
#endif //REG_DATA_LIMIT
    hkey = RCK(gParams->gRegRoot, fm_formatString(gParams->gRegPath));
    return RSB(hkey, fm_formatString(szName), pData, dataLen);
} /*fm_putRegData*/

/**********************************************************************/
/* getRegData(void *data, int dataLen, lpsz szValueName[, varargs]...)
/*  Gets a set of binary data from the registry, with printf-style
/*  formatting for the value name.
/*
/**********************************************************************/
int fm_getRegData(void *pData, int dataLen, LPCSTR szValueName,...)
{
    va_list ap;
    char szName[1024];
    int cb;
    HKEY hkey;

    va_start(ap, szValueName);
    cb = _vsnprintf(szName, sizeof(szName), szValueName, ap);
    szName[sizeof(szName)-1] = 0;   //ensure NUL-terminated
    va_end(ap);
    if (cb < 0) {
        //oops... expanded string is too long
        return ERROR_INVALID_DATA;
    }
    hkey = ROK(gParams->gRegRoot, fm_formatString(gParams->gRegPath));
    return RQB(hkey, fm_formatString(szName), pData, dataLen);
} /*fm_getRegData*/

/**********************************************************************/
/* enumRegSubKey(int index, lpsz szSubKey, int maxSubKeyLen)
/*  Gets name of n-th subkey under current registry key.
/*
/**********************************************************************/
int fm_enumRegSubKey(int index, LPSTR szSubKey, int maxSubKeyLen)
{
    HKEY hkey;

    hkey = ROK(gParams->gRegRoot, fm_formatString(gParams->gRegPath));
    return REK(hkey, index, szSubKey, maxSubKeyLen);
} /*fm_enumRegSubKey*/

/**********************************************************************/
/* enumRegValue(int index, lpsz szValueName, int maxValueNameLen,
/*              int *iType, int *cbData)
/*  Gets name, type, and data size of n-th value under current registry
/*  key.
/*
/**********************************************************************/
int fm_enumRegValue(int index, LPSTR szValueName, int maxValueNameLen,
                    int *iType, int *cbData)
{
    HKEY hkey;

    hkey = ROK(gParams->gRegRoot, fm_formatString(gParams->gRegPath));
    return REV(hkey, index, szValueName, &maxValueNameLen, iType, NULL, cbData);
} /*fm_enumRegValue*/

/**********************************************************************/
/* deleteRegSubKey(lpsz szSubKey[, varargs]...)
/*  Deletes specified subkey (and all its values) below the current
/*  key.  The subkey name is specified with printf-style formatting.
/*  The specified subkey must not have sub-subkeys (at least on NT;
/*  Win95 may allow deletion of an entire key subtree?).
/*
/**********************************************************************/
int fm_deleteRegSubKey(LPCSTR szSubKey,...)
{
    va_list ap;
    char szName[1024];
    int cb;
    HKEY hkey;

    va_start(ap, szSubKey);
    cb = _vsnprintf(szName, sizeof(szName), szSubKey, ap);
    szName[sizeof(szName)-1] = 0;   //ensure NUL-terminated
    va_end(ap);
    if (cb < 0) {
        //oops... expanded string is too long
        return ERROR_INVALID_DATA;
    }
    hkey = RCK(gParams->gRegRoot, fm_formatString(gParams->gRegPath));
    return RDK(hkey, fm_formatString(szName));
} /*fm_deleteRegSubKey*/

/**********************************************************************/
/* deleteRegValue(lpsz szValueName[, varargs]...)
/*  Deletes specified value under the current key.  The value name is
/*  specified with printf-style formatting.
/*
/**********************************************************************/
int fm_deleteRegValue(LPCSTR szValueName,...)
{
    va_list ap;
    char szName[1024];
    int cb;
    HKEY hkey;

    va_start(ap, szValueName);
    cb = _vsnprintf(szName, sizeof(szName), szValueName, ap);
    szName[sizeof(szName)-1] = 0;   //ensure NUL-terminated
    va_end(ap);
    if (cb < 0) {
        //oops... expanded string is too long
        return ERROR_INVALID_DATA;
    }
    hkey = RCK(gParams->gRegRoot, fm_formatString(gParams->gRegPath));
    return RDV(hkey, fm_formatString(szName));
} /*fm_deleteRegValue*/


INT_PTR fm_shellExec(INT_PTR szVerb, INT_PTR szFilename, INT_PTR szParams, INT_PTR szDefDir)
{
    INT_PTR ret = (INT_PTR)ShellExecute(fmc.hDlg, (char *)szVerb,
												  (char *)szFilename,
												  (char *)szParams,
												  (char *)szDefDir,
												  SW_SHOWNORMAL);
    if (ret <= 32) {
#ifdef VERBOSE
        ErrorOk("shellExec failed: %d", ret);
#endif
        return ret;   //failed
    }
    return -1;    //success
} /*fm_shellExec*/

int msgBox(UINT uType, LPCTSTR lpCaption, LPCTSTR lpFmt,...)
{
    char szBuffer[1024];
    va_list ap;

    va_start(ap, lpFmt);
    _vsnprintf(szBuffer, sizeof(szBuffer), lpFmt, ap);
    szBuffer[sizeof(szBuffer)-1] = 0;   //ensure NUL-terminated
    va_end(ap);
    return MessageBox(
        fmc.doingProxy?fmc.hDlg:NULL,   // handle of owner window
        fm_formatString(szBuffer),      // interpolated text for message box
        lpCaption,                      // address of title of message box  
        uType                           // style of message box
        | MB_TASKMODAL                  // Force task modal style.
       );
} /*msgBox*/

int fm_printf(LPCTSTR lpFmt,...)
{
    char szBuffer[1024];
    va_list ap;
    int nchars;

    va_start(ap, lpFmt);
    nchars = _vsnprintf(szBuffer, sizeof(szBuffer), lpFmt, ap);
    szBuffer[sizeof(szBuffer)-1] = 0;   //ensure NUL-terminated
    va_end(ap);
    MessageBox(
        fmc.doingProxy?fmc.hDlg:NULL,   // handle of owner window
        fm_formatString(szBuffer),      // interpolated text for message box
        
        //Added by Harald Heim, January 2003
        "stdout",           // address of title of message box  
        //"FilterMeister stdout",           // address of title of message box  

        0                               // style of message box
        | MB_TASKMODAL                  // Force task modal style.
       );
    return nchars;
} /*fm_printf*/

int Info(LPCTSTR lpFmt,...)
{
    char szBuffer[1024];
    va_list ap;

    va_start(ap, lpFmt);
    _vsnprintf(szBuffer, sizeof(szBuffer), lpFmt, ap);
    szBuffer[sizeof(szBuffer)-1] = 0;   //ensure NUL-terminated
    va_end(ap);
    return MessageBox(
        fmc.doingProxy?fmc.hDlg:NULL,   // handle of owner window
        fm_formatString(szBuffer),      // interpolated text for message box
        
        //Added by Harald Heim, January 12, 2003
         fmc.filterTitleText,
        //"FilterMeister Info",         // address of title of message box  

        MB_ICONINFORMATION              // style of message box
        | MB_TASKMODAL                  // Force task modal style.
        //| MB_TOPMOST                    // Force topmost window.
        //| MB_SETFOREGROUND              // Force foreground window.
       );
} /*Info*/

int Warn(LPCTSTR lpFmt,...)
{
    char szBuffer[1024];   
	va_list ap;

	va_start(ap, lpFmt);
    _vsnprintf(szBuffer, sizeof(szBuffer), lpFmt, ap);
    szBuffer[sizeof(szBuffer)-1] = 0;   //ensure NUL-terminated
    va_end(ap);
    return MessageBox(
        fmc.doingProxy?fmc.hDlg:NULL,   // handle of owner window
        fm_formatString(szBuffer),      // interpolated text for message box
        
        //Added by Harald Heim, January 12, 2003
         "Warning",
        //"FilterMeister Warning",      // address of title of message box  
        MB_ICONWARNING|MB_OKCANCEL      // style of message box
        | MB_TASKMODAL                  // Force task modal style.
       );
} /*Warn*/

/// Should add WarnOk() here....

int Error(LPCTSTR lpFmt,...)
{
    char szBuffer[1024];
    va_list ap;

    va_start(ap, lpFmt);
    _vsnprintf(szBuffer, sizeof(szBuffer), lpFmt, ap);
    szBuffer[sizeof(szBuffer)-1] = 0;   //ensure NUL-terminated
    va_end(ap);
    return MessageBox(
        fmc.doingProxy?fmc.hDlg:NULL,   // handle of owner window
        fm_formatString(szBuffer),      // interpolated text for message box
        
        //Added by Harald Heim, January 12, 2003        
        "Error",
        //"FilterMeister Error",            // address of title of message box  
        
        MB_ICONERROR|MB_ABORTRETRYIGNORE // style of message box
        | MB_TASKMODAL                  // Force task modal style.
       );
} /*Error*/

int ErrorOk(LPCTSTR lpFmt,...)
{
    char szBuffer[1024];
    va_list ap;

    va_start(ap, lpFmt);
    _vsnprintf(szBuffer, sizeof(szBuffer), lpFmt, ap);
    szBuffer[sizeof(szBuffer)-1] = 0;   //ensure NUL-terminated
    va_end(ap);
    return MessageBox(
        fmc.doingProxy?fmc.hDlg:NULL,   // handle of owner window
        fm_formatString(szBuffer),      // interpolated text for message box
        
        //Added by Harald Heim, January 12, 2003
        "Error",
        //"FilterMeister Error",            // address of title of message box  

        MB_ICONERROR|MB_OK              // style of message box
        | MB_TASKMODAL                  // Force task modal style.
       );
} /*ErrorOk*/

int YesNo(LPCTSTR lpFmt,...)
{
    char szBuffer[1024];
    va_list ap;

    va_start(ap, lpFmt);
    _vsnprintf(szBuffer, sizeof(szBuffer), lpFmt, ap);
    szBuffer[sizeof(szBuffer)-1] = 0;   //ensure NUL-terminated
    va_end(ap);
    return MessageBox(
        fmc.doingProxy?fmc.hDlg:NULL,   // handle of owner window
        fm_formatString(szBuffer),      // interpolated text for message box
        
        //Added by Harald Heim, January 12, 2003
         fmc.filterTitleText,
        //"FilterMeister",              // address of title of message box  

        MB_ICONQUESTION|MB_YESNO        // style of message box
        | MB_TASKMODAL                  // Force task modal style.
       );
} /*YesNo*/

int YesNoCancel(LPCTSTR lpFmt,...)
{
    char szBuffer[1024];
    va_list ap;

    va_start(ap, lpFmt);
    _vsnprintf(szBuffer, sizeof(szBuffer), lpFmt, ap);
    szBuffer[sizeof(szBuffer)-1] = 0;   //ensure NUL-terminated
    va_end(ap);
    return MessageBox(
        fmc.doingProxy?fmc.hDlg:NULL,   // handle of owner window
        fm_formatString(szBuffer),      // interpolated text for message box
        
        //Added by Harald Heim, January 12, 2003
         fmc.filterTitleText,
        //"FilterMeister",              // address of title of message box  

        MB_ICONQUESTION|MB_YESNOCANCEL  // style of message box
        | MB_TASKMODAL                  // Force task modal style.
       );
} /*YesNoCancel*/


/*****************************************************************************/
void InitContextRecord(void)
{
    // Initialize static portion of the FM context record.
    // (Needs to be done only once per DLL invocation.)	<== NO LONGER TRUE???
    int i;

    // Following private C-RTL routine converts float on top
    // of FPU stack to long int in EAX, but is not declared
    // in any public header file (it's only referenced by
    // compiler-generated code).
#ifdef _WIN64
	/// no _ftol ???  X64 FIXIT!!!
    extern long _ftol(void);
    extern long _ftol2(void);		//available in XP??
    extern long _ftol2_sse(void);	//available in XP??
    extern double X64_fmod_fix(double x, double y); //fixup for x64 fmod bug
#else //Win32
    extern long _ftol(void);
    extern long _ftol2(void);		//available in XP??
    extern long _ftol2_sse(void);	//available in XP??
#endif

	// Following private C-RTL routine checks the stack upon
	// procedure prolog entry to determine whether the stack
	// needs to be extended to accommodate the number of bytes
	// to be allocated in the prolog.  The number of bytes to
	// be allocated is passed in the eax/rax register.  For X64,
	// chkstk returns with no registers altered except R10 and R11.
	// The caller can then safely subtract the number of bytes in
	// rax from the rsp stack pointer.  For x86, the number of
	// bytes is subtracted from esp *before* chkstk returns, so
	// the caller does *not* subtract eax from esp upon return
	// (only eax is altered by the x86 chkstk routine.)
#ifdef _WIN64
	extern INT_PTR __chkstk(void);	//two underscores for Win64
#else //Win32
	extern INT_PTR _chkstk(void);	//one underscore for Win32
#endif

    // Initialize the info in the "frozen" section of the FM
    // context record.  This section should never be altered.
    // Any additions to the FM context record must *follow*
    // the frozen section.

    // initialize fmc.undefined with a signaling NaN.
    *(__int64 *)&fmc.undefined = 0x7ff7ffffffffffff; /* largest positive double SNaN */
    fmc.FMC_major_version = FMC_MAJOR_VERSION;
    fmc.FMC_minor_version = FMC_MINOR_VERSION;
    fmc.FMC_struct_size = sizeof(fmc);				//X64???
#ifdef _WIN64
	fmc.FMC_compile_platform = 64;	//compile platform is Win64	//not correct for STANDALONE!!!
#else
	fmc.FMC_compile_platform = 32;	//compile platform is Win32	//not correct for STANDALONE!!!
#endif
#ifdef STANDALONE
	fmc.FMC_target = fmc.FMC_compile_platform;
#else
	fmc.FMC_target = X64GEN ? 64 : 32;	// Gets set again before each compile.
#endif

#ifndef TRIG_ON_FLY
    // Precompute the trig tables...
    for (i=0; i < 1024; i++)
        fmc.pre_sin[i] = (int)(sin( (twopi/1024.0)*(double)i ) * 512.0 + 0.0); // round??

    for (i=0; i < 512; i++) {
        if ( i == 256)
            fmc.pre_tan[i] = (int)(tan( (twopi/1024.0)*(-255.5) ) * 1024.0 + 0.0);    // per Jens???
        else
            fmc.pre_tan[i] = (int)(tan( (twopi/1024.0)*(double)i ) * 1024.0 + 0.0);   // round?? can be neg!!
    }
#endif

    // Initialize constants

    // NDP control word constants, used for various rounding modes
    fmc.CW_ROUND = 0x133F;  //IC=affine, RC=00(round), PC=11(64bits), all 6 exceptions masked
    fmc.CW_FLOOR = 0x173F;  //IC=affine, RC=01(floor), PC=11(64bits), all 6 exceptions masked
    fmc.CW_CEIL  = 0x1B3F;  //IC=affine, RC=10(ceil),  PC=11(64bits), all 6 exceptions masked
    fmc.CW_CHOP  = 0x1F3F;  //IC=affine, RC=11(chop),  PC=11(64bits), all 6 exceptions masked

    // Save current control word (not a constant, but presumably 027F for MSVC).
    fmc.current_cw = (int)fm_fstcw();

    // The following "magic" constant has been carefully tuned for
    // use in rounding down when computing an integer square root,
    // over most (if not all) the domain of sqr().  See the code
    // for OP_SQR in the FM compiler module.
    fmc.magic5 = 0.49999999999999;

    // Following approximation to 512.0/Pi is used to compute d...
    fmc.d512byPi = 256.0 * 0.63661977236758134308;

    // Following 32-bit floating-point constant 1.0F is used by
    // the in-line code generated for powi().  We specifically
    // want a 32-bit float, not a 64-bit double, because (at least
    // on the 387), FDIVR_F 1.0F is faster than FDIVR_D 1.0.
    fmc.float1p0 = 1.0F;

    // Initialize built-in strings...
    fmc.filterAuthorText[0] = 0;
    fmc.filterCategoryText[0] = 0;
    fmc.filterTitleText[0] = 0;
    fmc.filterFilenameText[0] = 0;
    fmc.filterCopyrightText[0] = 0;
    fmc.filterDescriptionText[0] = 0;
    fmc.filterOrganizationText[0] = 0;
    fmc.filterURLText[0] = 0;
    fmc.filterVersionText[0] = 0;
    fmc.filterCaseText[0] = 0;
    fmc.filterImageModeText[0] = 0;
    fmc.filterHostText[0] = 0;
    fmc.filterInstallDir[0] = 0;
    fmc.filterUniqueID[0] = 0;

#ifndef DONT_SAVE_GLOBALS
    //Zero out the predefined globals (harmless, but probably unnecessary) */
    memset(&fmc.globvars, 0, sizeof(fmc.globvars));
#endif //DONT_SAVE_GLOBALS

    // Initialize literal pool...
    fmc.ndlits = 0;

    // Precompute the gamma table for gamma = 1.0...
    fm_setGamma(1.0);

    // Initialize vectors to FM built-in routines...

    // Standard FF built-ins.
    if (gStuff->imageMode == plugInModeRGB48 ||
        gStuff->imageMode == plugInModeGray16 ||
        //Added by Harald Heim, December 6, 2002
        gStuff->imageMode == plugInModeLab48 || 
        gStuff->imageMode == plugInModeCMYK64 ||
        gStuff->imageMode == plugInModeDeepMultichannel ||
        gStuff->imageMode == plugInModeDuotone16
        ) {

        // pixel depth = 16
        fmc.src = fm_src16;
        fmc.pset = fm_pset16;
        //Added by Harald Heim Feb 9, 2002
        fmc.pget = fm_pget16;
        fmc.pgetr = fm_pgetr16;

		fmc.tget = fm_tget16;
	    fmc.t2get = fm_t2get16;
		fmc.t3get = fm_t3get16;
		fmc.t4get = fm_t4get16;
		fmc.tset = fm_tset16;
		fmc.t2set = fm_t2set16;
		fmc.t3set = fm_t3set16;
		fmc.t4set = fm_t4set16;
    }
    else {
        // pixel depth = 8
        fmc.src = fm_src;
        fmc.pset = fm_pset;
        //Added by Harald Heim Feb 9, 2002
        fmc.pget = fm_pget;
        fmc.pgetr = fm_pgetr;

		fmc.tget = fm_tget;
	    fmc.t2get = fm_t2get;
		fmc.t3get = fm_t3get;
		fmc.t4get = fm_t4get;
		fmc.tset = fm_tset;
		fmc.t2set = fm_t2set;
		fmc.t3set = fm_t3set;
		fmc.t4set = fm_t4set;
    }
    
    fmc.rad = fm_rad;
    fmc.ctl = fm_ctl;
    fmc.val = fm_val;
    fmc.map = fm_map;
    fmc.min = fm_min;
    fmc.max = fm_max;
    fmc.fmin = fm_fmin;
    fmc.fmax = fm_fmax;
    fmc.abs = fm_abs;
    fmc.add = fm_add;
    fmc.sub = fm_sub;
    fmc.dif = fm_dif;
    fmc.rnd = fm_rnd;
    fmc.rst = fm_rst;
    fmc.mix = fm_mix;
    fmc.mix1 = fm_mix1;
    fmc.mix2 = fm_mix2;
    fmc.scl = fm_scl;
    fmc.sqr = fm_sqr;
    fmc.sin = fm_sin;
    fmc.cos = fm_cos;
    fmc.tan = fm_tan;
    fmc.r2x = fm_r2x;
    fmc.r2y = fm_r2y;
    fmc.c2d = fm_c2d;
    fmc.c2m = fm_c2m;
    fmc.get = fm_get;
    fmc.put = fm_put;
    fmc.cnv = fm_cnv;

    // FF+ built-ins.

    //Added by Harald Heim, February 4, 2002
    fmc.pointer_to_buffer = fm_pointer_to_buffer;
    //Added by Harald Heim, February 5, 2002
    fmc.contrast = fm_contrast;
    fmc.saturation = fm_saturation;
    fmc.posterize = fm_posterize;
    fmc.phaseshift = fm_phaseshift;
    fmc.blend = fm_blend;
    //Added by Harald Heim, February 6, 2002
    fmc.rgb2iuv = fm_rgb2iuv;
    fmc.egw = fm_egw;
    fmc.gray = fm_gray;
    fmc.solarize = fm_solarize;
    fmc.msk = fm_msk;
    //Added by Harald Heim, February 7, 2002
    fmc.tricos = fm_tricos;
    fmc.tri = fm_tri;
    fmc.grad2D = fm_grad2D;
    //Added by Harald Heim, February 8, 2002
    fmc.sinbell = fm_sinbell;
    fmc.iuv2rgb = fm_iuv2rgb;
    fmc.tone = fm_tone;
    fmc.rgb2cmyk = fm_rgb2cmyk;
    fmc.cmyk2rgb = fm_cmyk2rgb;
    //Added by Harald Heim, February 10, 2002
    fmc.xyzcnv = fm_xyzcnv;
    fmc.cell_initialize = fm_cell_initialize;
    //Added by Harald Heim, February 13, 2002
    fmc.set_edge_mode = fm_set_edge_mode;
    fmc.set_bitdepth_mode = fm_set_bitdepth_mode;
    fmc.set_psetp_mode = fm_set_psetp_mode;
	fmc.set_array_mode = fm_set_array_mode;

    fmc.bCircle = fm_bCircle;
    fmc.bRect = fm_bRect;
    fmc.bRect2 = fm_bRect2;
    fmc.bTriangle = fm_bTriangle;
    //Added by Harald Heim, March 8, 2002
    fmc.iget = fm_iget;
    //Added by Harald Heim, March 20, 2002
    fmc.refreshWindow = fm_refreshWindow;
    //Added by Harald Heim, April 02, 2002
    fmc.setZoom = fm_setZoom;
	fmc.setZoomEx = fm_setZoomEx;
    //Added by Harald Heim, April 06, 2002
    fmc.lockWindow = fm_lockWindow;
    fmc.refreshCtl = fm_refreshCtl;
    fmc.refreshRgn = fm_refreshRgn;
    //fmc.lockCtl = fm_lockCtl;
    //Added by Harald Heim, April 09, 2002
    fmc.allocArray = fm_allocArray;
    fmc.allocArrayPad = fm_allocArrayPad;
    fmc.freeArray = fm_freeArray;
    fmc.getArray = fm_getArray;
	fmc.fgetArray = fm_fgetArray;
    fmc.putArray = fm_putArray;
	fmc.fputArray = fm_fputArray;
    fmc.UNSAFE_getArray = fm_UNSAFE_getArray;
    fmc.UNSAFE_putArray = fm_UNSAFE_putArray;
	fmc.UNSAFE_fgetArray = fm_UNSAFE_fgetArray;
	fmc.UNSAFE_fputArray = fm_UNSAFE_fputArray;

	//fmc.fputArrayLE = fm_fputArrayLE;
	//fmc.fgetArrayLE = fm_fgetArrayLE;


    fmc.getArrayString = (fmf2)fm_getArrayString;
    fmc.putArrayString = (fmf3)fm_putArrayString;

    //Added by Harald Heim, Jun 4, 2002
    fmc.calcSBD = fm_calcSBD;
    fmc.freeSBD = fm_freeSBD;
    fmc.getSBD = fm_getSBD;
    fmc.getSBDangle = fm_getSBDangle;
    fmc.getSBDX = fm_getSBDX;
    fmc.getSBDY = fm_getSBDY;
    //Added by Harald Heim, Jun 6, 2002
    fmc.mouseOverWhenInvisible = fm_mouseOverWhenInvisible;
    //Added by Harald Heim, Aug 9, 2002
    fmc.getPreviewCoordX = fm_getPreviewCoordX;
    fmc.getPreviewCoordY = fm_getPreviewCoordY;
    //Added by Harald Heim, Aug 12, 2002
    fmc.getAsyncKeyState = fm_getAsyncKeyState;
    fmc.getAsyncKeyStateF = fm_getAsyncKeyStateF;
    //Added by Harald Heim, Aug 14, 2002
    //fmc.threeDto2D_X = fm_threeDto2D_X;
    //fmc.threeDto2D_Y = fm_threeDto2D_Y;

    fmc.getArrayDim = fm_getArrayDim;
    fmc.copyArray = fm_copyArray;
    fmc.fillArray = fm_fillArray;
    fmc.ffillArray = fm_ffillArray;
    fmc.ctlEnabled = fm_ctlEnabled;
    fmc.ctlEnabledAs = fm_ctlEnabledAs;	//MWVDLEE 2008-08-19
    //Added by Harald Heim, Dec 7, 2002
    fmc.rgb2lab = fm_rgb2lab;
    fmc.lab2rgb = fm_lab2rgb;
    //Added by Harald Heim, Dec 13, 2002
    fmc.setClickDrag = fm_setClickDrag;
    fmc.setTimerEvent = fm_setTimerEvent;
    fmc.getDisplaySettings = fm_getDisplaySettings;
    //Added by Harald Heim, Dec 14, 2002
    fmc.scrollPreview = fm_scrollPreview;
    //Added by Harald Heim, Dec 23, 2002
    fmc.rgb2hsl = fm_rgb2hsl;
    fmc.hsl2rgb = fm_hsl2rgb;
    fmc.ycbcr2rgb = fm_ycbcr2rgb;
    fmc.rgb2ycbcr = fm_rgb2ycbcr;
	//fmc.rgb2ypbpr = fm_rgb2ypbpr;	//implemented but not yet tested and released
	//fmc.ypbpr2rgb = fm_ypbpr2rgb;	//implemented but not yet tested and released
    //Added by Harald Heim, January 2003
    fmc.srcp = fm_srcp;
    fmc.psetp = fm_psetp;
    fmc.pgetp = fm_pgetp;
    fmc.getWindowsVersion = fm_getWindowsVersion;
    fmc.setPreviewCursor = fm_setPreviewCursor;
	fmc.getPreviewCursor = fm_getPreviewCursor;
    fmc.tsetp = fm_tsetp;
    fmc.tgetp = fm_tgetp;
    //Added by Harald Heim, March 9, 2003
    fmc.t2setp = fm_t2setp;
	fmc.t3setp = fm_t3setp;
	fmc.t4setp = fm_t4setp;
    fmc.t2getp = fm_t2getp;
	fmc.t3getp = fm_t3getp;
	fmc.t4getp = fm_t4getp;
    //Added by Harald Heim, March 12, 2003
    fmc.deleteRgn = fm_deleteRgn;
    //Added by Harald Heim, May 1, 2003
    fmc.setCtlDivisor = fm_setCtlDivisor;   
	//Added by M.W. van der Lee, June 29, 2008
	fmc.getCtlDivisor = fm_getCtlDivisor;	
    fmc.setCtlGamma = fm_setCtlGamma;

    fmc.getCtlCoord = fm_getCtlCoord;
    fmc.getCtlRange = fm_getCtlRange;
    fmc.getCtlPos = fm_getCtlPos;

	fmc.fr2x = fm_fr2x;
	fmc.fr2y = fm_fr2y;
	fmc.fc2m = fm_fc2m;
	fmc.fc2d = fm_fc2d;
	fmc.setEndCaps = fm_setEndCaps;
	fmc.setJoin = fm_setJoin;
	fmc.setInsideFrame = fm_setInsideFrame;
	fmc.setAngleArcFill = fm_setAngleArcFill;
	fmc.setAngleArc = fm_setAngleArc;

	fmc.countProcessors = fm_countProcessors;



#ifdef THEPLUGINSITE
    fmc.igetArray = fm_igetArray;

#if COLORWASHER
    fmc.ct2R = fm_ct2R;
    fmc.ct2G = fm_ct2G;
    fmc.ct2B = fm_ct2B;

    fmc.rgb2ct = fm_rgb2ct;
#endif
    
    fmc.cnvX_edge = fm_cnvX_edge;
    fmc.cnvY_edge = fm_cnvY_edge;

    //fmc.quickGauss = fm_quickGauss;
    fmc.fastMotion = fm_fastMotion;
    fmc.fastBox = fm_fastBox;
    fmc.fastGauss = fm_fastGauss;
	fmc.fastPreciseGauss = fm_fastPreciseGauss;
    fmc.fastMotionGauss = fm_fastMotionGauss;
    fmc.copyImage2Array = fm_copyImage2Array;

	fmc.calcEdgeMask = fm_calcEdgeMask;
	fmc.stretchContrast = fm_stretchContrast;
	fmc.calcYCC = fm_calcYCC;

    fmc.floodFill = fm_floodFill;
    fmc.noise = fm_noise;

#if 0
    fmc.setPixelDB = fm_setPixelDB;
    fmc.startSetPixelDB = fm_startSetPixelDB;
    fmc.endSetPixelDB = fm_endSetPixelDB;
    fmc.drawTextDB = fm_drawTextDB;
    fmc.setFontDB = fm_setFontDB;
#endif

    #ifdef CONTRASTMASTER
    fmc.calcSpecialMinMax = fm_calcSpecialMinMax;
    fmc.calcSpecialMinMaxFast = fm_calcSpecialMinMaxFast;
    fmc.calcTheMix = fm_calcTheMix;
    fmc.calcCMFinal = fm_calcCMFinal;
    //fmc.dynaContra = fm_dynaContra;
        fmc.processACE = fm_processACE;
    #endif

	#ifdef HDRpro
		//fmc.processHDR = fm_processHDR;
		//fmc.processHDR2 = fm_processHDR2;
	#endif

#endif

#ifdef IMGSOURCE
    fmc.openImage = fm_openImage;
    fmc.drawArray = fm_drawArray;
    fmc.applyFX = fm_applyFX;
#endif

#ifdef WATERMARK
    fmc.watermark = fm_watermark;
#endif


	fmc.copyResToArray = fm_copyResToArray;
	fmc.getResAddress = fm_getResAddress;
	fmc.getResSize = fm_getResSize;

    fmc.quickSort = fm_quickSort;
    fmc.quickMedian = fm_quickMedian;
    fmc.quickFill = fm_quickFill;
	fmc.quickFillArray = fm_quickFillArray;


    fmc.getDialogWidth= fm_getDialogWidth;
    fmc.getDialogHeight= fm_getDialogHeight;
    fmc.setDialogShowState = fm_setDialogShowState;
    fmc.PixelsToHDBUs = fm_PixelsToHDBUs;
    fmc.PixelsToVDBUs = fm_PixelsToVDBUs;
    fmc.setDialogMinMax = fm_setDialogMinMax;
    fmc.setDialogSizeGrip = fm_setDialogSizeGrip;

    fmc.startSetPixel = fm_startSetPixel;
    fmc.endSetPixel = fm_endSetPixel;
	fmc.setPixel = fm_setPixel;
	fmc.getPixel = fm_getPixel;
	fmc.setBitmap = fm_setBitmap;
	fmc.setBitmapStretch = fm_setBitmapStretch;
	fmc.setBitmapTransparent = fm_setBitmapTransparent;
	fmc.setBitmapTile = fm_setBitmapTile;
	fmc.setBitmapStretchTransparent = fm_setBitmapStretchTransparent;
	fmc.setText = fm_setText;
	fmc.setTextv = fm_setTextv;
    fmc.setFont = fm_setFont;
	fmc.setLine = fm_setLine;
	fmc.setEllipse = fm_setEllipse;
	fmc.setEllipseFill = fm_setEllipseFill;
	fmc.setRectFrame = fm_setRectFrame;
	fmc.setRectangle = fm_setRectangle;
	fmc.setFill = fm_setFill;
	fmc.setRectFill = fm_setRectFill;
	fmc.setRectGradient = fm_setRectGradient;
	fmc.setPenWidth = fm_setPenWidth;
	fmc.startSetPixelSS = fm_startSetPixelSS;
	fmc.startSetPixelBuffer = fm_startSetPixelBuffer;
	fmc.endSetPixelBuffer = fm_endSetPixelBuffer;
	fmc.getSetPixelWidth = fm_getSetPixelWidth;
	fmc.getSetPixelHeight = fm_getSetPixelHeight;
	fmc.setThinLineAA = fm_setThinLineAA;
	fmc.setThinEllipseAA = fm_setThinEllipseAA;
	fmc.blendRGBA = fm_blendRGBA;

	fmc.triggerThread = fm_triggerThread;
	fmc.waitForThread = fm_waitForThread;
	fmc.isThreadActive = fm_isThreadActive;
	fmc.getThreadRetVal = fm_getThreadRetVal;
	fmc.terminateThread = fm_terminateThread;

    fmc.createSync = fm_createSync;
    fmc.waitForSync = fm_waitForSync;
    fmc.deleteSync = fm_deleteSync;

    fmc.createCriticalSection = fm_createCriticalSection;
#ifdef HAVE_XP
    fmc.createCriticalSectionAndSpinCount = fm_createCriticalSectionSpinCount;
    fmc.setCriticalSectionSpinCount = fm_setCriticalSectionSpinCount;
#endif
#ifdef HAVE_VISTA
    fmc.createCriticalSectionEx = fm_createCriticalSectionEx;
#endif
    fmc.enterCriticalSection = fm_enterCriticalSection;
#ifdef HAVE_XP
    fmc.tryEnterCriticalSection = fm_tryEnterCriticalSection;
#endif
    fmc.leaveCriticalSection = fm_leaveCriticalSection;
    fmc.deleteCriticalSection = fm_deleteCriticalSection;

    fmc.interlockedIncrement = fm_interlockedIncrement;
    fmc.interlockedDecrement = fm_interlockedDecrement;
    fmc.interlockedExchange = fm_interlockedExchange;
    fmc.interlockedExchangeAdd = fm_interlockedExchangeAdd;
    fmc.interlockedCompareExchange = fm_interlockedCompareExchange;

    fmc.triggerEvent = fm_triggerEvent;
    fmc.requestRect = fm_requestRect;
    fmc.restoreRect = fm_restoreRect;


#if 1
    fmc.loadLib = fm_loadLib;
    fmc.freeLib = fm_freeLib;
    fmc.getLibFn = fm_getLibFn;
    fmc.callLibFmc = fm_callLibFmc;
    fmc.fcallLibFmc = (fmf1vd)fm_callLibFmc;
    fmc.callLib = fm_callLib;
    fmc.fcallLib = (fmf1vd)fm_callLib;
    fmc.getCpuReg = fm_getCpuReg;
    fmc.finit = fm_finit;
    fmc.fstsw = fm_fstsw;
    fmc.fstcw = fm_fstcw;
    fmc.fldcw = fm_fldcw;
#endif

	fmc.checkArrayAlloc = fm_checkArrayAlloc;

	fmc.getArrayAddress = fm_getArrayAddress;
	fmc.getBufferAddress = fm_getBufferAddress;



    fmc.setPreviewDragMode = fm_setPreviewDragMode;
    fmc.doEvents = fm_doEvents;

    fmc.hermiteInterpolate = fm_hermiteInterpolate;
    fmc.cubicInterpolate = fm_cubicInterpolate;
    fmc.cosineInterpolate = fm_cosineInterpolate;
    fmc.linearInterpolate = fm_linearInterpolate;

    fmc.getImageTitle = (fmf1)fm_getImageTitle;

    fmc.setCtlEditSize = fm_setCtlEditSize;
    fmc.setCtlFocus = fm_setCtlFocus;
    fmc.setEditBoxUpdate = fm_setEditBoxUpdate;
    fmc.checkCtlFocus = fm_checkCtlFocus;
	fmc.checkDialogFocus = fm_checkDialogFocus;
	
	fmc.getCtlText = fm_getCtlText;
	fmc.getCtlClass = fm_getCtlClass;
	fmc.getCtlItemCount = fm_getCtlItemCount;
	fmc.getCtlItemTop = fm_getCtlItemTop;
	fmc.setCtlItemTop = fm_setCtlItemTop;
	fmc.getCtlItemText = fm_getCtlItemText;
	fmc.setCtlItemText = fm_setCtlItemText;
	fmc.deleteCtlItem = fm_deleteCtlItem;
	fmc.deleteCtlItems = fm_deleteCtlItems;
	fmc.fillDir = fm_fillDir;

	fmc.getAppTheme = fm_getAppTheme;
	fmc.setDialogTheme = fm_setDialogTheme;
	fmc.setCtlTheme = fm_setCtlTheme;

	fmc.setDialogEvent = fm_setDialogEvent;
	fmc.clearDialogEvent = fm_clearDialogEvent;

	fmc.getSpecialFolder = fm_getSpecialFolder;

	fmc.setCtlStyle = fm_setCtlStyle;
	fmc.setCtlStyleEx = fm_setCtlStyleEx;
	fmc.clearCtlStyle = fm_clearCtlStyle;
	fmc.clearCtlStyleEx = fm_clearCtlStyleEx;

	fmc.setCtlBuddyStyle = fm_setCtlBuddyStyle;
	fmc.setCtlBuddyStyleEx = fm_setCtlBuddyStyleEx;
	fmc.clearCtlBuddyStyle = fm_clearCtlBuddyStyle;
	fmc.clearCtlBuddyStyleEx = fm_clearCtlBuddyStyleEx;

	fmc.getSysColor = fm_getSysColor;
	fmc.setCtlSysColor = fm_setCtlSysColor;
	fmc.setCtlFontSysColor = fm_setCtlFontSysColor;
    fmc.setDialogSysColor = fm_setDialogSysColor;

	fmc.getSysMem = fm_getSysMem;

	fmc.setToolTipDelay = fm_setToolTipDelay;

	fmc.getDialogPos = fm_getDialogPos;

	fmc.setCtlTab = fm_setCtlTab;
    fmc.getCtlTab = fm_getCtlTab;

	fmc.setCtlAnchor = fm_setCtlAnchor;
    fmc.updateAnchors = fm_updateAnchors;
    fmc.lockCtlScaling = fm_lockCtlScaling;
    fmc.scaleCtls = fm_scaleCtls;

    fmc.allocHost = fm_allocHost;
    fmc.lockHost = fm_lockHost;
    fmc.freeHost = fm_freeHost;

    fmc.setCtlScripting = fm_setCtlScripting;
    fmc.checkScriptVal = fm_checkScriptVal;
    fmc.enableScriptVal = fm_enableScriptVal;
    fmc.getScriptVal = fm_getScriptVal;
    fmc.setScriptVal = fm_setScriptVal;

    fmc.getDialogHandle= fm_getDialogHandle;
    fmc.getCtlHandle= fm_getCtlHandle;



    fmc.isqrt = fm_isqrt;
    fmc.isqrt_fast = (fmf0)isqrt_fast;	//_FASTCALL version
    fmc.cnvX = fm_cnvX;
    fmc.cnvY = fm_cnvY;
    //fmc.pget = fm_pget;       //(see above)
    /*fmc.tget = fm_tget;
    fmc.t2get = fm_t2get;
    fmc.tset = fm_tset;
    fmc.t2set = fm_t2set;*/
    //fmc.pgetr = fm_pgetr;       //(see above)
    fmc.tgetr = fm_tgetr;
    fmc.t2getr = fm_t2getr;
	fmc.t3getr = fm_t3getr;
	fmc.t4getr = fm_t4getr;
    fmc.psetr = fm_psetr;
    fmc.tsetr = fm_tsetr;
    fmc.t2setr = fm_t2setr;
	fmc.t3setr = fm_t3setr;
	fmc.t4setr = fm_t4setr;
    fmc.doForEveryRow = fm_doForEveryRow;
    fmc.doAction = (fmf1)DoAction;
    fmc.abort = fm_abort;
    fmc.testAbort = fm_testAbort;
    fmc.updateProgress = fm_updateProgress;
    fmc.updatePreview = fm_updatePreview;
    fmc.getFilterInstallDir = (fmf0)fm_getFilterInstallDir;
    fmc.appendEllipsis = (fmf1)fm_appendEllipsis;
    fmc.stripEllipsis = (fmf1)fm_stripEllipsis;
    fmc.formatString = (fmf1)fm_formatString;
    fmc.setCtlVal = fm_setCtlVal;
    fmc.setCtlDefVal = fm_setCtlDefVal;
    fmc.setCtlText = fm_setCtlText;
    fmc.setCtlTextv = fm_setCtlTextv;
    fmc.enableToolTipBalloon = fm_enableToolTipBalloon;
    fmc.setCtlToolTip = fm_setCtlToolTip;
    fmc.setCtlImage = fm_setCtlImage;
	fmc.getCtlImage = fm_getCtlImage;
    fmc.setCtlTicFreq = fm_setCtlTicFreq;
    fmc.setCtlFontColor = fm_setCtlFontColor;
    fmc.setCtlColor = fm_setCtlColor;
    fmc.getCtlColor = fm_getCtlColor;
    fmc.setCtlProperties = fm_setCtlProperties;
    fmc.clearCtlProperties = fm_clearCtlProperties;
    fmc.setCtlRange = fm_setCtlRange;
    fmc.setCtlLineSize = fm_setCtlLineSize;
    fmc.setCtlPageSize = fm_setCtlPageSize;

	fmc.setCtlThumbSize = fm_setCtlThumbSize;

    fmc.setCtlOrder = fm_setCtlOrder;

    fmc.createFont = fm_createFont;
    fmc.deleteFont = fm_deleteFont;
    fmc.setCtlFont = fm_setCtlFont;

    fmc.setCtlPos = fm_setCtlPos;
    fmc.setCtlPixelPos = fm_setCtlPixelPos;
    fmc.setCtlAction = fm_setCtlAction;
    fmc.enableCtl = fm_enableCtl;
    fmc.createCtl = fm_createCtl;
    fmc.initCtl = fm_initCtl;
    fmc.deleteCtl = fm_deleteCtl;
    fmc.resetAllCtls = fm_resetAllCtls;
    fmc.setDialogPos = fm_setDialogPos;
    fmc.setDialogStyle = fm_setDialogStyle;
    fmc.clearDialogStyle = fm_clearDialogStyle;
    fmc.setDialogStyleEx = fm_setDialogStyleEx;
    fmc.clearDialogStyleEx = fm_clearDialogStyleEx;
    fmc.setDialogText = fm_setDialogText;
    fmc.setDialogTextv = fm_setDialogTextv;
    fmc.setDialogColor = fm_setDialogColor;
    fmc.setDialogGradient = fm_setDialogGradient;
    fmc.setDialogImage = fm_setDialogImage;
    fmc.setDialogImageMode = fm_setDialogImageMode;
    fmc.setDialogDragMode = fm_setDialogDragMode;
    fmc.setDialogRegion = fm_setDialogRegion;
    fmc.createRectRgn = fm_createRectRgn;
    fmc.createRoundRectRgn = fm_createRoundRectRgn;
    fmc.createCircularRgn = fm_createCircularRgn;
    fmc.createEllipticRgn = fm_createEllipticRgn;
    fmc.createPolyRgn = fm_createPolyRgn;
    fmc.createDialogRgn = fm_createDialogRgn;
    fmc.createTitlebarRgn = fm_createTitlebarRgn;
    fmc.createMenubarRgn = fm_createMenubarRgn;
    fmc.createClientRgn = fm_createClientRgn;
    fmc.createStatusbarRgn = fm_createStatusbarRgn;
    fmc.playSoundWave = fm_playSoundWave;
    fmc.playSoundWaveLoop = fm_playSoundWaveLoop;
    fmc.playSoundWaveSync = fm_playSoundWaveSync;

    fmc.sleep = fm_sleep;
    fmc.RGB = fm_RGB;
    fmc.RGBA = fm_RGBA;
    fmc.Rval = fm_Rval;
    fmc.Gval = fm_Gval;
    fmc.Bval = fm_Bval;
    fmc.Aval = fm_Aval;
    fmc.getOpenFileName = (fmf15)fm_getOpenFileName;
    fmc.getSaveFileName = (fmf15)fm_getSaveFileName;
    fmc.chooseColor = (fmf2v)fm_chooseColor;
    fmc.shellExec = fm_shellExec;
    fmc.HDBUsToPixels = fm_HDBUsToPixels;
    fmc.VDBUsToPixels = fm_VDBUsToPixels;
    fmc.setGamma = fm_setGamma;
    fmc.gamma = fm_gamma;
    fmc.setRegRoot = (fmf1)fm_setRegRoot;
    fmc.getRegRoot = (fmf1)fm_getRegRoot;
    fmc.setRegPath = (fmf1v)fm_setRegPath;
    fmc.getRegPath = (fmf2)fm_getRegPath;
    fmc.putRegInt = (fmf2v)fm_putRegInt;
    fmc.getRegInt = (fmf2v)fm_getRegInt;
    fmc.putRegString = (fmf2v)fm_putRegString;
    fmc.getRegString = (fmf3v)fm_getRegString;
    fmc.putRegData = (fmf3v)fm_putRegData;
    fmc.getRegData = (fmf3v)fm_getRegData;
    fmc.enumRegSubKey = (fmf3)fm_enumRegSubKey;
    fmc.enumRegValue = (fmf5)fm_enumRegValue;
    fmc.deleteRegSubKey = (fmf1v)fm_deleteRegSubKey;
    fmc.deleteRegValue = (fmf1v)fm_deleteRegValue;
    fmc.msgBox = (fmf3v)msgBox;
    fmc.Info = (fmf1v)Info;
    fmc.Warn = (fmf1v)Warn;
    fmc.Error = (fmf1v)Error;
    fmc.ErrorOk = (fmf1v)ErrorOk;
    fmc.YesNo = (fmf1v)YesNo;
    fmc.YesNoCancel = (fmf1v)YesNoCancel;

    
    /* from <Winuser.h> */
    fmc.createPopupMenu = CreatePopupMenu;	//afh 01-may-2008
    fmc.createMenu = CreateMenu;
    fmc.destroyMenu = (fmf1)fm_destroyMenu; //thunk
    fmc.trackPopupMenu = fm_trackPopupMenu;
    fmc.insertMenuItem = (fmf5)fm_insertMenuItem;
    fmc.setMenu = fm_setMenu;
    

    /* from <Winbase.h> */
    fmc.findFirstFile = (fmf3)fm_findFirstFile;
    fmc.findNextFile = (fmf3)fm_findNextFile;
    fmc.findClose = (fmf1)FindClose;

    /*from <Winnls.h>*/
    fmc.getLocaleInfo = fm_getLocaleInfo; //thunk
    fmc.getSystemDefaultLCID = GetSystemDefaultLCID;	//need thunk for return type??
    fmc.getUserDefaultLCID = GetUserDefaultLCID;	//need thunk for return type??


    // C-RTL functions exposed to FM
    /* from <time.h> */
    fmc.clock = (fmf0)clock;	//need thunk for return type??
    fmc.strdate = (fmf1)_strdate;
    fmc.strtime = (fmf1)_strtime;
    fmc.time = (fmf1)time;
    /* from <string.h> */
    fmc.strcat = (fmf2)strcat;
    fmc.strchr = (fmf2)strchr;
    fmc.strcmp = (fmf2)strcmp;
    fmc.strcoll = (fmf2)strcoll;
    fmc.strcpy = (fmf2)strcpy;
    fmc.strcspn = (fmf2)strcspn;
    fmc.strdup = (fmf1)_strdup;    //needs free()
    fmc.strerror = (fmf1)strerror;
    fmc._strerror = (fmf1)_strerror; //not the same as strerror!
    fmc.stricmp = (fmf2)_stricmp;
    fmc.stricoll = (fmf2)_stricoll;
    fmc.strlen = (fmf1)strlen;
    fmc.strlwr = (fmf1)_strlwr;
    fmc.strncat = (fmf3)strncat;
    fmc.strncmp = (fmf3)strncmp;
    fmc.strncoll = (fmf3)_strncoll;
    fmc.strncpy = (fmf3)strncpy;
    fmc.strnicmp = (fmf3)_strnicmp;
    fmc.strnicoll = (fmf3)_strnicoll;
    fmc.strnset = (fmf3)_strnset;
    fmc.strpbrk = (fmf2)strpbrk;
    fmc.strrchr = (fmf2)strrchr;
    fmc.strrev = (fmf1)_strrev;
    fmc.strset = (fmf2)_strset;
    fmc.strspn = (fmf2)strspn;
    fmc.strstr = (fmf2)strstr;
    fmc.strtok = (fmf2)strtok;
    fmc.strupr = (fmf1)_strupr;
    fmc.strxfrm = (fmf3)strxfrm;
    /* from <memory.h> */
    fmc.memccpy = (fmf4)_memccpy;
    fmc.memchr = (fmf3)memchr;
    fmc.memcmp = (fmf3)memcmp;
    fmc.memcpy = (fmf3)memcpy;
    fmc.memicmp = (fmf3)_memicmp;
    fmc.memmove = (fmf3)memmove;
    fmc.memset = (fmf3)memset;
    /* from <math.h> */
    fmc.math_fabs = fabs;
    fmc.math_ceil = ceil;
    fmc.math_floor = floor;
    fmc.math_round = fm_round;
    fmc.math_chop = fm_chop;
#ifdef _WIN64
	//Trampoline to fix bug in MVSCRT x64 fmod function...
    fmc.math_fmod = X64_fmod_fix;
#else //Win32
    fmc.math_fmod = fmod;
#endif
    fmc.math_exp = exp;
    fmc.math_log = log;
    fmc.math_log10 = log10;
    fmc.math_ldexp = (fmfdid)ldexp;
    fmc.math_pow = pow;
    fmc.math_sqrt = sqrt;
    fmc.math_cos = cos;
    fmc.math_sin = sin;
    fmc.math_tan = tan;
    fmc.math_acos = acos;
    fmc.math_asin = asin;
    fmc.math_atan = atan;
    fmc.math_atan2 = atan2;
    fmc.math_cosh = cosh;
    fmc.math_sinh = sinh;
    fmc.math_tanh = tanh;
    fmc.math_hypot = _hypot;   //non-ANSI
    /* from <stdlib.h> */
    fmc.stdlib_rand = (fmf0)rand;	//need thunk for return type??
    fmc.stdlib_srand = (fmf1un)srand;
    fmc.strtod = (fmf2d)strtod;	//need thunk??
    fmc.strtol = (fmf3)strtol;	//need thunk??
    fmc.strtoul = (fmf3)strtoul;	//need thunk??
    /* from <malloc.h> */
    //fmc.alloca = (fmf1)_alloca;   //_alloca is an intrinsic, not in crtl???
    fmc.calloc = (fmf2)calloc;
    fmc.expand = (fmf2)_expand;
    fmc.free = (fmf1)free;
    fmc.malloc = (fmf1)malloc;
    fmc.msize = (fmf1)_msize;
    fmc.realloc = (fmf2)realloc;
    /* from <stdio.h> */
    fmc.stdio_clearerr = (fmf1)clearerr;
    fmc.stdio_fclose = (fmf1)fclose;
    fmc.stdio_fcloseall = (fmf0)_fcloseall;   //non-ANSI	//need thunk for return type??
    fmc.stdio_feof = (fmf1)feof;
    fmc.stdio_ferror = (fmf1)ferror;
    fmc.stdio_fflush = (fmf1)fflush;
    fmc.stdio_fgetc = (fmf1)fgetc;
    fmc.stdio_fgetpos = (fmf2)fgetpos;
    fmc.stdio_fgets = (fmf3)fgets;
    fmc.stdio_flushall = (fmf0)_flushall;    //non-ANSI		//need thunk for return type??
    fmc.stdio_fopen = (fmf2)fopen;
    fmc.stdio_fprintf = (fmf2v)fprintf;
    fmc.stdio_fputc = (fmf2)fputc;
    fmc.stdio_fputs = (fmf2)fputs;
    fmc.stdio_fread = (fmf4)fread;
    fmc.stdio_freopen = (fmf3)freopen;
    fmc.stdio_fscanf = (fmf2v)fscanf;
    fmc.stdio_fseek = (fmf3)fseek;
    fmc.stdio_fsetpos = (fmf2)fsetpos;
    fmc.stdio_ftell = (fmf1)ftell;
    fmc.stdio_fwrite = (fmf4)fwrite;
    fmc.stdio_getc = (fmf1)getc;
    //fmc.stdio_getchar = getchar;
    //fmc.stdio_gets = (fmf1)gets;
    //fmc.stdio_perror = (fmf1)perror;
    fmc.stdio_printf = (fmf1v)fm_printf;
    fmc.stdio_putc = (fmf2)putc;
    //fmc.stdio_putchar = putchar;
    //fmc.stdio_puts = (fmf1)puts;
    fmc.stdio_remove = (fmf1)remove;
    fmc.stdio_rename = (fmf2)rename;
    fmc.stdio_rewind = (fmf1)rewind;
    //fmc.stdio_scanf = (fmf1v)scanf;
    fmc.stdio_snprintf = (fmf3v)_snprintf;
    fmc.stdio_sprintf = (fmf2v)sprintf;
    fmc.stdio_sscanf = (fmf2v)sscanf;
    fmc.stdio_tmpfile = (fmf0)tmpfile;
    fmc.stdio_tmpnam = (fmf1)tmpnam;
    fmc.stdio_ungetc = (fmf2)ungetc;
	fmc.direct_chdir = (fmf1)_chdir;
	fmc.direct_getcwd = (fmf2)_getcwd;
	fmc.direct_mkdir = (fmf1)_mkdir;
	fmc.direct_rmdir = (fmf1)_rmdir;
	
    /* from <errno.h> */
    fmc.fm_errno = (fmf0)_errno;

    // private C-RTL functions
#ifdef _WIN64
	fmc.fm_chkstk = __chkstk;	//note *2* underscores for Win64
    ///fmc._ftol = _ftol;		//unresolved
    ///fmc._ftol2 = _ftol2;		//unresolved
    ///fmc._ftol2_sse = _ftol2_sse;	//unresolved
#else //Win32
	fmc.fm_chkstk = _chkstk;	//note *1* underscore for Win32
    fmc._ftol = _ftol;
    fmc._ftol2 = _ftol2;
    fmc._ftol2_sse = _ftol2_sse;
#endif


    // Win32 API functions exposed to FM
    fmc.CombineRgn = (wf4)CombineRgn;

    // Private FF+ built-in functions
    fmc.divCheck = (fmf0)gParams->CB_divCheck;

    // New FF+/FML built-in math functions
    fmc.math_iceil = fm_iceil;
    fmc.math_ifloor = fm_ifloor;
    fmc.math_iround = fm_iround;
    fmc.math_ichop = fm_ichop;
    fmc.math_ipow = fm_ipow;
    fmc.math_powi = fm_powi;

    // Event handlers
    fmc.onInitDialog = (fmf0)gParams->CB_initCode;				//not used???
    fmc.onCtl = (fmf0)gParams->CB_onCtl;						//not used???
    fmc.onZeroDivide = (fmf0)gParams->CB_onZeroDivide;			//used
    fmc.onDivideOverflow = (fmf0)gParams->CB_onDivideOverflow;	//used
    fmc.onFilterStart = (fmf0)gParams->CB_onFilterStart;		//not used???
    fmc.onFilterEnd = (fmf0)gParams->CB_onFilterEnd;			//not used???
    fmc.UserFunctions = (fmf0)gParams->CB_UserFunctions;		//not used for now

    //Following are not necessarily static, but we initialize
    //them here anyway for good hygiene...
    //These also get reinitialized later at the start of each tile.
    fmc.dmax = fmc.D = 1024;
    fmc.xmin = fmc.ymin = fmc.zmin = fmc.pmin = fmc.mmin = fmc.dmin = 0;
    fmc.rmax = fmc.gmax = fmc.bmax = fmc.amax = fmc.cmax = 255;	//should change with bit-depth??
    fmc.imax = fmc.umax = fmc.vmax = 255;
    fmc.rmin = fmc.gmin = fmc.bmin = fmc.amin = fmc.cmin = 0;
    fmc.imin = fmc.umin = fmc.vmin = 0;
    fmc.R = fmc.G = fmc.B = fmc.A = fmc.C = 255;    //conflict between FF and FF+???
    fmc.I = fmc.U = fmc.V = 255;
    fmc.t = fmc.tmin = 0;
    fmc.total = fmc.tmax = 1;

    //Initialize some other variables for good hygiene.
    fmc.n = 0;
    fmc.ctlMsg = 0;
    fmc.ctlCode = 0;
    fmc.ctlClass = CC_UNUSED;
    fmc.dividend = 0;
    fmc.divisor = 1;

    fmc.doingProxy = FALSE;
    fmc.doingScripting = 0;

    //initialize random number generator 
    srand(1);

    //initialize anonymous cells
    if (Dont_Initialize_Cells == 0)
    {
        for (i = 0; i < N_CELLS; i++) {
            fmc.cell[i] = 0;
        }
    }

    // Cache current control values in fmc.pre_ctl[].
    for (i=0; i < N_CTLS; i++) {
        fmc.pre_ctl[i] = gParams->ctl[i].val;
    }

} /*InitContextRecord*/

/*****************************************************************************/
/* Filter the area. */


//Commented out by Harald Heim, Jun 4, 2002
/*
#define maxBlurRadius 3

#if 0
#define mskVal(row,rowOff,col,colOff) \
            fmc.mskPtr[(rowOff)*fmc.mskRowDelta+(colOff)*fmc.mskColDelta]
#elif 0
#define mskVal(row,rowOff,col,colOff) \
            fmc.mskData[((row)+(rowOff))*fmc.mskRowDelta+((col)+(colOff))*fmc.mskColDelta]
#elif 1
static unsigned8 gMskVal(int row, int rowOff, int col, int colOff)
{
    // alert if out of bounds, for debugging...
#ifdef CHECK_GMSKVAL
#ifndef NOTRACE
    if ( row + rowOff <            -gBevWidth-maxBlurRadius   ||
         row + rowOff >    fmc.rows+gBevWidth+maxBlurRadius-1 ||
         col + colOff <            -gBevWidth-maxBlurRadius   ||
         col + colOff > fmc.columns+gBevWidth+maxBlurRadius-1 )
    {
        Trace (globals, "Bad maskData subscripts:\n %d, %d, %d, %d",
               row, rowOff, col, colOff);
    }
#endif
#endif

    if ((gNoMask || !(gParams->flags & AFH_USE_HOST_PADDING)) &&
        (row+rowOff < minRow || row+rowOff > maxRow ||
         col+colOff < minCol || col+colOff > maxCol) )
        return 0;   // Simulate the padding...
    else if (gNoMask)
        return 255; // Entire filterRect is inside the mask.
    else
        return fmc.mskPtr[rowOff*fmc.mskRowDelta + colOff*fmc.mskColDelta];
} //gMskVal
#define mskVal(row,rowOff,col,colOff) \
    gMskVal(row,rowOff,col,colOff)
#endif
*/


////////////////////////////////////////

void DoFilterRect (GPtr localGlobals)
{
    static int reentry_depth = 0;
    static int reentry_count = 0;

    if (gResult != noErr) {
        // Got an error somewhere along the way...
        // Get out quick in case the error was returned by
        // AdvanceState(), in which case the input/output data
        // pointers are bogus and can cause severe dyspepsia.
        return;
    }

    //check for reentry...
    if (reentry_depth++ > 0) {
        //reentered!!! Log the fact, and exit quickly...
        reentry_count++;
        reentry_depth--;
        return; //get out quick!!!
    }

    globals = localGlobals;

    if (!gStuff->outData) {
        // I think I've caught all cases where zero-dragging
        // can cause a GP in the filter code because outData
        // is set to 0 while dragging.  This is just a last-
        // minute sanity check to try to avoid  GP-ing on the
        // end user!  (Record any failures for later analysis.)
        static int unexpected_null_outData_aborts = 0;
        unexpected_null_outData_aborts++;
#if 1
        Info("Unexpected NULL outData, count = %d", unexpected_null_outData_aborts);
#endif
        reentry_depth--;
        return; //get out quick!!!
    }

    if (fmc.doingProxy) {
        //proxy is never tiled, so we can do this here...

#ifndef STANDALONE
        //If source needs to be recompiled, ask user and recompile
        //if requested...
        // (Keep this message in sync with the same message in
        // WM_COMMAND/IDOK.  But note that here we use just yes/no,
        // not yes/no/cancel...)
        if (!gParams->fValidCompile && gParams->fGotSourceCode) {
            char szMessage[512];
            // Get the message string...
            // (e.g., "The compiled code is not up-to-date.\n\n"
            //        "Would you like to compile it first?"
            LoadString(hDllInstance, IDS_RECOMPILE, szMessage, sizeof szMessage);
            switch (YesNo(szMessage))
            {
            case IDYES:
            case IDOK:
                // (re-)codegen and init before go...
                SendMessage(fmc.hDlg, WM_COMMAND, IDC_INIT, 0);
                break;
            case IDCANCEL:
            case IDABORT:
            case IDNO:
            default:
                break;  //proceed without re-compiling
            }//switch
        }
        //But what if compile fails (i.e., still !gParams->fValidCompile)?

#endif //ifndef STANDALONE

        InitBeforeFilter();
    }

    { //scope
    VRect scaledFilterRect;
    OSErr err = noErr;

    fmc.scaleFactor = gStuff->inputRate >> 16;
    fmc.srcPtr = (unsigned8 *) MyAdjustedInData;
    fmc.dstPtr = (unsigned8 *) gStuff->outData;
    fmc.mskData = (unsigned8 *) MyAdjustedMaskData;
	fmc.outRect32 = GetOutRect();	//needed??
    fmc.rows = fmc.outRect32.bottom - fmc.outRect32.top;
    fmc.columns = fmc.outRect32.right - fmc.outRect32.left;
    fmc.srcRowDelta = gStuff->inRowBytes;
    fmc.srcColDelta = gStuff->inHiPlane - gStuff->inLoPlane + 1;
    fmc.inPlanes = fmc.srcColDelta;
    fmc.dstRowDelta = gStuff->outRowBytes;
    fmc.dstColDelta = gStuff->planes;
    fmc.mskRowDelta = gStuff->maskRowBytes;
    fmc.mskColDelta = 1;

    { //new scope

#ifndef DONT_USE_MASK
    if (gStuff->haveMask)
    {   // mask stuff already set up for selection mask...
        gNoMask = FALSE;
    }
    else
#endif
    if (fmc.inPlanes >= 4)
    {   // Channel 3 is probably an alpha channel, so
        // use the alpha channel (layer transparency mask)
        // instead of the selection mask
        fmc.mskData = (unsigned8 *) MyAdjustedInData + 3;
        fmc.mskRowDelta = fmc.srcRowDelta;
        fmc.mskColDelta = fmc.srcColDelta;
        gNoMask = FALSE;
    }
    else
    {   // No selection mask and no transparency channel, so
        // we'll just fake a mask that includes the entire
        // filterRect (which will give us a nice rectangular
        // bevel).
        gNoMask = TRUE;
    }

	fmc.filterRect32 = GetFilterRect();	//needed?
	scaledFilterRect = fmc.filterRect32;
    scaleRect32(&scaledFilterRect, 1, fmc.scaleFactor);
    X0 = scaledFilterRect.left;
    Y0 = scaledFilterRect.top;
	fmc.inRect32 = GetInRect();	//needed??
    fmc.x_start = fmc.inRect32.left - X0;
    fmc.y_start = fmc.inRect32.top - Y0;
    fmc.X = fmc.xmax = scaledFilterRect.right - X0;
    fmc.Y = fmc.ymax = scaledFilterRect.bottom - Y0;
    //fmc.Z = fmc.zmax = fmc.pmax = (gStuff->planes > 4)?4:gStuff->planes; //What about CMYK + alpha???
    fmc.Z = fmc.zmax = fmc.pmax = gStuff->planes;
    fmc.M = fmc.mmax = (int)( sqrt((double)(fmc.X*fmc.X + fmc.Y*fmc.Y)) / 2 + 0.0 ); // don't round per Jens???
    fmc.dmax = fmc.D = 1024;
    fmc.xmin = fmc.ymin = fmc.zmin = fmc.pmin = fmc.mmin = fmc.dmin = 0;
    fmc.rmax = fmc.gmax = fmc.bmax = fmc.amax = fmc.cmax = 255;	//should change with bit-depth??
    fmc.imax = fmc.umax = fmc.vmax = 255;
    fmc.rmin = fmc.gmin = fmc.bmin = fmc.amin = fmc.cmin = 0;
    fmc.imin = fmc.umin = fmc.vmin = 0;
    fmc.R = fmc.G = fmc.B = fmc.A = fmc.C = 255;    //conflict between FF and FF+???
    fmc.I = fmc.U = fmc.V = 255;
    fmc.t = fmc.tmin = 0;
    fmc.total = fmc.tmax = 1;

#if 0   //already done in InitBeforeFilter.
    // Cache current control values in fmc.pre_ctl[].
    for (i=0; i < N_CTLS; i++) {
        fmc.pre_ctl[i] = gParams->ctl[i].val;
    }
#endif

    fmc.hDlg = MyHDlg; //already done in WM_INITDIALOG!!!

    // Compute min/max values for column and row indices that still lie
    // within the (scaled) filterRect (used to simulate padding of accesses
    // outside the bounds of filterRect).
	fmc.outRect32 = GetOutRect();	//needed??
    minCol = scaledFilterRect.left   - fmc.outRect32.left;
    maxCol = scaledFilterRect.right  - fmc.outRect32.left - 1;
    minRow = scaledFilterRect.top    - fmc.outRect32.top;
    maxRow = scaledFilterRect.bottom - fmc.outRect32.top - 1;

    {  

        // pass 2: ...

        fmc.writeAlpha = 
#ifndef DONT_USE_MASK
                !gStuff->haveMask &&
#endif
                !gNoMask;

        //if ((gParams->flags & AFH_OPTIMIZE_BEVEL))  //Co-opt this flag for timing...

        //Unconditionally copy input tile to output...
        //Do this with memcpy instead????
        fmc.y = fmc.y_start;
        for (fmc.row = 0; fmc.row < fmc.rows; fmc.row++)
        {
            if (gParams->fValidCompile) {
                // No progress indicator -- but do testAbort()
                if (fm_testAbort()) goto egress;
            }
            else if (fmc.doingProxy) {
                if (fm_updateProgress(fmc.row, fmc.rows)) goto egress;
            }
            else {
                if (fm_updateProgress(fmc.y - Y0, fmc.Y - Y0)) goto egress;
                fmc.y++;
            }

            fmc.srcPtr = (unsigned8 *) MyAdjustedInData + (fmc.row * fmc.srcRowDelta);
            fmc.dstPtr = (unsigned8 *) gStuff->outData + (fmc.row * fmc.dstRowDelta);
            fmc.mskPtr = fmc.mskData + (fmc.row * fmc.mskRowDelta);

            //Copy over all planes and possibly the alpha...

            /*if (gStuff->imageMode == plugInModeRGB48 ||
                gStuff->imageMode == plugInModeGray16|| 
                gStuff->imageMode == plugInModeLab48 || 
                gStuff->imageMode == plugInModeCMYK64 ||
                gStuff->imageMode == plugInModeDeepMultichannel ||
                gStuff->imageMode == plugInModeDuotone16){ */
            
            if (gStuff->imageMode > 9) {

                // pixel depth = 16
#define srcPtr16 ((unsigned16 *)fmc.srcPtr)
#define dstPtr16 ((unsigned16 *)fmc.dstPtr)
#define mskPtr16 ((unsigned16 *)fmc.mskPtr)

                for (fmc.col = 0; fmc.col < fmc.columns; fmc.col++)
                {
                    int plane;
                    for (plane = 0; plane < gStuff->planes; plane++)
                       dstPtr16[plane] = srcPtr16[plane];

#if 0
                    // No alpha in 16-bit mode?
                    if (fmc.writeAlpha) {   //redundant???
                        dstPtr16[3] = mskVal(fmc.row, 0, fmc.col, 0);
                    }
#endif

                    srcPtr16 += fmc.srcColDelta;
                    dstPtr16 += fmc.dstColDelta;
                    ///mskPtr16 += fmc.mskColDelta;    //wrong????

                } /* for fmc.col */
            }
            else {
                // pixel depth = 8
                for (fmc.col = 0; fmc.col < fmc.columns; fmc.col++)
                {
                    int plane;
                    for (plane = 0; plane < gStuff->planes; plane++)
                       fmc.dstPtr[plane] = fmc.srcPtr[plane];

                    if (fmc.writeAlpha) {   //redundant???
                        fmc.dstPtr[3] = mskVal(fmc.row, 0, fmc.col, 0);
                    }
                    fmc.srcPtr += fmc.srcColDelta;
                    fmc.dstPtr += fmc.dstColDelta;
                    fmc.mskPtr += fmc.mskColDelta;  //wrong????
                } /* for fmc.col */
            } /* if 8-bit depth */
        } /* for fmc.row */


        //call ForEveryTile...
        fmc.x = fmc.x_start;
        fmc.y = fmc.y_start;
        fmc.x_end = fmc.x_start + fmc.columns;
        fmc.y_end = fmc.y_start + fmc.rows;
        fmc.row = 0;
        fmc.col = 0;
        fmc.tbuf = NULL;
        fmc.t2buf = NULL;
        fmc.t3buf = NULL;
        fmc.t4buf = NULL;

        if (gParams->fValidCompile) {
            //need to handle longjmp back???
            INT_PTR res;
#if 0
            //Min version of tile buffers (same size as output tile)
            if (fmc.need_tbuf) {
                //allocate a (temporary) tile buffer...
                //Should use AllocateBuffer() ????
                do {
                    res = IDOK;

                    //some kind of dword padding (+ 3) added by Harald Heim, February 2003

					fmc.tbuf = calloc( fmc.rows * fmc.columns * fmc.Z + 3 , sizeof (fmc.tbuf[0]) );

                    if (!fmc.tbuf) {
                        res = Error("Could not allocate a tile buffer (%d bytes).",
                                fmc.rows * fmc.columns * fmc.Z * sizeof (fmc.tbuf[0]));
                    }
                } while (res == IDRETRY);
                if (res == IDABORT) {
                    gResult = memFullErr;
                    goto egress;
                }
#if 0
                else {
                    Info("Allocated a tile buffer (%d bytes).",
                            fmc.rows * fmc.columns * fmc.Z * sizeof (fmc.tbuf[0]));
                }
#endif
            }//if need tbuf
            if (fmc.need_t2buf) {
                //allocate a (temporary) tile buffer 2...
                //Should use AllocateBuffer() ????
                do {
                    res = IDOK;

                    //some kind of dword padding (+ 3) added by Harald Heim, February 2003
 
                    fmc.t2buf = calloc(fmc.rows * fmc.columns * fmc.Z + 3, sizeof (fmc.t2buf[0]));

                    if (!fmc.t2buf) {
                        res = Error("Could not allocate tile buffer 2 (%d bytes).",
                                fmc.rows * fmc.columns * fmc.Z * sizeof (fmc.t2buf[0]));
                    }
                } while (res == IDRETRY);
                if (res == IDABORT) {
                    gResult = memFullErr;
                    goto egress;
                }
#if 0
                else {
                    Info("Allocated tile buffer 2 (%d bytes).",
                            fmc.rows * fmc.columns * fmc.Z * sizeof (fmc.t2buf[0]));
                }
#endif
            }//if need t2buf
			if (fmc.need_t3buf) {
                //allocate a (temporary) tile buffer 3...
                //Should use AllocateBuffer() ????
                do {
                    res = IDOK;

                    //some kind of dword padding (+ 3) added by Harald Heim, February 2003
 
                    fmc.t3buf = calloc(fmc.rows * fmc.columns * fmc.Z + 3, sizeof (fmc.t3buf[0]));

                    if (!fmc.t3buf) {
                        res = Error("Could not allocate tile buffer 3 (%d bytes).",
                                fmc.rows * fmc.columns * fmc.Z * sizeof (fmc.t3buf[0]));
                    }
                } while (res == IDRETRY);
                if (res == IDABORT) {
                    gResult = memFullErr;
                    goto egress;
                }
#if 0
                else {
                    Info("Allocated tile buffer 3 (%d bytes).",
                            fmc.rows * fmc.columns * fmc.Z * sizeof (fmc.t3buf[0]));
                }
#endif
            }//if need t3buf
			if (fmc.need_t4buf) {
                //allocate a (temporary) tile buffer 4...
                //Should use AllocateBuffer() ????
                do {
                    res = IDOK;

                    //some kind of dword padding (+ 3) added by Harald Heim, February 2003
 
                    fmc.t4buf = calloc(fmc.rows * fmc.columns * fmc.Z + 3, sizeof (fmc.t3buf[0]));

                    if (!fmc.t4buf) {
                        res = Error("Could not allocate tile buffer 4 (%d bytes).",
                                fmc.rows * fmc.columns * fmc.Z * sizeof (fmc.t4buf[0]));
                    }
                } while (res == IDRETRY);
                if (res == IDABORT) {
                    gResult = memFullErr;
                    goto egress;
                }
#if 0
                else {
                    Info("Allocated tile buffer 4 (%d bytes).",
                            fmc.rows * fmc.columns * fmc.Z * sizeof (fmc.t4buf[0]));
                }
#endif
            }//if need t4buf
#else
            //Max version of tile buffers (same size as input tile)
            if (fmc.need_tbuf) {
                //allocate a (temporary) tile buffer...
                //Should use AllocateBuffer() ????
				
                do {
                    int bufferSize = fmc.imageMode>9 ? 2:1;
					
					res = IDOK;

                    //some kind of dword padding (+ 3) added by Harald Heim, February 2003
					
					if (gStuff->bufferProcs && gStuff->bufferProcs->allocateProc && gStuff->bufferProcs->lockProc && gStuff->bufferProcs->unlockProc && gStuff->bufferProcs->freeProc){ //Photoshop allocation
						if (AllocateBuffer( (fmc.X * fmc.Y * fmc.Z + 3) * bufferSize, &fmc.tbufID) == noErr){
							fmc.tbuf = LockBuffer(fmc.tbufID, true);
						} else { //Fall back to Windows allocation
							fmc.tbuf = calloc(fmc.X * fmc.Y * fmc.Z + 3, bufferSize);
							fmc.tbufID=0;
						}
					}else { //Windows allocation
						fmc.tbuf = calloc(fmc.X * fmc.Y * fmc.Z + 3, bufferSize); //sizeof (fmc.tbuf[0])
						fmc.tbufID=0;
					}


                    if (!fmc.tbuf) {
                        res = Error("Could not allocate a tile buffer (%d bytes).",
                                fmc.X * fmc.Y * fmc.Z * bufferSize); //sizeof (fmc.tbuf[0])
                    }
                } while (res == IDRETRY);
                if (res == IDABORT) {
                    gResult = memFullErr;
                    goto egress;
                }
#if 0
                else {
                    Info("Allocated a tile buffer (%d bytes).",
                            fmc.X * fmc.Y * fmc.Z * sizeof (fmc.tbuf[0]));
                }
#endif
            }//if need tbuf
            if (fmc.need_t2buf) {
                //allocate a (temporary) tile buffer 2...
                //Should use AllocateBuffer() ????
                do {
                    int bufferSize = fmc.imageMode>9 ? 2:1;
					res = IDOK;

                    //some kind of dword padding (+ 3) added by Harald Heim, February 2003

                    if (gStuff->bufferProcs && gStuff->bufferProcs->allocateProc && gStuff->bufferProcs->lockProc && gStuff->bufferProcs->unlockProc && gStuff->bufferProcs->freeProc){ //Photoshop allocation
						if (AllocateBuffer( (fmc.X * fmc.Y * fmc.Z + 3) * bufferSize, &fmc.t2bufID) == noErr){
							fmc.t2buf = LockBuffer(fmc.t2bufID, true);
						} else { //Fall back to Windows allocation
							fmc.t2buf = calloc(fmc.X * fmc.Y * fmc.Z + 3, bufferSize);
							fmc.t2bufID=0;
						}
					}else { //Windows allocation
						fmc.t2buf = calloc(fmc.X * fmc.Y * fmc.Z + 3, bufferSize); //sizeof (fmc.t2buf[0])
						fmc.t2bufID=0;
					}


                    if (!fmc.t2buf) {
                        res = Error("Could not allocate tile buffer 2 (%d bytes).",
                                fmc.X * fmc.Y * fmc.Z * bufferSize); //sizeof (fmc.t2buf[0])
                    }
                } while (res == IDRETRY);
                if (res == IDABORT) {
                    gResult = memFullErr;
                    goto egress;
                }
#if 0
                else {
                    Info("Allocated tile buffer 2 (%d bytes).",
                            fmc.X * fmc.Y * fmc.Z * sizeof (fmc.t2buf[0]));
                }
#endif
            }//if need t2buf
			if (fmc.need_t3buf) {
                //allocate a (temporary) tile buffer 3...
                //Should use AllocateBuffer() ????
                do {
                    int bufferSize = fmc.imageMode>9 ? 2:1;
					res = IDOK;

                    //some kind of dword padding (+ 3) added by Harald Heim, February 2003
					
					if (gStuff->bufferProcs && gStuff->bufferProcs->allocateProc && gStuff->bufferProcs->lockProc && gStuff->bufferProcs->unlockProc && gStuff->bufferProcs->freeProc){ //Photoshop allocation
						if (AllocateBuffer( (fmc.X * fmc.Y * fmc.Z + 3) * bufferSize, &fmc.t3bufID) == noErr){
							fmc.t3buf = LockBuffer(fmc.t3bufID, true);
						} else { //Fall back to Windows allocation
							fmc.t3buf = calloc(fmc.X * fmc.Y * fmc.Z + 3, bufferSize);
							fmc.t3bufID=0;
						}
					}else { //Windows allocation
						fmc.t3buf = calloc(fmc.X * fmc.Y * fmc.Z + 3, bufferSize); //sizeof (fmc.tbuf[0])
						fmc.t3bufID=0;
					}

                    if (!fmc.t3buf) {
                        res = Error("Could not allocate tile buffer 3 (%d bytes).",
                                fmc.X * fmc.Y * fmc.Z * bufferSize); //sizeof (fmc.t3buf[0])
                    }
                } while (res == IDRETRY);
                if (res == IDABORT) {
                    gResult = memFullErr;
                    goto egress;
                }
#if 0
                else {
                    Info("Allocated tile buffer 3 (%d bytes).",
                            fmc.X * fmc.Y * fmc.Z * sizeof (fmc.t3buf[0]));
                }
#endif
            }//if need t3buf
			if (fmc.need_t4buf) {
                //allocate a (temporary) tile buffer 4...
                //Should use AllocateBuffer() ????
                do {
                    int bufferSize = fmc.imageMode>9 ? 2:1;
					res = IDOK;

                    //some kind of dword padding (+ 3) added by Harald Heim, February 2003
					
					if (gStuff->bufferProcs && gStuff->bufferProcs->allocateProc && gStuff->bufferProcs->lockProc && gStuff->bufferProcs->unlockProc && gStuff->bufferProcs->freeProc){ //Photoshop allocation
						if (AllocateBuffer( (fmc.X * fmc.Y * fmc.Z + 3) * bufferSize, &fmc.t4bufID) == noErr){
							fmc.t4buf = LockBuffer(fmc.t4bufID, true);
						} else { //Fall back to Windows allocation
							fmc.t4buf = calloc(fmc.X * fmc.Y * fmc.Z + 3, bufferSize);
							fmc.t4bufID=0;
						}
					}else { //Windows allocation
						fmc.t4buf = calloc(fmc.X * fmc.Y * fmc.Z + 3, bufferSize); //sizeof (fmc.tbuf[0])
						fmc.t4bufID=0;
					}

                    if (!fmc.t4buf) {
                        res = Error("Could not allocate tile buffer 4 (%d bytes).",
                                fmc.X * fmc.Y * fmc.Z * bufferSize); //sizeof (fmc.t4buf[0])
                    }
                } while (res == IDRETRY);
                if (res == IDABORT) {
                    gResult = memFullErr;
                    goto egress;
                }
#if 0
                else {
                    Info("Allocated tile buffer 4 (%d bytes).",
                            fmc.X * fmc.Y * fmc.Z * sizeof (fmc.t4buf[0]));
                }
#endif
            }//if need t4buf
#endif

            //if (gParams->gInitialInvocation) scaleLock = false; //Activate Control Scaling again

			//Catch any longjmp's back...
			if (setjmp(Mde_error_exit_jb)) {
				//someone longjmp'ed back to us...
				goto egress;   //abort
			}
			//Call the ForEveryTile handler
            res = X86_call_fmf1(gParams->CB_forEveryTile, (INT_PTR)&fmc);
            if (!res) {
                //If ForEveryTile() returns false, call doForEveryRow...
                res = fm_doForEveryRow();
            }

        }//if fValidCompile
    }//scope


egress:
    // Free the temporary tile buffers if allocated...
    if (fmc.tbuf) {
        if (fmc.tbufID != 0){ //was allocated with Buffer Proc
			UnlockBuffer(fmc.tbufID);
			FreeBuffer(fmc.tbufID); 
		} else
			free (fmc.tbuf);

		fmc.tbufID = 0;
        fmc.tbuf = NULL;
#if 0
        Info("Freed the tile buffer.");
#endif
    }
    if (fmc.t2buf) {
        if (fmc.t2bufID != 0){ //was allocated with Buffer Proc
			UnlockBuffer(fmc.t2bufID);
			FreeBuffer(fmc.t2bufID); 
		} else
			free (fmc.t2buf);

		fmc.t2bufID = 0;
        fmc.t2buf = NULL;
#if 0
        Info("Freed tile buffer 2.");
#endif
    }
	if (fmc.t3buf) {
        if (fmc.t3bufID != 0){ //was allocated with Buffer Proc
			UnlockBuffer(fmc.t3bufID);
			FreeBuffer(fmc.t3bufID); 
		} else
			free (fmc.t3buf);

		fmc.t3bufID = 0;
        fmc.t3buf = NULL;
#if 0
        Info("Freed tile buffer 3.");
#endif
    }
	if (fmc.t4buf) {
        if (fmc.t4bufID != 0){ //was allocated with Buffer Proc
			UnlockBuffer(fmc.t4bufID);
			FreeBuffer(fmc.t4bufID); 
		} else
			free (fmc.t4buf);

		fmc.t4bufID = 0;
        fmc.t4buf = NULL;
#if 0
        Info("Freed tile buffer 4.");
#endif
    }
    } //scope
} //scope

if (fmc.doingProxy) {
    //Only one tile when doing proxy, so we can call the OnFilterEnd
    //handler here.

    //Call the OnFilterEnd handler...
    if (gParams->fValidCompile) {
        INT_PTR res;
		//Catch any longjmp's back...
		if (setjmp(Mde_error_exit_jb)) {
			//someone longjmp'ed back to us...
			goto egress2;   //abort
		}
        //for now, we ignore the return value
        res = X86_call_fmf1(gParams->CB_onFilterEnd, (INT_PTR)&fmc);
    }
}//doingProxy

egress2:
	reentry_depth--;
} /*DoFilterRect*/

/******************************************************/

static INT_PTR fm_doForEveryRow(void) {
    INT_PTR res;

    assert(gParams->fValidCompile);

    // re-initialize in case ForEveryTile() altered these values...
    fmc.y = fmc.y_start;

    for (fmc.row = 0; fmc.row < fmc.rows; fmc.row++, fmc.y++)
    {
        if (fmc.doingProxy) {
            if (res = fm_updateProgress(fmc.row, fmc.rows)) return res;
        }
        else {
            if (res = fm_updateProgress(fmc.y - Y0, fmc.Y - Y0)) return res;
        }

        fmc.srcPtr = (unsigned8 *) MyAdjustedInData + (fmc.row * fmc.srcRowDelta);
        fmc.dstPtr = (unsigned8 *) gStuff->outData + (fmc.row * fmc.dstRowDelta);
        fmc.mskPtr = fmc.mskData + (fmc.row * fmc.mskRowDelta);

        fmc.x = fmc.x_start;
        fmc.col = 0;        //redundant??? DoOneRow sets it also.
        fmc.writeAlpha = 
#ifndef DONT_USE_MASK
                !gStuff->haveMask &&
#endif
            !gNoMask;
        // call ForEveryRow() here.
        // or call it from within DoOneRow???
        // Then call DoOneRow to drive the ForEveryPixel code...
        //execute the compiled code...
        if (setjmp(Mde_error_exit_jb)) {
            //someone longjmp'ed back to us...
            return 1;   //failure
        }
        else {
            X86_call_fmf1(gParams->codeBuffer, (INT_PTR)gFmcp);
        }
    } /* for fmc.row */
    return 0;   //success
} /*fm_doForEveryRow*/

/******************************************************/

/* DoFilterRect for non-advanceState */

#ifndef DEMO

void DoFilterRectNoAdvanceState (GPtr globals)
{
#ifndef NOTRACE
    Trace (globals, "DoFilterRectNoAdvanceState:\n"
                    " Didn't expect to come here!");
#endif
}

#endif

/*****************************************************************************/

/* If we do all of the filtering during the start phase, the continue
   phase is negligible. */

/* Otherwise (no AdvanceState) Filters the area and requests the next area. */

#ifndef DEMO

void DoContinue (GPtr globals)
{
#ifndef NOTRACE
    Trace (globals, "DoContinue:\n"
                    " Didn't expect to come here!");
#endif
    
    if (TestAbort ())
    {
        gResult = 1;
        return;
    }
        
    if (!WarnAdvanceStateAvailable())
    {
        DoFilterRectNoAdvanceState (globals);
    
        if (!DoNextRect (globals))
            {
                PISetRect32(&fmc.inRect32, 0, 0, 0, 0);
                PISetRect32(&fmc.outRect32, 0, 0, 0, 0);
				SetOutRect(fmc.inRect32);
				SetOutRect(fmc.outRect32);
            }
    }
    else
    {
        PISetRect32(&fmc.inRect32, 0, 0, 0, 0);
        PISetRect32(&fmc.outRect32, 0, 0, 0, 0);
		SetOutRect(fmc.inRect32);
		SetOutRect(fmc.outRect32);
    }
}

#endif

/*****************************************************************************/

/* Requests first part of the image to be filtered. */

void DoInitialRect (GPtr globals)
{
#ifndef NOTRACE
    Trace (globals, "DoInitialRect:\n"
                    " Didn't expect to come here!");
#endif

	fmc.filterRect32 = GetFilterRect();	//needed?
    fmc.inRect32        = fmc.filterRect32;
    fmc.inRect32.bottom = fmc.inRect32.top + 1;
	SetInRect(fmc.inRect32);

    gStuff->inLoPlane = 0;
    gStuff->inHiPlane = gStuff->planes - 1;
    
    fmc.outRect32 = fmc.inRect32;
	SetOutRect(fmc.outRect32);

    gStuff->outLoPlane = gStuff->inLoPlane;
    gStuff->outHiPlane = gStuff->inHiPlane;

}
/*****************************************************************************/

/* Request the next area. */

Boolean DoNextRect (GPtr globals)
    {
#ifndef NOTRACE
    Trace (globals, "DoNextRect:\n"
                    " Didn't expect to come here!");
#endif

	fmc.inRect32 = GetInRect();
    fmc.inRect32.top    = fmc.inRect32.top    + 1;
    fmc.inRect32.bottom = fmc.inRect32.bottom + 1;
	SetOutRect(fmc.inRect32);
    
    fmc.outRect32 = fmc.inRect32;
	SetOutRect(fmc.outRect32);

    return fmc.inRect32.top < fmc.filterRect32.bottom;

    }

/*****************************************************************************/

/* Do any necessary clean-up. */

#ifndef DEMO

void DoFinish (GPtr globals)
{
    
    //Added by Harald Heim
    int i;

#ifndef NOTRACE
    Trace (globals, "DoFinish:\n"
                    " Nothing much to do!");
#endif

#ifdef SCRIPTABLE
    WriteScriptParameters();
#endif


	//Clean up threads
	fm_terminateThread(0);

    // Kill any async sound that might still be running...
    fm_playSoundWave(0);

    //Added by Harald Heim
    //Delete Array if available
    for (i=0; i<N_ARRAYS; i++) fm_freeArray(i);
    
    //Delete Fonts
    for (i=0; i<N_FONTS; i++){
            if (Font[i] != 0) DeleteObject(Font[i]);
    }

    //Added by Harald Heim
    //free distance buffer if necessary
    if (bufferID != 0)
    {
        UnlockBuffer(bufferID);
        FreeBuffer(bufferID);
        bufferID = 0;
    }

}

#endif

/*****************************************************************************/
/* Some utility functions to support proxies in dialog */

long long2fixed(long value) {
    return value << 16;
}


/*****************************************************************************/

void scaleRect(Rect *l, long n, long d) {					//What if coords are negative???
    // Carefully control rounding so scaled Rect is always
    // a (scaled) subset of the original Rect (we assume the
    // original Rect has no negative coordinates -- otherwise
    // we need to be even more careful!).
    l->left = (int16)((l->left * n + d - 1) / d);   // round up
    l->top = (int16)((l->top * n + d - 1) / d);     // round up
    l->right = (int16)((l->right * n) / d);         // round down
    l->bottom = (int16)((l->bottom * n) / d);       // round down
}

void scaleRect32(VRect *l, long n, long d) {				//What if coords are negative???
    // Carefully control rounding so scaled VRect is always
    // a (scaled) subset of the original VRect (we assume the
    // original VRect has no negative coordinates -- otherwise
    // we need to be even more careful!).
    l->left = ((l->left * n + d - 1) / d);   // round up
    l->top = ((l->top * n + d - 1) / d);     // round up
    l->right = ((l->right * n) / d);         // round down
    l->bottom = ((l->bottom * n) / d);       // round down
}

void shrinkRect(Rect *l, long x, long y) {
    l->left += (int16)x;
    l->top += (int16)y;
    l->right -= (int16)x;
    l->bottom -= (int16)y;
}

void shrinkRect32(VRect *l, long x, long y) {
    l->left += x;
    l->top += y;
    l->right -= x;
    l->bottom -= y;
}

void copyRect(Rect *l, const Rect *r) {	//make this a macro???
    l->left = r->left;
    l->top = r->top;
    l->right = r->right;
    l->bottom = r->bottom;
}

void copyRect32(VRect *l, const VRect *r) {	//make this a macro???
	*l = *r;
}

/*****************************************************************************/

/* Test for the display pixels proc and if it isn't available, show an alert. */

Boolean CheckProxiesSupported (DisplayPixelsProc displayPixels)
    {
    
    Boolean available = (displayPixels != NULL);
    
    if (!available)
        ShowAlert (16999);
        
    return available;
    
    }

/***********************************************************************************/

/* Computes the scaled down rectangle and the scale factor for the proxy */

void CalcProxyScaleFactor (const VRect *filterRect, Rect *proxyRect, long *scaleFactor)
    {

    int filterHeight;
    int filterWidth;
    int itemHeight;
    int itemWidth;
    VPoint fraction;

    shrinkRect (proxyRect, 1, 1);        // for the border

    filterHeight = filterRect->bottom - filterRect->top;
    filterWidth  = filterRect->right  - filterRect->left;
    
    itemHeight = proxyRect->bottom - proxyRect->top;
    itemWidth  = proxyRect->right  - proxyRect->left;
    
    if (itemHeight > filterHeight)
        itemHeight = filterHeight;
        
    if (itemWidth > filterWidth)
        itemWidth = filterWidth;

    fraction.h = (filterWidth + itemWidth - 1) / itemWidth;
    fraction.v = (filterHeight + itemHeight - 1) / itemHeight;

#if 0
    if (fraction.h > fraction.v) 
        *scaleFactor = fraction.h;
    else
        *scaleFactor = fraction.v;
#else
	*scaleFactor = max(fraction.h, fraction.v);
#endif

#if 0
	//For very thin images
	//afh fixed bug 01/06/2014
	if (filterHeight/(*scaleFactor)==0 || filterWidth/(*scaleFactor)==0){
		int c;
		for (c=*scaleFactor; c>=1; c--){
			if (filterHeight/c>0 && filterWidth/c>0) {
				*scaleFactor = c;
				break;
			}
		}
	}
#elif 0 //simpler implementation
	while (filterHeight/(*scaleFactor) == 0 || filterWidth/(*scaleFactor) == 0) {
		--(*scaleFactor);
	}
#elif 0 //even simpler implementation
	while (*scaleFactor > filterHeight || *scaleFactor > filterWidth) {
		--(*scaleFactor);
	}
#elif 1 //possibly the simplest implementation??
	*scaleFactor = min(*scaleFactor, min(filterHeight, filterWidth));
#endif

    //copyRect (proxyRect, filterRect); 
    //scaleRect (proxyRect, 1, *scaleFactor);

    }

/****************************************************************************/

